// Background task handling.
// Copyright © 2020 Core Technologies.
// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// Any editor function that wants to be called back to perform
// periodic work should use the functions here to link itself
// a callback that will be called during background processing.

#include	"includes.h"

typedef struct backgroundHandler
{
	BACKGROUND_FUNCTION
		*function;				// the background function to call
	void
		*userData;				// data passed to the function when it is called
	struct backgroundHandler
		*prev,					// links the list of background handlers together
		*next;
} BACKGROUND_HANDLER;

typedef struct backgroundHandlerList
{
	BACKGROUND_HANDLER
		*first,
		*last;
} BACKGROUND_HANDLER_LIST;

static BACKGROUND_HANDLER_LIST
	backgroundHandlerList;

static void BackgroundHandlerListUnlink(BACKGROUND_HANDLER *handler)
// Remove handler from the linked list of handlers.
{
	if(handler->prev)
	{
		handler->prev->next=handler->next;
	}
	else
	{
		backgroundHandlerList.first=handler->next;
	}
	if(handler->next)
	{
		handler->next->prev=handler->prev;
	}
	else
	{
		backgroundHandlerList.last=handler->prev;
	}
}

static void BackgroundHandlerListLinkToEnd(BACKGROUND_HANDLER *handler)
// Link handler to the end of the linked list of handlers.
{
	handler->next=NULL;
	if((handler->prev=backgroundHandlerList.last))
	{
		handler->prev->next=handler;			// point previous last item to this one
	}
	else										// backgroundHandlerList was empty, so this is first item too
	{
		backgroundHandlerList.first=handler;
	}
	backgroundHandlerList.last=handler;
}

void BackgroundDestroy(BACKGROUND_HANDLER *handler)
// Disconnect ourselves from the given task.
// NOTE: this does not need to kill the task, and it is usually desirable
// NOT to kill the task.
// This otherwise removes all task data structures from the editor.
{
	BackgroundHandlerListUnlink(handler);
	MDisposePtr(handler);
}

BACKGROUND_HANDLER *BackgroundCreate(BACKGROUND_FUNCTION *function,void *userData)
// Create a background handler and link it to the global list.
// The handler function will be called periodically until it is removed
// from the list by calling BackgroundDestroy.
// If this succeeds, it will return a pointer to the handler. If it
// fails, it will call ErrorSet, and return NULL.
{
	BACKGROUND_HANDLER
		*handler;

	if((handler=(BACKGROUND_HANDLER *)MNewPtr(sizeof(BACKGROUND_HANDLER))))
	{
		handler->function=function;
		handler->userData=userData;
		BackgroundHandlerListLinkToEnd(handler);
	}
	return(handler);
}

void BackgroundUpdate()
// Call any functions which want to perform work in the background.
// The handler that's called is allowed to remove itself from the list,
// or add any other handlers it would like.
// However it cannot remove any other handler.
{
	BACKGROUND_HANDLER
		*handler,
		*nextHandler;

	handler=backgroundHandlerList.first;
	while(handler)
	{
		nextHandler=handler->next;									// get pointer to next handler in case this one removes itself
		handler->function(handler->userData);
		handler=nextHandler;
	}
}

void BackgroundUnInit()
// Undo what BackgroundInit did.
// This will remove any handlers that remain on the list (none should).
{
	while(backgroundHandlerList.first)
	{
		BackgroundDestroy(backgroundHandlerList.first);
	}
}

bool BackgroundInit()
// Initialize background task handling.
{
	backgroundHandlerList.first=backgroundHandlerList.last=NULL;	// no background handlers yet
	return(true);
}
