// Chunk array management functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".
//
// NOTE: To manage the array data efficiently, and to keep memory
// from fragmenting, variable arrays of structures are kept in "Chunks".
// Each chunk allocates space from the free memory pool
// and the elements of the array are split among them. The chunks are linked
// to each other in order. When an element is inserted, additional chunks are allocated
// as necessary, and linked into the list at the appropriate location.
// Chunks keep track of how many elements are currently used within them.
// Chunks are not allowed to contain 0 elements.

#include	"includes.h"

typedef struct cArrayCacheEntry
{
	struct cArrayCacheEntry
		*next,						// used to link the cache entries (newest to oldest)
		*previous;
	CARRAY_CHUNK
		*chunk;						// the chunk whose "position" we know (NULL if this cache entry is invalid)
	uint64_t
		paramTotals[0];				// the total of each parameter for all chunks leading up to the cached chunk (variable size)
} CARRAY_CACHE_ENTRY;

// ----------------------------------------------------------------------------

void CArraySanity(CARRAY *cArray)
// Check the sanity of cArray.
{
	uint64_t
		paramTotals[cArray->numParams];
	CARRAY_CHUNK
		*chunk;
	CARRAY_CACHE_ENTRY
		*cacheEntry;
	uint32_t
		i;

	for(i=0;i<cArray->numParams;i++)
	{
		paramTotals[i]=0;
	}

	chunk=cArray->firstChunk;
	while(chunk)
	{
		if(chunk->cacheEntry)
		{
			if(chunk->cacheEntry->chunk==chunk)
			{
				for(i=0;i<cArray->numParams;i++)
				{
					if(chunk->cacheEntry->paramTotals[i]!=paramTotals[i])
					{
						fprintf(stderr,"Non-matching cache(%p) parameter %d: %" PRIu64 "!=%" PRIu64 "\n",chunk->cacheEntry,(int)i,chunk->cacheEntry->paramTotals[i],paramTotals[i]);
						abort();
					}
				}
			}
			else
			{
				fprintf(stderr,"Non-matching cache(%p) entry\n",chunk->cacheEntry);
				abort();
			}
		}
		for(i=0;i<cArray->numParams;i++)
		{
			paramTotals[i]+=chunk->params[i];
		}

		if(chunk->nextChunk==NULL)	// see if at the end
		{
			if(cArray->lastChunk!=chunk)
			{
				fprintf(stderr,"last chunk mismatch\n");
				abort();
			}
		}
		chunk=chunk->nextChunk;
	}
	cacheEntry=cArray->cache.first;
	while(cacheEntry)
	{
		if(cacheEntry->chunk&&(cacheEntry->chunk->cacheEntry!=cacheEntry))
		{
			fprintf(stderr,"cache(%p) entry mismatch\n",cacheEntry);
			abort();
		}
		cacheEntry=cacheEntry->next;
	}
	for(i=0;i<cArray->numParams;i++)
	{
		if(cArray->paramTotals[i]!=paramTotals[i])
		{
			fprintf(stderr,"Non-matching array paramTotal %d: %" PRIu64 "!=%" PRIu64 "\n",(int)i,cArray->paramTotals[i],paramTotals[i]);
			abort();
		}
	}
}

void CArrayCacheDump(CARRAY *cArray)
// Dump cache information.
{
	CARRAY_CACHE_ENTRY
		*cacheEntry;
	uint32_t
		i;

	cacheEntry=cArray->cache.first;
	while(cacheEntry)
	{
		if(cacheEntry->chunk)
		{
			fprintf(stderr,"cache(%p)\n",cacheEntry);
			for(i=0;i<cArray->numParams;i++)
			{
				fprintf(stderr,"    %" PRIu32 ": %" PRIu64 "\n",i,cacheEntry->paramTotals[i]);
			}
		}
		else
		{
			fprintf(stderr,"cache(%p) EMPTY\n",cacheEntry);
		}
		cacheEntry=cacheEntry->next;
	}
}

uint64_t CArrayTotalChunksGet(CARRAY *cArray)
// Count up the total number of chunks in the given array, and return it.
{
	uint64_t
		result;
	CARRAY_CHUNK
		*chunk;

	result=0;
	chunk=cArray->firstChunk;
	while(chunk)
	{
		result++;
		chunk=chunk->nextChunk;
	}
	return(result);
}

// ----------------------------------------------------------------------------
// Cache management

static uint64_t CacheEntryParameterDistance(CARRAY_CACHE_ENTRY *cacheEntry,uint32_t parameterIndex,uint64_t value)
// Return the absolute distance of the given parameter of the given
// cacheEntry (which must be valid) to the given value.
{
	if(cacheEntry->paramTotals[parameterIndex]<=value)
	{
		return(value-cacheEntry->paramTotals[parameterIndex]);
	}
	return(cacheEntry->paramTotals[parameterIndex]-value);
}

static CARRAY_CACHE_ENTRY *CArrayCacheLocateBest(CARRAY_CACHE *cache,uint32_t parameterIndex,uint64_t value,uint64_t *bestDistance)
// Find the cache entry with a given parameter closest to value.
// If there are no valid cache entries, return NULL.
{
	uint64_t
		distance;
	CARRAY_CACHE_ENTRY
		*cacheEntry,
		*bestEntry;

	bestEntry=NULL;
	(*bestDistance)=0;				// keep compiler quiet
	cacheEntry=cache->first;
	while(cacheEntry&&cacheEntry->chunk)
	{
		distance=CacheEntryParameterDistance(cacheEntry,parameterIndex,value);
		if((!bestEntry)||(distance<(*bestDistance)))
		{
			bestEntry=cacheEntry;
			(*bestDistance)=distance;
		}
		cacheEntry=cacheEntry->next;
	}
	return(bestEntry);
}

static void CacheEntryUnlink(CARRAY_CACHE *cache,CARRAY_CACHE_ENTRY *cacheEntry)
// Unlink the given cacheEntry from the cache.
{
	if(cacheEntry->previous)
	{
		cacheEntry->previous->next=cacheEntry->next;
	}
	else
	{
		cache->first=cacheEntry->next;
	}
	if(cacheEntry->next)
	{
		cacheEntry->next->previous=cacheEntry->previous;
	}
	else
	{
		cache->last=cacheEntry->previous;
	}
}

static void CacheEntryLinkToEnd(CARRAY_CACHE *cache,CARRAY_CACHE_ENTRY *cacheEntry)
// Link the given cacheEntry to the end of the list given in cache.
{
	cacheEntry->next=NULL;
	if((cacheEntry->previous=cache->last))
	{
		cache->last->next=cacheEntry;
	}
	else
	{
		cache->first=cacheEntry;
	}
	cache->last=cacheEntry;
}

static void CacheEntryLinkToStart(CARRAY_CACHE *cache,CARRAY_CACHE_ENTRY *cacheEntry)
// Link the given cacheEntry to the start of the list given in cache.
{
	cacheEntry->previous=NULL;
	if((cacheEntry->next=cache->first))
	{
		cache->first->previous=cacheEntry;
	}
	else
	{
		cache->last=cacheEntry;
	}
	cache->first=cacheEntry;
}

static void CacheEntryInvalidate(CARRAY_CACHE *cache,CARRAY_CACHE_ENTRY *cacheEntry)
// Mark a chunk entry which is currently valid as invalid, and relink it to the end of the list.
{
	cacheEntry->chunk->cacheEntry=NULL;			// mark the old cached chunk as not being cached
	cacheEntry->chunk=NULL;						// mark the entry as having no chunk
	CacheEntryUnlink(cache,cacheEntry);			// pull off the list
	CacheEntryLinkToEnd(cache,cacheEntry);		// add to the end
}

static void CacheEntryRefresh(CARRAY_CACHE *cache,CARRAY_CACHE_ENTRY *cacheEntry)
// Move the given cacheEntry to the start of the list.
{
	CacheEntryUnlink(cache,cacheEntry);			// pull off the list
	CacheEntryLinkToStart(cache,cacheEntry);	// add to the start
}

static void CArrayCacheEntryAdd(CARRAY *cArray,CARRAY_CHUNK *chunk,uint64_t *paramTotals)
// Add chunk/paramTotals to the cache. If there are no invalid entries,
// just replace the oldest.
// The added entry becomes the newest.
// This will avoid adding the entry if its position is identical to
// an existing one.
{
	CARRAY_CACHE_ENTRY
		*newEntry;
	uint32_t
		i;

	if(!chunk->cacheEntry)							// make sure the chunk we're caching is not already cached)
	{
		// replace the oldest entry
		if((newEntry=cArray->cache.last))			// make sure there's an entry to replace (it is allowed for a CARRAY to have no cache entries)
		{
			if(newEntry->chunk)						// was the entry in use?
			{
				newEntry->chunk->cacheEntry=NULL;	// mark the old chunk as no longer cached
			}
			newEntry->chunk=chunk;					// point at the new chunk
			newEntry->chunk->cacheEntry=newEntry;	// point the new chunk at this cache entry
			for(i=0;i<cArray->numParams;i++)
			{
				newEntry->paramTotals[i]=paramTotals[i];	// copy over the parameters
			}
			CacheEntryUnlink(&cArray->cache,newEntry);		// unlink from the end
			CacheEntryLinkToStart(&cArray->cache,newEntry);	// link to the start
		}
	}
}

static void CArrayCacheEntryDispose(CARRAY_CACHE_ENTRY *cacheEntry)
// Dispose of a cache entry created by CArrayCacheEntryCreate.
{
	MDisposePtr(cacheEntry);
}

static CARRAY_CACHE_ENTRY *CArrayCacheEntryCreate(uint32_t numParams)
// Create a cache entry, initialize it to empty, and return it.
// It there's a problem, ErrorSet and return NULL.
{
	CARRAY_CACHE_ENTRY
		*cacheEntry;

	if((cacheEntry=(CARRAY_CACHE_ENTRY *)MNewPtr(sizeof(CARRAY_CACHE_ENTRY)+numParams*sizeof(cacheEntry->paramTotals[0]))))
	{
		cacheEntry->next=NULL;
		cacheEntry->previous=NULL;
		cacheEntry->chunk=NULL;
	}
	return(cacheEntry);
}

void CArrayCacheEntriesUpdate(CARRAY *cArray,uint64_t updatePosition,int64_t *paramChanges)
// After the chunk array has been modified, this is called to
// update the cache.
// updatePosition is passed as the position of the modification.
// All chunks in the cache with starting positions
// which are at or after the updatePosition will have their parameter
// totals updated by paramChanges to bring them into line with
// the changes that occurred to the array.
{
	CARRAY_CACHE_ENTRY
		*cacheEntry;
	uint32_t
		i;

	cacheEntry=cArray->cache.first;
	while(cacheEntry&&cacheEntry->chunk)
	{
		if(cacheEntry->paramTotals[CA_ELEMENTS]>updatePosition)				// was cache entry after the update?
		{
			for(i=0;i<cArray->numParams;i++)
			{
				cacheEntry->paramTotals[i]+=paramChanges[i];				// adjust the totals
			}
		}
		cacheEntry=cacheEntry->next;
	}
}

static void CArrayCacheEntriesDispose(CARRAY *cArray)
// Dispose of the cache entries hanging off cArray
{
	CARRAY_CACHE_ENTRY
		*cacheEntry,
		*nextCacheEntry;

	cacheEntry=cArray->cache.first;
	while(cacheEntry)
	{
		nextCacheEntry=cacheEntry->next;
		CArrayCacheEntryDispose(cacheEntry);
		cacheEntry=nextCacheEntry;
	}
}

static bool CArrayCacheEntriesCreate(CARRAY *cArray,uint32_t numCacheEntries,uint32_t numParams)
// Create numCacheEntries cache entries, and link them to cArray.
// If there's a problem, ErrorSet and return false.
{
	bool
		fail;
	CARRAY_CACHE_ENTRY
		*newEntry;
	uint32_t
		i;

	cArray->cache.first=NULL;
	cArray->cache.last=NULL;
	fail=false;
	for(i=0;!fail&&(i<numCacheEntries);i++)
	{
		if((newEntry=CArrayCacheEntryCreate(numParams)))
		{
			CacheEntryLinkToEnd(&cArray->cache,newEntry);
		}
	}
	if(fail)
	{
		CArrayCacheEntriesDispose(cArray);
	}
	return(!fail);
}

// ----------------------------------------------------------------------------

static void CArrayChunkDispose(CARRAY_CHUNK *chunk)
// Function to dispose of CARRAY chunks created by CArrayChunkCreate.
{
	MDisposePtr(chunk);
}

static CARRAY_CHUNK *CArrayChunkCreate(uint32_t maxElements,uint32_t elementSize,uint32_t numParams)
// Function to create and initialize a chunk.
// If there's a problem, ErrorSet, and return NULL.
{
	uint8_t
		*buffer;
	CARRAY_CHUNK
		*chunk;
	uint32_t
		chunkHeaderSize,
		paramSize,
		dataSize;
	uint32_t
		i;

	chunkHeaderSize=((sizeof(CARRAY_CHUNK)+31)/32)*32;	// round up header size so params are placed on proper boundary (along with data)
	paramSize=numParams*sizeof(*chunk->params);
	dataSize=maxElements*elementSize;

	if((buffer=(uint8_t *)MNewPtr(chunkHeaderSize+paramSize+dataSize)))	// create memory for header, params, and data
	{
		chunk=(CARRAY_CHUNK *)buffer;

		chunk->nextChunk=NULL;
		chunk->previousChunk=NULL;
		chunk->cacheEntry=NULL;									// the new chunk is not in the cache

		chunk->params=(uint64_t *)&buffer[chunkHeaderSize];		// point to params
		chunk->data=&buffer[chunkHeaderSize+paramSize];			// point to element data

		for(i=0;i<numParams;i++)
		{
			chunk->params[i]=0;									// clear parameters for this chunk
		}

		return(chunk);
	}
	return(NULL);
}

// ----------------------------------------------------------------------------

static void CArrayChunkListDispose(CARRAY_CHUNK *startChunk)
// Dispose of a linked list of chunks.
{
	CARRAY_CHUNK
		*nextChunk;

	while(startChunk)
	{
		nextChunk=startChunk->nextChunk;
		CArrayChunkDispose(startChunk);
		startChunk=nextChunk;
	}
}

static bool CArrayChunkListCreate(CARRAY *cArray,CARRAY_CHUNK **startChunk,CARRAY_CHUNK **endChunk,uint64_t numChunks)
// Create a linked list of empty chunks.
// If there is a problem, set the error, free anything allocated, and return false.
{
	bool
		fail;
	CARRAY_CHUNK
		*chunk;

	(*startChunk)=(*endChunk)=NULL;
	fail=false;
	while(numChunks&&!fail)
	{
		if((chunk=CArrayChunkCreate(cArray->maxElementsPerChunk,cArray->elementSize,cArray->numParams)))
		{
			if(*endChunk)
			{
				(*endChunk)->nextChunk=chunk;
				chunk->previousChunk=(*endChunk);
				chunk->nextChunk=NULL;
				(*endChunk)=chunk;
			}
			else
			{
				(*startChunk)=(*endChunk)=chunk;
				chunk->previousChunk=NULL;
				chunk->nextChunk=NULL;
			}
			numChunks--;
		}
		else
		{
			CArrayChunkListDispose(*startChunk);				// remove the partially created chunk list
			fail=true;
		}
	}
	return(!fail);
}

/*
static void CArrayChunkListUnlink(CARRAY *cArray,CARRAY_CHUNK *startChunk,CARRAY_CHUNK *endChunk)
// Unlink the chunks from startChunk to endChunk from cArray.
{
	if(startChunk->previousChunk)
	{
		startChunk->previousChunk->nextChunk=endChunk->nextChunk;
	}
	else
	{
		cArray->firstChunk=endChunk->nextChunk;
	}

	if(endChunk->nextChunk)
	{
		endChunk->nextChunk->previousChunk=startChunk->previousChunk;
	}
	else
	{
		cArray->lastChunk=startChunk->previousChunk;
	}
	startChunk->previousChunk=endChunk->nextChunk=NULL;			// terminate the list
}
*/

// ----------------------------------------------------------------------------
// Searching

static void CArrayFinalParameterLocate(CARRAY *cArray,uint32_t parameterIndex,uint64_t value,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Locate the chunk in the given cArray which contains the value
// given by parameterIndex/value.
// outChunk/outParams enter this function as the place to start the search.
// outChunk will be updated to point to the located chunk, and outParams
// with the total of parameters that precede it.
// This may search forward or backward from the passed starting location.
// NOTE: if parameter is off the end of the list, then outChunk will
// return as NULL, and outParams will be the sum of all chunks in the list.
{
	uint32_t
		i;

	// move left as far as needed
	while(value<outParams[parameterIndex])						// need to move left from here?
	{
		(*outChunk)=(*outChunk)->previousChunk;					// move left
		for(i=0;i<cArray->numParams;i++)						// step back from a parameters perspective
		{
			outParams[i]-=(*outChunk)->params[i];
		}
	}

	// move right as far as needed (including off the end -- which shouldn't really happen because of the way this is called)
	while((*outChunk)&&(value>=outParams[parameterIndex]+(*outChunk)->params[parameterIndex]))	// need to move right from here?
	{
		for(i=0;i<cArray->numParams;i++)						// step forward from a parameters perspective
		{
			outParams[i]+=(*outChunk)->params[i];
		}
		(*outChunk)=(*outChunk)->nextChunk;						// move right
	}
}

static void CArrayStartingParameterLocate(CARRAY *cArray,uint32_t parameterIndex,uint64_t value,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Find the best starting chunk for a search based on the cache, and
// the starts and ends of the cArray itself. Fill in outChunk, and
// outParams with the best candidate.
{
	uint32_t
		i;
	CARRAY_CACHE_ENTRY
		*bestCacheEntry;
	uint64_t
		bestDistance;

	if(value<cArray->paramTotals[parameterIndex])				// make sure value is not off the end of the cArray
	{
		if((bestCacheEntry=CArrayCacheLocateBest(&cArray->cache,parameterIndex,value,&bestDistance))&&((bestDistance<value)&&(bestDistance<cArray->paramTotals[parameterIndex]-value)))
		{
			(*outChunk)=bestCacheEntry->chunk;					// copy the chunk
			for(i=0;i<cArray->numParams;i++)
			{
				outParams[i]=bestCacheEntry->paramTotals[i];	// copy the totals leading to the chunk
			}
			CacheEntryRefresh(&cArray->cache,bestCacheEntry);	// refresh this since we just used it
		}
		else	// no cache entries, or closer to the start or end of the cArray than to the cache
		{
			// Under some circumstances (deleting from the beginning of the array),
			// some parameters of the array can become unbalanced (think sparse array
			// with no sparse elements at the start, and a ton at the end). To deal
			// with those cases, the code here decides which end to start from
			// by splitting the difference of the totals in the array, and the totals
			// after the first chunk. At this point in the code, we're guaranteed
			// that there's at least one chunk in the array.
			if(value<=(cArray->firstChunk->params[parameterIndex]+(cArray->paramTotals[parameterIndex]-cArray->firstChunk->params[parameterIndex])/2))	// closer to start?
			{
				(*outChunk)=cArray->firstChunk;
				for(i=0;i<cArray->numParams;i++)
				{
					outParams[i]=0;
				}
			}
			else
			{
				(*outChunk)=cArray->lastChunk;					// point at the last chunk to start the search
				for(i=0;i<cArray->numParams;i++)
				{
					outParams[i]=cArray->paramTotals[i]-cArray->lastChunk->params[i];
				}
			}
		}
	}
	else	// off the end, just return it
	{
		(*outChunk)=NULL;
		for(i=0;i<cArray->numParams;i++)
		{
			outParams[i]=cArray->paramTotals[i];
		}
	}
}

static void CArrayParameterLocate(CARRAY *cArray,uint32_t parameterIndex,uint64_t value,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Locate the chunk in the cArray which contains the parameter value.
// Return outChunk, with outParams as the total of the parameters leading up to the chunk.
// NOTE: this only finds the start of the chunk. Locating the parameter within
// the chunk isn't done here.
{
	CArrayStartingParameterLocate(cArray,parameterIndex,value,outChunk,outParams);	// find the best place to start the search
	if(*outChunk)																	// make sure we could find a starting point (if not, the value is off the end of the array)
	{
		CArrayFinalParameterLocate(cArray,parameterIndex,value,outChunk,outParams);	// go hunt down the chunk given the starting point
	}
}

// ----------------------------------------------------------------------------

void CArrayChunkOffsetExtract(CARRAY_CHUNK **sourceChunk,uint64_t *sourceOffset,CARRAY *destCArray,void *dest,uint64_t destOffset,uint64_t numElements,uint64_t *paramTotals)
// Extract numElements from sourceChunk/sourceOffset into the buffer
// at dest/destOffset. If sourceChunk is passed as NULL (or becomes NULL
// while being traversed), zeroes will be inserted into the destination.
// If paramTotals is non-NULL, fill it with the totals of data moved.
// sourceChunk and sourceOffset will be pushed forward by this function.
{
	uint64_t
		numToMove;
	uint64_t
		localParamTotals[destCArray->numParams];
	uint32_t
		i;

	if(paramTotals)
	{
		for(i=0;i<destCArray->numParams;i++)
		{
			paramTotals[i]=0;
		}
	}
	while(numElements)
	{
		if(*sourceChunk)
		{
			numToMove=(*sourceChunk)->params[CA_ELEMENTS]-(*sourceOffset);	// work out how many elements to transfer this time
			if(numToMove>numElements)										// pin to the number remaining to transfer
			{
				numToMove=numElements;
			}
		}
		else
		{
			numToMove=numElements;
		}
		destCArray->elementsMoveFunction(destCArray,(*sourceChunk)?(*sourceChunk)->data:NULL,*sourceOffset,dest,destOffset,numToMove,paramTotals?localParamTotals:NULL);
		if(paramTotals)
		{
			for(i=0;i<destCArray->numParams;i++)						// bump the totals
			{
				paramTotals[i]+=localParamTotals[i];
			}
		}
		if(*sourceChunk)
		{
			(*sourceOffset)+=numToMove;
			if((*sourceOffset)>=(*sourceChunk)->params[CA_ELEMENTS])
			{
				(*sourceChunk)=(*sourceChunk)->nextChunk;
				(*sourceOffset)=0;
			}
		}
		destOffset+=numToMove;
		numElements-=numToMove;
	}
}

// ----------------------------------------------------------------------------

void CArrayExtract(CARRAY *cArray,uint64_t index,void *dest,uint64_t destOffset,uint64_t numElements,uint64_t *paramTotals)
// Extract numElements from cArray at index into the buffer
// at dest/destOffset.
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		offset;
	uint64_t
		localParamTotals[cArray->numParams];

	CArrayParameterLocate(cArray,CA_ELEMENTS,index,&chunk,localParamTotals);
	offset=index-localParamTotals[CA_ELEMENTS];
	CArrayChunkOffsetExtract(&chunk,&offset,cArray,dest,destOffset,numElements,paramTotals);
}

// ----------------------------------------------------------------------------
// Deletion cases

static void CArrayChunkDelete(CARRAY *cArray,CARRAY_CHUNK *chunk,int64_t *paramChanges)
// Delete an entire chunk from destCArray, update paramChanges.
{
	uint32_t
		i;

	// keep track of the total changes
	for(i=0;i<cArray->numParams;i++)
	{
		paramChanges[i]-=chunk->params[i];
	}

	// if chunk is in the cache, remove it
	if(chunk->cacheEntry)
	{
		CacheEntryInvalidate(&cArray->cache,chunk->cacheEntry);
	}

	// unlink the chunk from the cArray
	if(chunk->previousChunk)
	{
		chunk->previousChunk->nextChunk=chunk->nextChunk;
	}
	else
	{
		cArray->firstChunk=chunk->nextChunk;
	}
	if(chunk->nextChunk)
	{
		chunk->nextChunk->previousChunk=chunk->previousChunk;
	}
	else
	{
		cArray->lastChunk=chunk->previousChunk;
	}

	// let go of the memory the chunk occupied
	CArrayChunkDispose(chunk);
}

static void CArrayChunkElementsDelete(CARRAY *cArray,CARRAY_CHUNK *chunk,uint64_t offset,uint64_t numElements,int64_t *paramChanges)
// numElements within a single chunk are being deleted starting at offset.
// Remove the elements, and update paramChanges.
{
	uint64_t
		paramTotals[cArray->numParams];
	uint32_t
		i;

	cArray->elementsMoveFunction(cArray,chunk->data,offset,NULL,0,numElements,paramTotals);	// total up the params being deleted
	cArray->elementsMoveFunction(cArray,chunk->data,offset+numElements,chunk->data,offset,chunk->params[CA_ELEMENTS]-(offset+numElements),NULL);	// move back any elements at the end

	for(i=0;i<cArray->numParams;i++)
	{
		chunk->params[i]-=paramTotals[i];	// remove the deleted parameters from the chunk itself
		paramChanges[i]-=paramTotals[i];	// keep track of the overall amount of deleted parameters
	}
}

static void CArrayChunkOffsetDelete(CARRAY *cArray,CARRAY_CHUNK *chunk,uint64_t offset,uint64_t numElements,int64_t *paramChanges)
// Delete numElements starting from chunk/offset.
// numElements must be non-zero, chunk must be non-NULL, and offset must be valid within the chunk.
// This only deletes from the array -- it does not update the array
// itself, or the array's cache (other than to delete items from it if they
// got deleted here).
// paramChanges will be updated to report the total params changes (negative on delete).
// If there is a problem, ErrorSet, and return false.
{
	uint64_t
		numToDelete;
	CARRAY_CHUNK
		*nextChunk;

	if(offset)			// not deleting from the start of the chunk?
	{
		// NOTE: if this chunk is in the cache, it can stay there unmodified since the totals leading up to it are unmodified
		numToDelete=chunk->params[CA_ELEMENTS]-offset;									// assume deleting to end of the chunk
		if(numToDelete>numElements)														// pin to the requested number of elements
		{
			numToDelete=numElements;
		}
		CArrayChunkElementsDelete(cArray,chunk,offset,numToDelete,paramChanges);		// delete elements from the chunk

		numElements-=numToDelete;														// this many fewer elements to delete
		if(offset>=chunk->params[CA_ELEMENTS])											// need to push to next chunk?
		{
			chunk=chunk->nextChunk;
			offset=0;
		}
	}

	// deleting from the starts of chunks (offset is 0)
	while(numElements&&chunk)
	{
		numToDelete=chunk->params[CA_ELEMENTS];											// assume deleting to end of chunk
		if(numToDelete>numElements)
		{
			numToDelete=numElements;													// pin as needed
		}

		if(numToDelete>=chunk->params[CA_ELEMENTS])										// deleting the whole thing?
		{
			nextChunk=chunk->nextChunk;
			CArrayChunkDelete(cArray,chunk,paramChanges);								// go delete it (also deletes from the cache)
			chunk=nextChunk;
		}
		else
		{
			CArrayChunkElementsDelete(cArray,chunk,offset,numToDelete,paramChanges);	// just delete elements from the chunk (offset still at 0)
			if(chunk->cacheEntry)														// remove it from the cache if it was there, since its cached params may have been messed up by the preceding partial delete
			{
				CacheEntryInvalidate(&cArray->cache,chunk->cacheEntry);
			}
		}
		numElements-=numToDelete;
	}
}

void CArrayElementsDelete(CARRAY *cArray,uint64_t index,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Delete numElements from cArray starting at index.
// If non-NULL, return outChunk and outParams as if CArrayParameterLocate was called
// for the element just after the delete.
// NOTE: this will add an entry to the cache at the end of the delete.
{
	CARRAY_CHUNK
		*chunk,
		*previousChunk;
	uint64_t
		paramTotals[cArray->numParams];
	int64_t
		paramChanges[cArray->numParams];
	uint32_t
		i;

	CArrayParameterLocate(cArray,CA_ELEMENTS,index,&chunk,paramTotals);							// go hunt down the chunk given the starting point
	if(numElements&&chunk)																		// make sure request is for at least some elements, and we're not off the end of the world
	{
		previousChunk=chunk->previousChunk;														// keep pointer to previous chunk which we know will be untouched by the deletion (use it to update the cache at the end)

		// initialize the param changes array (it gets added to incrementally)
		for(i=0;i<cArray->numParams;i++)
		{
			paramChanges[i]=0;
		}

		CArrayChunkOffsetDelete(cArray,chunk,index-paramTotals[CA_ELEMENTS],numElements,paramChanges);	// go perform the actual deletion

		for(i=0;i<cArray->numParams;i++)														// update the array header to accurately reflect the param totals now
		{
			cArray->paramTotals[i]+=paramChanges[i];
		}

		CArrayCacheEntriesUpdate(cArray,index,paramChanges);									// adjust the cache as needed

		if(previousChunk&&previousChunk->nextChunk)												// now, if possible, place a cache entry at the point of the delete
		{
			CArrayCacheEntryAdd(cArray,previousChunk->nextChunk,paramTotals);					// add the delete point to the cache (paramTotals is still correct for the next chunk, no matter what happened during the delete)
		}

		// finally, set up the return chunk if needed
		if(outChunk||outParams)
		{
			if(previousChunk)
			{
				chunk=previousChunk->nextChunk;
			}
			else
			{
				chunk=cArray->firstChunk;
			}
			while(chunk&&(index>=paramTotals[CA_ELEMENTS]+chunk->params[CA_ELEMENTS]))	// see if we need to step to the next chunk (rare)
			{
				for(i=0;i<cArray->numParams;i++)
				{
					paramTotals[i]+=chunk->params[i];
				}
				chunk=chunk->nextChunk;
			}
		}
	}

	if(outChunk)
	{
		*outChunk=chunk;
	}
	if(outParams)
	{
		for(i=0;i<cArray->numParams;i++)
		{
			outParams[i]=paramTotals[i];
		}
	}
}

// ----------------------------------------------------------------------------
// Insertion cases
// These routines handle all of the possible cases that can occur while
// inserting chunks.

static void CArrayInsertWithRoom(CARRAY_CHUNK *sourceChunk,uint64_t sourceOffset,CARRAY *destCArray,CARRAY_CHUNK *destChunk,uint64_t destOffset,uint64_t numElements,int64_t *paramChanges)
// The chunk that is being inserted into contains enough free space to accept
// the entire insertion.
// NOTE: All the insert functions expect chunk offset to point to the next
// character to insert. This means the offset can enter and leave
// these functions pointed to the end of the data in the current chunk
// (as opposed to pushing to the start of the next chunk like all other
// functions do).
{
	uint64_t
		paramTotals[destCArray->numParams];
	uint32_t
		i;

	destCArray->elementsMoveFunction(destCArray,destChunk->data,destOffset,destChunk->data,destOffset+numElements,destChunk->params[CA_ELEMENTS]-destOffset,NULL);	// move data within the chunk to make room
	CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,destChunk->data,destOffset,numElements,paramTotals);	// extract elements into the chunk, count up params moved

	for(i=0;i<destCArray->numParams;i++)
	{
		destChunk->params[i]+=paramTotals[i];				// bump the totals for the chunk
		paramChanges[i]+=paramTotals[i];					// bump the total changes to be returned
	}
}

static void CArrayInsertWithRoomInTwo(CARRAY_CHUNK *sourceChunk,uint64_t sourceOffset,CARRAY *destCArray,CARRAY_CHUNK *destChunk,uint64_t destOffset,uint64_t numElements,int64_t *paramChanges)
// The chunk where the insertion starts, and the next chunk contain enough free space
// to allow the entire insertion.
// NOTE: All the insert functions expect chunk offset to point to the next
// character to insert. This means the offset can enter and leave
// these functions pointed to the end of the data in the current chunk
// (as opposed to pushing to the start of the next chunk like all other
// functions do).
{
	CARRAY_CHUNK
		*firstChunk,
		*secondChunk;
	uint64_t
		amountToMove;
	uint64_t
		distanceToMove;
	uint64_t
		paramTotals[destCArray->numParams];
	uint32_t
		i;

	firstChunk=destChunk;
	secondChunk=destChunk->nextChunk;

	if(numElements+destOffset<=destCArray->maxElementsPerChunk)				// see if the added elements can be contained completely within the first chunk
	{
		distanceToMove=numElements-(destCArray->maxElementsPerChunk-firstChunk->params[CA_ELEMENTS]);	// for speed, we would like to move as much as possible to the second chunk, but for memory efficiency, we would like to move as little as possible! (memory efficiency turns out to be better)

		destCArray->elementsMoveFunction(destCArray,secondChunk->data,0,secondChunk->data,distanceToMove,secondChunk->params[CA_ELEMENTS],NULL);	// move second chunk's data out of the way for the add

		destCArray->elementsMoveFunction(destCArray,firstChunk->data,firstChunk->params[CA_ELEMENTS]-distanceToMove,secondChunk->data,0,distanceToMove,paramTotals);	// move data from first chunk to second chunk, get totals moved into second chunk

		for(i=0;i<destCArray->numParams;i++)
		{
			firstChunk->params[i]-=paramTotals[i];							// remove the data from the first chunk
			secondChunk->params[i]+=paramTotals[i];							// add the data to the second chunk
		}

		destCArray->elementsMoveFunction(destCArray,firstChunk->data,destOffset,firstChunk->data,destOffset+numElements,firstChunk->params[CA_ELEMENTS]-destOffset,NULL);	// move data within the first chunk to make room

		CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,firstChunk->data,destOffset,numElements,paramTotals);	// extract elements into the chunk

		for(i=0;i<destCArray->numParams;i++)
		{
			firstChunk->params[i]+=paramTotals[i];							// bump data in the first chunk
			paramChanges[i]+=paramTotals[i];								// bump total change
		}
	}
	else
	{
		distanceToMove=numElements-(destCArray->maxElementsPerChunk-firstChunk->params[CA_ELEMENTS]);		// need to push second chunk data forward this distance

		destCArray->elementsMoveFunction(destCArray,secondChunk->data,0,secondChunk->data,distanceToMove,secondChunk->params[CA_ELEMENTS],NULL);	// move second chunk's data out of the way for the add

		amountToMove=firstChunk->params[CA_ELEMENTS]-destOffset;		// number of elements to move from first to second

		destCArray->elementsMoveFunction(destCArray,firstChunk->data,destOffset,secondChunk->data,destOffset+numElements-destCArray->maxElementsPerChunk,amountToMove,paramTotals);	// move data from first chunk to second chunk

		for(i=0;i<destCArray->numParams;i++)
		{
			firstChunk->params[i]-=paramTotals[i];							// removed this data from the first chunk
			secondChunk->params[i]+=paramTotals[i];							// added it to the second chunk
		}

		amountToMove=destCArray->maxElementsPerChunk-destOffset;			// number of elements to move into first

		CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,firstChunk->data,destOffset,amountToMove,paramTotals);	// extract elements into the chunk

		for(i=0;i<destCArray->numParams;i++)
		{
			firstChunk->params[i]+=paramTotals[i];							// bump data in the first chunk
			paramChanges[i]+=paramTotals[i];								// bump total change
		}

		amountToMove=numElements-(destCArray->maxElementsPerChunk-destOffset);	// number of elements to move into second

		CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,secondChunk->data,0,amountToMove,paramTotals);	// copy remaining elements into second chunk

		for(i=0;i<destCArray->numParams;i++)
		{
			secondChunk->params[i]+=paramTotals[i];							// bump data in the second chunk
			paramChanges[i]+=paramTotals[i];								// bump total change
		}
	}
}

static bool CArrayInsertWithNoRoom(CARRAY_CHUNK *sourceChunk,uint64_t sourceOffset,CARRAY *destCArray,CARRAY_CHUNK *destChunk,uint64_t destOffset,uint64_t numElements,int64_t *paramChanges)
// The elements to be inserted cannot fit within its two adjacent chunks.
// So, chunks will have to be created, and the elements put into them.
// This works by creating a list of chunks, and linking it to the end
// of the current destChunk, then moving the elements around to fill
// those chunks.
// NOTE: All the insert functions expect chunk offset to point to the next
// character to insert. This means the offset can enter and leave
// these functions pointed to the end of the data in the current chunk
// (as opposed to pushing to the start of the next chunk like all other
// functions do).
{
	CARRAY_CHUNK
		*startChunk,
		*endChunk;
	uint64_t
		numChunks,
		amountToMove,
		endOffset;
	uint64_t
		paramTotals[destCArray->numParams];
	uint32_t
		i;

	numChunks=(numElements-(destCArray->maxElementsPerChunk-(destChunk->params[CA_ELEMENTS]))+destCArray->maxElementsPerChunk-1)/destCArray->maxElementsPerChunk;
	if(CArrayChunkListCreate(destCArray,&startChunk,&endChunk,numChunks))	// create an empty list with just enough chunks to hold the data
	{
		// link the new list into the array of chunks
		if((endChunk->nextChunk=destChunk->nextChunk))
		{
			endChunk->nextChunk->previousChunk=endChunk;
		}
		else
		{
			destCArray->lastChunk=endChunk;
		}
		destChunk->nextChunk=startChunk;
		startChunk->previousChunk=destChunk;

		if(numElements<(destCArray->maxElementsPerChunk-destOffset))		// see if all of the elements we are adding can be added into the dest chunk (new list contains a single chunk)
		{
			amountToMove=destChunk->params[CA_ELEMENTS]+numElements-destCArray->maxElementsPerChunk;

			destCArray->elementsMoveFunction(destCArray,destChunk->data,destChunk->params[CA_ELEMENTS]-amountToMove,startChunk->data,0,amountToMove,paramTotals);	// move data from old last chunk to new last chunk
			destCArray->elementsMoveFunction(destCArray,destChunk->data,destOffset,destChunk->data,destOffset+numElements,destChunk->params[CA_ELEMENTS]-amountToMove-destOffset,NULL);	// move data within the original chunk to make room for new data

			for(i=0;i<destCArray->numParams;i++)
			{
				startChunk->params[i]=paramTotals[i];						// set parameters for new chunk
				destChunk->params[i]-=paramTotals[i];						// removed these from dest chunk
			}

			CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,destChunk->data,destOffset,numElements,paramTotals);	// extract elements into the chunk

			for(i=0;i<destCArray->numParams;i++)
			{
				destChunk->params[i]+=paramTotals[i];						// added to old chunk
				paramChanges[i]+=paramTotals[i];							// bump total change
			}
		}
		else
		{
			amountToMove=destChunk->params[CA_ELEMENTS]-destOffset;	// number of elements to move out of the old last chunk
			if((numElements+destOffset)/destCArray->maxElementsPerChunk<numChunks)	// special case to move the tail all the way into the end chunk, instead of letting it straddle the boundary
			{
				endOffset=0;
				destCArray->elementsMoveFunction(destCArray,destChunk->data,destOffset,endChunk->data,0,amountToMove,paramTotals);	// move data from old last chunk to new last chunk
			}
			else
			{
				endOffset=(numElements+destOffset)%destCArray->maxElementsPerChunk;
				destCArray->elementsMoveFunction(destCArray,destChunk->data,destOffset,endChunk->data,endOffset,amountToMove,paramTotals);	// move data from old last chunk to new last chunk
			}

			for(i=0;i<destCArray->numParams;i++)
			{
				endChunk->params[i]=paramTotals[i];							// set initial params for end chunk
				destChunk->params[i]-=paramTotals[i];						// removed params from here
			}

			amountToMove=destCArray->maxElementsPerChunk-destOffset;
			CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,destChunk->data,destOffset,amountToMove,paramTotals);	// extract elements into the chunk

			numElements-=amountToMove;										// this many fewer elements to move now

			for(i=0;i<destCArray->numParams;i++)
			{
				destChunk->params[i]+=paramTotals[i];						// added this data to the dest
				paramChanges[i]+=paramTotals[i];							// bump total change
			}

			// add the remaining data to the new list
			destChunk=startChunk;
			while(numElements)
			{
				amountToMove=destCArray->maxElementsPerChunk;				// assume we can move an entire chunk
				if(amountToMove>numElements)								// pin to what's remaining
				{
					amountToMove=numElements;
				}

				CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,destChunk->data,0,amountToMove,paramTotals);	// extract elements into the chunk

				for(i=0;i<destCArray->numParams;i++)
				{
					destChunk->params[i]+=paramTotals[i];					// added this data to the dest
					paramChanges[i]+=paramTotals[i];						// bump total change
				}

				numElements-=amountToMove;									// this many fewer elements to move now
				destChunk=destChunk->nextChunk;								// move to the next place to add
			}
		}
		return(true);														// done
	}
	return(false);
}

static bool CArrayInsertWithNoChunks(CARRAY_CHUNK *sourceChunk,uint64_t sourceOffset,CARRAY *destCArray,uint64_t numElements,int64_t *paramChanges)
// Insertion is happening in an empty array.
// Just create a chunk list, link it to destCArray, and fill it with the new elements.
// NOTE: All the insert functions expect chunk offset to point to the next
// element to insert. This means the offset can enter and leave
// these functions pointed to the end of the data in the current chunk
// (as opposed to pushing to the start of the next chunk like all other
// functions do).
{
	CARRAY_CHUNK
		*startChunk,
		*endChunk;
	uint64_t
		numChunks;
	uint64_t
		amountToMove;
	uint64_t
		paramTotals[destCArray->numParams];
	uint32_t
		i;

	numChunks=(numElements+destCArray->maxElementsPerChunk-1)/destCArray->maxElementsPerChunk;
	if(CArrayChunkListCreate(destCArray,&startChunk,&endChunk,numChunks))	// create an empty list with just enough chunks to hold the data
	{
		// link the new list into the array
		destCArray->firstChunk=startChunk;
		destCArray->lastChunk=endChunk;

		// add the data to the new list
		while(numElements)
		{
			amountToMove=destCArray->maxElementsPerChunk;					// assume we can move an entire chunk
			if(amountToMove>numElements)									// pin to what's remaining
			{
				amountToMove=numElements;
			}

			CArrayChunkOffsetExtract(&sourceChunk,&sourceOffset,destCArray,startChunk->data,0,amountToMove,paramTotals);	// extract elements into the chunk

			for(i=0;i<destCArray->numParams;i++)
			{
				startChunk->params[i]+=paramTotals[i];						// added this data to the dest
				paramChanges[i]+=paramTotals[i];							// bump total change
			}

			numElements-=amountToMove;										// this many fewer elements to move now
			startChunk=startChunk->nextChunk;								// move to the next place to add
		}
		return(true);														// done
	}
	return(false);
}

bool CArrayChunksInsert(CARRAY_CHUNK *sourceChunk,uint64_t sourceOffset,CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Insert numElements from sourceChunk/sourceOffset into destCArray at destIndex.
// If non-NULL, return outChunk/outParams as if CArrayParameterLocate
// was called on the first inserted element.
// If there is a problem, ErrorSet, and return false.
// NOTE: If sourceChunk/sourceOffset run out of elements (or sourceChunk is
// passed as NULL), zeroes will be inserted into the array at that point.
// NOTE: this will add an entry to the cache at the end of the insert.
// NOTE: to accommodate inserting unchunked data, sourceChunk is allowed to contain
// more than maxElementsPerChunk. Further, the only parameter field of
// sourceChunk which must exist and be valid is CA_ELEMENTS.
{
	bool
		fail;
	CARRAY_CHUNK
		*destChunk,
		*previousChunk,
		*nextChunk;
	uint64_t
		destParamTotals[destCArray->numParams],
		cacheParamTotals[destCArray->numParams];
	int64_t
		paramChanges[destCArray->numParams];
	uint64_t
		destOffset;
	uint32_t
		i;
	bool
		steppedBack;

	fail=false;
	CArrayParameterLocate(destCArray,CA_ELEMENTS,destIndex,&destChunk,destParamTotals);	// go hunt down the starting chunk
	if(numElements)														// only do more work if elements to insert
	{
		steppedBack=false;												// haven't stepped back to the previous chunk for an insert

		// initialize the param changes array (it gets added to incrementally)
		for(i=0;i<destCArray->numParams;i++)
		{
			paramChanges[i]=0;
		}

		if(!destChunk)													// if pointed to the end, back up to the last chunk
		{
			if((destChunk=destCArray->lastChunk))
			{
				destOffset=destChunk->params[CA_ELEMENTS];				// insert after any data
				for(i=0;i<destCArray->numParams;i++)					// step the totals back to the previous chunk
				{
					destParamTotals[i]-=destChunk->params[i];
				}
			}
		}
		else
		{
			destOffset=destIndex-destParamTotals[CA_ELEMENTS];			// start inserting here
		}
		if(destChunk)													// any chunks in the tree?
		{
			if((destOffset==0)&&((previousChunk=destChunk->previousChunk))&&(previousChunk->params[CA_ELEMENTS]<destCArray->maxElementsPerChunk))	// if inserting at the very start of a chunk, it is more efficient to look one back (if possible)
			{
				destChunk=previousChunk;								// step back to the previous chunk to start the insertion
				destOffset=destChunk->params[CA_ELEMENTS];				// start placing elements into the free space
				for(i=0;i<destCArray->numParams;i++)					// step the totals back to the previous chunk
				{
					destParamTotals[i]-=destChunk->params[i];
				}
				steppedBack=true;										// remember we stepped back so cache adjust works correctly below
			}

			if(numElements<=destCArray->maxElementsPerChunk-(destChunk->params[CA_ELEMENTS]))	// see if the elements can be placed in empty space of this chunk
			{
				CArrayCacheEntryAdd(destCArray,destChunk,destParamTotals);	// add a cache entry for this chunk
				CArrayInsertWithRoom(sourceChunk,sourceOffset,destCArray,destChunk,destOffset,numElements,paramChanges);	// drop elements in
			}
			else
			{
				if((nextChunk=destChunk->nextChunk))						// is there another chunk after this one?
				{
					if(numElements<=(2*destCArray->maxElementsPerChunk)-(destChunk->params[CA_ELEMENTS]+nextChunk->params[CA_ELEMENTS]))	// is there room in the combined free space of this chunk, and the next one?
					{
						if(nextChunk->cacheEntry)							// if the next chunk is in the cache, get rid of it, since we'll invalidate it with these changes
						{
							CacheEntryInvalidate(&destCArray->cache,nextChunk->cacheEntry);
						}

						if(nextChunk->nextChunk)							// add a cache entry for the next chunk's next (a chunk which won't be confused by the insert)
						{
							for(i=0;i<destCArray->numParams;i++)
							{
								cacheParamTotals[i]=destParamTotals[i]+destChunk->params[i]+nextChunk->params[i];
							}
							CArrayCacheEntryAdd(destCArray,nextChunk->nextChunk,cacheParamTotals);	// add a cache entry for this chunk
						}
						CArrayInsertWithRoomInTwo(sourceChunk,sourceOffset,destCArray,destChunk,destOffset,numElements,paramChanges);
					}
					else		// no room, we're going to need to add some chunks
					{
						// add a cache entry for the next chunk (which won't be confused by the insert)
						for(i=0;i<destCArray->numParams;i++)
						{
							cacheParamTotals[i]=destParamTotals[i]+destChunk->params[i];
						}
						CArrayCacheEntryAdd(destCArray,nextChunk,cacheParamTotals);

						if(CArrayInsertWithNoRoom(sourceChunk,sourceOffset,destCArray,destChunk,destOffset,numElements,paramChanges))
						{
							// push destChunk/destParamTotals forward if needed so that they encompass the point of insertion
							while(destChunk&&(destIndex>=destParamTotals[CA_ELEMENTS]+destChunk->params[CA_ELEMENTS]))	// see if we need to step to the next chunk (rare)
							{
								for(i=0;i<destCArray->numParams;i++)
								{
									destParamTotals[i]+=destChunk->params[i];
								}
								destChunk=destChunk->nextChunk;
							}
						}
						else
						{
							fail=true;
						}
					}
				}
				else	// last chunk does not have enough space to hold the elements (no need to add a cache entry at the end)
				{
					if(CArrayInsertWithNoRoom(sourceChunk,sourceOffset,destCArray,destChunk,destOffset,numElements,paramChanges))
					{
						// push destChunk/destParamTotals forward if needed so that they encompass the point of insertion
						while(destChunk&&(destIndex>=destParamTotals[CA_ELEMENTS]+destChunk->params[CA_ELEMENTS]))	// see if we need to step to the next chunk (rare)
						{
							for(i=0;i<destCArray->numParams;i++)
							{
								destParamTotals[i]+=destChunk->params[i];
							}
							destChunk=destChunk->nextChunk;
						}
					}
					else
					{
						fail=true;
					}
				}
			}
		}
		else
		{
			// inserting elements in an empty tree (no need to add a cache entry at the end)
			fail=!CArrayInsertWithNoChunks(sourceChunk,sourceOffset,destCArray,numElements,paramChanges);
			destChunk=destCArray->firstChunk;								// first inserted element is at the start of the first chunk
		}

		if(!fail)
		{
			for(i=0;i<destCArray->numParams;i++)							// update the array header to accurately reflect the param totals now
			{
				destCArray->paramTotals[i]+=paramChanges[i];
			}
			CArrayCacheEntriesUpdate(destCArray,steppedBack?destIndex-1:destIndex,paramChanges);	// adjust the cache as needed
		}
	}

	if(!fail)
	{
		if(outChunk)
		{
			*outChunk=destChunk;
		}
		if(outParams)
		{
			for(i=0;i<destCArray->numParams;i++)
			{
				outParams[i]=destParamTotals[i];
			}
		}
	}
	return(!fail);
}

bool CArrayInsert(CARRAY *sourceCArray,uint64_t sourceIndex,CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Insert numElements from sourceCArray at sourceIndex into destCArray at destIndex.
// If there is a problem, ErrorSet, and return false.
// NOTE: sourceCArray/sourceIndex must point to at least numElements, and must
// point to an array with the same type of elements as destCArray.
// NOTE: this will add an entry to the cache at the end of the insert.
// NOTE: to accommodate inserting unchunked data, source is allowed to contain
// chunks larger than maxElementsPerChunk where the only accurate parameter field is
// CA_ELEMENTS.
{
	CARRAY_CHUNK
		*sourceChunk;
	uint64_t
		sourceParamTotals[sourceCArray->numParams];
	uint64_t
		sourceOffset;

	CArrayParameterLocate(sourceCArray,CA_ELEMENTS,sourceIndex,&sourceChunk,sourceParamTotals);	// go hunt down the source chunk given the starting point
	sourceOffset=sourceIndex-sourceParamTotals[CA_ELEMENTS];	// get index into the source
	return(CArrayChunksInsert(sourceChunk,sourceOffset,destCArray,destIndex,numElements,outChunk,outParams));
}

bool CArrayRawInsert(const void *sourceData,CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Insert numElements sourceData into destCArray at destIndex.
// If there is a problem, ErrorSet, and return false.
// NOTE: sourceData must point to at least numElements.
// NOTE: this will add an entry to the cache at the end of the insert.
{
	uint64_t
		paramTotals[1];
	CARRAY_CHUNK
		pseudoChunk;

	paramTotals[CA_ELEMENTS]=numElements;

	pseudoChunk.nextChunk=NULL;
	pseudoChunk.previousChunk=NULL;
	pseudoChunk.cacheEntry=NULL;
	pseudoChunk.params=paramTotals;
	pseudoChunk.data=(void *)sourceData;
	return(CArrayChunksInsert(&pseudoChunk,0,destCArray,destIndex,numElements,outChunk,outParams));
}

bool CArrayEmptyInsert(CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Insert numElements of zeroes into destCArray at destIndex.
// It's assumed that the zeroes won't add to any param (other than CA_ELEMENTS).
// If there is a problem, ErrorSet, and return false.
// NOTE: this will add an entry to the cache at the end of the insert.
{
	return(CArrayChunksInsert(NULL,0,destCArray,destIndex,numElements,outChunk,outParams));
}

void CArrayLocate(CARRAY *cArray,uint32_t parameterIndex,uint64_t value,CARRAY_CHUNK **outChunk,uint64_t *outParams)
// Locate the chunk in the cArray which contains the parameter value.
// Return outChunk, with outParams as the total of the parameters leading up to the chunk.
// This will also add the located chunk to the cache.
// NOTE: this only finds the start of the chunk. Locating the parameter
// within the chunk is left to the caller.
// NOTE: for added speed when being called repeatedly for nearby
// positions, this performs a special case check against the most recent
// cache entry, and will return it immediately if it matches.
{
	uint32_t
		i;

	// see if the most recent cache entry just matches
	if((cArray->cache.first)&&(cArray->cache.first->chunk)&&(value>=cArray->cache.first->paramTotals[parameterIndex])&&(value<cArray->cache.first->paramTotals[parameterIndex]+cArray->cache.first->chunk->params[parameterIndex]))	// matches most recent?
	{
		(*outChunk)=cArray->cache.first->chunk;
		for(i=0;i<cArray->numParams;i++)
		{
			outParams[i]=cArray->cache.first->paramTotals[i];
		}
	}
	else	// most recent entry wasn't a match, do some more work to find the requested entry
	{
		CArrayParameterLocate(cArray,parameterIndex,value,outChunk,outParams);
		if(*outChunk)											// something located?
		{
			CArrayCacheEntryAdd(cArray,*outChunk,outParams);	// drop it into the cache
		}
	}
}

void CArrayDispose(CARRAY *cArray)
// Dispose of an array cArray created by CArrayCreate.
{
	CArrayCacheEntriesDispose(cArray);						// dispose of the cache entries
	CArrayChunkListDispose(cArray->firstChunk);				// get rid of any chunks hanging off the array
	MDisposePtr(cArray);									// get rid of the array itself
}

CARRAY *CArrayCreate(uint32_t maxElementsPerChunk,uint32_t elementSize,CARRAY_ELEMENTS_MOVE *elementsMoveFunction,uint32_t numParams,uint32_t numCacheEntries)
// Create an array cArray.
// If there is a problem, set the error, and return false.
{
	CARRAY
		*cArray;
	uint32_t
		i;

	if((cArray=(CARRAY *)MNewPtr(sizeof(CARRAY)+numParams*sizeof(cArray->paramTotals[0]))))	// create a CARRAY with room for numParams
	{
		cArray->firstChunk=cArray->lastChunk=(CARRAY_CHUNK *)NULL;	// no chunks in the array yet

		cArray->maxElementsPerChunk=maxElementsPerChunk;
		cArray->elementSize=elementSize;

		cArray->elementsMoveFunction=elementsMoveFunction;

		cArray->numParams=numParams;

		for(i=0;i<numParams;i++)									// clear parameter totals for the array
		{
			cArray->paramTotals[i]=0;
		}

		if(CArrayCacheEntriesCreate(cArray,numCacheEntries,numParams))	// create the cache entries
		{
			return(cArray);
		}
		MDisposePtr(cArray);										// get rid of the array itself
	}
	return((CARRAY *)NULL);
}
