// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// Chunk Arrays (CArray) are useful for managing dynamic arrays of
// objects. In e93, they're used for managing the text of buffers,
// the selections, styles, and also parts of the undo mechanism.
//
// The idea is that an arbitrary length array of objects needs to
// allow insertion/deletion quickly at any position in the array,
// and needs to allow efficient location of elements within the array,
// sometimes based on parameters of the array (e.g. locating line
// starts within the text).
//
// Using large linear buffers obviously doesn't work well for this
// as insertion and deletion involve moving arbitrary number of elements
// to make room in, or coalesce the arrays. Instead, CARRAYS break
// the array up into "chunks". The chunks for any given CARRAY are
// all of constant size, but may contain variable number of elements
// up to the maximum. The chunks are linked together in a list. As
// elements are inserted into the array, they're added to chunks where
// space is available, or new chunks are created and linked into the
// list as appropriate. This reduces the overhead of insertions and
// deletions, as it keeps from having to touch most of the uneffected
// elements in the array during these operations.

#define	CA_ELEMENTS				0	// this parameter always exists for each chunk in the array, and always allows searching the array for the chunk a given element is located in
#define	CA_NUM_INHERENT_PARAMS	1	// number of parameters inherent to CARRAY internal functions (user parameters added after this)

// NOTE: the CARRAY_ELEMENTS_MOVE function is defined to move
// elements of data from source to dest, while keeping track of
// params. Each of source, dest, and params may be passed as NULL.
// If source is passed as NULL, the function should copy zeroes
// (or equivalent for the array) into the elements. If dest
// is passed as NULL, the function does not move data, but just
// counts params. If params is NULL, the function does not count
// params during the move (makes the move faster).
// The move is defined to work correctly even if source and dest overlap.
typedef void CARRAY_ELEMENTS_MOVE(const struct cArray *cArray,const void *source,uint64_t sourceElementIndex,void *dest,uint64_t destElementIndex,uint64_t numElements,uint64_t *params);


typedef struct cArrayCacheEntry CARRAY_CACHE_ENTRY;

typedef struct cArrayChunk
{
	struct cArrayChunk
		*nextChunk,					// links the chunks together in the array
		*previousChunk;
	CARRAY_CACHE_ENTRY
		*cacheEntry;				// points to the cache entry for this chunk (or NULL if it is not in the cache)
	uint64_t
		*params;					// pointer to the parameters for this chunk (maintained by the user)
	void
		*data;						// pointer to the data for this chunk (maintained by the user)
} CARRAY_CHUNK;

typedef struct cArrayCache
{
	CARRAY_CACHE_ENTRY
		*first,						// linked list of cache entries, sorted from most recent hit to least
		*last;
} CARRAY_CACHE;

typedef struct cArray
{
	CARRAY_CHUNK
		*firstChunk,				// points to the first chunk in the array (or NULL if there are no chunks)
		*lastChunk;
	uint32_t
		maxElementsPerChunk,		// maximum number of elements in a chunk of this array
		elementSize;				// number of bytes per element
	CARRAY_ELEMENTS_MOVE
		*elementsMoveFunction;		// function to call to move elements (and count params)
	CARRAY_CACHE
		cache;						// used to make locating things in the array faster
	uint32_t
		numParams;					// number of parameters associated with the array
	uint64_t
		paramTotals[0];				// the sum of each parameter for all chunks in the array (variable length -- always at least 1 element (CA_ELEMENTS))
} CARRAY;

void CArraySanity(CARRAY *cArray);
uint64_t CArrayTotalChunksGet(CARRAY *cArray);
void CArrayCacheEntriesUpdate(CARRAY *cArray,uint64_t updatePosition,int64_t *paramChanges);
void CArrayChunkOffsetExtract(CARRAY_CHUNK **sourceChunk,uint64_t *sourceOffset,CARRAY *destCArray,void *dest,uint64_t destOffset,uint64_t numElements,uint64_t *paramTotals);
void CArrayExtract(CARRAY *cArray,uint64_t index,void *dest,uint64_t destOffset,uint64_t numElements,uint64_t *paramTotals);
void CArrayElementsDelete(CARRAY *cArray,uint64_t index,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams);
bool CArrayChunksInsert(CARRAY_CHUNK *sourceChunk,uint64_t sourceOffset,CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams);
bool CArrayInsert(CARRAY *sourceCArray,uint64_t sourceIndex,CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams);
bool CArrayRawInsert(const void *sourceData,CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams);
bool CArrayEmptyInsert(CARRAY *destCArray,uint64_t destIndex,uint64_t numElements,CARRAY_CHUNK **outChunk,uint64_t *outParams);
void CArrayLocate(CARRAY *cArray,uint32_t parameterIndex,uint64_t value,CARRAY_CHUNK **outChunk,uint64_t *outParams);
void CArrayDispose(CARRAY *cArray);
CARRAY *CArrayCreate(uint32_t maxElementsPerChunk,uint32_t elementSize,CARRAY_ELEMENTS_MOVE *elementsMoveFunction,uint32_t numParams,uint32_t numCacheEntries);

// ----------------------------------------------------------------------------
// static functions that speed access to the arrays

static inline void CArrayChunkParamSet(CARRAY_CHUNK *chunk,uint32_t parameterIndex,uint64_t value)
// Sets the given parameter for the given chunk.
{
	chunk->params[parameterIndex]=value;
}

static inline uint64_t CArrayChunkParamGet(CARRAY_CHUNK *chunk,uint32_t parameterIndex)
// Returns the given parameter for the given chunk.
{
	return(chunk->params[parameterIndex]);
}

static inline uint64_t CArrayChunkElementsGet(CARRAY_CHUNK *chunk)
// Returns the number of elements in the given chunk.
// This is just a convenience function.
{
	return(chunk->params[CA_ELEMENTS]);
}

static inline void CArrayParamTotalSet(CARRAY *cArray,uint32_t parameterIndex,uint64_t value)
// Sets the parameter given by parameterIndex for the array.
{
	cArray->paramTotals[parameterIndex]=value;
}

static inline uint64_t CArrayParamTotalGet(CARRAY *cArray,uint32_t parameterIndex)
// Returns the total of the given parameter for the given array.
{
	return(cArray->paramTotals[parameterIndex]);
}

static inline uint64_t CArrayTotalElementsGet(CARRAY *cArray)
// Returns the total number of elements in the given array.
// This is just a convenience function.
{
	return(cArray->paramTotals[CA_ELEMENTS]);
}
