// Cut, copy, and paste functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static EDITOR_BUFFER_HANDLE
	currentClipboardHandle;							// handle to the current clipboard

EDITOR_BUFFER *ClipboardCurrentGet()
// Return the current clipboard, or NULL if there is none.
{
	return(EditorBufferGetFromHandle(&currentClipboardHandle));
}

void ClipboardCurrentSet(EDITOR_BUFFER *clipboard)
// Set the current clipboard.
{
	EditorBufferHandleRelease(&currentClipboardHandle);				// release our grip on any buffer we held
	if(clipboard)
	{
		EditorBufferHandleGrab(&currentClipboardHandle,clipboard);	// hang onto this one
	}
}

bool ClipboardCopy(EDITOR_BUFFER *buffer,EDITOR_BUFFER *clipboard)
// Copy the current selection (if there is one) from buffer into clipboard.
// A "selection" is made in the clipboard text for every selection that is
// copied from the source text. These selections are used when columnar pasting.
// NOTE: it is possible for clipboard to be passed as NULL if no clipboard
// has been selected. This is considered an error.
// Also, if buffer and clipboard are the same, it is not legal to perform this
// operation.
{
	TEXT_STREAM
		stream;
	EDITOR_BUFFER
		*sourceBuffer,
		*destBuffer;
	SELECTION_UNIVERSE
		*sourceSelectionUniverse,
		*destSelectionUniverse;
	uint64_t
		sourceStartPosition,
		sourceLength,
		destStartPosition,
		destLength;
	bool
		sawInitialSelection,
		fail;

	fail=false;
	if(clipboard)
	{
		if(clipboard!=buffer)
		{
			destBuffer=clipboard;
			destSelectionUniverse=destBuffer->selectionUniverse;
			sourceBuffer=buffer;
			sourceSelectionUniverse=sourceBuffer->selectionUniverse;

			EditorReplaceStart(destBuffer);

			// kill off what was in the clipboard
			fail=!EditorTextReplace(destBuffer,0,TextUniverseTotalBytes(destBuffer->textUniverse),NULL,0,true);

			sawInitialSelection=false;
			sourceStartPosition=0;
			while(!fail&&SelectionAtOrAfterPositionGet(sourceSelectionUniverse,sourceStartPosition,&sourceStartPosition,&sourceLength))
			{
				// after every line but the last, place in a newline so that exported columnar selections work as expected
				if(sawInitialSelection)
				{
					fail=!EditorTextReplace(destBuffer,TextUniverseTotalBytes(destBuffer->textUniverse),0,(uint8_t *)"\n",1,true);
				}
				if(!fail)
				{
					TextStreamPositionSet(&stream,sourceBuffer->textUniverse,sourceStartPosition);	// point to the start of the selection in the source data
					destStartPosition=TextUniverseTotalBytes(destBuffer->textUniverse);
					if(EditorTextReplaceWithStream(destBuffer,destStartPosition,0,&stream,sourceLength,true))	// insert the text at the end of the destination
					{
						if(StyleCopy(sourceBuffer->styleUniverse,sourceStartPosition,destBuffer->styleUniverse,destStartPosition,sourceLength))
						{
							destLength=TextUniverseTotalBytes(destBuffer->textUniverse)-destStartPosition;
							if(destLength>0)
							{
								fail=!SelectionRangeSet(destSelectionUniverse,destStartPosition,destLength);	// create new selection
							}
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				sawInitialSelection=true;
				sourceStartPosition+=sourceLength;
			}
			EditorReplaceEnd(destBuffer);
		}
		else
		{
			ErrorSet("Clipboard must not be both source and destination");
			fail=true;
		}
	}
	else
	{
		ErrorSet("No current clipboard");
		fail=true;
	}
	if(!fail)
	{
		if(clipboard==ClipboardCurrentGet())		// only export the clipboard if it is the current one, and we did not fail
		{
			fail=!PlatformClipboardExport();
		}
	}
	return(!fail);
}

bool ClipboardCut(EDITOR_BUFFER *buffer,EDITOR_BUFFER *clipboard)
// Cut the current selection if there is one from buffer into clipboard.
// NOTE: it is possible for clipboard to be passed as NULL if no clipboard
// has been selected. This is considered an error.
// Also, if buffer and clipboard are the same, it is not legal to perform this
// operation.
{
	bool
		result;

	if((result=ClipboardCopy(buffer,clipboard)))
	{
		EditorReplaceStart(buffer);
		UndoGroupBegin(buffer);
		result=EditorSelectedTextDelete(buffer,buffer->selectionUniverse);
		UndoGroupEndStrict(buffer);
		EditorReplaceEnd(buffer);
	}
	return(result);
}

bool ClipboardColumnarPaste(EDITOR_BUFFER *buffer,EDITOR_VIEW *view,EDITOR_BUFFER *clipboard)
// Paste the contents of clipboard into buffer in columnar format.
// This means that each "selection" of the clipboard will be placed into buffer at
// the current cursor position, then the position is moved to the next line, and
// the process is repeated until all of the selections have been placed.
// NOTE: if there are selections in buffer, they are replaced one at a time.
// Any selections remaining after the paste is complete are deleted.
// NOTE: if there are no selections in the clipboard, then it is just pasted as one
// large blob.
// Finally, if view is passed as NULL, and the paste requires moving down a line
// at a time, this will just paste the remaining selections at the starts of lines.
// NOTE: it is possible for clipboard to be passed as NULL if no clipboard
// has been selected. This is considered an error.
// Also, if buffer and clipboard are the same, it is not legal to perform this
// operation.
{
	TEXT_STREAM
		sourceStream;
	EDITOR_BUFFER
		*sourceBuffer;
	SELECTION_UNIVERSE
		*sourceSelectionUniverse,
		*destSelectionUniverse;
	uint64_t
		sourceStartPosition,
		sourceLength,
		destStartPosition,
		destLength,
		lastReplacePosition,
		finalCursorPosition;
	uint64_t
		totalBytes,
		temp,
		lastNumLines,
		numReplaced,
		newPosition,
		line;
	int64_t
		desiredX;
	bool
		replacedOne,
		fail;

	fail=false;
	if(clipboard)
	{
		if(clipboard!=buffer)
		{
			sourceBuffer=clipboard;
			if(PlatformClipboardImport())									// give external world a chance to override clipboard
			{
				destSelectionUniverse=buffer->selectionUniverse;
				sourceSelectionUniverse=sourceBuffer->selectionUniverse;
				EditorReplaceStart(buffer);
				UndoGroupBegin(buffer);

				finalCursorPosition=lastReplacePosition=SelectionCursorPositionGet(destSelectionUniverse);	// initialize these
				if(!SelectionIsEmpty(sourceSelectionUniverse))		// see if selections in the clipboard
				{
					sourceStartPosition=0;							// get absolute position
					destStartPosition=0;
					numReplaced=0;
					replacedOne=false;								// tells if a replace has been done, so we know not to move the cursor position further

					// first, replace selections in the destination with selections from the source
					while(!fail&&SelectionAtOrAfterPositionGet(sourceSelectionUniverse,sourceStartPosition,&sourceStartPosition,&sourceLength)&&SelectionAtOrAfterPositionGet(destSelectionUniverse,0,&destStartPosition,&destLength))
					{
						TextStreamPositionSet(&sourceStream,sourceBuffer->textUniverse,sourceStartPosition);	// point to the start of the selection in the source data

						lastReplacePosition=destStartPosition;		// update this for later
						if(EditorTextReplaceWithStream(buffer,destStartPosition,destLength,&sourceStream,sourceLength,false))
						{
							if(StyleCopy(sourceBuffer->styleUniverse,sourceStartPosition,buffer->styleUniverse,destStartPosition,sourceLength))
							{
								if(!replacedOne)
								{
									finalCursorPosition=destStartPosition+sourceLength;	// place cursor after start of first replacement when this procedure is finished
									replacedOne=true;
								}
								numReplaced=sourceLength;			// number of bytes that were added at lastReplacePosition
							}
							else
							{
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						sourceStartPosition+=sourceLength;
					}

					// now, replace by moving down a line at a time
					line=TextUniversePositionToLine(buffer->textUniverse,lastReplacePosition,NULL);
					if(view)
					{
						PlatformViewTextPositionToXGet(view,lastReplacePosition,&desiredX);	// find the X offset we will attempt to insert at each time
					}

					if(replacedOne)		// if managed to replace, then move to the line BELOW the last replacement
					{
						line=TextUniversePositionToLine(buffer->textUniverse,lastReplacePosition+numReplaced,NULL);
						line++;		// start replacing on this line
					}

					while(!fail&&SelectionAtOrAfterPositionGet(sourceSelectionUniverse,sourceStartPosition,&sourceStartPosition,&sourceLength))
					{
						TextStreamPositionSet(&sourceStream,sourceBuffer->textUniverse,sourceStartPosition);	// point to the start of the selection in the source data
						destStartPosition=TextUniverseLineToPosition(buffer->textUniverse,line,NULL);
						if(view)
						{
							PlatformViewTextXToPositionGet(view,destStartPosition,desiredX,&newPosition,&temp);
						}
						else
						{
							newPosition=0;
						}

						destStartPosition+=newPosition;				// this is where the insertion should occur

						if(line>TextUniverseTotalLines(buffer->textUniverse))	// attempting to place on non-existent line?
						{
							if(EditorTextReplace(buffer,destStartPosition,0,(uint8_t *)"\n",1,false))	// stick in a newline at the end
							{
								destStartPosition++;				// then move past it
							}
							else
							{
								fail=true;
							}
						}
						lastNumLines=TextUniverseTotalLines(buffer->textUniverse);			// remember how many lines there were, so we can see how many were inserted
						if(!fail)
						{
							if(EditorTextReplaceWithStream(buffer,destStartPosition,0,&sourceStream,sourceLength,false))
							{
								if(StyleCopy(sourceBuffer->styleUniverse,sourceStartPosition,buffer->styleUniverse,destStartPosition,sourceLength))
								{
									if(!replacedOne)
									{
										finalCursorPosition=destStartPosition+sourceLength;	// place cursor after start of first replacement
										replacedOne=true;
									}
									line+=TextUniverseTotalLines(buffer->textUniverse)-lastNumLines+1;	// skip to next line, plus any inserted
								}
								else
								{
									fail=true;
								}
							}
							else
							{
								fail=true;
							}
						}
						sourceStartPosition+=sourceLength;
					}
				}
				else
				{
					totalBytes=TextUniverseTotalBytes(sourceBuffer->textUniverse);
					if(!SelectionAtOrAfterPositionGet(destSelectionUniverse,0,&destStartPosition,&destLength))
					{
						destStartPosition=finalCursorPosition;
						destLength=0;
					}
					TextStreamPositionSet(&sourceStream,sourceBuffer->textUniverse,0);				// point to the start of the selection in the source data
					if(EditorTextReplaceWithStream(buffer,destStartPosition,destLength,&sourceStream,totalBytes,false))
					{
						fail=!StyleCopy(sourceBuffer->styleUniverse,0,buffer->styleUniverse,destStartPosition,totalBytes);
					}
					else
					{
						fail=true;
					}
					finalCursorPosition=destStartPosition+totalBytes;								// place cursor after start of replacement
				}

				if(!fail)
				{
					if(EditorSelectedTextDelete(buffer,destSelectionUniverse))						// kill any remaining selections
					{
						SelectionCursorPositionSet(destSelectionUniverse,finalCursorPosition);		// drop cursor after start of first replacement
					}
					else
					{
						fail=true;
					}
				}

				UndoGroupEndStrict(buffer);
				EditorReplaceEnd(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			ErrorSet("Clipboard must not be both source and destination");
			fail=true;
		}
	}
	else
	{
		ErrorSet("No current clipboard");
		fail=true;
	}
	return(!fail);
}

void ClipboardImportEnd()
// Call this when you are done modifying the clipboard that was returned
// from ClipboardImportStart.
{
}

EDITOR_BUFFER *ClipboardImportStart()
// Call this from the gui to get the editor buffer that is
// the current clipboard, when something external to the editor has modified
// the system clipboard, and it is desired to get that data into the editor.
// It is legal for this routine to return NULL.
// In which case you should not attempt to place anything into the clipboard
// and you should not call ClipboardImportEnd.
{
	return(ClipboardCurrentGet());
}

void ClipboardExportEnd()
// Call this when you are done exporting the clipboard that was returned
// from ClipboardExportStart.
{
}

EDITOR_BUFFER *ClipboardExportStart()
// Call this from the gui to get the editor buffer that is
// the current clipboard, when you wish to export the current clipboard.
// If this returns NULL, you should not call ClipboardExportEnd.
{
	return(ClipboardCurrentGet());
}

void ClipboardUnInit()
// Undo whatever ClipboardInit did.
{
	ClipboardCurrentSet(NULL);			// clear the current clipboard, releasing any buffer it held
}

bool ClipboardInit()
// Initialize clipboard handling.
// If there is a problem, ErrorSet, and return false.
{
	EditorBufferHandleInit(&currentClipboardHandle);
	return(true);
}
