// Code point management
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct selectionElement
{
	SPARSE_ARRAY_HEADER
		saHeader;					// only need the header
} CODE_POINT_ELEMENT;


#define BIT_TABLE_ENTRIES			(256)					// number of entries in a bit table
#define BIT_TABLE_LENGTH			(BIT_TABLE_ENTRIES/32)	// number of 32-bit words in a bit table

#define	BitTableEntryGet(array,index)	((array)[(index)>>5]&(1<<((index)&0x1F)))		// get bit from field
#define	BitTableEntryClear(array,index)	((array)[(index)>>5]&=~(1<<((index)&0x1F)))		// clear bit in field
#define	BitTableEntrySet(array,index)	((array)[(index)>>5]|=(1<<((index)&0x1F)))		// set bit in field

// ----------------------------------------------------------------------------
// Code point lists are used to track arbitrary groups of 32-bit code points.
// These lists can be used for things like regular expression processing
// where an expression might provide an arbitrary list of characters
// which match at a given point.
//
// In general, code points with values <256 are common (all ASCII, and
// western European characters), and code points with values above that
// are rare. We want the common case to be as fast as possible, but must
// allow the rare cases. This is handled by using two different data
// structures. The first is a bit table (for all values <256). This is
// quick to index. The second structure is a sparse array. The sparse
// array has much greater overhead, but is still reasonably quick to
// access to see if a code point is in the list.

typedef struct codePointList
{
	uint32_t
		bits[BIT_TABLE_LENGTH];		// if a bit is set, the code point is in the list, otherwise not (used for the first BIT_TABLE_ENTRIES code points)
	SPARSE_ARRAY
		sparseArray;				// used for all code points after the first BIT_TABLE_ENTRIES
} CODE_POINT_LIST;

static void BitTableReset(uint32_t *bitTable)
// Reset all entries in the given bitTable.
{
	uint32_t
		i;

	for(i=0;i<BIT_TABLE_LENGTH;i++)
	{
		bitTable[i]=0;
	}
}

bool CodePointListTest(CODE_POINT_LIST *codePointList,uint32_t position)
// Return true if position of the given codePointList has been set.
{
	CODE_POINT_ELEMENT
		*element;
	uint64_t
		startPosition,
		numElements;

	if(position<BIT_TABLE_ENTRIES)								// is position in the bit table?
	{
		return(BitTableEntryGet(codePointList->bits,position));
	}
	else
	{
		SparseArrayRecordForPositionGet(&codePointList->sparseArray,position,&startPosition,&numElements,(void **)&element);
		return(element!=NULL);
	}
}

bool CodePointListTestAtOrAfterPositionGet(CODE_POINT_LIST *codePointList,uint32_t position,uint32_t *startPosition,uint64_t *numElements)
// Return the next range of positions at or after position which are
// occupied in the code point list. If there are no more ranges after position,
// this will return false.
{
	CODE_POINT_ELEMENT
		*element;
	uint64_t
		tempStartPosition,
		tempNumElements;

	while((position<BIT_TABLE_ENTRIES)&&!BitTableEntryGet(codePointList->bits,position))	// move through all empty entries of the bit table
	{
		position++;
	}
	if(position<BIT_TABLE_ENTRIES)		// still in the bit table?
	{
		*startPosition=position;
		while((position<BIT_TABLE_ENTRIES)&&BitTableEntryGet(codePointList->bits,position))	// move through all the continuous filled entries
		{
			position++;
		}
		*numElements=position-*startPosition;
		if(position>=BIT_TABLE_ENTRIES)	// got to the end of the bit table, move into the sparse array
		{
			if(SparseArrayRecordForPositionGet(&codePointList->sparseArray,position,&tempStartPosition,&tempNumElements,(void **)&element))
			{
				if(element)				// found something in the sparse array?
				{
					*numElements+=tempNumElements;	// add it in
				}
			}
		}
		return(true);
	}
	else	// position can only be in the sparse array
	{
		if(SparseArrayRecordAtOrAfterPositionGet(&codePointList->sparseArray,position,&tempStartPosition,&tempNumElements,(void **)&element))
		{
			*startPosition=tempStartPosition;
			*numElements=tempNumElements;
			return(true);
		}
	}
	return(false);
}

bool CodePointListRangeClear(CODE_POINT_LIST *codePointList,uint32_t startPosition,uint64_t numElements)
// Clear numElements code points from startPosition in the given list.
// If there is a problem, ErrorSet, return false.
{
	while(numElements&&(startPosition<BIT_TABLE_ENTRIES))		// handle bit table entries first
	{
		BitTableEntryClear(codePointList->bits,startPosition);
		startPosition++;
		numElements--;
	}
	return(SparseArrayRangeClear(&codePointList->sparseArray,startPosition,numElements));
}

bool CodePointListRangeSet(CODE_POINT_LIST *codePointList,uint32_t startPosition,uint64_t numElements)
// Set numElements code points from startPosition in the given list.
// If there is a problem, ErrorSet, return false.
{
	CODE_POINT_ELEMENT
		element;

	while(numElements&&(startPosition<BIT_TABLE_ENTRIES))		// handle bit table entries first
	{
		BitTableEntrySet(codePointList->bits,startPosition);
		startPosition++;
		numElements--;
	}
	return(SparseArrayRangeSet(&codePointList->sparseArray,startPosition,numElements,&element));
}

bool CodePointListsOr(CODE_POINT_LIST *destCodePointList,CODE_POINT_LIST *sourceCodePointList)
// Or all the elements of sourceCodePointList with destCodePointList.
// If there is a problem, ErrorSet, return false.
{
	uint32_t
		startPosition;
	uint64_t
		testPosition,
		numElements;
	bool
		result;

	result=true;
	testPosition=0;
	while(result&&(testPosition<0x100000000)&&CodePointListTestAtOrAfterPositionGet(sourceCodePointList,testPosition,&startPosition,&numElements))
	{
		result=CodePointListRangeSet(destCodePointList,startPosition,numElements);
		testPosition=startPosition+numElements;
	}
	return(result);
}

void CodePointListDispose(CODE_POINT_LIST *codePointList)
// Dispose of a CODE_POINT_LIST created by CodePointListCreate.
{
	SparseArrayUnInit(&codePointList->sparseArray);
	MDisposePtr(codePointList);
}

CODE_POINT_LIST *CodePointListCreate()
// Create a new empty CODE_POINT_LIST and return a pointer to it.
// If there's a problem, ErrorSet and return NULL.
{
	CODE_POINT_LIST
		*codePointList;

	if((codePointList=(CODE_POINT_LIST *)MNewPtr(sizeof(CODE_POINT_LIST))))
	{
		BitTableReset(codePointList->bits);
		if(SparseArrayInit(&codePointList->sparseArray,sizeof(CODE_POINT_ELEMENT)))
		{
			return(codePointList);
		}
		MDisposePtr(codePointList);
	}
	return(NULL);
}
