// editor '93
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static void EditorUnInit()
// Undo what EditorInit did.
{
	AbortUnInit();
	ClipboardUnInit();
	EditorBufferUnInit();
	SyntaxMapsUnInit();
	ViewColorSchemeUnInit();
	KeyBindUnInit();
	BackgroundUnInit();
	CodePointListDispose(wordSpaceList);
}

static bool EditorInit()
// Initialize all high-level editor concepts.
// If there is a problem, set the error and return false.
{
	if((wordSpaceList=CodePointListCreate()))
	{
		if(BackgroundInit())
		{
			if(KeyBindInit())
			{
				if(ViewColorSchemeInit())
				{
					if(SyntaxMapsInit())
					{
						if(EditorBufferInit())
						{
							if(ClipboardInit())
							{
								if(AbortInit())
								{
									return(true);
								}
								else
								{
									ErrorSet("Failed to initialize user abort");
								}
								ClipboardUnInit();
							}
							else
							{
								ErrorSet("Failed to initialize clipboard");
							}
							EditorBufferUnInit();
						}
						else
						{
							ErrorSet("Failed to initialize buffers");
						}
						SyntaxMapsUnInit();
					}
					else
					{
						ErrorSet("Failed to initialize syntax maps");
					}
					ViewColorSchemeUnInit();
				}
				else
				{
					ErrorSet("Failed to initialize color schemes");
				}
				KeyBindUnInit();
			}
			else
			{
				ErrorSet("Failed to initialize key bindings");
			}
			BackgroundUnInit();
		}
		else
		{
			ErrorSet("Failed to initialize background function handling");
		}
		CodePointListDispose(wordSpaceList);
	}
	else
	{
		ErrorSet("Failed to create wordSpaceList");
	}
	return(false);
}

int main(int argc,char *argv[])
// The e93 entry point.
{
	bool
		fail;
	uint32_t
		numPointers;

	fail=false;
	programName=argv[0];								// point to the program name forever more
	setlocale(LC_ALL,"");								// adjust to local customs
	if(PlatformEarlyInit())								// give process a chance to do whatever it would like (NOTE: if it returns false, exit without complaint (so it can fork if it wants to))
	{
		if(ErrorInit())
		{
			ErrorTraceClear();
			if(PlatformInit())							// handle low level GUI and environment init
			{
				ErrorTraceClear();
				if(EditorInit())						// handle high level initialization
				{
					if(PlatformLateInit())				// handle final platform initialization
					{
						ShellLoop(argc,argv);			// start the shell looping
						PlatformLateUnInit();
					}
					else
					{
						fprintf(stderr,"Failed to PlatformLateInit: %s\n",ErrorTraceGet());
					}
					EditorUnInit();
				}
				else
				{
					fprintf(stderr,"Failed to initialize editor: %s\n",ErrorTraceGet());
					fail=true;
				}
				PlatformUnInit();
			}
			else
			{
				fprintf(stderr,"Failed to PlatformInit: %s\n",ErrorTraceGet());
				fail=true;
			}
			ErrorUnInit();
		}
		else
		{
			fprintf(stderr,"Could not initialize error handlers\n");
			fail=true;
		}
		PlatformEarlyUnInit();
	}
	if((numPointers=MGetNumAllocatedPointers()))
	{
		fprintf(stderr,"Had %d pointer(s) allocated on exit\n",numPointers);
		fail=true;
	}
	if(fail)
	{
		return(1);										// tell OS bad things happened
	}
	return(0);
}
