// Edit module
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// The edit module handles the non-visible parts of editing not handled
// by text.c
// These include:
//  cursor position/selections
//  insertion and deletion of text
//  load/save
//  marks

#include	"includes.h"

static uint32_t
	scrollTimer;

uint64_t EditorOffsetFix(uint64_t offsetToFix,uint64_t startOffset,uint64_t oldEndOffset,uint64_t newEndOffset)
// Given an offset into a buffer which has just had a replacement of text
// from startOffset to oldEndOffset with text to newEndOffset, adjust
// offsetToFix so it points relatively to the same place it did
// previously. Return the adjusted offset.
{
	if(offsetToFix>=startOffset)			// if offsetToFix was before the change, then do not move it
	{
		if(offsetToFix>oldEndOffset)		// see if offsetToFix was past changed area (move it relatively)
		{
			offsetToFix=(offsetToFix+newEndOffset)-oldEndOffset;
		}
		else
		{
			offsetToFix=newEndOffset;		// if offsetToFix in middle of change, put it at the end
		}
	}
	return(offsetToFix);
}

static void EditorBufferIndicesFix(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t oldEndOffset,uint64_t newEndOffset)
// Given a change at startOffset, run through the selections, styles, anchors, etc for buffer,
// and fix them up so that they align with newEndOffset.
{
	EDITOR_MARK
		*mark;

	StylesAdjustForChange(buffer->styleUniverse,startOffset,oldEndOffset,newEndOffset);

	SelectionAdjustForChange(buffer->selectionUniverse,startOffset,oldEndOffset,newEndOffset);
	SelectionAdjustForChange(buffer->mathSelectionUniverse,startOffset,oldEndOffset,newEndOffset);
	SelectionAdjustForChange(buffer->auxSelectionUniverse,startOffset,oldEndOffset,newEndOffset);

	mark=buffer->marks;
	while(mark)
	{
		SelectionAdjustForChange(mark->selectionUniverse,startOffset,oldEndOffset,newEndOffset);	// keep all the marks updated too
		mark=mark->nextMark;
	}

	buffer->anchorStartPosition=EditorOffsetFix(buffer->anchorStartPosition,startOffset,oldEndOffset,newEndOffset);
	buffer->anchorEndPosition=EditorOffsetFix(buffer->anchorEndPosition,startOffset,oldEndOffset,newEndOffset);
}

static int64_t NewTopLineGetStrict(int64_t startLine,int64_t endLine,int64_t viewTopLine,uint64_t viewNumLines,bool useEnd)
// When homing the view vertically, this works out the best position for the top line.
{
	int64_t
		newTopLine;

	newTopLine=viewNumLines/4;
	if(!useEnd)													// see which position to home to
	{
		if(startLine<newTopLine)
		{
			newTopLine=0;
		}
		else
		{
			newTopLine=startLine-newTopLine;
		}
	}
	else
	{
		if(endLine<newTopLine)
		{
			newTopLine=0;
		}
		else
		{
			newTopLine=endLine-newTopLine;
		}
	}
	return(newTopLine);
}

static int64_t NewTopLineGetSemiStrict(int64_t startLine,int64_t endLine,int64_t viewTopLine,uint64_t viewNumLines,bool useEnd)
// When homing the view vertically, this works out the best position for the top line.
// This checks to see if the start and end are both in the view. If they are, it does nothing.
// If either is out of the view, this calls NewTopLineGetStrict.
{
	if((startLine<viewTopLine)||(endLine>=viewTopLine+(int64_t)viewNumLines))
	{
		return(NewTopLineGetStrict(startLine,endLine,viewTopLine,viewNumLines,useEnd));
	}
	return(viewTopLine);
}

static int64_t NewTopLineGetLenient(int64_t startLine,int64_t endLine,int64_t viewTopLine,uint64_t viewNumLines,bool useEnd)
// When homing the view vertically, this works out the best position for the top line.
// This tries to move as little as possible to get the start or end into the view.
{
	int64_t
		newTopLine;

	newTopLine=viewTopLine;
	if(!useEnd)															// see which position to home to
	{
		if(startLine<viewTopLine||!viewNumLines)						// line is off the top, or display is 0 lines tall
		{
			newTopLine=startLine;										// so set top to new line
		}
		else
		{
			if(startLine>=viewTopLine+(int64_t)viewNumLines)			// line is off the bottom, so place it at the bottom
			{
				newTopLine=startLine-viewNumLines+1;
			}
		}
	}
	else
	{
		if(endLine<viewTopLine||!viewNumLines)							// line is off the top, or display is 0 lines tall
		{
			newTopLine=endLine;											// so set top to new line
		}
		else
		{
			if(endLine>=viewTopLine+(int64_t)viewNumLines)				// line is off the bottom, so place it at the bottom
			{
				newTopLine=endLine-viewNumLines+1;
			}
		}
	}
	return(newTopLine);
}

static int64_t NewLeftPixelGetStrict(int64_t startPixel,int64_t endPixel,int64_t viewLeft,uint64_t viewWidth,bool useEnd)
// When homing the view horizontally, this works out the best position for the left pixel.
// This will try to center to start and end horizontally, but will force
// either the start or end onto the view depending on the setting of useEnd.
// NOTE: it is possible for endPixel to be less than startPixel. If that happens
// this still tries to center them on the view...
{
	int64_t
		newLeftPixel;
	int64_t
		pixelDistance;

	if(endPixel>=startPixel)
	{
		pixelDistance=endPixel-startPixel;
		newLeftPixel=startPixel-((int64_t)viewWidth-pixelDistance)/2;	// center
	}
	else
	{
		pixelDistance=startPixel-endPixel;
		newLeftPixel=endPixel-((int64_t)viewWidth-pixelDistance)/2;		// center
	}

	if(!useEnd)
	{
		if(newLeftPixel>startPixel)
		{
			newLeftPixel=startPixel;
		}
		if(newLeftPixel<startPixel-(int64_t)viewWidth)
		{
			newLeftPixel=startPixel-(int64_t)viewWidth;
		}
	}
	else
	{
		if(newLeftPixel>endPixel)
		{
			newLeftPixel=endPixel;
		}
		else if(newLeftPixel<endPixel-(int64_t)viewWidth)
		{
			newLeftPixel=endPixel-(int64_t)viewWidth;
		}
	}
	if(newLeftPixel<0)
	{
		newLeftPixel=0;
	}

	return(newLeftPixel);
}

static int64_t NewLeftPixelGetSemiStrict(int64_t startPixel,int64_t endPixel,int64_t viewLeft,uint64_t viewWidth,bool useEnd)
// When homing the view horizontally, this works out the best position for the left pixel.
// If start and end are on the view horizontally, this does nothing.
// Otherwise, this calls NewLeftPixelGetStrict.
{
	if((startPixel<viewLeft)||(startPixel>=viewLeft+(int64_t)viewWidth)||(endPixel<viewLeft)||(endPixel>=viewLeft+(int64_t)viewWidth))
	{
		return(NewLeftPixelGetStrict(startPixel,endPixel,viewLeft,viewWidth,useEnd));
	}
	return(viewLeft);
}

static int64_t NewLeftPixelGetLenient(int64_t startPixel,int64_t endPixel,int64_t viewLeft,uint64_t viewWidth,bool useEnd)
// When homing the view horizontally, this works out the best position for the left pixel.
// This attempts to scroll only a small amount to keep the position in the view.
{
	int64_t
		newLeftPixel;
	int64_t
		positionLeft;

	positionLeft=startPixel;
	if(useEnd)
	{
		positionLeft=endPixel;
	}
	if(positionLeft<viewLeft)
	{
		newLeftPixel=positionLeft-(positionLeft%HORIZONTAL_SCROLL_THRESHOLD);
	}
	else
	{
		if(positionLeft>=viewLeft+(int64_t)viewWidth)
		{
			newLeftPixel=((positionLeft-viewWidth)/HORIZONTAL_SCROLL_THRESHOLD+1)*HORIZONTAL_SCROLL_THRESHOLD;
		}
		else
		{
			newLeftPixel=viewLeft;		// leave it where it is
		}
	}
	return(newLeftPixel);
}


void EditorViewToRangeHome(EDITOR_VIEW *view,uint64_t startPosition,uint64_t endPosition,bool useEnd,int horizontalMovement,int verticalMovement)
// Move view so that startPosition to endPosition is displayed.
// Attempt to move the view in the ways given by horizontalMovement/verticalMovement
// if useEnd is true, make sure that endPosition is on the view even if it means that
// startPosition is forced off the view.
// NOTE: startPosition must be <= endPosition
// NOTE: This function attempts to move the view in the way that a human
// will find the most pleasant. Because of this, it uses some unusual logic
// and should not be used by any routine that needs any guarantee of where
// the view will end up.
{
	int64_t
		startLeftPixel,
		endLeftPixel;
	int64_t
		leftPixel,
		newLeftPixel;
	uint64_t
		numPixels;
	int64_t
		startLine,
		endLine,
		newTopLine,
		topLine;
	uint64_t
		numLines;

	ViewOriginGet(view,&leftPixel,&topLine);
	PlatformViewActiveBoundsGet(view,&numPixels,&numLines);

	startLine=TextUniversePositionToLine(view->parentBuffer->textUniverse,startPosition,NULL);
	startLeftPixel=0;
	if(horizontalMovement!=HT_NONE)								// see if horizontal information is needed
	{
		PlatformViewTextPositionToXGet(view,startPosition,&startLeftPixel);
	}

	if(startPosition!=endPosition)
	{
		endLine=TextUniversePositionToLine(view->parentBuffer->textUniverse,endPosition,NULL);
		endLeftPixel=0;
		if(horizontalMovement!=HT_NONE)							// see if horizontal information is needed
		{
			PlatformViewTextPositionToXGet(view,endPosition,&endLeftPixel);
		}
	}
	else
	{
		endLine=startLine;
		endLeftPixel=startLeftPixel;
	}

	// Handle vertical movement
	newTopLine=topLine;
	switch(verticalMovement)
	{
		case HT_NONE:
			break;
		case HT_LENIENT:
			newTopLine=NewTopLineGetLenient(startLine,endLine,topLine,numLines,useEnd);
			break;
		case HT_SEMI_STRICT:
			newTopLine=NewTopLineGetSemiStrict(startLine,endLine,topLine,numLines,useEnd);
			break;
		case HT_STRICT:
			newTopLine=NewTopLineGetStrict(startLine,endLine,topLine,numLines,useEnd);
			break;
	}

	// Handle horizontal movement
	newLeftPixel=leftPixel;
	switch(horizontalMovement)
	{
		case HT_NONE:
			break;
		case HT_LENIENT:
			newLeftPixel=NewLeftPixelGetLenient(startLeftPixel,endLeftPixel,leftPixel,numPixels,useEnd);
			break;
		case HT_SEMI_STRICT:
			newLeftPixel=NewLeftPixelGetSemiStrict(startLeftPixel,endLeftPixel,leftPixel,numPixels,useEnd);
			break;
		case HT_STRICT:
			newLeftPixel=NewLeftPixelGetStrict(startLeftPixel,endLeftPixel,leftPixel,numPixels,useEnd);
			break;
	}

	if((newTopLine!=topLine)||(newLeftPixel!=leftPixel))
	{
		ViewOriginSet(view,newLeftPixel,newTopLine);
	}
}

void EditorViewToSelectionHome(EDITOR_VIEW *view,bool useEnd,int horizontalMovement,int verticalMovement)
// Move view so that the selected text is displayed.
// If the selected text is larger than the view, useEnd tells which end should
// remain in the view.
{
	uint64_t
		startPosition,
		endPosition;

	SelectionEndPositionsGet(view->parentBuffer->selectionUniverse,&startPosition,&endPosition);
	EditorViewToRangeHome(view,startPosition,endPosition,useEnd,horizontalMovement,verticalMovement);
}

void EditorViewToSelectionEdgeHome(EDITOR_VIEW *view,bool useEnd,int horizontalMovement,int verticalMovement)
// Move view so that the start (or end) of the selection is displayed.
{
	uint64_t
		startPosition,
		endPosition;

	SelectionEndPositionsGet(view->parentBuffer->selectionUniverse,&startPosition,&endPosition);
	if(!useEnd)
	{
		endPosition=startPosition;
	}
	else
	{
		startPosition=endPosition;
	}
	EditorViewToRangeHome(view,startPosition,endPosition,false,horizontalMovement,verticalMovement);
}

void EditorVerticalScroll(EDITOR_VIEW *view,int64_t amountToScroll)
// Scroll the view vertically by amountToScroll.
// If amountToScroll<0, then scroll toward lower line numbers.
// The scroll will pin to the edges of the view.
{
	int64_t
		leftPixel;
	int64_t
		topLine;
	uint64_t
		totalLines;

	ViewOriginGet(view,&leftPixel,&topLine);
	totalLines=TextUniverseTotalLines(view->parentBuffer->textUniverse);
	if(amountToScroll>=0)
	{
		if((topLine+amountToScroll)<(int64_t)totalLines)
		{
			topLine+=amountToScroll;
		}
		else
		{
			topLine=totalLines;
		}
	}
	else
	{
		amountToScroll=-amountToScroll;
		if(topLine>amountToScroll)
		{
			topLine-=amountToScroll;
		}
		else
		{
			topLine=0;
		}
	}
	ViewOriginSet(view,leftPixel,topLine);
}

void EditorVerticalScrollByPages(EDITOR_VIEW *view,int64_t pagesToScroll)
// Page view vertically by pagesToScroll.
// If pagesToScroll<0, scroll toward lower line numbers in the view.
// The scroll will pin to the edges of the view.
{
	uint64_t
		numPixels;
	uint64_t
		numLines;

	PlatformViewActiveBoundsGet(view,&numPixels,&numLines);
	EditorVerticalScroll(view,pagesToScroll*(int64_t)numLines);
}

void EditorHorizontalScroll(EDITOR_VIEW *view,int64_t amountToScroll)
// Scroll the view horizontally by amountToScroll pixels.
// If amountToScroll<0, then scroll toward the left.
// The scroll will pin to the edges of the view.
{
	int64_t
		leftPixel;
	int64_t
		topLine;

	ViewOriginGet(view,&leftPixel,&topLine);
	if(amountToScroll>=0)
	{
		leftPixel+=amountToScroll;
	}
	else
	{
		if(leftPixel+amountToScroll>0)
		{
			leftPixel+=amountToScroll;
		}
		else
		{
			leftPixel=0;
		}
	}
	ViewOriginSet(view,leftPixel,topLine);
}

void EditorHorizontalScrollByPages(EDITOR_VIEW *view,int64_t pagesToScroll)
// Page view horizontally by pagesToScroll.
// If pagesToScroll<0, scroll toward the left.
// The scroll will pin to the edges of the view.
{
	uint64_t
		numPixels;
	uint64_t
		numLines;

	PlatformViewActiveBoundsGet(view,&numPixels,&numLines);
	EditorHorizontalScroll(view,pagesToScroll*numPixels);
}

void EditorSelectionChangeEnd(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse)
// A selection in buffer has finished being changed
// do whatever needs to be done.
{
	EDITOR_MARK
		*mark;

	if((mark=SelectionUniverseMarkGet(selectionUniverse)))
	{
		ViewsMarkChangeEnd(mark);
	}
	else
	{
		ViewsSelectionAreaChangeEnd(buffer);		// tell all the views of this buffer that the selection area is changed
		buffer->haveStartX=buffer->haveEndX=false;	// after selection changed, X offsets are no longer valid
		buffer->haveCurrentEnd=false;				// lose track of current end when selection changes
	}
}

void EditorSelectionChangeStart(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse)
// A selection in buffer is about to be changed in some way, do whatever is needed
// NOTE: a change in the selection could mean something as simple as a cursor position change.
// NOTE also: selection changes cannot be nested, and nothing except the selection is allowed
// to be altered during a selection change.
{
	EDITOR_MARK
		*mark;

	if((mark=SelectionUniverseMarkGet(selectionUniverse)))
	{
		ViewsMarkChangeStart(mark);
	}
	else
	{
		ViewsSelectionAreaChangeStart(buffer);		// tell all the views of this buffer that the selection area is changing
	}
}

static bool EditorTextReplaceWithStreamOrText(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,TEXT_STREAM *stream,uint8_t *textBuffer,uint64_t numBytes,bool allowUndoMerge)
// Replace the text from startOffset for length with numBytes of text from the given stream (or text buffer).
// If there is a problem, set the error, and return false.
{
	bool
		fail;

	fail=false;

	ViewsTextPieceReplaceStart(buffer,startOffset,length,numBytes);		// tell the views we're about to start a text replacement

	if(UndoDeleteRegister(buffer,startOffset,length,allowUndoMerge))	// remember what we are about to delete
	{
		TextUniverseDelete(buffer->textUniverse,startOffset,length);
		if((stream&&TextUniverseInsertFromStream(stream,buffer->textUniverse,startOffset,numBytes))||TextUniverseInsertFromText(textBuffer,buffer->textUniverse,startOffset,numBytes))
		{
			if(!UndoInsertRegister(buffer,startOffset,numBytes,allowUndoMerge))
			{
				fail=true;												// we do not cope well if this fails, just keep trying
			}
		}
		else
		{
			fail=true;
			numBytes=0;													// we have failed to insert any bytes
		}

		EditorBufferIndicesFix(buffer,startOffset,startOffset+length,startOffset+numBytes);	// fix up all indices which refer to text in the buffer (selections, styles, anchors, etc)

		if(!buffer->replaceHaveRange)									// if no changes yet, just log first change
		{
			buffer->replaceStartOffset=startOffset;
			buffer->replaceNumBytes=numBytes-length;
			buffer->replaceEndOffset=startOffset+length+buffer->replaceNumBytes;
			buffer->replaceHaveRange=true;								// remember that we saw a replacement
		}
		else
		{
			if(startOffset<buffer->replaceStartOffset)					// expand the range of changes
			{
				buffer->replaceStartOffset=startOffset;
			}
			buffer->replaceNumBytes+=numBytes-length;					// adjust for total number of bytes added/subtracted
			if(startOffset+length>buffer->replaceEndOffset)				// see if we need to push to a new end
			{
				buffer->replaceEndOffset=startOffset+numBytes;			// move to new end
			}
			else
			{
				buffer->replaceEndOffset+=numBytes-length;				// just adjust old end for new changes
			}
		}
	}
	else
	{
		fail=true;
	}

	ViewsTextPieceReplaceEnd(buffer,startOffset,length,numBytes);		// tell the views we've completed a text replacement

	return(!fail);
}

bool EditorTextReplaceWithStream(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,TEXT_STREAM *stream,uint64_t numBytes,bool allowUndoMerge)
// Replace the text from startOffset for length with numBytes of stream.
// If there is a problem, set the error, and return false.
{
	return(EditorTextReplaceWithStreamOrText(buffer,startOffset,length,stream,NULL,numBytes,allowUndoMerge));
}

bool EditorTextReplace(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,uint8_t *text,uint64_t numBytes,bool allowUndoMerge)
// Replace the text from startOffset for length with numBytes of text.
// If there is a problem, set the error, and return false.
{
	return(EditorTextReplaceWithStreamOrText(buffer,startOffset,length,NULL,text,numBytes,allowUndoMerge));
}

bool EditorFileReplace(EDITOR_BUFFER *buffer,uint64_t startOffset,uint64_t length,const char *path)
// Replace the text from startOffset for length with the text at path.
// If there is a problem, set the error, and return false.
{
	uint8_t
		*dataBuffer;
	EDITOR_FILE
		*file;
	uint64_t
		numRead;
	bool
		done,
		fail;

	fail=false;
	if((file=PlatformFileReadOpen(path)))									// open the file to load from
	{
		if((dataBuffer=(uint8_t *)MNewPtr(LOAD_SAVE_BUFFER_SIZE)))		// create a buffer to read into
		{
			done=false;
			while(!done&&!fail)											// read until we reach the end of the file
			{
				if(PlatformFileRead(file,dataBuffer,LOAD_SAVE_BUFFER_SIZE,&numRead))	// inhale as many bytes as we can fit into the buffer
				{
					if(EditorTextReplace(buffer,startOffset,length,dataBuffer,numRead,false))	// take the bytes we got, and insert them into the buffer
					{
						startOffset+=numRead;							// move these for next time
						length=0;
						if(!numRead)									// if no bytes were read, but we did not fail, then we are at the end of the file
						{
							done=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
			}
			MDisposePtr(dataBuffer);
		}
		else
		{
			fail=true;
		}
		PlatformFileClose(file);
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

void EditorReplaceEnd(EDITOR_BUFFER *buffer)
// When a replacement has been completed, call this, and it will
// update all the views as needed.
{
	uint64_t
		endOffset,
		numBytes;

	if(buffer->replaceHaveRange)						// did a replacement actually occur?
	{
		endOffset=buffer->replaceEndOffset-buffer->replaceNumBytes;		// convert current end offset to end offset as it would have looked before the text changed
		numBytes=buffer->replaceEndOffset-buffer->replaceStartOffset;	// convert this to a number of bytes of replacement of original text
		SyntaxMapUpdate(buffer,buffer->replaceStartOffset,endOffset,numBytes);
	}

	buffer->haveStartX=buffer->haveEndX=false;			// after text has been replaced, X offsets are no longer valid
	buffer->haveCurrentEnd=false;						// lose track of current end when text changes

	ViewsTextReplaceEnd(buffer);						// replacement is complete
}

void EditorReplaceStart(EDITOR_BUFFER *buffer)
// Call this when about to begin replacing text in a buffer (possibly many pieces).
// It will begin the process of remembering what has changed, so that
// syntax maps can be updated at the end.
{
	ViewsTextReplaceStart(buffer);						// let views know a replacement is coming
	buffer->replaceHaveRange=false;
	buffer->replaceStartOffset=0;
	buffer->replaceEndOffset=0;
	buffer->replaceNumBytes=0;
}

// ----------------------------------------------------------------------------

bool EditorSelectedTextDelete(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse)
// Delete all the selected text from the passed buffer, place the
// cursor at the start of the first selection deleted.
{
	uint64_t
		finalCursorPosition,
		endOffset;
	uint64_t
		startPosition,
		selectionLength;
	bool
		fail;

	fail=false;
	if(!SelectionIsEmpty(selectionUniverse))
	{
		SelectionEndPositionsGet(selectionUniverse,&finalCursorPosition,&endOffset);
		startPosition=0;
		while(!fail&&SelectionAtOrAfterPositionGet(selectionUniverse,0,&startPosition,&selectionLength))
		{
			fail=!EditorTextReplace(buffer,startPosition,selectionLength,NULL,0,false);
		}
		SelectionCursorPositionSet(selectionUniverse,finalCursorPosition);
	}
	return(!fail);
}

void EditorSelectionInfoGet(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t *startPosition,uint64_t *endPosition,uint64_t *startLine,uint64_t *endLine,uint64_t *startLinePosition,uint64_t *endLinePosition,uint64_t *totalSegments,uint64_t *totalSpan)
// Get useful information about the given selection.
{
	TEXT_STREAM
		stream;

	(*totalSegments)=SelectionSegmentsGet(selectionUniverse);
	(*totalSpan)=SelectionOccupiedElementsGet(selectionUniverse);
	if(*totalSegments)
	{
		SelectionEndPositionsGet(selectionUniverse,startPosition,endPosition);
	}
	else
	{
		(*startPosition)=(*endPosition)=SelectionCursorPositionGet(selectionUniverse);
	}
	(*startLine)=TextUniversePositionToLine(buffer->textUniverse,*startPosition,&stream);
	(*startLinePosition)=TextStreamPositionGet(&stream);
	(*endLine)=TextUniversePositionToLine(buffer->textUniverse,*endPosition,&stream);
	(*endLinePosition)=TextStreamPositionGet(&stream);
}

uint8_t *EditorNextSelectionToBuffer(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t *currentPosition,uint64_t additionalLength,uint64_t *actualLength,bool *atEnd)
// Create a buffer the size of the current selection in buffer + additionalLength,
// then read the bytes of the selection into the buffer, returning a pointer
// to the buffer. The buffer must be disposed of by the caller using MDisposePtr.
// If there is a problem, set the error, and return NULL.
// If atEnd is true, there are no more selections in the list.
// If there are no selections, return NULL, and set atEnd to true (do not ErrorSet).
// NOTE: if NULL comes back, and atEnd is false, it indicates an error.
{
	TEXT_STREAM
		stream;
	uint64_t
		numElements;
	uint8_t
		*dataBuffer;

	(*atEnd)=false;
	if(SelectionAtOrAfterPositionGet(selectionUniverse,*currentPosition,currentPosition,&numElements))	// see if there are selections
	{
		TextStreamPositionSet(&stream,buffer->textUniverse,*currentPosition);							// point the stream to the start of the selection in the source data
		(*actualLength)=numElements+additionalLength;
		if((dataBuffer=(uint8_t *)MNewPtr(*actualLength)))
		{
			TextStreamTextExtract(&stream,dataBuffer,numElements,NULL);
			(*currentPosition)+=numElements;			// update position
			return(dataBuffer);
		}
	}
	else
	{
		(*atEnd)=true;
	}
	return(NULL);
}

// ----------------------------------------------------------------------------
// Functions to handle view encoding

static uint32_t ViewTextStreamCodePoint(EDITOR_VIEW *view,const TEXT_STREAM *stream)
// Extract the code point from the current (synchronized) stream.
{
	TEXT_STREAM
		nextStream;
	uint32_t
		codePoint;

	if(view->useUtf8)
	{
		nextStream=*stream;
		codePoint=Utf8NextCodePoint(&nextStream);
	}
	else
	{
		codePoint=TextStreamChar(stream);
	}
	return(codePoint);
}

static void ViewTextStreamNext(EDITOR_VIEW *view,TEXT_STREAM *stream)
// Push the (synchronized) stream to the start of the next code point.
{
	if(view->useUtf8)
	{
		Utf8NextCodePoint(stream);
	}
	else
	{
		TextStreamNextChar(stream);
	}
}

static bool ViewTextStreamPrevious(EDITOR_VIEW *view,TEXT_STREAM *stream)
// Pull the synchronized stream back to the start of the previous code point.
{
	if(view->useUtf8)
	{
		if(TextStreamPreviousChar(stream))
		{
			Utf8Synchronize(stream);
			return(true);
		}
		return(false);
	}
	else
	{
		return(TextStreamPreviousChar(stream));
	}
}

static void ViewTextStreamSynchronize(EDITOR_VIEW *view,TEXT_STREAM *stream)
// Make sure the view text stream is synchronized to the current encoding.
{
	if(view->useUtf8)
	{
		if(!TextStreamAtEnd(stream))
		{
			Utf8Synchronize(stream);
		}
	}
}

// ----------------------------------------------------------------------------
// Cursor position/movement routines

bool EditorSelectionDelete(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse)
// Delete all the selected text in buffer specified by selectionUniverse.
// If there is a problem, ErrorSet and return false.
{
	bool
		result;

	EditorReplaceStart(buffer);
	UndoGroupBegin(buffer);
	result=EditorSelectedTextDelete(buffer,selectionUniverse);
	UndoGroupEndStrict(buffer);
	EditorReplaceEnd(buffer);
	return(result);
}

static bool IsWordSpace(uint32_t codePoint)
// Returns true if codePoint is considered to be the space between words.
{
	return(CodePointListTest(wordSpaceList,codePoint));
}

static uint64_t NextCharLeft(EDITOR_VIEW *view,uint64_t position)
// Return the position one character to the left of position.
{
	TEXT_STREAM
		stream;

	if(position)
	{
		TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position);			// point to the data
		ViewTextStreamSynchronize(view,&stream);
		ViewTextStreamPrevious(view,&stream);
		position=TextStreamPositionGet(&stream);
	}
	return(position);
}

static uint64_t NextCharRight(EDITOR_VIEW *view,uint64_t position)
// Return the position one character to the right of position.
{
	TEXT_STREAM
		stream;

	if(position<TextUniverseTotalBytes(view->parentBuffer->textUniverse))
	{
		TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position);			// point to the data
		ViewTextStreamSynchronize(view,&stream);
		ViewTextStreamNext(view,&stream);
		position=TextStreamPositionGet(&stream);
	}
	return(position);
}

static uint64_t NextWordLeft(EDITOR_VIEW *view,uint64_t position)
// Return the position one word to the left of position.
{
	TEXT_STREAM
		stream;
	uint32_t
		codePoint;
	bool
		found;

	if(position)
	{
		TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position-1);			// start looking one position back
		ViewTextStreamSynchronize(view,&stream);
		found=false;
		codePoint=ViewTextStreamCodePoint(view,&stream);
		if(IsWordSpace(codePoint))
		{
			position=TextStreamPositionGet(&stream);
			found=true;
		}
		while(!found&&(position=TextStreamPositionGet(&stream)))	// still ok to move back?
		{
			ViewTextStreamPrevious(view,&stream);				// step back
			codePoint=ViewTextStreamCodePoint(view,&stream);	// get what's there
			if(IsWordSpace(codePoint))							// word space?
			{
				found=true;
			}
		}
	}
	return(position);
}

static uint64_t NextWordRight(EDITOR_VIEW *view,uint64_t position)
// Return the position one word to the right of position.
{
	TEXT_STREAM
		stream;
	uint32_t
		codePoint;
	bool
		found;

	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position);			// look at the character in question
	ViewTextStreamSynchronize(view,&stream);
	found=false;
	if(!TextStreamAtEnd(&stream))
	{
		codePoint=ViewTextStreamCodePoint(view,&stream);
		ViewTextStreamNext(view,&stream);
		if(IsWordSpace(codePoint))
		{
			found=true;
		}
	}
	while(!found&&!TextStreamAtEnd(&stream))
	{
		codePoint=ViewTextStreamCodePoint(view,&stream);
		if(IsWordSpace(codePoint))
		{
			found=true;
		}
		else
		{
			ViewTextStreamNext(view,&stream);
		}
	}
	position=TextStreamPositionGet(&stream);
	return(position);
}

static uint64_t NextLineUp(EDITOR_VIEW *view,uint64_t position,bool *haveX,int64_t *desiredX)
// Return the position above position.
// NOTE: if there is no position above position, return the start of the text.
{
	uint64_t
		newPosition;
	uint64_t
		line,
		lineOffset,
		temp;

	line=TextUniversePositionToLine(view->parentBuffer->textUniverse,position,NULL);
	if(line)
	{
		if(!(*haveX))
		{
			PlatformViewTextPositionToXGet(view,position,desiredX);
			*haveX=true;
		}

		newPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,line-1,NULL);
		PlatformViewTextXToPositionGet(view,newPosition,*desiredX,&lineOffset,&temp);
		return(newPosition+lineOffset);
	}
	else
	{
		return(0);	// there is no line above this one
	}
}

static uint64_t NextLineDown(EDITOR_VIEW *view,uint64_t position,bool *haveX,int64_t *desiredX)
// Return the position below position.
// NOTE: if there is no position below position, return the end of the text.
{
	uint64_t
		newPosition;
	uint64_t
		line,
		lineOffset,
		temp;

	line=TextUniversePositionToLine(view->parentBuffer->textUniverse,position,NULL);
	if(line<TextUniverseTotalLines(view->parentBuffer->textUniverse))
	{
		if(!(*haveX))
		{
			PlatformViewTextPositionToXGet(view,position,desiredX);
			*haveX=true;
		}
		newPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,line+1,NULL);
		PlatformViewTextXToPositionGet(view,newPosition,*desiredX,&lineOffset,&temp);
		return(newPosition+lineOffset);
	}
	else
	{
		return(TextUniverseTotalBytes(view->parentBuffer->textUniverse));			// there is no line below this one, move to the end
	}
}

static uint64_t NextLineLeft(EDITOR_VIEW *view,uint64_t position)
// Return the position of the start of the line of position.
{
	TEXT_STREAM
		stream;
	bool
		found;

	if(position)
	{
		TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position-1);		// ignore starting character
		found=false;
		while(position&&!found)
		{
			if(TextStreamChar(&stream)=='\n')
			{
				found=true;
			}
			else
			{
				position--;
				TextStreamPreviousChar(&stream);
			}
		}
	}
	return(position);
}

static uint64_t NextLineRight(EDITOR_VIEW *view,uint64_t position)
// Return the position of the end of the line of position.
{
	TEXT_STREAM
		stream;
	bool
		found;

	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position);
	found=false;
	while(!TextStreamAtEnd(&stream)&&!found)
	{
		if(TextStreamChar(&stream)=='\n')
		{
			found=true;
		}
		else
		{
			position++;
			TextStreamNextChar(&stream);
		}
	}
	return(position);
}

static uint64_t NextParagraphLeft(EDITOR_VIEW *view,uint64_t position)
// Return the position of the start of the paragraph of position.
// NOTE: a paragraph is considered to be the text between
// empty lines.
{
	uint64_t
		totalBytes;
	TEXT_STREAM
		stream;
	uint64_t
		locatedPosition;
	bool
		foundPrevious,
		found;

	if((locatedPosition=position))
	{
		totalBytes=TextUniverseTotalBytes(view->parentBuffer->textUniverse);
		found=foundPrevious=false;
		if(position>=totalBytes)					// if searching back from the end, then step back
		{
			position=totalBytes-1;
			foundPrevious=true;
		}

		TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position);
		while(!found)
		{
			if((TextStreamChar(&stream)=='\n')&&foundPrevious)
			{
				found=true;
			}
			else
			{
				if(!(foundPrevious=(TextStreamChar(&stream)=='\n')))
				{
					locatedPosition=position;		// move to this place
				}
				if(position)
				{
					position--;
					TextStreamPreviousChar(&stream);
				}
				else
				{
					found=true;
				}
			}
		}
	}
	return(locatedPosition);
}

static uint64_t NextParagraphRight(EDITOR_VIEW *view,uint64_t position)
// Return the position of the end of the paragraph of position.
{
	TEXT_STREAM
		stream;
	uint64_t
		locatedPosition;
	bool
		foundPrevious,
		found;

	found=foundPrevious=false;
	locatedPosition=position;
	if(position)
	{
		position--;								// step back one
	}
	else
	{
		foundPrevious=true;						// if at start, then it is like having a newline previously
	}
	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,position);
	while(!TextStreamAtEnd(&stream)&&!found)
	{
		if(TextStreamChar(&stream)=='\n'&&foundPrevious)
		{
			found=true;
		}
		else
		{
			foundPrevious=(TextStreamChar(&stream)=='\n');
			position++;
			if(!foundPrevious)
			{
				locatedPosition=position;		// move to this place
			}
			TextStreamNextChar(&stream);
		}
	}
	return(locatedPosition);
}

static uint64_t NextPageUp(EDITOR_VIEW *view,uint64_t position,bool *haveX,int64_t *desiredX)
// Return the position that is one page up from position.
{
	uint64_t
		numPixels;
	uint64_t
		numLines;
	uint64_t
		newPosition;
	uint64_t
		line,
		lineOffset,
		temp;

	PlatformViewActiveBoundsGet(view,&numPixels,&numLines);
	line=TextUniversePositionToLine(view->parentBuffer->textUniverse,position,NULL);
	if(numLines)
	{
		if(line>numLines-1)
		{
			line-=numLines-1;
		}
		else
		{
			line=0;
		}
	}
	if(!(*haveX))
	{
		PlatformViewTextPositionToXGet(view,position,desiredX);
		*haveX=true;
	}

	newPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,line,NULL);
	PlatformViewTextXToPositionGet(view,newPosition,*desiredX,&lineOffset,&temp);
	return(newPosition+lineOffset);
}

static uint64_t NextPageDown(EDITOR_VIEW *view,uint64_t position,bool *haveX,int64_t *desiredX)
// Return the position that is one page down from position.
{
	uint64_t
		numPixels;
	uint64_t
		numLines,
		totalLines;
	uint64_t
		newPosition;
	uint64_t
		line,
		lineOffset,
		temp;

	PlatformViewActiveBoundsGet(view,&numPixels,&numLines);
	totalLines=TextUniverseTotalLines(view->parentBuffer->textUniverse);
	line=TextUniversePositionToLine(view->parentBuffer->textUniverse,position,NULL);
	if(numLines)
	{
		line+=numLines-1;
	}
	if(line>totalLines)
	{
		line=totalLines;
	}
	if(!(*haveX))
	{
		PlatformViewTextPositionToXGet(view,position,desiredX);
		*haveX=true;
	}

	newPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,line,NULL);
	PlatformViewTextXToPositionGet(view,newPosition,*desiredX,&lineOffset,&temp);
	return(newPosition+lineOffset);
}

static uint64_t RelativePositionGet(EDITOR_VIEW *view,uint64_t position,uint16_t relativeMode,bool *haveX,int64_t *desiredX)
// Return the position that relates to position in view by relativeMode.
// If haveX is true, desiredX is valid for moving up or down.
// This will return an updated haveX/desiredX.
{
	if(relativeMode&RPM_mBACKWARD)
	{
		switch(relativeMode&RPM_mMODE)
		{
			case RPM_CHAR:
				*haveX=false;
				return(NextCharLeft(view,position));
				break;
			case RPM_WORD:
				*haveX=false;
				return(NextWordLeft(view,position));
				break;
			case RPM_LINE:
				return(NextLineUp(view,position,haveX,desiredX));
				break;
			case RPM_LINEEDGE:
				*haveX=false;
				return(NextLineLeft(view,position));
				break;
			case RPM_PARAGRAPHEDGE:
				*haveX=false;
				return(NextParagraphLeft(view,position));
				break;
			case RPM_PAGE:
				return(NextPageUp(view,position,haveX,desiredX));
				break;
			case RPM_DOCEDGE:
				*haveX=false;
				return(0);
				break;
			default:
				*haveX=false;
				return(position);
				break;
		}
	}
	else
	{
		switch(relativeMode&RPM_mMODE)
		{
			case RPM_CHAR:
				*haveX=false;
				return(NextCharRight(view,position));
				break;
			case RPM_WORD:
				*haveX=false;
				return(NextWordRight(view,position));
				break;
			case RPM_LINE:
				return(NextLineDown(view,position,haveX,desiredX));
				break;
			case RPM_LINEEDGE:
				*haveX=false;
				return(NextLineRight(view,position));
				break;
			case RPM_PARAGRAPHEDGE:
				*haveX=false;
				return(NextParagraphRight(view,position));
				break;
			case RPM_PAGE:
				return(NextPageDown(view,position,haveX,desiredX));
				break;
			case RPM_DOCEDGE:
				*haveX=false;
				return(TextUniverseTotalBytes(view->parentBuffer->textUniverse));
				break;
			default:
				*haveX=false;
				return(position);
				break;
		}
	}
}

// selection boundary location routines

static void SelectionStreamWordLeft(EDITOR_VIEW *view,TEXT_STREAM *stream)
// Passed a view, and a stream pointing within it, move the stream
// to the word boundary on the left.
{
	bool
		found;

	if(!TextStreamAtEnd(stream))
	{
		if(!IsWordSpace(ViewTextStreamCodePoint(view,stream)))
		{
			ViewTextStreamPrevious(view,stream);
			found=false;
			while(!found)
			{
				if(IsWordSpace(ViewTextStreamCodePoint(view,stream)))
				{
					ViewTextStreamNext(view,stream);
					found=true;
				}
				else
				{
					found=!ViewTextStreamPrevious(view,stream);
				}
			}
		}
	}
}

static void SelectionStreamWordRight(EDITOR_VIEW *view,TEXT_STREAM *stream,bool selectNewLine)
// Passed a view, and a stream pointing within it, move the stream
// to the word boundary on the right.
{
	bool
		found;

	if(!TextStreamAtEnd(stream))
	{
		found=false;
		if(selectNewLine||(TextStreamChar(stream)!='\n'))
		{
			if(!IsWordSpace(ViewTextStreamCodePoint(view,stream)))
			{
				ViewTextStreamNext(view,stream);
				while(!found&&!TextStreamAtEnd(stream))
				{
					if(IsWordSpace(ViewTextStreamCodePoint(view,stream))||TextStreamChar(stream)=='\n')
					{
						found=true;
					}
					else
					{
						ViewTextStreamNext(view,stream);
					}
				}
			}
			else
			{
				ViewTextStreamNext(view,stream);
			}
		}
	}
}

static uint64_t NoMove(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Used to not move on the given side of the selection.
{
	return(betweenPosition);
}

static void ValueLeft(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint8_t leftValue,uint8_t rightValue,uint64_t *newPosition)
// Search to the left for the first unmatched leftValue.
// Adjust newPosition.
// If leftValue and rightValue are the same, this will stop at the first occurrence.
{
	TEXT_STREAM
		localStream;
	uint32_t
		codePoint;
	uint64_t
		matchCount;
	bool
		found;

	localStream=*stream;
	matchCount=0;
	found=false;
	while(!found&&ViewTextStreamPrevious(view,&localStream))
	{
		codePoint=ViewTextStreamCodePoint(view,&localStream);
		if(codePoint==leftValue)
		{
			if(matchCount)
			{
				matchCount--;
			}
			else
			{
				found=true;
			}
		}
		else
		{
			if(codePoint==rightValue)
			{
				matchCount++;
			}
		}
	}
	if(found)
	{
		ViewTextStreamNext(view,&localStream);
		(*newPosition)=TextStreamPositionGet(&localStream);
	}
}

static void ValueRight(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint8_t leftValue,uint8_t rightValue,uint64_t *newPosition)
// Search to the right for the first unmatched rightValue.
// Adjust newPosition.
// If leftValue and rightValue are the same, this will stop at the first occurrence.
{
	TEXT_STREAM
		localStream;
	uint32_t
		codePoint;
	uint64_t
		matchCount;
	bool
		found;

	localStream=*stream;
	matchCount=0;
	found=false;
	while(!found&&!TextStreamAtEnd(&localStream))
	{
		ViewTextStreamNext(view,&localStream);
		if(!TextStreamAtEnd(&localStream))
		{
			codePoint=ViewTextStreamCodePoint(view,&localStream);
			if(codePoint==rightValue)
			{
				if(matchCount)
				{
					matchCount--;
				}
				else
				{
					found=true;
				}
			}
			else
			{
				if(codePoint==leftValue)
				{
					matchCount++;
				}
			}
		}
	}
	if(found)
	{
		(*newPosition)=TextStreamPositionGet(&localStream);
	}
}

static bool SpecialWordMatchLeftLocate(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t *newPosition)
// See if the character being selected is special. If so, do special matching on it.
// This is used to match parenthesis, and brackets, and quotes, etc during the initial word
// select. Subsequent tracking is done normally.
{
	bool
		found;
	TEXT_STREAM
		localStream;

	found=false;
	switch(ViewTextStreamCodePoint(view,stream))
	{
		case '{':
		case '(':
		case '[':
		case '"':
		case '/':																	// c style comment
		case '\'':
			localStream=*stream;
			ViewTextStreamNext(view,&localStream);
			(*newPosition)=TextStreamPositionGet(&localStream);
			found=true;
			break;
		case '}':
			ValueLeft(view,stream,'{','}',newPosition);
			found=true;
			break;
		case ')':
			ValueLeft(view,stream,'(',')',newPosition);
			found=true;
			break;
		case ']':
			ValueLeft(view,stream,'[',']',newPosition);
			found=true;
			break;
	}
	return(found);
}

static bool SpecialWordMatchRightLocate(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t *newPosition)
// See if the character being selected is special. If so, do special matching on it.
// This is used to match parenthesis, and brackets, and quotes, etc during the initial word
// select. Subsequent tracking is done normally.
{
	bool
		found;

	found=false;
	switch(ViewTextStreamCodePoint(view,stream))
	{
		case '}':
		case ')':
		case ']':
			found=true;
			break;
		case '{':
			ValueRight(view,stream,'{','}',newPosition);
			found=true;
			break;
		case '(':
			ValueRight(view,stream,'(',')',newPosition);
			found=true;
			break;
		case '[':
			ValueRight(view,stream,'[',']',newPosition);
			found=true;
			break;
		case '"':
			ValueRight(view,stream,'"','"',newPosition);
			found=true;
			break;
		case '/':																	// c style comment
			ValueRight(view,stream,'/','/',newPosition);
			found=true;
			break;
		case '\'':
			ValueRight(view,stream,'\'','\'',newPosition);
			found=true;
			break;
	}
	return(found);
}

static uint64_t WordLeft(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Locate next word boundary to the left.
// Return the new between position.
{
	TEXT_STREAM
		stream;
	bool
		found;

	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,charPosition);		// look at the character in question
	ViewTextStreamSynchronize(view,&stream);
	if(!TextStreamAtEnd(&stream))
	{
		found=false;
		if(!(trackMode&TM_mREPEAT))														// if not in repeat, then see if clicked on something we want to match
		{
			found=SpecialWordMatchLeftLocate(view,&stream,&charPosition);
		}
		if(!found)
		{
			SelectionStreamWordLeft(view,&stream);
			charPosition=TextStreamPositionGet(&stream);
		}
	}
	return(charPosition);
}

static uint64_t WordRight(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Locate next word boundary to the right.
// Return the new between position.
{
	TEXT_STREAM
		stream;
	bool
		found;

	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,charPosition);
	ViewTextStreamSynchronize(view,&stream);
	if(!TextStreamAtEnd(&stream))
	{
		found=false;
		if(!(trackMode&TM_mREPEAT))									// if not in repeat, then see if clicked on something we want to match
		{
			found=SpecialWordMatchRightLocate(view,&stream,&charPosition);
		}
		if(!found)
		{
			SelectionStreamWordRight(view,&stream,true);
			charPosition=TextStreamPositionGet(&stream);
		}
	}
	return(charPosition);
}

static uint64_t LineLeft(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Locate next line boundary to the left, and including charPosition.
// Return the new between position.
{
	TEXT_STREAM
		stream;
	bool
		found;

	if(charPosition)
	{
		TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,charPosition-1);			// ignore starting character
		found=false;
		while(!found)
		{
			if(TextStreamChar(&stream)=='\n')
			{
				found=true;
			}
			else
			{
				charPosition--;
				found=!TextStreamPreviousChar(&stream);
			}
		}
	}
	return(charPosition);
}

static uint64_t LineRight(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Locate next line boundary to the right, and including charPosition.
// Return the new between position.
{
	TEXT_STREAM
		stream;
	bool
		found;

	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,charPosition);
	found=false;
	while(!TextStreamAtEnd(&stream)&&!found)
	{
		charPosition++;
		if(TextStreamChar(&stream)=='\n')
		{
			found=true;
		}
		else
		{
			TextStreamNextChar(&stream);
		}
	}
	return(charPosition);
}

static uint64_t AllLeft(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Move back to the start.
// Return the new between position.
{
	return(0);
}

static uint64_t AllRight(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode)
// Locate the end.
// Return the new between position.
{
	return(TextUniverseTotalBytes(view->parentBuffer->textUniverse));
}

// columnar selection functions
// NOTE: these are different from the functions above which return absolute positions in the
// text universe. These return offsets relative to the start of the line in question.

static uint64_t CNoMove(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition)
// Used to not move on the given side of the selection.
{
	return(betweenPosition);
}

static uint64_t ColumnarWordLeft(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition)
// Locate next word boundary to the left.
// Return the new between position.
{
	TEXT_STREAM
		localStream;

	localStream=*stream;
	TextStreamMoveRelative(&localStream,(int64_t)charPosition);		// look at the character in question
	SelectionStreamWordLeft(view,&localStream);
	charPosition=TextStreamPositionGet(&localStream)-TextStreamPositionGet(stream);

	return(charPosition);
}

static uint64_t ColumnarWordRight(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition)
// Locate next word boundary to the right.
// Return the new between position.
{
	TEXT_STREAM
		localStream;

	localStream=*stream;
	TextStreamMoveRelative(&localStream,(int64_t)charPosition);		// look at the character in question
	SelectionStreamWordRight(view,&localStream,false);
	charPosition=TextStreamPositionGet(&localStream)-TextStreamPositionGet(stream);

	return(charPosition);
}

static uint64_t ColumnarLineLeft(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition)
// Locate next line boundary to the left, and including charPosition.
// Return the new between position.
{
	return(0);
}

static uint64_t ColumnarLineRight(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition)
// Locate next line boundary to the right, and including charPosition.
// Return the new between position.
{
	TEXT_STREAM
		localStream;
	bool
		found;

	localStream=*stream;
	TextStreamMoveRelative(&localStream,(int64_t)charPosition);		// look at the character in question

	found=false;
	while(!TextStreamAtEnd(&localStream)&&!found)
	{
		if(TextStreamChar(&localStream)=='\n')
		{
			found=true;
		}
		else
		{
			charPosition++;
			TextStreamNextChar(&localStream);
		}
	}
	return(charPosition);
}

bool EditorSelectionSet(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t startPosition,uint64_t endPosition)
// Clear the current selection(s) if any, and set a new one.
// If there is a problem, ErrorSet, and return false.
{
	bool
		fail;

	fail=false;
	EditorSelectionChangeStart(buffer,selectionUniverse);
	SelectionDeleteAll(selectionUniverse);	// remove all old selections
	if(startPosition<endPosition)			// if selection contains at least one character, then create a new array element
	{
		fail=!SelectionRangeSet(selectionUniverse,startPosition,endPosition-startPosition);
	}
	else
	{
		SelectionCursorPositionSet(selectionUniverse,startPosition);		// set the cursor here
	}
	EditorSelectionChangeEnd(buffer,selectionUniverse);
	return(!fail);
}

static bool EditorNormalSelectionSet(EDITOR_VIEW *view,uint64_t startPosition,uint64_t endPosition,uint16_t trackMode)
// Set a "normal" select based on trackMode.
// If there is a problem, ErrorSet, and return false.
{
	EDITOR_BUFFER
		*buffer;
	SELECTION_UNIVERSE
		*selectionUniverse;
	bool
		fail;

	fail=false;
	buffer=view->parentBuffer;
	selectionUniverse=buffer->selectionUniverse;

	EditorSelectionChangeStart(buffer,selectionUniverse);
	SelectionDeleteAll(selectionUniverse);	// remove all old selections
	if(startPosition<endPosition)			// if selection contains at least one character, then create a new array element
	{
		fail=!SelectionRangeSet(selectionUniverse,startPosition,endPosition-startPosition);
	}
	else
	{
		SelectionCursorPositionSet(selectionUniverse,startPosition);		// set the cursor here
	}

	switch(trackMode&TM_mSELECT_MATH)
	{
		case TM_REPLACE:
			break;
		case TM_ADD:
			SelectionUniverseOr(buffer->mathSelectionUniverse,selectionUniverse);
			break;
		case TM_SUBTRACT:
			SelectionUniverseDestNotAnd(buffer->mathSelectionUniverse,selectionUniverse);
			break;
		case TM_XOR:
			SelectionUniverseXor(buffer->mathSelectionUniverse,selectionUniverse);
			break;
	}

	EditorSelectionChangeEnd(buffer,selectionUniverse);
	return(!fail);
}

static bool EditorColumnarSelectionSet(EDITOR_VIEW *view,int64_t leftX,int64_t rightX,uint64_t startLine,uint64_t endLine,uint16_t trackMode)
// Set a columnar selection based on trackMode.
// If there is a problem, ErrorSet, and return false.
{
	EDITOR_BUFFER
		*buffer;
	SELECTION_UNIVERSE
		*selectionUniverse;
	TEXT_STREAM
		stream;
	uint64_t
		currentPosition,
		startPosition,
		endPosition,
		startBetween,
		startChar,
		endBetween,
		endChar;
	uint64_t
		currentLine;
	uint64_t
		(*leftFunction)(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition),
		(*rightFunction)(EDITOR_VIEW *view,const TEXT_STREAM *stream,uint64_t betweenPosition,uint64_t charPosition);
	bool
		fail;

	fail=false;
	buffer=view->parentBuffer;
	selectionUniverse=buffer->selectionUniverse;

	switch(trackMode&TM_mSELECT_MODE)
	{
		case TM_CHAR:
			leftFunction=rightFunction=CNoMove;
			break;
		case TM_WORD:
			leftFunction=ColumnarWordLeft;
			rightFunction=ColumnarWordRight;
			break;
		case TM_LINE:
			leftFunction=ColumnarLineLeft;
			rightFunction=ColumnarLineRight;
			break;
		case TM_ALL:
			leftFunction=ColumnarLineLeft;
			rightFunction=ColumnarLineRight;
			startLine=0;
			endLine=TextUniverseTotalLines(buffer->textUniverse);
			break;
		default:
			leftFunction=rightFunction=CNoMove;
			break;
	}

	EditorSelectionChangeStart(buffer,selectionUniverse);

	SelectionDeleteAll(selectionUniverse);		// remove all old selections

	currentPosition=0;

	TextUniverseLineToPosition(buffer->textUniverse,startLine,&stream);	// find start of top line

	for(currentLine=startLine;!fail&&(currentLine<=endLine);currentLine++)
	{
		PlatformViewTextXToPositionGet(view,TextStreamPositionGet(&stream),leftX,&startBetween,&startChar);	// get line offsets to left and right of selection
		PlatformViewTextXToPositionGet(view,TextStreamPositionGet(&stream),rightX,&endBetween,&endChar);

		startPosition=TextStreamPositionGet(&stream)+leftFunction(view,&stream,startBetween,startChar);		// adjust the offsets based on the tracking mode
		endPosition=TextStreamPositionGet(&stream)+rightFunction(view,&stream,endBetween,endChar);

		if(currentLine==startLine)
		{
			SelectionCursorPositionSet(selectionUniverse,startPosition);
		}

		if(startPosition<currentPosition)		// if this extends back before a previous selection, then start immediately following it
		{
			startPosition=currentPosition;
		}
		if(endPosition<currentPosition)			// if this is too far back, then pull forward
		{
			endPosition=currentPosition;
		}
		if(startPosition<endPosition)			// make sure there is something to select
		{
			fail=!SelectionRangeSet(selectionUniverse,startPosition,endPosition-startPosition);
			currentPosition=endPosition;
		}

		TextStreamToNextLine(&stream);
	}

	switch(trackMode&TM_mSELECT_MATH)
	{
		case TM_REPLACE:
			break;
		case TM_ADD:
			SelectionUniverseOr(buffer->mathSelectionUniverse,selectionUniverse);
			break;
		case TM_SUBTRACT:
			SelectionUniverseDestNotAnd(buffer->mathSelectionUniverse,selectionUniverse);
			break;
		case TM_XOR:
			SelectionUniverseXor(buffer->mathSelectionUniverse,selectionUniverse);
			break;
	}

	EditorSelectionChangeEnd(buffer,selectionUniverse);

	return(!fail);
}

static void ColumnarXAdjust(EDITOR_BUFFER *buffer,int64_t textX,uint16_t trackMode)
// Update anchorX, columnarLeftX, columnarRightX of buffer based on
// the current tracking mode.
{
	if(!(trackMode&TM_mREPEAT))				// initial click -- work out anchor position
	{
		if(trackMode&TM_mCONTINUE)			// to continue find out which side is closest to the pointer X, and set the anchor to opposite end
		{
			if(ABS(textX-buffer->columnarLeftX)>=ABS(textX-buffer->columnarRightX))	// set anchor to the side furthest from the pointer position
			{
				buffer->anchorX=buffer->columnarLeftX;
			}
			else
			{
				buffer->anchorX=buffer->columnarRightX;
			}
		}
		else									// first click of new columnar selection -- anchor to the position given
		{
			buffer->anchorX=textX;
		}
	}
	// now, use the anchor and the point to determine the left, and right sides
	if(buffer->anchorX>=textX)
	{
		buffer->columnarLeftX=textX;
		buffer->columnarRightX=buffer->anchorX;
	}
	else
	{
		buffer->columnarLeftX=buffer->anchorX;
		buffer->columnarRightX=textX;
	}
}

static bool EditorNormalSelectionAdjust(EDITOR_VIEW *view,int64_t textX,int64_t textLine,uint16_t trackMode)
// Adjust/create the selection based on line, textX, and trackMode.
// If nothing would be selected, then place the cursor at the correct position.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		base,
		betweenPosition,
		charPosition,
		startPosition,
		endPosition;
	uint64_t
		(*leftFunction)(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode),
		(*rightFunction)(EDITOR_VIEW *view,uint64_t betweenPosition,uint64_t charPosition,uint16_t trackMode);

	buffer=view->parentBuffer;

	switch(trackMode&TM_mSELECT_MODE)
	{
		case TM_CHAR:
			leftFunction=rightFunction=NoMove;
			break;
		case TM_WORD:
			leftFunction=WordLeft;
			rightFunction=WordRight;
			break;
		case TM_LINE:
			leftFunction=LineLeft;
			rightFunction=LineRight;
			break;
		case TM_ALL:
			leftFunction=AllLeft;
			rightFunction=AllRight;
			break;
		default:
			leftFunction=rightFunction=NoMove;
			break;
	}

	ColumnarXAdjust(buffer,textX,trackMode);				// keep columnar X positions up to date (so if we switch modes, we'll have them)

	if(textLine<0)	// handle negative case
	{
		textX=0;
		textLine=0;
	}

	base=TextUniverseLineToPosition(buffer->textUniverse,textLine,NULL);
	PlatformViewTextXToPositionGet(view,base,textX,&betweenPosition,&charPosition);

	betweenPosition+=base;
	charPosition+=base;

	// adjust the anchor
	if(!(trackMode&TM_mREPEAT))								// during repeat, nothing is ever done to the anchor position
	{
		if(trackMode&TM_mCONTINUE)
		{
			if((trackMode&TM_mSELECT_MODE)==TM_CHAR)		// on initial click, work out what the anchors are
			{
				SelectionEndPositionsGet(buffer->selectionUniverse,&startPosition,&endPosition);	// find ends of old selection, drop anchors
				if(betweenPosition>=endPosition)			// after the end
				{
					buffer->anchorStartPosition=buffer->anchorEndPosition=startPosition;
				}
				else if(betweenPosition<=startPosition)		// before the start
				{
					buffer->anchorStartPosition=buffer->anchorEndPosition=endPosition;
				}
				else										// inside the current selection
				{
					if(endPosition-betweenPosition>=betweenPosition-startPosition)
					{
						buffer->anchorStartPosition=buffer->anchorEndPosition=endPosition;
					}
					else
					{
						buffer->anchorStartPosition=buffer->anchorEndPosition=startPosition;
					}
				}
			}
			else
			{
				startPosition=buffer->anchorStartPosition;
				endPosition=buffer->anchorEndPosition;
			}

			if(betweenPosition<buffer->anchorStartPosition)
			{
				startPosition=leftFunction(view,betweenPosition,charPosition,trackMode);
			}
			else if(betweenPosition>=buffer->anchorEndPosition)
			{
				endPosition=rightFunction(view,betweenPosition,charPosition,trackMode);
			}
		}
		else		// selection is beginning here
		{
			startPosition=leftFunction(view,betweenPosition,charPosition,trackMode);
			endPosition=rightFunction(view,betweenPosition,charPosition,trackMode);
			buffer->anchorStartPosition=startPosition;	// if starting a new selection, and not continuing from an old one, then drop anchor on the selection
			buffer->anchorEndPosition=endPosition;
		}
	}
	else	// repeating
	{
		startPosition=buffer->anchorStartPosition;
		endPosition=buffer->anchorEndPosition;

		if(betweenPosition<buffer->anchorStartPosition)
		{
			startPosition=leftFunction(view,betweenPosition,charPosition,trackMode);
		}
		else if(betweenPosition>=buffer->anchorEndPosition)
		{
			endPosition=rightFunction(view,betweenPosition,charPosition,trackMode);
		}
	}
	return(EditorNormalSelectionSet(view,startPosition,endPosition,trackMode));
}

static bool EditorColumnarSelectionAdjust(EDITOR_VIEW *view,int64_t textX,int64_t textLine,uint16_t trackMode)
// Adjust the columnar selection in the view based on the given global coordinates, and
// trackMode.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	uint64_t
		anchorLine,
		topLine,
		bottomLine;

	buffer=view->parentBuffer;

	ColumnarXAdjust(buffer,textX,trackMode);	// keep columnar X positions up to date

	if(textLine<0)							// handle negative case
	{
		textLine=0;
	}

	// work out what to do in "Y"
	if(!(trackMode&TM_mREPEAT))				// initial click -- work out anchor position
	{
		if(trackMode&TM_mCONTINUE)			// to continue find out which side of the rect is closest to the point, and set the anchor to opposite end
		{
			SelectionEndPositionsGet(buffer->selectionUniverse,&startPosition,&endPosition);	// find ends of old selection to determine where to drop anchor vertically

			topLine=TextUniversePositionToLine(buffer->textUniverse,startPosition,NULL);	// convert positions to lines
			bottomLine=TextUniversePositionToLine(buffer->textUniverse,endPosition,NULL);

			if(ABS(textLine-(int64_t)topLine)>=ABS(textLine-(int64_t)bottomLine))		// set anchor to the side furthest from the pointer position
			{
				buffer->anchorStartPosition=startPosition;
			}
			else
			{
				buffer->anchorStartPosition=endPosition;
			}
		}
		else									// first click of new columnar selection -- anchor to the position given
		{
			buffer->anchorStartPosition=TextUniverseLineToPosition(buffer->textUniverse,textLine,NULL);	// if starting a new selection, and not continuing from an old one, then drop anchor at the line of the click
		}
	}

	anchorLine=TextUniversePositionToLine(buffer->textUniverse,buffer->anchorStartPosition,NULL);	// convert anchor position to anchor line

	// now, use the anchor line and the pointer line to determine the top, and bottom lines of the selection
	if(anchorLine>=(uint64_t)textLine)
	{
		topLine=textLine;
		bottomLine=anchorLine;
	}
	else
	{
		topLine=anchorLine;
		bottomLine=textLine;
	}

	return(EditorColumnarSelectionSet(view,buffer->columnarLeftX,buffer->columnarRightX,topLine,bottomLine,trackMode));
}

static void ViewScrollToFollowPointer(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine)
// Given a view relative line and X offset, scroll the view.
{
	int64_t
		originLeftPixel,
		newLeftPixel;
	int64_t
		originTopLine,
		newTopLine;
	int64_t
		boundsLeftPixel,
		boundsTopLine;
	uint64_t
		boundsNumPixels,
		boundsNumLines;
	uint64_t
		totalTextLines;

	ViewOriginGet(view,&originLeftPixel,&originTopLine);
	PlatformViewScrollBoundsGet(view,&boundsLeftPixel,&boundsTopLine,&boundsNumPixels,&boundsNumLines);

	if(viewLine<boundsTopLine)
	{
		if(originTopLine)
		{
			newTopLine=originTopLine-1;
		}
		else
		{
			newTopLine=0;
		}
	}
	else
	{
		if(viewLine>=(int64_t)(boundsTopLine+boundsNumLines))
		{
			totalTextLines=TextUniverseTotalLines(view->parentBuffer->textUniverse);
			if(originTopLine<(int64_t)totalTextLines)
			{
				newTopLine=originTopLine+1;
			}
			else
			{
				newTopLine=totalTextLines;
			}
		}
		else
		{
			newTopLine=originTopLine;		// don't change the top line
		}
	}

	if(viewX<boundsLeftPixel)
	{
		if(originLeftPixel-HORIZONTAL_SCROLL_THRESHOLD>0)
		{
			newLeftPixel=originLeftPixel-HORIZONTAL_SCROLL_THRESHOLD;
		}
		else
		{
			newLeftPixel=0;
		}
	}
	else
	{
		if(viewX>=(int64_t)(boundsLeftPixel+boundsNumPixels))
		{
			newLeftPixel=originLeftPixel+HORIZONTAL_SCROLL_THRESHOLD;
		}
		else
		{
			newLeftPixel=originLeftPixel;
		}
	}
	if((newTopLine!=originTopLine)||(newLeftPixel!=originLeftPixel))
	{
		ViewOriginSet(view,newLeftPixel,newTopLine);
	}
}

static void ViewRateLimitedScrollToFollowPointer(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine)
// See if it is time to allow another scroll of the view.
{
	uint32_t
		currentTime;

	currentTime=PlatformMonotonicTime();
	if(currentTime-scrollTimer>=SCROLL_FOLLOW_DELAY)		// time for another scroll?
	{
		ViewScrollToFollowPointer(view,viewX,viewLine);		// do the scroll
		scrollTimer+=SCROLL_FOLLOW_DELAY;					// push forward time to next
		if(currentTime-scrollTimer>=SCROLL_FOLLOW_DELAY)	// if too much time passed, just catch up
		{
//			fprintf(stderr,"scroll time overflow\n");
			scrollTimer=currentTime;
		}
	}
}

static void ViewRateLimitedScrollReset()
// Reset the scroll rate limit timer.
{
	scrollTimer=PlatformMonotonicTime();
}

static bool EditorViewPointerTrackNormal(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine,uint16_t trackMode)
// Track the pointer in non-columnar mode.
{
	int64_t
		textX;
	int64_t
		textLine;
	static int64_t
		lastTextX;
	static int64_t
		lastTextLine;
	bool
		result;

	result=true;

	PlatformViewCoordinateXToTextCoordinateX(view,viewX,&textX);
	PlatformViewCoordinateLineToTextCoordinateLine(view,viewLine,&textLine);

	if(!(trackMode&TM_mREPEAT))		// first time through, reset scroll timer
	{
		ViewRateLimitedScrollReset();
	}

	if((!(trackMode&TM_mREPEAT))||(textLine!=lastTextLine)||(textX!=lastTextX))	// if first time, or point has moved, then track it
	{
		lastTextX=textX;
		lastTextLine=textLine;
		result=EditorNormalSelectionAdjust(view,textX,textLine,trackMode);
	}
	ViewRateLimitedScrollToFollowPointer(view,viewX,viewLine);
	return(result);
}

static bool EditorViewPointerTrackColumnar(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine,uint16_t trackMode)
// Track the pointer in columnar mode.
{
	int64_t
		textX;
	int64_t
		textLine;
	static int64_t
		lastTextX;
	static int64_t
		lastTextLine;
	bool
		result;

	result=true;

	PlatformViewCoordinateXToTextCoordinateX(view,viewX,&textX);
	PlatformViewCoordinateLineToTextCoordinateLine(view,viewLine,&textLine);

	if(!(trackMode&TM_mREPEAT))		// first time through, reset scroll timer
	{
		ViewRateLimitedScrollReset();
	}

	if((!(trackMode&TM_mREPEAT))||(textLine!=lastTextLine)||(textX!=lastTextX))	// if first time, or point has moved, then track it
	{
		lastTextX=textX;
		lastTextLine=textLine;
		result=EditorColumnarSelectionAdjust(view,textX,textLine,trackMode);
	}
	ViewRateLimitedScrollToFollowPointer(view,viewX,viewLine);
	return(result);
}

static bool EditorViewPointerTrackHand(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine,uint16_t trackMode)
// Track the pointer in hand mode.
{
	static int64_t
		baseViewXPosition,
		baseXPosition;
	static int64_t
		baseLine;
	static uint64_t
		baseViewTopLine;
	int64_t
		leftPixel,
		newLeftPixel;
	int64_t
		topLine,
		newTopLine;
	uint64_t
		numLines;

	ViewOriginGet(view,&leftPixel,&topLine);

	numLines=TextUniverseTotalLines(view->parentBuffer->textUniverse);	// only care about number of lines of text total (not number of lines in the view)

	if(!(trackMode&TM_mREPEAT))			// if first time through, get and remember the passed position
	{
		baseViewXPosition=leftPixel;	// remember where the view was when this started
		baseViewTopLine=topLine;
		baseXPosition=viewX;			// remember where the pointer was when this started
		baseLine=viewLine;
	}

	// do math carefully so that we can handle an unsigned number of lines in the buffer
	if(viewLine>baseLine)				// moving backward in the file
	{
		if((uint64_t)(viewLine-baseLine)<baseViewTopLine)	// ok to subtract w/o underflow?
		{
			newTopLine=baseViewTopLine-(uint64_t)(viewLine-baseLine);
		}
		else
		{
			newTopLine=0;
		}
	}
	else	// moving forward
	{
		if(((uint64_t)(baseLine-viewLine)<numLines)&&(baseViewTopLine<(numLines-(uint64_t)(baseLine-viewLine))))
		{
			newTopLine=baseViewTopLine+(uint64_t)(baseLine-viewLine);
		}
		else
		{
			newTopLine=numLines;
		}
	}

	newLeftPixel=baseViewXPosition-(viewX-baseXPosition);
	if(newLeftPixel<0)
	{
		newLeftPixel=0;
	}

	if((newTopLine!=topLine)||(newLeftPixel!=leftPixel))
	{
		ViewOriginSet(view,newLeftPixel,newTopLine);
	}
	return(true);
}

void EditorViewPointerTrackComplete(EDITOR_VIEW *view)
// Sent when pointer tracking ends in a view. This can clean up after the tracking.
{
	SelectionDeleteAll(view->parentBuffer->mathSelectionUniverse);	// clear math universe at end of tracking
}

bool EditorViewPointerTrack(EDITOR_VIEW *view,int64_t viewX,int64_t viewLine,uint16_t trackMode)
// Track the pointer in view, according to the current cursor/selection in view
// and trackMode.
{
	bool
		result;
	SELECTION_UNIVERSE
		*selectionUniverse;

	result=true;
	switch(trackMode&TM_mTRACK_MODE)
	{
		default:
		case TM_NORMAL:
		case TM_COLUMNAR:
			if((trackMode&TM_mSELECT_MATH)!=TM_REPLACE)	// performing math on the selection
			{
				// if first time through for math based selection, then move the current universe to
				// the math universe so we have the baseline which will be modified
				if(!(trackMode&TM_mREPEAT))
				{
					selectionUniverse=view->parentBuffer->mathSelectionUniverse;	// swap the selection and math universes
					view->parentBuffer->mathSelectionUniverse=view->parentBuffer->selectionUniverse;
					view->parentBuffer->selectionUniverse=selectionUniverse;
				}
			}
			if((trackMode&TM_mTRACK_MODE)==TM_NORMAL)
			{
				result=EditorViewPointerTrackNormal(view,viewX,viewLine,trackMode);
			}
			else
			{
				result=EditorViewPointerTrackColumnar(view,viewX,viewLine,trackMode);
			}
			break;
		case TM_HAND:
			result=EditorViewPointerTrackHand(view,viewX,viewLine,trackMode);
			break;
	}
	return(result);
}

bool EditorCursorMove(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode)
// Move the cursor according to relativeMode, unless there is a selection.
// If there is a selection, send the cursor to the top if relativeMode is backward.
// Send the cursor to the bottom of the selection if relativeMode is forward.
{
	uint64_t
		startPosition,
		endPosition;
	bool
		haveX;
	int64_t
		desiredX;
	bool
		result;

	result=true;
	if(!SelectionIsEmpty(selectionUniverse))
	{
		SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);		// find ends of old selection
		if(relativeMode&RPM_mBACKWARD)
		{
			result=EditorSelectionSet(view->parentBuffer,selectionUniverse,startPosition,startPosition);
		}
		else
		{
			result=EditorSelectionSet(view->parentBuffer,selectionUniverse,endPosition,endPosition);
		}
	}
	else
	{
		haveX=view->parentBuffer->haveStartX;			// pass these to relative position routine
		desiredX=view->parentBuffer->desiredStartX;
		startPosition=RelativePositionGet(view,SelectionCursorPositionGet(selectionUniverse),relativeMode,&haveX,&desiredX);
		result=EditorSelectionSet(view->parentBuffer,selectionUniverse,startPosition,startPosition);
		view->parentBuffer->haveStartX=haveX;			// update these
		view->parentBuffer->desiredStartX=desiredX;
	}
	return(result);
}

bool EditorSelectionMove(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode)
// Move the current end of the selection according to relativeMode.
// NOTE: if there is no current end, then use relative mode to determine
// what the current end should be.
{
	uint64_t
		tempPosition,
		startPosition,
		endPosition;
	bool
		tempBool,
		haveStartX,
		haveEndX,
		haveCurrentEnd;
	int64_t
		tempX,
		desiredStartX,
		desiredEndX;
	bool
		result;

	SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);		// find ends of old selection

	haveStartX=view->parentBuffer->haveStartX;
	desiredStartX=view->parentBuffer->desiredStartX;
	haveEndX=view->parentBuffer->haveEndX;
	desiredEndX=view->parentBuffer->desiredEndX;

	haveCurrentEnd=view->parentBuffer->haveCurrentEnd;

	if(!haveCurrentEnd)						// if there is no current end, take a cue from the desired direction
	{
		view->parentBuffer->currentIsStart=false;
		if(relativeMode&RPM_mBACKWARD)
		{
			view->parentBuffer->currentIsStart=true;
		}
		haveCurrentEnd=true;
	}

	if(view->parentBuffer->currentIsStart)
	{
		startPosition=RelativePositionGet(view,startPosition,relativeMode,&haveStartX,&desiredStartX);
	}
	else
	{
		endPosition=RelativePositionGet(view,endPosition,relativeMode,&haveEndX,&desiredEndX);
	}

	if(startPosition>endPosition)			// have the ends swapped?
	{
		tempPosition=startPosition;			// swap them here
		startPosition=endPosition;
		endPosition=tempPosition;

		tempBool=haveStartX;
		haveStartX=haveEndX;
		haveEndX=tempBool;

		tempX=desiredStartX;
		desiredStartX=desiredEndX;
		desiredEndX=tempX;

		view->parentBuffer->currentIsStart=!view->parentBuffer->currentIsStart;	// swap the current end
	}

	result=EditorSelectionSet(view->parentBuffer,selectionUniverse,startPosition,endPosition);

	view->parentBuffer->haveStartX=haveStartX;				// update these
	view->parentBuffer->desiredStartX=desiredStartX;
	view->parentBuffer->haveEndX=haveEndX;
	view->parentBuffer->desiredEndX=desiredEndX;

	view->parentBuffer->haveCurrentEnd=haveCurrentEnd;
	return(result);
}

bool EditorNormalSelectionExpand(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode)
// Expand the selection using relativeMode.
{
	uint64_t
		startPosition,
		endPosition;
	bool
		haveStartX,
		haveEndX;
	int64_t
		desiredStartX,
		desiredEndX;
	bool
		result;

	SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);		// find ends of old selection

	haveStartX=view->parentBuffer->haveStartX;
	desiredStartX=view->parentBuffer->desiredStartX;
	haveEndX=view->parentBuffer->haveEndX;
	desiredEndX=view->parentBuffer->desiredEndX;

	if(relativeMode&RPM_mBACKWARD)
	{
		startPosition=RelativePositionGet(view,startPosition,relativeMode,&haveStartX,&desiredStartX);
		view->parentBuffer->currentIsStart=true;
	}
	else
	{
		endPosition=RelativePositionGet(view,endPosition,relativeMode,&haveEndX,&desiredEndX);
		view->parentBuffer->currentIsStart=false;
	}
	result=EditorSelectionSet(view->parentBuffer,selectionUniverse,startPosition,endPosition);

	view->parentBuffer->haveStartX=haveStartX;				// update these
	view->parentBuffer->desiredStartX=desiredStartX;
	view->parentBuffer->haveEndX=haveEndX;
	view->parentBuffer->desiredEndX=desiredEndX;

	view->parentBuffer->haveCurrentEnd=true;
	return(result);
}

bool EditorNormalSelectionReduce(EDITOR_VIEW *view,SELECTION_UNIVERSE *selectionUniverse,uint16_t relativeMode)
// Reduce the selection using relativeMode.
{
	uint64_t
		startPosition,
		endPosition;
	bool
		haveStartX,
		haveEndX;
	int64_t
		desiredStartX,
		desiredEndX;
	bool
		result;

	SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);		// find ends of old selection

	haveStartX=view->parentBuffer->haveStartX;
	desiredStartX=view->parentBuffer->desiredStartX;
	haveEndX=view->parentBuffer->haveEndX;
	desiredEndX=view->parentBuffer->desiredEndX;

	if(relativeMode&RPM_mBACKWARD)
	{
		endPosition=RelativePositionGet(view,endPosition,relativeMode,&haveEndX,&desiredEndX);
		if(endPosition<=startPosition)
		{
			haveStartX=haveEndX=false;
			endPosition=startPosition;
		}
		view->parentBuffer->currentIsStart=false;
	}
	else
	{
		startPosition=RelativePositionGet(view,startPosition,relativeMode,&haveStartX,&desiredStartX);
		if(startPosition>=endPosition)
		{
			haveStartX=haveEndX=false;
			startPosition=endPosition;
		}
		view->parentBuffer->currentIsStart=true;
	}
	result=EditorSelectionSet(view->parentBuffer,selectionUniverse,startPosition,endPosition);

	view->parentBuffer->haveStartX=haveStartX;				// update these
	view->parentBuffer->desiredStartX=desiredStartX;
	view->parentBuffer->haveEndX=haveEndX;
	view->parentBuffer->desiredEndX=desiredEndX;

	view->parentBuffer->haveCurrentEnd=true;
	return(result);
}

bool EditorLineLocate(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint64_t line)
// Find the given line, and select it.
// If the line is past the end of the text, select the end of the text.
// NOTE: this thinks of lines as starting at 0, not 1
{
	TEXT_STREAM
		stream;
	uint64_t
		startPosition,
		endPosition;

	startPosition=TextUniverseLineToPosition(buffer->textUniverse,line,&stream);
	TextStreamToNextLine(&stream);
	endPosition=TextStreamPositionGet(&stream);
	return(EditorSelectionSet(buffer,selectionUniverse,startPosition,endPosition));
}

bool EditorSelectAll(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse)
// Select all of the text in buffer.
{
	return(EditorSelectionSet(buffer,selectionUniverse,0,TextUniverseTotalBytes(buffer->textUniverse)));
}

bool EditorDelete(EDITOR_VIEW *view,uint16_t relativeMode)
// Delete the character(s) between the current cursor position, and relativeMode away from it.
{
	EDITOR_BUFFER
		*buffer;
	SELECTION_UNIVERSE
		*selectionUniverse;
	uint64_t
		relativePosition,
		startDelete,
		endDelete;
	bool
		haveX;
	int64_t
		desiredX;
	bool
		fail;

	fail=false;
	buffer=view->parentBuffer;
	selectionUniverse=buffer->selectionUniverse;
	EditorReplaceStart(buffer);

	if(!SelectionIsEmpty(selectionUniverse))
	{
		UndoGroupBegin(buffer);
		fail=!EditorSelectedTextDelete(buffer,selectionUniverse);
		UndoGroupEndStrict(buffer);
		haveX=false;
		desiredX=0;							// keep compilers from complaining
	}
	else
	{
		haveX=buffer->haveStartX;			// pass these to relative position routine
		desiredX=buffer->desiredStartX;
		relativePosition=RelativePositionGet(view,SelectionCursorPositionGet(selectionUniverse),relativeMode,&haveX,&desiredX);

		if(relativePosition<SelectionCursorPositionGet(selectionUniverse))
		{
			startDelete=relativePosition;
			endDelete=SelectionCursorPositionGet(selectionUniverse);
		}
		else
		{
			startDelete=SelectionCursorPositionGet(selectionUniverse);
			endDelete=relativePosition;
		}
		if(startDelete<endDelete)
		{
			if(!EditorTextReplace(buffer,startDelete,endDelete-startDelete,NULL,0,true))
			{
				fail=true;
			}
		}
	}
	EditorReplaceEnd(buffer);
	buffer->haveStartX=haveX;			// update these
	buffer->desiredStartX=desiredX;

	return(!fail);
}

static bool EditorInsertAtSelection(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,uint8_t *text,uint64_t textLength)
// Insert text after the cursor or over the selection in the given selection universe.
{
	bool
		fail;

	fail=false;
	EditorReplaceStart(buffer);
	if(!SelectionIsEmpty(selectionUniverse))
	{
		UndoGroupBegin(buffer);
		if(EditorSelectedTextDelete(buffer,selectionUniverse))						// get rid of selections
		{
			fail=!EditorTextReplace(buffer,SelectionCursorPositionGet(selectionUniverse),0,text,textLength,true);
		}
		else
		{
			fail=true;
		}
		UndoGroupEnd(buffer);
	}
	else	// if not deleting a selection, then let undo group continue from where it left off
	{
		fail=!EditorTextReplace(buffer,SelectionCursorPositionGet(selectionUniverse),0,text,textLength,true);
	}
	EditorReplaceEnd(buffer);

	return(!fail);
}

bool EditorInsert(EDITOR_BUFFER *buffer,uint8_t *text,uint64_t textLength)
// Insert the text after the cursor, or over the selection if there is one.
{
	return(EditorInsertAtSelection(buffer,buffer->selectionUniverse,text,textLength));
}

bool EditorAuxInsert(EDITOR_BUFFER *buffer,uint8_t *text,uint64_t textLength)
// Insert the text after the aux cursor, or over the aux selection if there is one.
{
	return(EditorInsertAtSelection(buffer,buffer->auxSelectionUniverse,text,textLength));
}

bool EditorFileInsert(EDITOR_BUFFER *buffer,const char *path)
// Insert the data in the file at path after the cursor or over the selection
// in the selection universe.
// if there is a ErrorSet, and return false.
{
	SELECTION_UNIVERSE
		*selectionUniverse;
	bool
		fail;
	uint64_t
		replacePosition;

	fail=false;
	selectionUniverse=buffer->selectionUniverse;
	EditorReplaceStart(buffer);
	UndoGroupBegin(buffer);
	if(!SelectionIsEmpty(selectionUniverse))
	{
		fail=!EditorSelectedTextDelete(buffer,selectionUniverse);	// NOTE: we do not end the group here, so that the replace is in the same group, otherwise, it may or may not be depending on if the selection had more than one element
	}
	if(!fail)
	{
		replacePosition=SelectionCursorPositionGet(selectionUniverse);
		fail=!EditorFileReplace(buffer,replacePosition,0,path);
	}
	UndoGroupEndStrict(buffer);
	EditorReplaceEnd(buffer);
	return(!fail);
}

bool EditorAutoIndent(EDITOR_BUFFER *buffer)
// Drop a return into the text, and auto indent the next line.
// If there is a ErrorSet, and return false.
{
	TEXT_STREAM
		stream,
		startStream;
	SELECTION_UNIVERSE
		*selectionUniverse;
	uint64_t
		startReplace;
	uint64_t
		lineOffset;
	uint64_t
		numWhite;
	bool
		fail,
		done;
	uint8_t
		*insertBuffer;
	bool
		hadGroup;

	fail=hadGroup=false;
	selectionUniverse=buffer->selectionUniverse;

	EditorReplaceStart(buffer);
	if(!SelectionIsEmpty(selectionUniverse))
	{
		UndoGroupBegin(buffer);
		fail=!EditorSelectedTextDelete(buffer,selectionUniverse);
		hadGroup=true;
	}
	if(!fail)
	{
		startReplace=SelectionCursorPositionGet(selectionUniverse);

		TextUniversePositionToLine(buffer->textUniverse,startReplace,&stream);
		startStream=stream;
		lineOffset=startReplace-TextStreamPositionGet(&stream);					// work out how many bytes into the line we were
		numWhite=0;
		done=false;
		while(numWhite<lineOffset&&!done)										// run through counting the number of white space characters in this line, until the first non-white is seen, or until we reach the place where we will insert the return
		{
			switch(TextStreamChar(&stream))
			{
				case ' ':
				case '\t':
					numWhite++;
					break;
				default:
					done=true;
					break;
			}
			TextStreamNextChar(&stream);
		}
		if((insertBuffer=(uint8_t *)MNewPtr(numWhite+1)))						// make room for all the white space, and the CR
		{
			insertBuffer[0]='\n';												// jam in the CR

			TextStreamTextExtract(&startStream,&(insertBuffer[1]),numWhite,NULL);	// copy the white space from the line
			fail=!EditorTextReplace(buffer,startReplace,0,insertBuffer,numWhite+1,true);
			MDisposePtr(insertBuffer);
		}
		else
		{
			fail=true;
		}
	}
	if(hadGroup)
	{
		UndoGroupEnd(buffer);
	}
	EditorReplaceEnd(buffer);
	return(!fail);
}

// ----------------------------------------------------------------------------

void EditorBufferViewUnlink(EDITOR_VIEW *view)
// Disconnect view from its buffer.
{
	EDITOR_VIEW
		*previousView,
		*currentView;

	previousView=NULL;
	currentView=view->parentBuffer->firstView;
	while(currentView&&currentView!=view)
	{
		previousView=currentView;
		currentView=currentView->nextBufferView;
	}
	if(currentView)
	{
		if(previousView)
		{
			previousView->nextBufferView=view->nextBufferView;
		}
		else
		{
			view->parentBuffer->firstView=view->nextBufferView;
		}
	}
}

void EditorBufferViewLink(EDITOR_BUFFER *buffer,EDITOR_VIEW *view)
// Connect view to buffer (at the start).
{
	view->parentBuffer=buffer;
	view->nextBufferView=buffer->firstView;
	buffer->firstView=view;
}
