// Handle high level buffer stuff
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	TASK_DATA_BUFFER_SIZE	32768						// size of buffer used to haul in data from running tasks

// Some handy notes about buffers:
// shellBusy
//   This is used exclusively by the the shell. It indicates to the
//   shell that it is already doing something with the buffer, and that it is
//   not allowed to affect it in any way (useful when doing search and replace,
//   where replace is a shell script... keeps script from being able to modify
//   the buffer that is in the middle of being modified by replace command)
//   To that end, the buffer should be checked for "busy" before the shell attempts
//   to change the text it contains, or change its selection.
//   It should be set to busy whenever the shell is about to do something to the
//   buffer that could run a Tcl script.

static EDITOR_BUFFER
	*startBuffer;											// pointer to start of editor buffer list
static uint8_t
	taskData[TASK_DATA_BUFFER_SIZE];						// buffer used to read task data into

enum
{
	NO_TASK,
	NOT_FROM_FILE,
	DUPLICATE_NAME
};

static const char *errorDescriptions[]=
{
	"Buffer has no task",
	"Buffer is not linked to a file",
	"Duplicate buffer names are not allowed"
};

bool EditorBufferBusy(EDITOR_BUFFER *buffer)
// Check the given buffer's busy state.
// If busy, return true, else false.
{
	return(buffer->shellBusy);
}

void EditorBufferBusyClear(EDITOR_BUFFER *buffer)
// Clear the buffer's busy state.
{
	buffer->shellBusy=false;
}

void EditorBufferBusySet(EDITOR_BUFFER *buffer)
// Set the buffer's busy state.
{
	buffer->shellBusy=true;
}

EDITOR_BUFFER *EditorBufferGetNext(EDITOR_BUFFER *buffer)
// Return the next buffer, or NULL if none.
{
	return(buffer->nextBuffer);
}

EDITOR_BUFFER *EditorBufferGetFirst()
// Return the first buffer, or NULL if none.
{
	return(startBuffer);
}

EDITOR_BUFFER *EditorBufferLocate(const char *bufferName)
// Search through the list of buffers, and attempt to locate
// one with the given name.
// If one is found, return it, else return NULL.
{
	bool
		found;
	EDITOR_BUFFER
		*testBuffer;

	testBuffer=startBuffer;
	found=false;
	while(testBuffer&&!found)
	{
		if(strcmp(testBuffer->contentName,bufferName)==0)
		{
			found=true;
		}
		else
		{
			testBuffer=testBuffer->nextBuffer;
		}
	}
	return(testBuffer);
}

void EditorBufferHandleRelease(EDITOR_BUFFER_HANDLE *handle)
// Remove handle from the buffer it was linked to.
// NOTE: if handle is already released, this does nothing.
{
	EDITOR_BUFFER_HANDLE
		*previousHandle,
		*currentHandle;

	if(handle->buffer)			// make sure this handle is held
	{
		previousHandle=NULL;
		currentHandle=handle->buffer->bufferHandles;
		while(currentHandle&&currentHandle!=handle)
		{
			previousHandle=currentHandle;
			currentHandle=currentHandle->nextHandle;
		}
		if(currentHandle)
		{
			if(previousHandle)
			{
				previousHandle->nextHandle=currentHandle->nextHandle;
			}
			else
			{
				currentHandle->buffer->bufferHandles=currentHandle->nextHandle;
			}
		}
		handle->buffer=NULL;
	}
}

void EditorBufferHandleGrab(EDITOR_BUFFER_HANDLE *handle,EDITOR_BUFFER *buffer)
// Grab a handle to buffer.
{
	handle->buffer=buffer;
	handle->nextHandle=buffer->bufferHandles;
	buffer->bufferHandles=handle;
}

EDITOR_BUFFER *EditorBufferGetFromHandle(EDITOR_BUFFER_HANDLE *handle)
// Return the buffer that handle points to.
{
	return(handle->buffer);
}

void EditorBufferHandleInit(EDITOR_BUFFER_HANDLE *handle)
// Initialize a buffer handle so that it is in the released state.
{
	handle->buffer=NULL;
}

bool EditorBufferClear(EDITOR_BUFFER *buffer)
// Clear the buffer, wiping out any text that was previously there.
// Also kill any selection, reset the cursor, and set the "clean" point.
// If there is a problem, set the error, and return false.
{
	bool
		fail;

	fail=false;
	EditorReplaceStart(buffer);										// we are going to change text in a buffer
	fail=!EditorTextReplace(buffer,0,TextUniverseTotalBytes(buffer->textUniverse),NULL,0,false);	// remove all that was there
	EditorReplaceEnd(buffer);										// text change is complete
	EditorUndoClear(buffer);
	UndoCleanPointSet(buffer);
	return(!fail);
}

static bool LoadBuffer(EDITOR_BUFFER *buffer,const char *path)
// Load the buffer from path, wiping out any text that was previously there.
// Also kill any selection, reset the cursor, clear any marks, and set the buffer clean point.
// If there is a problem, set the error, and return false.
{
	bool
		fail;

	fail=false;
	EditorReplaceStart(buffer);										// we are going to change text in a buffer
	fail=!EditorFileReplace(buffer,0,TextUniverseTotalBytes(buffer->textUniverse),path);	// replace all that was there
	EditorReplaceEnd(buffer);										// text change is complete
	if(!fail)
	{
		if(EditorSelectionSet(buffer,buffer->selectionUniverse,0,0))	// move the cursor to the top after the change
		{
			EditorUndoClear(buffer);
			UndoCleanPointSet(buffer);
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

static bool SaveBuffer(EDITOR_BUFFER *buffer,const char *path,bool setClean)
// Save buffer to path.
// If setClean is true, then set the buffer as clean.
// If there is a problem, set the error, and return false.
{
	TEXT_STREAM
		stream;
	uint8_t
		*dataBuffer;
	EDITOR_FILE
		*file;
	uint64_t
		numWritten;
	uint64_t
		bytesLeft,
		bytesToWrite;
	bool
		fail;

	fail=false;

	TextStreamPositionSet(&stream,buffer->textUniverse,0);
	if((file=PlatformFileWriteOpen(path)))								// open the file to write to
	{
		if((dataBuffer=(uint8_t *)MNewPtr(LOAD_SAVE_BUFFER_SIZE)))		// create an intermediate buffer to speed the save
		{
			bytesLeft=TextUniverseTotalBytes(buffer->textUniverse);
			while(bytesLeft&&!fail)										// read until we reach the end of the file
			{
				if(bytesLeft>LOAD_SAVE_BUFFER_SIZE)						// see how many bytes to extract this time
				{
					bytesToWrite=LOAD_SAVE_BUFFER_SIZE;
				}
				else
				{
					bytesToWrite=bytesLeft;
				}
				TextStreamTextExtract(&stream,dataBuffer,bytesToWrite,NULL);	// read data from universe into buffer
				if(PlatformFileWrite(file,dataBuffer,bytesToWrite,&numWritten))	// write the buffer contents to the file
				{
					bytesLeft-=bytesToWrite;							// decrement bytes remaining to be written
				}
				else
				{
					fail=true;
				}
			}
			MDisposePtr(dataBuffer);
		}
		else
		{
			fail=true;
		}
		PlatformFileClose(file);
	}
	else
	{
		fail=true;
	}
	if(!fail&&setClean)
	{
		UndoCleanPointSet(buffer);				// make it clean again if asked to
	}
	return(!fail);
}

static void EditorBufferTaskDataInsert(EDITOR_BUFFER *buffer,uint8_t *data,uint64_t numBytes)
// Insert numBytes of data into buffer.
// Update any views of buffer that may need it.
// If there is a problem, ignore it.
{
	int64_t
		leftPixel;
	int64_t
		topLine;
	uint64_t
		numPixels;
	uint64_t
		numLines;
	EDITOR_VIEW
		*currentView;
	uint64_t
		startPosition,
		endPosition;
	int64_t
		startLine;

	// Run through the views, see which ones will need to be homed after the insert.
	// The ones to home, are those that had the cursor in view at the time of the insert.

	SelectionEndPositionsGet(buffer->selectionUniverse,&startPosition,&endPosition);
	currentView=buffer->firstView;					// get first view of this buffer (if any)
	while(currentView)								// walk through all views, see which will need to be homed
	{
		ViewOriginGet(currentView,&leftPixel,&topLine);
		PlatformViewActiveBoundsGet(currentView,&numPixels,&numLines);
		startLine=TextUniversePositionToLine(buffer->textUniverse,startPosition,NULL);
		currentView->wantHome=(startLine>=topLine&&startLine<topLine+(int64_t)numLines);	// see if the cursor is in the active part of the view
		currentView=currentView->nextBufferView;	// locate next view of this buffer
	}

	EditorAuxInsert(buffer,data,numBytes);			// drop the text into buffer

	SelectionEndPositionsGet(buffer->selectionUniverse,&startPosition,&endPosition);	// get selection position after the insertion
	// now home all of those views that need it
	currentView=buffer->firstView;					// get first view of this buffer (if any)
	while(currentView)								// walk through all views, update each one as needed
	{
		if(currentView->wantHome)
		{
			EditorViewToRangeHome(currentView,startPosition,endPosition,false,HT_NONE,HT_LENIENT);
		}
		currentView=currentView->nextBufferView;	// locate next view of this buffer
	}
}

void EditorBufferTaskDisconnect(EDITOR_BUFFER *buffer)
// If there's a task running on buffer, disconnect it.
{
	if(buffer->task)
	{
		BackgroundDestroy(buffer->taskBackgroundHandler);		// no longer need any background processing
		buffer->taskBackgroundHandler=NULL;
		PlatformTaskDisconnect(buffer->task);
		buffer->task=NULL;
		ViewsMetaChangeCallbacks(buffer);			// inform views that meta information about the buffer has changed (task status)
	}
}

static void EditorBufferTaskBackground(void *userData)
// Callback function that's enabled when a task is assigned to a
// buffer.
{
	EDITOR_BUFFER
		*buffer=(EDITOR_BUFFER *)userData;			// cast the user data into an editor buffer
	const char
		*terminationString;
	uint64_t
		bytesRead;

	if(PlatformTaskStatus(buffer->task,&terminationString))			// task still running normally?
	{
		if(PlatformTaskDataRead(buffer->task,taskData,sizeof(taskData),&bytesRead)&&bytesRead)	// got some data?
		{
			EditorBufferTaskDataInsert(buffer,taskData,bytesRead);	// drop data into the buffer
		}
	}
	else
	{
		while(PlatformTaskDataRead(buffer->task,taskData,sizeof(taskData),&bytesRead)&&bytesRead)	// scoop up any data left from the task
		{
			EditorBufferTaskDataInsert(buffer,taskData,bytesRead);	// drop data into the buffer
		}
		if(terminationString)										// see if task died under exceptional circumstances
		{
			bytesRead=snprintf((char *)taskData,sizeof(taskData),"%s\n",terminationString);
			EditorBufferTaskDataInsert(buffer,taskData,bytesRead);	// drop data into the buffer
		}
		EditorBufferTaskDisconnect(buffer);							// disconnect from the task now
	}
}

bool EditorBufferTaskEofWrite(EDITOR_BUFFER *buffer)
// Send an EOF to the task associated with buffer.
// If there's a problem, ErrorSet and return false.
{
	if(buffer->task)
	{
		if(PlatformTaskEofWrite(buffer->task))
		{
			return(true);
		}
	}
	else
	{
		ErrorSet(errorDescriptions[NO_TASK]);
	}
	return(false);
}

bool EditorBufferTaskDataWrite(EDITOR_BUFFER *buffer,const uint8_t *taskData,uint64_t taskDataLength)
// Send data to a running task in the given buffer.
// If there's a problem, ErrorSet and return false.
{
	if(buffer->task)
	{
		if(PlatformTaskDataWrite(buffer->task,taskData,taskDataLength))
		{
			return(true);
		}
	}
	else
	{
		ErrorSet(errorDescriptions[NO_TASK]);
	}
	return(false);
}

bool EditorBufferHasTask(EDITOR_BUFFER *buffer)
// Return true if buffer has a task associated with it.
{
	return(buffer->task?true:false);
}

bool EditorBufferTaskKill(EDITOR_BUFFER *buffer)
// Signal any task associated with buffer to die.
{
	if(buffer->task)
	{
		PlatformTaskKill(buffer->task);
		return(true);
	}
	else
	{
		ErrorSet(errorDescriptions[NO_TASK]);
	}
	return(false);
}

bool EditorBufferTaskSpawn(EDITOR_BUFFER *buffer,const char *commandString)
// Start a task given by commandString running in buffer.
// If there's a problem, ErrorSet and return false.
{
	if(!buffer->task)
	{
		if((buffer->taskBackgroundHandler=BackgroundCreate(EditorBufferTaskBackground,(void *)buffer)))
		{
			if((buffer->task=PlatformTaskSpawn((const uint8_t *)commandString)))
			{
				ViewsMetaChangeCallbacks(buffer);					// inform views that meta information about the buffer has changed (task status)
				return(true);
			}
			BackgroundDestroy(buffer->taskBackgroundHandler);
			buffer->taskBackgroundHandler=NULL;
		}
	}
	else
	{
		ErrorSet("Buffer already has a task");
	}
	return(false);
}

bool EditorBufferRevert(EDITOR_BUFFER *buffer)
// Revert buffer, no questions asked.
// If there is a problem, ErrorSet, and return false.
// NOTE: the window could be left with only part of the reverted text
// if there was a problem.
{
	if(buffer->fromFile)
	{
		if(LoadBuffer(buffer,buffer->contentName))
		{
			return(true);
		}
	}
	else
	{
		if(EditorBufferClear(buffer))
		{
			return(true);
		}
	}
	return(false);
}

bool EditorBufferSaveTo(EDITOR_BUFFER *buffer,const char *path)
// Save the text of buffer into path.
// NOTE: the text will be saved even if it is not dirty.
// If there is a problem, ErrorSet, and return false.
// NOTE: the buffer "clean" point is not moved by this call.
// NOTE: certain problems may cause the text to be only partially saved
// thus corrupting the file.
{
	return(SaveBuffer(buffer,path,false));
}

bool EditorBufferSaveAs(EDITOR_BUFFER *buffer,const char *path)
// Save the text of buffer into path. Change buffer's name
// to the absolute version of path, and update its window if it has one.
// NOTE: the text will be saved even if it is not dirty.
// If there is a problem, ErrorSet, and return false.
// NOTE: certain problems may cause the text to be only partially saved
// thus corrupting the file.
// NOTE: If after a save, the buffer's name becomes the same as that
// of another already existing buffer, the name change will fail, even though
// the contents of the buffer have been written into the file.
// This is a little ugly, but all editors have to cope with this problem somehow.
{
	EDITOR_BUFFER
		*otherBuffer;
	char
		*absolutePath;

	if(SaveBuffer(buffer,path,true))							// save the file using the regular path, mark it as clean
	{
		if((absolutePath=PlatformAbsolutePathCreate(path)))		// attempt to make an absolute path out of the one given (this could fail if someone deleted the file we just saved before we get here!)
		{
			if(!(otherBuffer=EditorBufferLocate(absolutePath))||(otherBuffer==buffer))	// see if there is another buffer down this path
			{
				MDisposePtr(buffer->contentName);				// wipe out the old content name
				buffer->contentName=absolutePath;				// point it at the new path
				buffer->fromFile=true;							// linked to file now
				EditorDocumentWindowTitleReassign(buffer);		// make new title to reflect new name
				return(true);
			}
			else
			{
				ErrorSet(errorDescriptions[DUPLICATE_NAME]);
			}
			MDisposePtr(absolutePath);
		}
	}
	return(false);
}

bool EditorBufferSave(EDITOR_BUFFER *buffer)
// Save the text of buffer into its file if it has one.
// If it does not have a file, this will fail.
{
	if(buffer->fromFile)
	{
		return(SaveBuffer(buffer,buffer->contentName,true));	// save buffer, mark it as clean
	}
	ErrorSet(errorDescriptions[NOT_FROM_FILE]);
	return(false);
}

static void EditorBufferDispose(EDITOR_BUFFER *buffer)
// Dispose of a buffer.
{
	EDITOR_BUFFER
		*previousBuffer,
		*testBuffer;
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER_HANDLE
		*handle;

	handle=buffer->bufferHandles;
	while(handle)											// for each handle to this buffer, unlink and clear it
	{
		handle->buffer=NULL;								// this buffer is gone...
		handle=handle->nextHandle;
	}

	VariableBindingTableUnInit(&(buffer->variableTable));	// remove variables hanging off this buffer
	if((window=PlatformDocumentWindowLocate(buffer)))
	{
		EditorDocumentWindowClose(window);					// close down the window if one exists
	}
	if(buffer->task)
	{
		EditorBufferTaskDisconnect(buffer);					// close down the connection to any task that may be attached to this buffer
	}
	if(buffer->syntaxInstance)								// if syntax instance has been assigned, remove it
	{
		SyntaxMapInstanceClose(buffer->syntaxInstance);
	}
	StyleUniverseClose(buffer->styleUniverse);
	while(buffer->marks)
	{
		EditorMarkClose(buffer,buffer->marks);				// get rid of all marks on this buffer
	}
	SelectionUniverseClose(buffer->auxSelectionUniverse);
	SelectionUniverseClose(buffer->mathSelectionUniverse);
	SelectionUniverseClose(buffer->selectionUniverse);
	UndoUniverseClose(buffer->undoUniverse);
	TextUniverseClose(buffer->textUniverse);
	MDisposePtr(buffer->contentName);

	previousBuffer=NULL;
	testBuffer=startBuffer;
	while(testBuffer&&testBuffer!=buffer)
	{
		previousBuffer=testBuffer;
		testBuffer=testBuffer->nextBuffer;
	}
	if(testBuffer==buffer)									// make sure we found it before trying to unlink
	{
		if(previousBuffer)
		{
			previousBuffer->nextBuffer=buffer->nextBuffer;
		}
		else
		{
			startBuffer=buffer->nextBuffer;
		}
	}

	MDisposePtr(buffer);
}

static EDITOR_BUFFER *EditorBufferCreate(const char *bufferName)
// Create a buffer (buffers contain both text and selection
// universes (which are created here)).
// If there is a problem, set the error, and return NULL.
{
	EDITOR_BUFFER
		*buffer;

	if((buffer=(EDITOR_BUFFER *)MNewPtr(sizeof(EDITOR_BUFFER))))
	{
		buffer->bufferHandles=NULL;									// not being held
		buffer->shellBusy=false;									// set this to a known value
		buffer->fromFile=false;

		if((buffer->contentName=(char *)MNewPtr(strlen(bufferName)+1)))
		{
			strcpy(buffer->contentName,bufferName);
			if((buffer->textUniverse=TextUniverseOpen()))			// create the text universe
			{
				if((buffer->undoUniverse=UndoUniverseOpen()))		// create the undo universe
				{
					if((buffer->selectionUniverse=SelectionUniverseOpen()))			// create the selection universe
					{
						if((buffer->mathSelectionUniverse=SelectionUniverseOpen()))		// create the math selection universe
						{
							if((buffer->auxSelectionUniverse=SelectionUniverseOpen()))	// create the aux selection universe (used for tasks, so they can write data into a place other than at the cursor)
							{
								buffer->marks=NULL;						// no marks on this buffer yet
								if((buffer->styleUniverse=StyleUniverseOpen()))			// create the style universe
								{
									buffer->syntaxInstance=NULL;		// no syntax highlight on this buffer yet
									buffer->taskBackgroundHandler=NULL;	// no task background handler running
									buffer->task=NULL;
									buffer->firstView=NULL;				// no views onto this buffer yet
									VariableBindingTableInit(&(buffer->variableTable));	// initialize the variable table

									buffer->replaceHaveRange=false;
									buffer->replaceStartOffset=0;
									buffer->replaceEndOffset=0;
									buffer->replaceNumBytes=0;

									buffer->anchorStartPosition=0;		// clear selection handling variables
									buffer->anchorEndPosition=0;
									buffer->anchorX=0;
									buffer->columnarLeftX=0;
									buffer->columnarRightX=0;
									buffer->haveStartX=false;
									buffer->haveEndX=false;
									buffer->desiredStartX=0;
									buffer->desiredEndX=0;
									buffer->haveCurrentEnd=false;
									buffer->currentIsStart=false;

									buffer->nextBuffer=startBuffer;
									startBuffer=buffer;
									return(buffer);
								}
								SelectionUniverseClose(buffer->auxSelectionUniverse);
							}
							SelectionUniverseClose(buffer->mathSelectionUniverse);
						}
						SelectionUniverseClose(buffer->selectionUniverse);
					}
					UndoUniverseClose(buffer->undoUniverse);
				}
				TextUniverseClose(buffer->textUniverse);
			}
			MDisposePtr(buffer->contentName);
		}
		MDisposePtr(buffer);
	}
	return(NULL);
}

EDITOR_BUFFER *EditorBufferNew(const char *bufferName)
// Create a new buffer with the given name.
// NOTE: if there is already a buffer with the given name, ErrorSet.
// If there is a problem, ErrorSet, return NULL.
{
	if(!EditorBufferLocate(bufferName))
	{
		return(EditorBufferCreate(bufferName));
	}
	else
	{
		ErrorSet(errorDescriptions[DUPLICATE_NAME]);
	}
	return(NULL);
}

void EditorBufferClose(EDITOR_BUFFER *buffer)
// Close the given editor buffer.
// If the buffer has a task, it will be disconnected.
// If the buffer has a window, it will be closed too.
// This will close the window without saving, even if the text has been modified.
{
	EditorBufferDispose(buffer);
}

EDITOR_BUFFER *EditorBufferOpen(const char *path)
// Open an editor buffer to the file at path.
// If there is a problem, set the error, and return NULL.
// NOTE: if path points to a buffer that is already open,
// then that buffer is simply returned.
{
	EDITOR_BUFFER
		*buffer;
	char
		*absolutePath;

	if((absolutePath=PlatformAbsolutePathCreate(path)))				// attempt to make an absolute path out of the one given
	{
		if((buffer=EditorBufferLocate(absolutePath)))
		{
			MDisposePtr(absolutePath);
			return(buffer);
		}
		else
		{
			if((buffer=EditorBufferCreate(absolutePath)))
			{
				if(LoadBuffer(buffer,absolutePath))
				{
					buffer->fromFile=true;
					MDisposePtr(absolutePath);
					return(buffer);
				}
				EditorBufferDispose(buffer);
			}
		}
		MDisposePtr(absolutePath);
	}
	return(NULL);
}

void EditorBufferUnInit()
// Undo whatever EditorBufferInit did.
{
	while(startBuffer)
	{
		if(startBuffer->bufferHandles)
		{
			fprintf(stderr,"Buffer dying with handles\n");	// this is a bug because no one should have a handle to a buffer at this moment
		}
		EditorBufferClose(startBuffer);						// get rid of any buffers left lying about
	}
}

bool EditorBufferInit()
// Initialize buffer handling.
// If there is a problem, ErrorSet, and return false.
{
	startBuffer=NULL;										// no start buffer as of yet
	return(true);
}
