// Key binding interface
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct keyBinding
{
	uint32_t
		modifiers,							// editor modifiers of the binding
		keyCode;							// key code of the binding
	struct keyBinding
		*previousBinding,					// pointer to previous binding in the global binding list
		*nextBinding,						// pointer to next binding in the global binding list (in user defined order)
		*nextHashBinding;					// pointer to next binding in local hash list
	char
		dataText[0];						// variable length array of characters of data
} KEY_BINDING;

static KEY_BINDING
	*keyBindingTable[256];					// array of heads of linked lists of key bindings (crude hash table to get some speed when locating a binding)
static KEY_BINDING
	*keyBindingListHead;					// head of list than runs through all entries

static KEY_BINDING **KeyBindHashListHeadGet(uint32_t keyCode)
// Perform hashing function on keyCode, return pointer to list head
// for bindings,
{
	return(&(keyBindingTable[keyCode&0xFF]));
}

static bool KeyBindPointersGet(uint32_t modifiers,uint32_t keyCode,KEY_BINDING ***hashListHead,KEY_BINDING **previousHashEntry,KEY_BINDING **entry)
// Locate a key binding in the binding table, return pointers to important entries.
{
	bool
		found;
	KEY_BINDING
		*currentEntry;

	*entry=NULL;													// keep gcc quiet
	currentEntry=*(*hashListHead=KeyBindHashListHeadGet(keyCode));	// point to the head of the linked list at this hash entry
	*previousHashEntry=NULL;
	found=false;
	while(currentEntry&&!found)
	{
		if((PlatformRemapEditorModifiers(currentEntry->modifiers)==modifiers)&&(currentEntry->keyCode==keyCode))
		{
			found=true;
			*entry=currentEntry;
		}
		else
		{
			*previousHashEntry=currentEntry;
			currentEntry=currentEntry->nextHashBinding;
		}
	}
	return(found);
}

char *KeyBindTextGet(KEY_BINDING *binding)
// Return a pointer to the key binding text for binding.
{
	return(binding->dataText);
}

void KeyBindEditorKeyGet(KEY_BINDING *binding,uint32_t *modifiers,uint32_t *keyCode)
// Return the key binding information.
{
	*modifiers=PlatformRemapEditorModifiers(binding->modifiers);
	*keyCode=binding->keyCode;
}

KEY_BINDING *KeyBindLocateNext(KEY_BINDING *currentBinding)
// Return the next key binding from the current one.
// If there is none, return NULL.
{
	return(currentBinding->nextBinding);
}

KEY_BINDING *KeyBindLocateFirst()
// Return the first key binding.
// If there is none, return NULL.
{
	return(keyBindingListHead);
}

KEY_BINDING *KeyBindLocate(uint32_t modifiers,uint32_t keyCode)
// Find a key binding that matches the keyCode, and modifierValue, and return it.
// If there is none, return NULL.
{
	KEY_BINDING
		**hashListHead,
		*previousHashEntry,
		*entry;

	if(KeyBindPointersGet(modifiers,keyCode,&hashListHead,&previousHashEntry,&entry))
	{
		return(entry);
	}
	return(NULL);
}

bool KeyBindDelete(uint32_t modifiers,uint32_t keyCode)
// Attempt to locate the given binding in the key bindings table, and
// remove it if it is found.
// If it is not located, return false.
{
	KEY_BINDING
		**hashListHead,
		*previousHashEntry,
		*entry;

	if(KeyBindPointersGet(modifiers,keyCode,&hashListHead,&previousHashEntry,&entry))
	{
		if(previousHashEntry)
		{
			previousHashEntry->nextHashBinding=entry->nextHashBinding;		// unlink the located entry
		}
		else
		{
			*hashListHead=entry->nextHashBinding;
		}
		if(entry->previousBinding)
		{
			if((entry->previousBinding->nextBinding=entry->nextBinding))
			{
				entry->nextBinding->previousBinding=entry->previousBinding;
			}
		}
		else
		{
			if((keyBindingListHead=entry->nextBinding))
			{
				keyBindingListHead->previousBinding=NULL;
			}
		}
		MDisposePtr(entry);
		return(true);
	}
	return(false);
}

bool KeyBindCreate(uint32_t modifiers,uint32_t keyCode,char *dataText)
// Search the key binding table for an entry that matches the given values, and delete it if
// it exists. Then add this entry to the table.
// If there is a problem, ErrorSet, and return false.
{
	KEY_BINDING
		**hashListHead,
		*newEntry;

	KeyBindDelete(modifiers,keyCode);							// try to delete one if it exists
	hashListHead=KeyBindHashListHeadGet(keyCode);				// point to head of list in hash table where this binding will go
	if((newEntry=(KEY_BINDING *)MNewPtr(sizeof(KEY_BINDING)+strlen(dataText)+1)))
	{
		newEntry->modifiers=modifiers;
		newEntry->keyCode=keyCode;
		newEntry->nextHashBinding=*hashListHead;
		strcpy(&(newEntry->dataText[0]),dataText);				// copy in the data
		*hashListHead=newEntry;									// link to head of hash list
		newEntry->previousBinding=NULL;
		if((newEntry->nextBinding=keyBindingListHead))			// point to the next entry
		{
			keyBindingListHead->previousBinding=newEntry;
		}
		keyBindingListHead=newEntry;
		return(true);
	}
	return(false);
}

// Strings that correspond to editor modifiers.
// NOTE: these must be in bit order of the modifiers themselves
// (lowest bit to highest bit).
static const char
	*modifierStringTable[]=
	{
		"Shift",
		"Ctrl",
		"Window",
		"Alt",
		"Cmd",
	};

bool KeyBindModifierNameToIndex(const char *name,uint32_t *modifierIndex)
// Look up name as the name of a modifier, if it is found, return true with
// modifierIndex filled in.
// If the name matches no modifier, ErrorSet, and return false.
{
	uint32_t
		i;
	bool
		result;

	result=false;
	for(i=0;(!result)&&(i<elementsof(modifierStringTable));i++)
	{
		if(strcasecmp(name,modifierStringTable[i])==0)		// match?
		{
			(*modifierIndex)=i;
			result=true;
		}
	}
	if(!result)
	{
		ErrorSet("Unmatched key modifier name: '%s'",name);
	}
	return(result);
}

uint32_t KeyBindModifierIndexToName(uint32_t modifierIndex,char *stringBuffer,uint32_t bufferLength)
// Fill stringBuffer (for up to bufferLength bytes) with the name of the
// modifier given by modifierIndex (which must be in range).
// Return the number of bytes (excluding 0 termination) written to stringBuffer.
{
	uint32_t
		result;

	result=snprintf(stringBuffer,bufferLength,"%s",modifierStringTable[modifierIndex]);
	if(bufferLength&&result>=bufferLength)		// if output filled the buffer, don't include 0 termination in the result
	{
		result=bufferLength-1;
	}
	return(result);
}

bool KeyBindNameToKey(const char *name,uint32_t *modifiers,uint32_t *keyCode)
// Given a string which is a complete key combination name (modifiers and key)
// convert it into the given values.
// If there's a problem, ErrorSet, and return false.
{
	uint32_t
		startIndex,
		currentIndex;
	bool
		gotName,
		result;
	char
		modifierName[256];
	uint32_t
		modifierNameLength;
	uint32_t
		modifierIndex;

	gotName=false;
	result=true;
	(*modifiers)=0;
	(*keyCode)=0;
	startIndex=0;
	while(result&&name[startIndex])
	{
		modifierNameLength=0;
		currentIndex=startIndex;
		while(name[currentIndex]&&(name[currentIndex]!='+'))	// move to the end, or the next separator
		{
			if(modifierNameLength<sizeof(modifierName)-1)		// copy modifier name to a buffer
			{
				modifierName[modifierNameLength]=name[currentIndex];
				modifierNameLength++;
			}
			currentIndex++;
		}
		modifierName[modifierNameLength]='\0';					// terminate the string
		if(name[currentIndex])									// at separator?
		{
			if((result=KeyBindModifierNameToIndex(modifierName,&modifierIndex)))
			{
				(*modifiers)|=(1<<modifierIndex);
			}
			startIndex=currentIndex+1;							// step over the separator
		}
		else						// at end
		{
			if((result=PlatformKeyNameToKeyCode(&name[startIndex],keyCode)))
			{
				gotName=true;
			}
			else
			{
				ErrorSet("Unmatched key name: '%s'",&name[startIndex]);
			}
			startIndex=currentIndex;	// at the end
		}
	}
	if(result&&!gotName)
	{
		ErrorSet("Empty key name");
		result=false;
	}
	return(result);
}

uint32_t KeyBindKeyToName(uint32_t modifiers,uint32_t keyCode,char *stringBuffer,uint32_t bufferLength)
// Given modifiers and a keyCode, render it's complete key combination name to
// a string.
// Return the number of bytes (excluding 0 termination) written to stringBuffer.
{
	uint32_t
		outputLength,
		printMaxLength,
		printLength;
	uint32_t
		i;

	outputLength=0;

	// Drop modifiers into the string
	i=elementsof(modifierStringTable);
	while(i)
	{
		i--;
		if(modifiers&(1<<i))
		{
			printMaxLength=bufferLength-outputLength;
			printLength=snprintf(&stringBuffer[outputLength],printMaxLength,"%s+",modifierStringTable[i]);
			if(printLength>printMaxLength)
			{
				printLength=printMaxLength;
			}
			outputLength+=printLength;
		}
	}

	// Drop keyCode into the string
	printMaxLength=bufferLength-outputLength;
	printLength=snprintf(&stringBuffer[outputLength],printMaxLength,"%s",PlatformKeyCodeToKeyName(keyCode));
	if(printLength>printMaxLength)
	{
		printLength=printMaxLength;
	}
	outputLength+=printLength;
	if(bufferLength&&outputLength>=bufferLength)		// if filled the buffer, don't include 0 termination in outputLength
	{
		outputLength=bufferLength-1;					// backup from termination
	}
	return(outputLength);
}

void KeyBindUnInit()
// Remove all key bindings from the bindings table.
{
	int
		i;
	KEY_BINDING
		*currentEntry,
		*nextEntry;

	for(i=0;i<256;i++)
	{
		currentEntry=keyBindingTable[i];
		while(currentEntry)
		{
			nextEntry=currentEntry->nextHashBinding;
			MDisposePtr(currentEntry);
			currentEntry=nextEntry;
		}
		keyBindingTable[i]=NULL;
	}
	keyBindingListHead=NULL;
}

bool KeyBindInit()
// Clear the key binding table so that there are no bindings present.
// If there is a problem, ErrorSet, and return false.
{
	int
		i;

	for(i=0;i<256;i++)
	{
		keyBindingTable[i]=NULL;
	}
	keyBindingListHead=NULL;
	return(true);
}
