// Named list functions
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// Create named lists of data structures.
// These can be useful when interfacing with Tcl.

#include	"includes.h"

typedef struct namedListItem
{
	void
		*data;						// data being pointed to by this element
	struct namedListItem
		*prev,
		*next;						// used to link the list together
	char
		name[0];					// variable length array, name of this element
} NAMED_LIST_ITEM;

typedef struct namedList
{
	NAMED_LIST_ITEM
		*first,
		*last;
} NAMED_LIST;

void *NamedListItemDataGet(NAMED_LIST_ITEM *namedListItem)
// Return a pointer to the data of the given item.
{
	return(namedListItem->data);
}

void NamedListItemDataSet(NAMED_LIST_ITEM *namedListItem,void *data)
// Update the pointer to data for the given item.
{
	namedListItem->data=data;
}

const char *NamedListItemNameGet(NAMED_LIST_ITEM *namedListItem)
// Return a pointer to the name of the given item.
{
	return(namedListItem->name);
}

NAMED_LIST_ITEM *NamedListItemLocateNextName(NAMED_LIST_ITEM *listItem,const char *name)
// Locate the next item on the given list with the given name. Comparisons
// are case sensitive. The search starts at listItem.
// If no matching item is located, NULL is returned.
{
	while(listItem)
	{
		if(strcmp(listItem->name,name)==0)
		{
			return(listItem);
		}
		listItem=listItem->next;
	}
	return(NULL);
}

NAMED_LIST_ITEM *NamedListItemLocateName(NAMED_LIST *namedList,const char *name)
// Locate the first item on the given list with the given name. Comparisons
// are case sensitive. If no matching item is located, NULL is returned.
{
	return(NamedListItemLocateNextName(namedList->first,name));
}

NAMED_LIST_ITEM *NamedListItemLocateNext(NAMED_LIST_ITEM *namedListItem)
// Return the next item on the given list. If there is no next item,
// return NULL.
{
	return(namedListItem->next);
}

NAMED_LIST_ITEM *NamedListItemLocateFirst(NAMED_LIST *namedList)
// Return the first item on the given list.
{
	return(namedList->first);
}

void NamedListItemDestroy(NAMED_LIST *namedList,NAMED_LIST_ITEM *namedListItem)
// Unlink the given namedListItem from the given namedList (of which it
// must be a part), and then delete the item.
// NOTE: the caller is responsible for freeing any data which was linked
// to the item (usually done before this is called).
{
	if(namedListItem->prev)
	{
		namedListItem->prev->next=namedListItem->next;
	}
	else
	{
		namedList->first=namedListItem->next;
	}
	if(namedListItem->next)
	{
		namedListItem->next->prev=namedListItem->prev;
	}
	else
	{
		namedList->last=namedListItem->prev;
	}
	MDisposePtr(namedListItem);
}

NAMED_LIST_ITEM *NamedListItemCreate(NAMED_LIST *namedList,const char *name)
// Create a list item with the given name, link it to the end of
// the list, and return it.
// The returned item will have a NULL data field.
// NOTE: this does not check to see if the passed name is a duplicate.
// The caller must do this if he cares not to have duplicate items on
// the list.
// If the item creation fails, ErrorSet is called and NULL is returned.
{
	NAMED_LIST_ITEM
		*namedListItem;
	int
		nameLength;

	nameLength=strlen(name);
	if((namedListItem=(NAMED_LIST_ITEM *)MNewPtr(sizeof(NAMED_LIST_ITEM)+nameLength+1)))
	{
		namedListItem->data=NULL;			// no data pointed to by the item
		strcpy(namedListItem->name,name);	// copy over the name
		namedListItem->next=NULL;
		if((namedListItem->prev=namedList->last))
		{
			namedListItem->prev->next=namedListItem;	// point previous last item to this one
		}
		else	// namedList was empty, so this is first item too
		{
			namedList->first=namedListItem;
		}
		namedList->last=namedListItem;
	}
	return(namedListItem);
}

void NamedListDestroy(NAMED_LIST *namedList)
// Destroy an *empty* list of named items.
// NOTE: the caller must have deleted up all the items on the list
// before this is called.
{
	MDisposePtr(namedList);
}

NAMED_LIST *NamedListCreate()
// Create an empty list of named items.
// If there's a problem, ErrorSet and return NULL.
{
	NAMED_LIST
		*namedList;

	if((namedList=(NAMED_LIST *)MNewPtr(sizeof(NAMED_LIST))))
	{
		namedList->first=NULL;		// no items on the list
		namedList->last=NULL;
	}
	return(namedList);
}
