// Dialog push button handler
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	MAX_BUTTON_NAME		256									// maximum length of button name (including terminator)

typedef struct button
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;										// function which is used to position this item in the dialog
	EDITOR_RECT
		rect;
	char
		buttonName[MAX_BUTTON_NAME];
	const GUI_IMAGE
		*image;													// image to draw for the button
	EDITOR_COLOR
		activeBackgroundColor,									// colors needed for button
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		pressedBackgroundColor,
		pressedForegroundColor,
		shadowColor;
	EDITOR_FONT
		*font;													// info about currently selected font
	bool
		highlight;												// used during tracking to remember if highlighted
	bool
		hasKey;													// true if this button has a key code associated with it
	uint32_t
		keyCode;												// equivalent key code for this button
	bool
		*pressedState;											// boolean set to true if button pressed
	buttonPressedFunction
		*pressedFunction;										// routine called when the button has been pressed
} BUTTON;

static bool ButtonPointInside(BUTTON *button,int32_t x,int32_t y)
// Return true if parent window relative x/y are in button,
// false if not.
{
	return(PointInRect(x,y,&button->rect));
}

static void ButtonInvalidate(BUTTON *button)
// Invalidate the button within its parent window.
{
	WindowRectInvalidate(DialogWindowGet(button->dialogItem.dialog),&button->rect);
}

static void ButtonRectangleSet(BUTTON *button)
// Work out the rectangle given the current window, and the image used for the
// button.
{
	if(button->positionFunction)	// see if non-null resize function specified
	{
		button->positionFunction(&button->dialogItem,&button->rect);	// get new rectangle
		ButtonInvalidate(button);									// invalidate it
	}
}

static void ButtonDraw(BUTTON *button,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Redraw button into imageBuffer.
{
	uint32_t
		ascent,
		descent;
	uint32_t
		stringLength,
		stringWidth;
	int
		stringX,
		stringY;
	EDITOR_COLOR
		drawColor;

	if(button->rect.w&&button->rect.h)
	{
		if(XRectInRegion(invalidRegion,button->rect.x,button->rect.y,button->rect.w,button->rect.h)!=RectangleOut)
		{
			if(button->highlight)
			{
				ImageDrawMask(button->image,button->pressedBackgroundColor,imageBuffer,button->rect.x+2,button->rect.y+2);
			}
			else
			{
				ImageDrawMask(button->image,button->shadowColor,imageBuffer,button->rect.x+2,button->rect.y+2);
				ImageDrawMask(button->image,button->activeBackgroundColor,imageBuffer,button->rect.x,button->rect.y);
			}

			FontHeightsGet(button->font,&ascent,&descent);

			stringLength=strlen(&(button->buttonName[0]));
			stringWidth=FontTextUtf8Width(button->font,&(button->buttonName[0]),stringLength);
			stringX=button->rect.x+(button->rect.w/2)-(stringWidth/2);
			stringY=button->rect.y+(button->rect.h/2)-(ascent+descent)/2+ascent;

			if(button->highlight)
			{
				stringX+=2;
				stringY+=2;
			}

			drawColor=button->activeForegroundColor;
			if(button->highlight)
			{
				drawColor=button->pressedForegroundColor;
			}
			FontTextUtf8Draw(button->font,imageBuffer,drawColor,stringX,stringY,&(button->buttonName[0]),stringLength);
		}
	}
}

static void ButtonPressed(BUTTON *button)
// Call this and button will behave as if it has been pressed.
{
	button->highlight=true;
	ButtonInvalidate(button);
	EditorWindowsUpdate();				// force an update so the button is visibly pressed while being processed (in case the process takes a long time w/o processing events)
	if(button->pressedState)
	{
		(*button->pressedState)=true;
	}
	if(button->pressedFunction)
	{
		button->pressedFunction(&button->dialogItem);
	}
	button->highlight=false;
	ButtonInvalidate(button);
}

static bool ButtonTrack(BUTTON *button,XEvent *event)
// See if the point given by event falls over button.
// If so, track it, and return true.
// If not, return false.
{
	bool
		onButton,
		wasOnButton;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	if((event->xbutton.button==Button1)&&ButtonPointInside(button,event->xbutton.x,event->xbutton.y))
	{
		button->highlight=true;
		ButtonInvalidate(button);
		wasOnButton=onButton=true;
		rootX=event->xbutton.x_root;
		rootY=event->xbutton.y_root;
		windowX=event->xbutton.x;
		windowY=event->xbutton.y;
		while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
		{
			onButton=ButtonPointInside(button,windowX,windowY);
			if(wasOnButton!=onButton)
			{
				button->highlight=onButton;
				ButtonInvalidate(button);
				wasOnButton=onButton;
			}
		}
		if(onButton)
		{
			ButtonPressed(button);
		}
		button->highlight=false;
		ButtonInvalidate(button);
		return(true);
	}
	return(false);
}

static bool ButtonKeyOverride(BUTTON *button,EDITOR_KEY *editorKey)
// This is called for buttons that may wish to grab keys before they go
// into the default focus of a dialog.
// true is returned if the button wanted the key.
{
	if(button->hasKey)
	{
		if(editorKey->keyCode==button->keyCode)
		{
			ButtonPressed(button);
			return(true);
		}
	}
	return(false);
}

// dialog interface routines

static void ButtonItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the button pointed to by item into imageBuffer.
{
	ButtonDraw((BUTTON *)item,imageBuffer,invalidRegion);
}

static void ButtonItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	BUTTON
		*button;

	button=(BUTTON *)item;
	ButtonInvalidate(button);				// invalidate where we're at currently
	ButtonRectangleSet(button);
}

void ButtonItemPress(DIALOG_ITEM *item)
// Do whatever would be done if the user pressed item.
{
	ButtonPressed((BUTTON *)item);
}

void ButtonItemDimensions(DIALOG_ITEM *item,uint32_t *width,uint32_t *height)
// Return the width and height of the button.
{
	BUTTON
		*button;

	button=(BUTTON *)item;
	*width=button->image->width+2;			// add 2 for the shadow
	*height=button->image->height+2;
}

static bool ButtonItemTrack(DIALOG_ITEM *item,XEvent *event)
// Track the button pointed to by item.
{
	return(ButtonTrack((BUTTON *)item,event));
}

static bool ButtonItemEarlyKey(DIALOG_ITEM *item,EDITOR_KEY *editorKey)
// See if this button wants the given editorKey, if so, return true.
{
	return(ButtonKeyOverride((BUTTON *)item,editorKey));
}

static void ButtonDispose(BUTTON *button)
// Unlink button from its parent window, and delete it.
// Invalidate the area of the window where button was.
{
	ButtonInvalidate(button);					// invalidate where it is now
	PlatformFontFree(button->font);
	MDisposePtr(button);
}

static void ButtonItemDispose(DIALOG_ITEM *item)
// Dispose of the local button structure pointed to by item.
{
	ButtonDispose((BUTTON *)item);
}

static BUTTON *ButtonCreate(DIALOG *dialog,BUTTON_DESCRIPTOR *descriptor)
// Create a button in dialog, invalidate the area of dialog
// where the button is to be placed.
// Return a pointer to it if successful.
// ErrorSet, and return NULL if not.
{
	BUTTON
		*button;

	if((button=(BUTTON *)MNewPtr(sizeof(BUTTON))))		// create button data structure
	{
		button->dialogItem.dialog=dialog;							// point back at the dialog
		button->dialogItem.disposeFunction=ButtonItemDispose;
		button->dialogItem.drawFunction=ButtonItemDraw;
		button->dialogItem.resizeFunction=ButtonItemResize;
		button->dialogItem.trackFunction=ButtonItemTrack;
		button->dialogItem.earlyKeyFunction=ButtonItemEarlyKey;
		button->dialogItem.wantFocus=false;
		button->dialogItem.focusChangeFunction=NULL;
		button->dialogItem.focusKeyFunction=NULL;
		button->dialogItem.focusPeriodicFunction=NULL;

		button->positionFunction=descriptor->positionFunction;
		strncpy(&(button->buttonName[0]),descriptor->buttonName,MAX_BUTTON_NAME);
		button->buttonName[MAX_BUTTON_NAME-1]='\0';		// terminate the string as needed

		button->activeBackgroundColor=descriptor->activeBackgroundColor;		// colors needed for button
		button->activeForegroundColor=descriptor->activeForegroundColor;
		button->inactiveBackgroundColor=descriptor->inactiveBackgroundColor;
		button->inactiveForegroundColor=descriptor->inactiveForegroundColor;
		button->pressedBackgroundColor=descriptor->pressedBackgroundColor;
		button->pressedForegroundColor=descriptor->pressedForegroundColor;
		button->shadowColor=descriptor->shadowColor;

		button->image=descriptor->image;				// copy the button image
		button->highlight=false;						// not highlighted
		button->hasKey=descriptor->hasKey;				// true if this button has a key code associated with it
		button->keyCode=descriptor->keyCode;			// equivalent key code for this button
		button->pressedState=descriptor->pressedState;	// points to a bool (or to NULL) that is set to true if this button has been pressed
		button->pressedFunction=descriptor->pressedFunction;	// copy the procedure to be called when the button is pressed
		if((button->font=PlatformFontLoad(descriptor->fontName)))
		{
			ButtonRectangleSet(button);
			return(button);
		}
		MDisposePtr(button);
	}
	return(NULL);
}

DIALOG_ITEM *ButtonItemCreate(DIALOG *dialog,void *descriptor)
// Create the button, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)ButtonCreate(dialog,(BUTTON_DESCRIPTOR *)descriptor));
}
