// Dialog check box handler
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	MAX_CHECK_BOX_NAME		256								// maximum length of check box name (including terminator)

typedef struct checkBox
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;										// function which is used to position this item in the dialog
	EDITOR_RECT
		rect;
	EDITOR_COLOR
		activeBackgroundColor,									// background when active
		activeForegroundColor,									// foreground when active
		inactiveBackgroundColor,								// background when inactive
		inactiveForegroundColor;								// foreground when inactive
	EDITOR_FONT
		*font;													// info about currently selected font
	char
		checkBoxName[MAX_CHECK_BOX_NAME];
	bool
		active;													// tells if the checkbox is "active" (can be clicked on, and drawn active)
	bool
		highlight;												// used during tracking to remember if highlighted
	bool
		hasKey;													// true if this check box has a key code associated with it
	uint32_t
		keyCode;												// equivalent key code for this button
	bool
		*pressedState;											// boolean set to true if check box pressed
	checkBoxPressedFunction
		*pressedFunction;										// routine called when the check box has been pressed (checked, or unchecked) (if NULL, won't be called)
} CHECK_BOX;

static bool CheckBoxPointInside(CHECK_BOX *checkBox,int32_t x,int32_t y)
// Return true if parent window relative x/y are in checkBox
// false if not.
{
	return(PointInRect(x,y,&checkBox->rect));
}

static void CheckBoxInvalidate(CHECK_BOX *checkBox)
// Invalidate the check box within its parent window.
{
	WindowRectInvalidate(DialogWindowGet(checkBox->dialogItem.dialog),&checkBox->rect);
}

static void CheckBoxRectangleSet(CHECK_BOX *checkBox)
// Work out the rectangle given the current window.
{
	if(checkBox->positionFunction)	// see if non-null resize function specified
	{
		checkBox->positionFunction(&checkBox->dialogItem,&checkBox->rect);	// get new rectangle
		CheckBoxInvalidate(checkBox);							// invalidate it
	}
}

static void CheckBoxDraw(CHECK_BOX *checkBox,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Redraw checkBox into imageBuffer.
{
	uint32_t
		ascent,
		descent;
	int32_t
		stringX,
		stringY;
	int32_t
		offsetX,
		offsetY;
	EDITOR_COLOR
		color;

	if(XRectInRegion(invalidRegion,checkBox->rect.x,checkBox->rect.y,checkBox->rect.w,checkBox->rect.h)!=RectangleOut)
	{
		ImageClipRegionRectSet(imageBuffer,&checkBox->rect);			// don't draw outside the checkbox area

		color=checkBox->inactiveBackgroundColor;
		if(checkBox->active)
		{
			color=checkBox->activeBackgroundColor;
		}

		ImageFill(imageBuffer,color,checkBox->rect.x,checkBox->rect.y,checkBox->rect.w,checkBox->rect.h);

		FontHeightsGet(checkBox->font,&ascent,&descent);

		offsetX=0;
		offsetY=0;
		if(checkBox->highlight)
		{
			offsetX=1;
			offsetY=1;
		}

		color=checkBox->inactiveForegroundColor;
		if(checkBox->active)
		{
			color=checkBox->activeForegroundColor;
		}

		if(checkBox->pressedState&&(*checkBox->pressedState))
		{
			ImageDrawMask(checkedBoxImage,color,imageBuffer,checkBox->rect.x+2+offsetX,checkBox->rect.y+(checkBox->rect.h-checkedBoxImage->height)/2+offsetY);
			stringX=checkBox->rect.x+checkedBoxImage->width+4;
		}
		else
		{
			ImageDrawMask(checkBoxImage,color,imageBuffer,checkBox->rect.x+2+offsetX,checkBox->rect.y+(checkBox->rect.h-checkBoxImage->height)/2+offsetY);
			stringX=checkBox->rect.x+checkBoxImage->width+4;
		}

		stringY=checkBox->rect.y+(checkBox->rect.h/2)-(ascent+descent)/2+ascent;

		FontTextUtf8Draw(checkBox->font,imageBuffer,color,stringX,stringY,&(checkBox->checkBoxName[0]),strlen(&(checkBox->checkBoxName[0])));

		ShadowRectangleOutline(imageBuffer,&checkBox->rect,BLACK,BLACK,1);
		ImageClipRegionSet(imageBuffer,None);							// unset clip area
	}
}

static bool CheckBoxTrack(CHECK_BOX *checkBox,XEvent *event)
// See if the point given by event falls over checkBox.
// If so, track it, and return true.
// If not, return false.
{
	bool
		onCheckBox,
		wasOnCheckBox;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	if((event->xbutton.button==Button1)&&CheckBoxPointInside(checkBox,event->xbutton.x,event->xbutton.y))
	{
		if(checkBox->active)
		{
			checkBox->highlight=true;
			CheckBoxInvalidate(checkBox);
			wasOnCheckBox=onCheckBox=true;
			rootX=event->xbutton.x_root;
			rootY=event->xbutton.y_root;
			windowX=event->xbutton.x;
			windowY=event->xbutton.y;
			while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
			{
				onCheckBox=CheckBoxPointInside(checkBox,windowX,windowY);
				if(wasOnCheckBox!=onCheckBox)
				{
					checkBox->highlight=onCheckBox;
					CheckBoxInvalidate(checkBox);
					wasOnCheckBox=onCheckBox;
				}
			}
			if(onCheckBox)
			{
				if(checkBox->pressedState)
				{
					(*checkBox->pressedState)=!(*checkBox->pressedState);	// toggle the state
				}
				if(checkBox->pressedFunction)
				{
					checkBox->pressedFunction(&checkBox->dialogItem);	// make callback if needed
				}
			}
			checkBox->highlight=false;
			CheckBoxInvalidate(checkBox);
			return(true);
		}
	}
	return(false);
}

static bool CheckBoxKeyOverride(CHECK_BOX *checkBox,EDITOR_KEY *editorKey)
// This is called for check boxes that may wish to grab keys before they go
// into the default focus of a dialog.
// true is returned if the check box wanted the key.
{
	if(checkBox->hasKey)
	{
		if(editorKey->keyCode==checkBox->keyCode)
		{
			checkBox->highlight=true;
			CheckBoxInvalidate(checkBox);
			if(checkBox->pressedState)
			{
				(*checkBox->pressedState)=true;
			}
			if(checkBox->pressedFunction)
			{
				checkBox->pressedFunction(&checkBox->dialogItem);
			}
			checkBox->highlight=false;
			CheckBoxInvalidate(checkBox);
			return(true);
		}
	}
	return(false);
}

// dialog interface routines

static void CheckBoxItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the check box pointed to by item.
{
	CheckBoxDraw((CHECK_BOX *)item,imageBuffer,invalidRegion);
}

static void CheckBoxItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	CHECK_BOX
		*checkBox;

	checkBox=(CHECK_BOX *)item;
	CheckBoxInvalidate(checkBox);					// invalidate where we're at currently
	CheckBoxRectangleSet(checkBox);
}

static bool CheckBoxItemTrack(DIALOG_ITEM *item,XEvent *event)
// Track the check box pointed to by item.
{
	return(CheckBoxTrack((CHECK_BOX *)item,event));
}

static bool CheckBoxItemEarlyKey(DIALOG_ITEM *item,EDITOR_KEY *editorKey)
// See if this check box wants the key given by event, if so, return true.
{
	return(CheckBoxKeyOverride((CHECK_BOX *)item,editorKey));
}

bool CheckBoxItemCheckedGet(DIALOG_ITEM *item)
// Return true of the given checkbox is checked.
{
	CHECK_BOX
		*checkBox;

	checkBox=(CHECK_BOX *)item;
	return(*checkBox->pressedState);
}

void CheckBoxItemCheckedSet(DIALOG_ITEM *item,bool checked)
// Set the checked status of the given checkbox.
{
	CHECK_BOX
		*checkBox;

	checkBox=(CHECK_BOX *)item;
	(*checkBox->pressedState)=checked;
	CheckBoxInvalidate(checkBox);
}

bool CheckBoxItemActiveGet(DIALOG_ITEM *item)
// Return true of the given checkbox is active.
{
	CHECK_BOX
		*checkBox;

	checkBox=(CHECK_BOX *)item;
	return(checkBox->active);
}

void CheckBoxItemActiveSet(DIALOG_ITEM *item,bool active)
// Adjust the active state of the checkbox.
{
	CHECK_BOX
		*checkBox;

	checkBox=(CHECK_BOX *)item;
	checkBox->active=active;
	CheckBoxInvalidate(checkBox);
}

static void CheckBoxDispose(CHECK_BOX *checkBox)
// Unlink checkBox from its parent window, and delete it.
// Invalidate the area of the window where checkBox was.
{
	CheckBoxInvalidate(checkBox);					// invalidate where it is now
	PlatformFontFree(checkBox->font);
	MDisposePtr(checkBox);
}

static void CheckBoxItemDispose(DIALOG_ITEM *item)
// Dispose of the local check box structure pointed to by item.
{
	CheckBoxDispose((CHECK_BOX *)item);
}

static CHECK_BOX *CheckBoxCreate(DIALOG *dialog,CHECK_BOX_DESCRIPTOR *descriptor)
// Create a check box in dialog, invalidate the area of dialog
// where the check box is to be placed.
// Return a pointer to it if successful.
// ErrorSet, and return NULL if not.
{
	CHECK_BOX
		*checkBox;

	if((checkBox=(CHECK_BOX *)MNewPtr(sizeof(CHECK_BOX))))	// create check box data structure
	{
		checkBox->dialogItem.dialog=dialog;							// point back at the dialog
		checkBox->dialogItem.disposeFunction=CheckBoxItemDispose;
		checkBox->dialogItem.drawFunction=CheckBoxItemDraw;
		checkBox->dialogItem.resizeFunction=CheckBoxItemResize;
		checkBox->dialogItem.trackFunction=CheckBoxItemTrack;
		checkBox->dialogItem.earlyKeyFunction=CheckBoxItemEarlyKey;
		checkBox->dialogItem.wantFocus=false;
		checkBox->dialogItem.focusChangeFunction=NULL;
		checkBox->dialogItem.focusKeyFunction=NULL;
		checkBox->dialogItem.focusPeriodicFunction=NULL;

		checkBox->positionFunction=descriptor->positionFunction;
		strncpy(&(checkBox->checkBoxName[0]),descriptor->checkBoxName,MAX_CHECK_BOX_NAME);
		checkBox->checkBoxName[MAX_CHECK_BOX_NAME-1]='\0';			// terminate the string as needed

		checkBox->activeBackgroundColor=descriptor->activeBackgroundColor;
		checkBox->activeForegroundColor=descriptor->activeForegroundColor;
		checkBox->inactiveBackgroundColor=descriptor->inactiveBackgroundColor;
		checkBox->inactiveForegroundColor=descriptor->inactiveForegroundColor;

		checkBox->active=descriptor->active;						// initial active state of the checkbox
		checkBox->highlight=false;									// not highlighted
		checkBox->hasKey=descriptor->hasKey;						// true if this check box has a key code associated with it
		checkBox->keyCode=descriptor->keyCode;						// equivalent key code for this check box
		checkBox->pressedState=descriptor->pressedState;			// points to a bool (or to NULL) that is set to true if this check box has been pressed
		checkBox->pressedFunction=descriptor->pressedFunction;		// copy the procedure to be called when the check box is checked/unchecked
		if((checkBox->font=PlatformFontLoad(descriptor->fontName)))
		{
			CheckBoxRectangleSet(checkBox);
			return(checkBox);
		}
		MDisposePtr(checkBox);
	}
	return(NULL);
}

DIALOG_ITEM *CheckBoxItemCreate(DIALOG *dialog,void *descriptor)
// Create the check box, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)CheckBoxCreate(dialog,(CHECK_BOX_DESCRIPTOR *)descriptor));
}
