// Dialog window handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	DEFAULT_STATIC_TEXT_FONT		"dejavu sans condensed:size=12"
#define	DEFAULT_TEXT_BOX_FONT			"liberation mono:size=9"
#define	DEFAULT_LIST_BOX_FONT			"dejavu sans condensed:size=10"
#define	DEFAULT_CHECK_BOX_FONT			"dejavu sans condensed:size=11:bold"
#define	DEFAULT_POPUP_FONT				"dejavu sans condensed:size=11:bold"
#define	DEFAULT_BUTTON_FONT				"dejavu sans condensed:size=12"

#define	FIND_COLOR_SCHEME_NAME			"_dialogFind"						// use this color scheme for the "find" view in the search dialog
#define	REPLACE_COLOR_SCHEME_NAME		"_dialogReplace"					// use this color scheme for the "replace" view in the search dialog
#define	SIMPLE_TEXT_COLOR_SCHEME_NAME	"_dialogSimpleText"					// use this color scheme for the view in the simple text dialog
#define	FILE_TEXT_COLOR_SCHEME_NAME		"_dialogFileName"					// use this color scheme for the view in the file dialog

typedef struct dialog
{
	EDITOR_WINDOW
		*parentWindow;		// the editor window that this dialog is in
	bool
		dialogComplete;		// set to true by dialog procs to tell when dialog is finished, and should be closed
	bool
		dialogAbort;		// set to true if the user closes the dialog by clicking on its close box, or hitting escape
	DIALOG_ITEM
		*firstItem,			// pointer to first item in the dialog (or NULL)
		*lastItem;			// pointer to last item in the dialog (or NULL)
	DIALOG_ITEM
		*focusItem;			// pointer to item with focus (or NULL)
	void
		*dialogLocal;		// place for local dialog routines to hang any data they need
} DIALOG;

static EDITOR_COLOR
	dialogBackgroundColor,					// configurable colors for all the various dialog pieces
	dialogStaticTextForegroundColor,
	dialogTextBoxActiveBorderColor,
	dialogTextBoxInactiveBorderColor,
	dialogListBoxActiveBackgroundColor,
	dialogListBoxActiveForegroundColor,
	dialogListBoxActiveBorderColor,
	dialogListBoxInactiveBackgroundColor,
	dialogListBoxInactiveForegroundColor,
	dialogListBoxInactiveBorderColor,
	dialogCheckBoxActiveBackgroundColor,
	dialogCheckBoxActiveForegroundColor,
	dialogCheckBoxInactiveBackgroundColor,
	dialogCheckBoxInactiveForegroundColor,
	dialogPopupActiveBackgroundColor,
	dialogPopupActiveForegroundColor,
	dialogPopupInactiveBackgroundColor,
	dialogPopupInactiveForegroundColor,
	dialogPopupBorderColor,
	dialogButtonActiveBackgroundColor,
	dialogButtonActiveForegroundColor,
	dialogButtonInactiveBackgroundColor,
	dialogButtonInactiveForegroundColor,
	dialogButtonPressedBackgroundColor,
	dialogButtonPressedForegroundColor,
	dialogButtonShadowColor;

static char
	dialogStaticTextFontName[MAX_FONT_NAME_LENGTH],
	dialogTextBoxFontName[MAX_FONT_NAME_LENGTH],
	dialogListBoxFontName[MAX_FONT_NAME_LENGTH],
	dialogCheckBoxFontName[MAX_FONT_NAME_LENGTH],
	dialogPopupFontName[MAX_FONT_NAME_LENGTH],
	dialogButtonFontName[MAX_FONT_NAME_LENGTH];

static int32_t
	dialogCenterX,		// center X and Y coordinates of dialog windows
	dialogCenterY;

// ----------------------------------------------------------------------------

static void DialogFocusChange(DIALOG *dialog,DIALOG_ITEM *newItem)
// Change the focus of dialog to newItem.
// If the focus is already at new item, do nothing.
// NOTE: this is DIALOG ITEM focus, not X window focus.
{
	DIALOG_ITEM
		*oldItem;

	oldItem=dialog->focusItem;
	if(oldItem!=newItem)							// make sure they are not the same, or both NULL
	{
		dialog->focusItem=newItem;
		if(oldItem)
		{
			if(oldItem->focusChangeFunction)
			{
				oldItem->focusChangeFunction(oldItem);	// tell him about it
			}
		}
		if(newItem)
		{
			if(newItem->focusChangeFunction)
			{
				newItem->focusChangeFunction(newItem);	// tell him about it
			}
		}
	}
}

static DIALOG_ITEM *DialogFocusItemLocateForward(DIALOG *dialog)
// Look forward from the dialog item which currently has the focus
// and return the next item which should get focus.
// If no items require focus, return NULL.
{
	DIALOG_ITEM
		*currentItem,
		*nextFocusItem;

	if((currentItem=dialog->focusItem))
	{
		currentItem=currentItem->nextItem;			// begin looking one beyond the old focus item (if there was an old one)
	}
	nextFocusItem=NULL;
	while(currentItem&&!nextFocusItem)
	{
		if(currentItem->wantFocus)
		{
			nextFocusItem=currentItem;				// this is the next one that wants the focus
		}
		currentItem=currentItem->nextItem;
	}
	currentItem=dialog->firstItem;					// start over at the beginning
	while(currentItem&&!nextFocusItem)
	{
		if(currentItem->wantFocus)
		{
			nextFocusItem=currentItem;				// this is the next one that wants the focus
		}
		currentItem=currentItem->nextItem;
	}
	return(nextFocusItem);
}

static void DialogFocusForward(DIALOG *dialog)
// Push the focus forward one notch, and alert the involved parties.
// NOTE: this is DIALOG ITEM focus, not X window focus.
{
	DIALOG_ITEM
		*nextFocusItem;

	nextFocusItem=DialogFocusItemLocateForward(dialog);	// this can return NULL, which will be passed to DialogFocusChange to select nothing
	DialogFocusChange(dialog,nextFocusItem);
}

static void DialogItemUnlink(DIALOG_ITEM *dialogItem)
// Unlink dialogItem from the dialog list.
{
	if(dialogItem->previousItem)
	{
		dialogItem->previousItem->nextItem=dialogItem->nextItem;
	}
	else
	{
		dialogItem->dialog->firstItem=dialogItem->nextItem;
	}
	if(dialogItem->nextItem)
	{
		dialogItem->nextItem->previousItem=dialogItem->previousItem;
	}
	else
	{
		dialogItem->dialog->lastItem=dialogItem->previousItem;
	}
}

static void DialogItemLink(DIALOG *dialog,DIALOG_ITEM *newItem)
// Link newItem to the end of the linked list of dialog items in dialog.
{
	newItem->nextItem=NULL;
	if((newItem->previousItem=dialog->lastItem))
	{
		dialog->lastItem->nextItem=newItem;
		dialog->lastItem=newItem;
	}
	else
	{
		dialog->firstItem=dialog->lastItem=newItem;
	}
}

static void DialogItemDispose(DIALOG_ITEM *dialogItem)
// Unlink, and dispose of a dialog item, created by DialogItemCreate.
{
	DialogItemUnlink(dialogItem);
	dialogItem->disposeFunction(dialogItem);
}

static DIALOG_ITEM *DialogItemCreate(DIALOG *dialog,DIALOG_ITEM *(*createFunction)(DIALOG *dialog,void *itemDescriptor),void *itemDescriptor)
// Create a new dialog item, link it to dialog, call the item's creation procedure.
// createFunction must not be NULL and must return true for this to succeed.
// If there is a problem, undo all that was done, ErrorSet, return NULL.
{
	DIALOG_ITEM
		*dialogItem;

	if((dialogItem=createFunction(dialog,itemDescriptor)))
	{
		DialogItemLink(dialog,dialogItem);
		return(dialogItem);
	}
	return(NULL);
}

static void DialogWindowDraw(DIALOG *dialog,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the dialog, and all items that request it.
// Drawing focus is handled from here.
{
	DIALOG_ITEM
		*currentItem;
	EDITOR_RECT
		rect;

	rect.x=0;
	rect.y=0;
	rect.w=dialog->parentWindow->w;
	rect.h=dialog->parentWindow->h;

	ImageFill(imageBuffer,dialogBackgroundColor,0,0,rect.w,rect.h);
//	ShadowRectangleOutline(imageBuffer,&rect,SHADOW_COLOR_BRIGHT,SHADOW_COLOR_DARK,2);	// drop in border

	currentItem=dialog->firstItem;
	while(currentItem)
	{
		if(currentItem->drawFunction)
		{
			currentItem->drawFunction(currentItem,imageBuffer,invalidRegion);
		}
		currentItem=currentItem->nextItem;
	}
}

static void DialogWindowUpdate(EDITOR_WINDOW *window)
// Make sure window is up to date.
{
	DIALOG
		*dialog;
	Window
		xWindow;
	GUI_IMAGE
		*imageBuffer;

	if(!XEmptyRegion(window->invalidRegion))					// see if it has an invalid area
	{
		dialog=(DIALOG *)window->windowInfo;
		xWindow=window->xWindow;								// get x window for this window
		if((imageBuffer=ImageCreate(window->w,window->h)))		// go get an image to draw into
		{
			DialogWindowDraw(dialog,imageBuffer,window->invalidRegion);
			ImageDrawToWindow(imageBuffer,xWindow,window->invalidRegion,window->w,window->h);
			ImageDestroy(imageBuffer);
		}
		XDestroyRegion(window->invalidRegion);					// clear out the invalid region now that the window has been drawn
		window->invalidRegion=XCreateRegion();
	}
}

static void DialogWindowAdjustForNewSize(EDITOR_WINDOW *window)
// If the window has changed size, then the
// controls will be needing update, so adjust them for the new
// size.
{
	DIALOG
		*dialog;
	DIALOG_ITEM
		*currentItem;

	dialog=(DIALOG *)window->windowInfo;

	currentItem=dialog->firstItem;
	while(currentItem)							// run through all the items and ask them to resize
	{
		if(currentItem->resizeFunction)
		{
			currentItem->resizeFunction(currentItem);
		}
		currentItem=currentItem->nextItem;
	}
}

static void DialogButtonPress(DIALOG *dialog,XEvent *event)
// Handle button press events for a dialog (see what the user is clicking
// on, and attempt to deal with it).
// This runs through the items backwards so that it finds the top-most
// item first.
{
	DIALOG_ITEM
		*currentItem;
	bool
		handled;

	handled=false;
	currentItem=dialog->lastItem;
	while(currentItem&&!handled)
	{
		if(currentItem->trackFunction)
		{
			handled=currentItem->trackFunction(currentItem,event);
		}
		currentItem=currentItem->previousItem;
	}
}

static void DialogKeyPress(DIALOG *dialog,EDITOR_KEY *editorKey)
// A key has arrived for window.
// Decide who it should go to, and handle it.
// If a focus change key comes in, and it is not grabbed early, it will
// be used to change the focus, and not sent to the focus item.
{
	DIALOG_ITEM
		*currentItem;
	bool
		handled;

	handled=false;
	currentItem=dialog->firstItem;
	while(currentItem&&!handled)
	{
		if(currentItem->earlyKeyFunction)
		{
			handled=currentItem->earlyKeyFunction(currentItem,editorKey);
		}
		currentItem=currentItem->nextItem;
	}
	if(!handled)
	{
		if(editorKey->keyCode==XKeysymToKeycode(xDisplay,XK_Tab))
		{
			DialogFocusForward(dialog);
			handled=true;
		}
		else if((dialog->focusItem)&&(dialog->focusItem->focusKeyFunction))
		{
			handled=dialog->focusItem->focusKeyFunction(dialog->focusItem,editorKey);
		}
		if(!handled)
		{
			if(editorKey->keyCode==XKeysymToKeycode(xDisplay,XK_Escape))	// escape will abort a dialog if not captured by anything else
			{
				dialog->dialogAbort=true;							// mark the abort
				dialog->dialogComplete=true;						// tell event loop we're done
			}
		}
	}
}

static void DialogPeriodicFunction(EDITOR_WINDOW *window)
// This is called periodically to flash cursors, or do whatever in the dialog
// when the dialog has focus.
{
	DIALOG
		*dialog;

	dialog=(DIALOG *)window->windowInfo;
	if(dialog->focusItem)
	{
		if(dialog->focusItem->focusPeriodicFunction)
		{
			dialog->focusItem->focusPeriodicFunction(dialog->focusItem);
		}
	}
}

static void DialogWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// Handle an event arriving for a dialog window.
{
	DIALOG
		*dialog;
	EDITOR_KEY
		editorKey;

	switch(event->type)
	{
		case ButtonPress:
			XRaiseWindow(xDisplay,window->xWindow);							// make sure dialog is on top
			XSetInputFocus(xDisplay,window->xWindow,RevertToPointerRoot,event->xbutton.time);	// give me the focus
			DialogButtonPress((DIALOG *)window->windowInfo,event);			// go handle the button press
			break;
		case KeyPress:
			if(KeyEventToEditorKey(&event->xkey,&editorKey))
			{
				DialogKeyPress((DIALOG *)window->windowInfo,&editorKey);	// go handle the key press
			}
			break;
		case ClientMessage:
			if(event->xclient.data.l[0]==(int)deleteWindowAtom)
			{
				dialog=(DIALOG *)(window->windowInfo);
				dialog->dialogAbort=true;									// mark the abort
				dialog->dialogComplete=true;								// tell event loop we're done
			}
			break;
		default:
			break;
	}
}

static void DialogEventsHandle(DIALOG *dialog)
// Handle events for dialog.
{
	dialog->dialogAbort=false;												// not aborted
	DialogFocusForward(dialog);												// place focus on first item that wants it
	EventLocalLoop(&(dialog->dialogComplete),dialog->parentWindow);			// process events here until the dialog completes
}

static void DialogExitPress(DIALOG_ITEM *item)
// Button was pressed, say that dialog is complete.
{
	item->dialog->dialogComplete=true;										// set completion flag to true
}

static void DialogDispose(DIALOG *dialog)
// Close a dialog window, kill any items it may have linked to it.
{
	while(dialog->firstItem)
	{
		DialogItemDispose(dialog->firstItem);								// remove any items
	}
	FocusAwayFrom(dialog->parentWindow);									// trick X by moving focus so window manager will not try to when this window dies
	XDestroyRegion(dialog->parentWindow->invalidRegion);					// destroy invalid region
	XDestroyWindow(xDisplay,dialog->parentWindow->xWindow);					// tell x to make the window go away
	WaitUntilDestroyed(dialog->parentWindow);
	WindowElementUnlink(dialog->parentWindow);								// unlink it from our list
	MDisposePtr(dialog->parentWindow);
	MDisposePtr(dialog);
}

static DIALOG *DialogCreate(EDITOR_RECT *rect,uint32_t minWidth,uint32_t minHeight,const char *title,EDITOR_COLOR background)
// Open a new empty dialog of the given size, with the given title.
// It should not be updated until all the items have been added to it.
// If there is a problem, ErrorSet, return NULL.
{
	DIALOG
		*dialog;
	EDITOR_WINDOW
		*window;
	XSetWindowAttributes
		attributes;
	uint32_t
		valueMask;
	Atom
		protocols[2];														// number of protocols
	XWMHints
		hints;																// hints to tell window manager how to treat these windows
	XSizeHints
		sizeHints;															// window size info (used to REALLY convince X that I know where I want my own window)
	XEvent
		xEvent;
	int32_t
		xOffset,
		yOffset;
	uint32_t
		screenWidth,
		screenHeight;

	if((dialog=(DIALOG *)MNewPtr(sizeof(DIALOG))))
	{
		dialog->dialogComplete=false;										// not done with this yet
		dialog->firstItem=dialog->lastItem=dialog->focusItem=NULL;
		if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))		// create editor window
		{
			WindowElementLink(window);										// link it to the global window list that we maintain
			window->windowType=EWT_DIALOG;									// set the window type to "dialog"
			window->w=rect->w;												// remember initial size for the window
			window->h=rect->h;

			valueMask=0;
			// Due to X's nasty error handling, we assume this create does not fail!!!

			PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&screenWidth,&screenHeight);
			window->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x+xOffset,rect->y+yOffset,rect->w,rect->h,2,DefaultDepth(xDisplay,xScreenNum),InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

			sizeHints.flags=USSize|USPosition|PMinSize;						// ### LIE to X, tell it the user requested the size and position so it will honor them
			sizeHints.x=rect->x+xOffset;
			sizeHints.y=rect->y+yOffset;
			sizeHints.width=rect->w;
			sizeHints.height=rect->h;
			sizeHints.min_width=minWidth;
			sizeHints.min_height=minHeight;
			XSetWMNormalHints(xDisplay,window->xWindow,&sizeHints);

//			XSetTransientForHint(xDisplay,window->xWindow,None);			// tell WM that this window is modal to all other windows in the application

			XStoreName(xDisplay,window->xWindow,title);
			XSelectInput(xDisplay,window->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask);	// want to see these events

			protocols[0]=takeFocusAtom;
			protocols[1]=deleteWindowAtom;
			XSetWMProtocols(xDisplay,window->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

			hints.flags=InputHint|StateHint|WindowGroupHint;
			hints.input=True;
			hints.initial_state=NormalState;
			hints.window_group=rootMenuWindow->xWindow;					// all dialogs belong to the group headed by the main menu window
			XSetWMHints(xDisplay,window->xWindow,&hints);

			window->invalidRegion=XCreateRegion();						// create new empty invalid region

			dialog->parentWindow=window;								// link the dialog to the window just created

			window->windowInfo=dialog;									// link the window to the dialog just created
			window->configureFunction=DialogWindowAdjustForNewSize;		// point to the function to call to configure the window
			window->updateFunction=DialogWindowUpdate;					// point to the function to call to update this window when it needs to be redrawn
			window->eventFunction=DialogWindowEvent;					// point to the function to call to handle events
			window->periodicFunction=DialogPeriodicFunction;			// point to the function to call to handle periodic updates

			XMapRaised(xDisplay,window->xWindow);						// map window to the top of the display
			WaitUntilMapped(window);									// wait until this window is actually mapped before continuing

			// This code asks the WM to please keep our dialog window on top
			xEvent.xclient.type=ClientMessage;
			xEvent.xclient.serial=0;
			xEvent.xclient.send_event=True;
			xEvent.xclient.display=xDisplay;
			xEvent.xclient.window=window->xWindow;
			xEvent.xclient.message_type=XInternAtom(xDisplay,"_NET_WM_STATE",True);
			xEvent.xclient.format=32;
			xEvent.xclient.data.l[0]=1;					// 1 _NET_WM_STATE_ADD
			xEvent.xclient.data.l[1]=XInternAtom(xDisplay,"_NET_WM_STATE_ABOVE",True);
			xEvent.xclient.data.l[2]=XInternAtom(xDisplay,"_NET_WM_STATE_MODAL",True);
			xEvent.xclient.data.l[3]=1;					// source indication (1 means application)
			xEvent.xclient.data.l[4]=0;
			XSendEvent(xDisplay,XDefaultRootWindow(xDisplay),False,SubstructureRedirectMask|SubstructureNotifyMask,&xEvent);

			return(dialog);
		}
		MDisposePtr(dialog);
	}
	ErrorSet("Failed to DialogCreate");
	return(NULL);
}

static DIALOG *DialogCreateCentered(uint32_t width,uint32_t height,uint32_t minWidth,uint32_t minHeight,const char *title,EDITOR_COLOR background)
// Open a new dialog window of width, height centered
// nicely on the screen.
// If there is a problem, ErrorSet, and return NULL.
{
	EDITOR_RECT
		dialogRect;

	dialogRect.x=dialogCenterX-width/2;
	dialogRect.y=dialogCenterY-height/2;
	dialogRect.w=width;
	dialogRect.h=height;
	return(DialogCreate(&dialogRect,minWidth,minHeight,title,background));
}

// ----------------------------------------------------------------------------

void DialogFocusTake(DIALOG_ITEM *item)
// An item in the dialog window has decided that it wants the focus, so change the focus to it.
{
	DialogFocusChange(item->dialog,item);
}

DIALOG_ITEM *DialogFocusItemGet(DIALOG *dialog)
// Return the focus item for a given dialog.
{
	return(dialog->focusItem);
}

EDITOR_WINDOW *DialogWindowGet(DIALOG *dialog)
// Return the window for a given dialog.
{
	return(dialog->parentWindow);
}

// ----------------------------------------------------------------------------

void DialogColorsGet(EDITOR_COLOR *backgroundColor)
// Get the color to be used to draw the backgrounds of dialog boxes.
{
	*backgroundColor=dialogBackgroundColor;
}

void DialogColorsSet(EDITOR_COLOR backgroundColor)
// Set the color to be used to draw the backgrounds of dialog boxes.
{
	dialogBackgroundColor=backgroundColor;
}

void DialogStaticTextColorsGet(EDITOR_COLOR *foregroundColor)
// Get the color to be used to draw the static text items of dialog boxes.
{
	*foregroundColor=dialogStaticTextForegroundColor;
}

void DialogStaticTextColorsSet(EDITOR_COLOR foregroundColor)
// Set the color to be used to draw the static text items of dialog boxes.
{
	dialogStaticTextForegroundColor=foregroundColor;
}

void DialogTextBoxColorsGet(EDITOR_COLOR *activeBorderColor,EDITOR_COLOR *inactiveBorderColor)
// Get the colors to be used to draw the text box items of dialog boxes.
{
	*activeBorderColor=dialogTextBoxActiveBorderColor;
	*inactiveBorderColor=dialogTextBoxInactiveBorderColor;
}

void DialogTextBoxColorsSet(EDITOR_COLOR activeBorderColor,EDITOR_COLOR inactiveBorderColor)
// Set the colors to be used to draw the text box items of dialog boxes.
{
	dialogTextBoxActiveBorderColor=activeBorderColor;
	dialogTextBoxInactiveBorderColor=inactiveBorderColor;
}

void DialogListBoxColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *activeBorderColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor,EDITOR_COLOR *inactiveBorderColor)
// Get the colors to be used to draw the list box items of dialog boxes.
{
	*activeBackgroundColor=dialogListBoxActiveBackgroundColor;
	*activeForegroundColor=dialogListBoxActiveForegroundColor;
	*activeBorderColor=dialogListBoxActiveBorderColor;
	*inactiveBackgroundColor=dialogListBoxInactiveBackgroundColor;
	*inactiveForegroundColor=dialogListBoxInactiveForegroundColor;
	*inactiveBorderColor=dialogListBoxInactiveBorderColor;
}

void DialogListBoxColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR activeBorderColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor,EDITOR_COLOR inactiveBorderColor)
// Set the colors to be used to draw the list box items of dialog boxes.
{
	dialogListBoxActiveBackgroundColor=activeBackgroundColor;
	dialogListBoxActiveForegroundColor=activeForegroundColor;
	dialogListBoxActiveBorderColor=activeBorderColor;
	dialogListBoxInactiveBackgroundColor=inactiveBackgroundColor;
	dialogListBoxInactiveForegroundColor=inactiveForegroundColor;
	dialogListBoxInactiveBorderColor=inactiveBorderColor;
}

void DialogCheckBoxColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor)
// Get the colors to be used to draw the check box items of dialog boxes.
{
	*activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
	*activeForegroundColor=dialogCheckBoxActiveForegroundColor;
	*inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
	*inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
}

void DialogCheckBoxColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor)
// Set the colors to be used to draw the check box items of dialog boxes.
{
	dialogCheckBoxActiveBackgroundColor=activeBackgroundColor;
	dialogCheckBoxActiveForegroundColor=activeForegroundColor;
	dialogCheckBoxInactiveBackgroundColor=inactiveBackgroundColor;
	dialogCheckBoxInactiveForegroundColor=inactiveForegroundColor;
}

void DialogPopupColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor,EDITOR_COLOR *borderColor)
// Get the colors to be used to draw the check box items of dialog boxes.
{
	*activeBackgroundColor=dialogPopupActiveBackgroundColor;
	*activeForegroundColor=dialogPopupActiveForegroundColor;
	*inactiveBackgroundColor=dialogPopupInactiveBackgroundColor;
	*inactiveForegroundColor=dialogPopupInactiveForegroundColor;
	*borderColor=dialogPopupBorderColor;
}

void DialogPopupColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor,EDITOR_COLOR borderColor)
// Set the colors to be used to draw the check box items of dialog boxes.
{
	dialogPopupActiveBackgroundColor=activeBackgroundColor;
	dialogPopupActiveForegroundColor=activeForegroundColor;
	dialogPopupInactiveBackgroundColor=inactiveBackgroundColor;
	dialogPopupInactiveForegroundColor=inactiveForegroundColor;
	dialogPopupBorderColor=borderColor;
}

void DialogButtonColorsGet(EDITOR_COLOR *activeBackgroundColor,EDITOR_COLOR *activeForegroundColor,EDITOR_COLOR *inactiveBackgroundColor,EDITOR_COLOR *inactiveForegroundColor,EDITOR_COLOR *pressedBackgroundColor,EDITOR_COLOR *pressedForegroundColor,EDITOR_COLOR *shadowColor)
// Get the colors to be used to draw the check box items of dialog boxes.
{
	*activeBackgroundColor=dialogButtonActiveBackgroundColor;
	*activeForegroundColor=dialogButtonActiveForegroundColor;
	*inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
	*inactiveForegroundColor=dialogButtonInactiveForegroundColor;
	*pressedBackgroundColor=dialogButtonPressedBackgroundColor;
	*pressedForegroundColor=dialogButtonPressedForegroundColor;
	*shadowColor=dialogButtonShadowColor;
}

void DialogButtonColorsSet(EDITOR_COLOR activeBackgroundColor,EDITOR_COLOR activeForegroundColor,EDITOR_COLOR inactiveBackgroundColor,EDITOR_COLOR inactiveForegroundColor,EDITOR_COLOR pressedBackgroundColor,EDITOR_COLOR pressedForegroundColor,EDITOR_COLOR shadowColor)
// Set the colors to be used to draw the check box items of dialog boxes.
{
	dialogButtonActiveBackgroundColor=activeBackgroundColor;
	dialogButtonActiveForegroundColor=activeForegroundColor;
	dialogButtonInactiveBackgroundColor=inactiveBackgroundColor;
	dialogButtonInactiveForegroundColor=inactiveForegroundColor;
	dialogButtonPressedBackgroundColor=pressedBackgroundColor;
	dialogButtonPressedForegroundColor=pressedForegroundColor;
	dialogButtonShadowColor=shadowColor;
}

// ----------------------------------------------------------------------------

const char *DialogStaticTextFontGet()
// Return the font name used for dialog static text.
{
	return(dialogStaticTextFontName);
}

void DialogStaticTextFontSet(const char *fontName)
// Set the font name used for dialog static text.
{
	strncpy(dialogStaticTextFontName,fontName,sizeof(dialogStaticTextFontName)-1);
	dialogStaticTextFontName[sizeof(dialogStaticTextFontName)-1]='\0';
}

const char *DialogTextBoxFontGet()
// Return the font name used for dialog text boxes.
{
	return(dialogTextBoxFontName);
}

void DialogTextBoxFontSet(const char *fontName)
// Set the font name used for dialog text boxes.
{
	strncpy(dialogTextBoxFontName,fontName,sizeof(dialogTextBoxFontName)-1);
	dialogTextBoxFontName[sizeof(dialogTextBoxFontName)-1]='\0';
}

const char *DialogListBoxFontGet()
// Return the font name used for dialog list boxes.
{
	return(dialogListBoxFontName);
}

void DialogListBoxFontSet(const char *fontName)
// Set the font name used for dialog list boxes.
{
	strncpy(dialogListBoxFontName,fontName,sizeof(dialogListBoxFontName)-1);
	dialogListBoxFontName[sizeof(dialogListBoxFontName)-1]='\0';
}

const char *DialogCheckBoxFontGet()
// Return the font name used for dialog check boxes.
{
	return(dialogCheckBoxFontName);
}

void DialogCheckBoxFontSet(const char *fontName)
// Set the font name used for dialog check boxes.
{
	strncpy(dialogCheckBoxFontName,fontName,sizeof(dialogCheckBoxFontName)-1);
	dialogCheckBoxFontName[sizeof(dialogCheckBoxFontName)-1]='\0';
}

const char *DialogPopupFontGet()
// Return the font name used for dialog popups.
{
	return(dialogPopupFontName);
}

void DialogPopupFontSet(const char *fontName)
// Set the font name used for dialog popups.
{
	strncpy(dialogPopupFontName,fontName,sizeof(dialogPopupFontName)-1);
	dialogPopupFontName[sizeof(dialogPopupFontName)-1]='\0';
}

const char *DialogButtonFontGet()
// Return the font name used for dialog buttons.
{
	return(dialogButtonFontName);
}

void DialogButtonFontSet(const char *fontName)
// Set the font name used for dialog buttons.
{
	strncpy(dialogButtonFontName,fontName,sizeof(dialogButtonFontName)-1);
	dialogButtonFontName[sizeof(dialogButtonFontName)-1]='\0';
}

// ----------------------------------------------------------------------------

static int32_t ButtonXGet(uint32_t index,uint32_t numButtons,uint32_t width,uint32_t windowWidth)
// Work out the X index for a button in a dialog window.
// Index is passed as the index of the button,
// numButtons is the maximum number of buttons in the window.
// All the buttons are assumed to be width pixels wide.
{
	if(windowWidth>=(20+width))
	{
		return(10+(windowWidth-(20+width))*index/(numButtons-1));
	}
	return(10);
}

// ----------------------------------------------------------------------------

static void DialogOkStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	if((int32_t)item->dialog->parentWindow->h>(50+rect->y))
	{
		rect->h=item->dialog->parentWindow->h-(50+rect->y);
	}
	else
	{
		rect->h=0;
	}
}

static void DialogOkSeparatorPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the separator.
{
	rect->x=10;
	rect->h=2;

	if(item->dialog->parentWindow->h>45)
	{
		rect->y=item->dialog->parentWindow->h-45;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
}

static void DialogOkButtonOkPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the ok button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/2-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

bool PlatformDialogOk(const char *text)
// Bring up an ok dialog, with text.
// Wait for the user to click on ok, then return.
// If there is a problem, ErrorSet, and return false.
{
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	SEPARATOR_DESCRIPTOR
		separatorDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	if((dialog=DialogCreateCentered(640,300,500,250," ",dialogBackgroundColor)))	// make new empty dialog
	{
		textDescriptor.positionFunction=DialogOkStatTextPosition;
		textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
		textDescriptor.string=text;
		textDescriptor.fontName=dialogStaticTextFontName;
		if(DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor))
		{
			separatorDescriptor.positionFunction=DialogOkSeparatorPosition;
			separatorDescriptor.horizontal=true;
			if(DialogItemCreate(dialog,SeparatorItemCreate,&separatorDescriptor))
			{
				buttonDescriptor.positionFunction=DialogOkButtonOkPosition;
				buttonDescriptor.buttonName="Ok";
				buttonDescriptor.image=buttonImage;
				buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
				buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
				buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
				buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
				buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
				buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
				buttonDescriptor.shadowColor=dialogButtonShadowColor;
				buttonDescriptor.fontName=dialogButtonFontName;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
				buttonDescriptor.pressedFunction=DialogExitPress;
				buttonDescriptor.pressedState=NULL;
				if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
				{
					DialogEventsHandle(dialog);						// deal with this dialog
					result=true;
				}
			}
		}
		DialogDispose(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------

static void DialogOkCancelStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	if((int32_t)item->dialog->parentWindow->h>(50+rect->y))
	{
		rect->h=item->dialog->parentWindow->h-(50+rect->y);
	}
	else
	{
		rect->h=0;
	}
}

static void DialogOkCancelSeparatorPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the separator.
{
	rect->x=10;
	rect->h=2;

	if(item->dialog->parentWindow->h>45)
	{
		rect->y=item->dialog->parentWindow->h-45;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
}

static void DialogOkCancelButtonOkPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the ok button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogOkCancelButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=4*item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

bool PlatformDialogOkCancel(const char *text,bool *cancel)
// Bring up an ok/cancel dialog, with text.
// If there is a problem, ErrorSet, return false.
// Otherwise, return true in cancel if the user
// canceled, false if not.
{
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	SEPARATOR_DESCRIPTOR
		separatorDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*cancel=false;
	if((dialog=DialogCreateCentered(640,300,500,250," ",dialogBackgroundColor)))			// make new empty dialog
	{
		textDescriptor.positionFunction=DialogOkCancelStatTextPosition;
		textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
		textDescriptor.string=text;
		textDescriptor.fontName=dialogStaticTextFontName;
		if(DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor))
		{
			separatorDescriptor.positionFunction=DialogOkCancelSeparatorPosition;
			separatorDescriptor.horizontal=true;
			if(DialogItemCreate(dialog,SeparatorItemCreate,&separatorDescriptor))
			{
				buttonDescriptor.positionFunction=DialogOkCancelButtonOkPosition;
				buttonDescriptor.buttonName="Ok";
				buttonDescriptor.image=buttonImage;
				buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
				buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
				buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
				buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
				buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
				buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
				buttonDescriptor.shadowColor=dialogButtonShadowColor;
				buttonDescriptor.fontName=dialogButtonFontName;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
				buttonDescriptor.pressedFunction=DialogExitPress;
				buttonDescriptor.pressedState=NULL;
				if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
				{
					buttonDescriptor.positionFunction=DialogOkCancelButtonCancelPosition;
					buttonDescriptor.buttonName="Cancel";
					buttonDescriptor.image=buttonImage;
					buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
					buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
					buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
					buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
					buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
					buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
					buttonDescriptor.shadowColor=dialogButtonShadowColor;
					buttonDescriptor.fontName=dialogButtonFontName;
					buttonDescriptor.hasKey=true;
					buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
					buttonDescriptor.pressedFunction=DialogExitPress;
					buttonDescriptor.pressedState=cancel;
					if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
					{
						DialogEventsHandle(dialog);					// deal with this dialog
						if(dialog->dialogAbort)
						{
							(*cancel)=true;							// return cancel if user closed dialog window
						}
						result=true;
					}
				}
			}
		}
		DialogDispose(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------

static void DialogYesNoCancelStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	if((int32_t)item->dialog->parentWindow->h>(50+rect->y))
	{
		rect->h=item->dialog->parentWindow->h-(50+rect->y);
	}
	else
	{
		rect->h=0;
	}
}

static void DialogYesNoCancelSeparatorPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the separator.
{
	rect->x=10;
	rect->h=2;

	if(item->dialog->parentWindow->h>45)
	{
		rect->y=item->dialog->parentWindow->h-45;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
}

static void DialogYesNoCancelButtonYesPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the yes button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(0,4,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogYesNoCancelButtonNoPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the no button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(1,4,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogYesNoCancelButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(3,4,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

bool PlatformDialogYesNoCancel(const char *text,bool *yes,bool *cancel)
// Bring up a yes, no, or cancel dialog, with text.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true the user canceled.
// Otherwise, if yes is true, yes was hit.
// Otherwise no was hit.
{
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	SEPARATOR_DESCRIPTOR
		separatorDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*yes=*cancel=false;
	if((dialog=DialogCreateCentered(640,300,500,250," ",dialogBackgroundColor)))			// make new empty dialog
	{
		textDescriptor.positionFunction=DialogYesNoCancelStatTextPosition;
		textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
		textDescriptor.string=text;
		textDescriptor.fontName=dialogStaticTextFontName;

		if(DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor))
		{
			separatorDescriptor.positionFunction=DialogYesNoCancelSeparatorPosition;
			separatorDescriptor.horizontal=true;
			if(DialogItemCreate(dialog,SeparatorItemCreate,&separatorDescriptor))
			{
				buttonDescriptor.positionFunction=DialogYesNoCancelButtonYesPosition;
				buttonDescriptor.buttonName="Yes";
				buttonDescriptor.image=buttonImage;
				buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
				buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
				buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
				buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
				buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
				buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
				buttonDescriptor.shadowColor=dialogButtonShadowColor;
				buttonDescriptor.fontName=dialogButtonFontName;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_y);
				buttonDescriptor.pressedFunction=DialogExitPress;
				buttonDescriptor.pressedState=yes;
				if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
				{
					buttonDescriptor.positionFunction=DialogYesNoCancelButtonNoPosition;
					buttonDescriptor.buttonName="No";
					buttonDescriptor.image=buttonImage;
					buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
					buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
					buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
					buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
					buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
					buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
					buttonDescriptor.shadowColor=dialogButtonShadowColor;
					buttonDescriptor.fontName=dialogButtonFontName;
					buttonDescriptor.hasKey=true;
					buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_n);
					buttonDescriptor.pressedFunction=DialogExitPress;
					buttonDescriptor.pressedState=NULL;
					if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
					{
						buttonDescriptor.positionFunction=DialogYesNoCancelButtonCancelPosition;
						buttonDescriptor.buttonName="Cancel";
						buttonDescriptor.image=buttonImage;
						buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
						buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
						buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
						buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
						buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
						buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
						buttonDescriptor.shadowColor=dialogButtonShadowColor;
						buttonDescriptor.fontName=dialogButtonFontName;
						buttonDescriptor.hasKey=true;
						buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
						buttonDescriptor.pressedFunction=DialogExitPress;
						buttonDescriptor.pressedState=cancel;
						if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
						{
							DialogEventsHandle(dialog);					// deal with this dialog
							if(dialog->dialogAbort)
							{
								(*cancel)=true;							// return cancel if user closed dialog window
							}
							result=true;
						}
					}
				}
			}
		}
		DialogDispose(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------

typedef struct
{
	DIALOG_ITEM
		*staticText;
} SIMPLE_TEXT_LOCAL;

static void DialogSimpleTextStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=StaticTextItemHeightGet(item,rect->w);	// work out the ideal height

	if(rect->h>item->dialog->parentWindow->h/2)		// limit the height if it is too large
	{
		rect->h=item->dialog->parentWindow->h/2;
	}
}

static void DialogSimpleTextTextBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the text box.
// It is placed just below the static text, and just above the buttons.
{
	SIMPLE_TEXT_LOCAL
		*local;
	EDITOR_RECT
		staticTextRect;
	int32_t
		bottomY;

	local=(SIMPLE_TEXT_LOCAL *)item->dialog->dialogLocal;
	StaticTextItemRectangleGet(local->staticText,&staticTextRect);

	rect->x=10;
	rect->y=staticTextRect.y+staticTextRect.h+5;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	bottomY=item->dialog->parentWindow->h-35-15;

	if(bottomY>rect->y)
	{
		rect->h=bottomY-rect->y;
	}
	else
	{
		rect->h=0;
	}
}

static void DialogSimpleTextButtonOkPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the ok button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSimpleTextButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=4*item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

bool PlatformDialogSimpleText(const char *title,char *enteredText,uint32_t stringBytes,bool *cancel)
// Open a string input dialog with title.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true, the user canceled.
// Otherwise, enteredText will be set to the text that the user entered in the box.
// If enteredText is set to something on entry, it will be placed in the box.
{
	SIMPLE_TEXT_LOCAL
		local;															// local state that callbacks can get to
	DIALOG
		*dialog;
	EDITOR_BUFFER
		*buffer;
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	uint64_t
		totalBytes;
	bool
		result;

	result=false;
	*cancel=false;
	if((buffer=EditorBufferNew("simple text")))		// create a buffer in which to collect the text
	{
		if(TextUniverseInsertFromText((uint8_t *)enteredText,buffer->textUniverse,TextUniverseTotalBytes(buffer->textUniverse),strlen(enteredText)))
		{
			if(EditorSelectAll(buffer,buffer->selectionUniverse))	// select it all
			{
				if((dialog=DialogCreateCentered(600,300,600,300," ",dialogBackgroundColor)))			// make new empty dialog
				{
					dialog->dialogLocal=(void *)&local;

					textDescriptor.positionFunction=DialogSimpleTextStatTextPosition;
					textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
					textDescriptor.string=title;
					textDescriptor.fontName=dialogStaticTextFontName;
					if((local.staticText=DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor)))
					{
						textBoxDescriptor.positionFunction=DialogSimpleTextTextBoxPosition;
						textBoxDescriptor.buffer=buffer;
						textBoxDescriptor.tabSize=4;
						textBoxDescriptor.fontName=dialogTextBoxFontName;
						textBoxDescriptor.colorSchemeName=SIMPLE_TEXT_COLOR_SCHEME_NAME;
						textBoxDescriptor.activeBorderColor=dialogTextBoxActiveBorderColor;
						textBoxDescriptor.inactiveBorderColor=dialogTextBoxInactiveBorderColor;
						if(DialogItemCreate(dialog,TextBoxItemCreate,&textBoxDescriptor))
						{
							buttonDescriptor.positionFunction=DialogSimpleTextButtonOkPosition;
							buttonDescriptor.buttonName="Ok";
							buttonDescriptor.image=buttonImage;
							buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
							buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
							buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
							buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
							buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
							buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
							buttonDescriptor.shadowColor=dialogButtonShadowColor;
							buttonDescriptor.fontName=dialogButtonFontName;
							buttonDescriptor.hasKey=true;
							buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
							buttonDescriptor.pressedFunction=DialogExitPress;
							buttonDescriptor.pressedState=NULL;
							if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
							{
								buttonDescriptor.positionFunction=DialogSimpleTextButtonCancelPosition;
								buttonDescriptor.buttonName="Cancel";
								buttonDescriptor.image=buttonImage;
								buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
								buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
								buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
								buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
								buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
								buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
								buttonDescriptor.shadowColor=dialogButtonShadowColor;
								buttonDescriptor.fontName=dialogButtonFontName;
								buttonDescriptor.hasKey=true;
								buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
								buttonDescriptor.pressedFunction=DialogExitPress;
								buttonDescriptor.pressedState=cancel;
								if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
								{
									DialogEventsHandle(dialog);					// deal with this dialog
									if(dialog->dialogAbort)
									{
										(*cancel)=true;							// return cancel if user closed dialog window
									}
									result=true;
								}
							}
						}
					}
					DialogDispose(dialog);
				}
			}
		}
		if(result&&!(*cancel))
		{
			TextUniverseTextExtract(buffer->textUniverse,0,(uint8_t *)enteredText,stringBytes,NULL);
			totalBytes=TextUniverseTotalBytes(buffer->textUniverse);
			if(totalBytes<stringBytes)
			{
				enteredText[totalBytes]='\0';
			}
			else
			{
				if(stringBytes)
				{
					enteredText[stringBytes-1]='\0';
				}
			}
		}
		EditorBufferClose(buffer);
	}
	return(result);
}

// ----------------------------------------------------------------------------

typedef struct
{
	DIALOG_ITEM
		*selectionExpressionCheckBox,
		*useOnlySubexpressionCheckBox,
		*subexpressionPopup;
} SEARCH_REPLACE_LOCAL;

static void DialogSearchReplaceFindStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceReplaceStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;

	if(item->dialog->parentWindow->h>150)
	{
		rect->y=(item->dialog->parentWindow->h-150)/2+10;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceTextBoxFindPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the find text box.
{
	int32_t
		yStart;
	uint32_t
		height;

	rect->x=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}

	if(item->dialog->parentWindow->h>150)
	{
		yStart=0;
		height=(uint32_t)(item->dialog->parentWindow->h-150)/2;
		if(height>30)			// leave 10 for space above label, 20 for text label
		{
			rect->y=yStart+30;	// move just below the text
			rect->h=height-30;	// take up remaining space
		}
		else
		{
			rect->y=0;
			rect->h=0;
		}
	}
	else
	{
		rect->y=0;
		rect->h=0;
	}
}

static void DialogSearchReplaceTextBoxReplacePosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the find text box.
{
	int32_t
		yStart;
	uint32_t
		height;

	rect->x=10;

	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}

	// get Y location where replace label and text window go
	if(item->dialog->parentWindow->h>150)
	{
		yStart=(item->dialog->parentWindow->h-150)/2;
		height=(uint32_t)(item->dialog->parentWindow->h-150)/2;
		if(height>30)			// leave 10 for space above label, 20 for text label
		{
			rect->y=yStart+30;	// move just below the text
			rect->h=height-30;	// take up remaining space
		}
		else
		{
			rect->y=0;
			rect->h=0;
		}
	}
	else
	{
		rect->y=0;
		rect->h=0;
	}
}

static void DialogSearchReplaceCheckBoxBackwardsPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the search backwards check box.
{
	rect->x=10;

	if(item->dialog->parentWindow->h>140)
	{
		rect->y=item->dialog->parentWindow->h-140;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceCheckBoxWrapPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the search wrap around check box.
{
	rect->x=10;

	if(item->dialog->parentWindow->h>115)
	{
		rect->y=item->dialog->parentWindow->h-115;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceCheckBoxIgnoreCasePosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the search ignore case check box.
{
	rect->x=10;

	if(item->dialog->parentWindow->h>90)
	{
		rect->y=item->dialog->parentWindow->h-90;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceCheckBoxLimitScopePosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the search limit scope check box.
{
	rect->x=10;

	if(item->dialog->parentWindow->h>65)
	{
		rect->y=item->dialog->parentWindow->h-65;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceCheckBoxExpressionPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the search expression check box.
{
	rect->x=item->dialog->parentWindow->w/2+10;

	if(item->dialog->parentWindow->h>140)
	{
		rect->y=item->dialog->parentWindow->h-140;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceCheckBoxUseSubexpressionPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the use subexpression check box.
{
	rect->x=item->dialog->parentWindow->w/2+10;

	if(item->dialog->parentWindow->h>115)
	{
		rect->y=item->dialog->parentWindow->h-115;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w*3/8>20)
	{
		rect->w=item->dialog->parentWindow->w*3/8-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplacePopupSubexpressionPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the subexpression popup.
{
	rect->x=item->dialog->parentWindow->w/2+10+item->dialog->parentWindow->w*3/8;

	if(item->dialog->parentWindow->h>115)
	{
		rect->y=item->dialog->parentWindow->h-115;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/8>20)
	{
		rect->w=item->dialog->parentWindow->w/8-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceCheckBoxTclPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the search Tcl replacement check box.
{
	rect->x=item->dialog->parentWindow->w/2+10;

	if(item->dialog->parentWindow->h>65)
	{
		rect->y=item->dialog->parentWindow->h-65;
	}
	else
	{
		rect->y=0;
	}

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogSearchReplaceButtonFindPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the find button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(0,6,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSearchReplaceButtonFindAllPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the find all button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(1,6,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSearchReplaceButtonReplacePosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the replace button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(2,6,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSearchReplaceButtonReplaceAllPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the replace all button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(3,6,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSearchReplaceButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=ButtonXGet(5,6,width,item->dialog->parentWindow->w);

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSearchReplaceCheckBoxExpressionPressed(DIALOG_ITEM *item)
// Called when the check-box status for selection expression (or subexpression) changes.
// This is used to update the subexpression checkbox's active state, and the
// popup's active state.
{
	SEARCH_REPLACE_LOCAL
		*local;
	bool
		expressionsSelected;
	bool
		subExpressionsSelected;

	local=(SEARCH_REPLACE_LOCAL *)item->dialog->dialogLocal;
	expressionsSelected=CheckBoxItemCheckedGet(local->selectionExpressionCheckBox);
	subExpressionsSelected=CheckBoxItemCheckedGet(local->useOnlySubexpressionCheckBox);

	CheckBoxItemActiveSet(local->useOnlySubexpressionCheckBox,expressionsSelected);
	PopupItemActiveSet(local->subexpressionPopup,expressionsSelected&&subExpressionsSelected);
}

static const char
	*subExpressionList[]=
	{
		"0",
		"1",
		"2",
		"3",
		"4",
		"5",
		"6",
		"7",
		"8",
		"9",
	};

bool PlatformDialogSearchReplace(EDITOR_BUFFER *searchBuffer,EDITOR_BUFFER *replaceBuffer,bool *backwards,bool *wrapAround,bool *ignoreCase,bool *limitScope,bool *selectionExpr,bool *useSubexpression,uint32_t *subexpressionIndex,bool *replaceFunction,uint16_t *searchType,bool *cancel)
// Open the search/replace dialog.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true, the user canceled.
// Otherwise, return the enumerated mode that the user chose, as well as
// the states of backwards, wrapAround, selectionExpr, ignoreCase,
// limitScope, replaceFunction.
{
	DIALOG
		*dialog;
	SEARCH_REPLACE_LOCAL
		local;															// local state that callbacks can get to
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	CHECK_BOX_DESCRIPTOR
		checkDescriptor;
	POPUP_DESCRIPTOR
		popupDescriptor;
	bool
		result;
	bool
		hadFind,
		hadFindAll,
		hadReplace,
		hadReplaceAll;

	result=false;
	*cancel=false;
	hadFind=false;
	hadFindAll=false;
	hadReplace=false;
	hadReplaceAll=false;

	if(EditorSelectAll(searchBuffer,searchBuffer->selectionUniverse))	// select the entire contents of the find buffer
	{
		if(EditorSelectAll(replaceBuffer,replaceBuffer->selectionUniverse))					// select the entire contents of the replace buffer
		{
			if((dialog=DialogCreateCentered(800,434,640,434," ",dialogBackgroundColor)))	// make new empty dialog
			{
				dialog->dialogLocal=(void *)&local;

				textDescriptor.positionFunction=DialogSearchReplaceFindStatTextPosition;
				textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
				textDescriptor.string="Find:";
				textDescriptor.fontName=dialogStaticTextFontName;
				if(DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor))
				{
					textBoxDescriptor.positionFunction=DialogSearchReplaceTextBoxFindPosition;
					textBoxDescriptor.buffer=searchBuffer;
					textBoxDescriptor.tabSize=4;
					textBoxDescriptor.fontName=dialogTextBoxFontName;
					textBoxDescriptor.colorSchemeName=FIND_COLOR_SCHEME_NAME;
					textBoxDescriptor.activeBorderColor=dialogTextBoxActiveBorderColor;
					textBoxDescriptor.inactiveBorderColor=dialogTextBoxInactiveBorderColor;
					if(DialogItemCreate(dialog,TextBoxItemCreate,&textBoxDescriptor))
					{
						textDescriptor.positionFunction=DialogSearchReplaceReplaceStatTextPosition;
						textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
						textDescriptor.string="Replace with:";
						textDescriptor.fontName=dialogStaticTextFontName;
						if(DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor))
						{
							textBoxDescriptor.positionFunction=DialogSearchReplaceTextBoxReplacePosition;
							textBoxDescriptor.buffer=replaceBuffer;
							textBoxDescriptor.tabSize=4;
							textBoxDescriptor.fontName=dialogTextBoxFontName;
							textBoxDescriptor.colorSchemeName=REPLACE_COLOR_SCHEME_NAME;
							textBoxDescriptor.activeBorderColor=dialogTextBoxActiveBorderColor;
							textBoxDescriptor.inactiveBorderColor=dialogTextBoxInactiveBorderColor;
							if(DialogItemCreate(dialog,TextBoxItemCreate,&textBoxDescriptor))
							{
								// check boxes
								checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxBackwardsPosition;
								checkDescriptor.checkBoxName="Search Backwards";
								checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
								checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
								checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
								checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
								checkDescriptor.fontName=dialogCheckBoxFontName;
								checkDescriptor.active=true;
								checkDescriptor.hasKey=false;
								checkDescriptor.pressedFunction=NULL;
								checkDescriptor.pressedState=backwards;
								if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
								{
									checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxWrapPosition;
									checkDescriptor.checkBoxName="Wrap Around";
									checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
									checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
									checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
									checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
									checkDescriptor.fontName=dialogCheckBoxFontName;
									checkDescriptor.active=true;
									checkDescriptor.hasKey=false;
									checkDescriptor.pressedFunction=NULL;
									checkDescriptor.pressedState=wrapAround;
									if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
									{
										checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxIgnoreCasePosition;
										checkDescriptor.checkBoxName="Ignore Case";
										checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
										checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
										checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
										checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
										checkDescriptor.fontName=dialogCheckBoxFontName;
										checkDescriptor.active=true;
										checkDescriptor.hasKey=false;
										checkDescriptor.pressedFunction=NULL;
										checkDescriptor.pressedState=ignoreCase;
										if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
										{
											checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxLimitScopePosition;
											checkDescriptor.checkBoxName="Limit Scope to Current Selection";
											checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
											checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
											checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
											checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
											checkDescriptor.fontName=dialogCheckBoxFontName;
											checkDescriptor.active=true;
											checkDescriptor.hasKey=false;
											checkDescriptor.pressedFunction=NULL;
											checkDescriptor.pressedState=limitScope;
											if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
											{
												checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxExpressionPosition;
												checkDescriptor.checkBoxName="Selection Expression";
												checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
												checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
												checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
												checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
												checkDescriptor.fontName=dialogCheckBoxFontName;
												checkDescriptor.active=true;
												checkDescriptor.hasKey=false;
												checkDescriptor.pressedFunction=DialogSearchReplaceCheckBoxExpressionPressed;
												checkDescriptor.pressedState=selectionExpr;
												if((local.selectionExpressionCheckBox=DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor)))
												{
													checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxUseSubexpressionPosition;
													checkDescriptor.checkBoxName="Select Only Subexpression";
													checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
													checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
													checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
													checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
													checkDescriptor.fontName=dialogCheckBoxFontName;
													checkDescriptor.active=true;
													checkDescriptor.hasKey=false;
													checkDescriptor.pressedFunction=DialogSearchReplaceCheckBoxExpressionPressed;
													checkDescriptor.pressedState=useSubexpression;
													if((local.useOnlySubexpressionCheckBox=DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor)))
													{
														popupDescriptor.positionFunction=DialogSearchReplacePopupSubexpressionPosition;
														popupDescriptor.activeBackgroundColor=dialogPopupActiveBackgroundColor;
														popupDescriptor.activeForegroundColor=dialogPopupActiveForegroundColor;
														popupDescriptor.inactiveBackgroundColor=dialogPopupInactiveBackgroundColor;
														popupDescriptor.inactiveForegroundColor=dialogPopupInactiveForegroundColor;
														popupDescriptor.borderColor=dialogPopupBorderColor;
														popupDescriptor.fontName=dialogPopupFontName;
														popupDescriptor.active=true;
														popupDescriptor.numElements=elementsof(subExpressionList);
														popupDescriptor.listElements=subExpressionList;
														popupDescriptor.currentElement=subexpressionIndex;
														popupDescriptor.pressedFunction=NULL;
														if((local.subexpressionPopup=DialogItemCreate(dialog,PopupItemCreate,&popupDescriptor)))
														{
															checkDescriptor.positionFunction=DialogSearchReplaceCheckBoxTclPosition;
															checkDescriptor.checkBoxName="Replace Text is Tcl Procedure";
															checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
															checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
															checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
															checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
															checkDescriptor.fontName=dialogCheckBoxFontName;
															checkDescriptor.active=true;
															checkDescriptor.hasKey=false;
															checkDescriptor.pressedFunction=NULL;
															checkDescriptor.pressedState=replaceFunction;
															if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
															{
																buttonDescriptor.positionFunction=DialogSearchReplaceButtonFindPosition;
																buttonDescriptor.buttonName="Find";
																buttonDescriptor.image=buttonImage;
																buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
																buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
																buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
																buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
																buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
																buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
																buttonDescriptor.shadowColor=dialogButtonShadowColor;
																buttonDescriptor.fontName=dialogButtonFontName;
																buttonDescriptor.hasKey=true;
																buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
																buttonDescriptor.pressedFunction=DialogExitPress;
																buttonDescriptor.pressedState=&hadFind;
																if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
																{
																	buttonDescriptor.positionFunction=DialogSearchReplaceButtonFindAllPosition;
																	buttonDescriptor.buttonName="Find All";
																	buttonDescriptor.image=buttonImage;
																	buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
																	buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
																	buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
																	buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
																	buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
																	buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
																	buttonDescriptor.shadowColor=dialogButtonShadowColor;
																	buttonDescriptor.fontName=dialogButtonFontName;
																	buttonDescriptor.hasKey=false;
																	buttonDescriptor.pressedFunction=DialogExitPress;
																	buttonDescriptor.pressedState=&hadFindAll;
																	if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
																	{
																		buttonDescriptor.positionFunction=DialogSearchReplaceButtonReplacePosition;
																		buttonDescriptor.buttonName="Replace";
																		buttonDescriptor.image=buttonImage;
																		buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
																		buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
																		buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
																		buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
																		buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
																		buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
																		buttonDescriptor.shadowColor=dialogButtonShadowColor;
																		buttonDescriptor.fontName=dialogButtonFontName;
																		buttonDescriptor.hasKey=false;
																		buttonDescriptor.pressedFunction=DialogExitPress;
																		buttonDescriptor.pressedState=&hadReplace;
																		if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
																		{
																			buttonDescriptor.positionFunction=DialogSearchReplaceButtonReplaceAllPosition;
																			buttonDescriptor.buttonName="Replace All";
																			buttonDescriptor.image=buttonImage;
																			buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
																			buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
																			buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
																			buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
																			buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
																			buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
																			buttonDescriptor.shadowColor=dialogButtonShadowColor;
																			buttonDescriptor.fontName=dialogButtonFontName;
																			buttonDescriptor.hasKey=false;
																			buttonDescriptor.pressedFunction=DialogExitPress;
																			buttonDescriptor.pressedState=&hadReplaceAll;
																			if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
																			{
																				buttonDescriptor.positionFunction=DialogSearchReplaceButtonCancelPosition;
																				buttonDescriptor.buttonName="Cancel";
																				buttonDescriptor.image=buttonImage;
																				buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
																				buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
																				buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
																				buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
																				buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
																				buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
																				buttonDescriptor.shadowColor=dialogButtonShadowColor;
																				buttonDescriptor.fontName=dialogButtonFontName;
																				buttonDescriptor.hasKey=true;
																				buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
																				buttonDescriptor.pressedFunction=DialogExitPress;
																				buttonDescriptor.pressedState=cancel;
																				if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
																				{
																					DialogSearchReplaceCheckBoxExpressionPressed(local.selectionExpressionCheckBox);	// make initial "callback" to set initial active states
																					DialogEventsHandle(dialog);					// deal with this dialog
																					if(dialog->dialogAbort)
																					{
																						(*cancel)=true;							// return cancel if user closed dialog window
																					}
																					result=true;
																				}
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
				DialogDispose(dialog);
			}
		}
	}
	if(result)
	{
		if(hadFind)
		{
			(*searchType)=ST_FIND;
		}
		else if(hadFindAll)
		{
			(*searchType)=ST_FIND_ALL;
		}
		else if(hadReplace)
		{
			(*searchType)=ST_REPLACE;
		}
		else if(hadReplaceAll)
		{
			(*searchType)=ST_REPLACE_ALL;
		}
	}
	return(result);
}

// ----------------------------------------------------------------------------

typedef struct
{
	DIALOG_ITEM
		*staticText,
		*okButton;
} SIMPLE_LIST_LOCAL;

static void DialogSimpleListStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=StaticTextItemHeightGet(item,rect->w);	// work out the ideal height

	if(rect->h>item->dialog->parentWindow->h/2)		// limit the height if it is too large
	{
		rect->h=item->dialog->parentWindow->h/2;
	}
}

static void DialogSimpleListListBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the list box area.
{
	SIMPLE_LIST_LOCAL
		*local;
	EDITOR_RECT
		staticTextRect;
	int32_t
		bottomY;

	local=(SIMPLE_LIST_LOCAL *)item->dialog->dialogLocal;
	StaticTextItemRectangleGet(local->staticText,&staticTextRect);
	rect->x=10;
	rect->y=staticTextRect.y+staticTextRect.h+5;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	bottomY=item->dialog->parentWindow->h-35-15;

	if(bottomY>rect->y)
	{
		rect->h=bottomY-rect->y;
	}
	else
	{
		rect->h=0;
	}
}

static void DialogSimpleListButtonOkPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the ok button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSimpleListButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=4*item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogSimpleListExitPressFunction(DIALOG_ITEM *item)
// When the simple list is double clicked in, it is just like
// pressing the Ok button, so we do that!
{
	SIMPLE_LIST_LOCAL
		*local;

	local=(SIMPLE_LIST_LOCAL *)item->dialog->dialogLocal;
	ButtonItemPress(local->okButton);
}

bool PlatformDialogSimpleList(const char *title,uint32_t numElements,const char **listElements,bool *selectedElements,bool *cancel)
// Open a list selection dialog box with two buttons (ok, cancel).
// The list dialog will be used to modify the selectionListArray, to tell
// which items were selected, and which were not.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true, the user canceled.
// Otherwise, selectedElements is modified to reflect which items
// have been selected.
{
	SIMPLE_LIST_LOCAL
		local;															// local state that callbacks can get to
	DIALOG
		*dialog;
	LIST_BOX_DESCRIPTOR
		listBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*cancel=false;

	if((dialog=DialogCreateCentered(800,700,600,300," ",dialogBackgroundColor)))	// make new empty dialog
	{
		dialog->dialogLocal=(void *)&local;

		textDescriptor.positionFunction=DialogSimpleListStatTextPosition;
		textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
		textDescriptor.string=title;
		textDescriptor.fontName=dialogStaticTextFontName;

		if((local.staticText=DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor)))
		{
			listBoxDescriptor.positionFunction=DialogSimpleListListBoxPosition;
			listBoxDescriptor.activeBackgroundColor=dialogListBoxActiveBackgroundColor;
			listBoxDescriptor.activeForegroundColor=dialogListBoxActiveForegroundColor;
			listBoxDescriptor.activeBorderColor=dialogListBoxActiveBorderColor;
			listBoxDescriptor.inactiveBackgroundColor=dialogListBoxInactiveBackgroundColor;
			listBoxDescriptor.inactiveForegroundColor=dialogListBoxInactiveForegroundColor;
			listBoxDescriptor.inactiveBorderColor=dialogListBoxInactiveBorderColor;
			listBoxDescriptor.numElements=numElements;
			listBoxDescriptor.listElements=listElements;
			listBoxDescriptor.selectedElements=selectedElements;
			listBoxDescriptor.fontName=dialogListBoxFontName;
			listBoxDescriptor.pressedFunction=DialogSimpleListExitPressFunction;
			if(DialogItemCreate(dialog,ListBoxItemCreate,&listBoxDescriptor))
			{
				buttonDescriptor.positionFunction=DialogSimpleListButtonOkPosition;
				buttonDescriptor.buttonName="Ok";
				buttonDescriptor.image=buttonImage;
				buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
				buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
				buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
				buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
				buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
				buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
				buttonDescriptor.shadowColor=dialogButtonShadowColor;
				buttonDescriptor.fontName=dialogButtonFontName;
				buttonDescriptor.hasKey=true;
				buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
				buttonDescriptor.pressedFunction=DialogExitPress;
				buttonDescriptor.pressedState=NULL;
				if((local.okButton=DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor)))	// remember this button
				{
					buttonDescriptor.positionFunction=DialogSimpleListButtonCancelPosition;
					buttonDescriptor.buttonName="Cancel";
					buttonDescriptor.image=buttonImage;
					buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
					buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
					buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
					buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
					buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
					buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
					buttonDescriptor.shadowColor=dialogButtonShadowColor;
					buttonDescriptor.fontName=dialogButtonFontName;
					buttonDescriptor.hasKey=true;
					buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
					buttonDescriptor.pressedFunction=DialogExitPress;
					buttonDescriptor.pressedState=cancel;
					if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
					{
						DialogEventsHandle(dialog);					// deal with this dialog
						if(dialog->dialogAbort)
						{
							(*cancel)=true;							// return cancel if user closed dialog window
						}
						result=true;
					}
				}
			}
		}
		DialogDispose(dialog);
	}
	return(result);
}

// ----------------------------------------------------------------------------
// Stuff dealing with open/save dialogs

// This structure is hung off of the open/save dialogs, it allows the dialog item
// procedures access to some things they need.

typedef struct
{
	EDITOR_BUFFER
		*buffer;
	char
		pathString[MAXPATHLEN+1];
	char
		**list;
	bool
		*selectionList;
	uint32_t
		numElements;
	char
		**outputList;
	uint32_t
		outputElements;
	bool
		justOne,
		justPaths;
	DIALOG_ITEM
		*pathText,
		*listBox,
		*textBox,
		*OkButton;
} FILE_DIALOG_PARAMETERS;

static void DialogFileStatTextTitlePosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogFileStatTextPathPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=30;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}

	rect->h=StaticTextItemHeightGet(item,rect->w);	// work out the ideal height
	if(rect->h>item->dialog->parentWindow->h/2)		// limit the height if it is too large
	{
		rect->h=item->dialog->parentWindow->h/2;
	}
}

static void DialogFileTextBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the file dialog path entry box.
{
	FILE_DIALOG_PARAMETERS
		*dialogParameters;
	EDITOR_RECT
		pathTextRect;

	dialogParameters=(FILE_DIALOG_PARAMETERS *)item->dialog->dialogLocal;
	StaticTextItemRectangleGet(dialogParameters->pathText,&pathTextRect);

	rect->x=10;
	rect->y=pathTextRect.y+pathTextRect.h+5;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=30;
}

static void DialogFileListBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the file dialog list box.
{
	FILE_DIALOG_PARAMETERS
		*dialogParameters;
	EDITOR_RECT
		pathTextRect;
	int32_t
		bottomY;

	dialogParameters=(FILE_DIALOG_PARAMETERS *)item->dialog->dialogLocal;
	StaticTextItemRectangleGet(dialogParameters->pathText,&pathTextRect);

	rect->x=10;
	rect->y=pathTextRect.y+pathTextRect.h+5+30+15;

	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}

	bottomY=item->dialog->parentWindow->h-35-15;

	if(bottomY>rect->y)
	{
		rect->h=bottomY-rect->y;
	}
	else
	{
		rect->h=0;
	}
}

static void DialogFileButtonOkPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the ok button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogFileButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=4*item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogFileButtonAcceptPathPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the accept path button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/2-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static char **DirectoryListProcess(const char *pathString,char **list,uint32_t inElements,uint32_t *outElements,bool justPaths)
// Given a path string, and a sorted raw list of files at the end of the path
// create a new list which contains the processed versions of each entry
// (place PATH_SEP after directories, eliminate files if justPaths is true).
// If there is a problem, ErrorSet, return NULL.
// NOTE: this does not dispose of the incoming list.
// the output list must be disposed of by calling StringListFree.
{
	char
		**outList;
	char
		tempPath[MAXPATHLEN+1];
	int
		length;
	uint32_t
		inElement;
	struct stat
		statOut;

	*outElements=0;
	if((outList=StringListNew()))
	{
		for(inElement=0;inElement<inElements;inElement++)
		{
			PathAndFileConcat(pathString,list[inElement],tempPath);
			if(stat(tempPath,&statOut)!=-1)				// if this fails, then do not copy the name
			{
				if(S_ISDIR(statOut.st_mode))
				{
					length=strlen(list[inElement]);
					if(length<MAXPATHLEN)
					{
						strcpy(tempPath,list[inElement]);
						tempPath[length++]=PATH_SEP;	// drop PATH_SEP onto the string
						tempPath[length]='\0';
						StringListAdd(tempPath,&outList,outElements);	// ignore errors here
					}
				}
				else
				{
					if(!justPaths)						// see if file names should be added
					{
						StringListAdd(list[inElement],&outList,outElements);	// ignore errors here
					}
				}
			}
		}
	}
	return(outList);
}

static void SelectionFilesDispose(char **list,bool *selectionList)
// Dispose of the two lists returned by SelectionFilesGet.
{
	MDisposePtr(selectionList);
	StringListFree(list);
}

static char **SelectionFilesGet(const char *pathString,uint32_t *numElements,bool **selectionList,bool justPaths)
// Given pathString, create a list of the files/directories at the end of the path.
// SelectionFilesDispose must be called on list, and selectionList to dispose of them.
// If there is a problem, ErrorSet, return NULL.
{
	char
		**rawList,
		**list;
	uint32_t
		rawElements;

	if((rawList=GlobAll(pathString,&rawElements)))
	{
		list=DirectoryListProcess(pathString,rawList,rawElements,numElements,justPaths);
		StringListFree(rawList);					// get rid of the raw list
		if(list)
		{
			if((*selectionList=(bool *)MNewPtrClr((*numElements)*sizeof(bool))))
			{
				return(list);
			}
			StringListFree(list);
		}
	}

	*selectionList=NULL;
	*numElements=0;

	return(NULL);
}

static void FileListPress(DIALOG_ITEM *item)
// When a selection has been made from the file list, this
// will decide what to do.
{
	FILE_DIALOG_PARAMETERS
		*dialogParameters;

	dialogParameters=(FILE_DIALOG_PARAMETERS *)item->dialog->dialogLocal;
	ButtonItemPress(dialogParameters->OkButton);		// behave as if Ok was pressed
}

static bool ListAddPathAndFile(const char *path,const char *file,char ***list,uint32_t *numElements)
// Like StringListAdd, but takes a path, and a filename which are
// concatenated into a complete path, and added to the list.
// If there is a problem, ErrorSet, and return false.
{
	char
		fullPath[MAXPATHLEN+1];

	PathAndFileConcat(path,file,fullPath);
	return(StringListAdd(fullPath,list,numElements));
}

static bool DialogFileOkInText(DIALOG *dialog,FILE_DIALOG_PARAMETERS *dialogParameters)
// Ok was pressed while the text was active.
{
	uint64_t
		bytesToExtract;
	char
		globPath[MAXPATHLEN+1];
	char
		**globList;
	uint32_t
		currentElement,
		globElements;
	bool
		fail;

	fail=false;

	if((dialogParameters->outputList=StringListNew()))
	{
		dialogParameters->outputElements=0;
		bytesToExtract=TextUniverseTotalBytes(dialogParameters->buffer->textUniverse);
		if(bytesToExtract>MAXPATHLEN)
		{
			bytesToExtract=MAXPATHLEN;
		}
		TextUniverseTextExtract(dialogParameters->buffer->textUniverse,0,(uint8_t *)globPath,bytesToExtract,NULL);
		if(bytesToExtract<MAXPATHLEN)
		{
			globPath[bytesToExtract]='\0';
		}
		else
		{
			globPath[MAXPATHLEN]='\0';
		}
		if((globList=Glob(&(dialogParameters->pathString[0]),globPath,false,true,&globElements)))
		{
			for(currentElement=0;!fail&&currentElement<globElements;currentElement++)
			{
				fail=!ListAddPathAndFile(&(dialogParameters->pathString[0]),globList[currentElement],&dialogParameters->outputList,&dialogParameters->outputElements);
			}
			StringListFree(globList);
		}
		else
		{
			fail=true;
		}
		if(fail)
		{
			StringListFree(dialogParameters->outputList);
			dialogParameters->outputElements=0;
		}
	}
	return(!fail);
}

static bool DialogFileOkInList(DIALOG *dialog,FILE_DIALOG_PARAMETERS *dialogParameters)
// Ok was pressed while the list was active.
// Create an output list which contains all the selected items.
// If there is a problem, ErrorSet, return false.
// NOTE: to make the interface seem more intuitive, if there
// are no selected items in the list, this will revert to making
// a list from the text box.
{
	uint32_t
		currentElement;
	bool
		noneSelected,
		fail;

	fail=false;
	noneSelected=true;
	if((dialogParameters->outputList=StringListNew()))
	{
		dialogParameters->outputElements=0;
		for(currentElement=0;!fail&&currentElement<dialogParameters->numElements;currentElement++)
		{
			if(dialogParameters->selectionList[currentElement])	// is this one selected?
			{
				noneSelected=false;
				fail=!ListAddPathAndFile(&(dialogParameters->pathString[0]),dialogParameters->list[currentElement],&dialogParameters->outputList,&dialogParameters->outputElements);
			}
		}
		if(fail||noneSelected)
		{
			StringListFree(dialogParameters->outputList);
			dialogParameters->outputElements=0;
		}
	}
	if(!fail)
	{
		if(noneSelected)									// if not failure, but nothing selected, use dialog text
		{
			return(DialogFileOkInText(dialog,dialogParameters));
		}
		else
		{
			return(true);
		}
	}
	return(false);
}

static void DialogFileOkPress(DIALOG_ITEM *item)
// The user pressed "Ok" in the dialog (or it was pressed for him)
// take appropriate action.
{
	DIALOG
		*dialog;
	DIR
		*directory;
	FILE_DIALOG_PARAMETERS
		*dialogParameters;
	bool
		fail;

	ErrorTraceClear();
	dialog=item->dialog;
	dialogParameters=(FILE_DIALOG_PARAMETERS *)dialog->dialogLocal;
	fail=false;
	if(dialog->focusItem==dialogParameters->listBox)		// if list has focus, then build a list out of what's selected
	{
		fail=!DialogFileOkInList(dialog,dialogParameters);
	}
	else
	{
		fail=!DialogFileOkInText(dialog,dialogParameters);
	}

	if(!fail)
	{
		if(dialogParameters->outputElements==1)				// if only one element, then see if directory
		{
			if((directory=opendir(dialogParameters->outputList[0])))	// see if it is a directory, and we are allowed to open it
			{
				closedir(directory);						// did not want it open, just wanted to see if we could
				if(!(dialog->focusItem==dialogParameters->listBox))
				{
					EditorBufferClear(dialogParameters->buffer);		// get rid of any text in the view
				}
				if(dialogParameters->list)
				{
					SelectionFilesDispose(dialogParameters->list,dialogParameters->selectionList);
				}
				realpath2(dialogParameters->outputList[0],&(dialogParameters->pathString[0]));	// this is the new path
				dialogParameters->list=SelectionFilesGet(&(dialogParameters->pathString[0]),&dialogParameters->numElements,&dialogParameters->selectionList,dialogParameters->justPaths);
				StaticTextItemTextReset(dialogParameters->pathText,&(dialogParameters->pathString[0]));
				DialogWindowAdjustForNewSize(dialog->parentWindow);		// changing the path can cause elements to resize
				ListBoxItemListsReset(dialogParameters->listBox,dialogParameters->numElements,(const char **)dialogParameters->list,dialogParameters->selectionList);
				MultipleClicksReset();						// reset multiple click timers
			}
			else
			{
				if(errno==ENOTDIR||errno==ENOENT)
				{
					if(!dialogParameters->justPaths)
					{
						dialog->dialogComplete=true;		// if not a directory, or known file, return it
					}
					else
					{
						PlatformMessageReport("%s\nNot a directory\n",dialogParameters->outputList[0]);	// let user know what's wrong
					}
				}
				else
				{
					ErrorStdCLibSet();						// report whatever error we got
					PlatformMessageReport("%s\n",ErrorTraceGet());	// let user know what's wrong
				}
			}
		}
		else
		{
			if(dialogParameters->justOne||dialogParameters->justPaths)
			{
				PlatformMessageReport("Path is ambiguous\n");		// got more elements than asked for, so complain, and keep going
			}
			else
			{
				dialog->dialogComplete=true;				// if anything other that 1 element, pass it back as the result
			}
		}
		if(!dialog->dialogComplete)
		{
			StringListFree(dialogParameters->outputList);	// not leaving, so kill the list
		}
	}
	else
	{
		PlatformMessageReport("%s\n",ErrorTraceGet());		// let user know what's wrong
	}
}

static bool DialogFile(const char *title,char *fullPath,uint32_t stringBytes,char ***list,bool justOne,bool justPaths,bool *cancel)
// Bring up a file open/save dialog.
// If there is a problem, ErrorSet, return false.
// Otherwise, cancel is true if the user canceled the dialog.
//
// Otherwise, if justPaths is true, return the final path chosen in the
// dialog in fullPath, truncating to stringBytes in length.
//
// Otherwise, listElements contains the pointer to the start of an array of
// pointers to char which are the selections.
// It must be disposed of at some later time with a call to DialogFilePathsFree.
// NOTE: the array is terminated with a pointer to NULL.
{
	DIALOG
		*dialog;
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	LIST_BOX_DESCRIPTOR
		listBoxDescriptor;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	FILE_DIALOG_PARAMETERS
		dialogParameters;
	bool
		result;
	char
		filePart[MAXPATHLEN+1];

	result=false;
	*cancel=false;

	dialogParameters.justOne=justOne;
	dialogParameters.justPaths=justPaths;

	if((dialogParameters.buffer=EditorBufferNew("file dialog")))			// create a buffer in which to collect the file name
	{
		PathAndFileSplit(fullPath,dialogParameters.pathString,filePart);	// split the incoming name into file and path components
		if(TextUniverseInsertFromText((uint8_t *)filePart,dialogParameters.buffer->textUniverse,TextUniverseTotalBytes(dialogParameters.buffer->textUniverse),strlen(filePart)))
		{
			if(EditorSelectAll(dialogParameters.buffer,dialogParameters.buffer->selectionUniverse))	// select it all
			{
				if((dialog=DialogCreateCentered(600,700,440,470," ",dialogBackgroundColor)))		// make new empty dialog
				{
					dialog->dialogLocal=&dialogParameters;				// hang parameter list off of the dialog

					dialogParameters.list=SelectionFilesGet(&(dialogParameters.pathString[0]),&dialogParameters.numElements,&dialogParameters.selectionList,justPaths);	// ignore failure, since list box will handle a NULL list

					textDescriptor.positionFunction=DialogFileStatTextTitlePosition;
					textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
					textDescriptor.string=title;
					textDescriptor.fontName=dialogStaticTextFontName;
					if(DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor))
					{
						textDescriptor.positionFunction=DialogFileStatTextPathPosition;
						textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
						textDescriptor.string=&(dialogParameters.pathString[0]);
						textDescriptor.fontName=dialogStaticTextFontName;
						if((dialogParameters.pathText=DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor)))
						{
							textBoxDescriptor.positionFunction=DialogFileTextBoxPosition;
							textBoxDescriptor.buffer=dialogParameters.buffer;
							textBoxDescriptor.tabSize=4;
							textBoxDescriptor.fontName=dialogTextBoxFontName;
							textBoxDescriptor.colorSchemeName=FILE_TEXT_COLOR_SCHEME_NAME;
							textBoxDescriptor.activeBorderColor=dialogTextBoxActiveBorderColor;
							textBoxDescriptor.inactiveBorderColor=dialogTextBoxInactiveBorderColor;
							if((dialogParameters.textBox=DialogItemCreate(dialog,TextBoxItemCreate,&textBoxDescriptor)))
							{
								listBoxDescriptor.positionFunction=DialogFileListBoxPosition;
								listBoxDescriptor.activeBackgroundColor=dialogListBoxActiveBackgroundColor;
								listBoxDescriptor.activeForegroundColor=dialogListBoxActiveForegroundColor;
								listBoxDescriptor.activeBorderColor=dialogListBoxActiveBorderColor;
								listBoxDescriptor.inactiveBackgroundColor=dialogListBoxInactiveBackgroundColor;
								listBoxDescriptor.inactiveForegroundColor=dialogListBoxInactiveForegroundColor;
								listBoxDescriptor.inactiveBorderColor=dialogListBoxInactiveBorderColor;
								listBoxDescriptor.numElements=dialogParameters.numElements;
								listBoxDescriptor.listElements=(const char **)dialogParameters.list;
								listBoxDescriptor.selectedElements=dialogParameters.selectionList;
								listBoxDescriptor.fontName=dialogListBoxFontName;
								listBoxDescriptor.pressedFunction=FileListPress;
								if((dialogParameters.listBox=DialogItemCreate(dialog,ListBoxItemCreate,&listBoxDescriptor)))
								{
									buttonDescriptor.positionFunction=DialogFileButtonOkPosition;
									buttonDescriptor.buttonName="Ok";
									buttonDescriptor.image=buttonImage;
									buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
									buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
									buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
									buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
									buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
									buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
									buttonDescriptor.shadowColor=dialogButtonShadowColor;
									buttonDescriptor.fontName=dialogButtonFontName;
									buttonDescriptor.hasKey=true;
									buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
									buttonDescriptor.pressedFunction=DialogFileOkPress;
									buttonDescriptor.pressedState=NULL;
									if((dialogParameters.OkButton=DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor)))	// remember this button
									{
										buttonDescriptor.positionFunction=DialogFileButtonCancelPosition;
										buttonDescriptor.buttonName="Cancel";
										buttonDescriptor.image=buttonImage;
										buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
										buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
										buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
										buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
										buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
										buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
										buttonDescriptor.shadowColor=dialogButtonShadowColor;
										buttonDescriptor.fontName=dialogButtonFontName;
										buttonDescriptor.hasKey=true;
										buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
										buttonDescriptor.pressedFunction=DialogExitPress;
										buttonDescriptor.pressedState=cancel;
										if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
										{
											if(justPaths)			// add another button for paths (use enter to select)
											{
												buttonDescriptor.positionFunction=DialogFileButtonAcceptPathPosition;
												buttonDescriptor.buttonName="Accept Path";
												buttonDescriptor.image=buttonImage;
												buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
												buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
												buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
												buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
												buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
												buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
												buttonDescriptor.shadowColor=dialogButtonShadowColor;
												buttonDescriptor.fontName=dialogButtonFontName;
												buttonDescriptor.hasKey=true;
												buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_KP_Enter);
												buttonDescriptor.pressedFunction=DialogExitPress;
												buttonDescriptor.pressedState=NULL;
												if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
												{
													DialogEventsHandle(dialog);	// deal with this dialog
													if(dialog->dialogAbort)
													{
														(*cancel)=true;			// return cancel if user closed dialog window
													}
													result=true;
												}
											}
											else
											{
												DialogEventsHandle(dialog);	// deal with this dialog
												if(dialog->dialogAbort)
												{
													(*cancel)=true;			// return cancel if user closed dialog window
												}
												result=true;
											}
										}
									}
								}
							}
						}
					}
					if(dialogParameters.list)							// delete any lists that are hanging around
					{
						SelectionFilesDispose(dialogParameters.list,dialogParameters.selectionList);
					}
					DialogDispose(dialog);
				}
			}
		}
		EditorBufferClose(dialogParameters.buffer);
	}
	if(justPaths)
	{
		if(stringBytes)
		{
			strncpy(fullPath,dialogParameters.pathString,stringBytes);	// return the path
			fullPath[stringBytes-1]='\0';
		}
	}
	if(result&&!(*cancel)&&(!justPaths))
	{
		*list=dialogParameters.outputList;
	}
	return(result);
}

static void DialogFilePathsFree(char **paths)
// Free an array of paths that was returned from DialogFile.
{
	StringListFree(paths);
}

bool PlatformDialogOpenFile(const char *title,char *fullPath,uint32_t stringBytes,char ***listElements,bool *cancel)
// Bring up a file open dialog.
// If there is a problem, ErrorSet, return false.
// Otherwise, cancel is true if the user canceled the dialog.
//
// Otherwise, listElements contains the pointer to the start of an array of
// pointers to char which are the selections.
// It must be disposed of at some later time with a call to PlatformDialogFreeOpenFilePaths.
// NOTE: the array is terminated with a pointer to NULL.
//
// NOTE: this is NOT allowed to change the current directory of the editor.
{
	return(DialogFile(title,fullPath,stringBytes,listElements,false,false,cancel));
}

void PlatformDialogFreeOpenFilePaths(char **paths)
// free an array of paths that was returned from PlatformDialogOpenFile
{
	DialogFilePathsFree(paths);
}

bool PlatformDialogSaveFile(const char *title,char *fullPath,uint32_t stringBytes,bool *cancel)
// Bring up a file save dialog.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true, the user canceled.
// Otherwise, return the full path name of the chosen file.
// NOTE: if the fullPath (including terminator) would be larger than stringBytes,
// then it will be truncated to fit.
// NOTE ALSO: fullPath may be sent to this routine with a pathname, and the save
// dialog will open to that path (as best as it can).
//
// NOTE: this is NOT allowed to change the current directory of the editor.
{
	char
		**paths;
	bool
		result;

	result=false;
	if(DialogFile(title,fullPath,stringBytes,&paths,true,false,cancel))
	{
		if(!(*cancel))
		{
			if(stringBytes)
			{
				strncpy(fullPath,paths[0],stringBytes);
				fullPath[stringBytes-1]='\0';
			}
			DialogFilePathsFree(paths);
		}
		result=true;
	}
	return(result);
}

bool PlatformDialogChoosePath(const char *title,char *fullPath,uint32_t stringBytes,bool *cancel)
// Bring up a path choose dialog.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true, the user canceled.
// Otherwise, return the full path name of the chosen path.
// NOTE: if the fullPath (including terminator) would be larger than stringBytes,
// then it will be truncated to fit.
// NOTE ALSO: fullPath may be sent to this routine with a pathname, and the
// dialog will open to that path (as best as it can).
//
// NOTE: this is NOT allowed to change the current directory of the editor.
{
	return(DialogFile(title,fullPath,stringBytes,NULL,true,true,cancel));
}

// ----------------------------------------------------------------------------

typedef struct
{
	DIALOG_ITEM
		*staticText,
		*okButton;
} CHOOSE_FONT_LOCAL;

static void DialogChooseFontStatTextPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the static text area.
{
	rect->x=10;
	rect->y=10;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=StaticTextItemHeightGet(item,rect->w);	// work out the ideal height

	if(rect->h>item->dialog->parentWindow->h/2)		// limit the height if it is too large
	{
		rect->h=item->dialog->parentWindow->h/2;
	}
}

static void DialogChooseFontListBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the list box area.
{
	CHOOSE_FONT_LOCAL
		*local;
	EDITOR_RECT
		staticTextRect;
	int32_t
		bottomY;

	local=(CHOOSE_FONT_LOCAL *)item->dialog->dialogLocal;
	StaticTextItemRectangleGet(local->staticText,&staticTextRect);
	rect->x=10;
	rect->y=staticTextRect.y+staticTextRect.h+5;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	bottomY=item->dialog->parentWindow->h-35-15-20-15-30-15;	// subtract off button height and separation distance, checkbox height and separation distance, sizebox height and separation distance

	if(bottomY>rect->y)
	{
		rect->h=bottomY-rect->y;
	}
	else
	{
		rect->h=0;
	}
}

static void DialogChooseFontSizeTextBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the font size text area.
{
	rect->x=10;
	rect->y=item->dialog->parentWindow->h-35-15-20-15-30;
	if(item->dialog->parentWindow->w>20)
	{
		rect->w=item->dialog->parentWindow->w-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=30;
}

static void DialogChooseFontBoldCheckBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the font bold check box area.
{
	rect->x=10;

	rect->y=item->dialog->parentWindow->h-35-15-20;

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogChooseFontObliqueCheckBoxPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the font oblique check box area.
{
	rect->x=item->dialog->parentWindow->w/2+10;

	rect->y=item->dialog->parentWindow->h-35-15-20;

	if(item->dialog->parentWindow->w/2>20)
	{
		rect->w=item->dialog->parentWindow->w/2-20;
	}
	else
	{
		rect->w=0;
	}
	rect->h=20;
}

static void DialogChooseFontButtonOkPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the ok button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogChooseFontButtonCancelPosition(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Position the cancel button.
{
	uint32_t
		width,
		height;

	ButtonItemDimensions(item,&width,&height);

	rect->w=width;		// set the width and height of the button
	rect->h=height;

	rect->x=4*item->dialog->parentWindow->w/5-rect->w/2;

	if(item->dialog->parentWindow->h>35)
	{
		rect->y=item->dialog->parentWindow->h-35;
	}
	else
	{
		rect->y=0;
	}
}

static void DialogChooseFontExitPressFunction(DIALOG_ITEM *item)
// When the simple list is double clicked in, it is just like
// pressing the Ok button, so we do that!
{
	CHOOSE_FONT_LOCAL
		*local;

	local=(CHOOSE_FONT_LOCAL *)item->dialog->dialogLocal;
	ButtonItemPress(local->okButton);
}

bool PlatformDialogChooseFont(const char *title,char *font,uint32_t stringBytes,bool *cancel)
// Bring up a font choose dialog.
// If there is a problem, ErrorSet, return false.
// Otherwise, if cancel is true, the user canceled.
// Otherwise, return the name of the chosen font in font.
// NOTE: font is a composite font name.
// NOTE: if the font (including terminator) would be larger than stringBytes,
// then it will be truncated to fit.
// NOTE ALSO: font may be sent to this routine with a font name, and the
// dialog will open to that font (as best as it can).
{
	char
		**fontList;
	uint32_t
		numElements;
	uint32_t
		i;
	bool
		found;
	EDITOR_BUFFER
		*buffer;
	char
		familyName[MAX_FONT_NAME_LENGTH];
	char
		compositeName[MAX_FONT_NAME_LENGTH];
	uint64_t
		bytesToExtract;
	char
		tempString[256];
	uint32_t
		size;
	bool
		bold,
		oblique;
	CHOOSE_FONT_LOCAL
		local;															// local state that callbacks can get to
	DIALOG
		*dialog;
	STATIC_TEXT_DESCRIPTOR
		textDescriptor;
	LIST_BOX_DESCRIPTOR
		listBoxDescriptor;
	TEXT_BOX_DESCRIPTOR
		textBoxDescriptor;
	CHECK_BOX_DESCRIPTOR
		checkDescriptor;
	BUTTON_DESCRIPTOR
		buttonDescriptor;
	bool
		result;

	result=false;
	*cancel=false;

	PlatformFontCompositeNameSplit(font,familyName,&size,&bold,&oblique);	// break incoming composite font name into components

	if((fontList=FontListCreate(&numElements)))
	{
		bool
			selectedElements[numElements];

		for(i=0;i<numElements;i++)
		{
			if(strcasecmp(fontList[i],familyName)==0)
			{
				selectedElements[i]=true;
			}
			else
			{
				selectedElements[i]=false;
			}
		}
		if((buffer=EditorBufferNew("choose font dialog")))						// create a buffer in which to collect the font size
		{
			snprintf(tempString,sizeof(tempString),"%d",size);					// get the size as a string
			if(TextUniverseInsertFromText((uint8_t *)tempString,buffer->textUniverse,TextUniverseTotalBytes(buffer->textUniverse),strlen(tempString)))
			{
				if(EditorSelectAll(buffer,buffer->selectionUniverse))			// select it all
				{
					if((dialog=DialogCreateCentered(600,700,600,300," ",dialogBackgroundColor)))	// make new empty dialog
					{
						dialog->dialogLocal=(void *)&local;

						textDescriptor.positionFunction=DialogChooseFontStatTextPosition;
						textDescriptor.foregroundColor=dialogStaticTextForegroundColor;
						textDescriptor.string=title;
						textDescriptor.fontName=dialogStaticTextFontName;
						if((local.staticText=DialogItemCreate(dialog,StaticTextItemCreate,&textDescriptor)))
						{
							listBoxDescriptor.positionFunction=DialogChooseFontListBoxPosition;
							listBoxDescriptor.activeBackgroundColor=dialogListBoxActiveBackgroundColor;
							listBoxDescriptor.activeForegroundColor=dialogListBoxActiveForegroundColor;
							listBoxDescriptor.activeBorderColor=dialogListBoxActiveBorderColor;
							listBoxDescriptor.inactiveBackgroundColor=dialogListBoxInactiveBackgroundColor;
							listBoxDescriptor.inactiveForegroundColor=dialogListBoxInactiveForegroundColor;
							listBoxDescriptor.inactiveBorderColor=dialogListBoxInactiveBorderColor;
							listBoxDescriptor.numElements=numElements;
							listBoxDescriptor.listElements=(const char **)fontList;
							listBoxDescriptor.selectedElements=selectedElements;
							listBoxDescriptor.fontName=dialogListBoxFontName;
							listBoxDescriptor.pressedFunction=DialogChooseFontExitPressFunction;
							if(DialogItemCreate(dialog,ListBoxItemCreate,&listBoxDescriptor))
							{
								textBoxDescriptor.positionFunction=DialogChooseFontSizeTextBoxPosition;
								textBoxDescriptor.buffer=buffer;
								textBoxDescriptor.tabSize=4;
								textBoxDescriptor.fontName=dialogTextBoxFontName;
								textBoxDescriptor.colorSchemeName=SIMPLE_TEXT_COLOR_SCHEME_NAME;
								textBoxDescriptor.activeBorderColor=dialogTextBoxActiveBorderColor;
								textBoxDescriptor.inactiveBorderColor=dialogTextBoxInactiveBorderColor;
								if(DialogItemCreate(dialog,TextBoxItemCreate,&textBoxDescriptor))
								{
									checkDescriptor.positionFunction=DialogChooseFontBoldCheckBoxPosition;
									checkDescriptor.checkBoxName="Bold";
									checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
									checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
									checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
									checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
									checkDescriptor.fontName=dialogCheckBoxFontName;
									checkDescriptor.active=true;
									checkDescriptor.hasKey=false;
									checkDescriptor.pressedFunction=NULL;
									checkDescriptor.pressedState=&bold;
									if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
									{
										checkDescriptor.positionFunction=DialogChooseFontObliqueCheckBoxPosition;
										checkDescriptor.checkBoxName="Oblique";
										checkDescriptor.activeBackgroundColor=dialogCheckBoxActiveBackgroundColor;
										checkDescriptor.activeForegroundColor=dialogCheckBoxActiveForegroundColor;
										checkDescriptor.inactiveBackgroundColor=dialogCheckBoxInactiveBackgroundColor;
										checkDescriptor.inactiveForegroundColor=dialogCheckBoxInactiveForegroundColor;
										checkDescriptor.fontName=dialogCheckBoxFontName;
										checkDescriptor.active=true;
										checkDescriptor.hasKey=false;
										checkDescriptor.pressedFunction=NULL;
										checkDescriptor.pressedState=&oblique;
										if(DialogItemCreate(dialog,CheckBoxItemCreate,&checkDescriptor))
										{
											buttonDescriptor.positionFunction=DialogChooseFontButtonOkPosition;
											buttonDescriptor.buttonName="Ok";
											buttonDescriptor.image=buttonImage;
											buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
											buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
											buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
											buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
											buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
											buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
											buttonDescriptor.shadowColor=dialogButtonShadowColor;
											buttonDescriptor.fontName=dialogButtonFontName;
											buttonDescriptor.hasKey=true;
											buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Return);
											buttonDescriptor.pressedFunction=DialogExitPress;
											buttonDescriptor.pressedState=NULL;
											if((local.okButton=DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor)))	// remember this button
											{
												buttonDescriptor.positionFunction=DialogChooseFontButtonCancelPosition;
												buttonDescriptor.buttonName="Cancel";
												buttonDescriptor.image=buttonImage;
												buttonDescriptor.activeBackgroundColor=dialogButtonActiveBackgroundColor;
												buttonDescriptor.activeForegroundColor=dialogButtonActiveForegroundColor;
												buttonDescriptor.inactiveBackgroundColor=dialogButtonInactiveBackgroundColor;
												buttonDescriptor.inactiveForegroundColor=dialogButtonInactiveForegroundColor;
												buttonDescriptor.pressedBackgroundColor=dialogButtonPressedBackgroundColor;
												buttonDescriptor.pressedForegroundColor=dialogButtonPressedForegroundColor;
												buttonDescriptor.shadowColor=dialogButtonShadowColor;
												buttonDescriptor.fontName=dialogButtonFontName;
												buttonDescriptor.hasKey=true;
												buttonDescriptor.keyCode=XKeysymToKeycode(xDisplay,XK_Escape);
												buttonDescriptor.pressedFunction=DialogExitPress;
												buttonDescriptor.pressedState=cancel;
												if(DialogItemCreate(dialog,ButtonItemCreate,&buttonDescriptor))
												{
													DialogEventsHandle(dialog);					// deal with this dialog
													if(!dialog->dialogAbort)					// if user didn't abort...
													{
														found=false;
														for(i=0;i<numElements&&!found;i++)		// hunt down the first selected element
														{
															if(selectedElements[i])
															{
																if((bytesToExtract=TextUniverseTotalBytes(buffer->textUniverse)))	// if the text wasn't cleared, extract a size from it
																{
																	if(bytesToExtract>sizeof(tempString)-1)
																	{
																		bytesToExtract=sizeof(tempString)-1;
																	}
																	TextUniverseTextExtract(buffer->textUniverse,0,(uint8_t *)tempString,bytesToExtract,NULL);
																	tempString[bytesToExtract]='\0';
																	size=strtol(tempString,NULL,0);	// get size back from buffer
																	if((int32_t)size<1)				// pin size to something >0
																	{
																		size=1;
																	}
																	if(size>255)					// also, don't let it get too huge
																	{
																		size=255;
																	}
																}
																PlatformFontComponentNameJoin(fontList[i],size,bold,oblique,compositeName);	// recombine
																strncpy(font,compositeName,stringBytes);
																font[stringBytes-1]='\0';
																found=true;
															}
														}
													}
													else
													{
														(*cancel)=true;							// return cancel if user closed dialog window
													}
													result=true;
												}
											}
										}
									}
								}
							}
						}
						DialogDispose(dialog);
					}
				}
			}
			EditorBufferClose(buffer);
		}
		FontListDispose(fontList);
	}
	return(result);
}

// ----------------------------------------------------------------------------

void PlatformDialogCenterPointSet(int32_t centerX,int32_t centerY)
// Set the absolute screen coordinates that will be used to center
// dialog boxes.
{
	dialogCenterX=centerX;
	dialogCenterY=centerY;
}

// ----------------------------------------------------------------------------

static bool DialogAddDefaultColorScheme(const char *schemeName)
// Create a default color scheme, and add it to the list of color schemes.
// If there's a problem, ErrorSet and return false.
{
	VIEW_COLOR_SCHEME
		*colorScheme;

	if((colorScheme=ViewColorSchemeCreate()))
	{
		ViewColorSchemeDrawColorsSet(colorScheme,true,GRAY_3,BLACK);
		ViewColorSchemeDrawColorsSet(colorScheme,false,GRAY_3,GRAY_1);
		ViewColorSchemeSelectionColorsSet(colorScheme,true,BLACK,GRAY_3);
		ViewColorSchemeSelectionColorsSet(colorScheme,false,GRAY_1,GRAY_3);
		if(ViewColorSchemeListAdd(colorScheme,schemeName))
		{
			return(true);
		}
		ViewColorSchemeDestroy(colorScheme);
	}
	return(false);
}

void DialogUnInit()
// Undo what DialogInit did.
// NOTE: this does not attempt to remove the view color schemes added
// by the init function.
{
}

bool DialogInit()
// Initialize dialog handling.
// NOTE: this must happen only after the editor is initialized, since it
// sets up view color schemes (an editor concept).
// NOTE: if this fails, it will ErrorSet and return false.
// It is possible for this to fail, and leave some of the color schemes it
// created on the global list. This isn't a problem, since color schemes
// are deleted from the list at editor exit.
{
	int32_t
		xOffset,
		yOffset;
	uint32_t
		screenWidth,
		screenHeight;

	// initialize the place where we'll pop up dialogs
	PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&screenWidth,&screenHeight);
	dialogCenterX=screenWidth/2;					// half the way across
	dialogCenterY=screenHeight/3;					// 1/3 the way down

	// set default dialog colors
	DialogColorsSet(BACKGROUND_COLOR);
	DialogStaticTextColorsSet(BLACK);
	DialogTextBoxColorsSet(WHITE,GRAY_3);
	DialogListBoxColorsSet(GRAY_3,BLACK,WHITE,GRAY_3,GRAY_1,GRAY_3);
	DialogCheckBoxColorsSet(GRAY_3,BLACK,GRAY_3,GRAY_1);
	DialogPopupColorsSet(GRAY_3,BLACK,GRAY_3,GRAY_1,BLACK);
	DialogButtonColorsSet(GRAY_3,BLACK,GRAY_3,GRAY_1,GRAY_1,BLACK,BLACK);

	// set defaults for all fonts
	DialogStaticTextFontSet(DEFAULT_STATIC_TEXT_FONT);
	DialogTextBoxFontSet(DEFAULT_TEXT_BOX_FONT);
	DialogListBoxFontSet(DEFAULT_LIST_BOX_FONT);
	DialogCheckBoxFontSet(DEFAULT_CHECK_BOX_FONT);
	DialogPopupFontSet(DEFAULT_POPUP_FONT);
	DialogButtonFontSet(DEFAULT_BUTTON_FONT);

	// set up default view color schemes
	if(DialogAddDefaultColorScheme(FIND_COLOR_SCHEME_NAME))
	{
		if(DialogAddDefaultColorScheme(REPLACE_COLOR_SCHEME_NAME))
		{
			if(DialogAddDefaultColorScheme(SIMPLE_TEXT_COLOR_SCHEME_NAME))
			{
				if(DialogAddDefaultColorScheme(FILE_TEXT_COLOR_SCHEME_NAME))
				{
					return(true);
				}
			}
		}
	}
	return(false);
}
