// Document window handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define		DEFAULT_STATUS_BAR_FONT		"dejavu sans condensed:size=11"	// default font to use for status bars

#define		STATUS_LEFT_MARGIN			10		// number of pixels from the left to start drawing status
#define		STATUS_RIGHT_MARGIN			10		// number of pixels from the right to start drawing status
#define		STATUS_MINIMUM_HEIGHT		16		// minimum height of the status line excluding STATUS_BAR_LINE_HEIGHT
#define		STATUS_BAR_LINE_HEIGHT		2		// number of pixels to use to draw line which separates status from view

typedef struct
{
	EDITOR_VIEW
		*view;									// the view onto a buffer
	bool
		verticalScrollBarOnLeft;				// gives the preference for vertical scroll bar placement
	SCROLL_BAR
		*verticalScrollBar,
		*horizontalScrollBar;
} GUI_DOCUMENT_WINDOW_INFO;

// Colors to use when drawing document windows.
static EDITOR_COLOR
	statusBackgroundColor,						// status bar colors
	statusForegroundColor,
	statusFlagActiveColor;

// Font to use when drawing document windows status bars.
static EDITOR_FONT
	*statusBarFont;								// font used in status bars

static bool DocumentWindowTitleSet(EDITOR_WINDOW *window,const char *title)
// Replace title of window.
// If there is a problem, return false.
{
	uint32_t
		length;

	// Set the name the old-fashioned way (which won't support UTF-8)
	XStoreName(xDisplay,window->xWindow,title);
	XSetIconName(xDisplay,window->xWindow,title);

	// Set the name the new-fangled way (which does support UTF-8, and should be picked up by mondern WMs)
	length=strlen(title);
	XChangeProperty(xDisplay,window->xWindow,wmNameAtom,utf8StringAtom,8,PropModeReplace,(unsigned char *)title,length);
	XChangeProperty(xDisplay,window->xWindow,iconNameAtom,utf8StringAtom,8,PropModeReplace,(unsigned char *)title,length);
	return(true);
}

static void DocumentWindowRectsGet(EDITOR_WINDOW *window,EDITOR_RECT *contentRect,EDITOR_RECT *viewRect,EDITOR_RECT *verticalRect,EDITOR_RECT *horizontalRect,EDITOR_RECT *statusRect)
// Get the rectangles for all the areas of a document window.
{
	uint32_t
		ascent,
		descent;
	uint32_t
		statusHeight;
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;

	contentRect->x=0;
	contentRect->y=0;
	contentRect->w=window->w;
	contentRect->h=window->h;

	statusRect->x=0;
	statusRect->y=0;
	statusRect->w=window->w;

	if(statusBarFont)
	{
		FontHeightsGet(statusBarFont,&ascent,&descent);
		statusHeight=ascent+descent;
	}
	else
	{
		statusHeight=0;
	}
	if(statusHeight<STATUS_MINIMUM_HEIGHT)
	{
		statusHeight=STATUS_MINIMUM_HEIGHT;
	}

	statusRect->h=statusHeight+STATUS_BAR_LINE_HEIGHT;

	viewRect->x=0;
	viewRect->y=statusRect->y+statusRect->h;				// leave room for status bar

	if(windowInfo->verticalScrollBarOnLeft)
	{
		viewRect->x+=SCROLL_BAR_WIDTH;						// if scroll bar on left, then step over it to begin the view
	}

	if(window->w>SCROLL_BAR_WIDTH)
	{
		viewRect->w=window->w-SCROLL_BAR_WIDTH;		// pull back for scroll bar, and right free area
	}
	else
	{
		viewRect->w=0;
	}
	if((uint32_t)window->h>(statusHeight+STATUS_BAR_LINE_HEIGHT+SCROLL_BAR_WIDTH))
	{
		viewRect->h=window->h-statusHeight-STATUS_BAR_LINE_HEIGHT-SCROLL_BAR_WIDTH;		// pull back for bottom scroll bar, and bottom free area
	}
	else
	{
		viewRect->h=0;										// view cant fit!
	}

	if(windowInfo->verticalScrollBarOnLeft)
	{
		verticalRect->x=0;
	}
	else
	{
		verticalRect->x=viewRect->x+viewRect->w;			// if on right, place just after the view
	}

	verticalRect->y=viewRect->y;
	verticalRect->w=SCROLL_BAR_WIDTH;
	verticalRect->h=viewRect->h;

	horizontalRect->x=viewRect->x;
	horizontalRect->y=viewRect->y+viewRect->h;
	horizontalRect->w=viewRect->w;
	horizontalRect->h=SCROLL_BAR_WIDTH;
}

static uint8_t HumanReadableNumber(char *string,uint64_t value)
// Render value into string as a human readable number (with commas every 3 digits).
// Return the number of characters that were output.
// NOTE: string must be at least 26 characters long (20 digits and 6 commas), since
// this is the largest number that can be represented in 64 bits.
// This does not output a terminating '\0'.
{
	uint8_t
		digitBuffer[20];
	uint8_t
		digitCount;
	uint8_t
		result;

	// work out how many digits the result will be, store them
	// in reverse order.
	digitCount=0;
	do
	{
		digitBuffer[digitCount]=value%10;
		digitCount++;
		value/=10;
	}
	while(value);

	// use the digits to generate the output
	result=0;
	while(digitCount)
	{
		digitCount--;
		string[result]=digitBuffer[digitCount]+'0';
		result++;
		if(digitCount&&((digitCount%3)==0))
		{
			string[result]=',';
			result++;
		}
	}
	return(result);
}

static uint8_t StatusLeftStringCreate(EDITOR_WINDOW *window,char *string)
// Create left-hand status string for window.
// NOTE: string must be large enough to hold the status.
// Return the number of bytes of string created (excluding termination).
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;
	EDITOR_BUFFER
		*buffer;
	uint8_t
		stringIndex;
	uint64_t
		startPosition,						// selection information
		endPosition,
		startLine,
		endLine,
		startLinePosition,
		endLinePosition,
		totalSegments,
		totalSpan;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;

	buffer=windowInfo->view->parentBuffer;

	EditorSelectionInfoGet(buffer,buffer->selectionUniverse,&startPosition,&endPosition,&startLine,&endLine,&startLinePosition,&endLinePosition,&totalSegments,&totalSpan);

	stringIndex=0;

	stringIndex+=sprintf(&string[stringIndex],"Lines:");
	stringIndex+=HumanReadableNumber(&string[stringIndex],TextUniverseTotalLines(buffer->textUniverse));
	stringIndex+=sprintf(&string[stringIndex]," Chars:");
	stringIndex+=HumanReadableNumber(&string[stringIndex],TextUniverseTotalBytes(buffer->textUniverse));

	if(!totalSegments)
	{
		stringIndex+=sprintf(&string[stringIndex]," Cursor(");
		stringIndex+=HumanReadableNumber(&string[stringIndex],startLine+1);
		stringIndex+=sprintf(&string[stringIndex],":");
		stringIndex+=HumanReadableNumber(&string[stringIndex],startPosition-startLinePosition);
		stringIndex+=sprintf(&string[stringIndex]," ");
		stringIndex+=HumanReadableNumber(&string[stringIndex],startPosition);
		stringIndex+=sprintf(&string[stringIndex],")");
	}
	else
	{
		stringIndex+=sprintf(&string[stringIndex]," Selections(Count:");
		stringIndex+=HumanReadableNumber(&string[stringIndex],totalSegments);
		stringIndex+=sprintf(&string[stringIndex]," TotalChars:");
		stringIndex+=HumanReadableNumber(&string[stringIndex],totalSpan);

		stringIndex+=sprintf(&string[stringIndex]," Start(");
		stringIndex+=HumanReadableNumber(&string[stringIndex],startLine+1);
		stringIndex+=sprintf(&string[stringIndex],":");
		stringIndex+=HumanReadableNumber(&string[stringIndex],startPosition-startLinePosition);
		stringIndex+=sprintf(&string[stringIndex]," ");
		stringIndex+=HumanReadableNumber(&string[stringIndex],startPosition);
		stringIndex+=sprintf(&string[stringIndex],")");

		stringIndex+=sprintf(&string[stringIndex]," End(");
		stringIndex+=HumanReadableNumber(&string[stringIndex],endLine+1);
		stringIndex+=sprintf(&string[stringIndex],":");
		stringIndex+=HumanReadableNumber(&string[stringIndex],endPosition-endLinePosition);
		stringIndex+=sprintf(&string[stringIndex]," ");
		stringIndex+=HumanReadableNumber(&string[stringIndex],endPosition);
		stringIndex+=sprintf(&string[stringIndex],"))");
	}
	return(stringIndex);
}

static uint8_t StatusRightStringCreate(EDITOR_WINDOW *window,char *string)
// Create right-hand status string for window.
// NOTE: string must be large enough to hold the status.
// Return the number of bytes of string created (excluding termination).
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;
	EDITOR_BUFFER
		*buffer;
	uint8_t
		stringIndex;
	SYNTAX_MAP
		*syntaxMap;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;

	buffer=windowInfo->view->parentBuffer;

	stringIndex=0;

	if(buffer->firstView)					// ### get tabsize for the *first* view -- when there are multiple views on a document window buffer, this will need to change
	{
		stringIndex+=sprintf(&string[stringIndex],"%s | Tabs:%" PRIu32,buffer->firstView->useUtf8?"UTF-8":"ASCII",buffer->firstView->tabSize);
	}
	if((syntaxMap=SyntaxMapGetAssigned(buffer)))
	{
		stringIndex+=sprintf(&string[stringIndex]," | %s",SyntaxMapNameGet(syntaxMap));
	}
	return(stringIndex);
}

static void StatusBarFlagDraw(GUI_IMAGE *imageBuffer,int32_t *currentX,int32_t y,uint32_t h,GUI_IMAGE *flagImage,EDITOR_COLOR backgroundColor)
// The status bar may contain various flags to indicate status like "document modified", or
// "task running". Call this to render a flag into the bar at *currentX. This will
// update *currentX to the width of the drawn flag.
{
	int32_t
		drawY;

	drawY=y+((int32_t)h-flagImage->height)/2;			// center the image in the draw area

	ImageFill(imageBuffer,backgroundColor,*currentX,y,flagImage->width,h);
	ImageFill(imageBuffer,BLACK,(*currentX)+flagImage->width,y,1,h);	// draw a separator bar

	ImageDraw(flagImage,imageBuffer,*currentX,drawY);

	(*currentX)+=flagImage->width+1;
}

static void StatusBarDraw(EDITOR_WINDOW *window,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the status bar for window.
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;
	EDITOR_BUFFER
		*buffer;
	int
		linePosition;
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	uint32_t
		ascent,
		descent;
	int32_t
		drawX;
	const char
		*commandString;
	char
		statusString[256];										// the status string
	uint8_t
		statusStringLength;
	uint32_t
		statusStringWidth;
	uint32_t
		fontHeight;
	int32_t
		fontY;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;

	buffer=windowInfo->view->parentBuffer;

	DocumentWindowRectsGet(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	if(XRectInRegion(invalidRegion,statusRect.x,statusRect.y,statusRect.w,statusRect.h)!=RectangleOut)
	{
		ImageClipRegionRectSet(imageBuffer,&statusRect);			// don't draw outside the status bar

		ImageFill(imageBuffer,statusBackgroundColor,statusRect.x,statusRect.y,statusRect.w,statusRect.h);	// fill with background

		drawX=statusRect.x;											// get place to start drawing
		StatusBarFlagDraw(imageBuffer,&drawX,statusRect.y,statusRect.h-STATUS_BAR_LINE_HEIGHT,bufferModifiedImage,UndoCleanPointAt(buffer)?statusBackgroundColor:statusFlagActiveColor);
		StatusBarFlagDraw(imageBuffer,&drawX,statusRect.y,statusRect.h-STATUS_BAR_LINE_HEIGHT,taskRunningImage,buffer->task?statusFlagActiveColor:statusBackgroundColor);

		if(statusBarFont)
		{
			FontHeightsGet(statusBarFont,&ascent,&descent);

			drawX+=STATUS_LEFT_MARGIN;
			fontHeight=ascent+descent;
			fontY=statusRect.y+(((statusRect.h-STATUS_BAR_LINE_HEIGHT)-fontHeight)/2)+ascent;

			statusStringLength=StatusLeftStringCreate(window,statusString);
			FontTextUtf8Draw(statusBarFont,imageBuffer,statusForegroundColor,drawX,fontY,statusString,statusStringLength);
			drawX+=FontTextUtf8Width(statusBarFont,statusString,statusStringLength);

			if(buffer->task)
			{
				commandString=PlatformTaskCommandLineGet(buffer->task);
				statusStringLength=0;
				while((statusStringLength<sizeof(statusString)-1)&&((*commandString)>=' '))			// copy to terminator, or control code
				{
					statusString[statusStringLength++]=*commandString;
					commandString++;
				}
				statusString[statusStringLength]='\0';												// terminate string
			}
			else
			{
				statusStringLength=StatusRightStringCreate(window,statusString);
			}
			statusStringWidth=FontTextUtf8Width(statusBarFont,statusString,statusStringLength);			// find out how wide the string is to right-justify it in the status bar
			if(drawX+statusStringWidth+STATUS_LEFT_MARGIN+STATUS_RIGHT_MARGIN<=statusRect.w)		// make sure we can draw the right string w/o overlapping the left string
			{
				FontTextUtf8Draw(statusBarFont,imageBuffer,statusForegroundColor,statusRect.w-statusStringWidth-STATUS_RIGHT_MARGIN,fontY,statusString,statusStringLength);
			}
		}

		// draw the line that separates the status bar from the view
		linePosition=statusRect.y+statusRect.h-STATUS_BAR_LINE_HEIGHT;
		ImageFill(imageBuffer,BLACK,statusRect.x,linePosition,statusRect.w,1);
		linePosition+=STATUS_BAR_LINE_HEIGHT-1;
		ImageFill(imageBuffer,WHITE,statusRect.x,linePosition,statusRect.w,1);

		ImageClipRegionSet(imageBuffer,None);						// unset clip area
	}
}

static void GrowAreaDraw(EDITOR_WINDOW *window,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the grow box area of the window.
{
	XRectangle
		growBox;
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	DocumentWindowRectsGet(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	growBox.x=verticalRect.x;
	growBox.y=horizontalRect.y;
	growBox.width=verticalRect.w;
	growBox.height=horizontalRect.h;
	ImageFill(imageBuffer,SCROLL_BAR_BACKGROUND,growBox.x,growBox.y,growBox.width,growBox.height);
}

static void DocumentWindowStatusAdjust(EDITOR_WINDOW *window)
// Adjust the status of the document window.
// This includes updating the status bar, and anything else status related.
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	DocumentWindowRectsGet(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);
	statusRect.h-=STATUS_BAR_LINE_HEIGHT;							// just invalidate the status area
	WindowRectInvalidate(window,&statusRect);
}

static void DocumentWindowViewStatusAdjust(EDITOR_VIEW *view)
// Adjust the status bar of the document window with the given view.
{
	DocumentWindowStatusAdjust(view->parentWindow);
}

static void DocumentWindowScrollThumbsAdjust(EDITOR_VIEW *view)
// Look at the data of the passed view, and
// adjust the thumbs of its scroll bars so that
// they are correct.
{
	uint32_t
		scrollBarLeftPixel;

	ScrollBarAdjust(((GUI_DOCUMENT_WINDOW_INFO *)view->parentWindow->windowInfo)->verticalScrollBar,TextUniverseTotalLines(view->parentBuffer->textUniverse),((PLATFORM_VIEW_INFO *)view->viewInfo)->topLine);
	// NOTE: since views allow for having a left pixel <0, but the scroll bars do not, peg the bar at 0
	if(((PLATFORM_VIEW_INFO *)view->viewInfo)->leftPixel>=0)
	{
		scrollBarLeftPixel=(uint32_t)((PLATFORM_VIEW_INFO *)view->viewInfo)->leftPixel;
	}
	else
	{
		scrollBarLeftPixel=0;
	}
	// since the view is allowed to scroll beyond the length of the horizontal scroll bar, pin the bar at the limit
	if(((PLATFORM_VIEW_INFO *)view->viewInfo)->leftPixel>H_SCROLL_MAX_WIDTH)
	{
		scrollBarLeftPixel=H_SCROLL_MAX_WIDTH;
	}
	ScrollBarAdjust(((GUI_DOCUMENT_WINDOW_INFO *)view->parentWindow->windowInfo)->horizontalScrollBar,H_SCROLL_MAX_WIDTH,scrollBarLeftPixel);
}

static void DocumentWindowScrollThumbsAndStatusAdjust(EDITOR_VIEW *view)
// Adjust the parts of the window that change when the view changes.
{
	DocumentWindowScrollThumbsAdjust(view);
	DocumentWindowViewStatusAdjust(view);
}

static void DocumentWindowDraw(EDITOR_WINDOW *window,GUI_IMAGE *imageBuffer,Region invalidRegion)
// A document window needs to be updated, this will draw it.
{
	EDITOR_VIEW
		*view;

	StatusBarDraw(window,imageBuffer,invalidRegion);
	view=((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view;
	ViewDraw(view,imageBuffer,invalidRegion);								// draw this view
	ScrollBarDraw(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar,imageBuffer,invalidRegion);	// draw the scroll bars
	ScrollBarDraw(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar,imageBuffer,invalidRegion);
	GrowAreaDraw(window,imageBuffer,invalidRegion);
}

static void DocumentWindowUpdate(EDITOR_WINDOW *window)
// Make sure the window is up to date.
{
	Window
		xWindow;
	GUI_IMAGE
		*imageBuffer;

	if(!XEmptyRegion(window->invalidRegion))							// see if it has an invalid area
	{
		xWindow=window->xWindow;
		if((imageBuffer=ImageCreate(window->w,window->h)))				// go get a pixmap
		{
			DocumentWindowDraw(window,imageBuffer,window->invalidRegion);
			ImageDrawToWindow(imageBuffer,xWindow,window->invalidRegion,window->w,window->h);	// copy the updates
			ImageDestroy(imageBuffer);
		}
		XDestroyRegion(window->invalidRegion);							// clear out the invalid region now that the window has been drawn
		window->invalidRegion=XCreateRegion();							// reset invalid region to empty
	}
}

static void DocumentWindowAdjustForNewSize(EDITOR_WINDOW *window)
// If the window has changed size, then the controls will be needing
// update, so adjust them for the new size.
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;

	DocumentWindowRectsGet(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);

	ViewBoundsSet(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view,&viewRect);		// change the bounds of the view

	ScrollBarReposition(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar,verticalRect.x,verticalRect.y,verticalRect.h,false);
	ScrollBarReposition(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar,horizontalRect.x,horizontalRect.y,horizontalRect.w,true);
}

static void DocumentWindowButtonPress(EDITOR_WINDOW *window,XEvent *event)
// Handle button press events for document windows (see what the user is clicking
// on, and attempt to deal with it).
{
	if(!ViewTrack(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view,event))
	{
		if(!ScrollBarTrack(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar,event))
		{
			ScrollBarTrack(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar,event);
		}
	}
}

static void DocumentPeriodicProc(EDITOR_WINDOW *window)
// This is called periodically to flash cursors, or do whatever in the document
// when the document has focus.
{
	ViewCursorTask(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view);			// flash the cursor in the view
}

static bool BoundKeyPressAttempt(EDITOR_KEY *editorKey)
// See if the keypress event can be handled by one of the bound keys.
// If so, handle it, and return true.
// If the key is not bound, return false.
{
	KEY_BINDING
		*entry;

	if((entry=KeyBindLocate(editorKey->modifiers,editorKey->keyCode)))
	{
		KeyBindEventHandle(entry);	// go handle it
		EditorWindowsUpdate();		// ### makes repeated operations easier to monitor
		return(true);
	}
	return(false);
}

static bool MenuKeyPressAttempt(EDITOR_KEY *editorKey)
// See if the keypress event can be handled by one of the menu items.
// If so, handle it, and return true.
// If no menu item wants to handle the key, return false.
{
	EDITOR_MENU
		*menu;

	if((menu=MenuKeyEventLocate(editorKey->modifiers,editorKey->keyCode)))
	{
		MenuEventHandle(menu);		// call the shell to handle this menu
		EditorWindowsUpdate();		// ### makes repeated operations easier to monitor
		return(true);
	}
	return(false);
}

static void DocumentWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// handle an event arriving for a document window.
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;
	EDITOR_BUFFER
		*buffer;
	EDITOR_KEY
		editorKey;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;

	buffer=windowInfo->view->parentBuffer;

	switch(event->type)
	{
		case ButtonPress:
			XSetInputFocus(xDisplay,window->xWindow,RevertToPointerRoot,event->xbutton.time);	// give it the focus
			DocumentWindowButtonPress(window,event);										// go handle the button press
			break;
		case KeyPress:
			if(KeyEventToEditorKey(&event->xkey,&editorKey))
			{
				if(!BoundKeyPressAttempt(&editorKey))										// try for bound key first -- bound keys more tightly control the modifiers they respond to
				{
					if(!MenuKeyPressAttempt(&editorKey))									// look for menu key if no bound key
					{
						ViewKeyEventHandle(windowInfo->view,&editorKey);					// give it to the view
					}
				}
			}
			break;
		case FocusIn:
			ViewActiveSet(windowInfo->view,true);
			break;
		case FocusOut:
			ViewActiveSet(windowInfo->view,false);
			break;
		case ClientMessage:
			if(event->xclient.send_event)
			{
				if(event->xclient.data.l[0]==(int)deleteWindowAtom)
				{
					ShellCommandHandle("close",1,&(buffer->contentName));				// tell shell to close this
				}
			}
			break;
		default:
			break;
	}
}

void DocumentWindowStatusBarColorsGet(EDITOR_COLOR *backgroundColor,EDITOR_COLOR *foregroundColor,EDITOR_COLOR *flagActiveColor)
// Get the colors to be used for the status bar.
{
	*backgroundColor=statusBackgroundColor;
	*foregroundColor=statusForegroundColor;
	*flagActiveColor=statusFlagActiveColor;
}

void DocumentWindowStatusBarColorsSet(EDITOR_COLOR backgroundColor,EDITOR_COLOR foregroundColor,EDITOR_COLOR flagActiveColor)
// Set the colors to be used for the status bar.
{
	EDITOR_WINDOW
		*window;

	statusBackgroundColor=backgroundColor;
	statusForegroundColor=foregroundColor;
	statusFlagActiveColor=flagActiveColor;

	window=windowListHead;
	while(window)
	{
		if(window->windowType==EWT_DOCUMENT)
		{
			DocumentWindowStatusAdjust(window);
		}
		window=window->nextWindow;
	}
}

const char *DocumentWindowStatusBarFontGet()
// Get the font to be used by document window status bars.
{
	return(PlatformFontCompositeNameGet(statusBarFont));
}

void DocumentWindowStatusBarFontSet(const char *fontName)
// Set the font to be used by document window status bars.
// The font will be used in all subsequently opened document windows.
// NOTE: if this fails, the status bar just won't be drawn.
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;

	if(statusBarFont)
	{
		PlatformFontFree(statusBarFont);
	}
	statusBarFont=PlatformFontLoad(fontName);

	// run through all document windows and update them
	window=windowListHead;
	while(window)
	{
		if(window->windowType==EWT_DOCUMENT)
		{
			DocumentWindowAdjustForNewSize(window);
			rect.x=0;
			rect.y=0;
			rect.w=window->w;
			rect.h=window->h;
			WindowRectInvalidate(window,&rect);	// invalidate the whole thing
		}
		window=window->nextWindow;
	}
}

void DocumentWindowScrollEditViewEvent(EDITOR_VIEW *view,uint32_t vKeyCode,bool repeating)
// Send a scroll event to the view.
{
	VIEW_EVENT
		record;
	EDITOR_VKEY
		keyData;

	if(!repeating)
	{
		record.eventType=VET_VKEY_DOWN;
	}
	else
	{
		record.eventType=VET_VKEY_REPEAT;
	}
	record.view=view;
	keyData.modifiers=0;			// no modifiers
	keyData.vKeyCode=vKeyCode;		// this is the virtual key
	record.eventData=&keyData;
	ViewEventHandle(&record);
}

static void ScrollEditViewUpEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be scrolled up.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLL_UP,repeating);
}

static void ScrollEditViewDownEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be scrolled down.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLL_DOWN,repeating);
}

static void ScrollEditViewLeftEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be scrolled left.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLL_LEFT,repeating);
}

static void ScrollEditViewRightEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be scrolled right.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_SCROLL_RIGHT,repeating);
}

static void ScrollEditViewPageUpEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be paged up.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENT_PAGE_UP,repeating);
}

static void ScrollEditViewPageDownEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be paged down.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENT_PAGE_DOWN,repeating);
}

static void ScrollEditViewPageLeftEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be paged left.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENT_PAGE_LEFT,repeating);
}

static void ScrollEditViewPageRightEvent(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be paged right.
{
	DocumentWindowScrollEditViewEvent((EDITOR_VIEW *)parameters,VVK_DOCUMENT_PAGE_RIGHT,repeating);
}

static void ScrollEditViewThumbVerticalEvent(SCROLL_BAR *scrollBar,uint64_t newValue,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be positioned vertically.
{
	VIEW_EVENT
		record;
	VIEW_POS_EVENT_DATA
		posData;

	record.eventType=VET_POSITION_VERTICAL;
	record.view=(EDITOR_VIEW *)parameters;
	posData.modifiers=0;
	posData.position=newValue;
	record.eventData=&posData;
	ViewEventHandle(&record);
}

static void ScrollEditViewThumbHorizontalEvent(SCROLL_BAR *scrollBar,uint64_t newValue,void *parameters)
// Tell the editor that the view in the window with scrollBar is to be positioned horizontally.
{
	VIEW_EVENT
		record;
	VIEW_POS_EVENT_DATA
		posData;

	record.eventType=VET_POSITION_HORIZONTAL;
	record.view=(EDITOR_VIEW *)parameters;
	posData.modifiers=0;
	posData.position=newValue;
	record.eventData=&posData;
	ViewEventHandle(&record);
}

static void DocumentWindowDispose(EDITOR_WINDOW *window)
// Close a given window, kill off the view.
{
	ScrollBarDispose(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->horizontalScrollBar);
	ScrollBarDispose(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->verticalScrollBar);
	ViewDispose(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view);	// get rid of the view, unlink it from everywhere
	XDestroyRegion(window->invalidRegion);						// destroy invalid region
	FocusAwayFrom(window);										// move focus off this window
	XDestroyWindow(xDisplay,window->xWindow);					// tell x to make the window go away
	WaitUntilDestroyed(window);
	WindowElementUnlink(window);								// unlink it from our list
	MDisposePtr(window->windowInfo);
	MDisposePtr(window);
}

static EDITOR_WINDOW *DocumentWindowCreate(EDITOR_BUFFER *buffer,EDITOR_RECT *rect,const char *title,bool useUtf8,uint32_t tabSize,const char *fontName,const char *colorSchemeName)
// Open a new document window with one EDITOR_VIEW onto buffer.
// Set the view(s) to use the given color scheme and tab size.
{
	EDITOR_RECT
		contentRect,
		viewRect,
		verticalRect,
		horizontalRect,
		statusRect;
	EDITOR_WINDOW
		*window;
	GUI_DOCUMENT_WINDOW_INFO
		*documentWindowInfo;
	SCROLL_BAR_DESCRIPTOR
		scrollDescriptor;
	EDITOR_VIEW_DESCRIPTOR
		viewDescriptor;
	XSetWindowAttributes
		attributes;
	uint32_t
		valueMask;
	Atom
		protocols[2];														// number of protocols
	XWMHints
		hints;																// hints to tell window manager how to treat these windows
	XSizeHints
		sizeHints;															// window size info (used to REALLY convince X that I know where I want my own window)
	int32_t
		xOffset,
		yOffset;
	uint32_t
		screenWidth,
		screenHeight;

	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))			// create editor window
	{
		if((documentWindowInfo=(GUI_DOCUMENT_WINDOW_INFO *)MNewPtr(sizeof(GUI_DOCUMENT_WINDOW_INFO))))	// make structure to contain document window info
		{
			WindowElementLink(window);									// link it to the global window list that we maintain
			window->windowType=EWT_DOCUMENT;							// set the editor window type to "document"
			window->w=rect->w;											// remember initial size for the window
			window->h=rect->h;
			window->windowInfo=documentWindowInfo;						// point to the edit info
			window->configureFunction=DocumentWindowAdjustForNewSize;	// point to the function to call to configure
			window->updateFunction=DocumentWindowUpdate;				// point to the function to call to update this window when it needs to be redrawn
			window->eventFunction=DocumentWindowEvent;					// point to the function to call to handle events
			window->periodicFunction=DocumentPeriodicProc;				// point to the function to call to handle periodic updates

			valueMask=0;

			PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&screenWidth,&screenHeight);
			// Due to X's nasty error handling, we assume this create does not fail!!!
			window->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x+xOffset,rect->y+yOffset,rect->w,rect->h,0,DefaultDepth(xDisplay,xScreenNum),InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

			sizeHints.flags=USSize|USPosition|PMinSize;
			sizeHints.x=rect->x+xOffset;
			sizeHints.y=rect->y+yOffset;
			sizeHints.width=rect->w;
			sizeHints.height=rect->h;
			sizeHints.min_width=1;
			sizeHints.min_height=1;
			XSetWMNormalHints(xDisplay,window->xWindow,&sizeHints);

			DocumentWindowTitleSet(window,title);
			XSelectInput(xDisplay,window->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask);	// want to see these events

			protocols[0]=takeFocusAtom;
			protocols[1]=deleteWindowAtom;
			XSetWMProtocols(xDisplay,window->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

			hints.flags=InputHint|StateHint|WindowGroupHint;
			hints.input=True;
			hints.initial_state=NormalState;
			hints.window_group=rootMenuWindow->xWindow;						// all documents belong to the group headed by the main menu window
			XSetWMHints(xDisplay,window->xWindow,&hints);

			WindowClassHintsSet(window);
			WindowIconSet(window);

			window->invalidRegion=XCreateRegion();							// create new empty invalid region

			documentWindowInfo->verticalScrollBarOnLeft=verticalScrollBarOnLeft;	// copy global preference
			DocumentWindowRectsGet(window,&contentRect,&viewRect,&verticalRect,&horizontalRect,&statusRect);	// get sizes of areas within the window

			viewDescriptor.buffer=buffer;
			viewDescriptor.rect=viewRect;											// place where the view is going
			viewDescriptor.active=true;												// create the view active
			viewDescriptor.useUtf8=useUtf8;											// tells if the view should use UTF8
			viewDescriptor.tabSize=tabSize;
			viewDescriptor.fontName=fontName;
			viewDescriptor.colorSchemeName=colorSchemeName;
			viewDescriptor.viewTextChangedVector=DocumentWindowScrollThumbsAndStatusAdjust;		// fix the thumbs, and status when the view text changes
			viewDescriptor.viewPositionChangedVector=DocumentWindowScrollThumbsAdjust;			// fix the thumbs when the view position changes
			viewDescriptor.viewSelectionChangedVector=DocumentWindowViewStatusAdjust;			// fix the status when the selection changes
			viewDescriptor.viewMetaChangedVector=DocumentWindowViewStatusAdjust;				// fix the status when the meta information (like tab size, or buffer's syntax map changes)

			if((documentWindowInfo->view=ViewCreate(window,&viewDescriptor)))		// create a view for this area, set up its font stuff
			{
				scrollDescriptor.x=verticalRect.x;									// set up description of the scroll bar
				scrollDescriptor.y=verticalRect.y;
				scrollDescriptor.length=verticalRect.h;
				scrollDescriptor.horizontal=false;
				scrollDescriptor.thumbMax=0;
				scrollDescriptor.thumbPosition=0;
				scrollDescriptor.stepLowerProc=ScrollEditViewUpEvent;
				scrollDescriptor.stepHigherProc=ScrollEditViewDownEvent;
				scrollDescriptor.pageLowerProc=ScrollEditViewPageUpEvent;
				scrollDescriptor.pageHigherProc=ScrollEditViewPageDownEvent;
				scrollDescriptor.thumbProc=ScrollEditViewThumbVerticalEvent;
				scrollDescriptor.procParameters=documentWindowInfo->view;

				if((documentWindowInfo->verticalScrollBar=ScrollBarCreate(window,&scrollDescriptor)))
				{
					scrollDescriptor.x=horizontalRect.x;
					scrollDescriptor.y=horizontalRect.y;
					scrollDescriptor.length=horizontalRect.w;
					scrollDescriptor.horizontal=true;
					scrollDescriptor.thumbMax=0;
					scrollDescriptor.thumbPosition=0;
					scrollDescriptor.stepLowerProc=ScrollEditViewLeftEvent;
					scrollDescriptor.stepHigherProc=ScrollEditViewRightEvent;
					scrollDescriptor.pageLowerProc=ScrollEditViewPageLeftEvent;
					scrollDescriptor.pageHigherProc=ScrollEditViewPageRightEvent;
					scrollDescriptor.thumbProc=ScrollEditViewThumbHorizontalEvent;
					scrollDescriptor.procParameters=documentWindowInfo->view;

					if((documentWindowInfo->horizontalScrollBar=ScrollBarCreate(window,&scrollDescriptor)))
					{
						DocumentWindowScrollThumbsAndStatusAdjust(documentWindowInfo->view);	// force the thumbs and status to be up to date
						XMapRaised(xDisplay,window->xWindow);							// map window to the display
						WaitUntilMapped(window);
						return(window);
					}
					ScrollBarDispose(documentWindowInfo->verticalScrollBar);
				}
				ViewDispose(documentWindowInfo->view);
			}
			XDestroyRegion(window->invalidRegion);					// destroy invalid region
			XDestroyWindow(xDisplay,window->xWindow);				// get rid of the window
			WaitUntilDestroyed(window);

			WindowElementUnlink(window);
			MDisposePtr(documentWindowInfo);
		}
		MDisposePtr(window);
	}
	return((EDITOR_WINDOW *)NULL);
}

// routines callable by the editor

void PlatformDocumentWindowUnminimize(EDITOR_WINDOW *window)
// Unminimize the passed window.
{
	EditorWindowUnminimize(window);
}

void PlatformDocumentWindowMinimize(EDITOR_WINDOW *window)
// Minimize the passed window.
{
	EditorWindowMinimize(window);
}

EDITOR_BUFFER *PlatformDocumentWindowBufferGet(EDITOR_WINDOW *window)
// Given a document window, return the buffer it is open to.
{
	GUI_DOCUMENT_WINDOW_INFO
		*windowInfo;
	EDITOR_BUFFER
		*buffer;

	windowInfo=(GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo;
	buffer=windowInfo->view->parentBuffer;
	return(buffer);
}

EDITOR_WINDOW *PlatformDocumentWindowLocate(EDITOR_BUFFER *buffer)
// Given an editor buffer, locate a document window that points to it.
// Return NULL if none exist.
// This hunts down the window by looking through the views of the buffer
// to find one which is in a document window.
{
	EDITOR_VIEW
		*view;
	EDITOR_WINDOW
		*window;

	view=buffer->firstView;
	while(view)
	{
		window=view->parentWindow;							// get the x window associated with this document window
		if(window->windowType==EWT_DOCUMENT)
		{
			return(window);
		}
		view=view->nextBufferView;
	}
	return(NULL);
}

EDITOR_WINDOW *PlatformDocumentWindowActiveGet()
// Return the document window that has the user's attention, or NULL if there is none.
// Depending on the system, this may or may not return the top-most document window.
// (For instance under Motif, the active document window may be the one with the
// cursor over it.)
{
	EDITOR_WINDOW
		*window;
	Window
		xWindow;
	int
		focusRevert;

	XGetInputFocus(xDisplay,&xWindow,&focusRevert);			// find out who has the focus
	if((window=WindowElementLocate(xWindow)))				// see if it is one of our windows
	{
		if(window->windowType==EWT_DOCUMENT)
		{
			return(window);									// if the window with focus is a document window, then return it
		}
	}
	return(EditorTopWindowTypeGet(EWT_DOCUMENT));			// otherwise, the active one is the top one
}

EDITOR_WINDOW *PlatformDocumentWindowTopGet()
// Return the top-most document window, or NULL if there is none.
{
	return(EditorTopWindowTypeGet(EWT_DOCUMENT));
}

void PlatformDocumentWindowTopSet(EDITOR_WINDOW *window)
// Set the passed document window to the top of the display.
{
	EditorTopWindowSet(window);
}

bool PlatformDocumentWindowSortedListGet(uint32_t *numElements,EDITOR_WINDOW ***list)
// Return a list of all document windows -- sorted top to bottom.
// If there is a problem, ErrorSet, and return false.
// Otherwise return true. If true is returned, list must be
// MDisposePtr'ed by the caller (even if numElements is 0).
{
	return(SortedEditorWindowTypeMaskListGet((1<<EWT_DOCUMENT),numElements,list));
}

EDITOR_VIEW *PlatformDocumentWindowCurrentViewGet(EDITOR_WINDOW *window)
// Passed a document window, return the current view of the window.
{
	return(((GUI_DOCUMENT_WINDOW_INFO *)window->windowInfo)->view);			// currently, document windows contain one view only
}

bool PlatformDocumentWindowTitleGet(EDITOR_WINDOW *window,char *title,uint32_t stringBytes)
// Copy window's title into title.
// If window's title is longer than stringBytes, return false.
{
	char
		*name;

	if(XFetchName(xDisplay,window->xWindow,&name))
	{
		strncpy(title,name,stringBytes);
		XFree(name);
	}
	else
	{
		strncpy(title,"no name",stringBytes);
	}
	if(stringBytes)
	{
		title[stringBytes-1]='\0';
	}
	return(true);
}

bool PlatformDocumentWindowTitleSet(EDITOR_WINDOW *window,const char *title)
// Replace title of window.
// If there is a problem, return false.
{
	return(DocumentWindowTitleSet(window,title));
}

bool PlatformDocumentWindowRectGet(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Get the rectangle of window.
// If there is a problem, ErrorSet, and return false.
// NOTE: once again, X windows makes this much more of a pain than it needs
// to be....
{
	Window
		childWindow;
	int32_t
		xOffset,
		yOffset;
	uint32_t
		screenWidth,
		screenHeight;

	rect->w=window->w;
	rect->h=window->h;

	// Convert our window coordinates to screen coordinates usable by XMoveResizeWindow()
	XTranslateCoordinates(xDisplay,window->xWindow,RootWindow(xDisplay,xScreenNum),0,0,&rect->x,&rect->y,&childWindow);

	PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&screenWidth,&screenHeight);
	rect->x-=xOffset;
	rect->y-=yOffset;

	return(true);
}

bool PlatformDocumentWindowRectSet(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Set window to rect.
// If there is a problem, leave rect unchanged, set the error
// and return false.
{
	int32_t
		xOffset,
		yOffset;
	uint32_t
		screenWidth,
		screenHeight;

	PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&screenWidth,&screenHeight);
	XMoveResizeWindow(xDisplay,window->xWindow,rect->x+xOffset,rect->y+yOffset,rect->w,rect->h);
	return(true);
}

void PlatformDocumentWindowClose(EDITOR_WINDOW *window)
// Close an open document window.
{
	DocumentWindowDispose(window);
}

EDITOR_WINDOW *PlatformDocumentWindowOpen(EDITOR_BUFFER *buffer,EDITOR_RECT *rect,const char *title,bool useUtf8,uint32_t tabSize,const char *fontName,const char *colorSchemeName)
// Open a document window onto buffer.
// If there is a problem, undo anything that was done,
// set the error, then return NULL.
{
	return(DocumentWindowCreate(buffer,rect,title,useUtf8,tabSize,fontName,colorSchemeName));
}

void DocumentWindowUnInit()
// Undo what DocumentWindowInit did.
{
	if(statusBarFont)
	{
		PlatformFontFree(statusBarFont);
	}
}

bool DocumentWindowInit()
// Initialize document window handling.
{
	// set up initial colors
	statusBackgroundColor=BACKGROUND_COLOR;		// status bar colors
	statusForegroundColor=BLACK;
	statusFlagActiveColor=0xFFFFFF00;
	statusBarFont=NULL;							// no font yet
	DocumentWindowStatusBarFontSet(DEFAULT_STATUS_BAR_FONT);
	return(true);
}
