// Low level event handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	BLINK_TIMEOUT_START		(SECOND*.4)	// when restarting the timeout value, this is used
#define	BLINK_TIMEOUT_CONTINUE	(SECOND*.3)	// when continuing timeouts, this value is used
#define	ABORT_CHECK_INTERVAL	(SECOND*.3)	// check for abort every 1/3 second

static uint32_t
	lastBlinkTime,							// tells when last blink occurred
	blinkTimeout;							// when this gets to 0, it is time to flash the cursor

static KeyCode
	abortKeyCode;							// keycode which matches the "abort" key
static uint32_t
	abortCheckTimer;						// tells when it is ok to check for an abort

static uint32_t
	commandModifierIndex;					// keeps track of which modifier key is used as the "command" key

// pseudo globals needed to handle multiple clicks

static uint32_t
	lastClickTime;
static Window
	lastClickWindow;
static int32_t
	lastClickX,
	lastClickY;
static uint8_t
	numClicks;
static uint32_t
	lastUserEventTime;					// keeps track of the event timestamp from the last user input (key/button)

#define	CLICK_SLOP		5				// number of pixels of slop in either direction
#define	CLICK_TIME		300				// number of milliseconds between clicks considered multiple

// ----------------------------------------------------------------------------

static uint32_t MonotonicTime()
// Return a 32-bit monotonic time that increases in milliseconds.
// This is used for various timers, and whatnot.
// NOTE: the code calling this relies on the returned 32-bit time to roll
// over cleanly from 0xFFFFFFFF to 0x00000000.
// This implementation likely won't roll over cleanly once time.tv_sec
// rolls over.
// However, we'll be okay since time_t (the type of time.tv_sec) is
// usually 64-bits these days, and it'll take an awful lot of time for
// that to roll over...
{
	timespec
		time;

	clock_gettime(CLOCK_MONOTONIC,&time);	// see what time it is
	return((time.tv_sec*1000)+(time.tv_nsec/1000000));
}


static bool WaitForEvent(uint32_t milliseconds)
// See if there's an X event pending, or if one of the task file descriptors
// has data in it. Return as soon as one of those things happens, or
// as soon as milliseconds expires. If the timeout expires with nothing
// ready, return false.
{
	unsigned int
		taskIndex;
	EDITOR_TASK
		*task;
	int
		result;

	result=XEventsQueued(xDisplay,QueuedAlready);
	if(result==0)									// make sure there are no pending X11 events already picked up
	{
		taskIndex=0;
		task=TaskFirst();							// count up the number of active tasks
		while(task)
		{
			taskIndex++;
			task=TaskNext(task);
		}

		struct pollfd
			ufds[taskIndex+1];						// make array to send to poll

		taskIndex=0;
		task=TaskFirst();							// fill in the task descriptors
		while(task)
		{
			ufds[taskIndex].fd=TaskOutputFileDescriptor(task);
			ufds[taskIndex].events=POLLIN;
			ufds[taskIndex].revents=0;
			taskIndex++;
			task=TaskNext(task);
		}

		ufds[taskIndex].fd=ConnectionNumber(xDisplay);	// fill in the final descriptor for the X socket
		ufds[taskIndex].events=POLLIN;
		ufds[taskIndex].revents=0;

		if((poll(&ufds[0],taskIndex,milliseconds))<=0)	// this only catches NEW X events, not ones that already sit in the queue
		{
			return(false);							// nothing ready
		}
	}
	return(true);
}

// ----------------------------------------------------------------------------

static Bool AbortEventPredicate(Display *display,XEvent *event,char *arg)
// Check the Event to see if it is an abort event
// if it is, return True
// NOTE: since this is an event predicate, it is not allowed to
// call back into XLib (See XCheckIfEvent man page).
// Because of this, we cannot call XLookupString, and must instead
// compare keycodes directly.
{
	if((event->type==KeyPress))
	{
		if(event->xkey.keycode==abortKeyCode)
		{
			return(True);
		}
	}
	return(False);
}

static Bool UpdateEventsPredicate(Display *display,XEvent *event,char *arg)
// Match all update events.
{
	if((event->type==Expose)||(event->type==GraphicsExpose)||(event->type==NoExpose)||(event->type==ConfigureNotify))
	{
		return(True);
	}
	return(False);
}

static Bool KeyEventsPredicate(Display *display,XEvent *event,char *arg)
// Match all keyboard press events.
{
	if(event->type==KeyPress)
	{
		return(True);
	}
	return(False);
}

static Bool ButtonEventPredicate(Display *display,XEvent *event,char *arg)
// Return True if the event is a button press, or a button release.
// All button events are looked at while in StillDown.
{
	if((event->type==ButtonPress)||(event->type==ButtonRelease)||(event->type==MotionNotify))
	{
		return(True);
	}
	return(False);
}

static Bool AllEventsPredicate(Display *display,XEvent *event,char *arg)
// Match all events (used in XCheckIfEvent below).
{
	return(True);
}


static void InvalidationsCollect(XEvent *event)
// Given an Expose, or GraphicsExpose event, find the window
// it belongs to, and add the event's invalid area to the window's
// invalid region.
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		invalidRect;

	if((window=WindowElementLocate(event->xany.window)))
	{
		invalidRect.x=event->xexpose.x;
		invalidRect.y=event->xexpose.y;
		invalidRect.w=event->xexpose.width;
		invalidRect.h=event->xexpose.height;
		WindowRectInvalidate(window,&invalidRect);
	}
}

static void WindowConfigure(XEvent *event)
// Given a ConfigureNotify event, find the window
// it belongs to, and tell it to adjust to its size.
{
	EDITOR_WINDOW
		*window;

	if((window=WindowElementLocate(event->xany.window)))
	{
		window->w=event->xconfigure.width;							// update locally kept width and height
		window->h=event->xconfigure.height;

		if(window->configureFunction)
		{
			window->configureFunction(window);						// call the window to configure
		}
	}
}

static void EventProcess(XEvent *event,EDITOR_WINDOW *modalWindow)
// Passed an event, do what is needed to take care of it.
// NOTE: if modalWindow is non-NULL, it indicates that all events other
// than invalidations, and configurations should be dropped for
// windows other than the given one.
{
	EDITOR_WINDOW
		*window;

	if((event->type==Expose)||(event->type==GraphicsExpose))
	{
		InvalidationsCollect(event);								// collect any rectangles from expose events
	}
	else if(event->type==ConfigureNotify)
	{
		WindowConfigure(event);										// go configure the window (without redrawing the invalidations, since it'll be faster and look better when windows are resized)
	}
	else if(event->type==SelectionRequest)
	{
		SelectionRequestHandle(event);								// process request for the selection (clipboard)
	}
	else
	{
		if(event->type==KeyPress||event->type==KeyRelease)			// keep track of time of last user input
		{
			lastUserEventTime=event->xkey.time;
		}
		else if(event->type==ButtonPress||event->type==ButtonRelease)
		{
			lastUserEventTime=event->xbutton.time;
		}

		if((window=WindowElementLocate(event->xany.window)))		// find the window the event has been sent to
		{
			if(!modalWindow||window==modalWindow)
			{
				if(window->eventFunction)							// see if it wants the event?
				{
					window->eventFunction(window,event);			// pass it along
				}
			}
		}
	}
}

static void PeriodicProc()
// When it is time to flash the cursor (or whatever), attempt to locate the window with input
// focus, and call it to do its thing.
{
	Window
		xWindow;
	int
		focusRevert;
	EDITOR_WINDOW
		*window;

	XGetInputFocus(xDisplay,&xWindow,&focusRevert);			// find out who has the focus
	if((window=WindowElementLocate(xWindow)))				// see if it is one of our windows
	{
		if(window->periodicFunction)
		{
			window->periodicFunction(window);
		}
	}
}

uint32_t XStateToEditorModifiers(unsigned int state)
// Convert modifiers and numClicks into modifier bits for the editor.
{
	uint32_t
		modifiers;

	modifiers=0;

	if(state&ShiftMask)
	{
		modifiers|=EEM_SHIFT;
	}
	if(state&ControlMask)
	{
		modifiers|=EEM_CTRL;
	}
	if(state&Mod1Mask)						// Alt key
	{
		modifiers|=EEM_ALT;
	}
	if(state&Mod4Mask)						// Window key
	{
		modifiers|=EEM_WINDOW;
	}
	return(modifiers);
}

static uint32_t KeyToVirtual(KeySym keySym,uint32_t modifiers)
// See if the given keySym maps to a virtual function for the editor.
// If no conversion makes sense, return VVK_NOP.
{
	uint32_t
		result;

	result=VVK_NOP;
	if(modifiers==PlatformRemapEditorModifiers(EEM_CMD))	// check for modified keys which have meaning (only have meaning when the command key is active, and no other modifier)
	{
		switch(keySym)
		{
			case XK_a:
				result=VVK_SELECT_ALL;
				break;
			case XK_z:
				result=VVK_UNDO_TOGGLE;
				break;
			case XK_x:
				result=VVK_CUT;
				break;
			case XK_c:
				result=VVK_COPY;
				break;
			case XK_v:
				result=VVK_PASTE;
				break;
			case XK_y:
				result=VVK_REDO;
				break;
			case XK_u:
				result=VVK_UNDO;
				break;
			default:
				break;
		}
	}
	if(result==VVK_NOP)						// if nothing found, look for special keys
	{
		switch(keySym)
		{
			case XK_Left:
			case XK_KP_Left:
				result=VVK_LEFT_ARROW;
				break;
			case XK_Right:
			case XK_KP_Right:
				result=VVK_RIGHT_ARROW;
				break;
			case XK_Up:
			case XK_KP_Up:
				result=VVK_UP_ARROW;
				break;
			case XK_Down:
			case XK_KP_Down:
				result=VVK_DOWN_ARROW;
				break;
			case XK_Prior:
			case XK_KP_Prior:
				result=VVK_PAGE_UP;
				break;
			case XK_Next:
			case XK_KP_Next:
				result=VVK_PAGE_DOWN;
				break;
			case XK_Home:
			case XK_KP_Home:
				result=VVK_HOME;
				break;
			case XK_End:
			case XK_KP_End:
				result=VVK_END;
				break;
			case XK_Insert:
			case XK_KP_Insert:
				result=VVK_INSERT;
				break;
			case XK_Delete:
			case XK_KP_Delete:
				result=VVK_DELETE;
				break;
			case XK_BackSpace:
				result=VVK_BACK_SPACE;
				break;
			case XK_Help:
				result=VVK_HELP;
				break;
			case XK_Undo:
				result=VVK_UNDO;
				break;
			case XK_Redo:
				result=VVK_REDO;
				break;
			case XK_Return:
			case XK_KP_Enter:
				result=VVK_RETURN;
				break;
			case XK_Tab:
				result=VVK_TAB;
				break;
			case XK_Escape:
				result=VVK_ESC;
				break;
			default:
				break;
		}
	}
	return(result);
}

bool KeyEventToEditorKey(XKeyEvent *event,EDITOR_KEY *editorKey)
// Take a keyboard event, and fill in an editor key structure.
// NOTE: if this returns false, it means the key event didn't yield
// anything useful (like pressing a modifier w/o another key).
{
	KeySym
		keySym;
	unsigned int
		xStateOut;
	int
		extra_rtrn;

	XkbLookupKeySym(xDisplay,event->keycode,event->state,&xStateOut,&keySym);	// convert to keysym, consume some modifiers
	if(!IsModifierKey(keySym))
	{
		editorKey->modifiers=XStateToEditorModifiers(event->state);
		editorKey->keyCode=event->keycode;
		editorKey->vKeyCode=KeyToVirtual(keySym,editorKey->modifiers);
		editorKey->numTextChars=XkbTranslateKeySym(xDisplay,&keySym,event->state,(char *)editorKey->text,sizeof(editorKey->text),&extra_rtrn);
		return(true);
	}
	return(false);
}

void EditorWindowsUpdate()
// Attend to changes brought about in our windows.
// Redraw all editor windows that have been invalidated by previous expose or
// graphics expose events.
{
	XEvent
		event;

	while(XCheckIfEvent(xDisplay,&event,UpdateEventsPredicate,NULL))		// get pending update messages
	{
		EventProcess(&event,NULL);									// go process them
	}
	InvalidWindowsRedraw();											// once all have been handled, do redraw
}

void CursorBlinkTimeReset()
// Reset the timeout until the next cursor blink arrives.
{
	lastBlinkTime=MonotonicTime();
	blinkTimeout=BLINK_TIMEOUT_START;								// number of ticks of system timer before flashing the cursor
}

Bool XCheckIfEventTimed(Display *display,XEvent *event_return,Bool (*predicate)(Display *,XEvent *,XPointer),XPointer arg,uint32_t milliseconds)
// Just like XCheckIfEvent, except this will wait up to milliseconds for
// an event to appear in the queue. Other events may enter the queue
// while waiting.
{
	struct pollfd
		ufds[1];
	Bool
		result;
	uint32_t
		startTime,
		totalTime;
	unsigned int
		numFds;

	result=XCheckIfEvent(display,event_return,predicate,arg);
	if(!result)
	{
		startTime=MonotonicTime();									// work out what time it is now
		totalTime=0;

		ufds[0].fd=ConnectionNumber(xDisplay);
		ufds[0].events=POLLIN;
		ufds[0].revents=0;

		numFds=1;
		while(!result&&(totalTime<milliseconds))
		{
			if((poll(&ufds[0],numFds,milliseconds-totalTime))>0)	// wake when new events enter the queue
			{
				result=XCheckIfEvent(display,event_return,predicate,arg);	// hunt for the event type we're looking for
			}
			totalTime=MonotonicTime()-startTime;
		}
	}
	return(result);
}

uint8_t MultipleClicksHandle(XEvent *event)
// Updates and returns numClicks which tells how many clicks were in roughly the same area
// of the same window, where each click followed the previous by no more than CLICK_TIME milliseconds.
// This is how double, triple, etc clicks are handled.
// NOTE: due to wrap around in the time field, this routine could possibly
// fail to work correctly once every 49.7 days.
{
	if((lastClickTime+CLICK_TIME)>event->xbutton.time&&
		(lastClickX>=(event->xbutton.x_root-CLICK_SLOP))&&(lastClickX<=(event->xbutton.x_root+CLICK_SLOP))&&
		(lastClickY>=(event->xbutton.y_root-CLICK_SLOP))&&(lastClickY<=(event->xbutton.y_root+CLICK_SLOP))&&
		event->xbutton.window==lastClickWindow)
	{
		numClicks++;
	}
	else
	{
		numClicks=0;
	}
	lastClickTime=event->xbutton.time;
	lastClickWindow=event->xbutton.window;
	lastClickX=event->xbutton.x_root;
	lastClickY=event->xbutton.y_root;
	return(numClicks);
}

void MultipleClicksReset()
// Clears the double click information, so that next click starts from scratch.
{
	lastClickTime=0;					// initialize multiple click variables
	lastClickWindow=(Window)0;
	lastClickX=lastClickY=0;
	numClicks=0;
}

void EditorDoBackground(uint32_t eventWaitTime)
// Do the background tasks of updating windows, tasks and timers.
{
	uint32_t
		currentTime;

	WaitForEvent(eventWaitTime);							// go wait up to this long for an event, or some task data to arrive
	BackgroundUpdate();										// let high level code perform background tasks
	ShellBackgroundUpdate();								// call the shell to do anything it wants
	currentTime=MonotonicTime();
	if((currentTime-lastBlinkTime)>=blinkTimeout)			// time to blink cursor?
	{
		lastBlinkTime=currentTime;
		blinkTimeout=BLINK_TIMEOUT_CONTINUE;				// reset the timer
		PeriodicProc();
	}
}

bool StillDown(Window xWindow,unsigned int button,int *rootX,int *rootY,int *windowX,int *windowY)
// Call this after receiving a Button Pressed event for button, and
// it will return true if the button is still held down.
// This will also collect pointer positions events, and use them
// to update rootX, rootY, windowX, windowY.
{
	XEvent
		event;
	Window
		root,
		child;
	unsigned int
		mask;
	bool
		locatedUp;

	EditorDoBackground(1);								// handle background stuff (keep it quick so tracking and scrolling don't suffer)
	EditorWindowsUpdate();								// make sure all windows get updated while spinning through this function
	locatedUp=false;
	XQueryPointer(xDisplay,xWindow,&root,&child,rootX,rootY,windowX,windowY,&mask);
	while(!locatedUp&&XCheckIfEvent(xDisplay,&event,ButtonEventPredicate,NULL))	// inhale all the events that are ready
	{
		if((event.type==ButtonRelease)&&(event.xbutton.button==button))
		{
			lastUserEventTime=event.xbutton.time;
			locatedUp=true;
		}
	}
	return(!locatedUp);
}

void EventLocalLoop(bool *done,EDITOR_WINDOW *modalWindow)
// Process events until the given bool is set true by some event, then
// return.
// NOTE: if modalWindow is non-NULL, it indicates that all events other
// than invalidations, and configurations should be dropped for
// windows other than the given one.
{
	XEvent
		event;

	while(!(*done))
	{
		EditorDoBackground(20);								// handle background stuff -- more slowly so we don't burn cycles while waiting for something to happen
		while(XCheckIfEvent(xDisplay,&event,AllEventsPredicate,NULL))	// process all waiting events
		{
			EventProcess(&event,modalWindow);
		}
		InvalidWindowsRedraw();								// once all events have been handled, do redraw
	}
}

uint32_t EventLastUserEventTime()
// Return the time of the last user input.
{
	return(lastUserEventTime);
}

// ----------------------------------------------------------------------------

uint32_t KeyCommandModifierGet()
// Return the index to the editor modifier key which is currently used
// as the command key.
{
	return(commandModifierIndex);
}

void KeyCommandModifierSet(uint32_t index)
// Set the index to the editor modifier key which is currently used
// as the command key.
{
	if(commandModifierIndex!=index)		// make sure it changed
	{
		commandModifierIndex=index;
		MenuCommandKeyChange();			// let menu code know the command key has changed so it can update if needed
	}
}

// ----------------------------------------------------------------------------

uint32_t PlatformRemapEditorModifiers(uint32_t modifiersIn)
// Remap modifiersIn, return the remapped version. This is called
// by the editor on the modifiers for any key binding before the
// modifiers are compared to those in a EDITOR_KEY event.
// In this X11 version, the function is used to allow remapping of
// the "Command" key (which most PC keyboards don't have) to another
// modifier (generally Alt or Control, but all others are allowed too).
{
	if(modifiersIn&(1<<EES_CMD))		// see if the command key is indicated
	{
		modifiersIn&=~(1<<EES_CMD);		// unset it
		modifiersIn|=(1<<KeyCommandModifierGet());	// add the modifier that command is currently mapped to
	}
	return(modifiersIn);
}

bool PlatformKeyNameToKeyCode(const char *name,uint32_t *keyCode)
// Given a string which is the name of an unmodified key on the keyboard,
// convert it into a keyCode (just a number specifying the physical key).
// If the name is not the name of any key on our keyboard, then return false.
{
	KeySym
		keySym;

	if((keySym=XStringToKeysym(name))!=NoSymbol)
	{
		*keyCode=XKeysymToKeycode(xDisplay,keySym);
		return(true);
	}
	return(false);
}

const char *PlatformKeyCodeToKeyName(uint32_t keyCode)
// Given a keycode, get its name.
// If the code is invalid, return NULL.
// NOTE: the returned string is static, and should not be modified or deleted.
{
	KeySym
		keySym;
	unsigned int
		mods_rtrn;

	XkbLookupKeySym(xDisplay,keyCode,0,&mods_rtrn,&keySym);
	return(XKeysymToString(keySym));
}

bool PlatformKeyPressGet(EDITOR_KEY *editorKey,bool wait,bool clearBuffered)
// Used by editor when it wants to get a key press.
// If a code is found, it is returned in keyCode/editorModifiers.
// If none is ready, return false.
// If wait is true, wait here until one is ready (true is always returned).
// If clearBuffered is true, then clear any keypresses that may have been in the
// buffer before proceeding.
{
	XEvent
		event;
	bool
		gotOne;

	gotOne=false;
	if(clearBuffered)
	{
		while(XCheckIfEvent(xDisplay,&event,KeyEventsPredicate,NULL))		// inhale all keypresses
			;
	}
	if(wait)
	{
		do
		{
			XIfEvent(xDisplay,&event,KeyEventsPredicate,NULL);
			lastUserEventTime=event.xkey.time;
			gotOne=KeyEventToEditorKey(&event.xkey,editorKey);
		} while(!gotOne);
	}
	else
	{
		while((!gotOne)&&XCheckIfEvent(xDisplay,&event,KeyEventsPredicate,NULL))
		{
			lastUserEventTime=event.xkey.time;
			gotOne=KeyEventToEditorKey(&event.xkey,editorKey);
		}
	}
	return(gotOne);
}

void PlatformEventLoop(int argc,char *argv[])
// Get the next event that the editor needs to know about and return it.
// If other events occur, this routine should stay here handling them until
// one specifically related to the editor arrives, it should then pass that
// event up to the editor for processing.
// If no events arrive, it is the responsibility of this routine to
// give time back to the OS if necessary.
{
	uint32_t
		i;

	// ### Before starting, we'd like the menu window to be drawn (it can look bad if it sits
	// half updated while we're opening the initial files).
	// You'd think that a single XSync followed by calls to pull in and process all configure
	// and update events would do the trick. But, for some reason I cannot yet explain, this
	// doesn't work. Instead, we need to issue multiple XSync's to push through all the queued
	// events. I'd really like to remove this hack.
	for(i=0;i<10;i++)
	{
		XSync(xDisplay,False);
		EditorWindowsUpdate();
	}

	ShellCommandHandle("initialargs",argc-1,&(argv[1]));	// pass off initial arguments to shell for processing

	EventLocalLoop(&timeToQuit,NULL);
}

void PlatformQuit()
// This is called from the editor when it wants to quit.
// It needs to inform the event loop that it should stop looping.
{
	timeToQuit=true;										// set the quit request flag
}

void PlatformEventsAbortClear()
// Call this to clear any pending abort status.
{
	XEvent
		event;

	while(XCheckIfEvent(xDisplay,&event,AbortEventPredicate,NULL))	// get all abort events out of the queue
		;
	abortCheckTimer=MonotonicTime();								// restart abort timer
}

bool PlatformEventsAbortCheck()
// This will return true if the user is requesting an abortion of the
// current job.
// This should be made to execute as efficiently as possible, since it will
// be called often during certain loops.
// To make it more efficient, this version checks the time, and if enough time
// has not passed, returns instantly.
{
	XEvent
		event;
	uint32_t
		currentTime;

	currentTime=MonotonicTime();
	if(currentTime-abortCheckTimer>=ABORT_CHECK_INTERVAL)
	{
		abortCheckTimer=currentTime;								// reset the timer
		if(XCheckIfEvent(xDisplay,&event,AbortEventPredicate,NULL))	// see if there is an abort event floating around in the queue
		{
			return(true);
		}
	}
	return(false);
}

uint32_t PlatformMonotonicTime()
// Return a 32-bit monotonic time that increases in milliseconds.
{
	return(MonotonicTime());
}

// ----------------------------------------------------------------------------

void EventsUnInit()
// Uninitialize event stuff.
{
}

bool EventsInit()
// Initialize event stuff.
// If there is a problem, ErrorSet, return false.
{
	abortKeyCode=XKeysymToKeycode(xDisplay,XK_Escape);				// keep the "abort" keycode used to abort various actions
	commandModifierIndex=EES_CTRL;									// set the initial modifier used for commands
	MultipleClicksReset();
	lastUserEventTime=0;
	return(true);
}
