// Code for generating lists of system fonts -- useful for selecting
// fonts in dialog boxes.
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

void FontListDispose(char **list)
// Dispose of a font list created by FontListCreate.
{
	StringListFree(list);
}

char **FontListCreate(uint32_t *numElements)
// Generate a list of font families.
// The list must be freed by calling FontListDispose.
// NOTE: the returned list will be sorted, and will not contain
// duplicates.
// If there is a problem, ErrorSet, and return NULL.
{
	FcPattern
		*pattern;
	FcObjectSet
		*objectSet;
	FcFontSet
		*fontSet;
	FcPattern
		*font;
	FcChar8
		*family;
	char
		**stringList;
	bool
		fail;
	char
		familyName[MAX_FONT_NAME_LENGTH];
	uint32_t
		i,
		outIndex;

	stringList=NULL;
	fontSet=NULL;
	if((pattern=FcPatternCreate()))
	{
		if((objectSet=FcObjectSetBuild(FC_FAMILY,NULL)))
		{
			if((fontSet=FcFontList(NULL,pattern,objectSet)))
			{
				if((stringList=StringListNew()))
				{
					*numElements=0;
					fail=false;
					for(i=0;!fail&&(i<(uint32_t)fontSet->nfont);i++)
					{
						font=fontSet->fonts[i];
						if(FcPatternGetString(font,FC_FAMILY,0,&family)==FcResultMatch)	// get family name
						{
							outIndex=0;
							while((outIndex<sizeof(familyName)-1)&&*family)				// copy locally, and make lower case
							{
								familyName[outIndex]=tolower(*family);
								family++;
								outIndex++;
							}
							familyName[outIndex]='\0';
							fail=!StringListAdd(familyName,&stringList,numElements);	// add lower case string to the list
						}
					}
					if(!fail)		// if all went well, sort the list, and remove duplicates
					{
						StringListSort(stringList,*numElements);
						(*numElements)=StringListRemoveDuplicates(stringList,*numElements);
					}
					else
					{
						StringListFree(stringList);
						stringList=NULL;			// return error
					}
				}
				FcFontSetDestroy(fontSet);
			}
			FcObjectSetDestroy(objectSet);
		}
		FcPatternDestroy(pattern);
	}
	return(stringList);
}
