// Font handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// This code thinks of each combination of font family, font size,
// boldness, and obliqueness as an independent EDITOR_FONT.
// Fonts are loaded by providing a composite name which includes
// all of the above information. Functions are provided to convert
// a composite name into its components, and vice versa.

// NOTE: Getting Xft to work with a single font was straightforward.
// However, working out what was needed to handle fallbacks when a glyph
// is not in that font was a complete pain. The fontconfig developer
// documentation provides no direction to developers about how this is
// meant to be handled. I ended up looking at the source for pango,
// emacs, and a terminal emulator called 'st' to work out how others do
// it.
// Pango and st use fontconfig's FcFontSort function to generate the list
// of fallback fonts. However, the handling of this is complex, and
// involves more bloat (caching of fonts, etc, etc...) than I want in
// e93.
// emacs takes a simpler approach, letting the user just specify the
// fonts that can be used for falling back.
// I prefer emac's approach given its simplicity, and that it puts the
// user directly in charge of which fonts get used for fallback
// (including none). That's what's implemented here.
// The concept of falling back to other fonts is contained solely in this
// file. None of the other code needs to be aware of it.

#include	"includes.h"

typedef struct fallbackFontName
{
	struct fallbackFontName
		*next;									// points to the next in the list (NULL if none)
	char
		familyName[0];							// variable length array that contains the 0 terminated font family name
} FALLBACK_FONT_NAME;

typedef struct fallbackFont
{
	XftFont
		*xftFont;								// font used as fallback for missing glyphs
	struct fallbackFont
		*next;									// next in the list (NULL if none)
} FALLBACK_FONT;

typedef struct editorFont
{
	NAMED_LIST_ITEM
		*listItem;								// link to name list item which holds the composite font name
	uint32_t
		usageCount;								// tells how many references there are to this font
	XftFont
		*xftFont;								// info about currently selected font
	FALLBACK_FONT
		*fallbackFonts;							// points to the head of the list of fallback fonts (NULL if none)
	uint32_t
		ascent,
		descent;
	bool
		charExists[256];						// cache of glyph existance of the first 256 glyphs of the font (presumably used the most often) (NOTE: this is the composite of the loaded font and any fallbacks)
	uint32_t
		charWidths[256];						// cache of widths of the first 256 glyphs of the font (presumably used the most often) (NOTE: this is the composite of the loaded font and any fallbacks)
} EDITOR_FONT;

static NAMED_LIST
	*loadedFontList;							// the list of fonts which are currently loaded

static FALLBACK_FONT_NAME
	*firstFallbackName,							// used to manage the list of font fallback names
	*lastFallbackName;

// ----------------------------------------------------------------------------

static EDITOR_FONT *FontLoadedLocate(const char *compositeName)
// See if an already loaded font can be located which has compositeName.
// If so, return a pointer to it, otherwise return NULL.
{
	NAMED_LIST_ITEM
		*listItem;

	if((listItem=NamedListItemLocateName(loadedFontList,compositeName)))
	{
		return((EDITOR_FONT *)NamedListItemDataGet(listItem));
	}
	return(NULL);
}

static void FontCompositeNameSplit(const char *compositeName,char *familyName,uint32_t *size,bool *bold,bool *oblique)
// Split a composite font name into components.
// compositeName can be any length, but should be less than MAX_FONT_NAME_LENGTH bytes long.
// familyName must be at least MAX_FONT_NAME_LENGTH bytes long.
{
	uint32_t
		index,
		outIndex;
	char
		stringBuffer[MAX_FONT_NAME_LENGTH];
	int
		var;

	index=0;
	while((index<MAX_FONT_NAME_LENGTH-1)&&compositeName[index]&&compositeName[index]!=':')
	{
		familyName[index]=tolower(compositeName[index]);	// all family names are strictly lower case
		index++;
	}
	familyName[index]='\0';			// terminate the family name

	// set defaults in case we don't encounter modifiers
	*size=9;
	*bold=false;
	*oblique=false;

	while(compositeName[index]&&compositeName[index]!=':')	// skip to first separator
	{
		index++;
	}
	if(compositeName[index])		// step over any separator
	{
		index++;
	}
	while(compositeName[index])		// process all modifier strings
	{
		outIndex=0;
		while(compositeName[index]&&compositeName[index]!=':')	// skip to first separator
		{
			if(!isspace(compositeName[index]))	// ignore whitespace
			{
				stringBuffer[outIndex]=tolower(compositeName[index]);
				outIndex++;
			}
			index++;
		}
		stringBuffer[outIndex]='\0';	// terminate output string
		if(compositeName[index])		// step over any separator
		{
			index++;
		}
		if(sscanf(stringBuffer,"size=%d",&var)==1)
		{
			*size=var;
		}
		else if(strcmp(stringBuffer,"bold")==0)
		{
			*bold=true;
		}
		else if(strcmp(stringBuffer,"oblique")==0)
		{
			*oblique=true;
		}
	}
}

static void FontComponentNameJoin(const char *familyName,uint32_t size,bool bold,bool oblique,char *compositeName)
// Take a component specified font, and convert it to the canonicalized
// composite name.
// compositeName must be at least MAX_FONT_NAME_LENGTH bytes long.
{
	snprintf(compositeName,MAX_FONT_NAME_LENGTH,"%s:size=%d%s%s",familyName,size,bold?":bold":"",oblique?":oblique":"");	// make composite string
}

// ----------------------------------------------------------------------------

static XftFont *FontForCodePointGet(EDITOR_FONT *font,uint32_t codePoint)
// Return the XftFont needed to render the given code point.
{
	FALLBACK_FONT
		*fallbackFont;

	if(!XftCharExists(xDisplay,font->xftFont,codePoint))
	{
		// hunt the fallbacks in order
		fallbackFont=font->fallbackFonts;
		while(fallbackFont)
		{
			if(XftCharExists(xDisplay,fallbackFont->xftFont,codePoint))
			{
				return(fallbackFont->xftFont);
			}
			fallbackFont=fallbackFont->next;
		}
	}
	return(font->xftFont);
}

uint32_t FontText32Width(EDITOR_FONT *font,const uint32_t *text,uint32_t length)
// Return the width (in pixels) of the given 32-bit text rendered in the given font.
// This will handle fallback fonts transparently.
{
	uint32_t
		i;
	uint32_t
		width;
	XGlyphInfo
		extents;
	uint32_t
		lastIndex;
	XftFont
		*currentFont,
		*lastFont;

	width=0;
	lastIndex=0;
	lastFont=NULL;
	for(i=0;i<length;i++)
	{
		if((currentFont=FontForCodePointGet(font,text[i]))!=lastFont)	// change in font?
		{
			if(lastFont)
			{
				XftTextExtents32(xDisplay,lastFont,(const FcChar32 *)&text[lastIndex],i-lastIndex,&extents);
				width+=extents.xOff;
			}
			lastIndex=i;
			lastFont=currentFont;
		}
	}
	if(lastFont)
	{
		XftTextExtents32(xDisplay,lastFont,(const FcChar32 *)&text[lastIndex],i-lastIndex,&extents);
		width+=extents.xOff;
	}
	return(width);
}

uint32_t FontTextUtf8Width(EDITOR_FONT *font,const char *text,uint32_t length)
// Return the width (in pixels) of the given UTF-8 encoded text rendered in the given font.
// This will handle fallback fonts transparently.
{
	uint32_t
		width;
	uint32_t
		buffer[256];
	uint32_t
		bufferIndex;
	int
		bytesUsed;
	FcChar32
		ucs4;

	width=0;
	bufferIndex=0;
	while(length&&(bytesUsed=FcUtf8ToUcs4((const FcChar8 *)text,&ucs4,length))>0)	// walk through all the characters of the input
	{
		buffer[bufferIndex]=ucs4;
		bufferIndex++;
		if(bufferIndex>=elementsof(buffer))							// filled the local buffer
		{
			width+=FontText32Width(font,buffer,bufferIndex);		// total length so far
			bufferIndex=0;
		}
		text+=bytesUsed;
		length-=bytesUsed;
	}
	if(bufferIndex)													// any left at the end?
	{
		width+=FontText32Width(font,buffer,bufferIndex);
	}
	return(width);
}

static bool FontCodePoint32WidthSlow(EDITOR_FONT *font,uint32_t codePoint,uint32_t *width)
// Check if the codePoint of the given font exists (looking it up in Xft). If it does,
// set width and return true. Otherwise, return false.
// This will handle fallback fonts transparently.
{
	FALLBACK_FONT
		*fallbackFont;
	XGlyphInfo
		extents;

	// do it the slow way
	if(XftCharExists(xDisplay,font->xftFont,codePoint))
	{
		XftTextExtents32(xDisplay,font->xftFont,(const FcChar32 *)&codePoint,1,&extents);
		*width=extents.xOff;
		return(true);
	}
	else	// hunt the fallbacks
	{
		fallbackFont=font->fallbackFonts;
		while(fallbackFont)
		{
			if(XftCharExists(xDisplay,fallbackFont->xftFont,codePoint))
			{
				XftTextExtents32(xDisplay,fallbackFont->xftFont,(const FcChar32 *)&codePoint,1,&extents);
				*width=extents.xOff;
				return(true);
			}
			fallbackFont=fallbackFont->next;
		}
	}
	return(false);
}

bool FontCodePoint32Width(EDITOR_FONT *font,uint32_t codePoint,uint32_t *width)
// Check if the codePoint of the given font exists. If it does,
// set width and return true. Otherwise, return false.
// This will handle fallback fonts transparently.
{
	if(codePoint<=0xFF)
	{
		if(font->charExists[codePoint])
		{
			*width=font->charWidths[codePoint];						// get width from cache (quick)
			return(true);
		}
		return(false);
	}
	return(FontCodePoint32WidthSlow(font,codePoint,width));			// do it the slow way
}

// ----------------------------------------------------------------------------

int FontText32Draw(EDITOR_FONT *font,GUI_IMAGE *imageBuffer,EDITOR_COLOR color,int x,int y,const uint32_t *text,uint32_t length)
// Draw length characters of the passed 32-bit text in the passed font into the passed
// imageBuffer in the given color (ARGB) at the given x and y location.
// Return the updated X coordinate.
// This will handle fallback fonts transparently.
{
	XRenderColor
		xrColor;
	XftColor
		xftColor;
	uint32_t
		i;
	XGlyphInfo
		extents;
	uint32_t
		lastIndex;
	XftFont
		*currentFont,
		*lastFont;

	ImageEditorColorToRenderColorAlpha(color,&xrColor);
	XftColorAllocValue(xDisplay,xVisualInfo32.visual,xColorMap32,&xrColor,&xftColor);

	lastIndex=0;
	lastFont=NULL;
	for(i=0;i<length;i++)
	{
		if((currentFont=FontForCodePointGet(font,text[i]))!=lastFont)	// change in font?
		{
			if(lastFont)
			{
				XftDrawString32(imageBuffer->xftDraw,&xftColor,lastFont,x,y,(const FcChar32 *)&text[lastIndex],i-lastIndex);
				XftTextExtents32(xDisplay,lastFont,(const FcChar32 *)&text[lastIndex],i-lastIndex,&extents);
				x+=extents.xOff;
			}
			lastIndex=i;
			lastFont=currentFont;
		}
	}
	if(lastFont)
	{
		XftDrawString32(imageBuffer->xftDraw,&xftColor,lastFont,x,y,(const FcChar32 *)&text[lastIndex],i-lastIndex);
		XftTextExtents32(xDisplay,lastFont,(const FcChar32 *)&text[lastIndex],i-lastIndex,&extents);
		x+=extents.xOff;
	}

	XftColorFree(xDisplay,xVisualInfo32.visual,xColorMap32,&xftColor);
	return(x);
}

int FontTextUtf8Draw(EDITOR_FONT *font,GUI_IMAGE *imageBuffer,EDITOR_COLOR color,int x,int y,const char *text,uint32_t length)
// Draw length bytes of the passed UTF-8 encoded text in the passed font into the passed
// imageBuffer in the given color (ARGB) at the given x and y location.
// Return the updated X coordinate.
// This will handle fallback fonts transparently.
{
	uint32_t
		buffer[256];
	uint32_t
		bufferIndex;
	int
		bytesUsed;
	FcChar32
		ucs4;

	bufferIndex=0;
	while(length&&(bytesUsed=FcUtf8ToUcs4((const FcChar8 *)text,&ucs4,length))>0)			// walk through all the characters of the input
	{
		buffer[bufferIndex]=ucs4;
		bufferIndex++;
		if(bufferIndex>=elementsof(buffer))										// filled the local buffer
		{
			x=FontText32Draw(font,imageBuffer,color,x,y,buffer,bufferIndex);	// draw buffer so far
			bufferIndex=0;
		}
		text+=bytesUsed;
		length-=bytesUsed;
	}
	if(bufferIndex)															// any left at the end?
	{
		x=FontText32Draw(font,imageBuffer,color,x,y,buffer,bufferIndex);	// draw whatever's left
	}
	return(x);
}

// ----------------------------------------------------------------------------

void FontHeightsGet(EDITOR_FONT *font,uint32_t *ascent,uint32_t *descent)
// Return the heights for the given font.
{
	*ascent=font->ascent;
	*descent=font->descent;
}

// ----------------------------------------------------------------------------

static void FallbackFontsFree(EDITOR_FONT *font)
// Free any fallback fonts associated with font.
{
	FALLBACK_FONT
		*currentFallbackFont,
		*nextFallbackFont;

	currentFallbackFont=font->fallbackFonts;
	while(currentFallbackFont)
	{
		nextFallbackFont=currentFallbackFont->next;
		XftFontClose(xDisplay,currentFallbackFont->xftFont);
		MDisposePtr(currentFallbackFont);
		currentFallbackFont=nextFallbackFont;
	}
	font->fallbackFonts=NULL;
}

static void FallbackFontsLoad(EDITOR_FONT *font,uint32_t size,bool bold,bool oblique)
// Load all the fallback fonts for the given font.
{
	FALLBACK_FONT_NAME
		*fallbackName;
	FALLBACK_FONT
		*currentFallbackFont,
		*nextFallbackFont;
	char
		canonicalCompositeName[MAX_FONT_NAME_LENGTH];

	fallbackName=firstFallbackName;
	currentFallbackFont=NULL;
	while(fallbackName)
	{
		FontComponentNameJoin(fallbackName->familyName,size,bold,oblique,canonicalCompositeName);	// construct composite name
		if((nextFallbackFont=(FALLBACK_FONT *)MNewPtr(sizeof(FALLBACK_FONT))))						// allocate fallback font record
		{
			if((nextFallbackFont->xftFont=XftFontOpenName(xDisplay,xScreenNum,canonicalCompositeName)))	// load the fallback font
			{
				nextFallbackFont->next=NULL;			// mark end of list
				// link to the passed font
				if(currentFallbackFont)
				{
					currentFallbackFont->next=nextFallbackFont;
				}
				else
				{
					font->fallbackFonts=nextFallbackFont;	// first record on the list
				}
				currentFallbackFont=nextFallbackFont;
			}
			else	// failed to open font
			{
				MDisposePtr(nextFallbackFont);
			}
		}
		fallbackName=fallbackName->next;
	}
}

// ----------------------------------------------------------------------------

void PlatformFontCompositeNameSplit(const char *compositeName,char *familyName,uint32_t *size,bool *bold,bool *oblique)
// Split a composite font name into components.
// familyName must be at least MAX_FONT_NAME_LENGTH bytes long.
{
	FontCompositeNameSplit(compositeName,familyName,size,bold,oblique);
}

void PlatformFontComponentNameJoin(const char *familyName,uint32_t size,bool bold,bool oblique,char *compositeName)
// Take a component specified font, and convert it to the canonicalized
// composite name.
// compositeName must be at least MAX_FONT_NAME_LENGTH bytes long.
{
	FontComponentNameJoin(familyName,size,bold,oblique,compositeName);
}

const char *PlatformFontCompositeNameGet(EDITOR_FONT *font)
// Return a pointer to the composite name of a loaded font.
// The pointer will stay in scope as long as the font stays loaded.
{
	return(NamedListItemNameGet(font->listItem));
}

void PlatformFontFree(EDITOR_FONT *font)
// Free a font allocated by PlatformFontLoad.
{
	if(!(--font->usageCount))
	{
		NamedListItemDestroy(loadedFontList,font->listItem);
		FallbackFontsFree(font);
		XftFontClose(xDisplay,font->xftFont);
		MDisposePtr(font);
	}
}

EDITOR_FONT *PlatformFontLoad(const char *compositeName)
// Attempt to load the font with the given composite name.
// NOTE: this will canonicalize the given composite name.
// If there is a problem, ErrorSet, return NULL.
{
	NAMED_LIST_ITEM
		*listItem;
	EDITOR_FONT
		*result;
	int
		i;
	char
		familyName[MAX_FONT_NAME_LENGTH];
	uint32_t
		size;
	bool
		bold,
		oblique;
	char
		canonicalCompositeName[MAX_FONT_NAME_LENGTH];

	// split and re-join the composite name to make it canonical, also get parameters associated with it
	FontCompositeNameSplit(compositeName,familyName,&size,&bold,&oblique);
	FontComponentNameJoin(familyName,size,bold,oblique,canonicalCompositeName);

	if((result=FontLoadedLocate(canonicalCompositeName)))		// see if we can find this one already loaded
	{
		result->usageCount++;									// add another user to this font structure
		return(result);
	}
	else
	{
		if((listItem=NamedListItemCreate(loadedFontList,canonicalCompositeName)))	// create a list item to link this font to
		{
			if((result=(EDITOR_FONT *)MNewPtr(sizeof(EDITOR_FONT))))	// allocate one
			{
				if((result->xftFont=XftFontOpenName(xDisplay,xScreenNum,canonicalCompositeName)))	// load the font
				{
					result->fallbackFonts=NULL;
					FallbackFontsLoad(result,size,bold,oblique);	// setup fallback fonts

					NamedListItemDataSet(listItem,result);			// place it onto the list
					result->listItem=listItem;
					result->usageCount=1;
					result->ascent=result->xftFont->ascent;
					result->descent=result->xftFont->descent;

					for(i=0;i<256;i++)
					{
						result->charExists[i]=FontCodePoint32WidthSlow(result,i,&result->charWidths[i]);
					}
					return(result);
				}
				MDisposePtr(result);
			}
			NamedListItemDestroy(loadedFontList,listItem);
		}
	}
	return(NULL);
}

EDITOR_FONT *PlatformFontDuplicate(EDITOR_FONT *original)
// Create a font structure which points to the same font as original.
// If there's a problem, return NULL.
// The returned font must eventually be freed by a call to
// PlatformFontFree.
{
	original->usageCount++;								// add another user to this font structure
	return(original);
}

// ----------------------------------------------------------------------------

bool FontsFallbackListAppend(const char *familyName)
// Add a font name to the end of the list of fallback names.
// On failure, ErrorSet and return false.
// NOTE: if this fails, it will leave the current list of fallback
// names untouched.
{
	uint32_t
		length;
	FALLBACK_FONT_NAME
		*fallbackName;

	length=strlen(familyName)+1;						// get the buffer length needed to hold the name and 0 termination
	if((fallbackName=(FALLBACK_FONT_NAME *)MNewPtr(sizeof(FALLBACK_FONT_NAME)+length)))
	{
		strcpy(fallbackName->familyName,familyName);	// copy over the family name
		fallbackName->next=NULL;						// mark the end of the list
		if(lastFallbackName)							// link to the end of the global list
		{
			lastFallbackName->next=fallbackName;		// point last entry at this one
			lastFallbackName=fallbackName;				// make this one the last entry
		}
		else
		{
			firstFallbackName=lastFallbackName=fallbackName;	// only entry on the list so far
		}
		return(true);									// success
	}
	return(false);
}

void FontsFallbackListClear()
// Delete the list of fallback font names that are used when any font is opened.
{
	FALLBACK_FONT_NAME
		*fallbackName,
		*nextFallbackName;

	fallbackName=firstFallbackName;
	while(fallbackName)
	{
		nextFallbackName=fallbackName->next;
		MDisposePtr(fallbackName);
		fallbackName=nextFallbackName;
	}
	firstFallbackName=NULL;
	lastFallbackName=NULL;
}

// ----------------------------------------------------------------------------

void FontsUnInit()
// Undo what FontsInit did.
{
	// NOTE: no fonts should be on the loadedFontList at this point
	NamedListDestroy(loadedFontList);					// kill the list itself
	FontsFallbackListClear();							// get rid of all fallback names
}

bool FontsInit()
// Set up font handling stuff.
{
	firstFallbackName=NULL;
	lastFallbackName=NULL;
	if((loadedFontList=NamedListCreate()))
	{
		return(true);
	}
	return(false);
}
