// Manage non-font drawing functions.
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// In the 26 years between writing e93 and the time I'm typing this
// comment (Oct. 2019), X11 has come a long way in terms of
// functionality. However, it has done so at the cost of considerable
// complexity in the form of extensions, etc...
// When trying to write this code, I spent many hours searching the
// Internet for documentation and example code to try to work out how to
// use the Xrender library and X rendering extension. My goal was just to
// be able to draw 24-bit RGB with alpha, and get that rendered into a
// TrueColor window (which may, or may not support alpha).
// It should be simple, and in the end, XRenderComposite does make it
// straightforward, but working through the details needed to use it is
// mind numbing. It's no wonder most people use other libraries built on
// top of this stuff to get it to work...
// Since e93 is meant to be lightweight and quick, I really didn't want
// to do that.
// As far as the Xrender library/extension documentation goes, it appears
// the only formal documents are these incredibly thin attempts:
//
// https://www.x.org/releases/X11R7.5/doc/libXrender/libXrender.txt
// https://keithp.com/~keithp/render/protocol.html
//
// To those that might have found this code on the same hunt I was on, it
// is possible to generate 32-bit offscreen buffers (on the X server),
// draw into those buffers with alpha, and then use XRenderComposite to
// copy the result to a window's drawable, even if that drawable is of a
// different depth (e.g. 24 bit).
// That's how all the code here works.

#include	"includes.h"

// ----------------------------------------------------------------------------

static void PixmapDestroy(Pixmap pixmap)
// Destroy of a pixmap created by PixmapCreate.
{
	XFreePixmap(xDisplay,pixmap);
}

static Pixmap PixmapCreate(uint32_t width,uint32_t height)
// Create a server-side pixmap.
// NOTE: it is possible for this to return None if the pixmap could
// not be created.
{
	Pixmap
		result;

	if(width<1)			// keep these from being 0 (X does not like that)
	{
		width=1;
	}
	if(height<1)		// keep these from being 0 (X does not like that)
	{
		height=1;
	}
	result=XCreatePixmap(xDisplay,RootWindow(xDisplay,xScreenNum),width,height,32);
	return(result);		// send back whatever we have (even if None)
}

// ----------------------------------------------------------------------------

void ImageEditorColorToRenderColorAlpha(EDITOR_COLOR color,XRenderColor *renderColor)
// Convert the given editor color (ARGB) to a pre-multiplied render color.
// 
{
	uint32_t
		alpha;
	uint32_t
		value;

	alpha=(color>>24)&0xFF;					// get alpha from MSB
	alpha=(alpha<<8)|alpha;					// extend to 16 bits
	renderColor->alpha=alpha;

	value=(color>>16)&0xFF;					// do same for R,G,B
	value=(value<<8)|value;
	value=(value*alpha)>>16;				// premultiply the alpha
	renderColor->red=value;

	value=(color>>8)&0xFF;
	value=(value<<8)|value;
	value=(value*alpha)>>16;
	renderColor->green=value;

	value=(color>>0)&0xFF;
	value=(value<<8)|value;
	value=(value*alpha)>>16;
	renderColor->blue=value;
}

void ImageEditorColorToRenderColor(EDITOR_COLOR color,XRenderColor *renderColor)
// Convert the given editor color (RGB) to a render color with alpha=0xFF.
{
	uint32_t
		alpha;
	uint32_t
		value;

	alpha=0xFF;
	alpha=(alpha<<8)|alpha;					// extend to 16 bits
	renderColor->alpha=alpha;

	value=(color>>16)&0xFF;					// do same for R,G,B
	value=(value<<8)|value;
	renderColor->red=value;

	value=(color>>8)&0xFF;
	value=(value<<8)|value;
	renderColor->green=value;

	value=(color>>0)&0xFF;
	value=(value<<8)|value;
	renderColor->blue=value;
}

// ----------------------------------------------------------------------------

void ImageClipRegionSet(GUI_IMAGE *image,Region region)
// Set the clip region for an image to region.
// Subsequent drawing will be clipped to this region.
// NOTE: the passed region may be deleted immediately after this
// is called.
// If the region is passed as "None", any clip region in the
// image will be cleared.
{
	XRenderPictureAttributes
		pictureAttributes;

	XftDrawSetClip(image->xftDraw,region);												// set it in the xftDraw (passing None clears it)
	if(region)
	{
		XRenderSetPictureClipRegion(xDisplay,image->picture,region);					// passing "None" does not clear it here -- we need to XRenderChangePicture instead
	}
	else
	{
		pictureAttributes.clip_mask=None;
		XRenderChangePicture(xDisplay,image->picture,CPClipMask,&pictureAttributes);	// clear the region
	}
}

void ImageClipRegionRectSet(GUI_IMAGE *image,EDITOR_RECT *rect)
// Set the clip region for an image to the passed rectangle.
// Subsequent drawing will be clipped to this region.
{
	Region
		region;
	XRectangle
		xRect;

	region=XCreateRegion();
	xRect.x=rect->x;
	xRect.y=rect->y;
	xRect.width=rect->w;
	xRect.height=rect->h;
	XUnionRectWithRegion(&xRect,region,region);						// create a region which is the passed rectangle
	ImageClipRegionSet(image,region);	
	XDestroyRegion(region);											// get rid of region now that it has been set
}

void ImageDestroy(GUI_IMAGE *image)
// Get rid of an GUI_IMAGE created by ImageCreate.
{
	XRenderFreePicture(xDisplay,image->picture);
	XftDrawDestroy(image->xftDraw);
	PixmapDestroy(image->pixmap);
	MDisposePtr(image);
}

GUI_IMAGE *ImageCreate(uint32_t width,uint32_t height)
// Create a 32-bit ARGB off-screen image on the server to draw into.
// If this fails, it will return NULL.
{
	GUI_IMAGE
		*image;
	XRenderPictFormat
		*pictFormat;
	XRenderPictureAttributes
		pictAttr;

	if((image=(GUI_IMAGE *)MNewPtr(sizeof(GUI_IMAGE))))										// create our local structure
	{
		image->width=width;
		image->height=height;
		if((image->pixmap=PixmapCreate(width,height))!=None)								// get a pixmap on the server
		{
			if((image->xftDraw=XftDrawCreate(xDisplay,image->pixmap,xVisualInfo32.visual,xColorMap32)))
			{
				if((pictFormat=XRenderFindStandardFormat(xDisplay,PictStandardARGB32)))		// get XRender format
				{
					pictAttr.component_alpha=True;
					image->picture=XRenderCreatePicture(xDisplay,image->pixmap,pictFormat,CPComponentAlpha,&pictAttr);
					return(image);															// return the structure
				}
				XftDrawDestroy(image->xftDraw);
			}
			PixmapDestroy(image->pixmap);
		}
		MDisposePtr(image);
	}
	return(NULL);
}

GUI_IMAGE *ImageRead(const char *path)
// Read an image from path into a GUI_IMAGE which can
// be used as the source or dest of various graphics/render functions.
// If there's a problem, ErrorSet and return NULL.
{
	XImage
		*xImage;
	GUI_IMAGE
		*image;
	GC
		context;

	image=NULL;
	if((xImage=ImageFileXImageRead(path)))
	{
		if((image=ImageCreate(xImage->width,xImage->height)))
		{
			context=XCreateGC(xDisplay,image->pixmap,0,NULL);
			XPutImage(xDisplay,image->pixmap,context,xImage,0,0,0,0,xImage->width,xImage->height);		// push the xImage into the pixmap on the server
			XFreeGC(xDisplay,context);		// don't need this context anymore
		}
		ImageFileXImageDestroy(xImage);
	}
	return(image);
}

// ----------------------------------------------------------------------------

void ImageDraw(const GUI_IMAGE *source,GUI_IMAGE *destination,int32_t destX,int32_t destY)
// Draw an image given by source into destination at destX, destY.
{
	XRenderComposite(xDisplay,PictOpOver,source->picture,None,destination->picture,0,0,0,0,destX,destY,source->width,source->height);
}

void ImageDrawMask(const GUI_IMAGE *source,EDITOR_COLOR color,GUI_IMAGE *destination,int32_t destX,int32_t destY)
// Treat source as a set of alphas which are multiplied by the components of color,
// then sent to the destination.
// This is useful for drawing symbols and icons in different colors.
{
	Picture
		maskPicture;
	XRenderColor
		renderColor;
	XRenderPictureAttributes
		pictureAttributes;

	ImageEditorColorToRenderColorAlpha(color,&renderColor);
	maskPicture=XRenderCreateSolidFill(xDisplay,&renderColor);
	pictureAttributes.component_alpha=True;
	XRenderChangePicture(xDisplay,maskPicture,CPComponentAlpha,&pictureAttributes);
	XRenderComposite(xDisplay,PictOpOver,maskPicture,source->picture,destination->picture,0,0,0,0,destX,destY,source->width,source->height);
	XRenderFreePicture(xDisplay,maskPicture);
}

void ImageDrawToWindow(const GUI_IMAGE *source,Drawable destination,Region invalidRegion,uint32_t width,uint32_t height)
// Draw an image into a window. This will handle windows with a depth of 24 or 32.
// invalidRegion tells the areas that should actually be altered.
{
	XRenderPictFormat
		*pictFormat;
	XRenderPictureAttributes
		pictAttr;
	Picture
		destPicture;

	pictFormat=NULL;
	if(DefaultDepth(xDisplay,xScreenNum)==24)
	{
		pictFormat=XRenderFindStandardFormat(xDisplay,PictStandardRGB24);
	}
	else if(DefaultDepth(xDisplay,xScreenNum)==32)
	{
		pictFormat=XRenderFindStandardFormat(xDisplay,PictStandardARGB32);
	}
	if(pictFormat)
	{
		pictAttr.component_alpha=False;
		destPicture=XRenderCreatePicture(xDisplay,destination,pictFormat,CPComponentAlpha,&pictAttr);
		XRenderSetPictureClipRegion(xDisplay,destPicture,invalidRegion);
		XRenderComposite(xDisplay,PictOpSrc,source->picture,None,destPicture,0,0,0,0,0,0,source->width,source->height);
		XRenderFreePicture(xDisplay,destPicture);
	}
}

void ImageFillAlpha(GUI_IMAGE *dest,EDITOR_COLOR color,int32_t x,int32_t y,uint32_t w,uint32_t h)
// Fill (with alpha) the given rectangle of the destination.
{
	XRenderColor
		renderColor;

	ImageEditorColorToRenderColorAlpha(color,&renderColor);
	XRenderFillRectangle(xDisplay,PictOpOver,dest->picture,&renderColor,x,y,w,h);
}

void ImageFill(GUI_IMAGE *dest,EDITOR_COLOR color,int32_t x,int32_t y,uint32_t w,uint32_t h)
// Fill the given rectangle of the destination w/o alpha.
// This is used when drawing the backgrounds of various windows.
{
	XRenderColor
		renderColor;

	ImageEditorColorToRenderColor(color,&renderColor);
	XRenderFillRectangle(xDisplay,PictOpSrc,dest->picture,&renderColor,x,y,w,h);
}

void ImageUnInit()
// Undo what ImageInit did.
{
}

bool ImageInit()
// Set up image handling.
{
	return(true);
}
