// Read images from .png files
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static void PngErrorCallback(png_structp pngPointer,png_const_charp message)
// Hideous callback needed to deal with libpng errors.
{
	ErrorSet("libpng error: %s",message);
	longjmp(png_jmpbuf(pngPointer),1);		// go back to the setjmp
}

static void ImageFilePngDestroy(uint8_t *pngImage)
// Free memory returned by ImageFilePngRead.
{
	MDisposePtr(pngImage);
}

static uint8_t *ImageFilePngRead(FILE *inputFile,uint32_t *width,uint32_t *height,uint32_t *rowBytes,uint32_t *colorType)
// Read in a PNG into a buffer and return the buffer. If there's a
// problem, ErrorSet and return NULL.
// NOTE: this uses libpng. IMHO, libpng is a mess. It is implemented in a
// way that shares almost none of my sensibility for how clean code
// should be implemented. This is especially evident in its error
// handling which uses longjump (ugh!).
{
	png_structp
		pngPointer;
	png_infop
		readInfoPointer;
	png_uint_32
		localWidth,
		localHeight;
	int
		bitDepth,
		localColorType,
		interlaceType,
		compressionType,
		filterType;
	uint32_t
		y;
	uint8_t
		*pngImage;

	pngImage=NULL;
	if((pngPointer=png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,PngErrorCallback,NULL)))
	{
		if(setjmp(png_jmpbuf(pngPointer))==0)
		{
			if((readInfoPointer=png_create_info_struct(pngPointer)))
			{
				png_init_io(pngPointer,inputFile);
				png_read_info(pngPointer,readInfoPointer);
				if(png_get_IHDR(pngPointer,readInfoPointer,&localWidth,&localHeight,&bitDepth,&localColorType,&interlaceType,&compressionType,&filterType))
				{
					(*width)=localWidth;
					(*height)=localHeight;
					(*rowBytes)=png_get_rowbytes(pngPointer,readInfoPointer);
//					fprintf(stdout,"width=%ld, height=%ld, bitDepth=%d, colorType=%d, rowBytes=%d\n",width,height,bitDepth,*colorType,*rowBytes);
					if((localColorType==PNG_COLOR_TYPE_RGB)||(localColorType==PNG_COLOR_TYPE_RGB_ALPHA))
					{
						(*colorType)=localColorType;							// set output value
						if(bitDepth==8)
						{
							if((pngImage=(uint8_t *)MNewPtr((*rowBytes)*localHeight)))	// make buffer to hold image data
							{
								uint8_t
									*rowPointers[localHeight];					// pointers to each row (needed by png_read_image);

								for(y=0;y<localHeight;y++)
								{
									rowPointers[y]=&pngImage[y*(*rowBytes)];	// make row pointers
								}

								png_read_image(pngPointer,rowPointers);	// read the image all at once, since it may be interlaced, and we don't want to cope with that here
							}
							else
							{
								ErrorSet("Failed to allocate image data");
							}
						}
						else
						{
							ErrorSet("Unexpected PNG bit depth (want 8)");
						}
					}
					else
					{
						ErrorSet("Unexpected PNG color type (want RGB, or RGBA)");
					}
				}
				else
				{
					ErrorSet("Failed to png_get_IHDR");
				}
			}
			else
			{
				ErrorSet("Failed to png_create_info_struct");
			}
		}
		else
		{
			// error is set in callback
		}
		png_destroy_read_struct(&pngPointer,&readInfoPointer,NULL);
	}
	else
	{
		ErrorSet("Failed to png_create_read_struct");
	}
	return(pngImage);
}

static XImage *ImageFilePngToXImage(uint8_t *pngImage,uint32_t width,uint32_t height,uint32_t rowBytes,uint32_t colorType)
// Convert a PNG image to an XImage.
{
	uint32_t
		outputRowBytes;
	uint32_t
		x,y;
	uint8_t
		r,g,b,a;
	uint8_t
		*inputPointer,
		*outputPointer;
	XImage
		*xImage;

	xImage=NULL;
	outputRowBytes=width*4;								// calculate number of bytes in an output row of the XImage
	if((xImage=XCreateImage(xDisplay,xVisualInfo32.visual,32,ZPixmap,0,NULL,width,height,32,outputRowBytes)))
	{
		if((xImage->data=(char *)MNewPtr(outputRowBytes*height)))	// make buffer to hold xImage data
		{
			outputPointer=(uint8_t *)xImage->data;
			for(y=0;y<height;y++)						// convert PNG data to XImage format
			{
				inputPointer=&pngImage[y*rowBytes];		// point to start of row of input data
				for(x=0;x<width;x++)
				{
					r=(*inputPointer++);
					g=(*inputPointer++);
					b=(*inputPointer++);
					if(colorType==PNG_COLOR_TYPE_RGB)	// handle RGB
					{
						a=0xFF;
					}
					else
					{
						a=(*inputPointer++);			// grab alpha from the input too
					}
					(*outputPointer++)=(b*a)>>8;		// premultiply, write in xImage order
					(*outputPointer++)=(g*a)>>8;
					(*outputPointer++)=(r*a)>>8;
					(*outputPointer++)=a;
				}
			}
		}
		else
		{
			ErrorSet("Failed to allocate image data");
			XDestroyImage(xImage);
			xImage=NULL;
		}
	}
	else
	{
		ErrorSet("Failed to XCreateImage");
	}
	return(xImage);
}

static unsigned long *ImageFilePngToCardinal32(uint8_t *pngImage,uint32_t width,uint32_t height,uint32_t rowBytes,uint32_t colorType,uint32_t *numElements)
// Convert a PNG image to 32-bit CARDINAL format needed by _NET_WM_ICON.
// This is useful for making the icon data that will be shown by the window manager.
// NOTE: the "unsigned long" type here is needed because the 32-bit CARDINAL
// type given to XChangeProperty, causes it to expect to be passed an array
// of longs. On a 64-bit build, these become 64-bit values. It's kind of
// wasteful as half the memory allocated is unused...
{
	uint32_t
		x,y;
	uint8_t
		r,g,b,a;
	uint8_t
		*inputPointer;
	unsigned long
		*cardinalImage,
		*outputPointer;

	(*numElements)=(2+width*height);				// number of unsigned longs in the output (add 2 for width and height)
	if((cardinalImage=(unsigned long *)MNewPtr((*numElements)*sizeof(unsigned long))))	// make the output buffer
	{
		outputPointer=cardinalImage;
		(*outputPointer++)=width;					// output the width and height first
		(*outputPointer++)=height;

		for(y=0;y<height;y++)						// convert PNG data to cardinal-32 format
		{
			inputPointer=&pngImage[y*rowBytes];		// point to start of row of input data
			for(x=0;x<width;x++)
			{
				r=(*inputPointer++);
				g=(*inputPointer++);
				b=(*inputPointer++);
				if(colorType==PNG_COLOR_TYPE_RGB)	// handle RGB
				{
					a=0xFF;
				}
				else
				{
					a=(*inputPointer++);			// grab alpha from the input too
				}
				(*outputPointer++)=(a<<24)|(r<<16)|(g<<8)|b;
			}
		}
	}
	else
	{
		ErrorSet("Failed to allocate Cardinal32 buffer");
	}
	return(cardinalImage);
}


void ImageFileXImageDestroy(XImage *xImage)
// Get rid of an XImage created by ImageFileXImageRead.
{
	MDisposePtr(xImage->data);
	xImage->data=NULL;		// keep XDestroyImage from trying to free the data
	XDestroyImage(xImage);
}

XImage *ImageFileXImageRead(const char *path)
// Inhale an image from path, build an Ximage, return that.
// If NULL is returned, path did not contain a readable image.
// This uses libpng to read RGBA PNGs.
{
	FILE
		*imageFile;
	uint8_t
		*pngImage;
	uint32_t
		width,
		height,
		rowBytes,
		colorType;
	XImage
		*xImage;

	xImage=NULL;
	if((imageFile=fopen(path,"rb")))
	{
		if((pngImage=ImageFilePngRead(imageFile,&width,&height,&rowBytes,&colorType)))
		{
			if((xImage=ImageFilePngToXImage(pngImage,width,height,rowBytes,colorType)))
			{
				// success
			}
			ImageFilePngDestroy(pngImage);
		}
		fclose(imageFile);
	}
	else
	{
		ErrorSet("Failed to open '%s': %s",path,strerror(errno));
	}
	return(xImage);
}

void ImageFileCardinal32Destroy(unsigned long *cardinalImage)
// Get rid of an 32-bit cardinal image created by ImageFileCardinal32Read.
{
	MDisposePtr(cardinalImage);
}

unsigned long *ImageFileCardinal32Read(const char *path,uint32_t *numElements)
// Inhale an image from path, convert to the 32-bit CARDINAL format needed by _NET_WM_ICON, return that.
// If NULL is returned, path did not contain a readable image.
// This uses libpng to read RGBA PNGs.
{
	FILE
		*imageFile;
	uint8_t
		*pngImage;
	uint32_t
		width,
		height,
		rowBytes,
		colorType;
	unsigned long
		*cardinalImage;

	cardinalImage=NULL;
	if((imageFile=fopen(path,"rb")))
	{
		if((pngImage=ImageFilePngRead(imageFile,&width,&height,&rowBytes,&colorType)))
		{
			if((cardinalImage=ImageFilePngToCardinal32(pngImage,width,height,rowBytes,colorType,numElements)))
			{
				// success
			}
			ImageFilePngDestroy(pngImage);
		}
		fclose(imageFile);
	}
	else
	{
		ErrorSet("Failed to open '%s': %s",path,strerror(errno));
	}
	return(cardinalImage);
}
