// Low level stuff needed to init machine
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static int XErrorEventHandler(Display *display,XErrorEvent *event)
// Handle protocol errors (generated by strange X implementations, or window managers).
{
//	char
//		errorMessage[256];

//	XGetErrorText(display,event->error_code,errorMessage,256);
//	fprintf(stderr,"X Error: %s\n",errorMessage);

	xFail=true;
	return(0);
}

void PlatformLateUnInit()
// Undo what PlatformLateInit did.
{
	DialogUnInit();
}

bool PlatformLateInit()
// This is called after the editor has initialized. It can perform
// any final initialization of the platform that wants to call into
// the editor.
{
	if(DialogInit())		// dialog init happens late because dialogs can use views, and to do that, the editor needs to be alive
	{
		return(true);
	}
	return(false);
}

void PlatformUnInit()
// Undo what PlatformInit did.
{
	XFreeCursor(xDisplay,arrowCursor);
	XFreeCursor(xDisplay,caretCursor);
	TasksUnInit();
	EventsUnInit();
	DocumentWindowUnInit();
	MenuUnInit();
	FontsUnInit();
	ImageAssetUnInit();
	ImageUnInit();
	XClipboardUnInit();
	ColorsUnInit();
	XFreeColormap(xDisplay,xColorMap32);
	XCloseDisplay(xDisplay);
}

bool PlatformInit()
// Initialize the machine, software libraries, etc...
// This is called before the editor inits, so nothing called
// here is allowed to call back into the editor, except for
// the error handling functions.
// If there is a problem, set the error and return false.
{
	windowListHead=NULL;						// no windows are open on the display yet
	XSetErrorHandler(XErrorEventHandler);		// set error handling routine
	if((xDisplay=XOpenDisplay(NULL)))			// connect to the default X server
	{
		xScreenNum=DefaultScreen(xDisplay);		// keep this around for handy future reference

		// define atoms needed by top-level windows
		wmNameAtom=XInternAtom(xDisplay,"_NET_WM_NAME",False);							// used to provide UTF8 title for window
		iconNameAtom=XInternAtom(xDisplay,"_NET_WM_ICON_NAME",False);					// used to provide UTF8 title for icon
		iconAtom=XInternAtom(xDisplay,"_NET_WM_ICON",False);							// used to provide color icon to window manager
		targetsAtom=XInternAtom(xDisplay,"TARGETS",False);								// used by some selection requestors to find out which targets (kinds of selection) we support
		selectionAtom=XInternAtom(xDisplay,"SELECTED_DATA",False);						// place where we receive selected data from other selection owners
		utf8StringAtom=XInternAtom(xDisplay,"UTF8_STRING",False);						// interface with programs that support UTF8 when importing clipboard
		incrAtom=XInternAtom(xDisplay,"INCR",False);									// incremental selection passing atom
		takeFocusAtom=XInternAtom(xDisplay,"WM_TAKE_FOCUS",False);
		deleteWindowAtom=XInternAtom(xDisplay,"WM_DELETE_WINDOW",False);

		if(XMatchVisualInfo(xDisplay,xScreenNum,32,TrueColor,&xVisualInfo32)!=0)		// make sure server supports 32-bit drawables with alpha (we use them for offscreen rendering)
		{
			xColorMap32=XCreateColormap(xDisplay,RootWindow(xDisplay,xScreenNum),xVisualInfo32.visual,AllocNone);

			if(ColorsInit())
			{
				if(XClipboardInit())
				{
					if(ImageInit())
					{
						if(ImageAssetInit())
						{
							if(FontsInit())
							{
								if(MenuInit())
								{
									if(DocumentWindowInit())
									{
										if(EventsInit())
										{
											if(TasksInit())
											{
												CursorBlinkTimeReset();					// reset the timeout value of the cursor blink timer
												showingBusy=0;							// reset cursor busy depth
												verticalScrollBarOnLeft=false;			// by default keep the scroll bar on the right
												caretCursor=XCreateFontCursor(xDisplay,XC_xterm);	// create the pointer cursor to use when inside view
												arrowCursor=XCreateFontCursor(xDisplay,XC_arrow);	// create the arrow cursor to use for other things
												timeToQuit=false;						// clear the quit request flag
												return(true);
											}
											EventsUnInit();
										}
										DocumentWindowUnInit();
									}
									MenuUnInit();
								}
								FontsUnInit();
							}
							ImageAssetUnInit();
						}
						ImageUnInit();
					}
					XClipboardUnInit();
				}
				ColorsUnInit();
				XFreeColormap(xDisplay,xColorMap32);
			}
		}
		else
		{
			ErrorSet("X server doesn't support 32-bit drawables");
		}
		XCloseDisplay(xDisplay);
	}
	else
	{
		ErrorSet("Could not open display");
	}
	return(false);
}

void PlatformEarlyUnInit()
// This is the very last thing that the editor calls before exiting.
{
}

bool PlatformEarlyInit()
// This is called before anything else in the editor, it can be used to fork off the
// editor as a separate process, or do anything which may need to be done very early.
// If this returns false, the editor will exit without complaint.
{
	return(true);
}

