// List box handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	LIST_BOX_TOP_OFFSET		4			// pixel offsets within the list box that keep the text of the items from drawing too close to the edge
#define	LIST_BOX_LEFT_OFFSET	4
#define	LIST_BOX_BOTTOM_OFFSET	4

#define	ITEM_SPACE_TOP			2			// pixels to add to each line of text
#define	ITEM_SPACE_BOTTOM		2

#define	MAX_LIST_CHAR_BUFFER	64			// max characters in the character buffer
#define	LIST_CHAR_TIMEOUT		(SECOND*.4)	// max time between consecutive typed characters
#define	LIST_BOX_BORDER_WIDTH	3			// width of border that surrounds the list box

typedef struct listBox
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;				// function which is used to position this item in the dialog
	EDITOR_RECT
		rect;							// position within the parent
	EDITOR_COLOR
		activeBackgroundColor,			// color to use when drawing background during focus
		activeForegroundColor,			// color to use when drawing foreground during focus
		activeBorderColor,				// color to draw around the border during focus
		inactiveBackgroundColor,		// color to use when drawing background during non focus
		inactiveForegroundColor,		// color to use when drawing foreground during non focus
		inactiveBorderColor;			// color to draw around the border during non focus
	bool
		verticalScrollBarOnLeft;		// gives the preference for vertical scroll bar placement
	SCROLL_BAR
		*scrollBar;						// pointer to the scrollbar attached to this list
	uint32_t
		numElements;					// tells how many elements are in the list
	const char
		**listElements;					// pointer to array of pointers to items in the list
	bool
		*selectedElements;				// pointer to array of bools which tell which items are selected
	uint32_t
		currentElement;					// this is the element last operated on
	uint32_t
		numLines;						// tells how many lines in the current font, and list box size can be displayed
	uint32_t
		topLine;						// tells which line is displayed at the top of the list box
	uint32_t
		lineHeight;						// height of a line of text
	uint32_t
		charBufferLength;				// number of characters in the charBuffer
	char
		charBuffer[MAX_LIST_CHAR_BUFFER];	// keeps track of the characters typed
	uint32_t
		charTimer;						// keeps track of the time between characters when hunting for an item (if the time is too long between characters, charBuffer is flushed)
	uint32_t
		scrollTimer;					// keeps track of the time between scrolls when tracking the pointer
	EDITOR_FONT
		*font;							// info about currently selected font
	listBoxPressedFunction
		*pressedFunction;				// routine called when the list items have been chosen (by double clicking in the list)
} LIST_BOX;


static void ListBoxInvalidate(LIST_BOX *listBox)
// Invalidate the list box within its parent window.
{
	WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&listBox->rect);
}

static void ListBoxControlRectsGet(LIST_BOX *listBox,EDITOR_RECT *borderRect,EDITOR_RECT *listContentRect,EDITOR_RECT *scrollBarRect)
// Return the rectangles that enclose the various pieces of listBox.
// If the list box starts getting too small, certain rects may be returned with 0 width or
// height.
{
	EDITOR_RECT
		insetRect;

	insetRect=(*borderRect)=listBox->rect;				// border rect is the whole thing

	insetRect.x+=LIST_BOX_BORDER_WIDTH;					// inset from the border to the content area of the list box
	insetRect.y+=LIST_BOX_BORDER_WIDTH;
	if(insetRect.w>2*LIST_BOX_BORDER_WIDTH)
	{
		insetRect.w-=2*LIST_BOX_BORDER_WIDTH;
	}
	else
	{
		insetRect.w=0;
	}
	if(insetRect.h>2*LIST_BOX_BORDER_WIDTH)
	{
		insetRect.h-=2*LIST_BOX_BORDER_WIDTH;
	}
	else
	{
		insetRect.h=0;
	}
	listContentRect->x=insetRect.x;
	listContentRect->y=insetRect.y;
	if(insetRect.w>SCROLL_BAR_WIDTH)
	{
		listContentRect->w=insetRect.w-SCROLL_BAR_WIDTH;
		if(listBox->verticalScrollBarOnLeft)
		{
			scrollBarRect->x=insetRect.x;
			listContentRect->x+=SCROLL_BAR_WIDTH;
		}
		else
		{
			scrollBarRect->x=insetRect.x+insetRect.w-SCROLL_BAR_WIDTH;
		}
		scrollBarRect->w=SCROLL_BAR_WIDTH;
	}
	else
	{
		listContentRect->w=0;
		scrollBarRect->x=insetRect.x;
		scrollBarRect->w=insetRect.w;
	}
	listContentRect->h=insetRect.h;

	scrollBarRect->y=insetRect.y;
	scrollBarRect->h=insetRect.h;
}

static void ListContentsInvalidate(LIST_BOX *listBox)
// Invalidate the list area of listBox.
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);
	WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&listContentRect);
}

static void ListBoxElementRectGet(LIST_BOX *listBox,uint32_t element,EDITOR_RECT *rect)
// Get back the rectangle for element in listBox.
// NOTE: if element is invalid, or element is not displayed,
// an empty rectangle is returned.
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	if(element>=listBox->topLine&&element<listBox->numElements&&element<listBox->topLine+listBox->numLines)
	{
		ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);

		rect->x=listContentRect.x;
		rect->y=listContentRect.y+LIST_BOX_TOP_OFFSET+(listBox->lineHeight)*(element-listBox->topLine);
		rect->w=listContentRect.w;
		rect->h=listBox->lineHeight;
	}
	else
	{
		rect->x=rect->y=0;
		rect->w=rect->h=0;
	}
}

static void ListBoxElementInfoRecalculate(LIST_BOX *listBox)
// Calculate some often needed constants.
{
	uint32_t
		ascent,
		descent;
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);
	FontHeightsGet(listBox->font,&ascent,&descent);

	listBox->lineHeight=ITEM_SPACE_TOP+ascent+descent+ITEM_SPACE_BOTTOM;

	// find out how many lines can be displayed in the box given the size, and font
	if(listContentRect.h>(LIST_BOX_TOP_OFFSET+LIST_BOX_BOTTOM_OFFSET))
	{
		listBox->numLines=(listContentRect.h-LIST_BOX_TOP_OFFSET-LIST_BOX_BOTTOM_OFFSET)/listBox->lineHeight;
	}
	else
	{
		listBox->numLines=0;
	}
}

static void ListBoxTopLineAdjust(LIST_BOX *listBox,uint32_t newTopLine)
// Set the top line of listBox to the given value (or the closest value that
// is legal).
// Update the scroll bar, and all that...
{
	uint32_t
		maxLines;

	if(newTopLine+listBox->numLines<listBox->numElements)
	{
		maxLines=listBox->numElements-listBox->numLines;
	}
	else
	{
		if(listBox->numElements>listBox->numLines)
		{
			newTopLine=listBox->numElements-listBox->numLines;
			maxLines=listBox->numElements-listBox->numLines;
		}
		else
		{
			newTopLine=0;
			maxLines=0;
		}
	}
	listBox->topLine=newTopLine;
	ScrollBarAdjust(listBox->scrollBar,maxLines,newTopLine);
}

static void ListBoxRectangleSet(LIST_BOX *listBox)
// Work out the rectangle given the current window.
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;

	if(listBox->positionFunction)	// see if non-null resize function specified
	{
		listBox->positionFunction(&listBox->dialogItem,&listBox->rect);	// get new rectangle
		ListBoxElementInfoRecalculate(listBox);							// recalculate some stuff
		ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);
		ScrollBarReposition(listBox->scrollBar,scrollBarRect.x,scrollBarRect.y,scrollBarRect.h,false);	// move scroll bar
		ListBoxTopLineAdjust(listBox,listBox->topLine);
		ListBoxInvalidate(listBox);					// invalidate it
	}
}

static void ListBoxTopLineSet(LIST_BOX *listBox,uint32_t newTopLine)
// Set the top line of listBox to the given value (or the closest value that
// is legal).
// Update the scroll bar, and all that...
{
	ListBoxTopLineAdjust(listBox,newTopLine);
	ListContentsInvalidate(listBox);				// invalidate contents
}

static void ListBoxHomeToCurrentElement(LIST_BOX *listBox)
// Make sure the current element is visible within listBox.
{
	if(listBox->currentElement<listBox->topLine)
	{
		ListBoxTopLineSet(listBox,listBox->currentElement);
	}
	else
	{
		if(listBox->currentElement>=listBox->topLine+listBox->numLines)
		{
			ListBoxTopLineSet(listBox,listBox->currentElement-listBox->numLines+1);
		}
	}
}

static void ListBoxHomeTowardCurrentElement(LIST_BOX *listBox)
// Move the list box toward the current element. This is useful while
// performing rate limited scrolls.
{
	if(listBox->currentElement<listBox->topLine)
	{
		ListBoxTopLineSet(listBox,listBox->topLine-1);
	}
	else
	{
		if(listBox->currentElement>=listBox->topLine+listBox->numLines)
		{
			ListBoxTopLineSet(listBox,listBox->topLine+1);
		}
	}
}

static void ListBoxCenterToCurrentElement(LIST_BOX *listBox)
// Place current element at the center of the list box.
{
	uint32_t
		centerElement;

	centerElement=0;
	if(listBox->currentElement>listBox->numLines/2)
	{
		centerElement=listBox->currentElement-listBox->numLines/2;
		if(centerElement>listBox->numElements-listBox->numLines+1)
		{
			centerElement=listBox->numElements-listBox->numLines+1;
		}
	}
	ListBoxTopLineSet(listBox,centerElement);
}

static void ListBoxSelectionsClear(LIST_BOX *listBox)
// Run through the elements of listBox, clearing the
// selections, invalidate any items that were changed.
{
	uint32_t
		currentElement;
	EDITOR_RECT
		itemRect;

	for(currentElement=0;currentElement<listBox->numElements;currentElement++)
	{
		if(listBox->selectedElements[currentElement])	// unselect any that were selected
		{
			listBox->selectedElements[currentElement]=false;
			ListBoxElementRectGet(listBox,currentElement,&itemRect);
			WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
		}
	}
}

static void ListBoxSelectionsSetAll(LIST_BOX *listBox)
// Run through the elements of listBox, setting all the
// selections, invalidate any items that were changed.
{
	uint32_t
		currentElement;
	EDITOR_RECT
		itemRect;

	for(currentElement=0;currentElement<listBox->numElements;currentElement++)
	{
		if(!listBox->selectedElements[currentElement])	// select any that were unselected
		{
			listBox->selectedElements[currentElement]=true;
			ListBoxElementRectGet(listBox,currentElement,&itemRect);
			WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
		}
	}
}

static bool PointedToElementGet(LIST_BOX *listBox,int32_t x,int32_t y,uint32_t *element,bool strict)
// Return element as the index to the item of listBox that
// falls under x,y in the parent window.
// NOTE: if strict is true, and no item falls under the point, return false.
// If strict is false, and no item falls under the point, return the item
// nearest the point (if the point is below all visible items, return
// the item just below the last visible item, if the point is above
// all visible items, return the item just above the first visible
// item.) The non-strict mode is used while scrolling.
{
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;
	uint32_t
		localElement;

	if(listBox->numElements)
	{
		ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);
		if(PointInRect(x,y,&listContentRect)||((!strict)&&(y>=listContentRect.y)&&(y<listContentRect.y+(int)listContentRect.h)))
		{
			if(y>=listContentRect.y+LIST_BOX_TOP_OFFSET)
			{
				localElement=(y-listContentRect.y-LIST_BOX_TOP_OFFSET)/listBox->lineHeight;
			}
			else
			{
				localElement=0;
			}
			if(localElement>=listBox->numLines)
			{
				if(listBox->numLines)
				{
					localElement=listBox->numLines-1;
				}
				else
				{
					return(false);		// no lines displayed in list!
				}
			}
			if(localElement+listBox->topLine<listBox->numElements)
			{
				*element=localElement+listBox->topLine;
			}
			else
			{
				*element=listBox->numElements-1;
			}
			return(true);
		}
		else
		{
			if(!strict)
			{
				if(y<listContentRect.y)
				{
					if(listBox->topLine)
					{
						*element=listBox->topLine-1;
					}
					else
					{
						*element=0;
					}
				}
				else
				{
					if(listBox->topLine+listBox->numLines<listBox->numElements)
					{
						*element=listBox->topLine+listBox->numLines;
					}
					else
					{
						*element=listBox->numElements-1;
					}
				}
				return(true);
			}
		}
	}
	return(false);
}

static bool ListBoxPointInside(LIST_BOX *listBox,int32_t x,int32_t y)
// Return true if parent window relative x/y are in listBox,
// false if not.
{
	return(PointInRect(x,y,&listBox->rect));
}

static void ListBoxDraw(LIST_BOX *listBox,GUI_IMAGE *imageBuffer,Region invalidRegion,bool hasFocus)
// Redraw listBox in its parent window, obey the invalid area of the parent's window.
{
	uint32_t
		ascent,
		descent;
	EDITOR_COLOR
		foreground,
		background;
	EDITOR_COLOR
		textColor;
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect,
		itemRect;
	uint32_t
		activeHeight,
		currentElement;

	ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);
	if(XRectInRegion(invalidRegion,borderRect.x,borderRect.y,borderRect.w,borderRect.h)!=RectangleOut)
	{
		if(!hasFocus)
		{
			ShadowRectangleOutline(imageBuffer,&borderRect,listBox->inactiveBorderColor,listBox->inactiveBorderColor,LIST_BOX_BORDER_WIDTH);	// drop in border
		}
		else
		{
			ShadowRectangleOutline(imageBuffer,&borderRect,listBox->activeBorderColor,listBox->activeBorderColor,LIST_BOX_BORDER_WIDTH);	// drop in border
		}

		ImageClipRegionRectSet(imageBuffer,&listContentRect);

		if(hasFocus)
		{
			foreground=listBox->activeForegroundColor;
			background=listBox->activeBackgroundColor;
		}
		else
		{
			foreground=listBox->inactiveForegroundColor;
			background=listBox->inactiveBackgroundColor;
		}

		// fill space at top and bottom where no text draws
		ImageFill(imageBuffer,background,listContentRect.x,listContentRect.y,listContentRect.w,LIST_BOX_TOP_OFFSET);

		if((listBox->numElements-listBox->topLine)>listBox->numLines)
		{
			activeHeight=LIST_BOX_TOP_OFFSET+(listBox->numLines*listBox->lineHeight);
		}
		else
		{
			activeHeight=LIST_BOX_TOP_OFFSET+((listBox->numElements-listBox->topLine)*listBox->lineHeight);
		}
		ImageFill(imageBuffer,background,listContentRect.x,listContentRect.y+activeHeight,listContentRect.w,listContentRect.h-activeHeight);

		FontHeightsGet(listBox->font,&ascent,&descent);

		for(currentElement=listBox->topLine;(currentElement<listBox->numElements)&&((currentElement-listBox->topLine)<listBox->numLines);currentElement++)
		{
			ListBoxElementRectGet(listBox,currentElement,&itemRect);
			if(!listBox->selectedElements[currentElement])					// see if this one is selected, if so, draw it as such
			{
				ImageFill(imageBuffer,background,itemRect.x,itemRect.y,itemRect.w,itemRect.h);
				textColor=foreground;
			}
			else
			{
				ImageFill(imageBuffer,foreground,itemRect.x,itemRect.y,itemRect.w,itemRect.h);
				textColor=background;
			}
			FontTextUtf8Draw(listBox->font,imageBuffer,textColor,itemRect.x+LIST_BOX_LEFT_OFFSET,itemRect.y+ITEM_SPACE_TOP+ascent,listBox->listElements[currentElement],strlen(listBox->listElements[currentElement]));
		}
		ImageClipRegionSet(imageBuffer,None);								// unset clip area

		ScrollBarDraw(listBox->scrollBar,imageBuffer,invalidRegion);
	}
}

static void ListRateLimitedScroll(LIST_BOX *listBox)
// Scroll the list box in a rate limited way.
{
	uint32_t
		currentTime;

	currentTime=PlatformMonotonicTime();
	if(currentTime-listBox->scrollTimer>=SCROLL_FOLLOW_DELAY)		// time for another scroll?
	{
		ListBoxHomeTowardCurrentElement(listBox);
		listBox->scrollTimer+=SCROLL_FOLLOW_DELAY;					// push forward time to next
		if(currentTime-listBox->scrollTimer>=SCROLL_FOLLOW_DELAY)	// if too much time passed, just catch up
		{
			listBox->scrollTimer=currentTime;
		}
	}
}

static void ListRateLimitedScrollReset(LIST_BOX *listBox)
// Reset the scroll rate limit timer.
{
	listBox->scrollTimer=PlatformMonotonicTime();					// reset the scroll timer
}

static void ListElementsTrackNormal(LIST_BOX *listBox,XEvent *event,uint32_t element)
// Track list elements normally.
{
	EDITOR_RECT
		itemRect;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	ListBoxSelectionsClear(listBox);
	listBox->selectedElements[element]=true;
	listBox->currentElement=element;
	ListBoxElementRectGet(listBox,element,&itemRect);
	WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
	rootX=event->xbutton.x_root;
	rootY=event->xbutton.y_root;
	windowX=event->xbutton.x;
	windowY=event->xbutton.y;
	while(StillDown(event->xbutton.window,event->xbutton.button,&rootX,&rootY,&windowX,&windowY))
	{
		if(PointedToElementGet(listBox,windowX,windowY,&element,false))
		{
			if(element!=listBox->currentElement)
			{
				listBox->selectedElements[listBox->currentElement]=false;
				ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
				WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);

				listBox->selectedElements[element]=true;
				ListBoxElementRectGet(listBox,element,&itemRect);
				WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);

				listBox->currentElement=element;
			}
			ListRateLimitedScroll(listBox);					// this will scroll the window if needed
		}
	}
	ListBoxHomeToCurrentElement(listBox);					// force current element into view on button release
}

static void ListElementsTrackXor(LIST_BOX *listBox,XEvent *event,uint32_t element)
// Track list elements in a mode which selects/deselects.
{
	uint32_t
		tempElement;
	EDITOR_RECT
		itemRect;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	bool
		fillMode;

	fillMode=listBox->selectedElements[element]=listBox->selectedElements[element]?false:true;
	listBox->currentElement=element;
	ListBoxElementRectGet(listBox,element,&itemRect);
	WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
	rootX=event->xbutton.x_root;
	rootY=event->xbutton.y_root;
	windowX=event->xbutton.x;
	windowY=event->xbutton.y;
	while(StillDown(event->xbutton.window,event->xbutton.button,&rootX,&rootY,&windowX,&windowY))
	{
		if(PointedToElementGet(listBox,windowX,windowY,&element,false))
		{
			if(element>listBox->currentElement)	// fill from current to new
			{
				for(tempElement=element;tempElement>listBox->currentElement;tempElement--)
				{
					listBox->selectedElements[tempElement]=fillMode;
					ListBoxElementRectGet(listBox,tempElement,&itemRect);
					WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
				}
				listBox->currentElement=element;
			}
			else
			{
				if(element<listBox->currentElement)
				{
					for(tempElement=element;tempElement<listBox->currentElement;tempElement++)
					{
						listBox->selectedElements[tempElement]=fillMode;
						ListBoxElementRectGet(listBox,tempElement,&itemRect);
						WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					}
					listBox->currentElement=element;
				}
			}
			ListRateLimitedScroll(listBox);				// this will scroll the window if needed
		}
	}
	ListBoxHomeToCurrentElement(listBox);					// force current element into view on button release
}

static void ListElementsTrackHand(LIST_BOX *listBox,XEvent *event,uint32_t element)
// Move around in the list box by tracking the mouse.
{
	int
		initialY;
	int
		lineOffset;
	uint32_t
		initialTopLine;
	uint32_t
		newTopLine;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	initialY=event->xbutton.y;
	initialTopLine=listBox->topLine;

	rootX=event->xbutton.x_root;
	rootY=event->xbutton.y_root;
	windowX=event->xbutton.x;
	windowY=event->xbutton.y;
	while(StillDown(event->xbutton.window,event->xbutton.button,&rootX,&rootY,&windowX,&windowY))
	{
		lineOffset=((initialY-windowY)/(int)listBox->lineHeight);
		if(lineOffset<0)
		{
			if(initialTopLine>=(uint32_t)-(lineOffset))
			{
				newTopLine=initialTopLine+lineOffset;
			}
			else
			{
				newTopLine=0;
			}
		}
		else
		{
			newTopLine=initialTopLine+lineOffset;
		}
		ListBoxTopLineSet(listBox,newTopLine);
	}
}

static bool ListElementsTrack(LIST_BOX *listBox,XEvent *event)
// Track the elements of the list.
{
	uint32_t
		element;
	EDITOR_RECT
		borderRect,
		listContentRect,
		scrollBarRect;
	uint8_t
		numClicks;
	uint32_t
		modifiers;

	ListBoxControlRectsGet(listBox,&borderRect,&listContentRect,&scrollBarRect);
	if(PointInRect(event->xbutton.x,event->xbutton.y,&listContentRect))
	{
		if(PointedToElementGet(listBox,event->xbutton.x,event->xbutton.y,&element,true))
		{
			numClicks=MultipleClicksHandle(event);
			modifiers=XStateToEditorModifiers(event->xbutton.state);
			switch(event->xbutton.button)
			{
				case Button1:
					if(!(modifiers&(EEM_SHIFT|EEM_CTRL)))		// ### later, make shift work additively
					{
						ListElementsTrackNormal(listBox,event,element);
					}
					else
					{
						ListElementsTrackXor(listBox,event,element);
					}
					if(numClicks&&listBox->pressedFunction)		// handle double-click of list item to exit
					{
						listBox->pressedFunction(&listBox->dialogItem);
					}
					break;
				case Button2:
					ListElementsTrackHand(listBox,event,element);
					break;
				case Button3:
					ListElementsTrackXor(listBox,event,element);
					break;
			}
			return(true);
		}
	}
	return(false);
}

static bool ListBoxTrack(LIST_BOX *listBox,XEvent *event)
// A button was pressed, see if it was in listBox, if so, track it, and return true.
// If it was not within the list box, return false.
{
	if(!ScrollBarTrack(listBox->scrollBar,event))
	{
		if(!ListElementsTrack(listBox,event))
		{
			return(false);
		}
	}
	return(true);
}

static void TypedCharactersReset(LIST_BOX *listBox)
// Reset the typed character search buffer for listBox.
{
	listBox->charBufferLength=0;						// number of characters in the charBuffer
	listBox->charTimer=PlatformMonotonicTime();					// reset the typed character timer
}

static bool BetterMatch(LIST_BOX *listBox,uint32_t element,uint32_t previousElement)
// Return true if element is a better match for the contents
// of currentBuffer, than previousElement.
// NOTE: if there is a tie, previousElement wins.
{
	int
		result;

	result=strncmp(listBox->charBuffer,listBox->listElements[previousElement],listBox->charBufferLength);	// see if char buffer so far is an exact match for the current best element
	if(result==0)
	{
		return(false);				// new element cannot be better
	}

	result=strcmp(listBox->listElements[element],listBox->listElements[previousElement]);	// see which of the two elements is larger
	if(result>0)					// new element is larger
	{
		return(strcmp(listBox->charBuffer,listBox->listElements[previousElement])>0);	// if char buffer is larger than previous, move to new
	}
	else if(result<0)				// old element is larger
	{
		return(strcmp(listBox->charBuffer,listBox->listElements[element])<=0);			// if char buffer is smaller than new, move to new
	}
	return(false);					// elements match, previous wins
}

static uint32_t ElementsSearch(LIST_BOX *listBox)
// Search through the elements of listBox, looking for
// the first, best match to charBuffer.
// Return the element located.
// NOTE: if numElements is 0, this will return 0.
{
	uint32_t
		currentElement,
		bestMatchElement;

	bestMatchElement=0;
	for(currentElement=1;currentElement<listBox->numElements;currentElement++)
	{
		if(BetterMatch(listBox,currentElement,bestMatchElement))
		{
			bestMatchElement=currentElement;
		}
	}
	return(bestMatchElement);
}

static void TypedCharCheck(LIST_BOX *listBox,char character)
// See where character should be added to the search buffer,
// add it, reset timers as needed, and hunt for the contents
// of the search buffer in listBox.
// Take the first, best match, set the current element to it,
// clear all other selections, and finally home the list to the current element.
// This allows the user to type characters while a list box is open,
// locating the element of the list that is the best match to what
// has been typed.
// NOTE: this does not depend on the list to be sorted in any way,
// but the results are better if it is.
{
	EDITOR_RECT
		itemRect;

	if((PlatformMonotonicTime()-listBox->charTimer)>=LIST_CHAR_TIMEOUT)
	{
		listBox->charBufferLength=0;								// reset index to character buffer if timeout has expired
	}
	if(listBox->charBufferLength+1<MAX_LIST_CHAR_BUFFER)
	{
		listBox->charBuffer[listBox->charBufferLength++]=character;	// if room, add the character to the buffer
		listBox->charBuffer[listBox->charBufferLength]='\0';		// terminate the string
	}
	if(listBox->numElements)
	{
		ListBoxSelectionsClear(listBox);
		listBox->currentElement=ElementsSearch(listBox);
		listBox->selectedElements[listBox->currentElement]=true;
		ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
		WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
		ListBoxCenterToCurrentElement(listBox); 					// center the selected element
	}
	listBox->charTimer=PlatformMonotonicTime();								// reset timer to current time
}

static void ListBoxScrollLower(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Scroll bar for this box is reporting a click.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	if(listBox->topLine>0)
	{
		ListBoxTopLineSet(listBox,listBox->topLine-1);
	}
}

static void ListBoxScrollHigher(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Scroll bar for this box is reporting a click.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	ListBoxTopLineSet(listBox,listBox->topLine+1);
}

static void ListBoxPageLower(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Scroll bar for this box is reporting a click.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	if(listBox->topLine>listBox->numLines)
	{
		ListBoxTopLineSet(listBox,listBox->topLine-listBox->numLines);
	}
	else
	{
		ListBoxTopLineSet(listBox,0);
	}
}

static void ListBoxPageHigher(SCROLL_BAR *scrollBar,bool repeating,void *parameters)
// Scroll bar for this box is reporting a click.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	ListBoxTopLineSet(listBox,listBox->topLine+listBox->numLines);
}

static void ListBoxThumbProc(SCROLL_BAR *scrollBar,uint64_t newValue,void *parameters)
// Scroll bar for this box is reporting a click in the thumb.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)parameters;
	ListBoxTopLineSet(listBox,newValue);
}

static void ListBoxKeyEventHandle(LIST_BOX *listBox,EDITOR_KEY *editorKey)
// A keyboard event is arriving for the list box, handle it here.
{
	EDITOR_RECT
		itemRect;

	if(listBox->numElements)
	{
		if(editorKey->vKeyCode!=VVK_NOP)					// see if the key maps to a virtual function
		{
			TypedCharactersReset(listBox);
			switch(editorKey->vKeyCode)
			{
				case VVK_UP_ARROW:
					if(listBox->selectedElements[listBox->currentElement])
					{
						if(listBox->currentElement)
						{
							ListBoxSelectionsClear(listBox);
							listBox->currentElement--;
							listBox->selectedElements[listBox->currentElement]=true;
							ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
							WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
						}
					}
					else
					{
						ListBoxSelectionsClear(listBox);
						listBox->selectedElements[listBox->currentElement]=true;
						ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
						WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					}
					ListBoxHomeToCurrentElement(listBox);
					break;
				case VVK_DOWN_ARROW:
					if(listBox->selectedElements[listBox->currentElement])
					{
						if(listBox->currentElement+1<listBox->numElements)
						{
							ListBoxSelectionsClear(listBox);
							listBox->currentElement++;
							listBox->selectedElements[listBox->currentElement]=true;
							ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
							WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
						}
					}
					else
					{
						ListBoxSelectionsClear(listBox);
						listBox->selectedElements[listBox->currentElement]=true;
						ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
						WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					}
					ListBoxHomeToCurrentElement(listBox);
					break;
				case VVK_PAGE_UP:
					if(listBox->selectedElements[listBox->currentElement])
					{
						if(listBox->currentElement)
						{
							if(listBox->currentElement>listBox->numLines)
							{
								listBox->currentElement-=listBox->numLines;
							}
							else
							{
								listBox->currentElement=0;
							}
							ListBoxSelectionsClear(listBox);
							listBox->selectedElements[listBox->currentElement]=true;
							ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
							WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
						}
					}
					else
					{
						ListBoxSelectionsClear(listBox);
						listBox->selectedElements[listBox->currentElement]=true;
						ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
						WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					}
					ListBoxHomeToCurrentElement(listBox);
					break;
				case VVK_PAGE_DOWN:
					if(listBox->selectedElements[listBox->currentElement])
					{
						if(listBox->currentElement+1<listBox->numElements)
						{
							if(listBox->currentElement+listBox->numLines<listBox->numElements)
							{
								listBox->currentElement+=listBox->numLines;
							}
							else
							{
								listBox->currentElement=listBox->numElements-1;
							}
							ListBoxSelectionsClear(listBox);
							listBox->selectedElements[listBox->currentElement]=true;
							ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
							WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
						}
					}
					else
					{
						ListBoxSelectionsClear(listBox);
						listBox->selectedElements[listBox->currentElement]=true;
						ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
						WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					}
					ListBoxHomeToCurrentElement(listBox);
					break;
				case VVK_HOME:
					ListBoxSelectionsClear(listBox);
					listBox->currentElement=0;
					listBox->selectedElements[listBox->currentElement]=true;
					ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
					WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					ListBoxHomeToCurrentElement(listBox);
					break;
				case VVK_END:
					ListBoxSelectionsClear(listBox);
					if(listBox->numElements)
					{
						listBox->currentElement=listBox->numElements-1;
						listBox->selectedElements[listBox->currentElement]=true;
					}
					else
					{
						listBox->currentElement=0;
					}
					ListBoxElementRectGet(listBox,listBox->currentElement,&itemRect);
					WindowRectInvalidate(DialogWindowGet(listBox->dialogItem.dialog),&itemRect);
					ListBoxHomeToCurrentElement(listBox);
					break;
				case VVK_SELECT_ALL:
					ListBoxSelectionsSetAll(listBox);
					break;
			}
		}
		else if(editorKey->numTextChars)		// does key map to text to search on?
		{
			TypedCharCheck(listBox,(char)editorKey->text[0]);	// just search on the first character
		}
	}
}

static void ListBoxListsReset(LIST_BOX *listBox,uint32_t numElements,const char **list,bool *selectedElements)
// Reset the list in listBox to list with selectedElements.
{
	uint32_t
		maxLines;

	listBox->numElements=numElements;
	listBox->listElements=list;
	listBox->selectedElements=selectedElements;

	TypedCharactersReset(listBox);
	listBox->topLine=listBox->currentElement=0;

	if(listBox->numLines<listBox->numElements)
	{
		maxLines=listBox->numElements-listBox->numLines;
	}
	else
	{
		if(listBox->numElements>listBox->numLines)
		{
			maxLines=listBox->numElements-listBox->numLines;
		}
		else
		{
			maxLines=0;
		}
	}
	ScrollBarAdjust(listBox->scrollBar,maxLines,0);
	ListContentsInvalidate(listBox);				// invalidate contents
}

// dialog interface routines

static void ListBoxItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the list box pointed to by item.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)item;
	ListBoxDraw(listBox,imageBuffer,invalidRegion,(item==DialogFocusItemGet(item->dialog)));
}

static void ListBoxItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)item;
	ListBoxInvalidate(listBox);				// invalidate where we're at currently
	ListBoxRectangleSet(listBox);
}

static bool ListBoxMouseWheelEventsHandle(LIST_BOX *listBox,XEvent *event)
// Added 2/12/2004 FEA.
// Call this to see if buttons 4 or 5 have been clicked. X sends button press events for buttons 4 and 5 when you
// scroll the mouse wheel. Returns true if it was a wheel event, else false.
{
	if(event->xbutton.button==Button4||event->xbutton.button==Button5)
	{
		if(event->xbutton.button==Button4)
		{
			if(listBox->topLine>3)
			{
				ListBoxTopLineSet(listBox,listBox->topLine-3);
			}
			else
			{
				ListBoxTopLineSet(listBox,0);
			}
		}
		else
		{
			ListBoxTopLineSet(listBox,listBox->topLine+3);
		}
		return(true);
	}
	return(false);
}

static bool ListBoxItemTrack(DIALOG_ITEM *item,XEvent *event)
// Attempt to track the list box pointed to by item.
{
	LIST_BOX
		*listBox;

	listBox=(LIST_BOX *)item;
	if(ListBoxPointInside(listBox,event->xbutton.x,event->xbutton.y))
	{
		ListRateLimitedScrollReset(listBox);
		if(!ListBoxMouseWheelEventsHandle(listBox,event))
		{
			DialogFocusTake(item);				// set the focus
			return(ListBoxTrack(listBox,event));
		}
		return(true);
	}
	return(false);
}

static void ListBoxItemFocusChange(DIALOG_ITEM *item)
// Focus status has changed for list box.
// Handle it here.
{
	ListBoxInvalidate((LIST_BOX *)item);
}

static bool ListBoxItemFocusKey(DIALOG_ITEM *item,EDITOR_KEY *editorKey)
// Give the key to the list box.
{
	ListBoxKeyEventHandle((LIST_BOX *)item,editorKey);
	return(true);
}

void ListBoxItemListsReset(DIALOG_ITEM *item,uint32_t numElements,const char **list,bool *selectedElements)
// Place a new list into item.
{
	ListBoxListsReset((LIST_BOX *)item,numElements,list,selectedElements);
}

static void ListBoxDispose(LIST_BOX *listBox)
// Unlink listBox from its parent window, and delete it.
// Invalidate the area of the window where the list box was.
{
	ListBoxInvalidate(listBox);							// invalidate where it is now

	ScrollBarDispose(listBox->scrollBar);
	PlatformFontFree(listBox->font);					// free the font

	MDisposePtr(listBox);
}

static void ListBoxItemDispose(DIALOG_ITEM *item)
// Dispose of the local list box structure pointed to by item.
{
	ListBoxDispose((LIST_BOX *)item);
}

static LIST_BOX *ListBoxCreate(DIALOG *dialog,LIST_BOX_DESCRIPTOR *descriptor)
// Create a list box in dialog, invalidate the area of dialog
// where the list box is to be placed.
// Return a pointer to it if successful.
// ErrorSet, return NULL if not.
{
	LIST_BOX
		*listBox;
	SCROLL_BAR_DESCRIPTOR
		scrollDescriptor;
	bool
		done;
	uint32_t
		index;

	if((listBox=(LIST_BOX *)MNewPtr(sizeof(LIST_BOX))))	// create list box data structure
	{
		listBox->dialogItem.dialog=dialog;							// point back at the dialog
		listBox->dialogItem.disposeFunction=ListBoxItemDispose;
		listBox->dialogItem.drawFunction=ListBoxItemDraw;
		listBox->dialogItem.resizeFunction=ListBoxItemResize;
		listBox->dialogItem.trackFunction=ListBoxItemTrack;
		listBox->dialogItem.earlyKeyFunction=NULL;
		listBox->dialogItem.wantFocus=true;
		listBox->dialogItem.focusChangeFunction=ListBoxItemFocusChange;
		listBox->dialogItem.focusKeyFunction=ListBoxItemFocusKey;
		listBox->dialogItem.focusPeriodicFunction=NULL;

		listBox->positionFunction=descriptor->positionFunction;

		listBox->activeForegroundColor=descriptor->activeForegroundColor;
		listBox->activeBackgroundColor=descriptor->activeBackgroundColor;
		listBox->activeBorderColor=descriptor->activeBorderColor;
		listBox->inactiveForegroundColor=descriptor->inactiveForegroundColor;
		listBox->inactiveBackgroundColor=descriptor->inactiveBackgroundColor;
		listBox->inactiveBorderColor=descriptor->inactiveBorderColor;

		listBox->verticalScrollBarOnLeft=verticalScrollBarOnLeft;	// copy global preference
		listBox->numElements=descriptor->numElements;
		listBox->listElements=descriptor->listElements;
		listBox->selectedElements=descriptor->selectedElements;

		TypedCharactersReset(listBox);								// reset keyboard input delays

		listBox->pressedFunction=descriptor->pressedFunction;

		if((listBox->font=PlatformFontLoad(descriptor->fontName)))
		{
			listBox->topLine=listBox->currentElement=0;

			done=false;
			for(index=0;!done&&index<listBox->numElements;index++)		// point to the first selected element at the start
			{
				if(listBox->selectedElements[index])
				{
					listBox->currentElement=index;
					done=true;
				}
			}

			scrollDescriptor.x=0;									// place scroll bar at 0 initially, resize later
			scrollDescriptor.y=0;
			scrollDescriptor.length=0;
			scrollDescriptor.horizontal=false;
			scrollDescriptor.thumbMax=listBox->numElements;
			scrollDescriptor.thumbPosition=listBox->topLine;
			scrollDescriptor.stepLowerProc=ListBoxScrollLower;
			scrollDescriptor.stepHigherProc=ListBoxScrollHigher;
			scrollDescriptor.pageLowerProc=ListBoxPageLower;
			scrollDescriptor.pageHigherProc=ListBoxPageHigher;
			scrollDescriptor.thumbProc=ListBoxThumbProc;
			scrollDescriptor.procParameters=(void *)listBox;

			if((listBox->scrollBar=ScrollBarCreate(DialogWindowGet(dialog),&scrollDescriptor)))
			{
				ListBoxRectangleSet(listBox);
				ListBoxCenterToCurrentElement(listBox);
				return(listBox);
			}
			PlatformFontFree(listBox->font);				// free the font
		}
		MDisposePtr(listBox);
	}
	return(NULL);
}

DIALOG_ITEM *ListBoxItemCreate(DIALOG *dialog,void *descriptor)
// Create the list box, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)ListBoxCreate(dialog,(LIST_BOX_DESCRIPTOR *)descriptor));
}
