// This takes care of all memory requests, and frees
// It indirects all editor requests for memory
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static uint64_t
	numAllocated=0;

void MMoveMem(const void *source,void *dest,uint64_t numBytes)
// Move memory from source to dest for numBytes.
// This must be smart enough to prevent overlapping areas from causing trouble.
{
	memmove(dest,source,(size_t)numBytes);
}

/* Comment in this code if memory leaks are discovered -- it will help track them down
typedef struct mallocDebug
{
	void
		*returnAddress;
	struct mallocDebug
		*next,
		*prev;
} MALLOC_DEBUG;

static MALLOC_DEBUG
	*head=NULL,
	*tail=NULL;

static void PointerListUnlink(MALLOC_DEBUG *header)
// Remove header from the linked list.
{
	if(header->prev)
	{
		header->prev->next=header->next;
	}
	else
	{
		head=header->next;
	}
	if(header->next)
	{
		header->next->prev=header->prev;
	}
	else
	{
		tail=header->prev;
	}
}

static void PointerListLinkToEnd(MALLOC_DEBUG *header)
// Add header to the linked list.
{
	header->next=NULL;
	if((header->prev=tail))
	{
		header->prev->next=header;			// point previous last item to this one
	}
	else									// headerList was empty, so this is first item too
	{
		head=header;
	}
	tail=header;
}

uint64_t MGetNumAllocatedPointers()
// Return the current number of allocated pointers.
{
	MALLOC_DEBUG
		*debug;

	debug=head;
	while(debug)
	{
		fprintf(stderr,"%p\n",debug->returnAddress);
		debug=debug->next;
	}
	if(head)
	{
		abort();
	}
	return(numAllocated);
}

void MDisposePtr(void *pointer)
// Dispose of memory allocated by MNewPtr.
// This is a debug version of the function frees the debug header
// as well.
{
	MALLOC_DEBUG
		*debug;

	if(pointer)
	{
		debug=(MALLOC_DEBUG *)(((uint8_t *)pointer)-sizeof(MALLOC_DEBUG));
		PointerListUnlink(debug);
		numAllocated--;
		free(debug);
	}
}

void *MNewPtr(uint64_t size)
// Call this to get a contiguous, aligned pointer to memory.
// If NULL is returned, the request could not be granted.
// If NULL is returned, GetError may be called to find out what went wrong.
// This is a debug version of the function which captures the return
// address of the caller -- useful when there's a leak.
{
	void
		*pointer;
	MALLOC_DEBUG
		*debug;

	pointer=NULL;
	if((debug=(MALLOC_DEBUG *)malloc(sizeof(MALLOC_DEBUG)+size)))
	{
		debug->returnAddress=__builtin_return_address(0);
		PointerListLinkToEnd(debug);
		pointer=&debug[1];
		numAllocated++;
	}
	else
	{
		ErrorSet("Failed to malloc %d bytes",size);		// could not allocate memory, so tell user
	}
	return(pointer);
}

void *MNewPtrClr(uint64_t size)
// Call this to get a contiguous, aligned pointer to cleared memory.
// If NULL is returned, the request could not be granted.
// If NULL is returned, GetError may be called to find out what went wrong.
// This is a debug version of the function which captures the return
// address of the caller -- useful when there's a leak.
{
	void
		*pointer;

	if((pointer=MNewPtr(size)))
	{
 		memset(pointer,0,size);
	}
	return(pointer);
}

static size_t MPointerSize(void *pointer)
// Given a pointer allocated by MNewPtr, return the length.
{
	void
		*actualPointer;

	actualPointer=((uint8_t *)pointer)-sizeof(MALLOC_DEBUG);
	return(malloc_usable_size(actualPointer)-sizeof(MALLOC_DEBUG));
}

void *MResizePtr(void *pointer,uint64_t size)
// Resize pointer to size, and return the resized pointer.
// If there is a problem, do not modify pointer, ErrorSet and return NULL.
{
	uint64_t
		usableSize;
	void
		*newPointer;

	usableSize=MPointerSize(pointer);
	if(size>usableSize)	// needs to get larger?
	{
		if((newPointer=(void *)MNewPtr(size)))
		{
			memcpy(newPointer,pointer,usableSize);
			MDisposePtr(pointer);
			return(newPointer);
		}
		else
		{
			ErrorSet("Failed to realloc %d bytes",size);	// could not allocate memory, so tell user
		}
		return(NULL);
	}
	return(pointer);
}
*/

uint64_t MGetNumAllocatedPointers()
// Return the current number of allocated pointers.
{
	return(numAllocated);
}

void MDisposePtr(void *pointer)
// Dispose of memory allocated by MNewPtr.
{
	if(pointer)
	{
		numAllocated--;
		free(pointer);
	}
}

void *MNewPtr(uint64_t size)
// Call this to get a contiguous, aligned pointer to memory.
// If NULL is returned, the request could not be granted.
// If NULL is returned, GetError may be called to find out what went wrong.
{
	void
		*pointer;

	if((pointer=(void *)malloc((size_t)size)))
	{
		numAllocated++;
	}
	else
	{
		ErrorSet("Failed to malloc %d bytes",size);		// could not allocate memory, so tell user
	}
	return(pointer);
}

void *MNewPtrClr(uint64_t size)
// Call this to get a contiguous, aligned pointer to cleared memory.
// If NULL is returned, the request could not be granted.
// If NULL is returned, GetError may be called to find out what went wrong.
{
	void
		*pointer;

	if((pointer=(void *)calloc((size_t)size,1)))
	{
		numAllocated++;
	}
	else
	{
		ErrorSet("Failed to calloc %d bytes",size);		// could not allocate memory, so tell user
	}
	return(pointer);
}

void *MResizePtr(void *pointer,uint64_t size)
// Resize pointer to size, and return the resized pointer.
// If there is a problem, do not modify pointer, ErrorSet and return NULL.
{
	void
		*newPointer;

	if((newPointer=(void *)realloc(pointer,(size_t)size)))
	{
		return(newPointer);
	}
	else
	{
		ErrorSet("Failed to realloc %d bytes",size);	// could not allocate memory, so tell user
	}
	return(NULL);
}
