// Menu handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	DEFAULT_MENU_FONT				"dejavu sans:size=12"	// default font to use for menus

#define MENU_BORDER_SIZE				2		// number of pixels of border around menus (not items, but whole menus)

#define	MENU_MIN_ITEM_COMMAND_SPACING	7		// minimum number of pixels to leave between menu item, and right justified command string
#define	MENU_ADDITIONAL_WIDTH			6		// additional pixel width added to each menu item to make it look better when drawn
#define	MENU_ADDITIONAL_HEIGHT			6		// additional pixel height added to each menu item to make it look better when drawn

// necessary prototypes
static void MenuSubWindowDispose(EDITOR_MENU *menu);
static EDITOR_WINDOW *MenuSubWindowCreate(EDITOR_MENU *owner,EDITOR_RECT *rect,const char *title);

// This definition sets EDITOR_MENU to the structure needed by the gui.
struct editorMenu								// actual structure of an editor menu
{
	EDITOR_WINDOW
		*window;								// if this menu currently has a window, it is pointed to by this (otherwise, its NULL)
	char
		*itemText;								// name of the item is stored here
	char
		*attributes;							// attribute string is stored here
	char
		*dataText;								// the text of the menu (used when menu is clicked on)
	bool
		hasKey;									// true if this item has a key code associated with it
	uint32_t
		modifiers,								// modifiers we are looking for (if any)
		keyCode;								// keyCode that we are looking for (if any)
	EDITOR_FONT
		*font;									// the font that is used for this menu item
	bool
		active;									// tells if the menu is active (can be selected) or not
	bool
		separator;								// tells if the menu item is a separator
	bool
		highlight;								// when a menu button is drawn, this tells if it should draw in its highlighted (pressed) state
	EDITOR_MENU									// linked list of menus
		*parent,
		*nextSibling,
		*previousSibling,
		*firstChild;
};

static EDITOR_MENU
	*firstMenu;

static char
	menuFontName[MAX_FONT_NAME_LENGTH];			// keeps the font name that's used for all menus which don't specify it otherwise
static EDITOR_COLOR
	menuBackgroundColor,						// background color for all menus that don't have one assigned
	menuForegroundColor,						// foreground color for all menus that don't have one assigned
	menuKeyColor,								// color to draw assigned keys in (shadow currently always drawn in black)
	menuShadowTopColor,							// top/left color for shadow outlines
	menuShadowBottomColor;						// bottom/right color for shadow outlines

static bool
	menuOpen;									// tells if we've got a menu open at the moment

enum
{
	BAD_RELATIONSHIP,
	BAD_PATH
};

static const char *errorDescriptions[]=
{
	"Invalid menu relationship",
	"Invalid menu path"
};

static EDITOR_MENU *MenuRelatedGet(EDITOR_MENU *menu,int menuRelationship)
// Return the menu with menuRelationship to menu.
// Return NULL if there is no menu with menuRelationship.
{
	EDITOR_MENU
		*outputMenu;

	outputMenu=NULL;
	switch(menuRelationship)
	{
		case MR_NEXT_SIBLING:
			if(menu->parent)
			{
				if(menu->nextSibling!=menu->parent->firstChild)
				{
					outputMenu=menu->nextSibling;
				}
			}
			else
			{
				if(menu->nextSibling!=firstMenu)
				{
					outputMenu=menu->nextSibling;
				}
			}
			break;
		case MR_PREVIOUS_SIBLING:
			if(menu->parent)
			{
				if(menu!=menu->parent->firstChild)
				{
					outputMenu=menu->previousSibling;
				}
			}
			else
			{
				if(menu!=firstMenu)
				{
					if(menu->previousSibling!=firstMenu)
					{
						outputMenu=menu->previousSibling;
					}
				}
				else
				{
					outputMenu=firstMenu;
				}
			}
			break;
		case MR_FIRST_CHILD:
			if(menu)
			{
				outputMenu=menu->firstChild;
			}
			else
			{
				outputMenu=firstMenu;
			}
			break;
		case MR_LAST_CHILD:
			if(menu)
			{
				if(menu->firstChild)
				{
					outputMenu=menu->firstChild->previousSibling;
				}
			}
			else
			{
				if(firstMenu)
				{
					outputMenu=firstMenu->previousSibling;
				}
			}
			break;
		default:
			break;
	}
	return(outputMenu);
}

static EDITOR_WINDOW *MenuWindowGet(EDITOR_MENU *menu)
// Return the window that belongs to menu if it has one, or NULL
// if it does not.
{
	if(menu)						// see if menu, or root
	{
		return(menu->window);
	}
	return(rootMenuWindow);
}

static void MenuItemSizeCalculate(EDITOR_MENU *menu,uint32_t *width,uint32_t *height)
// Get the ideal width and height for menu.
{
	uint32_t
		ascent,
		descent;
	char
		commandString[256];
	uint32_t
		stringLength;

	if(menu->separator)
	{
		(*height)=2;												// use exact height for separators
		(*width)=MENU_ADDITIONAL_WIDTH;
	}
	else
	{
		FontHeightsGet(menu->font,&ascent,&descent);

		(*height)=ascent+descent+MENU_ADDITIONAL_HEIGHT;			// get the height of the menu item (use total height of the font, plus some slop to make it look better)
		(*width)=FontTextUtf8Width(menu->font,menu->itemText,strlen(menu->itemText))+MENU_ADDITIONAL_WIDTH;

		if(menu->firstChild)										// if this menu has children, we need to draw the hierarchical right arrow, otherwise, we can use the command string
		{
			(*width)+=MENU_MIN_ITEM_COMMAND_SPACING;				// add spacing between item and command string
			(*width)+=subMenuImage->width;
			if((*height)<(uint32_t)subMenuImage->height)			// if the font is tiny, then adjust height
			{
				(*height)=subMenuImage->height;
			}
		}
		else
		{
			if(menu->hasKey)
			{
				(*width)+=MENU_MIN_ITEM_COMMAND_SPACING;			// add spacing between item and command string
				stringLength=KeyBindKeyToName(PlatformRemapEditorModifiers(menu->modifiers),menu->keyCode,commandString,sizeof(commandString));
				(*width)+=FontTextUtf8Width(menu->font,commandString,stringLength);
			}
		}
	}
}

static void MenuRectPin(EDITOR_RECT *rect)
// Make sure rect does not go off the display (as best as possible).
{
	uint32_t
		displayWidth,
		displayHeight;

	EditorDisplayDimensionsGet(&displayWidth,&displayHeight);
	if(rect->x+rect->w>displayWidth)
	{
		rect->x=displayWidth-rect->w;
	}
	if(rect->y+rect->h>displayHeight)
	{
		rect->y=displayHeight-rect->h;
	}
	if(rect->x<0)
	{
		rect->x=0;
	}
	if(rect->y<0)
	{
		rect->y=0;
	}
}

static void MenuItemRectCalculate(EDITOR_MENU *menu,EDITOR_RECT *rect)
// Get the rect of item as it will be drawn into its parent menu.
{
	EDITOR_MENU
		*currentMenu,
		*parentMenu;
	uint32_t
		tempWidth,
		tempHeight;
	int32_t
		currentY;

	rect->x=rect->y=MENU_BORDER_SIZE;									// offset by border
	rect->w=rect->h=0;
	currentY=0;
	parentMenu=menu->parent;
	currentMenu=MenuRelatedGet(parentMenu,MR_FIRST_CHILD);
	while(currentMenu)
	{
		MenuItemSizeCalculate(currentMenu,&tempWidth,&tempHeight);
		if(tempWidth>rect->w)
		{
			rect->w=tempWidth;											// find the maximum width of all the items
		}
		if(currentMenu==menu)											// when we get to our menu remember its Y position
		{
			rect->y+=currentY;
			rect->h=tempHeight;
		}
		currentY+=tempHeight;
		currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
	}
}

static void MenuWindowSizeCalculate(EDITOR_MENU *menu,uint32_t *width,uint32_t *height)
// Run though the list of children of menu, determining the size of each, use these
// sizes to get the composite size for menu.
{
	EDITOR_MENU
		*currentMenu;
	uint32_t
		tempWidth,
		tempHeight;

	(*width)=0;
	(*height)=0;
	currentMenu=MenuRelatedGet(menu,MR_FIRST_CHILD);
	while(currentMenu)
	{
		MenuItemSizeCalculate(currentMenu,&tempWidth,&tempHeight);
		if(tempWidth>(*width))
		{
			(*width)=tempWidth;											// find the maximum width of all the items
		}
		(*height)+=tempHeight;											// and the total height of all items
		currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
	}
	*width+=2*MENU_BORDER_SIZE;											// fancy border drawn around entire menu
	*height+=2*MENU_BORDER_SIZE;
}

static void MenuItemDraw(GUI_IMAGE *imageBuffer,EDITOR_MENU *menu)
// Use this to draw menu's item into the imageBuffer.
{
	EDITOR_RECT
		itemRect;
	uint32_t
		ascent,
		descent;
	uint32_t
		stringWidth;
	int
		itemX,
		itemY;
	int32_t
		xLeft,
		xRight,
		yTop,
		yBottom;
	char
		commandString[256];
	uint32_t
		stringLength;

	MenuItemRectCalculate(menu,&itemRect);								// get the rectangle for this item
	if(itemRect.w&&itemRect.h)
	{
		ImageFill(imageBuffer,menuBackgroundColor,itemRect.x,itemRect.y,itemRect.w,itemRect.h);

		if(menu->separator)
		{
			xLeft=itemRect.x;
			xRight=itemRect.x+itemRect.w-1;
			yTop=itemRect.y;
			yBottom=yTop+1;

			ImageFill(imageBuffer,BLACK,xLeft,yTop,xRight-xLeft,1);		// draw separator lines
			ImageFill(imageBuffer,WHITE,xLeft,yBottom,xRight-xLeft,1);
		}
		else
		{
			FontHeightsGet(menu->font,&ascent,&descent);

			stringLength=(uint32_t)strlen(menu->itemText);
			stringWidth=FontTextUtf8Width(menu->font,menu->itemText,stringLength);

			itemX=itemRect.x+(MENU_ADDITIONAL_WIDTH/2);
			itemY=itemRect.y+(itemRect.h/2)-(ascent+descent)/2+ascent;

			FontTextUtf8Draw(menu->font,imageBuffer,menuForegroundColor,itemX,itemY,menu->itemText,stringLength);

			if(menu->firstChild)
			{
				itemX=itemRect.x+(itemRect.w-subMenuImage->width-(MENU_ADDITIONAL_WIDTH/2));
				itemY=itemRect.y+(itemRect.h-subMenuImage->height)/2;
				ImageDraw(subMenuImage,imageBuffer,itemX,itemY);
			}
			else
			{
				if(menu->hasKey)
				{
					stringLength=KeyBindKeyToName(PlatformRemapEditorModifiers(menu->modifiers),menu->keyCode,commandString,sizeof(commandString));
					stringWidth=FontTextUtf8Width(menu->font,commandString,stringLength);

					itemX=itemRect.x+(itemRect.w-stringWidth-(MENU_ADDITIONAL_WIDTH/2));
					itemY=itemRect.y+(itemRect.h/2)-(ascent+descent)/2+ascent;

					FontTextUtf8Draw(menu->font,imageBuffer,GRAY_0,itemX+1,itemY+1,commandString,stringLength);
					FontTextUtf8Draw(menu->font,imageBuffer,menuKeyColor,itemX,itemY,commandString,stringLength);
				}
			}
			if(menu->highlight)
			{
				ShadowRectangleOutline(imageBuffer,&itemRect,menuShadowTopColor,menuShadowBottomColor,2);
			}
		}
	}
}

static void MenuWindowDraw(EDITOR_WINDOW *window,GUI_IMAGE *imageBuffer)
// Draw all of the buttons into window
// also draw the menu border.
{
	EDITOR_MENU
		*owner,							// the edit menu that owns this window
		*currentMenu;
	EDITOR_RECT
		windowRect;

	owner=(EDITOR_MENU *)window->windowInfo;
	currentMenu=MenuRelatedGet(owner,MR_FIRST_CHILD);
	while(currentMenu)
	{
		MenuItemDraw(imageBuffer,currentMenu);								// draw the items into the menu
		currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
	}

	windowRect.x=0;
	windowRect.y=0;
	windowRect.w=window->w;
	windowRect.h=window->h;
	ShadowRectangleOutline(imageBuffer,&windowRect,menuShadowTopColor,menuShadowBottomColor,MENU_BORDER_SIZE);
}

static void MenuWindowUpdate(EDITOR_WINDOW *window)
// Make sure window is up to date.
{
	Window
		xWindow;
	GUI_IMAGE
		*imageBuffer;

	if(!XEmptyRegion(window->invalidRegion))						// see if it has an invalid area
	{
		xWindow=window->xWindow;									// get x window for this window
		if((imageBuffer=ImageCreate(window->w,window->h)))	// create an image to be used to composite the window off-screen
		{
			MenuWindowDraw(window,imageBuffer);
			ImageDrawToWindow(imageBuffer,xWindow,window->invalidRegion,window->w,window->h);
			ImageDestroy(imageBuffer);
		}
		XDestroyRegion(window->invalidRegion);						// clear out the invalid region now that the window has been drawn
		window->invalidRegion=XCreateRegion();
	}
}

static void MenuWindowResize(EDITOR_MENU *menu)
// If menu has a window, call this to size and invalidate the window when a child of menu
// has changed in some way, or when the menu is created.
// If NULL is passed for menu, it indicates the root menu window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	XSizeHints
		sizeHints;

	rect.x=rect.y=0;

	if((window=MenuWindowGet(menu)))
	{
		MenuWindowSizeCalculate(menu,&rect.w,&rect.h);					// get the ideal dimensions of this window (based on its children)
		if(rect.w==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect.w=1;
		}
		if(rect.h==0)
		{
			rect.h=1;
		}

		sizeHints.flags=PMinSize|PMaxSize;
		sizeHints.min_width=rect.w;
		sizeHints.min_height=rect.h;
		sizeHints.max_width=rect.w;
		sizeHints.max_height=rect.h;
		XSetWMNormalHints(xDisplay,window->xWindow,&sizeHints);

		XResizeWindow(xDisplay,window->xWindow,rect.w,rect.h);			// set the window to the new size, now that child has changed

		WindowRectInvalidate(window,&rect);								// invalidate the entire contents of the window so it will be redrawn
	}
}

static EDITOR_MENU *MenuLocateWithKey(EDITOR_MENU *topLevel,uint32_t modifiers,uint32_t keyCode)
// Attempt to locate an active menu below topLevel, that has modifiers/keyCode defined.
// If one can be located, return it, else return NULL.
// ### this is probably too slow, since it is called with every key press.
{
	EDITOR_MENU
		*currentMenu,
		*locatedMenu;

	currentMenu=MenuRelatedGet(topLevel,MR_FIRST_CHILD);
	locatedMenu=NULL;
	while(currentMenu&&!locatedMenu)
	{
		if(!(locatedMenu=MenuLocateWithKey(currentMenu,modifiers,keyCode)))		// recurse to try to find child that has this key
		{
			if(currentMenu->hasKey&&currentMenu->active)
			{
				if((PlatformRemapEditorModifiers(currentMenu->modifiers)==modifiers)&&(currentMenu->keyCode==keyCode))	// see if this item has the key
				{
					locatedMenu=currentMenu;
				}
			}
		}
		currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
	}
	return(locatedMenu);
}

EDITOR_MENU *MenuKeyEventLocate(uint32_t modifiers,uint32_t keyCode)
// See if a menu can be located that corresponds to the given editorKey.
{
	EDITOR_MENU
		*menu;

	if((menu=MenuLocateWithKey(NULL,modifiers,keyCode)))
	{
		if(!menu->firstChild)										// must be leaf on the tree
		{
			return(menu);
		}
	}
	return(NULL);
}

static void MenuFontUnAssign(EDITOR_MENU *menu)
// Get rid of font assigned to a menu item.
{
	if(menu->font)
	{
		PlatformFontFree(menu->font);
		menu->font=NULL;				// use the default font
	}
}

static void MenuFontAssign(EDITOR_MENU *menu,const char *fontName)
// Given a font name, apply it to menu.
// If the menu already has a font assigned to it, get rid of it.
{
	if(menu->font)
	{
		PlatformFontFree(menu->font);
	}
	menu->font=PlatformFontLoad(fontName);
}

static void MenuKeyEquivalentUnAssign(EDITOR_MENU *menu)
// Remove key code assigned to menu.
{
	menu->hasKey=false;
}

static void MenuKeyEquivalentAssign(EDITOR_MENU *menu,const char *keyName)
// Given a key name, assign that key to menu.
// If the another menu already has the key assigned to it, get rid of it.
// If keyName does not map to a valid key, then the menu will have no key equivalent.
{
	uint32_t
		modifiers,
		keyCode;
	EDITOR_MENU
		*oldMenu;

	MenuKeyEquivalentUnAssign(menu);
	if(KeyBindNameToKey(keyName,&modifiers,&keyCode))
	{
		if((oldMenu=MenuLocateWithKey(NULL,modifiers,keyCode)))			// see if another menu item had this key, if so, take it
		{
			MenuKeyEquivalentUnAssign(oldMenu);
			MenuWindowResize(oldMenu->parent);							// update the old menu
		}
		menu->modifiers=modifiers;
		menu->keyCode=keyCode;
		menu->hasKey=true;
	}
}

static void MenuAttributeParameterExtract(const char *parameterInput,uint32_t *parameterIndex,char *parameterBuffer,uint32_t parameterBufferSize)
// Extract the next parameter from parameterInput[parameterIndex] into parameter buffer.
// If the parameter would be too long, just truncate it.
// Leave paramterIndex updated to point to the end of the given parameter, no matter
// how long it is.
{
	uint32_t
		outIndex;
	char
		character;
	bool
		done;

	outIndex=0;
	done=false;
	while((character=parameterInput[*parameterIndex])&&!done)	// make sure not end of string, or end of parameter
	{
		if(character!='\\')
		{
			(*parameterIndex)++;
			if(outIndex<parameterBufferSize)
			{
				parameterBuffer[outIndex++]=character;
			}
		}
		else
		{
			if(parameterInput[*parameterIndex+1]=='\\')			// see if quoted separator (not a separator at all then)
			{
				(*parameterIndex)+=2;
				if(outIndex<parameterBufferSize)
				{
					parameterBuffer[outIndex++]='\\';			// place separator into buffer
				}
			}
			else
			{
				done=true;										// ran into end of parameter list
			}
		}
	}
	if(outIndex<parameterBufferSize)
	{
		parameterBuffer[outIndex]='\0';							// terminate buffer
	}
	else
	{
		if(parameterBufferSize)
		{
			parameterBuffer[outIndex-1]='\0';					// terminate buffer by writing over last character
		}
	}
}

static void MenuAttributesUnRealize(EDITOR_MENU *menu)
// Undo anything that MenuAttributesRealize did.
{
	MenuFontUnAssign(menu);
	MenuKeyEquivalentUnAssign(menu);
}

static bool MenuAttributesRealize(EDITOR_MENU *menu)
// Attempt to fill the parts of menu's record that are described in its string.
// Currently supports:
// \Ffontname		specify the font to use when drawing the item
// \S				specify that this item is a menu separation line (text is not drawn)
// \Kkey			specify that mod1 <key> could be used to access the given menu
// If there is a problem, ErrorSet, return false.
{
	char
		parameterBuffer[256];						// max size for parameter string, anything larger gets truncated!
	uint32_t
		attributeIndex;
	char
		character;

	menu->separator=false;
	menu->hasKey=false;
	menu->modifiers=0;
	menu->keyCode=0;
	menu->font=NULL;								// no font yet

	attributeIndex=0;
	while((character=menu->attributes[attributeIndex]))
	{
		if(character!='\\')							// eat separators
		{
			MenuAttributeParameterExtract(menu->attributes,&attributeIndex,&(parameterBuffer[0]),256);
			switch(parameterBuffer[0])				// check command part
			{
				case 'F':
					MenuFontAssign(menu,&(parameterBuffer[1]));
					break;
				case 'S':
					menu->separator=true;
					break;
				case 'K':
					MenuKeyEquivalentAssign(menu,&(parameterBuffer[1]));
					break;
				default:
					break;
			}
		}
		else
		{
			attributeIndex++;
		}
	}

	if(!menu->font)									// if no font specified, use the default
	{
		MenuFontAssign(menu,menuFontName);
	}

	if(!menu->font)
	{
		ErrorSet("Failed to load menu font");
		return(false);
	}
	return(true);
}

static void MenuEnableStateSet(EDITOR_MENU *menu)
// Set the active state for menu.
{
}

static EDITOR_MENU *MenuItemLocate(EDITOR_MENU *parent,int32_t x,int32_t y)
// Locate the item (if any) that X and Y are over in parent.
// X and Y are relative to parent.
{
	EDITOR_MENU
		*currentMenu;
	EDITOR_RECT
		rect;
	bool
		found;

	currentMenu=MenuRelatedGet(parent,MR_FIRST_CHILD);
	found=false;
	while(currentMenu&&!found)
	{
		MenuItemRectCalculate(currentMenu,&rect);
		if(currentMenu->active&&!currentMenu->separator&&PointInRect(x,y,&rect))
		{
			found=true;
		}
		else
		{
			currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
		}
	}
	return(currentMenu);
}

static bool MenuTrackLocate(EDITOR_MENU *topLevel,int32_t rootX,int32_t rootY,EDITOR_MENU **menu)
// Given an X, and Y that are relative to the X root, run through the open
// menu window list, and return the menu item that the X and Y are over.
// If X,Y are not over any item, return menu=NULL.
// If false is returned, the given point did not fall in topLevel, or any of its sub windows.
{
	EDITOR_MENU
		*currentMenu;
	int32_t
		localX,
		localY;
	bool
		foundHigher;
	EDITOR_WINDOW
		*window;

	(*menu)=NULL;
	foundHigher=false;
	currentMenu=MenuRelatedGet(topLevel,MR_FIRST_CHILD);
	while(currentMenu&&!foundHigher)
	{
		if(currentMenu->window)
		{
			foundHigher=MenuTrackLocate(currentMenu,rootX,rootY,menu);
		}
		currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
	}
	if(!foundHigher)
	{
		if((window=MenuWindowGet(topLevel)))
		{
			GlobalToLocal(window->xWindow,rootX,rootY,&localX,&localY);	// convert global coordinates to those local to this window
			if(localX>=0&&localY>=0&&localX<(int32_t)window->w&&localY<(int32_t)window->h)
			{
				(*menu)=MenuItemLocate(topLevel,localX,localY);	// see if we can find an item here (either way, we are done because we saw the point in our window)
				foundHigher=true;
			}
		}
	}
	return(foundHigher);
}

static void MenuRectInvalidate(EDITOR_MENU *menu,EDITOR_RECT *rect)
// Invalidate the given rectangle of menu's window if menu has a window.
// NULL can be passed as menu, and indicates the root menu.
{
	EDITOR_WINDOW
		*window;

	if((window=MenuWindowGet(menu)))
	{
		WindowRectInvalidate(window,rect);		// invalidate the rectangle
	}
}

static void MenuItemInvalidate(EDITOR_MENU *menu)
// If menu is within a parent window, invalidate the rectangle of
// the parent that corresponds to menu.
// It is incorrect to call this with menu==NULL.
{
	EDITOR_RECT
		rect;

	MenuItemRectCalculate(menu,&rect);		// find the rectangle of this item within its parent
	MenuRectInvalidate(menu->parent,&rect);	// invalidate this rectangle of the parent's window (if he has one)
}

static void MenuItemUnpress(EDITOR_MENU *menu)
// When tracking has left a menu item, call this to unhighlight it.
// If an item has submenus open, it is not unpressed.
{
	if((!menu->window)&&menu->highlight)
	{
		menu->highlight=false;							// unhighlight this menu
		MenuItemInvalidate(menu);
	}
}

static void MenuSubsClose(EDITOR_MENU *menu)
// Close all of the windows for all inferiors of menu.
// The close is done depth first.
{
	EDITOR_MENU
		*currentMenu;

	currentMenu=MenuRelatedGet(menu,MR_FIRST_CHILD);
	while(currentMenu)
	{
		if(currentMenu->window)
		{
			MenuSubsClose(currentMenu);					// close down any this menu has first
			MenuSubWindowDispose(currentMenu);
		}
		if(currentMenu->highlight)
		{
			currentMenu->highlight=false;				// unhighlight this menu
			MenuItemInvalidate(currentMenu);
		}
		currentMenu=MenuRelatedGet(currentMenu,MR_NEXT_SIBLING);
	}
}

static void MenuSubOpen(EDITOR_MENU *menu)
// If menu is hierarchical, call this to open its submenu.
// An attempt is made to open the submenu to the right of menu.
// If it does not fit on the screen, some other method should be implemented.
{
	EDITOR_RECT
		rect;
	EDITOR_WINDOW
		*window;

	menu->highlight=true;								// highlight this menu
	MenuItemInvalidate(menu);
	if((window=MenuWindowGet(menu->parent))&&menu->firstChild)	// find parent's window, make sure this one has children
	{
		MenuItemRectCalculate(menu,&rect);				// get my rectangle within it
		LocalToGlobal(window->xWindow,rect.x+rect.w,rect.y,&rect.x,&rect.y);	// get global location for new window
		rect.y-=MENU_BORDER_SIZE;						// step the sub-menu up by the border size so it lines up with its parent
		rect.w=rect.h=0;								// initialize to zero
		MenuSubWindowCreate(menu,&rect,"submenu");
	}
}

static void MenuItemPress(EDITOR_MENU *menu)
// When a menu has been located, do what is needed to highlight it, and also
// if it has a submenu, open it (closing any submenus of any other menus at this level).
{
	if(!menu->window)									// if this one does not have a window, close any other windows at this level
	{
		MenuSubsClose(menu->parent);					// close any open submenus at this level
		MenuSubOpen(menu);								// open, or at least highlight this item
	}
	else
	{
		if(!menu->highlight)							// if it has a window, it may still be unhighlighted, so highlight it
		{
			menu->highlight=true;						// highlight this menu
			MenuItemInvalidate(menu);
		}
	}
}

static EDITOR_MENU *MenuWindowButtonPress(EDITOR_WINDOW *window,XEvent *event)
// Handle button press events for menu windows (see what the user is clicking
// on, and return it).
// window is passed in as the root menu window.
// NOTE: this does not close submenus when it exits, and it leaves the selected
// menu pressed.
{
	EDITOR_MENU
		*menu,
		*lastMenu;
	int
		rootX,
		rootY,
		windowX,
		windowY,
		lastX,
		lastY;
	int
		button;

	button=event->xbutton.button;
	MenuTrackLocate(NULL,event->xbutton.x_root,event->xbutton.y_root,&menu);	// find the menu item (if any) that was clicked on
	lastMenu=menu;
	if(menu)
	{
		MenuItemPress(menu);
	}
	rootX=event->xbutton.x_root;
	rootY=event->xbutton.y_root;
	windowX=event->xbutton.x;
	windowY=event->xbutton.y;
	lastX=rootX;
	lastY=rootY;

	while(StillDown(event->xbutton.window,button,&rootX,&rootY,&windowX,&windowY))
	{
		if(lastX!=rootX||lastY!=rootY)					// make sure something has changed
		{
			MenuTrackLocate(NULL,rootX,rootY,&menu);
			if(menu!=lastMenu)
			{
				if(lastMenu)
				{
					MenuItemUnpress(lastMenu);			// unpress the item that we are leaving
				}
				if(menu)
				{
					MenuItemPress(menu);				// if a new item was located, press it
				}
				lastMenu=menu;
			}
			lastX=rootX;
			lastY=rootY;
		}
	}
	return(menu);
}

static void MenusClose(Time time)
// If menus are open, close them all, release the pointer, and set the status to not open.
{
	if(menuOpen)
	{
		XUngrabPointer(xDisplay,time);
		menuOpen=false;
	}
	MenuSubsClose(NULL);										// get rid of sub menus
}

static void MenusOpen(Time time)
// If menus are not already open, then grab the pointer and set the status to open.
{
	if(!menuOpen)
	{
		XGrabPointer(xDisplay,rootMenuWindow->xWindow,False,ButtonPressMask|ButtonReleaseMask,GrabModeAsync,GrabModeAsync,None,None,time);
		menuOpen=true;
	}
}

static void MenuWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// Handle an event arriving for a menu window.
{
	EDITOR_KEY
		editorKey;
	EDITOR_MENU
		*menu;
	KEY_BINDING
		*entry;

	switch(event->type)
	{
		case ButtonPress:
			XRaiseWindow(xDisplay,window->xWindow);				// make sure menu is on top
			if((menu=MenuWindowButtonPress(window,event)))		// track the button press, see if it landed on a menu
			{
				if(menu->firstChild)							// see if we've landed on a menu that's not a leaf
				{
					MenusOpen(event->xbutton.time);
				}
				else	// landed on a leaf
				{
					MenusClose(event->xbutton.time);
					MenuEventHandle(menu);						// call the shell to handle this menu
				}
			}
			else		// click on something not a menu
			{
				MenusClose(event->xbutton.time);
			}
			break;
		case KeyPress:
			if(KeyEventToEditorKey(&event->xkey,&editorKey))			// convert X event to editorKey
			{
				if(editorKey.keyCode==XKeysymToKeycode(xDisplay,XK_Escape))	// user aborting menu selection
				{
					MenusClose(event->xkey.time);
				}
				else if((entry=KeyBindLocate(editorKey.modifiers,editorKey.keyCode)))
				{
					MenusClose(event->xkey.time);
					KeyBindEventHandle(entry);						// go handle it
				}
				else if((menu=MenuKeyEventLocate(editorKey.modifiers,editorKey.keyCode)))		// was this a menu key?
				{
					MenusClose(event->xkey.time);
					MenuEventHandle(menu);							// call the shell to handle this menu
				}
			}
			break;
		case ClientMessage:
			if(event->xclient.data.l[0]==(int)deleteWindowAtom)
			{
				ShellCommandHandle("quit",0,NULL);
			}
			break;
		default:
			break;
	}
}

static void MenuSubWindowDispose(EDITOR_MENU *menu)
// Remove the window associated with menu.
{
	EDITOR_WINDOW
		*window;

	if((window=menu->window))
	{
		XDestroyRegion(window->invalidRegion);					// destroy invalid region
		XDestroyWindow(xDisplay,window->xWindow);				// tell x to make the window go away
		WaitUntilDestroyed(window);								// wait for it to die
		WindowElementUnlink(window);							// unlink it from our list
		MDisposePtr(window);
		menu->window=NULL;										// point to no window
	}
}

static EDITOR_WINDOW *MenuSubWindowCreate(EDITOR_MENU *owner,EDITOR_RECT *rect,const char *title)
// This is called to create a sub menu window for owner.
// Sub menu windows do not have normal window decoration.
// NOTE: this checks to see if the submenu will go off the screen, and pins it
// into range.
{
	EDITOR_WINDOW
		*window;
	XSetWindowAttributes
		attributes;
	uint32_t
		valueMask;
	Atom
		protocols[1];													// number of protocols

	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))		// create editor window
	{
		WindowElementLink(window);										// link it to the global window list that we maintain
		window->windowType=EWT_MENU;									// set the window type to "menu"
		window->windowInfo=owner;										// point to the owner
		window->configureFunction=NULL;									// no configure function (window cannot change size by user action)
		window->updateFunction=MenuWindowUpdate;						// point to the function to call to update this window when it needs to be redrawn
		window->eventFunction=MenuWindowEvent;							// point to the function to call to handle events
		window->periodicFunction=NULL;									// no periodic activity for these windows

		attributes.save_under=True;
		attributes.override_redirect=True;
		valueMask=CWSaveUnder|CWOverrideRedirect;

		MenuWindowSizeCalculate(owner,&rect->w,&rect->h);

		if(rect->w==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect->w=1;
		}
		if(rect->h==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect->h=1;
		}
		MenuRectPin(rect);												// pin the rectangle to the screen

		window->w=rect->w;												// remember initial size for the window
		window->h=rect->h;

		// Due to X's nasty error handling, we assume this create does not fail!!!
		window->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x,rect->y,rect->w,rect->h,0,DefaultDepth(xDisplay,xScreenNum),InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

		XStoreName(xDisplay,window->xWindow,title);
		XSetIconName(xDisplay,window->xWindow,title);
		XSelectInput(xDisplay,window->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|KeyPressMask|StructureNotifyMask);	// want to see these events

		protocols[0]=deleteWindowAtom;
		XSetWMProtocols(xDisplay,window->xWindow,&(protocols[0]),1);	// tell X we would like to participate in these protocols

		window->invalidRegion=XCreateRegion();							// create new empty invalid region

		owner->window=window;											// point this menu at its window
		XDefineCursor(xDisplay,window->xWindow,arrowCursor);			// this is the cursor to use in the submenu
		XMapRaised(xDisplay,window->xWindow);							// map window to the top of the display
		WaitUntilMappedNoFocus(window);									// wait until this window is actually mapped before continuing
		return(window);
	}
	ErrorSet("Failed to MenuSubWindowCreate");
	return((EDITOR_WINDOW *)NULL);
}

static void MenuRootWindowDispose(EDITOR_WINDOW *window)
// Close a given root menu window.
{
	XDestroyRegion(window->invalidRegion);									// destroy invalid region
	FocusAwayFrom(window);
	XDestroyWindow(xDisplay,window->xWindow);								// tell x to make the window go away
	WaitUntilDestroyed(window);												// wait for it to die
	WindowElementUnlink(window);											// unlink it from our list
	MDisposePtr(window);
}

static EDITOR_WINDOW *MenuRootWindowCreate(EDITOR_RECT *rect,const char *title)
// Create a window which will be used as the root of a menu tree.
// If there is a problem, ErrorSet, return NULL.
{
	XSetWindowAttributes
		attributes;
	unsigned int
		valueMask;
	EDITOR_WINDOW
		*window;
	Atom
		protocols[2];													// number of protocols
	XWMHints
		hints;															// hints to tell window manager how to treat these windows
	XSizeHints
		sizeHints;														// window size info (used to REALLY convince X that I know where I want my own window)
	int32_t
		xOffset,
		yOffset;
	uint32_t
		screenWidth,
		screenHeight;

	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))		// create editor window
	{
		WindowElementLink(window);										// link it to the global window list that we maintain
		window->windowType=EWT_MENU;									// set the window type to "menu"
		window->windowInfo=NULL;										// root menu has no owner
		window->configureFunction=NULL;									// no configure function (window cannot change size by user action)
		window->updateFunction=MenuWindowUpdate;						// point to the function to call to update this window when it needs to be redrawn
		window->eventFunction=MenuWindowEvent;							// point to the function to call to handle events
		window->periodicFunction=NULL;									// no periodic activity for these windows

		valueMask=0;

		MenuWindowSizeCalculate(NULL,&rect->w,&rect->h);

		if(rect->w==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect->w=1;
		}
		if(rect->h==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect->h=1;
		}

		window->w=rect->w;												// remember initial size for the window
		window->h=rect->h;

		PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&screenWidth,&screenHeight);

		// Due to X's nasty error handling, we assume this create does not fail!!!
		window->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect->x+xOffset,rect->y+yOffset,rect->w,rect->h,0,DefaultDepth(xDisplay,xScreenNum),InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

		sizeHints.flags=USSize|USPosition|PMinSize|PMaxSize;		// ### LIE to X, tell it the user requested the size and position so it will honor them
		sizeHints.x=rect->x+xOffset;
		sizeHints.y=rect->y+yOffset;
		sizeHints.width=rect->w;
		sizeHints.height=rect->h;
		sizeHints.min_width=rect->w;
		sizeHints.min_height=rect->h;
		sizeHints.max_width=rect->w;
		sizeHints.max_height=rect->h;
		XSetWMNormalHints(xDisplay,window->xWindow,&sizeHints);

		XStoreName(xDisplay,window->xWindow,title);
		XSetIconName(xDisplay,window->xWindow,title);
		XSelectInput(xDisplay,window->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask|PropertyChangeMask);	// want to see these events

		protocols[0]=takeFocusAtom;
		protocols[1]=deleteWindowAtom;
		XSetWMProtocols(xDisplay,window->xWindow,&(protocols[0]),2);	// tell X we would like to participate in these protocols

		hints.flags=InputHint|StateHint|WindowGroupHint;
		hints.input=True;
		hints.initial_state=NormalState;
		hints.window_group=window->xWindow;
		XSetWMHints(xDisplay,window->xWindow,&hints);

		WindowClassHintsSet(window);
		WindowIconSet(window);

		window->invalidRegion=XCreateRegion();							// create new empty invalid region

		XMapRaised(xDisplay,window->xWindow);							// map window to the top of the display

		WaitUntilMapped(window);										// wait until this window is actually mapped before continuing
		return(window);
	}
	ErrorSet("Failed to MenuRootWindowCreate");
	return((EDITOR_WINDOW *)NULL);
}

static void MenuElementUnlink(EDITOR_MENU *menu)
// Unlink menu from the menu tree.
{
	if(menu->parent)
	{
		if(menu->parent->firstChild==menu)
		{
			if(menu->nextSibling!=menu)
			{
				menu->parent->firstChild=menu->nextSibling;
			}
			else
			{
				menu->parent->firstChild=NULL;
			}
		}
	}
	else
	{
		if(firstMenu==menu)
		{
			if(menu->nextSibling!=menu)
			{
				firstMenu=menu->nextSibling;
			}
			else
			{
				firstMenu=NULL;
			}
		}
	}
	menu->nextSibling->previousSibling=menu->previousSibling;
	menu->previousSibling->nextSibling=menu->nextSibling;
}

static bool MenuElementLink(EDITOR_MENU *menu,EDITOR_MENU *baseMenu,int menuRelationship)
// Link menu to baseMenu with menuRelationship.
// If baseMenu is NULL, then menuRelationship must be MR_FIRST_CHILD, or MR_LAST_CHILD as it is not possible to
// link menus as siblings of the root.
// If there is a problem, set the error, and return false.
{
	bool
		fail;

	fail=false;
	menu->firstChild=NULL;
	switch(menuRelationship)
	{
		case MR_NEXT_SIBLING:
			if(baseMenu)
			{
				menu->nextSibling=baseMenu->nextSibling;
				menu->nextSibling->previousSibling=menu;
				menu->previousSibling=baseMenu;
				baseMenu->nextSibling=menu;
				menu->parent=baseMenu->parent;
			}
			else
			{
				ErrorSet(errorDescriptions[BAD_RELATIONSHIP]);
				fail=true;
			}
			break;
		case MR_PREVIOUS_SIBLING:
			if(baseMenu)
			{
				menu->previousSibling=baseMenu->previousSibling;
				menu->previousSibling->nextSibling=menu;
				menu->nextSibling=baseMenu;
				baseMenu->previousSibling=menu;
				menu->parent=baseMenu->parent;
				if(menu->parent)
				{
					if(menu->parent->firstChild==baseMenu)
					{
						menu->parent->firstChild=menu;		// make us the first child if linked before old first child
					}
				}
				else
				{
					if(firstMenu==baseMenu)
					{
						firstMenu=menu;
					}
				}
			}
			else
			{
				ErrorSet(errorDescriptions[BAD_RELATIONSHIP]);
				fail=true;
			}
			break;
		case MR_FIRST_CHILD:
			menu->parent=baseMenu;								// set parent
			if(baseMenu)										// see if linking as a child of a current menu
			{
				if((menu->nextSibling=baseMenu->firstChild))
				{
					menu->previousSibling=menu->nextSibling->previousSibling;
					menu->nextSibling->previousSibling=menu;	// link me in as previous sibling of old first child if there was one
					menu->previousSibling->nextSibling=menu;	// link me in as next sibling
				}
				else
				{
					menu->previousSibling=menu;					// loop previous sibling to myself
					menu->nextSibling=menu;						// loop next sibling to myself
				}
				baseMenu->firstChild=menu;						// this is now the first child
			}
			else												// creating a child of the universe
			{
				if(firstMenu)
				{
					menu->nextSibling=firstMenu;
					menu->previousSibling=firstMenu->previousSibling;
					menu->previousSibling->nextSibling=menu;
					firstMenu->previousSibling=menu;
				}
				else
				{
					menu->previousSibling=menu;
					menu->nextSibling=menu;
				}
				firstMenu=menu;
			}
			break;
		case MR_LAST_CHILD:
			menu->parent=baseMenu;								// set parent
			if(baseMenu)										// see if linking as a child of a current menu
			{
				if((menu->nextSibling=baseMenu->firstChild))
				{
					menu->previousSibling=menu->nextSibling->previousSibling;
					menu->nextSibling->previousSibling=menu;	// link me in as previous sibling of old first child if there was one
					menu->previousSibling->nextSibling=menu;	// link me in as next sibling
				}
				else
				{
					menu->previousSibling=menu;					// loop previous sibling to myself
					menu->nextSibling=menu;						// loop next sibling to myself
					baseMenu->firstChild=menu;
				}
			}
			else												// creating a child of the universe
			{
				if(firstMenu)
				{
					menu->nextSibling=firstMenu;
					menu->previousSibling=firstMenu->previousSibling;

					menu->previousSibling->nextSibling=menu;

					firstMenu->previousSibling=menu;
				}
				else
				{
					menu->previousSibling=menu;
					menu->nextSibling=menu;
					firstMenu=menu;
				}
			}
			break;
		default:
			ErrorSet(errorDescriptions[BAD_RELATIONSHIP]);
			fail=true;
			break;
	}
	return(!fail);
}

static void MenuElementDispose(EDITOR_MENU *menu)
// Dispose of an editor menu structure.
{
	MenuSubWindowDispose(menu);
	MenuAttributesUnRealize(menu);
	MDisposePtr(menu->dataText);
	MDisposePtr(menu->attributes);
	MDisposePtr(menu->itemText);
	MDisposePtr(menu);
}

static EDITOR_MENU *MenuElementCreate(const char *name,const char *attributes,const char *dataText)
// Create an editor menu structure (not linked to anything, but fill in
// name, and attributes).
// If there is a problem, set the error, and return NULL.
{
	EDITOR_MENU
		*menu;

	if((menu=(EDITOR_MENU *)MNewPtr(sizeof(EDITOR_MENU))))				// create editor menu element
	{
		if((menu->itemText=(char *)MNewPtr(strlen(name)+1)))
		{
			strcpy(menu->itemText,name);								// copy the name
			if((menu->attributes=(char *)MNewPtr(strlen(attributes)+1)))
			{
				strcpy(menu->attributes,attributes);					// copy the attributes
				if((menu->dataText=(char *)MNewPtr(strlen(dataText)+1)))
				{
					strcpy(menu->dataText,dataText);					// copy the data text for this menu
					menu->active=menu->highlight=false;					// item is not active, and not highlighted by default
					menu->window=NULL;									// this menu currently has no window
					if(MenuAttributesRealize(menu))							// parse through the attribute string, set up font, and command string
					{
						return(menu);
					}
					MDisposePtr(menu->dataText);
				}
				MDisposePtr(menu->attributes);
			}
			MDisposePtr(menu->itemText);
		}
		MDisposePtr(menu);
	}
	return(NULL);
}

static EDITOR_MENU *MenuCreateAtBase(EDITOR_MENU *baseMenu,int menuRelationship,const char *name,const char *attributes,const char *dataText,bool active)
// Create a new menu with name, as menuRelationship to baseMenu
// and return a pointer to it.
// If baseMenu is NULL it indicates the top of the hierarchy.
// Therefore, menuRelationship must be MR_FIRST_CHILD, or MR_LAST_CHILD as it is not possible to
// create menus which are siblings of the root.
// attributes is a machine specific string that describes machine specific attributes of the menu.
// If active is false, the menu should be created inactive.
// If there is a problem, set the error, and return NULL.
{
	EDITOR_MENU
		*menu;

	if((menu=MenuElementCreate(name,attributes,dataText)))
	{
		menu->active=active;										// set the active state
		if(MenuElementLink(menu,baseMenu,menuRelationship))
		{
			MenuWindowResize(menu->parent);							// if the parent of menu has a window, then update the window so that this child item appears
			if(menu->parent)
			{
				if((menu->previousSibling==menu)&&(menu->nextSibling==menu))	// if this is the only child of the parent, then the parent needs to be updated to show he has children
				{
					MenuWindowResize(menu->parent->parent);			// creating this item has forced my parent to change (to show he has children)
				}
			}
			return(menu);
		}
		MenuElementDispose(menu);
	}
	return(NULL);
}

static void MenuRealize(EDITOR_MENU *menu)
// Realize menu attributes (and any children it may have).
// If NULL is passed, realize ALL menus.
// NOTE: this calls itself recursively.
{
	EDITOR_MENU
		*subMenu;

	subMenu=MenuRelatedGet(menu,MR_FIRST_CHILD);					// see if this one has any children
	while(subMenu)
	{
		MenuRealize(subMenu);
		subMenu=MenuRelatedGet(subMenu,MR_NEXT_SIBLING);
	}
	if(menu)
	{
		MenuAttributesUnRealize(menu);
		MenuAttributesRealize(menu);
	}
	MenuWindowResize(menu);
}

// ----------------------------------------------------------------------------

void MenuCommandKeyChange()
// Call this to let the menu code know the command key has changed -- it may need
// to redraw the root.
{
	MenuRealize(NULL);
}

void MenuColorsGet(EDITOR_COLOR *backgroundColor,EDITOR_COLOR *foregroundColor,EDITOR_COLOR *keyColor,EDITOR_COLOR *shadowTopColor,EDITOR_COLOR *shadowBottomColor)
// Return the colors currently in use by the menus.
{
	*backgroundColor=menuBackgroundColor;
	*foregroundColor=menuForegroundColor;
	*keyColor=menuKeyColor;
	*shadowTopColor=menuShadowTopColor;
	*shadowBottomColor=menuShadowBottomColor;
}

void MenuColorsSet(EDITOR_COLOR backgroundColor,EDITOR_COLOR foregroundColor,EDITOR_COLOR keyColor,EDITOR_COLOR shadowTopColor,EDITOR_COLOR shadowBottomColor)
// Set the colors to draw the menus in.
{
	menuBackgroundColor=backgroundColor;
	menuForegroundColor=foregroundColor;
	menuKeyColor=keyColor;
	menuShadowTopColor=shadowTopColor;
	menuShadowBottomColor=shadowBottomColor;
	MenuRealize(NULL);
}

const char *MenuFontGet()
// Return the name of the font currently in use by the menus.
{
	return(menuFontName);
}

void MenuFontSet(const char *fontName)
// Set the font to be used for all menu items that don't explicitly call one out.
{
	strncpy(menuFontName,fontName,sizeof(menuFontName)-1);
	menuFontName[sizeof(menuFontName)-1]='\0';
	MenuRealize(NULL);
}

void MenuRootMove(int32_t x,int32_t y)
// X specific function to position the menu window.
{
	XMoveWindow(xDisplay,rootMenuWindow->xWindow,x,y);
}

void MenuRootRaise()
// X specific menu function to pull menus to the top of the window pile.
{
	XRaiseWindow(xDisplay,rootMenuWindow->xWindow);				// raise the root menu window to the top
}

// ----------------------------------------------------------------------------

EDITOR_MENU *PlatformMenuRelatedGet(EDITOR_MENU *menu,int menuRelationship)
// Return the menu with menuRelationship to menu.
// Return NULL if there is no menu with menuRelationship.
{
	return(MenuRelatedGet(menu,menuRelationship));
}

void PlatformMenuActiveSet(EDITOR_MENU *menu,bool active)
// Set menu active state.
// If a menu is set inactive, and it contains sub-menus, they should be inactive because
// their parent is now inactive, but they should not have their active
// state changed. This way, when the parent becomes active again, the children
// will resume their old states.
{
	menu->active=active;
	MenuEnableStateSet(menu);
}

char *PlatformMenuNameGet(EDITOR_MENU *menu)
// Return a pointer to the name of menu.
{
	return(menu->itemText);
}

bool PlatformMenuNameSet(EDITOR_MENU *menu,const char *name)
// Set the name for menu.
// If this fails, leave the old name, set the error, and return false.
{
	char
		*newName;
	if((newName=(char *)MNewPtr(strlen(name)+1)))
	{
		strcpy(newName,name);
		MDisposePtr(menu->itemText);
		menu->itemText=newName;
		return(true);
	}
	return(false);
}

char *PlatformMenuAttributesGet(EDITOR_MENU *menu)
// Return a pointer to the attributes string for menu.
{
	return(menu->attributes);
}

bool PlatformMenuAttributesSet(EDITOR_MENU *menu,const char *attributes)
// Set the attributes for menu.
// If this fails, set the error, and return false.
{
	char
		*newAttributes;

	if((newAttributes=(char *)MNewPtr(strlen(attributes)+1)))
	{
		strcpy(newAttributes,attributes);
		MDisposePtr(menu->attributes);
		menu->attributes=newAttributes;
		return(true);
	}
	return(false);
}

char *PlatformMenuDataTextGet(EDITOR_MENU *menu)
// Return a pointer to the data text string for menu.
{
	return(menu->dataText);
}

bool PlatformMenuDataTextSet(EDITOR_MENU *menu,const char *dataText)
// Set the data text for menu.
// If this fails, leave the data text, set the error, and return false.
{
	char
		*newDataText;

	if((newDataText=(char *)MNewPtr(strlen(dataText)+1)))
	{
		strcpy(newDataText,dataText);
		MDisposePtr(menu->dataText);
		menu->dataText=newDataText;
		return(true);
	}
	return(false);
}

bool PlatformMenuGet(int argc,const char *argv[],EDITOR_MENU **menu)
// Traverse the elements of the menu path given by argc,argv
// and attempt to locate the menu that they point to.
// If the path is invalid, return false.
// Otherwise return true with menu set to the located item.
// NOTE: on entry, menu is used as the starting point for
// the search. If it is passed as NULL, it indicates that the search
// should start at the top.
{
	EDITOR_MENU
		*currentMenu,
		*testMenu;
	int
		elementIndex;
	bool
		matching;

	currentMenu=testMenu=*menu;
	matching=true;
	for(elementIndex=0;(elementIndex<argc)&&matching;elementIndex++)
	{
		if(currentMenu)							// locate root of current menu
		{
			testMenu=currentMenu->firstChild;
		}
		else
		{
			testMenu=firstMenu;
		}
		matching=false;
		while(testMenu&&!matching)
		{
			if(strcmp(argv[elementIndex],testMenu->itemText)==0)
			{
				matching=true;					// located a match, this is the menu we want
			}
			else
			{
				testMenu=PlatformMenuRelatedGet(testMenu,MR_NEXT_SIBLING);
			}
		}
		currentMenu=testMenu;
	}
	*menu=testMenu;
	return(matching);
}


void PlatformMenuDispose(EDITOR_MENU *menu)
// Remove menu (and any children it may have).
// If NULL is passed, dispose of ALL menus.
// NOTE: this calls itself recursively.
{
	EDITOR_MENU
		*parent,
		*subMenu;

	while((subMenu=PlatformMenuRelatedGet(menu,MR_LAST_CHILD)))	// see if this one has any children
	{
		PlatformMenuDispose(subMenu);
	}
	if(menu)
	{
		parent=menu->parent;							// remember the parent, so after this is disposed of, he can be updated
		MenuElementUnlink(menu);						// unlink this item
		MenuElementDispose(menu);						// get rid of it
		if(parent)										// see if parent is not root
		{
			if(parent->window&&!parent->firstChild)		// if the parent has a window, but now has no children, kill the window
			{
				MenuSubWindowDispose(parent);			// get rid of the window if it contained the last child
			}
			if(!parent->firstChild)
			{
				MenuWindowResize(parent->parent);		// tell the parent's parent to resize (to show that there are no children)
			}
		}
		MenuWindowResize(parent);						// if the parent of menu has a window, then update the window so that this child item disappears
	}
}

EDITOR_MENU *PlatformMenuCreate(int argc,const char *argv[],int menuRelationship,const char *name,const char *attributes,const char *dataText,bool active)
// Create a new menu with name, as menuRelationship to the passed argc,argv path.
// The path is absolute within the menu hierarchy.
// Attributes is a machine specific string that describes machine specific attributes of the menu.
// If active is false, the menu should be created inactive.
// If there is a problem, set the error, and return false.
{
	EDITOR_MENU
		*parentMenu,
		*testMenu;
	bool
		fail;

	parentMenu=NULL;									// start at the top
	if(PlatformMenuGet(argc,argv,&parentMenu))			// locate item to link this one to
	{
		testMenu=parentMenu;
		fail=false;
		if(menuRelationship==MR_NEXT_SIBLING||menuRelationship==MR_PREVIOUS_SIBLING)	// see if hunting for sibling
		{
			if(parentMenu)								// if he wants sibling, better not be at top
			{
				testMenu=testMenu->parent;				// back up one
				if(PlatformMenuGet(1,&name,&testMenu))	// see if an item already exists with this name in the place we are putting it
				{
					if(testMenu!=parentMenu)			// cannot delete the one we want to add to!!
					{
						PlatformMenuDispose(testMenu);	// kill it before adding new item
					}
					else
					{
						ErrorSet(errorDescriptions[BAD_RELATIONSHIP]);
						fail=true;
					}
				}
			}
			else
			{
				ErrorSet(errorDescriptions[BAD_RELATIONSHIP]);
				fail=true;
			}
		}
		else
		{
			if(PlatformMenuGet(1,&name,&testMenu))		// see if an item already exists with this name in the place we are putting it
			{
				PlatformMenuDispose(testMenu);			// kill it before adding new item
			}
		}
		if(!fail)
		{
			return(MenuCreateAtBase(parentMenu,menuRelationship,name,attributes,dataText,active));
		}
	}
	else
	{
		ErrorSet(errorDescriptions[BAD_PATH]);
	}
	return(NULL);
}

// ----------------------------------------------------------------------------

void MenuUnInit()
// Uninitialize the editor menuing system.
{
	PlatformMenuDispose(NULL);									// get rid of any menus that were created
	MenuRootWindowDispose(rootMenuWindow);						// get rid of the root of the menu windows
}

bool MenuInit()
// Initialize the editor menuing system.
// If there is a problem, ErrorSet, return false.
{
	EDITOR_RECT
		rect;

	firstMenu=NULL;												// no menus exist yet
	menuOpen=false;												// no menu open currently

	strcpy(menuFontName,DEFAULT_MENU_FONT);						// default menu font

	// set up default menu colors
	menuForegroundColor=BLACK;
	menuBackgroundColor=BACKGROUND_COLOR;
	menuKeyColor=WHITE;
	menuShadowTopColor=WHITE;
	menuShadowBottomColor=GRAY_0;

	rect.x=10;													// this is the desired place for the main menu window
	rect.y=20;
	rect.w=1;													// make tiny window, because we have no menus yet
	rect.h=1;
	if((rootMenuWindow=MenuRootWindowCreate(&rect,programName)))	// attempt to create the root of the menu windows
	{
		return(true);
	}
	ErrorSet("Failed to MenuInit");
	return(false);
}
