// Dialog popup menu handler
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define POPUP_SHADOW_THICKNESS			1						// number of pixels of shadow around popup area, and menus

#define	POPUP_ADDITIONAL_WIDTH			6						// additional pixel width added to each menu item to make it look better when drawn
#define	POPUP_ADDITIONAL_HEIGHT			6						// additional pixel height added to each menu item to make it look better when drawn

typedef struct popup
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;										// function which is used to position this item in the dialog
	EDITOR_RECT
		rect;													// the rectangle for the item itself
	EDITOR_COLOR
		activeBackgroundColor,									// background when active
		activeForegroundColor,									// foreground when active
		inactiveBackgroundColor,								// background when inactive
		inactiveForegroundColor,								// foreground when inactive
		borderColor;											// color for borders
	EDITOR_FONT
		*font;													// info about currently selected font
	bool
		active;													// tells if the popup is "active" (can be clicked on, and drawn active)
	uint32_t
		numElements;											// tells how many elements are in the list
	const char
		**listElements;											// pointer to array of pointers to items in the list
	uint32_t
		*currentElement;										// currently selected element in the popup (updated by popup tracking)
	popupPressedFunction
		*pressedFunction;										// routine called when an item has been chosen
	EDITOR_WINDOW
		*menuWindow;											// the popup menu window (created during click)
	bool
		menuTrackDone;											// set true when done tracking the menu item
	bool
		haveTrackElement;										// true if an element is currently selected during tracking
	uint32_t
		trackElement;											// if haveTrackElement is true, tells which item is selected
} POPUP;

static bool PopupPointInside(POPUP *popup,int32_t x,int32_t y)
// Return true if parent window relative x/y are in popup,
// false if not.
{
	return(PointInRect(x,y,&popup->rect));
}

static void PopupInvalidate(POPUP *popup)
// Invalidate the popup within its parent window.
{
	WindowRectInvalidate(DialogWindowGet(popup->dialogItem.dialog),&popup->rect);
}

static void PopupRectangleSet(POPUP *popup)
// Work out the rectangle given the current window.
{
	if(popup->positionFunction)	// see if non-null resize function specified
	{
		popup->positionFunction(&popup->dialogItem,&popup->rect);	// get new rectangle
		PopupInvalidate(popup);									// invalidate it
	}
}

static int32_t PopupBaselineCalculate(POPUP *popup,uint32_t ascent,uint32_t descent)
// Work out the vertical distance from the top of the popup window
// to the baseline where the popup string is drawn.
{
	return(popup->rect.y+(popup->rect.h/2)-(ascent+descent)/2+ascent);	// center the string vertically
}

static void PopupDraw(POPUP *popup,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Redraw the popup area into imageBuffer.
{
	uint32_t
		ascent,
		descent;
	uint32_t
		stringLength,
		stringWidth;
	int
		stringX,
		stringY;
	EDITOR_COLOR
		color;

	if(popup->rect.w&&popup->rect.h)
	{
		if(XRectInRegion(invalidRegion,popup->rect.x,popup->rect.y,popup->rect.w,popup->rect.h)!=RectangleOut)
		{
			ImageClipRegionRectSet(imageBuffer,&popup->rect);			// don't draw outside the popup area

			color=popup->inactiveBackgroundColor;
			if(popup->active)
			{
				color=popup->activeBackgroundColor;
			}

			ImageFill(imageBuffer,color,popup->rect.x,popup->rect.y,popup->rect.w,popup->rect.h);

			FontHeightsGet(popup->font,&ascent,&descent);

			stringLength=strlen(popup->listElements[*popup->currentElement]);
			stringWidth=FontTextUtf8Width(popup->font,popup->listElements[*popup->currentElement],stringLength);
			stringX=popup->rect.x+(popup->rect.w/2)-(stringWidth/2);
			stringY=PopupBaselineCalculate(popup,ascent,descent);

			color=popup->inactiveForegroundColor;
			if(popup->active)
			{
				color=popup->activeForegroundColor;
			}

			FontTextUtf8Draw(popup->font,imageBuffer,color,stringX,stringY,popup->listElements[*popup->currentElement],stringLength);

			ShadowRectangleOutline(imageBuffer,&popup->rect,popup->borderColor,popup->borderColor,POPUP_SHADOW_THICKNESS);
			ImageClipRegionSet(imageBuffer,None);							// unset clip area
		}
	}
}

static void PopupMenuWindowInvalidate(POPUP *popup)
// Invalidate the popup menu window.
{
	EDITOR_RECT
		rect;													// the rectangle for the item itself

	rect.x=0;
	rect.y=0;
	rect.w=popup->menuWindow->w;
	rect.h=popup->menuWindow->h;
	WindowRectInvalidate(popup->menuWindow,&rect);
}

static uint32_t PopupMenuWindowItemHeight(POPUP *popup,uint32_t ascent,uint32_t descent)
// Return the height of each item in the popup menu window.
{
	return(ascent+descent+POPUP_ADDITIONAL_HEIGHT);
}

static int32_t PopupMenuWindowItemTopCalculate(POPUP *popup,uint32_t ascent,uint32_t descent,uint32_t itemIndex)
// Work out the vertical distance from the top the popup *menu* window
// to the top of the item given by itemIndex.
{
	return(itemIndex*(PopupMenuWindowItemHeight(popup,ascent,descent)));
}

static int32_t PopupMenuWindowItemBaselineCalculate(POPUP *popup,uint32_t ascent,uint32_t descent,uint32_t itemIndex)
// Work out the vertical distance from the top the popup *menu* window
// to the baseline of the item given by itemIndex.
{
	int32_t
		yOffset;

	yOffset=PopupMenuWindowItemTopCalculate(popup,ascent,descent,itemIndex);	// get offset to the top
	yOffset+=PopupMenuWindowItemHeight(popup,ascent,descent)/2-(ascent+descent)/2+ascent;	// center the text vertically in its spot
	return(yOffset);
}

static void PopupMenuWindowItemRectCalculate(POPUP *popup,uint32_t ascent,uint32_t descent,uint32_t itemIndex,EDITOR_RECT *rect)
// Work out the rectangle for the given item in the popup menu window.
{
	rect->x=0;
	rect->y=PopupMenuWindowItemTopCalculate(popup,ascent,descent,itemIndex);
	rect->w=popup->rect.w;
	rect->h=PopupMenuWindowItemHeight(popup,ascent,descent);
}

static void PopupMenuWindowSizeCalculate(POPUP *popup,uint32_t ascent,uint32_t descent,uint32_t *width,uint32_t *height)
// Work out the width and height of the popup window.
{
	(*width)=popup->rect.w;
	(*height)=popup->numElements*(PopupMenuWindowItemHeight(popup,ascent,descent));
}

static void PopupMenuWindowDraw(EDITOR_WINDOW *window,GUI_IMAGE *imageBuffer)
// Draw the contents of the popup window.
{
	POPUP
		*popup;
	EDITOR_RECT
		rect;
	uint32_t
		ascent,
		descent;
	uint32_t
		i;
	int32_t
		stringX,
		stringY;
	uint32_t
		stringLength,
		stringWidth;

	popup=(POPUP *)window->windowInfo;
	FontHeightsGet(popup->font,&ascent,&descent);
	ImageFill(imageBuffer,popup->activeBackgroundColor,0,0,window->w,window->h);	// fill the background of the window
	rect.x=0;
	rect.y=0;
	rect.w=window->w;
	rect.h=window->h;
	ShadowRectangleOutline(imageBuffer,&rect,popup->borderColor,popup->borderColor,POPUP_SHADOW_THICKNESS);
	for(i=0;i<popup->numElements;i++)
	{
		stringLength=strlen(popup->listElements[i]);
		stringWidth=FontTextUtf8Width(popup->font,popup->listElements[i],stringLength);
		stringX=(popup->rect.w/2)-(stringWidth/2);
		stringY=PopupMenuWindowItemBaselineCalculate(popup,ascent,descent,i);
		FontTextUtf8Draw(popup->font,imageBuffer,popup->activeForegroundColor,stringX,stringY,popup->listElements[i],stringLength);
		if((popup->haveTrackElement)&&(i==popup->trackElement))
		{
			PopupMenuWindowItemRectCalculate(popup,ascent,descent,i,&rect);
			ShadowRectangleOutline(imageBuffer,&rect,popup->borderColor,popup->borderColor,2);
		}
	}
}

static void PopupMenuWindowUpdate(EDITOR_WINDOW *window)
// Called to redraw the popup menu window.
{
	Window
		xWindow;
	GUI_IMAGE
		*imageBuffer;

	if(!XEmptyRegion(window->invalidRegion))						// see if it has an invalid area
	{
		xWindow=window->xWindow;									// get x window for this window
		if((imageBuffer=ImageCreate(window->w,window->h)))			// create an image to be used to composite the window off-screen
		{
			PopupMenuWindowDraw(window,imageBuffer);
			ImageDrawToWindow(imageBuffer,xWindow,window->invalidRegion,window->w,window->h);
			ImageDestroy(imageBuffer);
		}
		XDestroyRegion(window->invalidRegion);						// clear out the invalid region now that the window has been drawn
		window->invalidRegion=XCreateRegion();
	}
}

static void PopupRectPin(EDITOR_RECT *rect)
// Make sure rect does not go off the display (as best as possible).
{
	uint32_t
		displayWidth,
		displayHeight;

	EditorDisplayDimensionsGet(&displayWidth,&displayHeight);
	if(rect->x+rect->w>displayWidth)
	{
		rect->x=displayWidth-rect->w;
	}
	if(rect->y+rect->h>displayHeight)
	{
		rect->y=displayHeight-rect->h;
	}
	if(rect->x<0)
	{
		rect->x=0;
	}
	if(rect->y<0)
	{
		rect->y=0;
	}
}

static bool PopupMenuWindowItemLocate(POPUP *popup,int32_t rootX,int32_t rootY,uint32_t *itemIndex)
// Given an X, and Y that are relative to the X11 root, work out which popup item
// the point is over, set itemIndex, and return true.
// If X,Y are not over any item, return false.
{
	int32_t
		localX,
		localY;
	uint32_t
		ascent,
		descent;
	uint32_t
		i;
	EDITOR_RECT
		rect;

	FontHeightsGet(popup->font,&ascent,&descent);
	GlobalToLocal(popup->menuWindow->xWindow,rootX,rootY,&localX,&localY);	// convert global coordinates to those local to the menu window
	for(i=0;i<popup->numElements;i++)
	{
		PopupMenuWindowItemRectCalculate(popup,ascent,descent,i,&rect);
		if(PointInRect(localX,localY,&rect))
		{
			*itemIndex=i;
			return(true);
		}
	}
	return(false);
}

static void PopupMenuWindowButtonPress(POPUP *popup,XEvent *event)
// On second click of the popup menu, with the mouse button still
// down, track the user interaction. If the button is released outside
// the menu, then don't update the current item. If it is released
// within a menu item, select that item.
{
	uint32_t
		newItem;
	int
		rootX,
		rootY,
		windowX,
		windowY,
		lastX,
		lastY;
	int
		button;

	button=event->xbutton.button;
	if((popup->haveTrackElement=PopupMenuWindowItemLocate(popup,event->xbutton.x_root,event->xbutton.y_root,&popup->trackElement)))	// find the item (if any) that is under the pointer at initial click
	{
		PopupMenuWindowInvalidate(popup);					// invalidate the menu window to redraw with the selected track element
	}
	rootX=event->xbutton.x_root;
	rootY=event->xbutton.y_root;
	windowX=event->xbutton.x;
	windowY=event->xbutton.y;
	lastX=rootX;
	lastY=rootY;
	while(StillDown(event->xbutton.window,button,&rootX,&rootY,&windowX,&windowY))
	{
		if(lastX!=rootX||lastY!=rootY)						// make sure something has changed
		{
			if(PopupMenuWindowItemLocate(popup,rootX,rootY,&newItem))	// find the item (if any) that is under the pointer currently
			{
				if((!popup->haveTrackElement)||(newItem!=popup->trackElement))
				{
					popup->haveTrackElement=true;
					popup->trackElement=newItem;
					PopupMenuWindowInvalidate(popup);
				}
			}
			else
			{
				if(popup->haveTrackElement)
				{
					popup->haveTrackElement=false;
					PopupMenuWindowInvalidate(popup);
				}
			}
			lastX=rootX;
			lastY=rootY;
		}
	}
}

static void PopupMenuWindowEvent(EDITOR_WINDOW *window,XEvent *event)
// Handle an event arriving for a popup menu window.
// These only happen while the pointer is gabbed.
{
	POPUP
		*popup;

	switch(event->type)
	{
		case ButtonPress:
			popup=(POPUP *)window->windowInfo;
			PopupMenuWindowButtonPress(popup,event);
			XUngrabPointer(xDisplay,event->xbutton.time);
			XUngrabKeyboard(xDisplay,event->xbutton.time);
			popup->menuTrackDone=true;
			break;
		default:
			break;
	}
}

static void PopupMenuWindowDispose(EDITOR_WINDOW *window)
// Remove the given popup window.
{
	XDestroyRegion(window->invalidRegion);							// destroy invalid region
	XDestroyWindow(xDisplay,window->xWindow);						// tell x to make the window go away
	WaitUntilDestroyed(window);										// wait for it to die
	WindowElementUnlink(window);									// unlink it from our list
	MDisposePtr(window);
}

static EDITOR_WINDOW *PopupMenuWindowCreate(POPUP *popup)
// This is called to create the popup menu window.
// Popup menu windows do not have normal window decoration.
// NOTE: this checks to see if the popup will go off the screen, and pins it
// into range.
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	XSetWindowAttributes
		attributes;
	uint32_t
		ascent,
		descent;
	uint32_t
		valueMask;
	Atom
		protocols[1];													// number of protocols

	popup->haveTrackElement=false;										// until further notice, don't highlight any element in the window
	if((window=(EDITOR_WINDOW *)MNewPtr(sizeof(EDITOR_WINDOW))))		// create editor window
	{
		WindowElementLink(window);										// link it to the global window list that we maintain
		window->windowType=EWT_TRANSIENT;								// set the editor window type
		window->windowInfo=popup;										// set window info pointer
		window->configureFunction=NULL;									// no configure function as window does not spontaneously change size
		window->updateFunction=PopupMenuWindowUpdate;					// point to the function to call to update this window when it needs to be redrawn
		window->eventFunction=PopupMenuWindowEvent;						// point to the function to call to handle events in the popup menu window
		window->periodicFunction=NULL;									// no periodic activity for these windows

		attributes.save_under=True;
		attributes.override_redirect=True;
		valueMask=CWSaveUnder|CWOverrideRedirect;

		FontHeightsGet(popup->font,&ascent,&descent);
		rect.x=popup->rect.x;
		rect.y=PopupBaselineCalculate(popup,ascent,descent)-PopupMenuWindowItemBaselineCalculate(popup,ascent,descent,*popup->currentElement);	// get Y for window that lines up the baselines
		LocalToGlobal(DialogWindowGet(popup->dialogItem.dialog)->xWindow,rect.x,rect.y,&rect.x,&rect.y);	// make global

		PopupMenuWindowSizeCalculate(popup,ascent,descent,&rect.w,&rect.h);	// work out the width and height of the window
		PopupRectPin(&rect);											// pin the rectangle to the screen

		if(rect.w==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect.w=1;
		}
		if(rect.h==0)													// ### this is because X will not allow opening windows of 0 width or height!
		{
			rect.h=1;
		}
		window->w=rect.w;												// remember initial size for the window
		window->h=rect.h;

		// Due to X's nasty error handling, we assume this create does not fail!!!
		window->xWindow=XCreateWindow(xDisplay,RootWindow(xDisplay,xScreenNum),rect.x,rect.y,rect.w,rect.h,0,DefaultDepth(xDisplay,xScreenNum),InputOutput,DefaultVisual(xDisplay,xScreenNum),valueMask,&attributes);

//		XSetTransientForHint(xDisplay,window->xWindow,None);			// tell WM that this window is modal to all other windows in the application

		XStoreName(xDisplay,window->xWindow,"popup");
		XSetIconName(xDisplay,window->xWindow,"");
		XSelectInput(xDisplay,window->xWindow,ExposureMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|KeyPressMask|StructureNotifyMask);	// want to see these events

		protocols[0]=deleteWindowAtom;
		XSetWMProtocols(xDisplay,window->xWindow,&(protocols[0]),1);	// tell X we would like to participate in these protocols

		window->invalidRegion=XCreateRegion();							// create new empty invalid region

		XDefineCursor(xDisplay,window->xWindow,arrowCursor);			// this is the cursor to use in the popup
		XMapRaised(xDisplay,window->xWindow);							// map window to the top of the display
		WaitUntilMappedNoFocus(window);									// wait until this window is actually mapped before continuing
		return(window);
	}
	ErrorSet("Failed to PopupMenuWindowCreate");
	return((EDITOR_WINDOW *)NULL);
}

static void PopupEventLoop(POPUP *popup,Time time)
// User has brought up the popup, but not yet selected an item.
// Process events, attempting to complete the selection process.
// NOTE: this will not return unless the user cancels (by clicking outside
// the popup), or selects an item from the popup.
{
	XGrabPointer(xDisplay,popup->menuWindow->xWindow,False,ButtonPressMask|ButtonReleaseMask,GrabModeAsync,GrabModeAsync,None,None,time);
	XGrabKeyboard(xDisplay,popup->menuWindow->xWindow,False,GrabModeAsync,GrabModeAsync,time);
	popup->menuTrackDone=false;
	EventLocalLoop(&popup->menuTrackDone,popup->menuWindow);
}

static void PopupMenuWindowInitialTrack(POPUP *popup,XEvent *event)
// On creation of the popup menu, with the mouse button still
// down, track the user interaction. If they release the button
// before making a new selection, leave the window open, and wait
// for additional clicks. If they make a new selection, but then
// release somewhere else, just cancel. If they make a new selection
// and release within it, then take that as their selection.
{
	bool
		newItemSelected;		// tells if the user selected something other than the initial item
	uint32_t
		newItem;
	int
		rootX,
		rootY,
		windowX,
		windowY,
		lastX,
		lastY;
	int
		button;

	newItemSelected=false;
	button=event->xbutton.button;
	if((popup->haveTrackElement=PopupMenuWindowItemLocate(popup,event->xbutton.x_root,event->xbutton.y_root,&popup->trackElement)))	// find the item (if any) that is under the pointer at initial click
	{
		PopupMenuWindowInvalidate(popup);					// invalidate the menu window to redraw with the selected track element
	}
	rootX=event->xbutton.x_root;
	rootY=event->xbutton.y_root;
	windowX=event->xbutton.x;
	windowY=event->xbutton.y;
	lastX=rootX;
	lastY=rootY;
	while(StillDown(event->xbutton.window,button,&rootX,&rootY,&windowX,&windowY))
	{
		if(lastX!=rootX||lastY!=rootY)						// make sure something has changed
		{
			if(PopupMenuWindowItemLocate(popup,rootX,rootY,&newItem))	// find the item (if any) that is under the pointer currently
			{
				if((!popup->haveTrackElement)||(newItem!=popup->trackElement))
				{
					newItemSelected=true;					// got something new (means we don't need to leave the menu window open on button release)
					popup->haveTrackElement=true;
					popup->trackElement=newItem;
					PopupMenuWindowInvalidate(popup);
				}
			}
			else
			{
				if(popup->haveTrackElement)
				{
					newItemSelected=true;					// got something new (means we don't need to leave the menu window open on button release)
					popup->haveTrackElement=false;
					PopupMenuWindowInvalidate(popup);
				}
			}
			lastX=rootX;
			lastY=rootY;
		}
	}

	if(!newItemSelected)									// see if we need to leave the window open, and wait for an additional click
	{
		PopupEventLoop(popup,event->xbutton.time);			// go get an additional click
	}

	if(popup->haveTrackElement)								// is an element selected?
	{
		*popup->currentElement=popup->trackElement;			// set this element as the current one
		if(popup->pressedFunction)
		{
			popup->pressedFunction(&popup->dialogItem);
		}
	}
}

static bool PopupTrack(POPUP *popup,XEvent *event)
// See if the point given by the button event falls over popup.
// If so, create a popup menu window, track it, and return true.
// If not, return false.
{
	if((event->xbutton.button==Button1)&&PopupPointInside(popup,event->xbutton.x,event->xbutton.y))	// click inside the popup item?
	{
		if(popup->active)
		{
			if((popup->menuWindow=PopupMenuWindowCreate(popup)))	// create the popup menu window
			{
				PopupMenuWindowInitialTrack(popup,event);			// handle initial tracking of the click
			}
			PopupMenuWindowDispose(popup->menuWindow);
			PopupInvalidate(popup);									// redraw the popup, since the item may have now changed
		}
		return(true);
	}
	return(false);
}

// dialog interface routines

static void PopupItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the popup pointed to by item into imageBuffer.
{
	PopupDraw((POPUP *)item,imageBuffer,invalidRegion);
}

static void PopupItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	POPUP
		*popup;

	popup=(POPUP *)item;
	PopupInvalidate(popup);				// invalidate where we're at currently
	PopupRectangleSet(popup);
}

static bool PopupItemTrack(DIALOG_ITEM *item,XEvent *event)
// Track the popup pointed to by item.
{
	return(PopupTrack((POPUP *)item,event));
}

bool PopupItemActiveGet(DIALOG_ITEM *item)
// Return true of the given popup is active.
{
	POPUP
		*popup;

	popup=(POPUP *)item;
	return(popup->active);
}

void PopupItemActiveSet(DIALOG_ITEM *item,bool active)
// Adjust the active state of the popup.
{
	POPUP
		*popup;

	popup=(POPUP *)item;
	popup->active=active;
	PopupInvalidate(popup);
}

static void PopupDispose(POPUP *popup)
// Unlink popup from its parent window, and delete it.
// Invalidate the area of the window where popup was.
{
	PopupInvalidate(popup);					// invalidate where it is now
	PlatformFontFree(popup->font);
	MDisposePtr(popup);
}

static void PopupItemDispose(DIALOG_ITEM *item)
// Dispose of the local popup structure pointed to by item.
{
	PopupDispose((POPUP *)item);
}

static POPUP *PopupCreate(DIALOG *dialog,POPUP_DESCRIPTOR *descriptor)
// Create a popup in dialog, invalidate the area of dialog
// where the popup is to be placed.
// Return a pointer to it if successful.
// ErrorSet, and return NULL if not.
{
	POPUP
		*popup;

	if((popup=(POPUP *)MNewPtr(sizeof(POPUP))))					// create popup data structure
	{
		popup->dialogItem.dialog=dialog;						// point back at the dialog
		popup->dialogItem.disposeFunction=PopupItemDispose;
		popup->dialogItem.drawFunction=PopupItemDraw;
		popup->dialogItem.resizeFunction=PopupItemResize;
		popup->dialogItem.trackFunction=PopupItemTrack;
		popup->dialogItem.earlyKeyFunction=NULL;
		popup->dialogItem.wantFocus=false;
		popup->dialogItem.focusChangeFunction=NULL;
		popup->dialogItem.focusKeyFunction=NULL;
		popup->dialogItem.focusPeriodicFunction=NULL;

		popup->positionFunction=descriptor->positionFunction;

		popup->activeBackgroundColor=descriptor->activeBackgroundColor;			// copy colors
		popup->activeForegroundColor=descriptor->activeForegroundColor;
		popup->inactiveBackgroundColor=descriptor->inactiveBackgroundColor;
		popup->inactiveForegroundColor=descriptor->inactiveForegroundColor;
		popup->borderColor=descriptor->borderColor;

		popup->active=descriptor->active;						// set initial active state

		popup->numElements=descriptor->numElements;				// tells how many elements are in the list
		popup->listElements=descriptor->listElements;			// pointer to array of pointers to items in the list
		popup->currentElement=descriptor->currentElement;		// the currently selected element in the list
		popup->haveTrackElement=false;

		if(*(popup->currentElement)>=descriptor->numElements)	// pin the current element
		{
			if(descriptor->numElements)
			{
				*popup->currentElement=descriptor->numElements-1;
			}
			else
			{
				*popup->currentElement=0;
			}
		}

		popup->pressedFunction=descriptor->pressedFunction;		// copy the procedure to be called when the popup is pressed
		if((popup->font=PlatformFontLoad(descriptor->fontName)))
		{
			PopupRectangleSet(popup);
			return(popup);
		}
		MDisposePtr(popup);
	}
	return(NULL);
}

DIALOG_ITEM *PopupItemCreate(DIALOG *dialog,void *descriptor)
// Create the popup, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)PopupCreate(dialog,(POPUP_DESCRIPTOR *)descriptor));
}
