// Scroll bar handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

// Some terminology:
// low arrow is the arrow that attempts to move the thumb to lower values
// high arrow moves it higher

#define	SCROLL_INITIAL_DELAY		(SECOND*.2)		// number of ticks to wait before repeating
#define	SCROLL_REPEAT_DELAY			(SECOND*.01)	// number of ticks to between repeats

static void ScrollBarControlRectsGet(SCROLL_BAR *scrollBar,EDITOR_RECT *lowArrowRect,EDITOR_RECT *highArrowRect,EDITOR_RECT *thumbRect,EDITOR_RECT *slideRect,EDITOR_RECT *pageLowRect,EDITOR_RECT *pageHighRect)
// Return the rectangles that enclose the various pieces of scrollBar.
// If the scroll bar starts getting too small, certain rects may be returned with 0 width or
// height.
{
	uint32_t
		lowLength,
		highLength,
		thumbLength,
		thumbSpace;

	if(scrollBar->length<2*ARROW_LENGTH)			// problem fitting arrows?
	{
		lowLength=(scrollBar->length+1)/2;			// get the length of the two arrows, and work backwards from there
		highLength=scrollBar->length-lowLength;
	}
	else
	{
		lowLength=highLength=ARROW_LENGTH;
	}
	if(scrollBar->horizontal)
	{
		lowArrowRect->x=scrollBar->x;
		lowArrowRect->w=lowLength;
		lowArrowRect->y=scrollBar->y;
		lowArrowRect->h=SCROLL_BAR_WIDTH;

		highArrowRect->w=highLength;
		highArrowRect->x=(scrollBar->x+scrollBar->length)-highLength;
		highArrowRect->y=scrollBar->y;
		highArrowRect->h=SCROLL_BAR_WIDTH;

		slideRect->x=lowArrowRect->x+lowArrowRect->w;
		slideRect->w=highArrowRect->x-slideRect->x;
		slideRect->y=scrollBar->y;
		slideRect->h=SCROLL_BAR_WIDTH;

		thumbLength=(uint64_t)slideRect->w/(scrollBar->thumbMax+1);

		if(thumbLength<THUMB_LENGTH)
		{
			thumbLength=THUMB_LENGTH;
		}

		if(thumbLength>=slideRect->w)
		{
			thumbRect->w=slideRect->w;
		}
		else
		{
			thumbRect->w=thumbLength;
		}
		thumbSpace=slideRect->w-thumbRect->w;
		if(scrollBar->thumbMax)
		{
			thumbRect->x=slideRect->x+(((uint64_t)thumbSpace*scrollBar->thumbPosition)/scrollBar->thumbMax);
		}
		else
		{
			thumbRect->x=slideRect->x;
		}
		thumbRect->y=scrollBar->y;
		thumbRect->h=SCROLL_BAR_WIDTH;

		pageLowRect->x=slideRect->x;
		pageLowRect->y=slideRect->y;
		pageLowRect->w=thumbRect->x-slideRect->x;
		pageLowRect->h=slideRect->h;

		pageHighRect->x=thumbRect->x+thumbRect->w;
		pageHighRect->y=slideRect->y;
		pageHighRect->w=slideRect->x+slideRect->w-pageHighRect->x;
		pageHighRect->h=slideRect->h;
	}
	else
	{
		lowArrowRect->x=scrollBar->x;
		lowArrowRect->w=SCROLL_BAR_WIDTH;
		lowArrowRect->y=scrollBar->y;
		lowArrowRect->h=lowLength;

		highArrowRect->x=scrollBar->x;
		highArrowRect->w=SCROLL_BAR_WIDTH;
		highArrowRect->y=(scrollBar->y+scrollBar->length)-highLength;
		highArrowRect->h=highLength;

		slideRect->x=scrollBar->x;
		slideRect->w=SCROLL_BAR_WIDTH;
		slideRect->y=lowArrowRect->y+lowArrowRect->h;
		slideRect->h=highArrowRect->y-slideRect->y;

		thumbRect->x=scrollBar->x;
		thumbRect->w=SCROLL_BAR_WIDTH;

		thumbLength=(uint64_t)slideRect->h/(scrollBar->thumbMax+1);

		if(thumbLength<THUMB_LENGTH)
		{
			thumbLength=THUMB_LENGTH;
		}

		if(thumbLength>=slideRect->h)
		{
			thumbRect->h=slideRect->h;
		}
		else
		{
			thumbRect->h=thumbLength;
		}
		thumbSpace=slideRect->h-thumbRect->h;

		if(scrollBar->thumbMax)
		{
			thumbRect->y=slideRect->y+(((uint64_t)thumbSpace*scrollBar->thumbPosition)/scrollBar->thumbMax);
		}
		else
		{
			thumbRect->y=slideRect->y;
		}

		pageLowRect->x=slideRect->x;
		pageLowRect->y=slideRect->y;
		pageLowRect->w=slideRect->w;
		pageLowRect->h=thumbRect->y-slideRect->y;

		pageHighRect->x=slideRect->x;
		pageHighRect->y=thumbRect->y+thumbRect->h;
		pageHighRect->w=slideRect->w;
		pageHighRect->h=slideRect->y+slideRect->h-pageHighRect->y;
	}
}

bool ScrollBarPointInside(SCROLL_BAR *scrollBar,int32_t x,int32_t y)
// Return true if parent window relative x/y are in scrollBar,
// false if not.
{
	EDITOR_RECT
		rect;

	rect.x=scrollBar->x;
	rect.y=scrollBar->y;
	if(scrollBar->horizontal)
	{
		rect.w=scrollBar->length;
		rect.h=SCROLL_BAR_WIDTH;
	}
	else
	{
		rect.w=SCROLL_BAR_WIDTH;
		rect.h=scrollBar->length;
	}
	return(PointInRect(x,y,&rect));
}

static void ScrollBarInvalidate(SCROLL_BAR *scrollBar)
// Invalidate the scroll bar within its parent window.
{
	EDITOR_RECT
		invalidRect;

	invalidRect.x=scrollBar->x;
	invalidRect.y=scrollBar->y;
	if(scrollBar->horizontal)
	{
		invalidRect.w=scrollBar->length;
		invalidRect.h=SCROLL_BAR_WIDTH;
	}
	else
	{
		invalidRect.w=SCROLL_BAR_WIDTH;
		invalidRect.h=scrollBar->length;
	}
	WindowRectInvalidate(scrollBar->parentWindow,&invalidRect);
}

static void LeftArrowDraw(GUI_IMAGE *imageBuffer,EDITOR_RECT *rect,bool highlight)
// Draw a left arrow control into rect of imageBuffer.
{
	if(rect->w&&rect->h)
	{
		ImageDraw(leftArrowImage,imageBuffer,rect->x,rect->y);
	}
}

static void RightArrowDraw(GUI_IMAGE *imageBuffer,EDITOR_RECT *rect,bool highlight)
// Draw a right arrow control into rect of imageBuffer.
{
	if(rect->w&&rect->h)
	{
		ImageDraw(rightArrowImage,imageBuffer,rect->x,rect->y);
	}
}

static void UpArrowDraw(GUI_IMAGE *imageBuffer,EDITOR_RECT *rect,bool highlight)
// Draw an up arrow control into rect of imageBuffer.
{
	if(rect->w&&rect->h)
	{
		ImageDraw(upArrowImage,imageBuffer,rect->x,rect->y);
	}
}

static void DownArrowDraw(GUI_IMAGE *imageBuffer,EDITOR_RECT *rect,bool highlight)
// Draw a down arrow control into rect of imageBuffer.
{
	if(rect->w&&rect->h)
	{
		ImageDraw(downArrowImage,imageBuffer,rect->x,rect->y);
	}
}

static void PageIndicatorDraw(GUI_IMAGE *imageBuffer,EDITOR_RECT *rect,bool horizontal,bool highlight)
// Draw a page control into rect of imageBuffer.
{
	EDITOR_COLOR
		color;

	if(rect->w&&rect->h)
	{
		if(!highlight)
		{
			color=SCROLL_BAR_BACKGROUND;
		}
		else
		{
			color=GRAY_1;
		}
		if(horizontal)
		{
			if(rect->h>1)
			{
				ImageFill(imageBuffer,color,rect->x,rect->y+1,rect->w,rect->h-2);	// fill the inset rectangle
			}
			ImageFill(imageBuffer,BLACK,rect->x,rect->y,rect->w,1);					// draw top
			ImageFill(imageBuffer,BLACK,rect->x,rect->y+rect->h-1,rect->w,1);		// draw bottom
		}
		else
		{
			if(rect->w>1)
			{
				ImageFill(imageBuffer,color,rect->x+1,rect->y,rect->w-2,rect->h);	// fill the inset rectangle
			}
			ImageFill(imageBuffer,BLACK,rect->x,rect->y,1,rect->h);					// draw left
			ImageFill(imageBuffer,BLACK,rect->x+rect->w-1,rect->y,1,rect->h);		// draw right
		}
	}
}

static void ThumbDraw(GUI_IMAGE *imageBuffer,EDITOR_RECT *rect,bool highlight)
// Draw a thumb control into rect of imageBuffer.
{
	EDITOR_COLOR
		color;

	if(rect->w&&rect->h)
	{
		ImageFill(imageBuffer,BLACK,rect->x,rect->y,rect->w,rect->h);
		if(rect->w>1&&rect->h>1)
		{
			if(!highlight)
			{
				color=GRAY_3;
			}
			else
			{
				color=WHITE;
			}
			ImageFill(imageBuffer,color,rect->x+1,rect->y+1,rect->w-2,rect->h-2);
		}
		ImageFill(imageBuffer,WHITE,rect->x+1,rect->y+1,rect->w-2,1);
		ImageFill(imageBuffer,WHITE,rect->x+1,rect->y+1,1,rect->h-2);
	}
}

void ScrollBarDraw(SCROLL_BAR *scrollBar,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Redraw scrollBar into imageBuffer.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if(scrollBar->horizontal)
	{
		LeftArrowDraw(imageBuffer,&lowArrowRect,scrollBar->highlightLowArrow);
		RightArrowDraw(imageBuffer,&highArrowRect,scrollBar->highlightHighArrow);
	}
	else
	{
		UpArrowDraw(imageBuffer,&lowArrowRect,scrollBar->highlightLowArrow);
		DownArrowDraw(imageBuffer,&highArrowRect,scrollBar->highlightHighArrow);
	}
	PageIndicatorDraw(imageBuffer,&pageLowRect,scrollBar->horizontal,scrollBar->highlightLowPage);
	PageIndicatorDraw(imageBuffer,&pageHighRect,scrollBar->horizontal,scrollBar->highlightHighPage);
	ThumbDraw(imageBuffer,&thumbRect,scrollBar->highlightThumb);
}

static bool LowArrowTrack(SCROLL_BAR *scrollBar,XEvent *event)
// Track the low arrow of scrollBar.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;
	uint32_t
		currentTime,
		scrollTime,
		repeatTime;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if((event->xbutton.button==Button1)&&PointInRect(event->xbutton.x,event->xbutton.y,&lowArrowRect))
	{
		scrollBar->highlightLowArrow=true;
		ScrollBarInvalidate(scrollBar);
		if(scrollBar->stepLowerProc)
		{
			scrollBar->stepLowerProc(scrollBar,false,scrollBar->procParameters);
		}
		repeatTime=SCROLL_INITIAL_DELAY;
		scrollTime=PlatformMonotonicTime();
		while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
		{
			currentTime=PlatformMonotonicTime();
			if(currentTime-scrollTime>=repeatTime)
			{
				if(scrollBar->stepLowerProc)
				{
					scrollBar->stepLowerProc(scrollBar,true,scrollBar->procParameters);
				}
				scrollTime=currentTime;
				repeatTime=SCROLL_REPEAT_DELAY;
			}
		}
		scrollBar->highlightLowArrow=false;
		ScrollBarInvalidate(scrollBar);
		return(true);
	}
	return(false);
}

static bool HighArrowTrack(SCROLL_BAR *scrollBar,XEvent *event)
// Track the high arrow of scrollBar.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;
	uint32_t
		currentTime,
		scrollTime,
		repeatTime;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if((event->xbutton.button==Button1)&&PointInRect(event->xbutton.x,event->xbutton.y,&highArrowRect))
	{
		scrollBar->highlightHighArrow=true;
		ScrollBarInvalidate(scrollBar);
		if(scrollBar->stepHigherProc)
		{
			scrollBar->stepHigherProc(scrollBar,false,scrollBar->procParameters);
		}
		repeatTime=SCROLL_INITIAL_DELAY;
		scrollTime=PlatformMonotonicTime();
		while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
		{
			currentTime=PlatformMonotonicTime();
			if(currentTime-scrollTime>=repeatTime)
			{
				if(scrollBar->stepHigherProc)
				{
					scrollBar->stepHigherProc(scrollBar,true,scrollBar->procParameters);
				}
				scrollTime=currentTime;
				repeatTime=SCROLL_REPEAT_DELAY;
			}
		}
		scrollBar->highlightHighArrow=false;
		ScrollBarInvalidate(scrollBar);
		return(true);
	}
	return(false);
}

static bool LowPageTrack(SCROLL_BAR *scrollBar,XEvent *event)
// Track the low page of scrollBar.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;
	uint32_t
		currentTime,
		scrollTime,
		repeatTime;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if((event->xbutton.button==Button1)&&PointInRect(event->xbutton.x,event->xbutton.y,&pageLowRect))
	{
		scrollBar->highlightLowPage=true;
		ScrollBarInvalidate(scrollBar);
		if(scrollBar->pageLowerProc)
		{
			scrollBar->pageLowerProc(scrollBar,false,scrollBar->procParameters);
		}
		repeatTime=SCROLL_INITIAL_DELAY;
		scrollTime=PlatformMonotonicTime();
		while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
		{
			currentTime=PlatformMonotonicTime();
			if(currentTime-scrollTime>=repeatTime)
			{
				if(scrollBar->pageLowerProc)
				{
					scrollBar->pageLowerProc(scrollBar,true,scrollBar->procParameters);
				}
				scrollTime=currentTime;
				repeatTime=SCROLL_REPEAT_DELAY;
			}
		}
		scrollBar->highlightLowPage=false;
		ScrollBarInvalidate(scrollBar);
		return(true);
	}
	return(false);
}

static bool HighPageTrack(SCROLL_BAR *scrollBar,XEvent *event)
// Track the high page of scrollBar.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;
	uint32_t
		currentTime,
		scrollTime,
		repeatTime;
	int
		rootX,
		rootY,
		windowX,
		windowY;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if((event->xbutton.button==Button1)&&PointInRect(event->xbutton.x,event->xbutton.y,&pageHighRect))
	{
		scrollBar->highlightHighPage=true;
		ScrollBarInvalidate(scrollBar);
		if(scrollBar->pageHigherProc)
		{
			scrollBar->pageHigherProc(scrollBar,false,scrollBar->procParameters);
		}
		repeatTime=SCROLL_INITIAL_DELAY;
		scrollTime=PlatformMonotonicTime();
		while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
		{
			currentTime=PlatformMonotonicTime();
			if(currentTime-scrollTime>=repeatTime)
			{
				if(scrollBar->pageHigherProc)
				{
					scrollBar->pageHigherProc(scrollBar,true,scrollBar->procParameters);
				}
				scrollTime=currentTime;
				repeatTime=SCROLL_REPEAT_DELAY;
			}
		}
		scrollBar->highlightHighPage=false;
		ScrollBarInvalidate(scrollBar);
		return(true);
	}
	return(false);
}

static uint64_t ThumbValueCalculate(uint32_t currentPixel,uint32_t numPixels,uint64_t thumbMax)
// Given a pixel position within a range of pixels, return the thumbPosition within
// thumbMax.
{
	if(numPixels)
	{
		return((((uint64_t)currentPixel*thumbMax)+(uint64_t)(numPixels/2))/(uint64_t)numPixels);
	}
	else
	{
		return(0);
	}
}

static bool MetaThumbTrack(SCROLL_BAR *scrollBar,XEvent *event)
// See if a button other than Button1 is down in the scroll bar, if so
// track the thumb of scrollBar.
// While tracking the thumb of scrollBar, it is rude for the application to change
// the position of the bar on the screen,
// if this is done, the results will be unpredictable.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;
	int
		rootX,
		rootY,
		windowX,
		windowY,
		testIndex;
	uint32_t
		scrollablePixels;
	uint64_t
		lastValue,
		newValue;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if((event->xbutton.button!=Button1)&&PointInRect(event->xbutton.x,event->xbutton.y,&slideRect))
	{
		scrollBar->highlightThumb=true;
		ScrollBarInvalidate(scrollBar);
		rootX=event->xbutton.x_root;
		rootY=event->xbutton.y_root;
		windowX=event->xbutton.x;
		windowY=event->xbutton.y;
		if(scrollBar->horizontal)
		{
			scrollablePixels=slideRect.w-thumbRect.w;
			lastValue=ThumbValueCalculate(thumbRect.x-slideRect.x,scrollablePixels,scrollBar->thumbMax);
			while(StillDown(event->xbutton.window,event->xbutton.button,&rootX,&rootY,&windowX,&windowY))
			{
				testIndex=windowX-slideRect.x;				// get current pointer position
				testIndex-=thumbRect.w/2;					// add to initial offset
				if(testIndex<0)								// pin to range
				{
					testIndex=0;
				}
				if(testIndex>(int)scrollablePixels)
				{
					testIndex=scrollablePixels;
				}
				newValue=ThumbValueCalculate((uint32_t)testIndex,scrollablePixels,scrollBar->thumbMax);
				if(newValue!=lastValue)
				{
					if(scrollBar->thumbProc)
					{
						scrollBar->thumbProc(scrollBar,newValue,scrollBar->procParameters);
					}
					lastValue=newValue;
				}
			}
		}
		else
		{
			scrollablePixels=slideRect.h-thumbRect.h;
			lastValue=ThumbValueCalculate(thumbRect.y-slideRect.y,scrollablePixels,scrollBar->thumbMax);
			while(StillDown(event->xbutton.window,event->xbutton.button,&rootX,&rootY,&windowX,&windowY))
			{
				testIndex=windowY-slideRect.y;				// get current pointer position
				testIndex-=thumbRect.h/2;					// add to initial offset
				if(testIndex<0)								// pin to range
				{
					testIndex=0;
				}
				if(testIndex>(int)scrollablePixels)
				{
					testIndex=scrollablePixels;
				}
				newValue=ThumbValueCalculate((uint32_t)testIndex,scrollablePixels,scrollBar->thumbMax);
				if(newValue!=lastValue)
				{
					if(scrollBar->thumbProc)
					{
						scrollBar->thumbProc(scrollBar,newValue,scrollBar->procParameters);
					}
					lastValue=newValue;
				}
			}
		}
		scrollBar->highlightThumb=false;
		ScrollBarInvalidate(scrollBar);
		return(true);
	}
	return(false);
}

static bool ThumbTrack(SCROLL_BAR *scrollBar,XEvent *event)
// Track the thumb of scrollBar.
// While tracking the thumb of scrollBar, it is rude for the application to change
// the position of the bar on the screen.
// If this is done, the results will be unpredictable.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;
	int
		rootX,
		rootY,
		windowX,
		windowY,
		testIndex;
	uint32_t
		scrollablePixels,
		initialOffset;
	uint64_t
		lastValue,
		newValue;

	ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);
	if((event->xbutton.button==Button1)&&PointInRect(event->xbutton.x,event->xbutton.y,&thumbRect))
	{
		scrollBar->highlightThumb=true;
		ScrollBarInvalidate(scrollBar);
		rootX=event->xbutton.x_root;
		rootY=event->xbutton.y_root;
		windowX=event->xbutton.x;
		windowY=event->xbutton.y;
		if(scrollBar->horizontal)
		{
			scrollablePixels=slideRect.w-thumbRect.w;
			initialOffset=thumbRect.x-slideRect.x;
			lastValue=ThumbValueCalculate(initialOffset,scrollablePixels,scrollBar->thumbMax);
			while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
			{
				testIndex=windowX-event->xbutton.x;			// get current pointer position
				testIndex+=(int32_t)initialOffset;			// add to initial offset
				if(testIndex<0)								// pin to range
				{
					testIndex=0;
				}
				if(testIndex>(int)scrollablePixels)
				{
					testIndex=scrollablePixels;
				}
				newValue=ThumbValueCalculate((uint32_t)testIndex,scrollablePixels,scrollBar->thumbMax);
				if(newValue!=lastValue)
				{
					if(scrollBar->thumbProc)
					{
						scrollBar->thumbProc(scrollBar,newValue,scrollBar->procParameters);
					}
					lastValue=newValue;
				}
			}
		}
		else
		{
			scrollablePixels=slideRect.h-thumbRect.h;
			initialOffset=thumbRect.y-slideRect.y;
			lastValue=ThumbValueCalculate(initialOffset,scrollablePixels,scrollBar->thumbMax);
			while(StillDown(event->xbutton.window,Button1,&rootX,&rootY,&windowX,&windowY))
			{
				testIndex=windowY-event->xbutton.y;			// get current pointer position
				testIndex+=(int32_t)initialOffset;			// add to initial offset
				if(testIndex<0)								// pin to range
				{
					testIndex=0;
				}
				if(testIndex>(int)scrollablePixels)
				{
					testIndex=scrollablePixels;
				}
				newValue=ThumbValueCalculate((uint32_t)testIndex,scrollablePixels,scrollBar->thumbMax);
				if(newValue!=lastValue)
				{
					if(scrollBar->thumbProc)
					{
						scrollBar->thumbProc(scrollBar,newValue,scrollBar->procParameters);
					}
					lastValue=newValue;
				}
			}
		}
		scrollBar->highlightThumb=false;
		ScrollBarInvalidate(scrollBar);
		return(true);
	}
	return(false);
}

bool ScrollBarTrack(SCROLL_BAR *scrollBar,XEvent *event)
// A button was pressed, see if it was in scrollBar, if so, track it, and return true.
// If it was not within the scroll bar, return false.
{
	if(!LowArrowTrack(scrollBar,event))
	{
		if(!HighArrowTrack(scrollBar,event))
		{
			if(!MetaThumbTrack(scrollBar,event))
			{
				if(!LowPageTrack(scrollBar,event))
				{
					if(!HighPageTrack(scrollBar,event))
					{
						if(!ThumbTrack(scrollBar,event))
						{
							return(false);
						}
					}
				}
			}
		}
	}
	return(true);
}

void ScrollBarReposition(SCROLL_BAR *scrollBar,int32_t x,int32_t y,uint32_t length,bool horizontal)
// Change the position of scrollBar within its parent.
// Be nice, and only invalidate it if it actually changes.
{
	if((scrollBar->x!=x)||(scrollBar->y!=y)||(scrollBar->length!=length)||(scrollBar->horizontal!=horizontal))
	{
		ScrollBarInvalidate(scrollBar);					// invalidate where it is now
		scrollBar->x=x;
		scrollBar->y=y;
		scrollBar->length=length;
		scrollBar->horizontal=horizontal;
		ScrollBarInvalidate(scrollBar);					// invalidate where it is going
	}
}

void ScrollBarAdjust(SCROLL_BAR *scrollBar,uint64_t thumbMax,uint64_t thumbPosition)
// Adjust the thumbMax, and position of scrollBar.
// Be nice, and only invalidate it if it actually changes.
{
	EDITOR_RECT
		lowArrowRect,
		highArrowRect,
		thumbRect,
		slideRect,
		pageLowRect,
		pageHighRect;

	if((scrollBar->thumbMax!=thumbMax)||(scrollBar->thumbPosition!=thumbPosition))
	{
		ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);	// invalidate where the thumb was
		WindowRectInvalidate(scrollBar->parentWindow,&pageLowRect);
		WindowRectInvalidate(scrollBar->parentWindow,&thumbRect);
		WindowRectInvalidate(scrollBar->parentWindow,&pageHighRect);
		scrollBar->thumbMax=thumbMax;
		scrollBar->thumbPosition=thumbPosition;
		ScrollBarControlRectsGet(scrollBar,&lowArrowRect,&highArrowRect,&thumbRect,&slideRect,&pageLowRect,&pageHighRect);	// invalidate where the thumb is going
		WindowRectInvalidate(scrollBar->parentWindow,&pageLowRect);
		WindowRectInvalidate(scrollBar->parentWindow,&thumbRect);
		WindowRectInvalidate(scrollBar->parentWindow,&pageHighRect);
	}
}

void ScrollBarDispose(SCROLL_BAR *scrollBar)
// Unlink scrollBar from its parent window, and delete it.
// Invalidate the area of the window where the scroll bar was.
{
	ScrollBarInvalidate(scrollBar);					// invalidate where it is now
	MDisposePtr(scrollBar);
}

SCROLL_BAR *ScrollBarCreate(EDITOR_WINDOW *window,SCROLL_BAR_DESCRIPTOR *descriptor)
// Create a scroll bar in window, invalidate the area of window
// where the scroll bar is to be placed.
// Return a pointer to it if successful.
// ErrorSet, return NULL if not.
{
	SCROLL_BAR
		*scrollBar;

	if((scrollBar=(SCROLL_BAR *)MNewPtr(sizeof(SCROLL_BAR))))		// create scroll bar data structure
	{
		scrollBar->parentWindow=window;
		scrollBar->x=descriptor->x;
		scrollBar->y=descriptor->y;
		scrollBar->length=descriptor->length;
		scrollBar->horizontal=descriptor->horizontal;
		scrollBar->thumbMax=descriptor->thumbMax;
		scrollBar->thumbPosition=descriptor->thumbPosition;

		scrollBar->highlightLowArrow=false;
		scrollBar->highlightHighArrow=false;
		scrollBar->highlightLowPage=false;
		scrollBar->highlightHighPage=false;
		scrollBar->highlightThumb=false;

		scrollBar->thumbProc=descriptor->thumbProc;
		scrollBar->stepLowerProc=descriptor->stepLowerProc;
		scrollBar->stepHigherProc=descriptor->stepHigherProc;
		scrollBar->pageLowerProc=descriptor->pageLowerProc;
		scrollBar->pageHigherProc=descriptor->pageHigherProc;
		scrollBar->procParameters=descriptor->procParameters;
		ScrollBarInvalidate(scrollBar);
	}
	return(scrollBar);
}
