// Separator handling (draw separation lines for dialogs)
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct separator
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;										// function which is used to position this item in the dialog
	bool
		horizontal;												// true if separator is horizontal
	int32_t
		x,
		y;
	uint32_t
		length;
} SEPARATOR;

static void SeparatorInvalidate(SEPARATOR *separator)
// Invalidate separator within its parent window.
{
	EDITOR_RECT
		rect;

	rect.x=separator->x;
	rect.y=separator->y;

	if(separator->horizontal)
	{
		rect.w=separator->length;
		rect.h=2;
	}
	else
	{
		rect.w=2;
		rect.h=separator->length;
	}
	WindowRectInvalidate(DialogWindowGet(separator->dialogItem.dialog),&rect);
}

static void SeparatorSizeSet(SEPARATOR *separator)
// Work out the size given the current window.
{
	EDITOR_RECT
		rect;

	if(separator->positionFunction)	// see if non-null position function specified
	{
		separator->positionFunction(&separator->dialogItem,&rect);
		separator->x=rect.x;
		separator->y=rect.y;
		separator->length=rect.w;
		SeparatorInvalidate(separator);								// invalidate it
	}
}

static void SeparatorDraw(SEPARATOR *separator,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Redraw separator into imageBuffer.
// This ignores invalidRegion, since checking and clipping against it
// it likely slower than just doing the draw.
{
	if(separator->horizontal)
	{
		ImageFill(imageBuffer,BLACK,separator->x,separator->y,separator->length,1);
		ImageFill(imageBuffer,WHITE,separator->x,separator->y+1,separator->length,1);
	}
	else
	{
		ImageFill(imageBuffer,BLACK,separator->x,separator->y,1,separator->length);
		ImageFill(imageBuffer,WHITE,separator->x+1,separator->y,1,separator->length);
	}
}

// dialog interface routines

static void SeparatorItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the separator pointed to by item.
{
	SeparatorDraw((SEPARATOR *)item,imageBuffer,invalidRegion);
}

static void SeparatorItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	SEPARATOR
		*separator;

	separator=(SEPARATOR *)item;
	SeparatorInvalidate(separator);				// invalidate where we're at currently
	SeparatorSizeSet(separator);
}

static void SeparatorDispose(SEPARATOR *separator)
// Unlink separator from its parent window, and delete it.
// Invalidate the area of the window where separator was.
{
	SeparatorInvalidate(separator);									// invalidate where it is now
	MDisposePtr(separator);
}

static void SeparatorItemDispose(DIALOG_ITEM *item)
// Dispose of the local separator structure pointed to by item.
{
	SeparatorDispose((SEPARATOR *)item);
}

static SEPARATOR *SeparatorCreate(DIALOG *dialog,SEPARATOR_DESCRIPTOR *descriptor)
// Create a separator item in dialog, invalidate the area of dialog
// where the separator is to be placed.
// Return a pointer to it if successful.
// ErrorSet, and return NULL if not.
{
	SEPARATOR
		*separator;

	if((separator=(SEPARATOR *)MNewPtr(sizeof(SEPARATOR))))			// create data structure
	{
		separator->dialogItem.dialog=dialog;							// point back at the dialog
		separator->dialogItem.disposeFunction=SeparatorItemDispose;
		separator->dialogItem.drawFunction=SeparatorItemDraw;
		separator->dialogItem.resizeFunction=SeparatorItemResize;
		separator->dialogItem.trackFunction=NULL;
		separator->dialogItem.earlyKeyFunction=NULL;
		separator->dialogItem.wantFocus=false;
		separator->dialogItem.focusChangeFunction=NULL;
		separator->dialogItem.focusKeyFunction=NULL;
		separator->dialogItem.focusPeriodicFunction=NULL;

		separator->positionFunction=descriptor->positionFunction;
		separator->horizontal=descriptor->horizontal;

		SeparatorSizeSet(separator);
		return(separator);
	}
	return(NULL);
}

DIALOG_ITEM *SeparatorItemCreate(DIALOG *dialog,void *descriptor)
// Create the separator, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)SeparatorCreate(dialog,(SEPARATOR_DESCRIPTOR *)descriptor));
}
