// Shell commands specific to X version of e93
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	TEMP_STRING_BUFFER_LENGTH	1024

enum
{
	VSBP_LEFT,
	VSBP_RIGHT,
};

static TOKEN_LIST verticalScrollBarPlacements[]=
{
	{"left",VSBP_LEFT},
	{"right",VSBP_RIGHT},
	{"",0}
};

static bool NibbleGet(char character,uint32_t *nibble)
// Interpret character as a hex nibble, and return it.
// If there is a problem return false.
{
	if(character>='0'&&character<='9')
	{
		*nibble=character-'0';
		return(true);
	}
	else
	{
		if(character>='A'&&character<='F')
		{
			*nibble=10+character-'A';
			return(true);
		}
		else
		{
			if(character>='a'&&character<='f')
			{
				*nibble=10+character-'a';
				return(true);
			}
		}
	}
	return(false);
}

static bool ColorGet(Tcl_Interp *localInterpreter,Tcl_Obj *const object,EDITOR_COLOR *color)
// Convert object to a color value.
// Return false if the conversion failed, and fill the Tcl result.
{
	uint32_t
		index,
		nibble;
	char
		character;
	bool
		fail;
	char
		*string;

	fail=false;
	string=Tcl_GetString(object);
	if(!PlatformColorNameToColor(string,color))
	{
		*color=0;
		index=0;
		while(!fail&&(character=string[index]))
		{
			if(NibbleGet(character,&nibble))
			{
				(*color)<<=4;
				(*color)|=nibble;
				index++;
			}
			else
			{
				fail=true;
			}
		}
		if(!index)					// no characters in the input string
		{
			fail=true;
		}
	}
	if(fail)
	{
		Tcl_AppendResult(localInterpreter,"Invalid color:",string,NULL);
	}
	return(!fail);
}

static bool TclInt32Get(Tcl_Interp *localInterpreter,Tcl_Obj *const object,int32_t *number)
// convert object to a int32_t.
// Return false if the conversion failed, and fill the Tcl result.
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(int32_t)longValue;
		return(true);
	}
	return(false);
}

static int Cmd_ScrollBarPlacement(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Specify preferred placement for vertical scroll bars.
{
	int
		placement;

	if(objc==2)
	{
		if(TokenMatch(Tcl_GetString(objv[1]),verticalScrollBarPlacements,&placement))
		{
			verticalScrollBarOnLeft=(placement==VSBP_LEFT);	// set global preference
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Invalid placement",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"placement");
	}
	return(TCL_ERROR);
}

static int Cmd_CommandKeyGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Report the modifier key which is currently mapped to the "command" function.
{
	char
		modifierName[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		KeyBindModifierIndexToName(KeyCommandModifierGet(),modifierName,sizeof(modifierName));
		Tcl_AppendResult(localInterpreter,modifierName,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_CommandKeySet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the modifier key which is currently mapped to the "command" function.
{
	uint32_t
		modifierIndex;

	if(objc==2)
	{
		ErrorTraceClear();
		if(KeyBindModifierNameToIndex(Tcl_GetString(objv[1]),&modifierIndex))
		{
			KeyCommandModifierSet(modifierIndex);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"modifierName");
	}
	return(TCL_ERROR);
}

static int Cmd_RaiseRootMenu(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Raise the root menu to the top window on the screen.
{
	if(objc==1)
	{
		MenuRootRaise();
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_MoveRootMenu(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Move the root menu to a new location on the screen.
{
	int32_t
		x,
		y;

	if(objc==3)
	{
		if(TclInt32Get(localInterpreter,objv[1],&x))
		{
			if(TclInt32Get(localInterpreter,objv[2],&y))
			{
				MenuRootMove(x,y);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"x y");
	}
	return(TCL_ERROR);
}

static int Cmd_MenuFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Report the font currently in use by the menus.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,MenuFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_MenuFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set font to use for menus and update the menus.
{
	if(objc==2)
	{
		MenuFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_MenuColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by the menus.
{
	EDITOR_COLOR
		backgroundColor,
		foregroundColor,
		keyColor,
		shadowTopColor,
		shadowBottomColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		MenuColorsGet(&backgroundColor,&foregroundColor,&keyColor,&shadowTopColor,&shadowBottomColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)backgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)foregroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)keyColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)shadowTopColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X",(int)shadowBottomColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_MenuColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set colors to use for menus and update the menus.
{
	EDITOR_COLOR
		backgroundColor,
		foregroundColor,
		keyColor,
		shadowTopColor,
		shadowBottomColor;

	if(objc==6)
	{
		if(ColorGet(localInterpreter,objv[1],&backgroundColor))
		{
			if(ColorGet(localInterpreter,objv[2],&foregroundColor))
			{
				if(ColorGet(localInterpreter,objv[3],&keyColor))
				{
					if(ColorGet(localInterpreter,objv[4],&shadowTopColor))
					{
						if(ColorGet(localInterpreter,objv[5],&shadowBottomColor))
						{
							MenuColorsSet(backgroundColor,foregroundColor,keyColor,shadowTopColor,shadowBottomColor);
							return(TCL_OK);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"backgroundColor foregroundColor keyColor shadowTopColor shadowBottomColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog boxes.
{
	EDITOR_COLOR
		backgroundColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogColorsGet(&backgroundColor);

		snprintf(tempString,sizeof(tempString),"%08X",(int)backgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog boxes.
{
	EDITOR_COLOR
		backgroundColor;

	if(objc==2)
	{
		if(ColorGet(localInterpreter,objv[1],&backgroundColor))
		{
			DialogColorsSet(backgroundColor);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"backgroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogStaticTextColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog static text items.
{
	EDITOR_COLOR
		foregroundColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogStaticTextColorsGet(&foregroundColor);

		snprintf(tempString,sizeof(tempString),"%08X",(int)foregroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogStaticTextColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog static text items.
{
	EDITOR_COLOR
		foregroundColor;

	if(objc==2)
	{
		if(ColorGet(localInterpreter,objv[1],&foregroundColor))
		{
			DialogStaticTextColorsSet(foregroundColor);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"foregroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogTextBoxColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog text box items.
{
	EDITOR_COLOR
		activeBorderColor,
		inactiveBorderColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogTextBoxColorsGet(&activeBorderColor,&inactiveBorderColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeBorderColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X",(int)inactiveBorderColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogTextBoxColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog text box items.
{
	EDITOR_COLOR
		activeBorderColor,
		inactiveBorderColor;

	if(objc==3)
	{
		if(ColorGet(localInterpreter,objv[1],&activeBorderColor))
		{
			if(ColorGet(localInterpreter,objv[2],&inactiveBorderColor))
			{
					DialogTextBoxColorsSet(activeBorderColor,inactiveBorderColor);
					return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"activeBorderColor inactiveBorderColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogListBoxColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog list box items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		activeBorderColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		inactiveBorderColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogListBoxColorsGet(&activeBackgroundColor,&activeForegroundColor,&activeBorderColor,&inactiveBackgroundColor,&inactiveForegroundColor,&inactiveBorderColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeBorderColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X",(int)inactiveBorderColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogListBoxColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog list box items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		activeBorderColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		inactiveBorderColor;

	if(objc==7)
	{
		if(ColorGet(localInterpreter,objv[1],&activeBackgroundColor))
		{
			if(ColorGet(localInterpreter,objv[2],&activeForegroundColor))
			{
				if(ColorGet(localInterpreter,objv[3],&activeBorderColor))
				{
					if(ColorGet(localInterpreter,objv[4],&inactiveBackgroundColor))
					{
						if(ColorGet(localInterpreter,objv[5],&inactiveForegroundColor))
						{
							if(ColorGet(localInterpreter,objv[6],&inactiveBorderColor))
							{
								DialogListBoxColorsSet(activeBackgroundColor,activeForegroundColor,activeBorderColor,inactiveBackgroundColor,inactiveForegroundColor,inactiveBorderColor);
								return(TCL_OK);
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"activeBackgroundColor activeForegroundColor activeBorderColor inactiveBackgroundColor inactiveForegroundColor inactiveBorderColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogCheckBoxColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog check box items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogCheckBoxColorsGet(&activeBackgroundColor,&activeForegroundColor,&inactiveBackgroundColor,&inactiveForegroundColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X",(int)inactiveForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogCheckBoxColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog check box items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor;

	if(objc==5)
	{
		if(ColorGet(localInterpreter,objv[1],&activeBackgroundColor))
		{
			if(ColorGet(localInterpreter,objv[2],&activeForegroundColor))
			{
				if(ColorGet(localInterpreter,objv[3],&inactiveBackgroundColor))
				{
					if(ColorGet(localInterpreter,objv[4],&inactiveForegroundColor))
					{
						DialogCheckBoxColorsSet(activeBackgroundColor,activeForegroundColor,inactiveBackgroundColor,inactiveForegroundColor);
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"activeBackgroundColor activeForegroundColor inactiveBackgroundColor inactiveForegroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogPopupColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog popup items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		borderColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogPopupColorsGet(&activeBackgroundColor,&activeForegroundColor,&inactiveBackgroundColor,&inactiveForegroundColor,&borderColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X",(int)borderColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogPopupColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog popup items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		borderColor;

	if(objc==6)
	{
		if(ColorGet(localInterpreter,objv[1],&activeBackgroundColor))
		{
			if(ColorGet(localInterpreter,objv[2],&activeForegroundColor))
			{
				if(ColorGet(localInterpreter,objv[3],&inactiveBackgroundColor))
				{
					if(ColorGet(localInterpreter,objv[4],&inactiveForegroundColor))
					{
						if(ColorGet(localInterpreter,objv[5],&borderColor))
						{
							DialogPopupColorsSet(activeBackgroundColor,activeForegroundColor,inactiveBackgroundColor,inactiveForegroundColor,borderColor);
							return(TCL_OK);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"activeBackgroundColor activeForegroundColor inactiveBackgroundColor inactiveForegroundColor borderColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogButtonColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors used by dialog button items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		pressedBackgroundColor,
		pressedForegroundColor,
		shadowColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DialogButtonColorsGet(&activeBackgroundColor,&activeForegroundColor,&inactiveBackgroundColor,&inactiveForegroundColor,&pressedBackgroundColor,&pressedForegroundColor,&shadowColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)activeForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)inactiveForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)pressedBackgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)pressedForegroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X",(int)shadowColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogButtonColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors used by dialog button items.
{
	EDITOR_COLOR
		activeBackgroundColor,
		activeForegroundColor,
		inactiveBackgroundColor,
		inactiveForegroundColor,
		pressedBackgroundColor,
		pressedForegroundColor,
		shadowColor;

	if(objc==8)
	{
		if(ColorGet(localInterpreter,objv[1],&activeBackgroundColor))
		{
			if(ColorGet(localInterpreter,objv[2],&activeForegroundColor))
			{
				if(ColorGet(localInterpreter,objv[3],&inactiveBackgroundColor))
				{
					if(ColorGet(localInterpreter,objv[4],&inactiveForegroundColor))
					{
						if(ColorGet(localInterpreter,objv[5],&pressedBackgroundColor))
						{
							if(ColorGet(localInterpreter,objv[6],&pressedForegroundColor))
							{
								if(ColorGet(localInterpreter,objv[7],&shadowColor))
								{
									DialogButtonColorsSet(activeBackgroundColor,activeForegroundColor,inactiveBackgroundColor,inactiveForegroundColor,pressedBackgroundColor,pressedForegroundColor,shadowColor);
									return(TCL_OK);
								}
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"activeBackgroundColor activeForegroundColor inactiveBackgroundColor inactiveForegroundColor pressedBackgroundColor pressedForegroundColor shadowColor");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogStaticTextFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the font used for static text items.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DialogStaticTextFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogStaticTextFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the font used for static text items.
{
	if(objc==2)
	{
		DialogStaticTextFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogTextBoxFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the font used for dialog text boxes.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DialogTextBoxFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogTextBoxFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the font used for dialog text boxes.
{
	if(objc==2)
	{
		DialogTextBoxFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogListBoxFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the font used for dialog list boxes.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DialogListBoxFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogListBoxFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the font used for dialog list boxes.
{
	if(objc==2)
	{
		DialogListBoxFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogCheckBoxFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get font to use for dialog check boxes.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DialogCheckBoxFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogCheckBoxFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set font to use for dialog check boxes.
{
	if(objc==2)
	{
		DialogCheckBoxFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogPopupFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get font to use for dialog popups.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DialogPopupFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogPopupFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set font to use for dialog popups.
{
	if(objc==2)
	{
		DialogPopupFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_DialogButtonFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get font to use for dialog buttons.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DialogButtonFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogButtonFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set font to use for dialog buttons.
{
	if(objc==2)
	{
		DialogButtonFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}


static int Cmd_StatusBarFontGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set font to use for document window status bars.
{
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,DocumentWindowStatusBarFontGet(),NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_StatusBarFontSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set font to use for document window status bars.
{
	if(objc==2)
	{
		DocumentWindowStatusBarFontSet(Tcl_GetString(objv[1]));
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"font");
	}
	return(TCL_ERROR);
}

static int Cmd_StatusBarColorsGet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get colors to use for the status bar.
{
	EDITOR_COLOR
		backgroundColor,
		foregroundColor,
		flagColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		DocumentWindowStatusBarColorsGet(&backgroundColor,&foregroundColor,&flagColor);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)backgroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)foregroundColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);

		snprintf(tempString,sizeof(tempString),"%08X ",(int)flagColor);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_StatusBarColorsSet(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set colors to use for the status bar.
{
	EDITOR_COLOR
		backgroundColor,
		foregroundColor,
		flagColor;

	if(objc==4)
	{
		if(ColorGet(localInterpreter,objv[1],&backgroundColor))
		{
			if(ColorGet(localInterpreter,objv[2],&foregroundColor))
			{
				if(ColorGet(localInterpreter,objv[3],&flagColor))
				{
					DocumentWindowStatusBarColorsSet(backgroundColor,foregroundColor,flagColor);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"backgroundColor foregroundColor flagColor");
	}
	return(TCL_ERROR);
}

static int Cmd_FallbackFonts(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Provide the family names of fonts used for fallback when glyphs cannot be located in the currently chosen font.
{
	int
		i;

	FontsFallbackListClear();			// clear the current list
	for(i=1;i<objc;i++)
	{
		FontsFallbackListAppend(Tcl_GetString(objv[i]));	// put these on the list
	}
	return(TCL_OK);
}

static int Cmd_FontList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return a list of fonts.
// The list is guaranteed to be sorted, and all items will be unique.
{
	char
		**fontList;
	uint32_t
		numItems,
		i;
	bool
		fail;

	fail=false;
	if(objc==1)
	{
		if((fontList=FontListCreate(&numItems)))
		{
			for(i=0;i<numItems;i++)
			{
				Tcl_AppendElement(localInterpreter,fontList[i]);
			}
			FontListDispose(fontList);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

// routines callable by the editor

bool PlatformShellCommandsAdd(Tcl_Interp *interpreter)
// If a given environment desires additional shell commands
// it can add them here.
// If there is a problem, ErrorSet, and return false.
{
	Tcl_CreateObjCommand(interpreter,"scrollbarplacement",Cmd_ScrollBarPlacement,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"commandkeyget",Cmd_CommandKeyGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"commandkeyset",Cmd_CommandKeySet,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"raiserootmenu",Cmd_RaiseRootMenu,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"moverootmenu",Cmd_MoveRootMenu,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"menufontget",Cmd_MenuFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"menufontset",Cmd_MenuFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"menucolorsget",Cmd_MenuColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"menucolorsset",Cmd_MenuColorsSet,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"dialogcolorsget",Cmd_DialogColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogcolorsset",Cmd_DialogColorsSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogstatictextcolorsget",Cmd_DialogStaticTextColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogstatictextcolorsset",Cmd_DialogStaticTextColorsSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogtextboxcolorsget",Cmd_DialogTextBoxColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogtextboxcolorsset",Cmd_DialogTextBoxColorsSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialoglistboxcolorsget",Cmd_DialogListBoxColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialoglistboxcolorsset",Cmd_DialogListBoxColorsSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogcheckboxcolorsget",Cmd_DialogCheckBoxColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogcheckboxcolorsset",Cmd_DialogCheckBoxColorsSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogpopupcolorsget",Cmd_DialogPopupColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogpopupcolorsset",Cmd_DialogPopupColorsSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogbuttoncolorsget",Cmd_DialogButtonColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogbuttoncolorsset",Cmd_DialogButtonColorsSet,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"dialogstatictextfontget",Cmd_DialogStaticTextFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogstatictextfontset",Cmd_DialogStaticTextFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogtextboxfontget",Cmd_DialogTextBoxFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogtextboxfontset",Cmd_DialogTextBoxFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialoglistboxfontget",Cmd_DialogListBoxFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialoglistboxfontset",Cmd_DialogListBoxFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogcheckboxfontget",Cmd_DialogCheckBoxFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogcheckboxfontset",Cmd_DialogCheckBoxFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogpopupfontget",Cmd_DialogPopupFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogpopupfontset",Cmd_DialogPopupFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogbuttonfontget",Cmd_DialogButtonFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"dialogbuttonfontset",Cmd_DialogButtonFontSet,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"statusbarfontget",Cmd_StatusBarFontGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"statusbarfontset",Cmd_StatusBarFontSet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"statusbarcolorsget",Cmd_StatusBarColorsGet,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"statusbarcolorsset",Cmd_StatusBarColorsSet,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"fallbackfonts",Cmd_FallbackFonts,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"fontlist",Cmd_FontList,NULL,NULL);
	return(true);
}





