// Static text handler for dialogs
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct staticText
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;										// function which is used to position this item in the dialog
	EDITOR_COLOR
		foregroundColor;										// color to use when drawing foreground
	EDITOR_RECT
		rect;													// item's rectangle within the dialog
	EDITOR_FONT
		*font;													// info about currently selected font
	unsigned int
		stringBufferLength;										// tells the length of the string buffer (may be longer than the string that's in the buffer)
	char
		*string;												// pointer to static text string
} STATIC_TEXT;

static void StaticTextInvalidate(STATIC_TEXT *text)
// Invalidate text within its parent window.
{
	WindowRectInvalidate(DialogWindowGet(text->dialogItem.dialog),&text->rect);
}

static void StaticTextRectangleSet(STATIC_TEXT *text)
// Work out the rectangle given the current window.
{
	if(text->positionFunction)									// see if non-null position function specified
	{
		text->positionFunction(&text->dialogItem,&text->rect);	// get new rectangle
		StaticTextInvalidate(text);								// invalidate it
	}
}

static const char *StaticTextNextBreakPoint(EDITOR_FONT *font,uint32_t width,const char *string)
// Passed a font, a width, and pointer to a UTF-8 string, move the pointer forward to point
// to the next place to break the string. When the end of the string is reached, the
// return pointer will point to the \0 termination.
// NOTE: unless passed a pointer to a \0 terminator, this will always skip at least one character.
{
	uint32_t
		currentWidth;
	uint32_t
		index;
	uint32_t
		bestResult;
	bool
		done;

	index=0;
	bestResult=0;
	done=false;
	while(!done)
	{
		if(string[index])
		{
			if(string[index]!='\n')
			{
				currentWidth=FontTextUtf8Width(font,string,index+1);	// get the width (inefficiently, but handles UTF-8)
				if(currentWidth<width)
				{
					if((const unsigned char)string[index]<=' '||string[index]=='/')	// good place to break?
					{
						index++;					// step over the character
						bestResult=index;			// remember this as the best break point we saw so far
					}
					else
					{
						index++;					// step over the character
					}
				}
				else								// passed the width
				{
					if(bestResult)					// had a good place to break?
					{
						index=bestResult;			// return this as the break point
					}
					else							// if no best result, there was no good break point, just force one
					{
						if(index>1)					// make sure we end up using at least 1 character
						{
							index--;				// step back to the point where we didn't go over the edge
						}
					}
					done=true;						// done
				}
			}
			else									// saw EOL
			{
				index++;							// skip over the newline
				done=true;							// done
			}
		}
		else
		{
			done=true;
		}
	}
	return(&string[index]);
}

static void StaticTextDraw(STATIC_TEXT *text,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Redraw text in its parent window, obey the invalid area of the parent's window.
// This handles wrap-around of text (whitespace), and paths (/'s).
{
	uint32_t
		ascent,
		descent;
	const char
		*stringPointer,
		*nextStringPointer;
	uint32_t
		length;
	int
		stringX,
		stringY;

	if(XRectInRegion(invalidRegion,text->rect.x,text->rect.y,text->rect.w,text->rect.h)!=RectangleOut)
	{
		FontHeightsGet(text->font,&ascent,&descent);

		stringX=text->rect.x;
		stringY=text->rect.y+ascent;

		stringPointer=text->string;
		while((*stringPointer)&&(stringY<text->rect.y+(int)text->rect.h))	// stop drawing when we get below the bottom edge
		{
			nextStringPointer=StaticTextNextBreakPoint(text->font,text->rect.w,stringPointer);
			length=nextStringPointer-stringPointer;		// work out how many characters to draw
			if(length&&stringPointer[length-1]=='\n')	// don't draw newlines
			{
				length--;
			}
			FontTextUtf8Draw(text->font,imageBuffer,text->foregroundColor,stringX,stringY,stringPointer,length);
			stringY+=ascent+descent;
			stringPointer=nextStringPointer;
		}
	}
}

static void StaticTextReset(STATIC_TEXT *text,const char *newString)
// Place a new static text string into text.
// NOTE: if the memory allocation fails, the string will remain unchanged.
{
	unsigned int
		newBufferLength;
	char
		*newBuffer;

	newBufferLength=strlen(newString)+1;				// get new size for the string buffer
	if(newBufferLength>text->stringBufferLength)		// need to resize the buffer?
	{
		if((newBuffer=(char *)MResizePtr((void *)text->string,newBufferLength)))	// success?
		{
			text->stringBufferLength=newBufferLength;	// update the size
			text->string=newBuffer;						// update the pointer
		}
	}

	if(newBufferLength<=text->stringBufferLength)		// room to copy the string?
	{
		strcpy(text->string,newString);					// copy over the passed string
		StaticTextInvalidate(text);
	}
}

// dialog interface routines

static void StaticTextItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the static text pointed to by item.
{
	StaticTextDraw((STATIC_TEXT *)item,imageBuffer,invalidRegion);
}

static void StaticTextItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	STATIC_TEXT
		*text;

	text=(STATIC_TEXT *)item;
	StaticTextInvalidate(text);				// invalidate where we're at currently
	StaticTextRectangleSet(text);
}

void StaticTextItemTextReset(DIALOG_ITEM *item,char *newString)
// Reset the text of item to newString, invalidate the item.
{
	StaticTextReset((STATIC_TEXT *)item,newString);
}

void StaticTextItemRectangleGet(DIALOG_ITEM *item,EDITOR_RECT *rect)
// Return the current rectanlge of the static text item.
{
	*rect=((STATIC_TEXT *)item)->rect;
}

uint32_t StaticTextItemHeightGet(DIALOG_ITEM *item,uint32_t width)
// Return the height in pixels needed to draw item within the passed width.
{
	STATIC_TEXT
		*text;
	uint32_t
		numLines;
	uint32_t
		ascent,
		descent;
	const char
		*stringPointer;

	text=(STATIC_TEXT *)item;
	numLines=0;
	stringPointer=text->string;
	while(*stringPointer)
	{
		stringPointer=StaticTextNextBreakPoint(text->font,width,stringPointer);
		numLines++;
	}

	FontHeightsGet(text->font,&ascent,&descent);
	return(numLines*(ascent+descent));
}

static void StaticTextDispose(STATIC_TEXT *text)
// Unlink text from its parent window, and delete it.
// Invalidate the area of the window where text was.
{
	StaticTextInvalidate(text);							// invalidate where it is now
	PlatformFontFree(text->font);						// free the font
	MDisposePtr(text->string);							// get rid of the string
	MDisposePtr(text);
}

static void StaticTextItemDispose(DIALOG_ITEM *item)
// Dispose of the local static text structure pointed to by item.
{
	StaticTextDispose((STATIC_TEXT *)item);
}

static STATIC_TEXT *StaticTextCreate(DIALOG *dialog,STATIC_TEXT_DESCRIPTOR *descriptor)
// Create a static text item in dialog, invalidate the area of dialog
// where the text is to be placed.
// Return a pointer to it if successful.
// ErrorSet, and return NULL if not.
{
	STATIC_TEXT
		*text;

	if((text=(STATIC_TEXT *)MNewPtr(sizeof(STATIC_TEXT))))			// create static text data structure
	{
		text->dialogItem.dialog=dialog;							// point back at the dialog
		text->dialogItem.disposeFunction=StaticTextItemDispose;
		text->dialogItem.drawFunction=StaticTextItemDraw;
		text->dialogItem.resizeFunction=StaticTextItemResize;
		text->dialogItem.trackFunction=NULL;
		text->dialogItem.earlyKeyFunction=NULL;
		text->dialogItem.wantFocus=false;
		text->dialogItem.focusChangeFunction=NULL;
		text->dialogItem.focusKeyFunction=NULL;
		text->dialogItem.focusPeriodicFunction=NULL;

		text->foregroundColor=descriptor->foregroundColor;

		text->positionFunction=descriptor->positionFunction;

		text->stringBufferLength=strlen(descriptor->string)+1;
		if((text->string=(char *)MNewPtr(text->stringBufferLength)))	// allocate a buffer for the string
		{
			strcpy(text->string,descriptor->string);					// copy it over
			if((text->font=PlatformFontLoad(descriptor->fontName)))
			{
				StaticTextRectangleSet(text);
				return(text);
			}
			MDisposePtr(text->string);
		}
		MDisposePtr(text);
	}
	return(NULL);
}

DIALOG_ITEM *StaticTextItemCreate(DIALOG *dialog,void *descriptor)
// Create the static text, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)StaticTextCreate(dialog,(STATIC_TEXT_DESCRIPTOR *)descriptor));
}
