// String list functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

void StringListFree(char **list)
// Free a list of strings created by calling StringListNew.
{
	uint32_t
		index;

	index=0;
	while(list&&list[index])
	{
		MDisposePtr(list[index]);
		index++;
	}
	MDisposePtr(list);
}

static int StringsCompare(const void *i,const void *j)
// Compare two elements of the list.
{
	char
		*elementA,
		*elementB;

	elementA=*((char **)i);
	elementB=*((char **)j);

	return(strcmp(elementA,elementB));
}

uint32_t StringListRemoveDuplicates(char **list,uint32_t numElements)
// Run through a sorted list, and remove any duplicates.
// Return the new number of elements in the list.
{
	uint32_t
		i,
		outIndex;

	if(numElements)
	{
		outIndex=0;
		for(i=1;i<numElements;i++)
		{
			if(strcmp(list[i],list[outIndex])!=0)
			{
				outIndex++;
				list[outIndex]=list[i];
			}
			else
			{
				MDisposePtr(list[i]);							// free the duplicate item
			}
		}
		outIndex++;
		list[outIndex]=NULL;									// keep the list terminated with a NULL
		return(outIndex);
	}
	return(0);
}

void StringListSort(char **list,uint32_t numElements)
// Run through any elements of list, and sort them
// by calling qsort.
{
	qsort((char *)list,numElements,sizeof(char **),StringsCompare);
}

bool StringListReplace(const char *string,char **list,uint32_t element)
// Place string into list at the given position, freeing
// any string that was there.
// NOTE: element must be in range.
// If there is a problem leave the entry untouched, ErrorSet, and return false.
{
	char
		*newElement;

	if((newElement=(char *)MNewPtr(strlen(string)+1)))
	{
		strcpy(newElement,string);
		MDisposePtr(list[element]);
		list[element]=newElement;
		return(true);
	}
	return(false);
}

bool StringListAdd(const char *string,char ***list,uint32_t *numElements)
// Add string to list.
// At any point, list can be deleted by calling StringListFree.
// If there is a problem, do not add the element, ErrorSet, and return false.
{
	char
		**newList;
	uint32_t
		newLength;

	newLength=sizeof(char *)*((*numElements)+2);
	if((newList=(char **)MResizePtr(*list,newLength)))
	{
		(*list)=newList;
		if(((*list)[*numElements]=(char *)MNewPtr(strlen(string)+1)))
		{
			strcpy((*list)[(*numElements)++],string);
			(*list)[*numElements]=NULL;									// keep the list terminated with a NULL
			return(true);
		}
	}
	return(false);
}

char **StringListNew()
// Create an empty string list.
// If there is a problem, ErrorSet, return NULL
{
	char
		**list;

	if((list=(char **)MNewPtr(sizeof(char *))))
	{
		list[0]=NULL;
	}
	return(list);
}
