// Routines to allow tasks to run under the editor
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define		MAX_TASK_WRITE_TIME	(SECOND/2)		// how long we'll allow a task to go away w/o accepting our write data before complaining

// This definition sets EDITOR_TASK to the structure needed by the gui.
typedef struct editorTask
{
	pid_t
		pid;					// process id of task
	bool
		closedStdin;			// set true if EOF has been sent to task by closing our half of its input pipe
	int
		fdin[2],				// file handles used to pipe data between us and the task (if one is active)
		fdout[2];
	bool
		disconnected;			// tells if the window open to the task was closed before the task finished
	bool
		finished;				// tells if the task finished (we got a SIGCHLD for it)
	int
		status;					// gives the completion status for the task
	EDITOR_TASK
		*prev,					// links the list of tasks together
		*next;
	char
		commandLine[0];			// command given to start the task
} EDITOR_TASK;

typedef struct taskList
{
	EDITOR_TASK
		*first,
		*last;
} TASK_LIST;

enum
{
	STDIN_CLOSED,
};

static const char *errorDescriptions[]=
{
	"Task's stdin is closed",
};

static TASK_LIST
	taskList;							// list of tasks

static void UnBlockSigchld()
// Reallow SIGCHLD once critical modifications are complete.
{
	sigset_t
		set;

	sigemptyset(&set);
	sigaddset(&set,SIGCHLD);
	pthread_sigmask(SIG_UNBLOCK,&set,NULL);
}

static void BlockSigchld()
// Hold off SIGCHLD while we're modifying critical data structures.
{
	sigset_t
		set;

	sigemptyset(&set);
	sigaddset(&set,SIGCHLD);
	pthread_sigmask(SIG_BLOCK,&set,NULL);
}

static void TaskListUnlink(EDITOR_TASK *task)
// Remove task from the linked list of tasks.
{
	BlockSigchld();
	if(task->prev)
	{
		task->prev->next=task->next;
	}
	else
	{
		taskList.first=task->next;
	}
	if(task->next)
	{
		task->next->prev=task->prev;
	}
	else
	{
		taskList.last=task->prev;
	}
	UnBlockSigchld();
}

static void TaskListLinkToEnd(EDITOR_TASK *task)
// Link task to the end of the linked list of tasks.
{
	BlockSigchld();
	task->next=NULL;
	if((task->prev=taskList.last))
	{
		task->prev->next=task;			// point previous last item to this one
	}
	else								// taskList was empty, so this is first item too
	{
		taskList.first=task;
	}
	taskList.last=task;
	UnBlockSigchld();
}

static void TaskListSweep()
// Sweep any disconnected/finished tasks from the task list.
{
	EDITOR_TASK
		*nextTask,
		*task;

	task=taskList.first;
	while(task)
	{
		nextTask=task->next;
		if(task->disconnected&&task->finished)
		{
			TaskListUnlink(task);
			MDisposePtr(task);
		}
		task=nextTask;
	}
}

static void TaskUnusedParentPipesClose(EDITOR_TASK *task)
// Close the half of the pipes that are not needed on the parent side.
{
	close(task->fdin[0]);
	close(task->fdout[1]);
}

static void TaskUnusedChildPipesClose(EDITOR_TASK *task)
// Close the half of the pipes that are not needed on the child side.
{
	close(task->fdin[1]);
	close(task->fdout[0]);
}

static void TaskRemainingPipesClose(EDITOR_TASK *task)
// Once the child is gone, this is used to close any pipes that remain open.
{
	if(!task->closedStdin)						// this part of the pipe can be closed before we get here, if an EOF has been sent to the task
	{
		close(task->fdin[1]);
	}
	close(task->fdout[0]);
}

static void TaskPipesClose(EDITOR_TASK *task)
// close pipes opened by TaskPipesOpen.
{
	close(task->fdin[0]);
	close(task->fdin[1]);
	close(task->fdout[0]);
	close(task->fdout[1]);
}

static bool TaskPipesOpen(EDITOR_TASK *task)
// Create pipes that will connect us to a task.
// If there is a problem, ErrorSet, return false.
{
	int
		flags;

	if(pipe(&(task->fdin[0]))==0)
	{
		if(pipe(&(task->fdout[0]))==0)
		{
			flags=fcntl(task->fdin[1],F_GETFL,0);
			flags|=O_NDELAY;						// turn off blocking on write of task's stdin
			fcntl(task->fdin[1],F_SETFL,flags);

			flags=fcntl(task->fdout[0],F_GETFL,0);
			flags|=O_NDELAY;						// turn off blocking on read of task's stdout
			fcntl(task->fdout[0],F_SETFL,flags);

			task->closedStdin=false;				// have not closed the input to the task
			return(true);
		}
		else
		{
			ErrorStdCLibSet();
		}
		close(task->fdin[0]);
		close(task->fdin[1]);
	}
	else
	{
		ErrorStdCLibSet();
	}
	return(false);
}

// ----------------------------------------------------------------------------

int TaskOutputFileDescriptor(EDITOR_TASK *task)
// Return the file descriptor where the given task is writing its output.
{
	return(task->fdout[0]);
}

EDITOR_TASK *TaskNext(EDITOR_TASK *task)
// Return the next task in the list after task (or NULL if none exists).
// NOTE: this skips disconnected tasks.
{
	while(task->next)
	{
		task=task->next;
		if(!task->disconnected)
		{
			return(task);
		}
	}
	return(NULL);
}

EDITOR_TASK *TaskFirst()
// Return the first task in the list (or NULL if none exists).
// NOTE: this skips disconnected tasks.
{
	if(taskList.first)
	{
		if(taskList.first->disconnected)
		{
			return(TaskNext(taskList.first));
		}
		else
		{
			return(taskList.first);
		}
	}
	return(NULL);
}

// ----------------------------------------------------------------------------

bool PlatformTaskDataRead(EDITOR_TASK *task,uint8_t *buffer,uint64_t bufferLength,uint64_t *bytesRead)
// Attempt to read data that may be pending from task into buffer, for at
// most bufferLength bytes. Return *bytesRead as the number of bytes actually
// read. If this returns false, it indicates an error in the call to read.
// In this case, ErrorSet will have been called.
// NOTE: the test for errno==EWOULDBLOCK is needed, because
// read does not want to return the fact that it read 0 bytes in
// non-blocking mode.
// If it did return 0, we might think it hit EOF.
// NOTE: the task that's spawned sees its stdout as a pipe
// (non-interactive). As such it'll fully-buffer its output to stdout. If
// the task crashes, a lot of the stdout output that would have been seen
// on a terminal gets lost, since it was in the stdout buffer at the time
// of the crash. On a terminal, stdout is line buffered, so most "printf"
// statements make it out to the terminal before the crash. It might be
// useful to have e93 use a tty instead of a pipe to connect to the process,
// but this would also cause some processes like 'ls' to output in human
// format (multi-column) instead of machine-format...
{
	ssize_t
		result;

	if(((result=read(task->fdout[0],buffer,bufferLength))>=0)||(errno==EWOULDBLOCK))
	{
		if(result>0)
		{
			(*bytesRead)=result;
		}
		else
		{
			(*bytesRead)=0;
		}
		return(true);
	}
	else
	{
		ErrorStdCLibSet();
	}
	return(false);
}

bool PlatformTaskStatus(EDITOR_TASK *task,const char **terminationString)
// Check the status of the given task to see if it is still running.
// If it is, return true. If it is not, return false, with terminationString
// set to NULL if it terminated normally, or pointing to a descriptive
// string if it terminated due to a signal.
{
	if(!task->finished)
	{
		return(true);
	}
	else
	{
		(*terminationString)=NULL;
		if(WIFSIGNALED(task->status))			// if process died due to a signal, let the user know
		{
			(*terminationString)=strsignal(WTERMSIG(task->status));
		}
	}
	return(false);
}

bool PlatformTaskEofWrite(EDITOR_TASK *task)
// Send an EOF to the input of the given task.
// If there is some problem, ErrorSet, return false.
{
	if(!task->closedStdin)						// if closed already, complain
	{
		close(task->fdin[1]);					// close the side of the tasks stdin pipe that we write into, this will send an EOF to the task
		task->closedStdin=true;
		return(true);
	}
	else
	{
		ErrorSet(errorDescriptions[STDIN_CLOSED]);
	}
	return(false);
}

bool PlatformTaskDataWrite(EDITOR_TASK *task,const uint8_t *data,uint64_t numBytes)
// Send numBytes to the standard input of the given task.
// If there is a problem writing data, ErrorSet, return false.
// NOTE: this works in non-blocking mode. If the tasks's pipe
// stops accepting new data within a given timeout period, we'll
// return an error.
{
	int
		result;
	uint32_t
		writeTimer,
		currentTime;
	bool
		fail;

	fail=false;
	if(!task->closedStdin)							// make sure it has not been closed
	{
		writeTimer=PlatformMonotonicTime();
		while(numBytes&&!fail)
		{
			currentTime=PlatformMonotonicTime();
			result=write(task->fdin[1],data,numBytes);	// write data to "stdin" of task
			if(result>0)							// wrote something?
			{
				data+=result;						// push forward in the data
				numBytes-=result;					// reduce the number left to write
				writeTimer=currentTime;				// reset timeout each time the task takes some data
			}
			else if(result<0)						// error?
			{
				if(errno==EAGAIN)					// did the call just block?
				{
					if(currentTime-writeTimer<MAX_TASK_WRITE_TIME)	// if not blocked too long, keep going
					{
						EditorDoBackground(20);
					}
					else
					{
						ErrorSet("Write timeout");	// complain that the task isn't taking the data we're trying to give (don't want to sit here indefinitely if a task isn't reading stdin)
						fail=true;
					}
				}
				else
				{
					ErrorStdCLibSet();				// complain that the write failed
					fail=true;
				}
			}
		}
	}
	else
	{
		ErrorSet(errorDescriptions[STDIN_CLOSED]);
		fail=true;
	}
	return(!fail);
}

const char *PlatformTaskCommandLineGet(EDITOR_TASK *task)
// Return a pointer to the command line string for the command that started task.
{
	return(task->commandLine);
}

void PlatformTaskKill(EDITOR_TASK *task)
// Signal the given task to quit.
{
	kill(-(task->pid),SIGHUP);					// be nice, just send HUP, don't go for the total SIGKILL
}

void PlatformTaskDisconnect(EDITOR_TASK *task)
// Disconnect ourselves from the given task.
// NOTE: this does not kill the task, and it is usually desirable
// NOT to kill the task.
// Instead, this closes the pipes to the task, and then marks
// the task as disconnected in our task list. We keep the task in
// the list so that if it does finish while the editor is running,
// we'll know it was ours (and not Tcl's), and can reap it appropriately
// in SIGCHLD. NOTE: that the SIGCHLD handler cannot delete the task
// from our list after it has finished. We'll sweep dead tasks
// out of the list on the next attempt to run one.
{
	TaskRemainingPipesClose(task);
	task->disconnected=true;							// mark the task as having been disconnected so we can clean up after it if it eventually quits
}

EDITOR_TASK *PlatformTaskSpawn(const uint8_t *commandLine)
// Spawn off a task and return it if successful.
// ErrorSet, and return NULL if there was a problem.
{
	EDITOR_TASK
		*task;
	int
		forkResult;
	size_t
		commandLineLength;

	TaskListSweep();									// clean any dead tasks out of the list
	commandLineLength=strlen((const char *)commandLine);
	if((task=(EDITOR_TASK *)MNewPtr(sizeof(EDITOR_TASK)+commandLineLength+1)))	// create task record -- leave room to copy command line (and 0 termination) at the end
	{
		task->disconnected=false;						// task is not disconnected
		task->finished=false;							// task is not complete until we get a SIGCHLD for it
		strcpy(task->commandLine,(const char *)commandLine);	// copy over the command line
		malloc_trim(0);									// in Linux (at least as of 4.8.0-49), fork seems to hang if the heap is huge (been opening very large files) -- this seems to help
		if(TaskPipesOpen(task))							// open pipes to the task
		{
			if((forkResult=fork()))						// parent process gets child ID back
			{
				if(forkResult>0)
				{
					TaskUnusedParentPipesClose(task);	// close the half of the pipes which are meaningless to us
					task->pid=forkResult;				// set the process ID
					TaskListLinkToEnd(task);			// add to linked list of tasks
					return(task);
				}
				else
				{
					ErrorStdCLibSet();
				}
			}
			else										// child process gets 0 back
			{
				TaskUnusedChildPipesClose(task);		// close the half of the pipes which are meaningless to us
				setsid();								// make a process group with this as the leader
				dup2(task->fdin[0],0);					// create stdin
				dup2(task->fdout[1],1);					// create stdout
				dup2(task->fdout[1],2);					// create stderr
				execl("/bin/sh","sh","-c",commandLine,NULL);	// start the shell running in the child fork (if this works, we're done)
				_exit(-1);								// our task did not start, so bail out (use _exit to avoid flushing twice)
			}
			TaskPipesClose(task);						// failed to fork, close down pipes
		}
		MDisposePtr(task);
	}
	return(NULL);
}

// ----------------------------------------------------------------------------

static void TaskSignalHandler(int signal,siginfo_t *sigInfo,void *context)
// Signal handler called when a task of ours dies.
// NOTE: this gets signaled when any child of our task dies. This includes
// children created by Tcl. We don't want to reap Tcl's children because it
// bothers Tcl and makes it complain when waitpid fails. So, when a signal
// arrives here, we look to see if it was from a task that was started
// by this code, and only reap it in that case.
{
	EDITOR_TASK
		*task;
	int
		status;

	if((sigInfo->si_code!=CLD_STOPPED)&&(sigInfo->si_code!=CLD_CONTINUED))
	{
		task=taskList.first;
		while(task&&(task->pid!=sigInfo->si_pid))		// hunt down the task
		{
			task=task->next;
		}
		if(task)										// if found, mark it as finished, and set the status
		{
			status=0;
			waitpid(task->pid,&status,WNOHANG);			// tell the kernel to reap the task
			task->status=status;						// set the completion status
			task->finished=true;						// mark this task as finished
		}
	}
}

void TasksUnInit()
// Uninitialize the task handling code.
// This will disconnect any tasks that may still be running
// when it is called. Then it will clear the list of
// tasks.
{
	EDITOR_TASK
		*task;

	task=taskList.first;
	while(task)
	{
		if(!task->disconnected)
		{
			PlatformTaskDisconnect(task);
		}
		task=task->next;
	}
	while((task=taskList.first))
	{
		TaskListUnlink(task);
		MDisposePtr(task);
	}
}

bool TasksInit()
// Initialize the task handling code.
// This initializes the list of tasks, and sets up a signal handler
// to collect the status of tasks when they die.
{
	struct sigaction
		sigAction;

	taskList.first=taskList.last=NULL;				// no tasks currently

	memset(&sigAction,0,sizeof(sigAction));			// clear all fields (including those we may not know about)
	sigAction.sa_sigaction=TaskSignalHandler;
	sigemptyset(&sigAction.sa_mask);
	sigAction.sa_flags=SA_NOCLDSTOP|SA_RESTART|SA_SIGINFO;

	sigaction(SIGCHLD,&sigAction,NULL);				// install the handler

	return(true);
}
