// Text box handling for dialogs
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	BORDER_WIDTH			3

typedef struct textBox
{
	DIALOG_ITEM
		dialogItem;
	DIALOG_POSITION_FUNCTION
		*positionFunction;				// function which is used to position this item in the dialog
	EDITOR_RECT
		rect;							// position and size within parent window
	EDITOR_VIEW
		*view;							// the view that this text box item made
	EDITOR_COLOR
		activeBorderColor,
		inactiveBorderColor;
} TEXT_BOX;

static void TextBoxInvalidate(TEXT_BOX *textBox)
// Invalidate the text box within its parent window.
{
	WindowRectInvalidate(DialogWindowGet(textBox->dialogItem.dialog),&textBox->rect);
}

static void TextBoxControlRectsGet(TEXT_BOX *textBox,EDITOR_RECT *borderRect,EDITOR_RECT *textContentRect)
// Return the rectangles that enclose the various pieces of textBox.
// If the text box starts getting too small, certain rects may be returned with 0 width or
// height.
{
	EDITOR_RECT
		insetRect;

	insetRect=*borderRect=textBox->rect;					// border rect is the whole thing

	insetRect.x+=BORDER_WIDTH;								// inset from the border to the content area of the text box
	insetRect.y+=BORDER_WIDTH;
	if(insetRect.w>2*BORDER_WIDTH)
	{
		insetRect.w-=2*BORDER_WIDTH;
	}
	else
	{
		insetRect.w=0;
	}
	if(insetRect.h>2*BORDER_WIDTH)
	{
		insetRect.h-=2*BORDER_WIDTH;
	}
	else
	{
		insetRect.h=0;
	}
	*textContentRect=insetRect;
}

static void TextBoxRectangleSet(TEXT_BOX *textBox)
// Work out the rectangle given the current window.
{
	EDITOR_RECT
		borderRect,
		textContentRect;

	if(textBox->positionFunction)			// see if non-null resize function specified
	{
		textBox->positionFunction(&textBox->dialogItem,&textBox->rect);	// get new rectangle
		TextBoxControlRectsGet(textBox,&borderRect,&textContentRect);
		ViewBoundsSet(textBox->view,&textContentRect);		// move the view over
		TextBoxInvalidate(textBox);							// invalidate it
	}
}

static bool TextBoxPointInside(TEXT_BOX *textBox,int32_t x,int32_t y)
// Return true if parent window relative x/y are in textBox
// false if not.
{
	return(PointInRect(x,y,&textBox->rect));
}

static void TextBoxDraw(TEXT_BOX *textBox,GUI_IMAGE *imageBuffer,Region invalidRegion,bool hasFocus)
// Redraw textBox into imageBuffer.
{
	EDITOR_RECT
		borderRect,
		textContentRect;

	TextBoxControlRectsGet(textBox,&borderRect,&textContentRect);
	if(XRectInRegion(invalidRegion,borderRect.x,borderRect.y,borderRect.w,borderRect.h)!=RectangleOut)
	{
		if(!hasFocus)
		{
			ShadowRectangleOutline(imageBuffer,&borderRect,textBox->inactiveBorderColor,textBox->inactiveBorderColor,BORDER_WIDTH);
		}
		else
		{
			ShadowRectangleOutline(imageBuffer,&borderRect,textBox->activeBorderColor,textBox->activeBorderColor,BORDER_WIDTH);
		}
		ViewDraw(textBox->view,imageBuffer,invalidRegion);
	}
}

static bool TextBoxTrack(TEXT_BOX *textBox,XEvent *event)
// A button was pressed, see if it was in textBox, if so, track it, and return true.
// If it was not within the text box, return false.
{
	return(ViewTrack(textBox->view,event));	// only thing tracking in the text box is the view
}

static void TextBoxKeyEventHandle(TEXT_BOX *textBox,EDITOR_KEY *editorKey)
// A keyboard event is arriving for the text box, handle it here.
{
	ViewKeyEventHandle(textBox->view,editorKey);
}

static void TextBoxActiveSet(TEXT_BOX *textBox,bool active)
// Set the "active" status of the text box (called when focus changes).
{
	TextBoxInvalidate(textBox);					// invalidate the whole thing -- border needs to be redrawn with highlight colors
	ViewActiveSet(textBox->view,active);
}

static void TextBoxPeriodicProcHandle(TEXT_BOX *textBox)
// Call this to handle periodic things a text box must do.
{
	ViewCursorTask(textBox->view);
}

// dialog interface routines

static void TextBoxItemDraw(DIALOG_ITEM *item,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the text box pointed to by item.
{
	TextBoxDraw((TEXT_BOX *)item,imageBuffer,invalidRegion,(item==DialogFocusItemGet(item->dialog)));	// draw the text box
}

static void TextBoxItemResize(DIALOG_ITEM *item)
// When the dialog window size changes, call this to update the item itself.
// This will invalidate the old and new item areas.
{
	TEXT_BOX
		*textBox;

	textBox=(TEXT_BOX *)item;
	TextBoxInvalidate(textBox);					// invalidate where we're at currently
	TextBoxRectangleSet(textBox);
}

static bool TextBoxItemTrack(DIALOG_ITEM *item,XEvent *event)
// Attempt to track the text box pointed to by item.
{
	if(TextBoxPointInside((TEXT_BOX *)item,event->xbutton.x,event->xbutton.y))
	{
		DialogFocusTake(item);
		return(TextBoxTrack((TEXT_BOX *)item,event));
	}
	return(false);
}

static void TextBoxItemFocusChange(DIALOG_ITEM *item)
// The focus is changing into, or out of this text box item, deal with it.
{
	TextBoxActiveSet((TEXT_BOX *)item,item==DialogFocusItemGet(item->dialog));
}

static bool TextBoxItemFocusKey(DIALOG_ITEM *item,EDITOR_KEY *editorKey)
// Give the key to the text box.
{
	TextBoxKeyEventHandle((TEXT_BOX *)item,editorKey);
	return(true);
}

static void TextBoxItemPeriodicFunction(DIALOG_ITEM *item)
// This is used to flash the cursor of the view.
{
	TextBoxPeriodicProcHandle((TEXT_BOX *)item);
}

static void TextBoxDispose(TEXT_BOX *textBox)
// Unlink textBox from its parent window, and delete it.
// Invalidate the area of the window where the text box was.
{
	TextBoxInvalidate(textBox);					// invalidate where it is now
	ViewDispose(textBox->view);
	MDisposePtr(textBox);
}

static void TextBoxItemDispose(DIALOG_ITEM *item)
// Dispose of the local text box structure pointed to by item.
{
	TextBoxDispose((TEXT_BOX *)item);
}

static TEXT_BOX *TextBoxCreate(DIALOG *dialog,TEXT_BOX_DESCRIPTOR *descriptor)
// Create a text box in dialog, invalidate the area of dialog
// where the text box is to be placed.
// Return a pointer to it if successful.
// ErrorSet, return NULL if not.
{
	TEXT_BOX
		*textBox;
	EDITOR_VIEW_DESCRIPTOR
		viewDescriptor;

	if((textBox=(TEXT_BOX *)MNewPtr(sizeof(TEXT_BOX))))			// create text box data structure
	{
		textBox->dialogItem.dialog=dialog;						// point back at the dialog
		textBox->dialogItem.disposeFunction=TextBoxItemDispose;
		textBox->dialogItem.drawFunction=TextBoxItemDraw;
		textBox->dialogItem.resizeFunction=TextBoxItemResize;
		textBox->dialogItem.trackFunction=TextBoxItemTrack;
		textBox->dialogItem.earlyKeyFunction=NULL;
		textBox->dialogItem.wantFocus=true;
		textBox->dialogItem.focusChangeFunction=TextBoxItemFocusChange;
		textBox->dialogItem.focusKeyFunction=TextBoxItemFocusKey;
		textBox->dialogItem.focusPeriodicFunction=TextBoxItemPeriodicFunction;		// set up cursor flash task

		textBox->positionFunction=descriptor->positionFunction;

		textBox->activeBorderColor=descriptor->activeBorderColor;					// colors needed for text box
		textBox->inactiveBorderColor=descriptor->inactiveBorderColor;

		viewDescriptor.buffer=descriptor->buffer;				// the buffer on which to view

		viewDescriptor.rect.x=0;								// initially make empty view
		viewDescriptor.rect.y=0;
		viewDescriptor.rect.w=0;
		viewDescriptor.rect.h=0;
		viewDescriptor.active=false;							// tells if the view is created active or not
		viewDescriptor.useUtf8=true;							// tells if the view is should interpret UTF8
		viewDescriptor.tabSize=descriptor->tabSize;
		viewDescriptor.fontName=descriptor->fontName;			// set up the font name
		viewDescriptor.colorSchemeName=descriptor->colorSchemeName;	// set up the color scheme
		viewDescriptor.viewTextChangedVector=NULL;
		viewDescriptor.viewSelectionChangedVector=NULL;
		viewDescriptor.viewPositionChangedVector=NULL;
		viewDescriptor.viewMetaChangedVector=NULL;

		if((textBox->view=ViewCreate(DialogWindowGet(dialog),&viewDescriptor)))
		{
			TextBoxRectangleSet(textBox);
			return(textBox);
		}
		MDisposePtr(textBox);
	}
	return(NULL);
}

DIALOG_ITEM *TextBoxItemCreate(DIALOG *dialog,void *descriptor)
// Create the textBox, and return it.
// If there is a problem, ErrorSet, return false.
{
	return((DIALOG_ITEM *)TextBoxCreate(dialog,(TEXT_BOX_DESCRIPTOR *)descriptor));
}
