// Platform specific view management
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// This code handles drawing and events within "views".
//
// A view displays the text of an editor buffer (in various styles),
// along with the cursor/selections, and marks (which are just
// represented as other selections). The cursor shape and size, as well
// as how the selections and marks and styles look are all up to the code
// here. The editor code above here does not know or care.
//
// When the high-level editor code is about to change text, styles,
// selections, or marks in a buffer (or tab size for a view), it will
// call a "start" function for the type of change (text, selection,
// etc...) make the change, and then call an "end" function to indicate
// that change is complete. The calls will cause the view to become
// invalidated in the areas of the change.
// NOTE: the calls to the various start functions are *not* nestable.
// The editor will call a single start function, then perform the change
// and call the corresponding end function before ever calling another
// start function.
//
// This also handles os-specific events in the view, converting them to
// generic editor events, and passing them back up.
//
// Some terms are used in this code which could stand to be defined:
//
// "Window Coordinate" -- a signed 64-bit X, and Y that are 0,0 at the
// top left of a window, and increment to the right and down respectively.
//
// "View Coordinate" -- a signed 64-bit X, and Y that are 0,0 at the top
// left of the view, and increment to the right and down respectively.
// NOTE the view Y coordinate can also be expressed as a 64-bit viewLine
// in some circumstances.
//
// "View Origin" -- a signed 64-bit "left pixel", and a signed 64-bit
// top line.
//
// "Text Coordinate" -- a signed 64-bit X pixel position in the text, and
// a signed 64-bit line number. X is 0 at the left of a line of text, and
// increments moving to the right. Line numbers are counted starting from
// 0 at the top of the text, and incrementing downwards. Negative values
// are allowed during conversions between coordinate systems, but
// generally don't show up otherwise.
//
// "Text Position" -- an unsigned 64-bit character index into the text
// with 0 representing the first character.
//
// There are functions in the code to convert between the various positions
// and coordinate systems.

#include	"includes.h"

#define	ACTIVE_LEFT_MARGIN		10											// margin held from left of bounds to active area
#define	ACTIVE_RIGHT_MARGIN		10											// margin held from right of bounds to active area
#define	ACTIVE_TOP_MARGIN		4											// margin held from top of bounds to active area
#define	ACTIVE_BOTTOM_MARGIN	4											// margin held from bottom of bounds to active area

#define	EXPANSION_BUFFER_SIZE	8192										// number of bytes used in the expansion buffer (must be larger than EXPANSION_MAX_ADD)
#define	EXPANSION_MAX_ADD		256											// maximum number of bytes that will ever be added to the expansion buffer in one call to GlyphExpand (tabs can expand to this size)

static const char hexConv[]="0123456789ABCDEF";								// used for speedy hex conversions (for display of non-ASCII codes)

typedef struct textExpansionState
{
	uint32_t
		*buffer;															// place to expand text into (NULL if not outputting)
	uint32_t
		bufferIndex;														// offset into the staging buffer
	EDITOR_FONT
		*font;																// font being used for the expansion
	uint32_t
		tabSize;															// number of spaces to use per tab when expanding
	uint32_t
		columnNumber;														// running column number
	uint64_t
		runningWidth;														// running width (in pixels)
} TEXT_EXPANSION_STATE;

// When text is being built for output, this buffer
// is used to hold it.
static uint32_t
	gceExpansionBuffer[EXPANSION_BUFFER_SIZE];								// text is expanded into here for more efficient output

// ----------------------------------------------------------------------------
// Text expansion (expands tabs and special characters)

static bool CodePointAddToBuffer(TEXT_EXPANSION_STATE *expansionState,uint32_t codePoint)
// Add the passed 32-bit codePoint to the staging buffer.
// The codePoint is only added if it has a non-zero width.
// If the codePoint is added, return true.
{
	uint32_t
		width;

	if(FontCodePoint32Width(expansionState->font,codePoint,&width))
	{
		if(expansionState->buffer)
		{
			expansionState->buffer[expansionState->bufferIndex++]=codePoint;
		}
		expansionState->runningWidth+=width;
		expansionState->columnNumber++;
		return(true);
	}
	return(false);
}

static void GlyphExpand(uint32_t codePoint,TEXT_EXPANSION_STATE *expansionState)
// Add the necessary glyphs to the expansion buffer to render the given codePoint.
// Update expansionState.
// NOTE: when this is called, there must always be enough room in the expansion
// buffer for the glyphs that get added.
{
	uint32_t
		i,
		numTabSpaces;
	bool
		printHex;

	printHex=false;
	// handle all control characters we care about
	if(codePoint=='\t')							// handle tabs
	{
		if(expansionState->tabSize)
		{
			numTabSpaces=expansionState->tabSize-(expansionState->columnNumber%expansionState->tabSize);
			for(i=0;i<numTabSpaces;i++)
			{
				CodePointAddToBuffer(expansionState,' ');		// space over to tab stop
			}
		}
		else
		{
			printHex=true;
		}
	}
	else
	{
		printHex=!CodePointAddToBuffer(expansionState,codePoint);
	}

	if(printHex)
	{
		CodePointAddToBuffer(expansionState,'[');
		if(codePoint>0xFFFF)
		{
			CodePointAddToBuffer(expansionState,hexConv[(codePoint>>20)&0x0F]);
			CodePointAddToBuffer(expansionState,hexConv[(codePoint>>16)&0x0F]);
		}
		if(codePoint>0xFF)
		{
			CodePointAddToBuffer(expansionState,hexConv[(codePoint>>12)&0x0F]);
			CodePointAddToBuffer(expansionState,hexConv[(codePoint>>8)&0x0F]);
		}
		CodePointAddToBuffer(expansionState,hexConv[(codePoint>>4)&0x0F]);
		CodePointAddToBuffer(expansionState,hexConv[codePoint&0x0F]);
		CodePointAddToBuffer(expansionState,']');
	}
}

static void TextExpansionStateInit(TEXT_EXPANSION_STATE *expansionState)
// Initialize a text expansion state.
{
	expansionState->buffer=NULL;
	expansionState->bufferIndex=0;
	expansionState->font=NULL;
	expansionState->tabSize=0;
	expansionState->columnNumber=0;
	expansionState->runningWidth=0;
}

// ----------------------------------------------------------------------------

static bool LineCodePointFetch(TEXT_STREAM *stream,bool useUtf8,uint64_t maxPosition,uint32_t *codePoint)
// Fetch the next code point from the line of text given by stream.
// Return false if fetching the code point would read off the end of the line,
// or if it would push stream beyond maxPosition.
// If true is returned, stream will be updated to point past the last
// byte of input used. Otherwise it will be left unmodified.
{
	bool
		result;

	result=false;
	if(((*codePoint)=TextStreamChar(stream))!='\n')		// make sure we're not at the end of a line
	{
		if(useUtf8)										// want UTF-8?
		{
			result=Utf8LimitedNextCodePoint(stream,maxPosition,codePoint);
		}
		else
		{
			if((result=(TextStreamPositionGet(stream)<maxPosition)))
			{
				TextStreamNextChar(stream);				// just use this one up
			}
		}
	}
	return(result);
}

// ----------------------------------------------------------------------------

static EDITOR_COLOR ColorsSum(EDITOR_COLOR colorA,EDITOR_COLOR colorB)
// When marks overlap, this is called to blend the colors of the overlapping
// marks together.
{
	unsigned int
		red,
		green,
		blue;

	red=((colorA>>16)&0xFF)+((colorB>>16)&0xFF);
	green=((colorA>>8)&0xFF)+((colorB>>8)&0xFF);
	blue=(colorA&0xFF)+(colorB&0xFF);
	if(red>255)
	{
		red=255;
	}
	if(green>255)
	{
		green=255;
	}
	if(blue>255)
	{
		blue=255;
	}
	return((red<<16)|(green<<8)|blue);
}

static EDITOR_FONT *FontForStyleGet(EDITOR_VIEW *view,bool haveStyle,uint32_t desiredStyle)
// Return the font that should be used for the given style.
// If there is none specified, then use the default for the view.
{
	EDITOR_FONT
		*font;

	if((!haveStyle)||!(font=ViewColorSchemeFontsStyleFontGet(view->colorSchemeFonts,desiredStyle)))	// font assigned to the style?
	{
		font=ViewColorSchemeFontsViewFontGet(view->colorSchemeFonts);
	}
	return(font);
}

static EDITOR_COLOR BackgroundForStyleGet(EDITOR_VIEW *view,bool haveStyle,uint32_t style,bool active)
// Return the background color that should be used for the given style.
// If there is none specified, then use the default.
{
	EDITOR_COLOR
		background,
		foreground;

	if(haveStyle)
	{
		ViewColorSchemeStyleDrawColorsGet(view->colorScheme,style,active,&background,&foreground);
		if(!background)
		{
			ViewColorSchemeDrawColorsGet(view->colorScheme,active,&background,&foreground);
		}
	}
	else
	{
		ViewColorSchemeDrawColorsGet(view->colorScheme,active,&background,&foreground);
	}
	return(background);
}

static EDITOR_COLOR ForegroundForStyleGet(EDITOR_VIEW *view,bool haveStyle,uint32_t style,bool active)
// Return the foreground color that should be used for the given style.
// If there is none specified, then use the default for the view.
{
	EDITOR_COLOR
		background,
		foreground;

	if(haveStyle)
	{
		ViewColorSchemeStyleDrawColorsGet(view->colorScheme,style,active,&background,&foreground);
		if(!foreground)
		{
			ViewColorSchemeDrawColorsGet(view->colorScheme,active,&background,&foreground);
		}
	}
	else
	{
		ViewColorSchemeDrawColorsGet(view->colorScheme,active,&background,&foreground);
	}
	return(foreground);
}

static bool PointInView(EDITOR_VIEW *view,int32_t x,int32_t y)
// Return true if parent window relative x/y are in view
// false if not.
{
	return(PointInRect(x,y,&(view->viewInfo->bounds)));
}

// ----------------------------------------------------------------------------
// Conversions

static void ViewCoordinateXPin(EDITOR_VIEW *view,int64_t viewX,int64_t *pinnedViewX)
// Pin the given view X coordinate to the view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	if(viewX<0)
	{
		viewX=0;
	}
	else if(viewX>viewInfo->bounds.w)
	{
		viewX=viewInfo->bounds.w;
	}
	(*pinnedViewX)=viewX;
}

static void ViewCoordinateLineToViewCoordinateY(EDITOR_VIEW *view,int64_t viewLine,int64_t *viewY)
// Convert a view line coordinate to a view Y coordinate.
// This works properly for negative viewLine.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	(*viewY)=(viewLine*viewInfo->lineHeight)+ACTIVE_TOP_MARGIN;
}

static void ViewCoordinateYToViewCoordinateLine(EDITOR_VIEW *view,int64_t viewY,int64_t *viewLine)
// Convert a view Y coordinate to a view line coordinate.
// This works properly for negative viewY.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	viewY-=ACTIVE_TOP_MARGIN;
	if(viewY<0)
	{
		viewY-=viewInfo->lineHeight;							// account for cases about 0
	}
	(*viewLine)=viewY/viewInfo->lineHeight;
}

static void ViewCoordinateXToTextCoordinateX(EDITOR_VIEW *view,int64_t viewX,int64_t *textX)
// Convert a view X coordinate to a text X coordinate.
// This works properly for negative viewX.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	(*textX)=viewX+viewInfo->leftPixel-ACTIVE_LEFT_MARGIN;
}

static void ViewCoordinateLineToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewLine,int64_t *textLine)
// Convert a view line coordinate to a text line coordinate.
// This works properly for negative viewLine.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	(*textLine)=viewLine+viewInfo->topLine;
}

static void ViewCoordinateYToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewY,int64_t *textLine)
// Convert a view Y coordinate to a text line coordinate.
// This works properly for negative viewY.
{
	int64_t
		viewLine;

	ViewCoordinateYToViewCoordinateLine(view,viewY,&viewLine);
	ViewCoordinateLineToTextCoordinateLine(view,viewLine,textLine);
}

static void ViewCoordinateToTextCoordinate(EDITOR_VIEW *view,int64_t viewX,int64_t viewY,int64_t *textX,int64_t *textLine)
// Convert a set of view coordinates into a set of text coordinates.
// This takes into account the current origin of the view.
// This works properly for negative viewX and viewY.
{
	ViewCoordinateXToTextCoordinateX(view,viewX,textX);
	ViewCoordinateYToTextCoordinateLine(view,viewY,textLine);
}

static void ViewCoordinateToWindowCoordinate(EDITOR_VIEW *view,int64_t viewX,int64_t viewY,int64_t *windowX,int64_t *windowY)
// Given a point in the view, convert it to a point in the window
// that contains the view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;									// point at the information record for this view
	(*windowX)=viewX+viewInfo->bounds.x;
	(*windowY)=viewY+viewInfo->bounds.y;
}

static void WindowCoordinateToViewCoordinate(EDITOR_VIEW *view,int64_t windowX,int64_t windowY,int64_t *viewX,int64_t *viewY)
// Given a point local to the window that contains view,
// convert it into a view position.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;									// point at the information record for this view
	(*viewX)=windowX-viewInfo->bounds.x;
	(*viewY)=windowY-viewInfo->bounds.y;
}

static void TextCoordinateXToViewCoordinateX(EDITOR_VIEW *view,int64_t textX,int64_t *viewX)
// Convert a text X coordinate into a view X coordinate.
// This takes into account the current origin of the view.
// This works properly for negative textX.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	(*viewX)=textX-viewInfo->leftPixel+ACTIVE_LEFT_MARGIN;
}

static void TextCoordinateLineToViewCoordinateY(EDITOR_VIEW *view,int64_t textLine,int64_t *viewY)
// Convert a text line coordinate into a view Y coordinate.
// This takes into account the current origin of the view.
// This works properly for negative textLine.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	ViewCoordinateLineToViewCoordinateY(view,textLine-viewInfo->topLine,viewY);
}

/* unused currently
static void TextCoordinateLineToViewCoordinateLine(EDITOR_VIEW *view,int64_t textLine,int64_t *viewLine)
// Convert a text line coordinate to a view line coordinate.
// This works properly for negative textLine.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	(*viewLine)=textLine-viewInfo->topLine;
}
*/

static void TextCoordinateToViewCoordinate(EDITOR_VIEW *view,int64_t textX,int64_t textLine,int64_t *viewX,int64_t *viewY)
// Convert a set of text coordinates into a set of view coordinates.
{
	TextCoordinateXToViewCoordinateX(view,textX,viewX);
	TextCoordinateLineToViewCoordinateY(view,textLine,viewY);
}

static bool TextStreamExpandPixels(EDITOR_VIEW *view,TEXT_STREAM *stream,TEXT_EXPANSION_STATE *expansionState,int64_t numPixels)
// Given a stream, and expansionState which describes the current
// expansion, move the stream and expansionState to the start of the
// glyph at textX.
// NOTE: this is only implemented for moves in the forward direction
// (textX>=expansionState->runningWidth).
// NOTE: if this hits the end of the line in stream, or runs off the end
// of stream, expansion stops, and false is returned.
{
	bool
		haveStyle,
		inLine,
		loopDone;
	TEXT_STREAM
		lastStream;
	uint32_t
		codePoint;
	int32_t
		lastRunningWidth;
	uint32_t
		lastColumnNumber;
	uint64_t
		stylePosition,
		numStyleElements;
	uint64_t
		maxPosition;
	uint32_t
		currentStyle;

	haveStyle=StyleRangeGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&stylePosition,&numStyleElements,&currentStyle);
	stylePosition+=numStyleElements;							// style will change at this position (if numStyleElements==0, we're at the end of the style universe)

	lastRunningWidth=expansionState->runningWidth;
	lastColumnNumber=expansionState->columnNumber;
	lastStream=*stream;

	expansionState->font=FontForStyleGet(view,haveStyle,currentStyle);
	maxPosition=TextUniverseTotalBytes(stream->universe);

	loopDone=false;
	inLine=true;

	while(!loopDone&&inLine&&!TextStreamAtEnd(stream))			// run until we get to the position, hit EOL, or hit EOF
	{
		if(numStyleElements&&(TextStreamPositionGet(stream)>=stylePosition))	// time to update style? (check numStyleElements to see if we're at the end of the style universe)
		{
			haveStyle=StyleRangeGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&stylePosition,&numStyleElements,&currentStyle);
			stylePosition+=numStyleElements;					// style will change at this position

			expansionState->font=FontForStyleGet(view,haveStyle,currentStyle);
		}

		if((inLine=LineCodePointFetch(stream,view->useUtf8,maxPosition,&codePoint)))
		{
			GlyphExpand(codePoint,expansionState);
			if((int64_t)expansionState->runningWidth>numPixels)	// see if we pushed far enough
			{
				expansionState->runningWidth=lastRunningWidth;	// step back to the previous location, and remember we're done
				expansionState->columnNumber=lastColumnNumber;
				*stream=lastStream;
				loopDone=true;
			}
			else
			{
				lastRunningWidth=expansionState->runningWidth;	// remember these
				lastColumnNumber=expansionState->columnNumber;
				lastStream=*stream;
			}
		}
	}
	return(loopDone);
}

static void TextStreamExpandCharacters(EDITOR_VIEW *view,TEXT_STREAM *stream,TEXT_EXPANSION_STATE *expansionState,uint64_t maxPosition,bool limitMax,uint64_t limitX)
// Given a stream (synchornized to the start of a codePoint), and
// expansionState which describes the current expansion, move the stream
// and expansionState forward until maxPosition.
// If limitMax is set, then stop as soon as limitX is hit/passed.
// NOTE: if maxPosition is *within* a codePoint, the stream will be
// pushed only up to the start of that codePoint.
{
	uint32_t
		codePoint;
	bool
		haveStyle;
	uint64_t
		stylePosition,
		numStyleElements;
	uint32_t
		currentStyle;

	haveStyle=StyleRangeGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&stylePosition,&numStyleElements,&currentStyle);
	stylePosition+=numStyleElements;						// style will change at this position (if numStyleElements==0, we're at the end of the style universe)

	expansionState->font=FontForStyleGet(view,haveStyle,currentStyle);

	while((TextStreamPositionGet(stream)<maxPosition)&&!TextStreamAtEnd(stream)&&(!limitMax||(expansionState->runningWidth<limitX))&&LineCodePointFetch(stream,view->useUtf8,maxPosition,&codePoint))	// run out to the given number of characters, or max position
	{
		if(numStyleElements&&TextStreamPositionGet(stream)>stylePosition)	// time to update style? (check numStyleElements to see if we're at the end of the style universe)
		{
			haveStyle=StyleRangeGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&stylePosition,&numStyleElements,&currentStyle);	// get new style
			stylePosition+=numStyleElements;				// style will change again at this position

			expansionState->font=FontForStyleGet(view,haveStyle,currentStyle);
		}
		GlyphExpand(codePoint,expansionState);
	}
}

static void TextPositionToTextCoordinate(EDITOR_VIEW *view,uint64_t position,int64_t *textX,int64_t *textLine,bool limitMax,uint64_t limitX)
// Convert a text position into an X offset, and line number.
// If limitMax is true, limitX will limit how far along the line we'll search when
// trying to locate the character.
// If the limit is hit, the search will stop at that point.
{
	TEXT_STREAM
		stream;
	TEXT_EXPANSION_STATE
		expansionState;

	TextExpansionStateInit(&expansionState);
	expansionState.tabSize=view->tabSize;

	(*textLine)=TextUniversePositionToLine(view->parentBuffer->textUniverse,position,&stream);	// point stream to the start of the line
	TextStreamExpandCharacters(view,&stream,&expansionState,position,limitMax,limitX);	// run out to the position
	(*textX)=expansionState.runningWidth;
}

// ----------------------------------------------------------------------------
// Drawing

static bool ViewCursorWindowRect(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection,EDITOR_RECT *cursorRect)
// If the cursor of selection is in the view, fill cursorRect with the rect of the
// cursor (in window coordinates).
// If the cursor is not in the view, return false.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	TEXT_STREAM
		stream;
	TEXT_EXPANSION_STATE
		expansionState;
	int64_t
		textX,
		textLine,
		viewX,
		viewY,
		windowX,
		windowY;

	viewInfo=view->viewInfo;												// point at the information record for this view

	textLine=TextUniversePositionToLine(view->parentBuffer->textUniverse,SelectionCursorPositionGet(selection),&stream);	// point stream to the start of the line the cursor is on
	if((textLine>=viewInfo->topLine)&&(textLine<viewInfo->topLine+viewInfo->numLines))	// make sure cursor is on the view vertically
	{
		TextExpansionStateInit(&expansionState);
		expansionState.tabSize=view->tabSize;

		ViewCoordinateXToTextCoordinateX(view,viewInfo->bounds.w,&textX);				// get maximum X position to hunt (stop early if cursor is to the right of the view)
		TextStreamExpandCharacters(view,&stream,&expansionState,SelectionCursorPositionGet(selection),true,textX);	// run out to the cursor, or until just past the end of the view

		textX=expansionState.runningWidth;												// get text X location of the cursor
		TextCoordinateToViewCoordinate(view,textX,textLine,&viewX,&viewY);				// convert back to view coordinates

		if((viewX>=0)&&(viewX<viewInfo->bounds.w))										// see if on view horizontally
		{
			ViewCoordinateToWindowCoordinate(view,viewX,viewY,&windowX,&windowY);		// convert to window coordinates

			cursorRect->x=windowX;														// convert view coordinates to window coordinates
			cursorRect->w=1;
			cursorRect->y=windowY;
			cursorRect->h=viewInfo->lineHeight;
			return(true);
		}
	}
	return(false);
}

static void ViewCursorDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,SELECTION_UNIVERSE *selection)
// Draw the given selection cursor wherever it happens to be in view into
// the passed pixmap constrained by the passed graphics context.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	EDITOR_RECT
		cursorRect;

	viewInfo=view->viewInfo;												// point at the information record for this view
	if((SelectionIsEmpty(selection))&&viewInfo->blinkState&&view->active)
	{
		if(ViewCursorWindowRect(view,selection,&cursorRect))				// get the window-coordinate rect of the cursor on the view
		{
			ImageFill(imageBuffer,ForegroundForStyleGet(view,false,0,true),cursorRect.x,cursorRect.y,cursorRect.w,cursorRect.h);
		}
	}
}

static bool MarksBackgroundColorAtPositionGet(EDITOR_VIEW *view,uint64_t position,EDITOR_COLOR *color)
// Work out what the mark background color contribution is for the given text position.
// This looks at all visible marks, and sums their individual background color
// components. If there are no visible marks at the position, false is returned.
{
	bool
		sawMark;
	MARK_VISIBILITY
		*visibleMark;

	sawMark=false;
	visibleMark=view->headVisibility;					// get head of visible marks
	while(visibleMark)
	{
		if(SelectionAtPositionGet(visibleMark->parentMark->selectionUniverse,position))	// mark active?
		{
			if(!sawMark)								// if first mark, just copy the color
			{
				if(view->active)
				{
					*color=visibleMark->activeBackgroundColor;
				}
				else
				{
					*color=visibleMark->inactiveBackgroundColor;
				}
				sawMark=true;
			}
			else										// sum in other colors
			{
				if(view->active)
				{
					*color=ColorsSum(*color,visibleMark->activeBackgroundColor);
				}
				else
				{
					*color=ColorsSum(*color,visibleMark->inactiveBackgroundColor);
				}
			}
		}
		visibleMark=visibleMark->nextMark;				// move to next mark hanging off of this view
	}
	return(sawMark);
}

static void ViewTextPieceDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,uint32_t *buffer,uint64_t bufferLength,bool selected,bool haveStyle,uint32_t style,int64_t viewX,int64_t viewY)
// Draw the piece of text in the staging buffer into the view in the given style and selection.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor;
	int64_t
		windowX,
		windowY;

	viewInfo=view->viewInfo;							// point at the information record for this view

	if(selected)
	{
		ViewColorSchemeSelectionColorsGet(view->colorScheme,view->active,&backgroundColor,&foregroundColor);	// get selection foreground color
		if(!foregroundColor)							// if not specified, then just use normal foreground color
		{
			if(backgroundColor)
			{
				foregroundColor=ForegroundForStyleGet(view,haveStyle,style,view->active);
			}
			else	// neither foreground or background color specified for selection, use inverse
			{
				foregroundColor=BackgroundForStyleGet(view,haveStyle,style,view->active);
			}
		}
	}
	else
	{
		foregroundColor=ForegroundForStyleGet(view,haveStyle,style,view->active);
	}
	ViewCoordinateToWindowCoordinate(view,viewX,viewY,&windowX,&windowY);
	FontText32Draw(FontForStyleGet(view,haveStyle,style),imageBuffer,foregroundColor,windowX,windowY+viewInfo->lineAscent,buffer,bufferLength);	// draw to window buffer
}

static void ViewTextLineDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,uint64_t textLine,TEXT_STREAM *stream)
// Read text from the line at stream, draw into the view.
// This just draws the text foreground, not the background (which
// should already have been drawn).
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	TEXT_EXPANSION_STATE
		expansionState;
	int64_t
		textX,
		maxTextX,
		lastX;
	int64_t
		viewX,
		viewY;
	uint32_t
		codePoint;
	bool
		haveCurrentStyle,
		haveNextStyle;
	uint32_t
		currentStyle,
		nextStyle;
	bool
		currentSelection,
		nextSelection;
	uint64_t
		maxPosition;

	viewInfo=view->viewInfo;							// point at the information record for this view
	ViewCoordinateXToTextCoordinateX(view,viewInfo->bounds.w,&maxTextX);		// get maximum X position to draw

	TextExpansionStateInit(&expansionState);
	expansionState.tabSize=view->tabSize;
	maxPosition=TextUniverseTotalBytes(stream->universe);

	ViewCoordinateXToTextCoordinateX(view,0,&textX);							// get the text X offset to the left-most pixel of the view
	TextStreamExpandPixels(view,stream,&expansionState,textX);					// push out to the starting character to draw

	expansionState.buffer=gceExpansionBuffer;			// write expanded text to this buffer
	expansionState.bufferIndex=0;						// begin at start of buffer

	currentSelection=SelectionAtPositionGet(view->parentBuffer->selectionUniverse,TextStreamPositionGet(stream));
	haveCurrentStyle=StyleAtPositionGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&currentStyle);

	expansionState.font=FontForStyleGet(view,haveCurrentStyle,currentStyle);

	lastX=expansionState.runningWidth;					// start drawing here

	while(!TextStreamAtEnd(stream)&&(expansionState.runningWidth<(uint64_t)maxTextX)&&LineCodePointFetch(stream,view->useUtf8,maxPosition,&codePoint))
	{
		// get the selection and style for the code point just fetched by looking at the selection and style data for its last byte
		nextSelection=SelectionAtPositionGet(view->parentBuffer->selectionUniverse,TextStreamPositionGet(stream)-1);
		haveNextStyle=StyleAtPositionGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream)-1,&nextStyle);

		if((haveCurrentStyle!=haveNextStyle)||(haveCurrentStyle&&(currentStyle!=nextStyle))||(currentSelection!=nextSelection)||(expansionState.bufferIndex>EXPANSION_BUFFER_SIZE-EXPANSION_MAX_ADD))	// is something changing? (if so, output text so far) (this will also output if the staging buffer is getting full)
		{
			TextCoordinateToViewCoordinate(view,lastX,textLine,&viewX,&viewY);		// convert text coordinates to view coordinates
			ViewTextPieceDraw(view,imageBuffer,expansionState.buffer,expansionState.bufferIndex,currentSelection,haveCurrentStyle,currentStyle,viewX,viewY);

			lastX=expansionState.runningWidth;			// update the draw index for next time
			expansionState.bufferIndex=0;				// begin at start of buffer

			currentSelection=nextSelection;
			haveCurrentStyle=haveNextStyle;
			currentStyle=nextStyle;

			expansionState.font=FontForStyleGet(view,haveCurrentStyle,currentStyle);
		}
		GlyphExpand(codePoint,&expansionState);			// expand the glyph into the buffer now
	}

	// ran off the end of the line, or the end of the universe
	if(expansionState.bufferIndex)
	{
		TextCoordinateToViewCoordinate(view,lastX,textLine,&viewX,&viewY);		// convert text coordinates to view coordinates
		ViewTextPieceDraw(view,imageBuffer,expansionState.buffer,expansionState.bufferIndex,currentSelection,haveCurrentStyle,currentStyle,viewX,viewY);
	}
}

static void ViewTextDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer)
// Draw the text for all lines into view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	TEXT_STREAM
		stream;
	uint64_t
		i;

	viewInfo=view->viewInfo;			// point at the information record for this view

	TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine,&stream);	// point to data for the top line
	for(i=0;i<viewInfo->numLines;i++)
	{
		ViewTextLineDraw(view,imageBuffer,viewInfo->topLine+i,&stream);
		TextStreamToNextLine(&stream);	// move to the next line
	}
}

static void ViewSelectionSegmentDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,EDITOR_COLOR color,int64_t textStartX,int64_t textStartLine,int64_t textEndX,int64_t textEndLine)
// Draw a selection segment given by textStartLine,textStartX and
// textEndLine,textEndX into the view.
// textStartLine must always be <= textEndLine.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	int64_t
		viewStartX,
		viewEndX,
		viewY,
		textLineEndX,
		textLineBelowEndX,
		viewLineEndX,
		viewLineBelowEndX;
	int64_t
		windowX,
		windowY;

	viewInfo=view->viewInfo;			// point at the information record for this view

	if(textStartLine==textEndLine)		// handle simple case of selection on a single line
	{
		if((textStartLine>=viewInfo->topLine)&&(textStartLine<viewInfo->topLine+viewInfo->numLines))	// verify line is in the view
		{
			// make coordinates view local
			TextCoordinateLineToViewCoordinateY(view,textStartLine,&viewY);

			TextCoordinateXToViewCoordinateX(view,textStartX,&viewStartX);
			ViewCoordinateXPin(view,viewStartX,&viewStartX);

			TextCoordinateXToViewCoordinateX(view,textEndX,&viewEndX);
			ViewCoordinateXPin(view,viewEndX,&viewEndX);

			if((viewEndX-viewStartX)>1)	// something left to draw?
			{
				ViewCoordinateToWindowCoordinate(view,viewStartX,viewY,&windowX,&windowY);
				ImageFill(imageBuffer,color,windowX,windowY,viewEndX-viewStartX-1,viewInfo->lineHeight-1);
			}
		}
	}
	else	// starts on one line, ends on another
	{
		if(textStartLine<viewInfo->topLine)			// starting before the view?
		{
			textStartX=0;							// move X back to the margin
			textStartLine=viewInfo->topLine;		// move to the first line of the view
		}
		if(textEndLine>viewInfo->topLine+viewInfo->numLines)	// if off the end, then pin
		{
			textEndX=viewInfo->leftPixel+viewInfo->bounds.w;	// move to the right margin
			textEndLine=viewInfo->topLine+viewInfo->numLines;
		}
		while((textStartLine<=textEndLine)&&(textStartLine<viewInfo->topLine+viewInfo->numLines))	// move through the lines
		{
			if(textStartLine<textEndLine)			// see if before the last line
			{
				textLineEndX=viewInfo->leftPixel+viewInfo->bounds.w;	// line draws to the edge
				if(textStartLine+1<textEndLine)		// is next line a full line?
				{
					textLineBelowEndX=viewInfo->leftPixel+viewInfo->bounds.w;
				}
				else
				{
					textLineBelowEndX=textEndX-1;	// line draws to the end (pull back one pixel for segment edge)
				}
			}
			else
			{
				textLineEndX=textEndX-1;			// line draws to the end (pull back one pixel for segment edge)
				textLineBelowEndX=0;
			}

			TextCoordinateLineToViewCoordinateY(view,textStartLine,&viewY);	// work out the Y for this draw

			TextCoordinateXToViewCoordinateX(view,textStartX,&viewStartX);
			ViewCoordinateXPin(view,viewStartX,&viewStartX);

			TextCoordinateXToViewCoordinateX(view,textLineEndX,&viewLineEndX);
			ViewCoordinateXPin(view,viewLineEndX,&viewLineEndX);

			TextCoordinateXToViewCoordinateX(view,textLineBelowEndX,&viewLineBelowEndX);
			ViewCoordinateXPin(view,viewLineBelowEndX,&viewLineBelowEndX);

			if(viewStartX<viewLineBelowEndX)	// is there some of this line that needs to be drawn full height?
			{
				ViewCoordinateToWindowCoordinate(view,viewStartX,viewY,&windowX,&windowY);
				ImageFill(imageBuffer,color,windowX,windowY,viewLineBelowEndX-viewStartX,viewInfo->lineHeight);
				viewStartX=viewLineBelowEndX;	// move over
			}
			if(viewStartX<viewLineEndX)			// remainder of line to be drawn with reduced height?
			{
				ViewCoordinateToWindowCoordinate(view,viewStartX,viewY,&windowX,&windowY);
				ImageFill(imageBuffer,color,windowX,windowY,viewLineEndX-viewStartX,viewInfo->lineHeight-1);
			}
			textStartX=0;
			textStartLine++;
		}
	}
}

static void ViewSelectionDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,SELECTION_UNIVERSE *selection)
// Draw the visible part of the selection.
// The selection is drawn as a number of outlined segments.
// As such, it is more efficient to draw the segments themselves than
// to try to draw the selection line by line as other draws do.
// This function works pretty hard to extract only those selection segments
// that intersect the view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor,
		selectionColor;
	TEXT_STREAM
		currentStream,
		endStream;
	TEXT_EXPANSION_STATE
		currentExpansionState,
		endExpansionState;
	int64_t
		minTextX,
		maxTextX;
	uint64_t
		currentPosition,
		currentLineEndPosition;
	uint64_t
		selectionPosition,
		numElements;
	bool
		done,
		doneLine;
	int64_t
		selectionStartTextX,
		selectionEndTextX;
	int64_t
		selectionStartTextLine,
		selectionEndTextLine;
	int64_t
		viewCurrentTextLine,
		viewEndTextLine;

	viewInfo=view->viewInfo;											// point at the information record for this view

	ViewColorSchemeSelectionColorsGet(view->colorScheme,view->active,&backgroundColor,&foregroundColor);
	if(backgroundColor)
	{
		selectionColor=backgroundColor;
	}
	else	// no selection background color specified, so use invert mode (select with the foreground color)
	{
		ViewColorSchemeDrawColorsGet(view->colorScheme,view->active,&backgroundColor,&foregroundColor);
		selectionColor=foregroundColor;
	}

	ViewCoordinateXToTextCoordinateX(view,0,&minTextX);											// get minimum X position to hunt
	ViewCoordinateXToTextCoordinateX(view,viewInfo->bounds.w,&maxTextX);						// get maximum X position to hunt

	// Run through all the selection segments that intersect the view.
	ViewCoordinateLineToTextCoordinateLine(view,0,&viewCurrentTextLine);						// start at line 0 of the view
	ViewCoordinateLineToTextCoordinateLine(view,viewInfo->numLines,&viewEndTextLine);			// one past the end of the view

	done=false;
	while((!done)&&(viewCurrentTextLine<viewEndTextLine))				// still have lines to look at?
	{
		TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewCurrentTextLine,&currentStream);	// point to the start of the current line of the view
		if(!TextStreamAtEnd(&currentStream))							// make sure there's a line
		{
			TextExpansionStateInit(&currentExpansionState);
			currentExpansionState.tabSize=view->tabSize;

			TextStreamExpandPixels(view,&currentStream,&currentExpansionState,minTextX);					// find the position in the text that is as close to the left edge of the view as possible
			currentPosition=TextStreamPositionGet(&currentStream);											// start looking for selections at this point

			endStream=currentStream;
			endExpansionState=currentExpansionState;
			TextStreamExpandPixels(view,&endStream,&endExpansionState,maxTextX);							// find the position in the text that is as close to the right edge of the view as possible
			currentLineEndPosition=TextStreamPositionGet(&endStream);										// this, along with currentPosition define a range to compare to the selection to check for intersect

			doneLine=false;
			while(!doneLine)
			{
				if(SelectionAtOrAfterPositionGet(selection,currentPosition,&selectionPosition,&numElements))	// see if there's a selection intersecting currentPosition, or after it
				{
					if(selectionPosition<=currentLineEndPosition)											// intersection?
					{
						if((selectionPosition>=currentPosition)&&(selectionPosition+numElements<=currentLineEndPosition))							// start and end both in view and both on the current line? (we can make this case go much faster)
						{
							TextStreamExpandCharacters(view,&currentStream,&currentExpansionState,selectionPosition,false,0);	// get width to start of selection
							selectionStartTextX=currentExpansionState.runningWidth;
							TextStreamExpandCharacters(view,&currentStream,&currentExpansionState,selectionPosition+numElements,false,0);	// get width to end of selection
							selectionEndTextX=currentExpansionState.runningWidth;
							selectionStartTextLine=selectionEndTextLine=viewCurrentTextLine;				// both on current line
						}
						else	// slow case
						{
							TextPositionToTextCoordinate(view,selectionPosition,&selectionStartTextX,&selectionStartTextLine,true,maxTextX);			// find out where this selection begins (line and pixel)
							TextPositionToTextCoordinate(view,selectionPosition+numElements,&selectionEndTextX,&selectionEndTextLine,true,maxTextX);	// find out where this selection ends (line and pixel)
						}

						ViewSelectionSegmentDraw(view,imageBuffer,selectionColor,selectionStartTextX,selectionStartTextLine,selectionEndTextX,selectionEndTextLine);	// go render it
						if(selectionPosition+numElements<=currentLineEndPosition)					// see if possibly more selections on the same line
						{
							currentPosition=selectionPosition+numElements;
						}
						else	// no more selections on the same line, move to the next line, or the selection end line, whichever's greater
						{
							viewCurrentTextLine++;	// move to the next line
							if(viewCurrentTextLine<selectionEndTextLine)	// if the end line is greater, move there
							{
								viewCurrentTextLine=selectionEndTextLine;
							}
							doneLine=true;
						}
					}
					else	// selection doesn't intersect -- must be completely after the line, move to the line where it starts, or the next line whichever's greater
					{
						viewCurrentTextLine++;	// move to the next line
						selectionStartTextLine=TextUniversePositionToLine(view->parentBuffer->textUniverse,selectionPosition,NULL);	// find out what line the start is on
						if(viewCurrentTextLine<selectionStartTextLine)
						{
							viewCurrentTextLine=selectionStartTextLine;
						}
						doneLine=true;
					}
				}
				else
				{
					doneLine=true;															// no more selections, we're completely done
					done=true;
				}
			}
		}
		else
		{
			done=true;
		}
	}
}

static void ViewBackgroundPieceDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,uint64_t width,bool haveMarkColor,EDITOR_COLOR markColor,bool haveStyle,uint32_t style,int64_t viewX,int64_t viewY)
// Draw the background into the view in the given markColor and style.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	EDITOR_COLOR
		background;
	int64_t
		windowX,
		windowY;

	if(haveMarkColor||style)										// drawing something other than the default background (which has already been filled)
	{
		viewInfo=view->viewInfo;									// point at the information record for this view

		if(!haveMarkColor)
		{
			background=BackgroundForStyleGet(view,haveStyle,style,view->active);
		}
		else
		{
			background=markColor;
		}

		ViewCoordinateToWindowCoordinate(view,viewX,viewY,&windowX,&windowY);
		ImageFill(imageBuffer,background,windowX,windowY,width,viewInfo->lineHeight);
	}
}

static void ViewBackgroundLineDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,uint32_t textLine,TEXT_STREAM *stream)
// Draw the background for the line of text given at stream
// into the view. This includes any the background for
// any styles and marks in the text, but does not
// include the current selection -- that's drawn separately.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	TEXT_EXPANSION_STATE
		expansionState;
	int64_t
		textX,
		maxTextX,
		lastX;
	int64_t
		viewX,
		viewY;
	uint32_t
		codePoint;
	bool
		haveCurrentStyle,
		haveNextStyle;
	uint32_t
		currentStyle,
		nextStyle;
	bool
		currentMark,
		nextMark;
	EDITOR_COLOR
		currentMarkColor,
		nextMarkColor;
	uint64_t
		maxPosition;

	viewInfo=view->viewInfo;				// point at the information record for this view
	ViewCoordinateXToTextCoordinateX(view,viewInfo->bounds.w,&maxTextX);		// get maximum X position to draw

	TextExpansionStateInit(&expansionState);
	expansionState.tabSize=view->tabSize;
	maxPosition=TextUniverseTotalBytes(stream->universe);

	ViewCoordinateXToTextCoordinateX(view,0,&textX);							// get the text X offset to the left-most pixel of the view
	TextStreamExpandPixels(view,stream,&expansionState,textX);					// push out to the starting character to draw

	currentMarkColor=nextMarkColor=0;		// keep compiler quiet
	currentMark=MarksBackgroundColorAtPositionGet(view,TextStreamPositionGet(stream),&currentMarkColor);
	haveCurrentStyle=StyleAtPositionGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&currentStyle);

	expansionState.font=FontForStyleGet(view,haveCurrentStyle,currentStyle);

	lastX=expansionState.runningWidth;

	while(!TextStreamAtEnd(stream)&&(expansionState.runningWidth<(uint64_t)maxTextX)&&LineCodePointFetch(stream,view->useUtf8,maxPosition,&codePoint))	// fill in all of the background needed
	{
		nextMark=MarksBackgroundColorAtPositionGet(view,TextStreamPositionGet(stream)-1,&nextMarkColor);
		haveNextStyle=StyleAtPositionGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream)-1,&nextStyle);

		if((haveCurrentStyle!=haveNextStyle)||((haveCurrentStyle)&&currentStyle!=nextStyle)||(currentMark!=nextMark)||(currentMark&&(currentMarkColor!=nextMarkColor)))	// is something about to change? (if so, output background so far)
		{
			TextCoordinateToViewCoordinate(view,lastX,textLine,&viewX,&viewY);		// convert text coordinates to view coordinates
			ViewBackgroundPieceDraw(view,imageBuffer,expansionState.runningWidth-lastX,currentMark,currentMarkColor,haveCurrentStyle,currentStyle,viewX,viewY);

			lastX=expansionState.runningWidth;		// move over

			currentMark=nextMark;
			currentMarkColor=nextMarkColor;
			haveCurrentStyle=haveNextStyle;
			currentStyle=nextStyle;

			expansionState.font=FontForStyleGet(view,haveCurrentStyle,currentStyle);
		}
		GlyphExpand(codePoint,&expansionState);
	}

	// ran off the end of the line, or the end of the universe
	if(expansionState.runningWidth-lastX)
	{
		TextCoordinateToViewCoordinate(view,lastX,textLine,&viewX,&viewY);			// convert text coordinates to view coordinates
		ViewBackgroundPieceDraw(view,imageBuffer,expansionState.runningWidth-lastX,currentMark,currentMarkColor,haveCurrentStyle,currentStyle,viewX,viewY);
		lastX=expansionState.runningWidth;			// move over
	}

	if(lastX<maxTextX)								// do we need to draw out to the end of the line?
	{
		currentMark=MarksBackgroundColorAtPositionGet(view,TextStreamPositionGet(stream),&currentMarkColor);
		haveCurrentStyle=StyleAtPositionGet(view->parentBuffer->styleUniverse,TextStreamPositionGet(stream),&currentStyle);	// get style and selection information for newline (or off of end)
		TextCoordinateToViewCoordinate(view,lastX,textLine,&viewX,&viewY);			// convert text coordinates to view coordinates
		ViewBackgroundPieceDraw(view,imageBuffer,maxTextX-lastX,currentMark,currentMarkColor,haveCurrentStyle,currentStyle,viewX,viewY);
	}
}

static void ViewBackgroundLinesDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer)
// Draw the background for all lines into view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	TEXT_STREAM
		stream;
	uint32_t
		i;

	viewInfo=view->viewInfo;			// point at the information record for this view

	TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine,&stream);	// point to data for this line
	for(i=0;i<viewInfo->numLines;i++)
	{
		ViewBackgroundLineDraw(view,imageBuffer,viewInfo->topLine+i,&stream);
		TextStreamToNextLine(&stream);
	}
}

static void ViewBackgroundDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer)
// Draw the overall background of the view -- all pixels that
// don't get touched by anything else.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor;

	viewInfo=view->viewInfo;											// point at the information record for this view

	ViewColorSchemeDrawColorsGet(view->colorScheme,view->active,&backgroundColor,&foregroundColor);
	ImageFill(imageBuffer,backgroundColor,viewInfo->bounds.x,viewInfo->bounds.y,viewInfo->bounds.w,viewInfo->bounds.h);
}

void ViewDraw(EDITOR_VIEW *view,GUI_IMAGE *imageBuffer,Region invalidRegion)
// Draw the given view into imageBuffer.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;											// point at the information record for this view

	if(XRectInRegion(invalidRegion,viewInfo->bounds.x,viewInfo->bounds.y,viewInfo->bounds.w,viewInfo->bounds.h)!=RectangleOut)
	{
		ImageClipRegionRectSet(imageBuffer,&viewInfo->bounds);			// don't draw outside the view

		ViewBackgroundDraw(view,imageBuffer);							// draw the overall view background
		ViewBackgroundLinesDraw(view,imageBuffer);						// draw the background for the lines (including all visible marks)
		ViewSelectionDraw(view,imageBuffer,view->parentBuffer->selectionUniverse);	// draw the selection
		ViewTextDraw(view,imageBuffer);									// draw the text
		ViewCursorDraw(view,imageBuffer,view->parentBuffer->selectionUniverse);

		ImageClipRegionSet(imageBuffer,None);							// unset clip area
	}
}

// ----------------------------------------------------------------------------
// Invalidation

static void ViewPortionInvalidate(EDITOR_VIEW *view,uint64_t viewStartLine,uint64_t viewEndLine,uint32_t viewStartPixel,uint32_t viewEndPixel)
// Invalidate the area of the view between viewStartLine, viewEndLine, and viewStartPixel, viewEndPixel.
// NOTE: viewStartLine, viewEndLine, viewStartPixel, and viewEndPixel are all relative to the displayed view
// so that viewStartLine of 0 always means the top line, and viewStartPixel of 0 always
// means the left-most pixel.
// NOTE: viewStartLine must be <= viewEndLine, viewStartPixel must be <= viewEndPixel.
// The invalidated area is exclusive of viewEndLine, and viewEndPixel.
{
	uint32_t
		totalPixels;
	PLATFORM_VIEW_INFO
		*viewInfo;
	int64_t
		viewY,
		windowX,
		windowY;
	EDITOR_RECT
		invalidRect;

	viewInfo=view->viewInfo;						// point at the information record for this view

	totalPixels=viewInfo->bounds.w;
	if(viewInfo->numLines&&totalPixels)				// make sure we can actually invalidate
	{
		if(viewStartLine>viewInfo->numLines)
		{
			viewStartLine=viewInfo->numLines;
		}
		if(viewEndLine>viewInfo->numLines)
		{
			viewEndLine=viewInfo->numLines;
		}

		if(viewStartPixel>totalPixels)
		{
			viewStartPixel=totalPixels;
		}
		if(viewEndPixel>totalPixels)
		{
			viewEndPixel=totalPixels;
		}

		ViewCoordinateLineToViewCoordinateY(view,viewStartLine,&viewY);	// convert line to Y at top of line
		ViewCoordinateToWindowCoordinate(view,viewStartPixel,viewY,&windowX,&windowY);	// convert to window coordinates

		invalidRect.x=windowX;
		invalidRect.y=windowY;
		invalidRect.w=viewEndPixel-viewStartPixel;
		invalidRect.h=(viewEndLine-viewStartLine)*viewInfo->lineHeight;
		WindowRectInvalidate(view->parentWindow,&invalidRect);
	}
}

static void ViewCursorInvalidate(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection)
// Passed a view, invalidate the visible area of the view
// covered by the cursor.
{
	EDITOR_RECT
		cursorRect;

	if(SelectionIsEmpty(selection))
	{
		if(ViewCursorWindowRect(view,selection,&cursorRect))		// get the window-coordinate rect of the cursor on the view
		{
			WindowRectInvalidate(view->parentWindow,&cursorRect);
		}
	}
}

static void ViewInvalidate(EDITOR_VIEW *view)
// Invalidate the entire area of the view.
{
	WindowRectInvalidate(view->parentWindow,&view->viewInfo->bounds);
}

// ----------------------------------------------------------------------------

static void ViewFontInfoRecalculate(EDITOR_VIEW *view)
// Recalculate the cached information for view
// from its current size and color scheme.
{
	int32_t
		viewHeight;
	EDITOR_FONT
		*font;
	uint32_t
		ascent,
		descent;
	uint32_t
		maxDescent;
	uint32_t
		i;
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=(PLATFORM_VIEW_INFO *)view->viewInfo;					// point at the information record for this view

	font=ViewColorSchemeFontsViewFontGet(view->colorSchemeFonts);	// get the base font for the view
	FontHeightsGet(font,&ascent,&descent);

	viewInfo->lineAscent=ascent;
	maxDescent=descent;

	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		if((font=ViewColorSchemeFontsStyleFontGet(view->colorSchemeFonts,i)))	// see if a font is assigned for this style
		{
			FontHeightsGet(font,&ascent,&descent);

			if(viewInfo->lineAscent<(ascent))
			{
				viewInfo->lineAscent=ascent;
			}
			if(maxDescent<(descent))
			{
				maxDescent=descent;
			}
		}
	}
	viewInfo->lineHeight=viewInfo->lineAscent+maxDescent;

	if(viewInfo->lineHeight<=0)
	{
		viewInfo->lineHeight=1;									// lines are minimally one pixel tall (so we do not get an infinite number of lines on the view)
	}

	if(viewInfo->bounds.w>ACTIVE_LEFT_MARGIN+ACTIVE_RIGHT_MARGIN)
	{
		viewInfo->numPixels=viewInfo->bounds.w-(ACTIVE_LEFT_MARGIN+ACTIVE_RIGHT_MARGIN);
	}
	else
	{
		viewInfo->numPixels=0;
	}

	if(viewInfo->bounds.h>ACTIVE_TOP_MARGIN+ACTIVE_BOTTOM_MARGIN)
	{
		viewHeight=viewInfo->bounds.h-ACTIVE_TOP_MARGIN-ACTIVE_BOTTOM_MARGIN;	// find out how much can be used for the text
		viewInfo->numLines=viewHeight/viewInfo->lineHeight;						// get integer number of lines in this view
	}
	else
	{
		viewInfo->numLines=0;
	}

	viewInfo->startPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine,NULL);
	viewInfo->endPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine+viewInfo->numLines,NULL);
}

// ----------------------------------------------------------------------------
// Cursor blinking

static void ViewCursorBlink(EDITOR_VIEW *view)
// Calling this will cause the cursor in view to change blink states, updating it as needed.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=(PLATFORM_VIEW_INFO *)view->viewInfo;				// point at the information record for this view
	viewInfo->blinkState=!viewInfo->blinkState;				// invert blink state
	ViewCursorInvalidate(view,view->parentBuffer->selectionUniverse);	// invalidate the area of the window where the cursor is currently, so we'll see the change on the next redraw
}

static void ViewCursorBlinkReset(EDITOR_VIEW *view)
// When this is called, it will reset the cursor blinking, force the
// cursor to the blink-on state, and reset the blink timeout.
// This is used to make the cursor show up when events occur, and
// to stop it blinking while tracking the mouse, etc...
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=(PLATFORM_VIEW_INFO *)view->viewInfo;				// point at the information record for this view
	if(!viewInfo->blinkState)								// if it is currently off, blink it to on
	{
		ViewCursorBlink(view);
	}
	CursorBlinkTimeReset();
}

void ViewCursorTask(EDITOR_VIEW *view)
// Handle blinking the cursor for view.
{
	ViewCursorBlink(view);
}

// ----------------------------------------------------------------------------
// Events

static bool ViewMouseWheelEventsHandle(EDITOR_VIEW *view,XEvent *event)
// Call this to see if mouse wheel has been rolled.
// X sends button press events 4 (for UP) and 5 (for DOWN) when you scroll the mouse wheel.
// Returns true if the event was a mouse wheel event.
{
	if((event->xbutton.button==Button4)||(event->xbutton.button==Button5))
	{
		DocumentWindowScrollEditViewEvent(view,(event->xbutton.button==Button4)?VVK_SCROLL_UP:VVK_SCROLL_DOWN,false);
		DocumentWindowScrollEditViewEvent(view,(event->xbutton.button==Button4)?VVK_SCROLL_UP:VVK_SCROLL_DOWN,false);
		DocumentWindowScrollEditViewEvent(view,(event->xbutton.button==Button4)?VVK_SCROLL_UP:VVK_SCROLL_DOWN,false);
		return(true);
	}
	return(false);
}

bool ViewTrack(EDITOR_VIEW *view,XEvent *event)
// Call this to see if the click was in the view, and if so, track it.
// If the click was not in the view, return false.
{
	VIEW_EVENT
		record;
	VIEW_CLICK_EVENT_DATA
		clickData;
	PLATFORM_VIEW_INFO
		*viewInfo;
	int
		rootX,
		rootY,
		windowX,
		windowY;
	int64_t
		viewX,
		viewY,
		viewLine;
	uint32_t
		modifiers;
	uint8_t
		button,
		numClicks;

	if(PointInView(view,event->xbutton.x,event->xbutton.y))
	{
		if(!ViewMouseWheelEventsHandle(view,event))
		{
			numClicks=MultipleClicksHandle(event);
			modifiers=XStateToEditorModifiers(event->xbutton.state);
			switch(event->xbutton.button)
			{
				case Button1:
					button=0;
					break;
				case Button2:
					button=1;
					break;
				case Button3:
					button=2;
					break;
				default:
					button=0;
					break;
			}
			clickData.modifiers=modifiers;
			clickData.button=button;
			clickData.numClicks=numClicks;

			record.eventType=VET_CLICK;
			record.view=view;
			record.eventData=&clickData;
			viewInfo=(PLATFORM_VIEW_INFO *)(view->viewInfo);
			WindowCoordinateToViewCoordinate(record.view,event->xbutton.x,event->xbutton.y,&viewX,&viewY);
			ViewCoordinateYToViewCoordinateLine(view,viewY,&viewLine);				// make Y into line

			if((viewLine>=0)&&(viewLine<viewInfo->numLines))						// make sure initial click is in the view vertically
			{
				clickData.xClick=viewX;
				clickData.yClick=viewLine;
				ViewEventHandle(&record);

				rootX=event->xbutton.x_root;
				rootY=event->xbutton.y_root;
				windowX=event->xbutton.x;
				windowY=event->xbutton.y;

				while(StillDown(event->xbutton.window,event->xbutton.button,&rootX,&rootY,&windowX,&windowY))		// send repeat events for the view click
				{
					record.view=view;
					record.eventType=VET_CLICK_HOLD;
					record.eventData=&clickData;

					WindowCoordinateToViewCoordinate(record.view,windowX,windowY,&viewX,&viewY);
					ViewCoordinateYToViewCoordinateLine(view,viewY,&viewLine);	// pass view Y as a line to the upper level code, since that's all it cares about
					clickData.xClick=viewX;
					clickData.yClick=viewLine;

					clickData.modifiers=modifiers;
					ViewEventHandle(&record);
				}
				record.eventType=VET_CLICK_RELEASE;
				record.view=view;
				ViewEventHandle(&record);
			}
		}
		return(true);
	}
	return(false);
}

void ViewKeyEventHandle(EDITOR_VIEW *view,EDITOR_KEY *editorKey)
// Handle an incoming key for view.
{
	VIEW_EVENT
		viewEvent;

	viewEvent.view=view;										// point at the view
	viewEvent.eventType=VET_KEY_DOWN;							// let it know what type
	viewEvent.eventData=editorKey;
	ViewEventHandle(&viewEvent);								// pass this off to high level handler
}

void ViewBoundsSet(EDITOR_VIEW *view,EDITOR_RECT *newBounds)
// Set the bounds of view to newBounds.
// Invalidate area where the view used to be but now is not
// and invalidate new areas of the view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;
	EDITOR_RECT
		*viewBounds;
	unsigned int
		width,
		height;

	viewInfo=view->viewInfo;									// point at the information record for this view

	viewBounds=&(viewInfo->bounds);								// point at the view bounds rect

	if((newBounds->x!=viewBounds->x)||(newBounds->y!=viewBounds->y)||(newBounds->w!=viewBounds->w)||(newBounds->h!=viewBounds->h))
	{
		WindowRectInvalidate(view->parentWindow,viewBounds);	// invalidate old area
		WindowRectInvalidate(view->parentWindow,newBounds);		// invalidate new area

		(*viewBounds)=*newBounds;								// copy new rect into place
		ViewFontInfoRecalculate(view);							// update additional information

		if(!(width=viewInfo->bounds.w))
		{
			width=1;
		}
		if(!(height=viewInfo->bounds.h))
		{
			height=1;
		}
		XMoveResizeWindow(xDisplay,viewInfo->cursorWindow,viewInfo->bounds.x,viewInfo->bounds.y,width,height);	// adjust the cursor change window
	}
}


// ----------------------------------------------------------------------------
// Routines used by the editor proper

void PlatformViewTextXToPositionGet(EDITOR_VIEW *view,uint64_t linePosition,int64_t textX,uint64_t *betweenOffset,uint64_t *charOffset)
// Given linePosition which points to the byte at the start of a line, and textX
// as a pixel position along the line, determine two offsets:
// betweenOffset is the offset of the character boundary closest to textX,
// charOffset is the offset of the character closest to the textX.
// If linePosition is past EOF, both offsets will be returned as 0.
{
	TEXT_STREAM
		stream;
	uint32_t
		codePoint;
	TEXT_EXPANSION_STATE
		expansionState;
	int64_t
		lastX;
	uint32_t
		charWidth;
	bool
		inLine;
	uint64_t
		maxPosition;

	TextExpansionStateInit(&expansionState);
	expansionState.tabSize=view->tabSize;
	maxPosition=TextUniverseTotalBytes(view->parentBuffer->textUniverse);

	TextStreamPositionSet(&stream,view->parentBuffer->textUniverse,linePosition);	// make a stream pointing to the start of the line

	inLine=TextStreamExpandPixels(view,&stream,&expansionState,textX);				// expand out to the position
	(*betweenOffset)=(*charOffset)=TextStreamPositionGet(&stream)-linePosition;
	lastX=expansionState.runningWidth;
	if(inLine)																		// make sure we didn't run off the end of the line
	{
		LineCodePointFetch(&stream,view->useUtf8,maxPosition,&codePoint);
		GlyphExpand(codePoint,&expansionState);
	}
	charWidth=expansionState.runningWidth-lastX;
	if(inLine&&(textX>(lastX+((int32_t)charWidth/2))))
	{
		(*betweenOffset)=TextStreamPositionGet(&stream)-linePosition;
	}
}

void PlatformViewTextPositionToXGet(EDITOR_VIEW *view,uint64_t position,int64_t *textX)
// Given a character position within text, determine the text X offset to the start of the character.
// NOTE: this function requires a view, since the selected fonts and styles in the view impact
// the returned X offset.
// If position is passed in past the end of the buffer, this will return the width of the last line.
{
	int64_t
		lineNumber;

	TextPositionToTextCoordinate(view,position,textX,&lineNumber,false,0);
}

// ----------------------------------------------------------------------------
// High level update handling

void PlatformViewCoordinateXToTextCoordinateX(EDITOR_VIEW *view,int64_t viewX,int64_t *textX)
// Convert a view X coordinate to a text X coordinate.
// This works properly for negative viewX.
{
	ViewCoordinateXToTextCoordinateX(view,viewX,textX);
}

void PlatformViewCoordinateLineToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewLine,int64_t *textLine)
// Convert a view line coordinate to a text line coordinate.
// This works properly for negative viewLine.
{
	ViewCoordinateLineToTextCoordinateLine(view,viewLine,textLine);
}

void PlatformViewCoordinateYToTextCoordinateLine(EDITOR_VIEW *view,int64_t viewY,int64_t *textLine)
// Convert a view Y coordinate to a text line coordinate.
// This works properly for negative viewY.
{
	ViewCoordinateYToTextCoordinateLine(view,viewY,textLine);
}

void PlatformViewCoordinateToTextCoordinate(EDITOR_VIEW *view,int64_t viewX,int64_t viewY,int64_t *textX,int64_t *textLine)
// Convert a set of view coordinates into a set of text coordinates.
// This takes into account the current origin of the view.
// This works properly for negative viewX and viewY.
{
	ViewCoordinateToTextCoordinate(view,viewX,viewY,textX,textLine);
}

void PlatformViewMetaChangeEnd(EDITOR_VIEW *view)
// The meta data for the given view has been updated.
// metadata is currently either tab size, or text encoding.
{
	ViewInvalidate(view);									// just lazily invalidate the whole thing
}

void PlatformViewMetaChangeStart(EDITOR_VIEW *view)
// Some metadata for the given view is about to be updated.
// metadata is currently either tab size, or text encoding.
{
}

void PlatformViewStyleAreaChangeEnd(EDITOR_VIEW *view,STYLE_UNIVERSE *style)
// The areas covered by different styles in the view have finished changing...
{
	ViewInvalidate(view);									// just lazily invalidate the whole thing
}

void PlatformViewStyleAreaChangeStart(EDITOR_VIEW *view,STYLE_UNIVERSE *style)
// The areas covered by different styles in the view are about to change.
{
}

void PlatformViewSelectionAreaChangeEnd(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection)
// The area covered by selection (or its cursor position) has changed
// for view.
// NOTE: selection may point to the user selection, or to any mark
// in the buffer.
// NOTE: if selection is passed as NULL, it indicates that the selection was
// deleted between the call to PlatformViewSelectionAreaChangeStart, and the
// call here. The original selection should be invalidated.
{
	ViewCursorBlinkReset(view);								// reset cursor blink on selection change
	ViewInvalidate(view);									// just lazily invalidate the whole thing
}

void PlatformViewSelectionAreaChangeStart(EDITOR_VIEW *view,SELECTION_UNIVERSE *selection)
// The area covered by selection (or its cursor position) is about to
// change for view.
// NOTE: selection may point to the user selection, or to any mark
// in the buffer.
// NOTE: if selection is passed as NULL, it is an indication that the editor
// is about to create a new selection (which will be passed to PlatformViewSelectionAreaChangeEnd).
// The entire area of the new selection should be invalidated when the end function
// is called.
{
}

void PlatformViewColorSchemeChangeEnd(EDITOR_VIEW *view)
// The color scheme and fonts for a view have been changed.
{
	ViewFontInfoRecalculate(view);							// update font information that's cached for the view
	ViewInvalidate(view);									// just lazily invalidate the whole thing
}

void PlatformViewColorSchemeChangeStart(EDITOR_VIEW *view)
// The color scheme and fonts for a view are about to change.
{
}

void PlatformViewTextPieceReplaceEnd(EDITOR_VIEW *view,uint64_t position,uint64_t startLength,uint64_t endLength)
// A text piece replacement has been completed to the buffer for this view.
// position points to the spot in the buffer's text where the change
// occurred. startLength tells how many bytes were removed.
// endLength tells how many bytes were added back.
// NOTE: changing text also implies possible changes to the styles
// associated with the text.
// NOTE: changing the text further implies a possible change to any
// selections and marks in the text in the area of the replacement.
// This needs to invalidate around the text change so that any
// selections that may have changed along with the text replacement
// will also be redrawn.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;								// point at the information record for this view
	if(position+startLength>=viewInfo->startPosition)		// make sure change is not completely before the start
	{
		if(position<=viewInfo->endPosition)					// make sure change is not completely after the end, if it is, we can completely ignore it
		{
			if(viewInfo->startLine>=viewInfo->topLine)		// see if change is happening at or after the top line
			{
				if(TextUniverseTotalLines(view->parentBuffer->textUniverse)==viewInfo->totalLinesAtStart)	// no net difference in number of lines?
				{
					// NOTE: changes can reach backward (depending on the font), so just invalidate the whole line
					ViewPortionInvalidate(view,viewInfo->startLine-viewInfo->topLine,viewInfo->endLine-viewInfo->topLine+1,0,viewInfo->bounds.w);	// invalidate the lines
				}
				else	// number of lines changed, so invalidate to bottom of view
				{
					ViewPortionInvalidate(view,viewInfo->startLine-viewInfo->topLine,viewInfo->numLines+1,0,viewInfo->bounds.w);	// invalidate to the bottom
				}
			}
			else											// change is before the top line, extending into the view
			{
				viewInfo->topLine=viewInfo->startLine;		// just move the view to the start line of the change
				ViewInvalidate(view);						// redraw the whole thing
			}

			// update cache of positions
			viewInfo->startPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine,NULL);
			viewInfo->endPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine+viewInfo->numLines,NULL);
			ViewCursorBlinkReset(view);						// reset cursor blink on text change in the view
			ViewCursorInvalidate(view,view->parentBuffer->selectionUniverse);	// change was in the view, make sure the cursor gets updated (potentially at its new location)
		}
	}
	else	// change is completely before the start, use it to update position cache quickly
	{
		viewInfo->startPosition+=endLength-startLength;		// adjust start position
		viewInfo->endPosition+=endLength-startLength;		// adjust end position
		viewInfo->topLine+=TextUniverseTotalLines(view->parentBuffer->textUniverse)-viewInfo->totalLinesAtStart;	// adjust the top line of the view based on the number of lines that were deleted, and added
	}
}

void PlatformViewTextPieceReplaceStart(EDITOR_VIEW *view,uint64_t position,uint64_t startLength,uint64_t endLength)
// The replacement of a piece of text is about to occur in the buffer
// for this view.
// position points to the spot in the buffer's text where the change
// will occur. startLength tells how many bytes will be removed.
// endLength tells how many bytes will be added back.
// NOTE: changing text also implies possible changes to the styles
// associated with the text.
// NOTE: changing the text further implies a possible change to any
// selections and marks in the text in the area of the replacement.
// This needs to invalidate around the text change so that any
// selections that may have changed along with the text replacement
// will also be redrawn.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;								// point at the information record for this view
	viewInfo->totalLinesAtStart=TextUniverseTotalLines(view->parentBuffer->textUniverse);	// hold the total number of lines in the universe before the replacement

	if(position+startLength>=viewInfo->startPosition)		// make sure change is not completely before the start, if it is, we can ignore it at this point
	{
		if(position<=viewInfo->endPosition)					// make sure change is not completely after the end, if it is, we can completely ignore it
		{
			ViewCursorInvalidate(view,view->parentBuffer->selectionUniverse);	// change is in the view, make sure the cursor gets updated if it is also in the view
			viewInfo->startLine=TextUniversePositionToLine(view->parentBuffer->textUniverse,position,NULL);			// get the line of the change (slow, but only happens for changes intersecting the view)
			viewInfo->endLine=TextUniversePositionToLine(view->parentBuffer->textUniverse,position+startLength,NULL);
		}
	}
}

void PlatformViewTextReplaceEnd(EDITOR_VIEW *view)
// The replacement of (potentially multiple pieces) of text in the view has
// completed. Do what's needed to update things.
{
	// nothing needed for this implementation -- text changes are only handled a piece at a time
}

void PlatformViewTextReplaceStart(EDITOR_VIEW *view)
// The replacement of (potentially multiple pieces) of text in the view is
// beginning. This is called to prepare. After this has been called, it can
// expect that PlatformViewTextPieceReplaceStart/PlatformViewTextPieceReplaceEnd will
// be called in pairs for each piece of text that is changed. Once all the
// changes are complete, PlatformViewTextReplaceEnd will be called.
{
	// nothing needed for this implementation -- text changes are only handled a piece at a time
}

void PlatformViewFocusChange(EDITOR_VIEW *view)
// When the focus on a view changes, this gets called to update it.
{
	ViewCursorBlinkReset(view);								// reset cursor blink on focus change
	ViewInvalidate(view);									// redraw the whole thing -- active status will be taken into account
}

// ----------------------------------------------------------------------------
// Changes to where the view points in the text, and to the size of the
// view.

void PlatformViewScrollBoundsGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine,uint64_t *numPixels,uint64_t *numLines)
// Get the bounds of the area of the view inside of which a click
// doesn't cause any scrolling.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	*leftPixel=0;
	*numPixels=viewInfo->bounds.w;
	*topLine=0;
	*numLines=viewInfo->numLines;
}

void PlatformViewActiveBoundsGet(EDITOR_VIEW *view,uint64_t *numPixels,uint64_t *numLines)
// Get the number of pixels and lines in the active part of the passed view.
// This is useful when paging in the view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	*numPixels=viewInfo->numPixels;
	*numLines=viewInfo->numLines;
}

void PlatformViewOriginGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine)
// Get the current origin of the given view.
{
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;
	*leftPixel=viewInfo->leftPixel;
	*topLine=viewInfo->topLine;
}

void PlatformViewOriginSet(EDITOR_VIEW *view,int64_t newLeftPixel,int64_t newTopLine)
// Set the upper left corner of view.
// NOTE: newTopLine, and newLeftPixel will be forced into bounds by this routine.
{
	uint64_t
		numLines;
	PLATFORM_VIEW_INFO
		*viewInfo;

	viewInfo=view->viewInfo;									// point at the information record for this view
	numLines=TextUniverseTotalLines(view->parentBuffer->textUniverse);

	if(newTopLine>(int64_t)numLines)
	{
		newTopLine=numLines;
	}

	if(viewInfo->leftPixel!=newLeftPixel||viewInfo->topLine!=newTopLine)	// see if anything actually changed
	{
		viewInfo->leftPixel=newLeftPixel;
		viewInfo->topLine=newTopLine;
		ViewInvalidate(view);									// invalidate the entire view
		viewInfo->startPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine,NULL);
		viewInfo->endPosition=TextUniverseLineToPosition(view->parentBuffer->textUniverse,viewInfo->topLine+viewInfo->numLines,NULL);
	}
}

// ----------------------------------------------------------------------------

void PlatformViewInfoDispose(EDITOR_VIEW *view)
// Unlink view from everywhere, and dispose of it.
{
	XDestroyWindow(xDisplay,view->viewInfo->cursorWindow);	// tell x to make the cursor window go away
	MDisposePtr(view->viewInfo);
}

bool PlatformViewInfoCreate(EDITOR_VIEW *view,EDITOR_RECT *boundsRect)
// Called after all editor specific fields of a new view have been filled in.
// This creates and stuffs local GUI specific information, and sets up the view in the GUI.
// If there is a problem, ErrorSet, and return false.
{
	unsigned int
		width,
		height;

	if((view->viewInfo=(PLATFORM_VIEW_INFO *)MNewPtr(sizeof(PLATFORM_VIEW_INFO))))
	{
		view->viewInfo->leftPixel=0;							// init text to the origin
		view->viewInfo->topLine=0;
		view->viewInfo->bounds=*boundsRect;						// copy bounds rectangle

		view->viewInfo->blinkState=false;						// next time it blinks, it will blink to on
		view->viewInfo->nextBlinkTime=0;						// next time it checks, it will blink

		ViewFontInfoRecalculate(view);							// update view information based on fonts and stuff

		if(!(width=view->viewInfo->bounds.w))
		{
			width=1;
		}
		if(!(height=view->viewInfo->bounds.h))
		{
			height=1;
		}

		// create a window which has the sole purpose of changing the cursor to a caret when it enters the view
		view->viewInfo->cursorWindow=XCreateWindow(xDisplay,view->parentWindow->xWindow,view->viewInfo->bounds.x,view->viewInfo->bounds.y,width,height,0,0,InputOnly,DefaultVisual(xDisplay,xScreenNum),0,NULL);
		XDefineCursor(xDisplay,view->viewInfo->cursorWindow,caretCursor);	// this is the cursor to use in the view
		XMapRaised(xDisplay,view->viewInfo->cursorWindow);		// map window to the top of the pile

		return(true);
	}
	return(false);
}
