// Generic window handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static bool
	hadInvalidation=false;

void EditorDisplayDimensionsGet(uint32_t *width,uint32_t *height)
// Return the dimensions of the total display (across all screens).
// NOTE: this information is looked up each time it is needed. Although
// that's a little inefficient, it means that the display size can be adjusted
// and e93 will adapt on the fly.
{
	XWindowAttributes
		attributes;

	XGetWindowAttributes(xDisplay,RootWindow(xDisplay,xScreenNum),&attributes);
	*width=attributes.width;
	*height=attributes.height;
}

void WindowRectInvalidate(EDITOR_WINDOW *window,EDITOR_RECT *rect)
// Invalidate rect of window, so it is drawn the next time we redraw.
{
	XRectangle
		windowRect;

	if(rect->w&&rect->h)	// NOTE: this gets around a bug in X11 (does not properly handle adding 0 width or height to a region!)
	{
		windowRect.x=rect->x;
		windowRect.y=rect->y;
		windowRect.width=rect->w;
		windowRect.height=rect->h;
		XUnionRectWithRegion(&windowRect,window->invalidRegion,window->invalidRegion);	// add to invalid region of the window
		hadInvalidation=true;
	}
}

void WindowRegionInvalidate(EDITOR_WINDOW *window,Region region)
// Invalidate region of window, so it is drawn the next time we redraw.
{
	XUnionRegion(region,window->invalidRegion,window->invalidRegion);	// make this part invalid
	hadInvalidation=true;
}

void InvalidWindowsRedraw()
// Run though the window list, and redraw anything that
// has invalid area.
// NOTE: for efficiency, this checks a global which tells
// if any invalidations have happened, and if not, it does nothing.
{
	EDITOR_WINDOW
		*window;

	if(hadInvalidation)
	{
		hadInvalidation=false;									// clear now, so updates can cause further invalidation (not that they should want to, but just to be correct)
		window=windowListHead;									// update all windows that need it
		while(window)
		{
			if(window->updateFunction)
			{
				window->updateFunction(window);						// call the window to update
			}
			window=window->nextWindow;
		}
	}
}

void WindowClassHintsSet(EDITOR_WINDOW *window)
// Set the class hints for this window.
// Used to create a class of windows so that window list applications can group things properly.
{
	XClassHint
		classHints;
	char
		uniqueClass[64];

	classHints.res_name=(char *)"e93";
	sprintf(uniqueClass,"e93 (%d)",getpid());
	classHints.res_class=uniqueClass;
	XSetClassHint(xDisplay,window->xWindow,&classHints);
}

void WindowIconSet(EDITOR_WINDOW *window)
// Set the icon for the given window.
{
	unsigned char
		*iconData;
	uint32_t
		iconElements;

	iconData=ImageAssetIconGet(&iconElements);
	XChangeProperty(xDisplay,window->xWindow,iconAtom,XA_CARDINAL,32,PropModeReplace,iconData,iconElements);
}

static Bool MapEventPredicate(Display *display,XEvent *event,char *arg)
// Use this to locate the mapping event for the given window.
{
	if((event->type==MapNotify)&&(event->xmap.window==*((Window *)arg)))
	{
		return(True);
	}
	return(False);
}

static Bool FocusEventPredicate(Display *display,XEvent *event,char *arg)
// Use this to locate the focus event for the given window.
{
	if((event->type==FocusIn)&&(event->xfocus.window==*((Window *)arg)))
	{
		return(True);
	}
	return(False);
}

static Bool DestroyEventPredicate(Display *display,XEvent *event,char *arg)
// Use this to locate the destroy event for the given window (actually, this
// gets ANY event for the window -- that way, no stray events are left after
// the window is gone).
{
	if(event->xmap.window==*((Window *)arg))
	{
		return(True);
	}
	return(False);
}

void WaitUntilMappedNoFocus(EDITOR_WINDOW *window)
// Wait until a mapping event for window arrives.
{
	XEvent
		event;

	XIfEvent(xDisplay,&event,MapEventPredicate,(char *)&window->xWindow);	// wait here until our window is mapped
}

void WaitUntilMapped(EDITOR_WINDOW *window)
// Wait until a mapping event for window arrives, then set the focus to that window, and
// wait for it to complete.
{
	XEvent
		event;

	XIfEvent(xDisplay,&event,MapEventPredicate,(char *)&window->xWindow);	// wait here until our window is mapped
	XSetInputFocus(xDisplay,window->xWindow,RevertToPointerRoot,CurrentTime);
	XIfEvent(xDisplay,&event,FocusEventPredicate,(char *)&window->xWindow);	// wait here until our window has focus
}

void WaitUntilDestroyed(EDITOR_WINDOW *window)
// Wait until a the destroy event for window arrives, eat all other events arriving for the window
// while we wait.
{
	XEvent
		event;

	do
	{
		XIfEvent(xDisplay,&event,DestroyEventPredicate,(char *)&window->xWindow);	// wait here until our window is destroyed
	}
	while(event.type!=DestroyNotify);
}

void WindowElementLink(EDITOR_WINDOW *window)
// Passed an EDITOR_WINDOW, link it to the start of the global list.
{
	if(windowListHead)								// point previous back to new
	{
		windowListHead->previousWindow=window;
	}
	window->nextWindow=windowListHead;				// point to the old head of the list
	window->previousWindow=NULL;					// no previous window
	windowListHead=window;							// point the head at the new element
}

void WindowElementUnlink(EDITOR_WINDOW *window)
// Passed an EDITOR_WINDOW, unlink it from the global list.
{
	if(window->nextWindow)
	{
		window->nextWindow->previousWindow=window->previousWindow;
	}
	if(window->previousWindow)
	{
		window->previousWindow->nextWindow=window->nextWindow;
	}
	else
	{
		windowListHead=window->nextWindow;			// this one was at the head, so update
	}
}

static uint32_t WindowElementTypeMaskCount(uint16_t windowTypeMask)
// Return the total number of editor windows that match windowTypeMask.
{
	uint32_t
		count;
	EDITOR_WINDOW
		*window;

	count=0;
	window=windowListHead;
	while(window)
	{
		if(windowTypeMask&(1<<window->windowType))
		{
			count++;
		}
		window=window->nextWindow;
	}
	return(count);
}

EDITOR_WINDOW *WindowElementLocate(Window xWindow)
// Passed an xWindow, locate the editor window associated with it
// (if there is one). If there is none, return NULL.
{
	EDITOR_WINDOW
		*window;
	bool
		found;

	window=windowListHead;
	found=false;
	while(window&&!found)
	{
		if(window->xWindow==xWindow)
		{
			found=true;
		}
		else
		{
			window=window->nextWindow;
		}
	}
	return(window);
}

void EditorWindowMinimize(EDITOR_WINDOW *window)
// Iconify the passed window.
{
	XIconifyWindow(xDisplay,window->xWindow,xScreenNum);
}

void EditorWindowUnminimize(EDITOR_WINDOW *window)
// Undo what EditorWindowMinimize did.
{
	XMapWindow(xDisplay,window->xWindow);
}

void EditorTopWindowSet(EDITOR_WINDOW *window)
// Set the passed document window to the top of the display
// NOTE: X windows has evolved since e93 was written. It has become even
// more needlessly complex, and now window managers don't always honor
// XRaiseWindow requests (Metacity, I'm looking at you).
// The Window Manager authors apparently feel it is the duty of the WM to
// police which applications can raise windows and when.
// To me, this is nonsense. When I call XRaiseWindow, it should raise
// the damn window.
// Now however, to try to convince these misguided WMs, e93 attempts to follow:
// http://standards.freedesktop.org/wm-spec/wm-spec-1.3.html#id2506353
// by sending a _NET_ACTIVE_WINDOW message to the root window.
// If it works, the effect will be the same as having called XRaiseWindow --
// just more complex and difficult to follow.
{
	XEvent
		xEvent;

	XMapWindow(xDisplay,window->xWindow);					// make sure it's not iconified -- Thanks Michael Manning
	XRaiseWindow(xDisplay,window->xWindow);					// place it on top
	XSetInputFocus(xDisplay,window->xWindow,RevertToPointerRoot,CurrentTime);	// give it the focus

	xEvent.xclient.type=ClientMessage;
	xEvent.xclient.serial=0;
	xEvent.xclient.send_event=True;
	xEvent.xclient.display=xDisplay;
	xEvent.xclient.window=window->xWindow;
	xEvent.xclient.message_type=XInternAtom(xDisplay,"_NET_ACTIVE_WINDOW",True);
	xEvent.xclient.format=32;
	xEvent.xclient.data.l[0]=1;					// 1 tells the WM we're an application
	xEvent.xclient.data.l[1]=EventLastUserEventTime();	// send timestamp of last user event
	xEvent.xclient.data.l[2]=0;					// for now, send the currently active window as 0, later if it matters, figure out what we really need
	xEvent.xclient.data.l[3]=0;
	xEvent.xclient.data.l[4]=0;
	XSendEvent(xDisplay,XDefaultRootWindow(xDisplay),False,SubstructureRedirectMask|SubstructureNotifyMask,&xEvent);
}

static void SortedEditorWindowTypeMaskListFillRecurse(Window *siblings,unsigned int numSiblings,uint16_t windowTypeMask,uint32_t maxElements,uint32_t *numElements,EDITOR_WINDOW **list)
// Run through the siblings list from right to left (top to bottom),
// looking to see if any matches an editor window. If it does, add the window
// to *list, and keep going. For those that don't match, recurse down
// and keep looking. 
// This calls itself recursively.
{
	bool
		match[numSiblings];
	Window
		root,
		parent,
		*children;
	unsigned int
		siblingIndex,
		numChildren;
	EDITOR_WINDOW
		*window;

	// search this level (breadth first)
	siblingIndex=numSiblings;
	while(((*numElements)<maxElements)&&siblingIndex)
	{
		siblingIndex--;
		if((window=WindowElementLocate(siblings[siblingIndex])))
		{
			if(windowTypeMask&(1<<window->windowType))							// does the type match?
			{
				list[*numElements]=window;										// remember this one was found
				(*numElements)++;
			}
			match[siblingIndex]=true;											// matched one of ours at this level, no need to traverse deeper from here
		}
		else
		{
			match[siblingIndex]=false;											// no match -- move deeper if possible
		}
	}

	// now recurse down as needed
	siblingIndex=numSiblings;
	while(((*numElements)<maxElements)&&siblingIndex)
	{
		siblingIndex--;
		if(!match[siblingIndex])												// if no match for this sibling, then recurse into it
		{
			if(XQueryTree(xDisplay,siblings[siblingIndex],&root,&parent,&children,&numChildren))	// get the children of this window (window could be going away, so we must allow this to fail)
			{
				if(children)
				{
					SortedEditorWindowTypeMaskListFillRecurse(children,numChildren,windowTypeMask,maxElements,numElements,list);	// search the level below
					XFree(children);
				}
			}
		}
	}
}

static void SortedEditorWindowTypeMaskListFill(uint16_t windowTypeMask,uint32_t maxElements,uint32_t *numElements,EDITOR_WINDOW **list)
// Fill **list with pointers to editor windows of the types given by windowTypeMask, sorted
// from top to bottom. Stop after maxElements have been put into the list.
{
	Window
		root;

	(*numElements)=0;
	root=RootWindow(xDisplay,xScreenNum);
	SortedEditorWindowTypeMaskListFillRecurse(&root,1,windowTypeMask,maxElements,numElements,list);
}

EDITOR_WINDOW *EditorTopWindowTypeGet(uint16_t windowType)
// Return the top-most window of windowType, or NULL if there is none.
{
	EDITOR_WINDOW
		*window;
	uint32_t
		numElements;

	window=NULL;
	SortedEditorWindowTypeMaskListFill((1<<windowType),1,&numElements,&window);
	return(window);
}

bool SortedEditorWindowTypeMaskListGet(uint16_t windowTypeMask,uint32_t *numElements,EDITOR_WINDOW ***list)
// Return a list of window pointers, sorted top to bottom for the types given by windowTypeMask.
// If there is a problem, ErrorSet, and return false.
// If true is returned, then list must be MDisposePtr'ed later (even if numElements
// is 0).
{
	uint32_t
		maxListElements;
	bool
		fail;

	fail=false;
	maxListElements=WindowElementTypeMaskCount(windowTypeMask);							// get the max number of elements
	if(((*list)=(EDITOR_WINDOW **)MNewPtr(sizeof(EDITOR_WINDOW *)*maxListElements)))	// create a list of pointers large enough to hold the maximum number of windows located
	{
		SortedEditorWindowTypeMaskListFill(windowTypeMask,maxListElements,numElements,*list);
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

EDITOR_WINDOW *EditorFocusWindowGet()
// Return the editor window with the focus, or the top window
// if none has focus.
// If there is no top window, return NULL.
{
	EDITOR_WINDOW
		*window;
	EDITOR_WINDOW
		*locatedWindow;
	Window
		xWindow;
	int
		focusRevert;

	XGetInputFocus(xDisplay,&xWindow,&focusRevert);			// find out who has the focus
	if((locatedWindow=WindowElementLocate(xWindow)))		// see if it is one of ours
	{
		if((locatedWindow->windowType==EWT_DOCUMENT)||(locatedWindow->windowType==EWT_DIALOG))
		{
			return(locatedWindow);							// if there is one of our documents/dialogs with focus, it is preferred
		}
		else
		{
			if((window=EditorTopWindowTypeGet(EWT_DIALOG)))		// try to find the top-most dialog
			{
				return(window);
			}
			else
			{
				if((window=EditorTopWindowTypeGet(EWT_DOCUMENT)))	// try to find the top-most document
				{
					return(window);
				}
				else
				{
					return(locatedWindow);					// no document/dialog, so give back whatever it was we found in the first place
				}
			}
		}
	}
	if((window=EditorTopWindowTypeGet(EWT_DIALOG)))			// if focus window was not ours try to find the top-most dialog
	{
		return(window);
	}
	else
	{
		if((window=EditorTopWindowTypeGet(EWT_DOCUMENT)))	// try to find the top-most document
		{
			return(window);
		}
	}
	return(rootMenuWindow);									// as a last resort, give back the root menu
}

void FocusAwayFrom(EDITOR_WINDOW *window)
// Set the focus to the topmost window that is not window.
// This is done to fool X into not readjusting the focus when we
// delete window.
{
	EDITOR_WINDOW
		*windows[2];										// get two top-most windows in here
	uint32_t
		numElements,
		i;
	bool
		haveFocus;
	Window
		focusWindow;

	SortedEditorWindowTypeMaskListFill((1<<EWT_DOCUMENT)|(1<<EWT_DIALOG)|(1<<EWT_MENU),2,&numElements,windows);	// get the two top-most windows (one of them is guaranteed to not be the passed window)
	haveFocus=false;
	for(i=0;i<numElements;i++)
	{
		if(windows[i]!=window)
		{
			focusWindow=windows[i]->xWindow;
			haveFocus=true;
		}
	}
	if(haveFocus)
	{
		XSetInputFocus(xDisplay,focusWindow,RevertToPointerRoot,CurrentTime);	// give focus to this window
		XSync(xDisplay,false);
	}
}

// ----------------------------------------------------------------------------

void PlatformScreenOffsetAndDimensionsGet(int32_t *xOffset,int32_t *yOffset,uint32_t *width,uint32_t *height)
// Return the offset and dimensions of the primary screen that the editor is
// running on.
// NOTE: On multi-head displays, it is possible that the primary screen
// (the one that e93 is opening windows onto) is not at offset 0,0 within
// the entire display.
// NOTE: this information is looked up each time it is needed. Although
// that's a little inefficient, it means that the screens can be adjusted
// and e93 will adapt on the fly.
{
	XineramaScreenInfo
		*screenInfo;
	int
		numHeads;
	XWindowAttributes
		attributes;

	if((screenInfo=XineramaQueryScreens(xDisplay,&numHeads)))	// success implies numHeads>0
	{
		// use the 0th display as the primary one
		*xOffset=screenInfo[0].x_org;
		*yOffset=screenInfo[0].y_org;
		*width=screenInfo[0].width;
		*height=screenInfo[0].height;
		XFree(screenInfo);
	}
	else	// single head
	{
		XGetWindowAttributes(xDisplay,RootWindow(xDisplay,xScreenNum),&attributes);
		*xOffset=0;
		*yOffset=0;
		*width=attributes.width;
		*height=attributes.height;
	}
}
