// X windows clipboard/selection handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

// Some notes about X11 clipboard/selection handling.
//
// First, as is typical for concepts in X, the clipboard is both complex,
// and poorly thought out.
// The application is left to do a lot of work, and a number of the steps
// can go wrong (or be done wrong) in subtle ways.
// See: http://tronche.com/gui/x/icccm/sec-2.html
//
// X has two concepts of the clipboard. The first involves what's
// selected in a window (PRIMARY), and the second is more like the
// traditional cut/copy/paste (CLIPBOARD).
//
// Two concepts aren't better than one. They cause interoperation
// problems between programs. Some programs don't understand the
// cut/copy/paste model, while others don't easily support the odd
// selection model. For instance e93 allows the user to have selections
// present in every text window. e93 doesn't unhighlight a selection when
// another is made in a different window. Although this is natural, and
// very useful, it doesn't play very well with the X selection concept.
// xterm and xpdf on the other hand don't support the cut/copy/paste model.
// Instead, they only use the selection concept.
// Because of this, the user is left with an inconsistent system, and the
// frustrations caused by it.
// To help work around this, e93 tries to support both models at the
// same time. When the user asks the editor to cut or copy, e93 will claim
// ownership of both PRIMARY, and CLIPBOARD. So, someone pasting
// into a word processor, or clicking the middle mouse button in xterm will
// get a behavior that is unsurprising.
// On the flip side, when importing the clipboard, e93 looks to see if the
// PRIMARY selection is owned by another window. If so, it'll import that
// selection as its clipboard, otherwise it continues using its internal
// clipboard.
// The reason e93 doesn't import from the CLIPBOARD is that most programs
// seem to have a way to modify PRIMARY, but not all seem to be able to
// modify CLIPBOARD, and we want to be able to import from as many
// programs as possible. This isn't perfect, but works reasonably well.

#define	SELECTION_RECEIVE_TIMEOUT	(SECOND*5)		// amount of time to wait before deciding that the selection owner is not cooperating
#define	SELECTION_SEND_TIMEOUT		(SECOND*5)		// amount of time to wait before deciding that the selection requestor is not cooperating

static unsigned long
	maxXSelectionChunkSize;							// tells the maximum number of bytes we'll transfer in one chunk when sending selections to X

static Bool NotifyPredicate(Display *display,XEvent *event,char *arg)
// Use this to wait for the notification that the selection has been converted.
{
	if(event->type==SelectionNotify&&event->xselection.requestor==*((Window *)arg))
	{
		return(True);
	}
	return(False);
}

static bool SelectionNotifyGet(Window xWindow,XEvent *event)
// Wait here until either a time out, or until we get notification of
// selection data placed on a property of the given window.
// Return event, or false if there is a timeout.
{
	uint32_t
		selectionTimer;
	bool
		gotEvent;

	selectionTimer=PlatformMonotonicTime();							// reset to current time

	gotEvent=false;
	while(((PlatformMonotonicTime()-selectionTimer)<SELECTION_RECEIVE_TIMEOUT)&&!gotEvent)	// wait here until the notification event arrives, or we time out
	{
		gotEvent=XCheckIfEventTimed(xDisplay,event,NotifyPredicate,(char *)&xWindow,SECOND/100);
	}
	return(gotEvent);
}

static Bool PropertyPredicate(Display *display,XEvent *event,char *arg)
// Use this to wait for the our property to change, so we know more data
// has arrived during incremental selection update.
{
	if(event->type==PropertyNotify&&event->xproperty.window==*((Window *)arg))
	{
		return(True);
	}
	return(False);
}

static bool PropertyChangeNotifyGet(Window xWindow,XEvent *event)
// Wait here until either a time out, or until we get notification of
// property change in the given window (NOTE: property delete messages
// are read, and ignored here).
// Return event, or false if there is a timeout.
{
	uint32_t
		selectionTimer;
	bool
		gotEvent;

	selectionTimer=PlatformMonotonicTime();					// reset to current time

	gotEvent=false;
	while(((PlatformMonotonicTime()-selectionTimer)<SELECTION_RECEIVE_TIMEOUT)&&!gotEvent)	// wait here until the notification event arrives, or we time out
	{
		if(XCheckIfEventTimed(xDisplay,event,PropertyPredicate,(char *)&xWindow,SECOND/100))
		{
			if(event->xproperty.state==PropertyNewValue)
			{
				gotEvent=true;						// only have the event when it is a new value
			}
		}
	}
	return(gotEvent);
}

static void PropertyNotifyFlush(Display *display,Window xWindow)
// Remove all property notify events that are in the queue
// for xWindow on display.
{
	XEvent
		event;

	while(XCheckIfEvent(display,&event,PropertyPredicate,(char *)&xWindow))
		;
}

static bool PropertyDeleteNotifyGet(Display *display,Window xWindow,XEvent *event)
// Wait here until either a time out, or until we get notification of
// property delete in the given window (NOTE: property change messages
// are read, and ignored here).
// Return event, or false if there is a timeout.
{
	uint32_t
		selectionTimer;
	bool
		gotEvent;

	selectionTimer=PlatformMonotonicTime();							// reset to current time

	gotEvent=false;
	while(((PlatformMonotonicTime()-selectionTimer)<SELECTION_SEND_TIMEOUT)&&!gotEvent)	// wait here until the notification event arrives, or we time out
	{
		if(XCheckIfEventTimed(display,event,PropertyPredicate,(char *)&xWindow,SECOND/100))
		{
			if(event->xproperty.state==PropertyDelete)
			{
				gotEvent=true;						// only have the event when it is a new value
			}
		}
	}
	return(gotEvent);
}

static bool SelectionPropertyInhale(Window xWindow,Atom atom,EDITOR_BUFFER *clipboard,uint32_t *numBytesInhaled)
// Read all the bytes of the selection property that are arriving, and append them to
// clipboard. If there is a problem, report it here, return false.
{
	Atom
		returnType;
	unsigned long
		numItems,
		numItemsLeft,
		offset;
	int
		formatReturn;
	unsigned char
		*data;
	bool
		done,
		fail;

	done=fail=false;
	offset=0;
	*numBytesInhaled=0;
	while(!done&&!fail)
	{
		if(XGetWindowProperty(xDisplay,xWindow,selectionAtom,offset,16384,False,atom,&returnType,&formatReturn,&numItems,&numItemsLeft,&data)==Success)
		{
			if(returnType==atom&&formatReturn==8)
			{
				(*numBytesInhaled)+=numItems;
				offset+=numItems>>2;

				EditorTextReplace(clipboard,TextUniverseTotalBytes(clipboard->textUniverse),0,(uint8_t *)data,numItems,false);	// take the bytes we got, and insert them into the clipboard at the end
				done=!numItemsLeft;
			}
			else
			{
				ErrorSet("Got unexpected atom from selection owner: %d",returnType);
				fail=true;
			}
			XFree(data);
		}
		else
		{
			ErrorSet("Failed to get selection property");
			fail=true;
		}
	}
	return(!fail);
}

static bool INCRSelectionImport(Atom atom,EDITOR_BUFFER *clipboard)
// The selection that is being passed to us is so large that the server
// cannot handle it, so fall back to the INCR solution. This works by
// passing the selection incrementally between the clients, making sure
// never to overflow the server with data.
// If there is a problem getting the selection, report it here, and return false.
{
	XEvent
		event;
	Atom
		returnType;
	unsigned long
		numItems,
		numItemsLeft;
	int
		formatReturn;
	unsigned char
		*incrData;
	bool
		done,
		fail;
	Window
		rootXWindow;
	uint32_t
		numBytesInhaled;

	rootXWindow=rootMenuWindow->xWindow;
	fail=false;
	// read to get INCR value which tells how much data will be transferred
	if(XGetWindowProperty(xDisplay,rootXWindow,selectionAtom,0,1,False,incrAtom,&returnType,&formatReturn,&numItems,&numItemsLeft,&incrData)==Success)
	{
		PropertyNotifyFlush(xDisplay,rootXWindow);						// get rid of any pending notify events
		XDeleteProperty(xDisplay,rootXWindow,selectionAtom);			// once the property is read, delete it (this tells sender that we are ready for more)

		if(returnType==incrAtom&&formatReturn==32)
		{
			EditorReplaceStart(clipboard);
			EditorTextReplace(clipboard,0,TextUniverseTotalBytes(clipboard->textUniverse),NULL,0,false);	// clear the clipboard
			done=fail=false;
			while(!done&&!fail)
			{
				if(PropertyChangeNotifyGet(rootXWindow,&event))			// wait for it to change, to see if there is more data
				{
					if(SelectionPropertyInhale(rootXWindow,atom,clipboard,&numBytesInhaled))
					{
						if(numBytesInhaled==0)
						{
							done=true;									// done when property of 0 size is encountered
						}
					}
					else
					{
						fail=true;
					}
					XDeleteProperty(xDisplay,rootXWindow,selectionAtom);	// delete this, so we can go again
				}
				else
				{
					ErrorSet("Timed out waiting for reply from selection owner");
					fail=true;
				}
			}
			EditorReplaceEnd(clipboard);
		}
		else
		{
			ErrorSet("Got unexpected atom from selection owner: %d",returnType);
			fail=true;
		}
		XFree(incrData);
	}
	else
	{
		ErrorSet("Failed to get selection property");
		fail=true;
	}
	return(!fail);
}

static bool NormalSelectionImport(Atom atom,EDITOR_BUFFER *clipboard)
// The selection owner has decided to pass the selection to us in the
// normal way (used when the selection is small enough to be held by the server).
// If there is a problem getting the selection, report it here, and return false.
{
	Window
		rootXWindow;
	uint32_t
		numBytesInhaled;
	bool
		fail;

	rootXWindow=rootMenuWindow->xWindow;
	EditorReplaceStart(clipboard);
	EditorTextReplace(clipboard,0,TextUniverseTotalBytes(clipboard->textUniverse),NULL,0,false);	// clear the clipboard
	fail=!SelectionPropertyInhale(rootXWindow,atom,clipboard,&numBytesInhaled);
	XDeleteProperty(xDisplay,rootXWindow,selectionAtom);								// get rid of the property
	EditorReplaceEnd(clipboard);
	return(!fail);
}

static bool AttemptSelectionConversion(Window window,Atom atom)
// See if we can get the selection owner to convert the selection to the type
// given by atom.
// If we can, return true, otherwise return false.
{
	XEvent
		event;

	XConvertSelection(xDisplay,XA_PRIMARY,atom,selectionAtom,window,CurrentTime);

	if(SelectionNotifyGet(window,&event))			// wait until notified of selection placed on our window
	{
		if(event.xselection.property==selectionAtom)	// if the one with the selection could convert it, then we have it now
		{
			return(true);
		}
	}
	return(false);
}

static bool FindSelectionConversion(Window window,Atom *atom)
// See if we can get the selection owner to convert the selection to a
// type we understand. If so, return true with atom set to the type.
// If not, return false.
{
	if(AttemptSelectionConversion(window,utf8StringAtom))
	{
		*atom=utf8StringAtom;
		return(true);
	}
	if(AttemptSelectionConversion(window,XA_STRING))
	{
		*atom=XA_STRING;
		return(true);
	}
	return(false);
}

static void SelectionRequestRefuse(XEvent *event)
// Complain to the requestor that we do not wish to fulfill his pitiful
// request for data.
{
	XEvent
		notifyEvent;

	notifyEvent.xselection.type=SelectionNotify;
	notifyEvent.xselection.serial=0;
	notifyEvent.xselection.send_event=False;
	notifyEvent.xselection.display=event->xselectionrequest.display;
	notifyEvent.xselection.requestor=event->xselectionrequest.requestor;
	notifyEvent.xselection.selection=event->xselectionrequest.selection;
	notifyEvent.xselection.target=event->xselectionrequest.target;
	notifyEvent.xselection.property=None;
	notifyEvent.xselection.time=event->xselectionrequest.time;
	XSendEvent(event->xselectionrequest.display,event->xselectionrequest.requestor,False,0,&notifyEvent);
}

static void TargetsRequestHandle(XEvent *event)
// Handle sending the list of targets we support.
// Currently, we only support the TARGETS target, the STRING target, and the UTF8_STRING target.
{
	Atom
		types[]={targetsAtom,XA_STRING,utf8StringAtom};	// the types of targets we support
	XEvent
		response;							// event we send to respond to the request

	XChangeProperty(event->xselectionrequest.display,event->xselectionrequest.requestor,event->xselectionrequest.property,XA_ATOM,32,PropModeReplace,(unsigned char *)types,(int)(sizeof(types)/sizeof(types[0])));

	// set values for the response event
	response.xselection.property=event->xselectionrequest.property;
	response.xselection.type=SelectionNotify;
	response.xselection.display=event->xselectionrequest.display;
	response.xselection.requestor=event->xselectionrequest.requestor;
	response.xselection.selection=event->xselectionrequest.selection;
	response.xselection.target=event->xselectionrequest.target;
	response.xselection.time=event->xselectionrequest.time;

	XSendEvent(event->xselectionrequest.display,event->xselectionrequest.requestor,0,0,&response);	// send the response event
	XFlush(event->xselectionrequest.display);
}

static void NormalSelectionRequestHandle(XEvent *event,EDITOR_BUFFER *clipboard)
// Handle sending the contents of the selection in the traditional way.
{
	uint8_t
		*buffer;
	uint64_t
		totalBytes;
	XEvent
		notifyEvent;
	bool
		sentOk;

	sentOk=false;
	totalBytes=TextUniverseTotalBytes(clipboard->textUniverse);
	if((buffer=(uint8_t *)MNewPtr(totalBytes)))										// create a buffer to give to X
	{
		TextUniverseTextExtract(clipboard->textUniverse,0,buffer,totalBytes,NULL);		// read data from universe into buffer
		XChangeProperty(event->xselectionrequest.display,event->xselectionrequest.requestor,event->xselectionrequest.property,event->xselectionrequest.target,8,PropModeReplace,(unsigned char *)buffer,totalBytes);

		notifyEvent.xselection.type=SelectionNotify;
		notifyEvent.xselection.serial=0;
		notifyEvent.xselection.send_event=False;
		notifyEvent.xselection.display=event->xselectionrequest.display;
		notifyEvent.xselection.requestor=event->xselectionrequest.requestor;
		notifyEvent.xselection.selection=event->xselectionrequest.selection;
		notifyEvent.xselection.target=event->xselectionrequest.target;
		notifyEvent.xselection.property=event->xselectionrequest.property;
		notifyEvent.xselection.time=event->xselectionrequest.time;
		XSendEvent(event->xselectionrequest.display,event->xselectionrequest.requestor,False,0,&notifyEvent);
		sentOk=true;
		MDisposePtr(buffer);
	}
	if(!sentOk)
	{
		SelectionRequestRefuse(event);
	}
}

static void INCRSelectionRequestHandle(XEvent *event,EDITOR_BUFFER *clipboard)
// Handle sending the contents of the selection in the incremental way.
{
	TEXT_STREAM
		stream;
	uint8_t
		*buffer;
	XEvent
		notifyEvent;
	unsigned long
		bytesThisTime,
		bytesToDo;
	bool
		fail;

	bytesToDo=TextUniverseTotalBytes(clipboard->textUniverse);			// make local to hold total bytes, pass a pointer to ChangeProperty routine
	if((buffer=(uint8_t *)MNewPtr(maxXSelectionChunkSize)))				// create a buffer to give to X
	{
		XSelectInput(event->xselectionrequest.display,event->xselectionrequest.requestor,PropertyChangeMask);
		PropertyNotifyFlush(event->xselectionrequest.display,event->xselectionrequest.requestor);		// get rid of any pending notify events
		XChangeProperty(event->xselectionrequest.display,event->xselectionrequest.requestor,event->xselectionrequest.property,incrAtom,32,PropModeReplace,(unsigned char *)&bytesToDo,1);	// send INCR to requestor to let him know we will do it the long way

		notifyEvent.xselection.type=SelectionNotify;
		notifyEvent.xselection.serial=0;
		notifyEvent.xselection.send_event=False;
		notifyEvent.xselection.display=event->xselectionrequest.display;
		notifyEvent.xselection.requestor=event->xselectionrequest.requestor;
		notifyEvent.xselection.selection=event->xselectionrequest.selection;
		notifyEvent.xselection.target=event->xselectionrequest.target;
		notifyEvent.xselection.property=event->xselectionrequest.property;
		notifyEvent.xselection.time=event->xselectionrequest.time;
		XSendEvent(event->xselectionrequest.display,event->xselectionrequest.requestor,False,0,&notifyEvent);

		TextStreamPositionSet(&stream,clipboard->textUniverse,0);		// start at the beginning
		fail=false;
		while(bytesToDo&&!fail)
		{
			if(bytesToDo>maxXSelectionChunkSize)
			{
				bytesThisTime=maxXSelectionChunkSize;
			}
			else
			{
				bytesThisTime=bytesToDo;
			}
			TextStreamTextExtract(&stream,buffer,bytesThisTime,NULL);	// read data from universe into buffer
			if(PropertyDeleteNotifyGet(event->xselectionrequest.display,event->xselectionrequest.requestor,&notifyEvent))
			{
				XChangeProperty(event->xselectionrequest.display,event->xselectionrequest.requestor,event->xselectionrequest.property,event->xselectionrequest.target,8,PropModeReplace,(unsigned char *)buffer,bytesThisTime);
				bytesToDo-=bytesThisTime;
			}
			else
			{
				fail=true;
			}
		}
		if(!fail)
		{
			if(PropertyDeleteNotifyGet(event->xselectionrequest.display,event->xselectionrequest.requestor,&notifyEvent))
			{
				XChangeProperty(event->xselectionrequest.display,event->xselectionrequest.requestor,event->xselectionrequest.property,event->xselectionrequest.target,8,PropModeReplace,(unsigned char *)buffer,0);
			}
		}
		XSelectInput(event->xselectionrequest.display,event->xselectionrequest.requestor,0);	// no longer want events for this window
		MDisposePtr(buffer);
	}
	else
	{
		SelectionRequestRefuse(event);
	}
}

// ----------------------------------------------------------------------------

void SelectionRequestHandle(XEvent *event)
// Someone wants the contents of a selection (which we own).
// Send the current clipboard contents to them.
{
	EDITOR_BUFFER
		*clipboard;

	if((event->xselectionrequest.selection==XA_PRIMARY)||(event->xselectionrequest.selection==XA_CLIPBOARD(xDisplay)))	// handle those selections we know about, refuse the rest
	{
		if((event->xselectionrequest.target==targetsAtom))
		{
			TargetsRequestHandle(event);	// client asking what targets we support
		}
		else if((event->xselectionrequest.target==XA_STRING)||(event->xselectionrequest.target==utf8StringAtom))
		{
			if((clipboard=ClipboardExportStart()))
			{
				if(TextUniverseTotalBytes(clipboard->textUniverse)>maxXSelectionChunkSize)		// if too large, do it the incremental way
				{
					INCRSelectionRequestHandle(event,clipboard);
				}
				else
				{
					NormalSelectionRequestHandle(event,clipboard);
				}
				ClipboardExportEnd();
			}
			else	// failed to start exporting the clipboard
			{
				SelectionRequestRefuse(event);
			}
		}
		else	// don't know about the target that's being requested
		{
			SelectionRequestRefuse(event);
		}
	}
	else	// don't know about the selection that's being requested
	{
		SelectionRequestRefuse(event);
	}
}

// ----------------------------------------------------------------------------

bool PlatformClipboardImport()
// When the editor is about to do a paste, or something else
// that requires it to look at the contents of the clipboard
// it will call this to possibly import outside contents into
// the current clipboard.
// NOTE: this should return true if it succeeds, false if it fails.
// NOTE: if no importing is done, this should still return true.
{
	Window
		xWindow,
		rootXWindow;
	bool
		fail;
	Atom
		conversionType,
		returnType;
	unsigned long
		numItems,
		numItemsLeft;
	int
		formatReturn;
	unsigned char
		*data;
	EDITOR_BUFFER
		*clipboard;

	fail=false;
	xWindow=XGetSelectionOwner(xDisplay,XA_PRIMARY);			// see if we own the selection, if so, do not import it
	rootXWindow=rootMenuWindow->xWindow;
	if(xWindow!=None&&xWindow!=rootXWindow)						// someone else owns it, so import it
	{
		if((clipboard=ClipboardImportStart()))
		{
			if(FindSelectionConversion(rootXWindow,&conversionType))
			{
				if(XGetWindowProperty(xDisplay,rootXWindow,selectionAtom,0,0,False,AnyPropertyType,&returnType,&formatReturn,&numItems,&numItemsLeft,&data)==Success)
				{
					if(returnType==conversionType&&formatReturn==8)
					{
						NormalSelectionImport(conversionType,clipboard);	// attempt to import the selection the traditional way
					}
					else
					{
						if(returnType==incrAtom&&formatReturn==32)			// see if other client wants to do this incrementally
						{
							INCRSelectionImport(conversionType,clipboard);	// attempt to import the selection the incremental way
						}
						else
						{
							ErrorSet("Unknown selection format (Atom=%d)",returnType);
							fail=true;
						}
					}
					XFree(data);
				}
				else
				{
					ErrorSet("Failed to get selection property");
					fail=true;
				}
			}
			else
			{
				ErrorSet("Selection owner couldn't convert to string");
				fail=true;
			}
			ClipboardImportEnd();
		}
	}
	return(!fail);
}

bool PlatformClipboardExport()
// When the editor changes the contents of a clipboard, this is called
// so the GUI can export the clipboard into its current environment
// if there is a problem, return false.
{
	XSetSelectionOwner(xDisplay,XA_CLIPBOARD(xDisplay),rootMenuWindow->xWindow,CurrentTime);		// let the world know that the clipboard is mine (yes, CurrentTime should not be used, but there is no good alternative, since scripts can change the clipboard at any time)
	XSetSelectionOwner(xDisplay,XA_PRIMARY,rootMenuWindow->xWindow,CurrentTime);					// also take the primary selection
	return(true);
}

void XClipboardUnInit()
// Undo what ClipboardInit did.
{
	// nothing needed here
}

bool XClipboardInit()
// Initialize handling of X11 clipboard/selections.
{
	// consider selections larger than a quarter of the maximum request size to be "large". See ICCCM section 2.5
	if(!(maxXSelectionChunkSize=XExtendedMaxRequestSize(xDisplay)/4))	// if that's not available, then use unextended version
	{
		maxXSelectionChunkSize=XMaxRequestSize(xDisplay)/4;
	}
	return(true);
}
