// Regular expression handling
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define MAX_STRING_PER_NODE			64					// maximum number of characters in a string within a node
#define	MAX_RECURSION_DEPTH			32768				// maximum amount of recursion we deem allowable

typedef enum
{
	// NOTE: the order of these IS important, there are tables below that match this order
	PT_STRING=0,				// matching constant string
	PT_LIST,					// matching one character to see if it is in a list
	PT_PREVIOUSMATCH,			// matching against data found in previous match
	PT_SUBEXPRESSIONSTART,		// matching against data found in a sub expression
	PT_SUBEXPRESSIONEND,		// end of sub expression match
	PT_EDGE,					// edge match
} PATTERN_TYPE;

// Mode used with PT_EDGE that describes how the test should be performed
typedef enum
{
	EM_LEFT,					// match if the character on the left is on the list
	EM_RIGHT,					// match if the character on the right is on the list
	EM_LEFT_BOTH,				// match if the character on the left is on the list, and the character on the right isn't
	EM_RIGHT_BOTH,				// match if the character on the right is on the list, and the character on the left isn't
	EM_TOGGLE,					// match if one side is on the list and the other isn't
} EDGE_MODE;

typedef struct
{
	TEXT_STREAM
		startStream;			// stream that tells where this match occurred (stream position is invalid until there's a match)
	uint64_t
		numBytes;				// number of bytes matched by this register
	uint64_t
		minimumMatchLength,		// used only during analysis, tells the minimum length of a pattern that could be matched by this register (0 means it can match nothing)
		maximumMatchLength;		// used only during analysis, tells the maximum length of a pattern that could be matched by this register
	bool
		infiniteMatchLength;	// used only during analysis, tells if the maximum length of a pattern could be infinite (in which case maximumMatchLength should be ignored)
} REGISTER;

typedef struct expressionNode
{
	PATTERN_TYPE
		type;					// type of pattern this element represents
	uint8_t
		dataLength;				// length of string data if it exists
	union
	{
		struct
		{
			uint32_t
				data[MAX_STRING_PER_NODE];	// code points of the string to compare against (if string type)
		} re_string;
		struct
		{
			CODE_POINT_LIST
				*codePointList;		// tell if a character is or is not to be accepted (in a list)
		} re_list;
		struct
		{
			REGISTER
				*expRegister;		// points to the register which holds the start and end of the last match we are to compare against (for PT_PREVIOUSMATCH type)
		} re_previous;
		struct
		{
			struct expressionListHeader
				*subExpressionList;	// pointer to expression list for this subexpression
			uint64_t
				currentCount;		// current number of successful iterations of the subexpression
			REGISTER
				*expRegister;		// points to the register to remember match of subexpression in (NULL if none remembered)
			TEXT_STREAM
				lastStream;			// stream that was active last time
			uint64_t
				lastNumToSearch;	// number of bytes we were asked to search last time
			bool
				justEntered;		// tells if just entered this subexpression
		} re_subexpression;
		struct
		{
			EDGE_MODE
				mode;				// tells how to perform the match
			bool
				docEdgeMatches;		// if true, document edge is considered the same as a list match, otherwise it isn't
			CODE_POINT_LIST
				*codePointList;		// list of characters that must match on the matching edge (and must not match on the other side if testBothSides is set)
		} re_edge;
	};
	bool
		matchSpecified;			// user has specified a match limit for this node
	uint64_t
		matchMin,				// minimum number of times to match pattern (0 means we dont even have to match this)
		matchMax;				// maximum number of times to match pattern (0 means infinite)
	struct expressionNode
		*nextExpression,		// next expression along the line at this level
		*parentExpression;		// parent expression, or NULL if at top
} EXPRESSION_NODE;

typedef struct expressionListHeader
{
	EXPRESSION_NODE
		*parentExpression;		// parent that got us here, or NULL if at top
	EXPRESSION_NODE
		*firstExpression;		// first expression in the list for this level
	struct expressionListHeader
		*nextAlternate;			// next alternate expression list for this level (NULL for no more)
} EXPRESSION_LIST_HEADER;

typedef struct compiledExpression
{
	bool
		ignoreCase;				// tells if the expression was compiled to ignore case (used during matching to know if we should translate or not)
	EXPRESSION_LIST_HEADER
		*firstList;				// points to the first top-level expression list
	uint32_t
		numRegisters;			// incremented every time a register is defined during compile
	bool
		registerDefined[MAX_REGISTERS];	// used during compilation to remember which registers have been defined already
	REGISTER
		expRegisters[MAX_REGISTERS];	// the registers that keep track of various match points
	CODE_POINT_LIST
		*charsMatched;			// contains list of all characters that can possibly be matched by this expression
	CODE_POINT_LIST
		*startCharsMatched;		// contains list of all characters that can possibly be matched by this expression at the start
	uint64_t
		minimumMatchLength,		// tells the minimum length of a pattern that could be matched by this expression (0 means it can match nothing)
		maximumMatchLength;		// tells the maximum length of a pattern that could be matched by this expression
	bool
		infiniteMatchLength;	// tells if the maximum length of a pattern could be infinite (in which case maximumMatchLength should be ignored)
} COMPILED_EXPRESSION;

typedef struct
{
	bool
		useUtf8;				// tells if comparisons should be made by interpreting text as UTF-8
	bool
		ignoreCase;				// tells if case should be ignored when matching
	bool
		leftEdge,				// tells if search is starting on logical left edge (used for edge match operations < and ^)
		rightEdge;				// tells if search is starting on logical right edge (used for edge match operations > and $)
	bool
		foundMatch;				// set true if a match is located
	uint64_t
		initialNumToSearch;		// tells how many bytes we were asked to search through at the start
	TEXT_STREAM
		endStream;				// points just past the match
	uint32_t
		matchRecursionDepth;	// used to keep track of how far we are recursing during match attempts, allows us to fail without running out of stack
	ABORT_TEST_FUNCTION
		*abortFunction;			// function to test for abort (or NULL if none)
} RECURSE_GLOBALS;

enum
{
	UNBALANCED_PAREN,
	UNBALANCED_SQUARE_BRACKET,
	BAD_RANGE_EXPRESSION,
	BAD_REPETITION,
	INVALID_REGISTER,
	DANGLING_QUOTE,
	INVALID_DIGIT,
	INVALID_HEX,
	MISSING_DIGIT,
	STACK_OVERFLOW,
	SEARCH_ABORTED
};

static const char *errorDescriptions[]=
{
	"Unbalanced ()",
	"Unbalanced [ or [^",
	"Invalid {} expression",
	"Repetition operator *, +, ?, or {} applied to nothing",
	"Invalid register specification",
	"Incomplete \\ expression",
	"Invalid digit",
	"Invalid \\x expression",
	"Digit expected, but not found",
	"Expression stack overflow",
	"Search Aborted"
};

static void RegexFreeNode(EXPRESSION_NODE *expressionNode)
// Free the node given by expressionNode.
{
	MDisposePtr(expressionNode);
}

static EXPRESSION_NODE *RegexAllocateNode()
// Allocate, and return an expression node.
// If there is a problem, ErrorSet, and return NULL.
{
	return((EXPRESSION_NODE *)MNewPtr(sizeof(EXPRESSION_NODE)));
}

static void RegexFreeList(EXPRESSION_LIST_HEADER *expressionList)
// Free the list given by expressionList.
{
	MDisposePtr(expressionList);
}

static EXPRESSION_LIST_HEADER *RegexAllocateList()
// Allocate, and return an expression list header.
// If there is a problem, ErrorSet, and return NULL.
{
	return((EXPRESSION_LIST_HEADER *)MNewPtr(sizeof(EXPRESSION_LIST_HEADER)));
}

static bool RegexHexDigitGet(TEXT_STREAM *stream,uint8_t *nibble)
// Get a hex digit from the input stream, and convert it to a nibble.
// If there is a problem, set the error, and return false.
{
	bool
		fail;
	uint32_t
		codePoint;

	fail=false;
	codePoint='\0';
	if(!TextStreamAtEnd(stream))
	{
		codePoint=SearchToLower(Utf8NextCodePoint(stream));
		if(codePoint>='a')
		{
			codePoint-='a'-10;			// make letter into number
			if(codePoint>15)
			{
				ErrorSet(errorDescriptions[INVALID_DIGIT]);
				fail=true;
			}
		}
		else
		{
			if(codePoint>='0')
			{
				codePoint-='0';
				if(codePoint>9)
				{
					ErrorSet(errorDescriptions[INVALID_DIGIT]);
					fail=true;
				}
			}
			else
			{
				ErrorSet(errorDescriptions[INVALID_DIGIT]);
				fail=true;
			}
		}
	}
	else
	{
		ErrorSet(errorDescriptions[MISSING_DIGIT]);
		fail=true;
	}
	(*nibble)=codePoint;
	return(!fail);
}

static bool RegexQuotedCharacterSingleHexConvert(TEXT_STREAM *stream,uint32_t *codePoint)
// Try to interpret the stream as 'xx' which is a 2 digit hex
// sequence. If successful, fill in codePoint, and return true.
// If this fails, ErrorSet and return false.
// stream is always pushed past the last character read.
{
	uint8_t
		highNibble,
		lowNibble;
	bool
		result;

	result=false;
	if(RegexHexDigitGet(stream,&highNibble))
	{
		if(RegexHexDigitGet(stream,&lowNibble))
		{
			*codePoint=highNibble*16+lowNibble;			// create actual character
			result=true;
		}
	}
	return(result);
}

static bool RegexQuotedCharacterLongHexConvert(TEXT_STREAM *stream,uint32_t *codePoint)
// Try to interpret the stream as '{xxxxxxxx}' which is an arbitrarily long
// string of hex digits. If successful, fill in codePoint, and return true.
// If this fails, ErrorSet and return false.
// stream is always pushed past the last character read.
{
	TEXT_STREAM
		tempStream;
	uint32_t
		streamCodePoint;
	bool
		done,
		result;
	uint8_t
		nibble;

	result=true;
	(*codePoint)=0;
	if(!TextStreamAtEnd(stream))
	{
		streamCodePoint=Utf8NextCodePoint(stream);
		if(streamCodePoint=='{')
		{
			done=false;
			while(!done&&result)
			{
				if(!TextStreamAtEnd(stream))
				{
					tempStream=*stream;
					streamCodePoint=Utf8NextCodePoint(stream);
					if(streamCodePoint=='}')
					{
						done=true;
					}
					else
					{
						*stream=tempStream;
						if(RegexHexDigitGet(stream,&nibble))
						{
							(*codePoint)=((*codePoint)<<4)|nibble;
						}
						else
						{
							result=false;
						}
					}
				}
				else
				{
					ErrorSet(errorDescriptions[MISSING_DIGIT]);
					result=false;
				}
			}
		}
		else
		{
			ErrorSet(errorDescriptions[INVALID_HEX]);
			result=false;
		}
	}
	else
	{
		ErrorSet(errorDescriptions[INVALID_HEX]);
		result=false;
	}
	return(result);
}

static bool RegexQuotedCharacterHexConvert(TEXT_STREAM *stream,uint32_t *codePoint)
// Convert the parameter of a \x request.
// If successful, fill in codePoint, and return true.
// If this fails, ErrorSet and return false.
// stream is always pushed past the last character read.
{
	TEXT_STREAM
		tempStream;
	bool
		result;

	tempStream=*stream;
	if(!(result=RegexQuotedCharacterSingleHexConvert(stream,codePoint)))		// failed to do it the short way?
	{
		ErrorTraceClear();
		*stream=tempStream;														// put back at the start
		result=RegexQuotedCharacterLongHexConvert(stream,codePoint);			// see if it can be done the long way
	}
	return(result);
}

static bool RegexQuotedCharacterConvert(TEXT_STREAM *stream,uint32_t *codePoint)
// Convert character after a \ into what it actually represents
// (for instance \n means newline).
// This can fail if there are no more characters after the \, or
// if the characters do not make sense.
{
	bool
		fail;

	fail=false;
	if(!TextStreamAtEnd(stream))
	{
		*codePoint=Utf8NextCodePoint(stream);
		switch(*codePoint)
		{
			case 'b':
				*codePoint=0x08;
				break;
			case 'n':
				*codePoint='\n';
				break;
			case 'r':
				*codePoint='\r';
				break;
			case 't':
				*codePoint='\t';
				break;
			case 'x':
				fail=!RegexQuotedCharacterHexConvert(stream,codePoint);
				break;
		}
	}
	else
	{
		ErrorSet(errorDescriptions[DANGLING_QUOTE]);
		fail=true;
	}
	return(!fail);
}

static EXPRESSION_NODE *RegexStringAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,uint32_t codePoint)
// Add codePoint to currentNode. If currentNode is NULL, or full, or not of the correct type
// then create a new node, and make codePoint the first of the string.
// NOTE: if currentNode was full, the new node is linked to it's nextExpression.
// Return the node that codePoint was added to.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if(compiledExpression->ignoreCase)
	{
		codePoint=SearchToLower(codePoint);
	}

	if(currentNode&&(currentNode->type==PT_STRING)&&(!currentNode->matchSpecified)&&(currentNode->dataLength<MAX_STRING_PER_NODE))
	{
		node=currentNode;							// there is space in the current node, so add this in
		node->re_string.data[node->dataLength++]=codePoint;	// drop in the data, bump the amount
	}
	else
	{
		if((node=RegexAllocateNode()))
		{
			node->type=PT_STRING;					// create a node, and apply the correct type to it
			node->parentExpression=parent;			// set the parent
			node->dataLength=1;						// so far, it contains one byte of string data
			node->re_string.data[0]=codePoint;				// drop in the data
			node->matchSpecified=false;				// no match count specified for this node yet
			node->matchMin=1;						// until further notice, this must match exactly once
			node->matchMax=1;
			node->nextExpression=NULL;				// this is now the last expression
			if(currentNode)
			{
				currentNode->nextExpression=node;	// link to previous node if one existed
			}
		}
	}
	return(node);
}

static void RegexRangeAddToList(EXPRESSION_NODE *node,bool invert,uint32_t startPosition,uint64_t numElements)
// Add the bytes between rangeLow, and rangeHigh (inclusive) to the list in node.
{
	if(invert)
	{
		CodePointListRangeClear(node->re_list.codePointList,startPosition,numElements);
	}
	else
	{
		CodePointListRangeSet(node->re_list.codePointList,startPosition,numElements);
	}
}

static EXPRESSION_NODE *RegexListAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,TEXT_STREAM *stream)
// List operator has started, parse it, and create a list node in the list.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;
	uint32_t
		codePoint,
		lastCodePoint;
	bool
		invert,
		done,
		fail;
	uint64_t
		startPosition;

	if(!TextStreamAtEnd(stream))
	{
		if((node=RegexAllocateNode()))
		{
			node->type=PT_LIST;							// create a node, and apply the correct type to it
			node->parentExpression=parent;				// set the parent
			node->matchSpecified=false;					// no match count specified for this node yet
			node->matchMin=1;							// until further notice, this must match exactly once
			node->matchMax=1;
			node->nextExpression=NULL;					// this is now the last expression

			if((node->re_list.codePointList=CodePointListCreate()))	// create the list
			{
				done=fail=false;
				invert=false;
				startPosition=TextStreamPositionGet(stream);	// remember where we started
				lastCodePoint=0;							// start range at 0
				while(!done&&!fail&&!TextStreamAtEnd(stream))
				{
					codePoint=Utf8NextCodePoint(stream);
					switch(codePoint)
					{
						case '\\':
							if(RegexQuotedCharacterConvert(stream,&codePoint))
							{
								if(compiledExpression->ignoreCase)
								{
									codePoint=SearchToLower(codePoint);
								}
								RegexRangeAddToList(node,invert,codePoint,1);
							}
							else
							{
								fail=true;
							}
							break;
						case '^':
							if(TextStreamPositionGet(stream)-1==startPosition)					// at start??
							{
								invert=true;				// set not flag
								CodePointListRangeSet(node->re_list.codePointList,0,UTF8_NUM_CODEPOINTS);	// allow everything
								codePoint=0;				// keep last char pointed to the start
							}
							else
							{
								RegexRangeAddToList(node,invert,codePoint,1);
							}
							break;
						case '-':
							if(!TextStreamAtEnd(stream))	// get the next codePoint for the range (if there is one)
							{
								codePoint=Utf8NextCodePoint(stream);
								switch(codePoint)
								{
									case '\\':
										if(RegexQuotedCharacterConvert(stream,&codePoint))
										{
											if(compiledExpression->ignoreCase)
											{
												codePoint=SearchToLower(codePoint);
											}
											if(codePoint>=lastCodePoint)
											{
												RegexRangeAddToList(node,invert,lastCodePoint,(codePoint-lastCodePoint)+1);
											}
										}
										else
										{
											fail=true;
										}
										break;
									case ']':
										codePoint=UTF8_NUM_CODEPOINTS-1;
										RegexRangeAddToList(node,invert,lastCodePoint,(codePoint-lastCodePoint)+1);		// range pointing into space is assumed max value
										done=true;
										break;
									default:
										if(compiledExpression->ignoreCase)
										{
											codePoint=SearchToLower(codePoint);
										}
										if(codePoint>=lastCodePoint)
										{
											RegexRangeAddToList(node,invert,lastCodePoint,(codePoint-lastCodePoint)+1);
										}
										break;
								}
							}
							else
							{
								ErrorSet(errorDescriptions[UNBALANCED_SQUARE_BRACKET]);
								fail=true;
							}
							break;
						case ']':
							done=true;						// scanning is done (empty lists are allowed)
							break;
						default:							// otherwise it is normal, and just gets added in
							if(compiledExpression->ignoreCase)
							{
								codePoint=SearchToLower(codePoint);
							}
							RegexRangeAddToList(node,invert,codePoint,1);
							break;
					}
					lastCodePoint=codePoint;				// remember this for next time around the loop
				}
				if(!done&&!fail)							// ran out of things to scan
				{
					ErrorSet(errorDescriptions[UNBALANCED_SQUARE_BRACKET]);
					fail=true;
				}
				if(!fail)
				{
					if(currentNode)
					{
						currentNode->nextExpression=node;	// link to previous node if one existed
					}
					return(node);
				}
				CodePointListDispose(node->re_list.codePointList);
			}
			RegexFreeNode(node);
		}
	}
	else
	{
		ErrorSet(errorDescriptions[UNBALANCED_SQUARE_BRACKET]);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexAnyAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,TEXT_STREAM *stream)
// Create a list that matches any codePoint except \n.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_LIST;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression
		if((node->re_list.codePointList=CodePointListCreate()))	// create the list
		{
			CodePointListRangeSet(node->re_list.codePointList,0,UTF8_NUM_CODEPOINTS);	// set all entries in the array
			CodePointListRangeClear(node->re_list.codePointList,'\n',1);				// remove newline from the list
			if(currentNode)
			{
				currentNode->nextExpression=node;		// link to previous node if one existed
			}
			return(node);
		}
		RegexFreeNode(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexPreviousMatchAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,REGISTER *expRegister)
// Link in a new node that matches previously matched text.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_PREVIOUSMATCH;				// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression
		node->re_previous.expRegister=expRegister;			// set pointer to the register that needs to be matched
		if(currentNode)
		{
			currentNode->nextExpression=node;		// link to previous node if one existed
		}
		return(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexStartOfWordMatchAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// Link in a new node that matches the start of a word.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_EDGE;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		node->re_edge.mode=EM_LEFT_BOTH;			// match at word boundary
		node->re_edge.docEdgeMatches=true;			// document edge is considered a word boundary

		if((node->re_edge.codePointList=CodePointListCreate()))	// create the list
		{
			CodePointListsOr(node->re_edge.codePointList,wordSpaceList);	// copy the word space table

			if(currentNode)
			{
				currentNode->nextExpression=node;		// link to previous node if one existed
			}
			return(node);
		}
		RegexFreeNode(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexEndOfWordMatchAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// Link in a new node that matches the end of a word.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_EDGE;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		node->re_edge.mode=EM_RIGHT_BOTH;			// match at word boundary
		node->re_edge.docEdgeMatches=true;			// document edge is considered a word boundary

		if((node->re_edge.codePointList=CodePointListCreate()))	// create the list
		{
			CodePointListsOr(node->re_edge.codePointList,wordSpaceList);	// copy the word space table

			if(currentNode)
			{
				currentNode->nextExpression=node;		// link to previous node if one existed
			}
			return(node);
		}
		RegexFreeNode(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexWordBoundaryMatchAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// Link in a new node that matches the start or end of a word.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_EDGE;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		node->re_edge.mode=EM_TOGGLE;				// match at start or end
		node->re_edge.docEdgeMatches=true;			// document edge is considered a word boundary

		if((node->re_edge.codePointList=CodePointListCreate()))	// create the list
		{
			CodePointListsOr(node->re_edge.codePointList,wordSpaceList);	// copy the word space table

			if(currentNode)
			{
				currentNode->nextExpression=node;		// link to previous node if one existed
			}
			return(node);
		}
		RegexFreeNode(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexStartOfLineMatchAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// Link in a new node that matches the start of a line.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_EDGE;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		node->re_edge.mode=EM_LEFT;					// look for start of line to the left (don't care what's on the right)
		node->re_edge.docEdgeMatches=true;			// document edge is considered a line boundary

		if((node->re_edge.codePointList=CodePointListCreate()))	// create the list
		{
			CodePointListRangeSet(node->re_edge.codePointList,'\n',1);

			if(currentNode)
			{
				currentNode->nextExpression=node;		// link to previous node if one existed
			}
			return(node);
		}
		RegexFreeNode(node);
	}
	return(NULL);
}

static EXPRESSION_NODE *RegexEndOfLineMatchAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode)
// Link in a new node that matches the end of a line.
// If there is a problem, ErrorSet, and return NULL.
{
	EXPRESSION_NODE
		*node;

	if((node=RegexAllocateNode()))
	{
		node->type=PT_EDGE;							// create a node, and apply the correct type to it
		node->parentExpression=parent;				// set the parent
		node->matchSpecified=false;					// no match count specified for this node yet
		node->matchMin=1;							// until further notice, this must match exactly once
		node->matchMax=1;
		node->nextExpression=NULL;					// this is now the last expression

		node->re_edge.mode=EM_RIGHT;				// look for start of line to the right (don't care what's on the left)
		node->re_edge.docEdgeMatches=true;			// document edge is considered a line boundary

		if((node->re_edge.codePointList=CodePointListCreate()))	// create the list
		{
			CodePointListRangeSet(node->re_edge.codePointList,'\n',1);

			if(currentNode)
			{
				currentNode->nextExpression=node;		// link to previous node if one existed
			}
			return(node);
		}
		RegexFreeNode(node);
	}
	return(NULL);
}

static void RegexDigitsInhale(TEXT_STREAM *stream,uint64_t *result)
// Inhale as many digits as possible from stream.
// Return result.
// NOTE: it is ok for there to be no digits during the read, in this case
// result will return as 0, and stream will not be changed.
{
	TEXT_STREAM
		tempStream;
	bool
		done;
	uint32_t
		codePoint;

	done=false;
	*result=0;
	while(!done)
	{
		if(!TextStreamAtEnd(stream))
		{
			tempStream=*stream;								// keep this in case we need to put it back
			codePoint=Utf8NextCodePoint(stream);
			if(codePoint>='0'&&codePoint<='9')
			{
				*result=*result*10+(codePoint-'0');			// add this digit to the result
			}
			else
			{
				*stream=tempStream;							// unwind the last character read
				done=true;
			}
		}
		else
		{
			done=true;
		}
	}
}

static bool RegexRangeParse(TEXT_STREAM *stream,uint64_t *rangeLow,uint64_t *rangeHigh)
// Repetition range operator has started, read in the ranges, and return them in
// rangeLow, and rangeHigh.
// If there is a problem, ErrorSet, and return false.
{
	uint32_t
		codePoint;

	*rangeHigh=0;											// this means infinity
	RegexDigitsInhale(stream,rangeLow);						// attempt to read the low number in the range
	if(!TextStreamAtEnd(stream))							// see if any more codePoints (looking for , or })
	{
		codePoint=Utf8NextCodePoint(stream);
		switch(codePoint)
		{
			case ',':
				RegexDigitsInhale(stream,rangeHigh);		// attempt to read the high number in the range
				if(!TextStreamAtEnd(stream)&&Utf8NextCodePoint(stream)=='}')		// see if any more codePoints (looking for })
				{
					if((*rangeHigh==0)||(*rangeHigh>=*rangeLow))		// make sure low is <= high
					{
						return(true);
					}
				}
				ErrorSet(errorDescriptions[BAD_RANGE_EXPRESSION]);
				break;
			case '}':
				if((*rangeHigh=*rangeLow))					// range with just one number means exactly that many times (0 is invalid)
				{
					return(true);
				}
				else
				{
					ErrorSet(errorDescriptions[BAD_RANGE_EXPRESSION]);
				}
				break;
			default:
				ErrorSet(errorDescriptions[BAD_RANGE_EXPRESSION]);
				break;
		}
	}
	else
	{
		ErrorSet(errorDescriptions[BAD_RANGE_EXPRESSION]);
	}
	return(false);
}

// this is called recursively from below, so we must prototype it
static EXPRESSION_LIST_HEADER *RegexSubExpressionListParse(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,TEXT_STREAM *stream,uint32_t currentRegister);	// prototype this to keep compiler happy

static EXPRESSION_NODE *RegexSubExpressionAdd(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,EXPRESSION_NODE *currentNode,TEXT_STREAM *stream)
// Create a subexpression node, and build the subexpression into it.
// If there is a problem, ErrorSet, and return NULL.
{
	uint32_t
		currentRegister;
	EXPRESSION_NODE
		*startNode,
		*endNode;

	if((startNode=RegexAllocateNode()))
	{
		startNode->type=PT_SUBEXPRESSIONSTART;				// create a start node, and apply the correct type to it
		startNode->parentExpression=parent;
		startNode->matchSpecified=false;					// no match count specified for this node yet (none ever will be)
		startNode->matchMin=1;								// set to some default values (never looked at)
		startNode->matchMax=1;
		if((startNode->nextExpression=endNode=RegexAllocateNode()))
		{
			endNode->type=PT_SUBEXPRESSIONEND;				// create a node, and apply the correct type to it
			endNode->parentExpression=parent;
			endNode->matchSpecified=false;					// no match count specified for this node yet
			endNode->matchMin=1;							// until further notice, this must match exactly once
			endNode->matchMax=1;
			endNode->nextExpression=NULL;					// this is now the last expression

			currentRegister=compiledExpression->numRegisters;
			if(currentRegister<MAX_REGISTERS)				// see if there is a register
			{
				endNode->re_subexpression.expRegister=&(compiledExpression->expRegisters[currentRegister]);		// remember the register
				compiledExpression->numRegisters++;			// increment the number of registers in use
			}
			else
			{
				endNode->re_subexpression.expRegister=NULL;			// no register for this one
			}

			if((endNode->re_subexpression.subExpressionList=RegexSubExpressionListParse(compiledExpression,endNode,stream,currentRegister)))		// get subexpression
			{
				if(currentNode)
				{
					currentNode->nextExpression=startNode;	// link to previous node if one existed
				}
				return(startNode);
			}
			RegexFreeNode(endNode);
		}
		RegexFreeNode(startNode);
	}
	return(NULL);
}

static void RegexListsFree(EXPRESSION_LIST_HEADER *list)
// Given a pointer to the top of an expression list, free
// all subnodes, and all alternates below.
// Then free list.
// NOTE: this calls itself recursively for sublists.
{
	EXPRESSION_LIST_HEADER
		*currentList,
		*nextAlternate;
	EXPRESSION_NODE
		*currentNode,
		*nextNode;

	currentList=list;
	while(currentList)
	{
		nextAlternate=currentList->nextAlternate;
		currentNode=currentList->firstExpression;
		while(currentNode)
		{
			nextNode=currentNode->nextExpression;
			if(currentNode->type==PT_LIST)						// free list code point list
			{
				CodePointListDispose(currentNode->re_list.codePointList);
			}
			else if(currentNode->type==PT_EDGE)					// free edge code point list
			{
				CodePointListDispose(currentNode->re_edge.codePointList);
			}
			else if(currentNode->type==PT_SUBEXPRESSIONEND)		// free sub-list
			{
				RegexListsFree(currentNode->re_subexpression.subExpressionList);		// free all the nodes of this subexpression
			}
			RegexFreeNode(currentNode);
			currentNode=nextNode;								// move to the next horizontally
		}
		RegexFreeList(currentList);
		currentList=nextAlternate;
	}
}

static EXPRESSION_LIST_HEADER *RegexSubExpressionListParse(COMPILED_EXPRESSION *compiledExpression,EXPRESSION_NODE *parent,TEXT_STREAM *stream,uint32_t currentRegister)
// Parse stream, and build lists of nodes for compiledExpression.
// (If compiledExpression is NULL, then it is assumed that we are compiling a sublist.)
// Return the head of the created list of expression nodes.
// If there is a problem, ErrorSet, and return NULL.
{
	TEXT_STREAM
		tempStream;
	EXPRESSION_LIST_HEADER
		*startList,
		*currentList;						// current list in the list of alternates
	EXPRESSION_NODE
		*currentNode;
	uint32_t
		codePoint;
	uint32_t
		expRegisterIndex;
	bool
		done,
		fail;

	fail=done=false;
	if((startList=RegexAllocateList()))
	{
		startList->parentExpression=parent;
		startList->firstExpression=NULL;
		startList->nextAlternate=NULL;
		currentList=startList;								// point to the current list that is being built
		currentNode=NULL;
		while(!fail&&!done&&!TextStreamAtEnd(stream))
		{
			codePoint=Utf8NextCodePoint(stream);
			switch(codePoint)
			{
				case '(':
					fail=!((currentNode=RegexSubExpressionAdd(compiledExpression,parent,currentNode,stream)));
					break;
				case ')':
					if(parent)								// if not at the top, then take this as the end of our group
					{
						if(currentRegister<MAX_REGISTERS)	// define this register now
						{
							compiledExpression->registerDefined[currentRegister]=true;
						}
						done=true;
					}
					else
					{
						ErrorSet(errorDescriptions[UNBALANCED_PAREN]);
						fail=true;
					}
					break;
				case '{':
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_EDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						fail=!RegexRangeParse(stream,&currentNode->matchMin,&currentNode->matchMax);
					}
					else
					{
						ErrorSet(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '*':									// match 0 or more
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_EDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						currentNode->matchMin=0;			// no requirement to match any
						currentNode->matchMax=0;			// match as many as possible
					}
					else
					{
						ErrorSet(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '+':									// match 1 or more
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_EDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						currentNode->matchMin=1;			// match at least 1
						currentNode->matchMax=0;			// match as many as possible
					}
					else
					{
						ErrorSet(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '?':									// match 0 or 1
					if(currentNode&&!currentNode->matchSpecified&&currentNode->type!=PT_EDGE)
					{
						currentNode->matchSpecified=true;	// match has been specified
						currentNode->matchMin=0;			// no requirement to match any
						currentNode->matchMax=1;			// match at most 1
					}
					else
					{
						ErrorSet(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
					break;
				case '[':
					fail=!(currentNode=RegexListAdd(compiledExpression,parent,currentNode,stream));
					break;
				case '|':
					if((currentList->nextAlternate=RegexAllocateList()))	// create a new list, link it to the old, and make it current
					{
						currentList=currentList->nextAlternate;
						currentList->parentExpression=parent;
						currentList->nextAlternate=NULL;
						currentList->firstExpression=NULL;
						currentNode=NULL;					// no longer in the middle of a node
					}
					else
					{
						fail=true;
					}
					break;
				case '.':
					fail=!(currentNode=RegexAnyAdd(compiledExpression,parent,currentNode,stream));
					break;
				case '<':
					fail=!(currentNode=RegexStartOfWordMatchAdd(compiledExpression,parent,currentNode));
					break;
				case '>':
					fail=!(currentNode=RegexEndOfWordMatchAdd(compiledExpression,parent,currentNode));
					break;
				case '^':
					fail=!(currentNode=RegexStartOfLineMatchAdd(compiledExpression,parent,currentNode));
					break;
				case '$':
					fail=!(currentNode=RegexEndOfLineMatchAdd(compiledExpression,parent,currentNode));
					break;
				case '\\':
					if(!TextStreamAtEnd(stream))
					{
						tempStream=*stream;
						codePoint=Utf8NextCodePoint(stream);
						if(codePoint>='0'&&codePoint<='9')								// see if sub pattern match
						{
							expRegisterIndex=codePoint-'0';								// make ASCII digit into binary number
							if(expRegisterIndex<MAX_REGISTERS&&compiledExpression->registerDefined[expRegisterIndex])	// make sure it is in range, and that the given register has already been defined (this will guarantee that it is matched before we need it)
							{
								fail=!(currentNode=RegexPreviousMatchAdd(compiledExpression,parent,currentNode,&compiledExpression->expRegisters[expRegisterIndex]));
							}
							else
							{
								ErrorSet(errorDescriptions[INVALID_REGISTER]);
								fail=true;
							}
						}
						else if(codePoint=='b')				// handle word boundary spec
						{
							fail=!(currentNode=RegexWordBoundaryMatchAdd(compiledExpression,parent,currentNode));
						}
						else
						{
							*stream=tempStream;				// step back
							if(RegexQuotedCharacterConvert(stream,&codePoint))
							{
								fail=!(currentNode=RegexStringAdd(compiledExpression,parent,currentNode,codePoint));
							}
							else
							{
								fail=true;
							}
						}
					}
					else
					{
						ErrorSet(errorDescriptions[DANGLING_QUOTE]);
						fail=true;
					}
					break;
				default:
					fail=!(currentNode=RegexStringAdd(compiledExpression,parent,currentNode,codePoint));	// just take the codePoint as part of a string
					break;
			}
			if(currentNode)
			{
				if(!currentList->firstExpression)
				{
					currentList->firstExpression=currentNode;
				}
				if(currentNode->type==PT_SUBEXPRESSIONSTART)							// if just had start of expression, move to the end
				{
					currentNode=currentNode->nextExpression;
				}
			}
		}
		if(!fail&&!done&&parent)														// reached the end, but in middle of group somewhere
		{
			ErrorSet(errorDescriptions[UNBALANCED_PAREN]);
			fail=true;
		}
		if(!fail)
		{
			return(startList);
		}
		RegexListsFree(startList);
	}
	return(NULL);
}

static void MoveOver(EXPRESSION_LIST_HEADER *list,uint32_t depth)
// Move over by depth spaces.
{
	uint32_t
		i;

	for(i=0;i<depth;i++)
	{
		fprintf(stderr," ");
	}
}

static void CodePointListPrint(CODE_POINT_LIST *codePointList)
// Dump out a list of code points.
{
	uint32_t
		startPosition;
	uint64_t
		testPosition,
		numElements;

	testPosition=0;
	while((testPosition<UTF8_NUM_CODEPOINTS)&&CodePointListTestAtOrAfterPositionGet(codePointList,testPosition,&startPosition,&numElements))
	{
		fprintf(stderr,"{%d %d} ",startPosition,(uint32_t)numElements);
		testPosition=startPosition+numElements;
	}
	fprintf(stderr,"\n");
}

static void NodePrint(EXPRESSION_LIST_HEADER *list,EXPRESSION_NODE *node,uint32_t depth)
// Print formatted node information.
{
	uint32_t
		i;

	switch(node->type)
	{
		case PT_STRING:
			MoveOver(list,depth);
			fprintf(stderr,"STRING %" PRIu64 "/%" PRIu64 " '",node->matchMin,node->matchMax);
			for(i=0;i<node->dataLength;i++)
			{
				if(node->re_string.data[i]>=0x20&&node->re_string.data[i]<=0x7F)
				{
					fprintf(stderr,"%c",node->re_string.data[i]);
				}
				else
				{
					fprintf(stderr,"[%08X]",node->re_string.data[i]);
				}
			}
			fprintf(stderr,"'\n");
			break;
		case PT_LIST:
			MoveOver(list,depth);
			fprintf(stderr,"LIST %" PRIu64 "/%" PRIu64 " ",node->matchMin,node->matchMax);
			CodePointListPrint(node->re_list.codePointList);
			break;
		case PT_PREVIOUSMATCH:
			MoveOver(list,depth);
			fprintf(stderr,"PREVIOUSMATCH %" PRIu64 "/%" PRIu64 "\n",node->matchMin,node->matchMax);
			break;
		case PT_SUBEXPRESSIONSTART:
			MoveOver(list,depth);
			fprintf(stderr,"SUBEXPRESSIONSTART %" PRIu64 "/%" PRIu64 "\n",node->matchMin,node->matchMax);
			break;
		case PT_SUBEXPRESSIONEND:
			MoveOver(list,depth);
			fprintf(stderr,"SUBEXPRESSIONEND %" PRIu64 "/%" PRIu64 "\n",node->matchMin,node->matchMax);
			break;
		case PT_EDGE:
			MoveOver(list,depth);
			fprintf(stderr,"EDGE %" PRIu64 "/%" PRIu64 "\n",node->matchMin,node->matchMax);
			break;
		default:
			MoveOver(list,depth);
			fprintf(stderr,"????? (%u)\n",node->type);
			break;
	}
}

static void ExpressionListPrint(EXPRESSION_LIST_HEADER *list,uint32_t depth)
// Dump debugging output of the compiled expression.
// NOTE: this calls itself.
{
	EXPRESSION_LIST_HEADER
		*currentList;
	EXPRESSION_NODE
		*currentNode;

	currentList=list;
	while(currentList)
	{
		currentNode=currentList->firstExpression;
		while(currentNode)
		{
			if(currentNode->type==PT_SUBEXPRESSIONEND)
			{
				ExpressionListPrint(currentNode->re_subexpression.subExpressionList,depth+1);		// print the sublist
			}
			NodePrint(currentList,currentNode,depth);
			currentNode=currentNode->nextExpression;
		}
		currentList=currentList->nextAlternate;
	}
}

void RegexPrintCompiledExpression(COMPILED_EXPRESSION *expression)
// Dump debugging output of the compiled expression.
{
	ExpressionListPrint(expression->firstList,0);
	if(expression->infiniteMatchLength)
	{
		fprintf(stderr,"Expression max match is infinite\n");
	}
	else
	{
		fprintf(stderr,"Expression max match length=%" PRIu64 "\n",expression->maximumMatchLength);
	}
	fprintf(stderr,"Expression min match length=%" PRIu64 "\n",expression->minimumMatchLength);
	fprintf(stderr,"Expression matching start chars=");
	CodePointListPrint(expression->startCharsMatched);
	fprintf(stderr,"Expression matching chars=");
	CodePointListPrint(expression->charsMatched);
}

static void RegexExpressionListAnalyze(COMPILED_EXPRESSION *expression,EXPRESSION_LIST_HEADER *list,uint64_t *minLength,uint64_t *maxLength,bool *infiniteLength,bool haveStartChar)
// Analyze the given expression list, recursively calling this routine as needed.
{
	EXPRESSION_LIST_HEADER
		*currentList;
	EXPRESSION_NODE
		*currentNode;
	uint64_t
		localMinLength,
		localMaxLength,
		alternateMinLength,
		alternateMaxLength,
		subExpressionMinLength,
		subExpressionMaxLength;
	bool
		subExpressionInfiniteLength;
	uint32_t
		i;
	bool
		firstAlternate;

	currentList=list;
	firstAlternate=true;

	localMinLength=0;
	localMaxLength=0;

	while(currentList)
	{
		alternateMinLength=0;
		alternateMaxLength=0;
		currentNode=currentList->firstExpression;
		while(currentNode)
		{
			switch(currentNode->type)
			{
				case PT_STRING:
					if((!haveStartChar)&&(alternateMinLength==0))	// if at start codePoint of expression, add codePoint to list of possible start codePoints
					{
						CodePointListRangeSet(expression->startCharsMatched,currentNode->re_string.data[0],1);
					}
					alternateMinLength+=(currentNode->dataLength-1)+currentNode->matchMin;	// last codePoint must occur this many times
					if(currentNode->matchMax)
					{
						alternateMaxLength+=currentNode->dataLength+(currentNode->matchMax-1);	// last codePoint gets repeated
					}
					else
					{
						*infiniteLength=true;
					}
					for(i=0;i<currentNode->dataLength;i++)
					{
						CodePointListRangeSet(expression->charsMatched,currentNode->re_string.data[i],1);
					}
					break;
				case PT_LIST:
					if((!haveStartChar)&&(alternateMinLength==0))	// if at start codePoint of expression, add codePoint to list of possible start codePoints
					{
						CodePointListsOr(expression->startCharsMatched,currentNode->re_list.codePointList);	// add bits in
					}
					alternateMinLength+=currentNode->matchMin;
					if(currentNode->matchMax)
					{
						alternateMaxLength+=currentNode->matchMax;
					}
					else
					{
						*infiniteLength=true;
					}
					CodePointListsOr(expression->charsMatched,currentNode->re_list.codePointList);	// add bits in
					break;
				case PT_PREVIOUSMATCH:
					alternateMinLength+=currentNode->matchMin*currentNode->re_previous.expRegister->minimumMatchLength;
					if(currentNode->matchMax)
					{
						*infiniteLength|=currentNode->re_previous.expRegister->infiniteMatchLength;
						alternateMaxLength+=currentNode->matchMax*currentNode->re_previous.expRegister->maximumMatchLength;
					}
					else
					{
						*infiniteLength=true;
					}
					break;
				case PT_SUBEXPRESSIONSTART:
					break;
				case PT_SUBEXPRESSIONEND:
					subExpressionMinLength=0;
					subExpressionMaxLength=0;
					subExpressionInfiniteLength=false;
					RegexExpressionListAnalyze(expression,currentNode->re_subexpression.subExpressionList,&subExpressionMinLength,&subExpressionMaxLength,&subExpressionInfiniteLength,(haveStartChar||alternateMinLength>0));	// analyze the sublist
					if(currentNode->re_subexpression.expRegister)
					{
						currentNode->re_subexpression.expRegister->minimumMatchLength=subExpressionMinLength;	// remember the min length for this sub-expression
						currentNode->re_subexpression.expRegister->maximumMatchLength=subExpressionMaxLength;	// remember the max length for this sub-expression
						currentNode->re_subexpression.expRegister->infiniteMatchLength=subExpressionInfiniteLength;	// remember if this could match infinitely
					}
					alternateMinLength+=currentNode->matchMin*subExpressionMinLength;
					if(currentNode->matchMax&&!subExpressionInfiniteLength)
					{
						alternateMaxLength+=currentNode->matchMax*subExpressionMaxLength;
					}
					else
					{
						*infiniteLength=true;
					}
					break;
				case PT_EDGE:
					break;
				default:
					break;
			}
			currentNode=currentNode->nextExpression;
		}

		if(firstAlternate||alternateMinLength<localMinLength)
		{
			localMinLength=alternateMinLength;
		}
		if(alternateMaxLength>localMaxLength)
		{
			localMaxLength=alternateMaxLength;
		}
		firstAlternate=false;

		currentList=currentList->nextAlternate;
	}

	*minLength+=localMinLength;
	*maxLength+=localMaxLength;
}

static void RegexExpressionAnalyze(COMPILED_EXPRESSION *expression)
// Work out which codePoints can possibly occur in expression.
// Also, work out the maximum length of a match of expression.
// NOTE: if the maximum match length is infinite, set maximumMatchLength to 0.
{
	CodePointListRangeClear(expression->charsMatched,0,UTF8_NUM_CODEPOINTS);		// clear the bit table of possible matches
	CodePointListRangeClear(expression->startCharsMatched,0,UTF8_NUM_CODEPOINTS);	// clear the bit table of possible matches at the start

	expression->minimumMatchLength=0;
	expression->maximumMatchLength=0;
	expression->infiniteMatchLength=false;
	RegexExpressionListAnalyze(expression,expression->firstList,&expression->minimumMatchLength,&expression->maximumMatchLength,&expression->infiniteMatchLength,false);	// recursively analyze the lists
}

void RegexFree(COMPILED_EXPRESSION *expression)
// Free all nodes used by expression.
// This must be called for every call to RECompile to deallocate the
// memory used by expression.
{
	RegexListsFree(expression->firstList);			// free the lists
	CodePointListDispose(expression->startCharsMatched);
	CodePointListDispose(expression->charsMatched);
	MDisposePtr(expression);						// remove the expression header
}

COMPILED_EXPRESSION *RegexCompileStream(TEXT_STREAM *stream,bool ignoreCase)
// Compile the expression in stream, and return a pointer to the
// node of the start of the expression.
// If there is a problem, call ErrorSet, and return NULL.
{
	COMPILED_EXPRESSION
		*compiledExpression;
	uint32_t
		i;

	if((compiledExpression=(COMPILED_EXPRESSION *)MNewPtr(sizeof(COMPILED_EXPRESSION))))
	{
		if((compiledExpression->charsMatched=CodePointListCreate()))
		{
			if((compiledExpression->startCharsMatched=CodePointListCreate()))
			{
				compiledExpression->ignoreCase=ignoreCase;					// remember if this expression was compiled to ignore case

				compiledExpression->numRegisters=0;							// no registers encountered yet
				for(i=0;i<MAX_REGISTERS;i++)
				{
					compiledExpression->registerDefined[i]=false;			// no registers have been defined yet
				}
				if((compiledExpression->firstList=RegexSubExpressionListParse(compiledExpression,NULL,stream,0)))
				{
					RegexExpressionAnalyze(compiledExpression);				// look at the expression, and work out various pieces of information which can be used to determine how far back one needs to go from a text change to locate an expression which includes the change
//					RegexPrintCompiledExpression(compiledExpression);		// dump out the expression
					return(compiledExpression);
				}
				CodePointListDispose(compiledExpression->startCharsMatched);
			}
			CodePointListDispose(compiledExpression->charsMatched);
		}
		MDisposePtr(compiledExpression);
	}
	return(NULL);
}

COMPILED_EXPRESSION *RegexCompileText(uint8_t *expression,uint64_t numBytes,bool ignoreCase)
// Compile numBytes of expression, and return a pointer to the
// node of the start of the expression.
// If there is a problem, call ErrorSet, and return NULL.
{
	COMPILED_EXPRESSION
		*compiledExpression;
	TEXT_UNIVERSE
		*tempUniverse;
	TEXT_STREAM
		stream;

	compiledExpression=NULL;
	if((tempUniverse=TextUniverseOpen()))	// create a text universe
	{
		if(TextUniverseInsertFromText(expression,tempUniverse,0,numBytes))	// copy in the expression text
		{
			TextStreamPositionSet(&stream,tempUniverse,0);					// point to the start
			compiledExpression=RegexCompileStream(&stream,ignoreCase);		// go compile the expression
		}
		TextUniverseClose(tempUniverse);	// get rid of the temp universe
	}
	return(compiledExpression);
}

// ----------------------------------------------------------------------------
// MATCHING FUNCTIONS

static EXPRESSION_NODE *RegexNextNodeGet(EXPRESSION_NODE *node)
// Get the next expression node at this level, or move up a level if possible.
{
	EXPRESSION_NODE
		*newNode;

	if(!(newNode=node->nextExpression))
	{
		newNode=node->parentExpression;
	}
	return(newNode);
}

// these are called recursively from below, so they must be prototyped
static bool RegexAlternativesMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_LIST_HEADER *list,RECURSE_GLOBALS *recurseGlobals);
static bool RegexStringNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals);
static bool RegexListNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals);
static bool RegexPreviousNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals);
static bool RegexSubStartNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals);
static bool RegexSubEndNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals);
static bool RegexEdgeNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals);

// array of pointers to matching routines
static bool
	(*matchRoutines[])(const TEXT_STREAM *,uint64_t,EXPRESSION_NODE *,RECURSE_GLOBALS *)=
	{
		RegexStringNodeMatch,
		RegexListNodeMatch,
		RegexPreviousNodeMatch,
		RegexSubStartNodeMatch,
		RegexSubEndNodeMatch,
		RegexEdgeNodeMatch
	};

// macro function that branches to next match routine from current one
#define	RegexCallNextNode(stream,numToSearch,node,recurseGlobals)	matchRoutines[node->type](stream,numToSearch,node,recurseGlobals);

static bool RegexMatchCodePoints(bool ignoreCase,uint32_t codePointA,uint32_t codePointB)
// See if codePointA matches codePointB. This will handle ignoring case as needed.
{
	if(ignoreCase)
	{
		codePointA=SearchToLower(codePointA);
		codePointB=SearchToLower(codePointB);
	}
	return(codePointA==codePointB);
}

static bool RegexMatchCodePointWithList(bool ignoreCase,uint32_t codePoint,CODE_POINT_LIST *codePointList)
// See if codePoint is in codePointList. This will handle ignoring case as needed.
{
	if(ignoreCase)
	{
		codePoint=SearchToLower(codePoint);
	}
	return(CodePointListTest(codePointList,codePoint));
}

static bool RegexStringNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match node with the data at stream.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	TEXT_STREAM
		localStream;
	uint64_t
		i,
		endPosition;
	bool
		fail;
	uint32_t
		codePoint,
		lastCodePoint;
	EXPRESSION_NODE
		*nextExpression;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		localStream=*stream;
		endPosition=TextStreamPositionGet(&localStream)+numToSearch;

		fail=false;
		// match the constant part of the string (if one exists)
		recurseGlobals->foundMatch=true;
		i=0;
		while(recurseGlobals->foundMatch&&(i<(uint8_t)(node->dataLength-1)))
		{
			if(TextStreamPositionGet(&localStream)<endPosition)
			{
				if(recurseGlobals->useUtf8)
				{
					codePoint=Utf8NextCodePoint(&localStream);
				}
				else
				{
					codePoint=TextStreamChar(&localStream);
					TextStreamNextChar(&localStream);				// move forward in the input
				}
				recurseGlobals->foundMatch=RegexMatchCodePoints(recurseGlobals->ignoreCase,codePoint,node->re_string.data[i]);	// check for match
				i++;
			}
			else
			{
				recurseGlobals->foundMatch=false;
			}
		}

		if(recurseGlobals->foundMatch)								// make sure we made it through the first part
		{
			lastCodePoint=node->re_string.data[node->dataLength-1];
			i=0;													// reset match count
			while(recurseGlobals->foundMatch&&(TextStreamPositionGet(&localStream)<endPosition)&&((!node->matchMax)||(i<node->matchMax)))	// read as many as possible until we hit max, or fail
			{
				if(recurseGlobals->useUtf8)
				{
					codePoint=Utf8NextCodePoint(&localStream);
				}
				else
				{
					codePoint=TextStreamChar(&localStream);
					TextStreamNextChar(&localStream);				// move forward in the input
				}
				if((recurseGlobals->foundMatch=RegexMatchCodePoints(recurseGlobals->ignoreCase,codePoint,lastCodePoint)))	// check for match
				{
					i++;											// bump the count
				}
				else	// failed to match -- step back one
				{
					TextStreamPreviousChar(&localStream);			// move back in the input
					if(recurseGlobals->useUtf8)
					{
						Utf8Synchronize(&localStream);
					}
				}
			}
			if(i>=node->matchMin)									// see if we got the minimum number of matches
			{
				if((nextExpression=RegexNextNodeGet(node)))			// see if there is another node after this one (if so, recurse into it)
				{
					fail=!RegexCallNextNode(&localStream,endPosition-TextStreamPositionGet(&localStream),nextExpression,recurseGlobals);
					while(!fail&&!recurseGlobals->foundMatch&&i>node->matchMin)		// see if possible to backup and try again
					{
						i--;
						TextStreamPreviousChar(&localStream);		// back up 1 codePoint
						if(recurseGlobals->useUtf8)
						{
							Utf8Synchronize(&localStream);
						}
						fail=!RegexCallNextNode(&localStream,endPosition-TextStreamPositionGet(&localStream),nextExpression,recurseGlobals);
					}
				}
				else												// no other nodes, this was at the end, so it is done
				{
					recurseGlobals->endStream=localStream;			// copy the stream
					recurseGlobals->foundMatch=true;				// if no more nodes, then this list is complete, and it matched
				}
			}
			else
			{
				recurseGlobals->foundMatch=false;					// did not find minimum number of matches, so this fails here and now
			}
		}
		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool RegexListNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match node with the data at stream.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	TEXT_STREAM
		localStream;
	uint64_t
		i,
		endPosition;
	bool
		fail;
	uint32_t
		codePoint;
	EXPRESSION_NODE
		*nextExpression;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		localStream=*stream;
		endPosition=TextStreamPositionGet(&localStream)+numToSearch;

		fail=false;
		recurseGlobals->foundMatch=true;							// start by assuming a match
		i=0;
		while(recurseGlobals->foundMatch&&(TextStreamPositionGet(&localStream)<endPosition)&&((!node->matchMax)||(i<node->matchMax)))	// read as many as possible until we hit max, or fail
		{
			if(recurseGlobals->useUtf8)
			{
				codePoint=Utf8NextCodePoint(&localStream);
			}
			else
			{
				codePoint=TextStreamChar(&localStream);
				TextStreamNextChar(&localStream);					// move forward in the input
			}
			if((recurseGlobals->foundMatch=RegexMatchCodePointWithList(recurseGlobals->ignoreCase,codePoint,node->re_list.codePointList)))	// check for match
			{
				i++;
			}
			else
			{
				TextStreamPreviousChar(&localStream);				// back up 1 codePoint
				if(recurseGlobals->useUtf8)
				{
					Utf8Synchronize(&localStream);
				}
			}
		}
		if(i>=node->matchMin)										// see if we got the minimum number of matches
		{
			if((nextExpression=RegexNextNodeGet(node)))				// see if there is another node after this one (if so, recurse into it)
			{
				fail=!RegexCallNextNode(&localStream,endPosition-TextStreamPositionGet(&localStream),nextExpression,recurseGlobals);
				while(!fail&&!recurseGlobals->foundMatch&&i>node->matchMin)			// see if possible to backup and try again
				{
					i--;
					TextStreamPreviousChar(&localStream);			// back up 1 codePoint
					if(recurseGlobals->useUtf8)
					{
						Utf8Synchronize(&localStream);
					}
					fail=!RegexCallNextNode(&localStream,endPosition-TextStreamPositionGet(&localStream),nextExpression,recurseGlobals);
				}
			}
			else													// no other nodes, this was at the end, so it is done
			{
				recurseGlobals->endStream=localStream;				// copy the stream
				recurseGlobals->foundMatch=true;					// if no more nodes, then this list is complete, and it matched
			}
		}
		else
		{
			recurseGlobals->foundMatch=false;						// did not find minimum number of matches, so this fails here and now
		}
		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool RegexPreviousNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match node with the data at stream.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	TEXT_STREAM
		localStream;
	uint64_t
		i,
		nodeMatched,
		endPosition;
	bool
		fail;
	EXPRESSION_NODE
		*nextExpression;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		localStream=*stream;
		endPosition=TextStreamPositionGet(&localStream)+numToSearch;

		fail=false;
		recurseGlobals->foundMatch=true;										// start by assuming a match
		i=0;
		while(!fail&&recurseGlobals->foundMatch&&((!node->matchMax)||(i<node->matchMax)))	// match as many as possible until we hit max, or fail
		{
			if(!TextStreamAtEnd(&node->re_previous.expRegister->startStream)&&(endPosition-TextStreamPositionGet(&localStream))>=node->re_previous.expRegister->numBytes)	// see if it is defined, and enough bytes to test, if not, matching fails
			{
				if((recurseGlobals->foundMatch=LiteralMatch(&localStream,&node->re_previous.expRegister->startStream,node->re_previous.expRegister->numBytes,false,recurseGlobals->ignoreCase,&recurseGlobals->endStream)))
				{
					i++;
					localStream=recurseGlobals->endStream;	// move to new position

					if((node->re_previous.expRegister->numBytes==0)&&!node->matchMax)	// if allowed to match infinite times, and matched 0 bytes, that's an error (we'd spin forever)
					{
						ErrorSet(errorDescriptions[BAD_REPETITION]);
						fail=true;
					}
				}
			}
			else
			{
				recurseGlobals->foundMatch=false;
			}
		}
		if(i>=node->matchMin)								// see if we got the minimum number of matches
		{
			if((nextExpression=RegexNextNodeGet(node)))		// see if there is another node after this one (if so, recurse into it)
			{
				fail=!RegexCallNextNode(&localStream,endPosition-TextStreamPositionGet(&localStream),nextExpression,recurseGlobals);
				while(!fail&&!recurseGlobals->foundMatch&&i>node->matchMin)	// see if possible to backup and try again
				{
					i--;
					nodeMatched=node->re_previous.expRegister->numBytes;
					TextStreamMoveRelative(&localStream,-nodeMatched);	// back up nodeMatched codePoints
					fail=!RegexCallNextNode(&localStream,endPosition-TextStreamPositionGet(&localStream),nextExpression,recurseGlobals);
				}
			}
			else											// no other nodes, this was at the end, so it is done
			{
				recurseGlobals->endStream=localStream;		// copy the stream
				recurseGlobals->foundMatch=true;			// if no more nodes, then this list is complete, and it matched
			}
		}
		else
		{
			recurseGlobals->foundMatch=false;				// did not find minimum number of matches, so this fails here and now
		}
		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool RegexSubStartNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match node with the data at stream.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	uint64_t
		oldCount;
	bool
		fail;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		node->nextExpression->re_subexpression.justEntered=true;		// set up the end of the subexpression so that it knows we just hit the start
		oldCount=node->nextExpression->re_subexpression.currentCount;	// keep track of how many times this expression matched last time
		fail=!RegexCallNextNode(stream,numToSearch,node->nextExpression,recurseGlobals);
		node->nextExpression->re_subexpression.currentCount=oldCount;	// restore match count
		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool RegexSubEndNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match node with the data at stream.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	bool
		fail;
	EXPRESSION_NODE
		*nextExpression;
	REGISTER
		oldRegister;
	TEXT_STREAM
		oldLastStream;									// used to keep the contents of the old register in case we back up
	uint64_t
		oldLastNumToSearch;

	oldLastNumToSearch=0;

	oldRegister.numBytes=0;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=recurseGlobals->foundMatch=false;
		if(node->re_subexpression.justEntered)
		{
			node->re_subexpression.currentCount=0;					// start off the counter
			node->re_subexpression.lastStream=*stream;				// remember where we were, so the register can be set if needed
			node->re_subexpression.lastNumToSearch=numToSearch;
			node->re_subexpression.justEntered=false;				// next time this is seen, we will know it came from a successful completion!
			if(RegexAlternativesMatch(stream,numToSearch,node->re_subexpression.subExpressionList,recurseGlobals))
			{
				if(!recurseGlobals->foundMatch)			// if could not match, then see if we can move forward from here
				{
					if(!node->matchMin)					// if no matches needed, then we have success
					{
						if((nextExpression=RegexNextNodeGet(node)))		// see if there is another node after this one (if so, recurse into it)
						{
							fail=!RegexCallNextNode(stream,numToSearch,nextExpression,recurseGlobals);
						}
						else
						{
							recurseGlobals->foundMatch=true;
							recurseGlobals->endStream=*stream;
						}
					}
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			node->re_subexpression.currentCount++;					// made another successful match
			if(node->re_subexpression.expRegister)					// see if this is one that is actually remembered
			{
				oldRegister.startStream=node->re_subexpression.expRegister->startStream;		// copy old register contents (in case we need to back out)
				oldRegister.numBytes=node->re_subexpression.expRegister->numBytes;

				oldLastStream=node->re_subexpression.lastStream;
				oldLastNumToSearch=node->re_subexpression.lastNumToSearch;

				node->re_subexpression.expRegister->startStream=node->re_subexpression.lastStream;
				node->re_subexpression.expRegister->numBytes=node->re_subexpression.lastNumToSearch-numToSearch;

				node->re_subexpression.lastStream=*stream;
				node->re_subexpression.lastNumToSearch=numToSearch;
			}
			if((!node->matchMax)||node->re_subexpression.currentCount<node->matchMax)
			{
				if(RegexAlternativesMatch(stream,numToSearch,node->re_subexpression.subExpressionList,recurseGlobals))
				{
					if(!recurseGlobals->foundMatch)							// if could not match, then see if we can move forward from here
					{
						if(node->re_subexpression.currentCount>=node->matchMin)
						{
							if((nextExpression=RegexNextNodeGet(node)))		// see if there is another node after this one (if so, recurse into it)
							{
								fail=!RegexCallNextNode(stream,numToSearch,nextExpression,recurseGlobals);
							}
							else
							{
								recurseGlobals->endStream=*stream;
								recurseGlobals->foundMatch=true;
							}
						}
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				if((nextExpression=RegexNextNodeGet(node)))	// see if there is another node after this one (if so, recurse into it)
				{
					fail=!RegexCallNextNode(stream,numToSearch,nextExpression,recurseGlobals);
				}
				else
				{
					recurseGlobals->endStream=*stream;
					recurseGlobals->foundMatch=true;
				}
			}
			if(!recurseGlobals->foundMatch)
			{
				node->re_subexpression.currentCount--;				// this match failed, so back out

				if(node->re_subexpression.expRegister)				// see if this is one that is actually remembered
				{
					node->re_subexpression.lastStream=oldLastStream;
					node->re_subexpression.lastNumToSearch=oldLastNumToSearch;

					node->re_subexpression.expRegister->startStream=oldRegister.startStream;
					node->re_subexpression.expRegister->numBytes=oldRegister.numBytes;
				}
			}
		}
		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}


static bool RegexLeftEdgeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Check to see if the character to the left of the stream matches the
// node.
{
	TEXT_STREAM
		localStream;										// used to keep from changing stream
	uint32_t
		codePoint;
	bool
		result;

	if((numToSearch==recurseGlobals->initialNumToSearch)&&recurseGlobals->leftEdge)	// if at the start of the search, and we've been told it's a left edge...
	{
		result=node->re_edge.docEdgeMatches;			// return a match if document edge can match
	}
	else if(TextStreamPositionGet(stream))				// possible to reach back and look at the character?
	{
		localStream=*stream;
		TextStreamPreviousChar(&localStream);			// back up from current position, and look for start codePoint
		if(recurseGlobals->useUtf8)
		{
			Utf8Synchronize(&localStream);				// move back to start of UTF8 code point
			codePoint=Utf8NextCodePoint(&localStream);
		}
		else
		{
			codePoint=TextStreamChar(&localStream);
		}
		result=RegexMatchCodePointWithList(false,codePoint,node->re_edge.codePointList);	// see if left edge codePoint
	}
	else
	{
		result=node->re_edge.docEdgeMatches;			// return a match if document edge can match
	}
	return(result);
}

static bool RegexRightEdgeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Check to see if the character to the right of the stream matches the
// node.
{
	TEXT_STREAM
		localStream;									// used to keep from changing stream
	uint32_t
		codePoint;
	bool
		result;

	if(!numToSearch&&recurseGlobals->rightEdge)			// if at the end of the search, and we've been told it's a right edge...
	{
		result=node->re_edge.docEdgeMatches;			// return a match if document edge can match
	}
	else if(!TextStreamAtEnd(stream))					// possible to reach forward and look at the character?
	{
		localStream=*stream;
		if(recurseGlobals->useUtf8)						// fetch the character, and see if it is in the list
		{
			codePoint=Utf8NextCodePoint(&localStream);
		}
		else
		{
			codePoint=TextStreamChar(&localStream);
		}
		result=RegexMatchCodePointWithList(false,codePoint,node->re_edge.codePointList);	// see if right edge codePoint?
	}
	else
	{
		result=node->re_edge.docEdgeMatches;			// return a match if document edge can match
	}
	return(result);
}

static bool RegexEdgeNodeMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_NODE *node,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match node with the data at stream.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	bool
		fail;
	bool
		matchedLeft,
		matchedRight;
	EXPRESSION_NODE
		*nextExpression;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		fail=false;
		recurseGlobals->foundMatch=false;

		switch(node->re_edge.mode)
		{
			case EM_LEFT:
				recurseGlobals->foundMatch=RegexLeftEdgeMatch(stream,numToSearch,node,recurseGlobals);
				break;
			case EM_RIGHT:
				recurseGlobals->foundMatch=RegexRightEdgeMatch(stream,numToSearch,node,recurseGlobals);
				break;
			case EM_LEFT_BOTH:
				recurseGlobals->foundMatch=RegexLeftEdgeMatch(stream,numToSearch,node,recurseGlobals)&&!RegexRightEdgeMatch(stream,numToSearch,node,recurseGlobals);
				break;
			case EM_RIGHT_BOTH:
				recurseGlobals->foundMatch=RegexRightEdgeMatch(stream,numToSearch,node,recurseGlobals)&&!RegexLeftEdgeMatch(stream,numToSearch,node,recurseGlobals);
				break;
			case EM_TOGGLE:
				matchedLeft=RegexLeftEdgeMatch(stream,numToSearch,node,recurseGlobals)&&!RegexRightEdgeMatch(stream,numToSearch,node,recurseGlobals);
				matchedRight=RegexRightEdgeMatch(stream,numToSearch,node,recurseGlobals)&&!RegexLeftEdgeMatch(stream,numToSearch,node,recurseGlobals);
				recurseGlobals->foundMatch=(matchedLeft&&!matchedRight)||(matchedRight&&!matchedLeft);
				break;
		}
		if(recurseGlobals->foundMatch)
		{
			recurseGlobals->endStream=*stream;				// return position of match
			if((nextExpression=RegexNextNodeGet(node)))		// see if there is another node after this one (if so, recurse into it)
			{
				fail=!RegexCallNextNode(stream,numToSearch,nextExpression,recurseGlobals);
			}
		}

		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool RegexAlternativesMatch(const TEXT_STREAM *stream,uint64_t numToSearch,EXPRESSION_LIST_HEADER *list,RECURSE_GLOBALS *recurseGlobals)
// Attempt to match alternatives at the given level.
// If the pattern could be matched, return true in recurseGlobals->foundMatch.
// NOTE: do not search through more than numToSearch bytes.
// recurseGlobals->endStream will point just past the end of the match if one was located.
// This tries each alternative in order until a match is located, or no match could be found.
// If there is some sort of hard failure, this routine will ErrorSet, and return false.
{
	EXPRESSION_LIST_HEADER
		*currentList;
	bool
		fail;

	if(++recurseGlobals->matchRecursionDepth<MAX_RECURSION_DEPTH)
	{
		if(!recurseGlobals->abortFunction||!recurseGlobals->abortFunction())
		{
			currentList=list;
			recurseGlobals->foundMatch=false;
			fail=false;
			while(currentList&&!recurseGlobals->foundMatch&&!fail)
			{
				if(currentList->firstExpression)					// if non-empty list of expressions, go handle it
				{
					fail=!RegexCallNextNode(stream,numToSearch,currentList->firstExpression,recurseGlobals);
				}
				else	// list is empty, so it matches, and we continue with the parent's next node
				{
					if(currentList->parentExpression)				// go back to the parent
					{
						fail=!RegexCallNextNode(stream,numToSearch,currentList->parentExpression,recurseGlobals);
					}
					else											// no other nodes, this was at the end, so it is done
					{
						recurseGlobals->endStream=*stream;
						recurseGlobals->foundMatch=true;			// if list is empty, then it matches
					}
				}
				currentList=currentList->nextAlternate;
			}
		}
		else
		{
			ErrorSet(errorDescriptions[SEARCH_ABORTED]);
			fail=true;												// abort happened
		}
		recurseGlobals->matchRecursionDepth--;
	}
	else
	{
		ErrorSet(errorDescriptions[STACK_OVERFLOW]);
		fail=true;
	}
	return(!fail);
}

static bool RegexMatch(const TEXT_STREAM *stream,COMPILED_EXPRESSION *compiledExpression,uint64_t numToSearch,RECURSE_GLOBALS *recurseGlobals)
// This is the regular expression match routine.
// It will return true in recurseGlobals->foundMatch if a match is found.
// If there is a hard failure, ErrorSet will be called, and false returned.
// NOTE: this can be called when stream points to nothing valid (for instance in the case
// when there is no text to search). It is even possible for this to find
// a match in that case (as in the case when the expression is '$').
// Of course, if stream doesn't point to anything, numToSearch must be 0.
{
	uint32_t
		i;

	for(i=0;i<compiledExpression->numRegisters;i++)
	{
		TextStreamPositionClear(&compiledExpression->expRegisters[i].startStream);	// no registers have been matched yet
	}

	return(RegexAlternativesMatch(stream,numToSearch,compiledExpression->firstList,recurseGlobals));
}

bool RegexRegisterMatchGet(COMPILED_EXPRESSION *compiledExpression,uint32_t expRegisterIndex,TEXT_STREAM *startStream,uint64_t *matchLength)
// After a call to a regular expression search which succeeds, this
// may be called to locate information about the tagged matches.
// If the passed register matched, set startStream and matchLength to point to the start
// of the match and the length, and return true.
// Otherwise, just return false.
{
	if(expRegisterIndex<compiledExpression->numRegisters)
	{
		if(!TextStreamAtEnd(&compiledExpression->expRegisters[expRegisterIndex].startStream))
		{
			*startStream=compiledExpression->expRegisters[expRegisterIndex].startStream;
			*matchLength=compiledExpression->expRegisters[expRegisterIndex].numBytes;
			return(true);
		}
	}
	return(false);
}

bool RegexSearchForward(const TEXT_STREAM *stream,COMPILED_EXPRESSION *compiledExpression,bool useUtf8,bool leftEdge,bool rightEdge,uint64_t numToSearchStart,uint64_t numToSearch,bool *foundMatch,TEXT_STREAM *startStream,TEXT_STREAM *endStream,ABORT_TEST_FUNCTION *abortFunction)
// Search for compiledExpression starting at stream.
// Return true in foundMatch if it is found.
// leftEdge tells if the starting point is considered a left edge for matching purposes.
// rightEdge tells if the ending point is considered a right edge for matching purposes.
// numToSearchStart gives the maximum number of bytes to look through trying to find the start of a match.
// numToSearch gives the maximum number of bytes to look through.
// startStream gives the address in the text which matched.
// endStream gives the address of the end of the match+1.
// If abortFunction is non NULL, it will be called during the search, and an error
// will be returned if it ever returns true.
// If there is a hard error, ErrorSet will be called, and false returned.
{
	RECURSE_GLOBALS
		recurseGlobals;
	TEXT_STREAM
		localStream,
		tempStream;
	bool
		fail,
		done;
	uint64_t
		endPositionStart,
		endPosition;
	uint32_t
		codePoint;

	// to save time/stack while calling recursively, this structure is used to hold variables that are global to the recursive search routines, and only a pointer to this is passed
	recurseGlobals.useUtf8=useUtf8;
	recurseGlobals.ignoreCase=compiledExpression->ignoreCase;
	recurseGlobals.leftEdge=leftEdge;
	recurseGlobals.rightEdge=rightEdge;
	recurseGlobals.foundMatch=false;
	TextStreamPositionClear(&recurseGlobals.endStream);
	recurseGlobals.matchRecursionDepth=0;						// clear recursion depth
	recurseGlobals.abortFunction=abortFunction;

	localStream=*stream;
	endPositionStart=TextStreamPositionGet(&localStream)+numToSearchStart;
	endPosition=TextStreamPositionGet(&localStream)+numToSearch;

	fail=false;

	if(compiledExpression->minimumMatchLength)					// this case can be sped up
	{
		while(!fail&&!recurseGlobals.foundMatch&&TextStreamPositionGet(&localStream)<endPositionStart)
		{
			if(useUtf8)
			{
				tempStream=localStream;
				codePoint=Utf8NextCodePoint(&tempStream);
			}
			else
			{
				codePoint=TextStreamChar(&localStream);
			}

			if(RegexMatchCodePointWithList(compiledExpression->ignoreCase,codePoint,compiledExpression->startCharsMatched))	// do some quick tests to bail out now if possible
			{
				recurseGlobals.initialNumToSearch=endPosition-TextStreamPositionGet(&localStream);
				if(RegexMatch(&localStream,compiledExpression,recurseGlobals.initialNumToSearch,&recurseGlobals))
				{
					if(!recurseGlobals.foundMatch)
					{
						// move forward in the input, and try a match there
						if(useUtf8)
						{
							Utf8NextCodePoint(&localStream);
						}
						else
						{
							TextStreamNextChar(&localStream);
						}
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				// move forward in the input, and try a match there
				if(useUtf8)
				{
					Utf8NextCodePoint(&localStream);
				}
				else
				{
					TextStreamNextChar(&localStream);
				}
			}
			recurseGlobals.leftEdge=false;
		}
	}
	else
	{
		done=false;
		do
		{
			recurseGlobals.initialNumToSearch=endPosition-TextStreamPositionGet(&localStream);
			if(RegexMatch(&localStream,compiledExpression,recurseGlobals.initialNumToSearch,&recurseGlobals))
			{
				recurseGlobals.leftEdge=false;
				if(!recurseGlobals.foundMatch)
				{
					if(TextStreamPositionGet(&localStream)<endPositionStart)	// still possible to move forward?
					{
						// move forward in the input, and try a match there
						if(useUtf8)
						{
							Utf8NextCodePoint(&localStream);
						}
						else
						{
							TextStreamNextChar(&localStream);
						}
					}
					else	// no match found, and at the end, so done
					{
						done=true;
					}
				}
			}
			else
			{
				fail=true;
			}
		}
		while(!fail&&!done&&!recurseGlobals.foundMatch);
	}
	*startStream=localStream;
	*endStream=recurseGlobals.endStream;
	*foundMatch=recurseGlobals.foundMatch;

	return(!fail);
}

bool RegexSearchBackward(const TEXT_STREAM *stream,COMPILED_EXPRESSION *compiledExpression,bool useUtf8,bool leftEdge,bool rightEdge,uint64_t numToSearch,bool *foundMatch,TEXT_STREAM *startStream,TEXT_STREAM *endStream,ABORT_TEST_FUNCTION *abortFunction)
// Search backward for the expression return true in foundMatch if it is found.
// This repeatedly calls RegexMatch.
// If there is a hard error, ErrorSet will be called, and false returned.
{
	RECURSE_GLOBALS
		recurseGlobals;
	TEXT_STREAM
		localStream;
	bool
		fail;
	uint64_t
		endPosition;

	// to save time/stack while calling recursively, this structure is used to hold variables that are global to the recursive search routines, and only a pointer to this is passed
	recurseGlobals.useUtf8=useUtf8;
	recurseGlobals.ignoreCase=compiledExpression->ignoreCase;
	recurseGlobals.rightEdge=rightEdge;
	recurseGlobals.foundMatch=false;
	TextStreamPositionClear(&recurseGlobals.endStream);
	recurseGlobals.matchRecursionDepth=0;					// clear recursion depth
	recurseGlobals.abortFunction=abortFunction;

	localStream=*stream;
	endPosition=TextStreamPositionGet(&localStream)-numToSearch;

	// do initial test because regex can match 0 bytes
	recurseGlobals.leftEdge=(TextStreamPositionGet(&localStream)==endPosition)?leftEdge:false;
	recurseGlobals.initialNumToSearch=TextStreamPositionGet(stream)-TextStreamPositionGet(&localStream);
	fail=!RegexMatch(&localStream,compiledExpression,recurseGlobals.initialNumToSearch,&recurseGlobals);
	while(!fail&&!recurseGlobals.foundMatch&&TextStreamPositionGet(&localStream)>endPosition)	// if not found, and still something to look through, then do it
	{
		TextStreamPreviousChar(&localStream);				// move backward in the input
		if(useUtf8)											// step back to previous codePoint
		{
			Utf8Synchronize(&localStream);
		}
		recurseGlobals.leftEdge=(TextStreamPositionGet(&localStream)==endPosition)?leftEdge:false;
		recurseGlobals.initialNumToSearch=TextStreamPositionGet(stream)-TextStreamPositionGet(&localStream);
		fail=!RegexMatch(&localStream,compiledExpression,recurseGlobals.initialNumToSearch,&recurseGlobals);
	}
	*startStream=localStream;
	*endStream=recurseGlobals.endStream;
	*foundMatch=recurseGlobals.foundMatch;

	return(!fail);
}

void RegexHuntBackToFarthestPossibleStart(const TEXT_STREAM *stream,COMPILED_EXPRESSION *compiledExpression,uint64_t *numBack,TEXT_STREAM *endStream)
// Given a stream into some text, this will move backwards through the text to a point which
// is the farthest possible starting point for a regular expression which matches at the passed point.
// numBack is the number of bytes backwards that we moved, and endStream points to
// the starting point of a search.
// NOTE: this is too slow and ugly.
// NOTE: this does not need to work for UTF8, since it is never used in that context.
{
	TEXT_STREAM
		localStream;
	bool
		done;
	uint64_t
		currentBack;

	localStream=*stream;
	*numBack=0;
	*endStream=localStream;

	if(TextStreamPositionGet(&localStream))										// make sure there's room to move back further
	{
		done=false;
		currentBack=0;

		while(!done&&(compiledExpression->infiniteMatchLength||(currentBack<compiledExpression->maximumMatchLength)))
		{
			if(RegexMatchCodePointWithList(compiledExpression->ignoreCase,TextStreamChar(&localStream),compiledExpression->charsMatched))	// see if this is a matching codePoint
			{
				*endStream=localStream;
				*numBack=currentBack;

				currentBack++;

				done=!TextStreamPreviousChar(&localStream);						// done if we've hit the start
			}
			else
			{
				done=true;		// non matching, so this is the spot
			}
		}
	}
}
