//	Search/replace functions
//	Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

enum
{
	MATCHED_NOTHING,
};

static const char *errorDescriptions[]=
{
	"Find matched nothing -- could not continue",
};

typedef struct
{
	TEXT_UNIVERSE
		*searchForText;				// pointer to the search for text universe
	TEXT_UNIVERSE
		*replaceWithText;			// pointer to the replace with text universe
	bool
		useUtf8;					// true if we treat search text as UTF-8 (compare UTF-8 code points instead of bytes)
	bool
		ignoreCase;					// true if we should ignore case, false if not
	COMPILED_EXPRESSION
		*expression;				// if type is regex, this points to the compiled expression, otherwise it is NULL
	bool
		useSubexpression;			// if true, and if a regex search, then set matchUseOffset and matchUseLength for the subexpression given by subexpressionIndex
	uint32_t
		subexpressionIndex;			// the index of the subexpression which is to be used as the result of the search
	char
		*procedure;					// if non NULL, this points to a procedure to use instead of the replace text
	ABORT_TEST_FUNCTION
		*abortFunction;				// if non NULL, gives function to call to test for abort during search
	bool
		foundMatch;					// tells if the search found a match
	TEXT_STREAM
		matchStartStream,			// points to the start of a match (regardless of the state of useSubexpression)
		matchEndStream,				// points to the end of a match (regardless of the state of useSubexpression)
		matchUseStream;				// points to the part of the match to use (takes into account useSubexpression)
	uint64_t
		matchLength,				// gives the length of the match (regardless of the state of useSubexpression)
		matchUseLength;				// gives the length of the part of the match which is to be used as the result of the search (takes into account useSubexpression)
} SEARCH_RECORD;


uint32_t SearchToUpper(uint32_t codePoint)
// Convert the given codePoint to upper-case.
{
	return(Tcl_UniCharToUpper(codePoint));
}

uint32_t SearchToLower(uint32_t codePoint)
// Convert the given codePoint to lower-case.
{
	return(Tcl_UniCharToLower(codePoint));
}

bool LiteralMatch(const TEXT_STREAM *stream,const TEXT_STREAM *forTextStream,uint64_t numToSearch,bool useUtf8,bool ignoreCase,TEXT_STREAM *endStream)
// See if the text at stream matches the text in forTextStream.
// If there's a match, return true with endStream updated to point to the
// position in the stream immediately following the match.
// NOTE: there must be at least numToSearch bytes remaining in both stream and forTextStream.
{
	TEXT_STREAM
		forTextStreamCopy;
	uint32_t
		codePointA,
		codePointB;
	uint64_t
		endPosition;
	bool
		matching;

	matching=true;
	*endStream=*stream;													// copy input to output
	forTextStreamCopy=*forTextStream;									// copy the "for" stream locally so the passed structure doesn't get modified
	endPosition=TextStreamPositionGet(&forTextStreamCopy)+numToSearch;
	while((TextStreamPositionGet(&forTextStreamCopy)<endPosition)&&matching)
	{
		if(useUtf8)
		{
			codePointA=Utf8NextCodePoint(endStream);
			codePointB=Utf8NextCodePoint(&forTextStreamCopy);
		}
		else
		{
			codePointA=TextStreamChar(endStream);
			TextStreamNextChar(endStream);
			codePointB=TextStreamChar(&forTextStreamCopy);
			TextStreamNextChar(&forTextStreamCopy);
		}
		if(ignoreCase)
		{
			codePointA=SearchToLower(codePointA);
			codePointB=SearchToLower(codePointB);
		}
		matching=(codePointA==codePointB);
	}
	return(matching);
}

bool SearchForwardLiteral(const TEXT_STREAM *stream,const TEXT_STREAM *forTextStream,uint64_t forTextLength,uint64_t numToSearch,bool useUtf8,bool ignoreCase,bool *foundMatch,TEXT_STREAM *startStream,TEXT_STREAM *endStream,ABORT_TEST_FUNCTION *abortFunction)
// Search forward in stream for forTextStream.
// Look at up to numToSearch bytes of source data.
// If a match is located, return *foundMatch=true, with startStream pointing
// to the start of the match, and endStream pointing to the character immediately
// following the match.
// If there is a failure, ErrorSet, and return false.
{
	uint64_t
		endPosition;
	uint32_t
		abortCounter;
	bool
		fail;

	fail=false;
	*foundMatch=false;
	abortCounter=0;
	*startStream=*stream;

	if(forTextLength&&(forTextLength<=numToSearch))
	{
		endPosition=TextStreamPositionGet(startStream)+numToSearch-forTextLength;
		while(!fail&&!(*foundMatch)&&(TextStreamPositionGet(startStream)<=endPosition))
		{
			if(LiteralMatch(startStream,forTextStream,forTextLength,useUtf8,ignoreCase,endStream))
			{
				*foundMatch=true;
			}
			else
			{
				if(useUtf8)
				{
					Utf8NextCodePoint(startStream);			// step over UTF8 codepoint
				}
				else
				{
					TextStreamNextChar(startStream);		// step over the byte
				}

				abortCounter++;

				if(abortCounter>4095)					// limit how often we'll check for abort
				{
					if(abortFunction&&abortFunction())
					{
						ErrorSet("Search Aborted");
						fail=true;
					}
					abortCounter=0;
				}
			}
		}
	}
	return(!fail);
}

bool SearchBackwardLiteral(const TEXT_STREAM *stream,const TEXT_STREAM *forTextStream,uint64_t forTextLength,uint64_t numToSearch,bool useUtf8,bool ignoreCase,bool *foundMatch,TEXT_STREAM *startStream,TEXT_STREAM *endStream,ABORT_TEST_FUNCTION *abortFunction)
// Search backward in stream for forTextStream.
// Look at up to numToSearch bytes of source data.
// If a match is located, return *foundMatch=true, with startStream pointing
// to the start of the match, and endStream pointing to the character immediately
// following the match.
// If there is a failure, ErrorSet, and return false.
{
	uint64_t
		endPosition;
	uint32_t
		abortCounter;
	bool
		fail;

	fail=false;
	*foundMatch=false;
	abortCounter=0;
	*startStream=*stream;

	if(forTextLength&&(forTextLength<=numToSearch))
	{
		endPosition=TextStreamPositionGet(startStream)-numToSearch;
		TextStreamMoveRelative(startStream,-(forTextLength-1));		// back up in the source by forTextLength

		while(!fail&&!(*foundMatch)&&(TextStreamPositionGet(startStream)>endPosition))
		{
			TextStreamPreviousChar(startStream);				// move backward in the input
			if(useUtf8)											// step back to previous codePoint
			{
				Utf8Synchronize(startStream);
			}

			if(LiteralMatch(startStream,forTextStream,forTextLength,useUtf8,ignoreCase,endStream))
			{
				*foundMatch=true;
			}
			else
			{
				abortCounter++;

				if(abortCounter>4095)							// limit how often we'll check for abort
				{
					if(abortFunction&&abortFunction())
					{
						ErrorSet("Search Aborted");
						fail=true;
					}
					abortCounter=0;
				}
			}
		}
	}
	return(!fail);
}

static bool SearchForward(const TEXT_STREAM *stream,SEARCH_RECORD *searchRecord,bool leftEdge,bool rightEdge,uint64_t numToSearch)
// Search forward given searchRecord.
// Update searchRecord to describe what was or was not found.
// Return false if there was some sort of hard failure.
{
	bool
		fail;
	TEXT_STREAM
		forTextStream;

	fail=false;

	if(searchRecord->expression)
	{
		if(RegexSearchForward(stream,searchRecord->expression,searchRecord->useUtf8,leftEdge,rightEdge,numToSearch,numToSearch,&searchRecord->foundMatch,&searchRecord->matchStartStream,&searchRecord->matchEndStream,searchRecord->abortFunction))
		{
			searchRecord->matchLength=TextStreamPositionGet(&searchRecord->matchEndStream)-TextStreamPositionGet(&searchRecord->matchStartStream);
			if(searchRecord->useSubexpression)
			{
				if(RegexRegisterMatchGet(searchRecord->expression,searchRecord->subexpressionIndex,&searchRecord->matchUseStream,&searchRecord->matchUseLength))	// was sub-expression located?
				{
					// success
				}
				else
				{
					searchRecord->matchUseStream=searchRecord->matchEndStream;	// if not located, match only the very end
					searchRecord->matchUseLength=0;			// empty match
				}
			}
			else	// not using subexpression, so return the whole thing
			{
				searchRecord->matchUseStream=searchRecord->matchStartStream;
				searchRecord->matchUseLength=searchRecord->matchLength;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		TextStreamPositionSet(&forTextStream,searchRecord->searchForText,0);
		if(SearchForwardLiteral(stream,&forTextStream,TextUniverseTotalBytes(searchRecord->searchForText),numToSearch,searchRecord->useUtf8,searchRecord->ignoreCase,&searchRecord->foundMatch,&searchRecord->matchStartStream,&searchRecord->matchEndStream,searchRecord->abortFunction))
		{
			searchRecord->matchLength=TextStreamPositionGet(&searchRecord->matchEndStream)-TextStreamPositionGet(&searchRecord->matchStartStream);
			searchRecord->matchUseStream=searchRecord->matchStartStream;
			searchRecord->matchUseLength=searchRecord->matchLength;
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

static bool SearchBackward(const TEXT_STREAM *stream,SEARCH_RECORD *searchRecord,bool leftEdge,bool rightEdge,uint64_t numToSearch)
// Search backward given searchRecord.
// Update searchRecord to describe what was or was not found.
// Return false if there was some sort of hard failure.
{
	bool
		fail;
	TEXT_STREAM
		forTextStream;

	fail=false;
	if(searchRecord->expression)
	{
		if(RegexSearchBackward(stream,searchRecord->expression,searchRecord->useUtf8,leftEdge,rightEdge,numToSearch,&searchRecord->foundMatch,&searchRecord->matchStartStream,&searchRecord->matchEndStream,searchRecord->abortFunction))
		{
			searchRecord->matchLength=TextStreamPositionGet(&searchRecord->matchEndStream)-TextStreamPositionGet(&searchRecord->matchStartStream);
			if(searchRecord->useSubexpression)
			{
				if(RegexRegisterMatchGet(searchRecord->expression,searchRecord->subexpressionIndex,&searchRecord->matchUseStream,&searchRecord->matchUseLength))	// was sub-expression located?
				{
					// success
				}
				else
				{
					searchRecord->matchUseStream=searchRecord->matchEndStream;	// if not located, match only the very end
					searchRecord->matchUseLength=0;			// empty match
				}
			}
			else	// not using subexpression, so return the whole thing
			{
				searchRecord->matchUseStream=searchRecord->matchStartStream;
				searchRecord->matchUseLength=searchRecord->matchLength;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		TextStreamPositionSet(&forTextStream,searchRecord->searchForText,0);
		if(SearchBackwardLiteral(stream,&forTextStream,TextUniverseTotalBytes(searchRecord->searchForText),numToSearch,searchRecord->useUtf8,searchRecord->ignoreCase,&searchRecord->foundMatch,&searchRecord->matchStartStream,&searchRecord->matchEndStream,searchRecord->abortFunction))
		{
			searchRecord->matchLength=TextStreamPositionGet(&searchRecord->matchEndStream)-TextStreamPositionGet(&searchRecord->matchStartStream);
			searchRecord->matchUseStream=searchRecord->matchStartStream;
			searchRecord->matchUseLength=searchRecord->matchLength;
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

static void SearchRecordDispose(SEARCH_RECORD *searchRecord)
// Get rid of searchRecord, and anything allocated under it
{
	if(searchRecord->expression)		// if there is a selection expression, get rid of it
	{
		RegexFree(searchRecord->expression);
	}
	if(searchRecord->procedure)			// if there is a procedure, get rid of it
	{
		MDisposePtr(searchRecord->procedure);
	}
	MDisposePtr(searchRecord);
}

static SEARCH_RECORD *SearchRecordCreate(TEXT_UNIVERSE *searchForText,TEXT_UNIVERSE *replaceWithText,bool useUtf8,bool ignoreCase,bool selectionExpr,bool useSubexpression,uint32_t subexpressionIndex,bool replaceProc,ABORT_TEST_FUNCTION *abortFunction)
// Create a search record that describes the search, and remembers some stuff about it.
// If there is a problem, ErrorSet, and return NULL.
{
	SEARCH_RECORD
		*searchRecord;
	COMPILED_EXPRESSION
		*compiledExpression;
	uint64_t
		totalBytes;
	TEXT_STREAM
		stream;
	bool
		fail;

	fail=false;
	if((searchRecord=(SEARCH_RECORD *)MNewPtr(sizeof(SEARCH_RECORD))))
	{
		searchRecord->searchForText=searchForText;			// remember these for later
		searchRecord->replaceWithText=replaceWithText;
		searchRecord->useUtf8=useUtf8;
		searchRecord->ignoreCase=ignoreCase;
		searchRecord->expression=NULL;						// assume no selection expression yet
		searchRecord->useSubexpression=useSubexpression;
		searchRecord->subexpressionIndex=subexpressionIndex;
		searchRecord->procedure=NULL;						// assume no procedure yet
		searchRecord->abortFunction=abortFunction;
		searchRecord->foundMatch=false;						// no match yet
		TextStreamPositionClear(&searchRecord->matchStartStream);
		TextStreamPositionClear(&searchRecord->matchEndStream);
		TextStreamPositionClear(&searchRecord->matchUseStream);
		searchRecord->matchLength=0;
		searchRecord->matchUseLength=0;

		if(replaceProc)										// if treating replace as procedure, we need to grab the text from the replace buffer, and place it into memory
		{
			totalBytes=TextUniverseTotalBytes(replaceWithText);
			if((searchRecord->procedure=(char *)MNewPtr(totalBytes+1)))	// make a buffer
			{
				TextUniverseTextExtract(replaceWithText,0,(uint8_t *)searchRecord->procedure,totalBytes,NULL);
				searchRecord->procedure[totalBytes]='\0';	// terminate the string
			}
			else
			{
				fail=true;
			}
		}
		if(!fail)
		{
			if(selectionExpr)
			{
				TextStreamPositionSet(&stream,searchForText,0);			// point to the start
				if((compiledExpression=RegexCompileStream(&stream,ignoreCase)))
				{
					searchRecord->expression=compiledExpression;
					return(searchRecord);
				}
			}
			else
			{
				return(searchRecord);
			}
		}
		if(searchRecord->procedure)			// if there is a procedure, get rid of it
		{
			MDisposePtr(searchRecord->procedure);
		}
		MDisposePtr(searchRecord);
	}
	return(NULL);
}

bool EditorFind(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,TEXT_UNIVERSE *searchForText,bool useUtf8,bool backward,bool wrapAround,bool ignoreCase,bool limitScope,bool selectionExpr,bool useSubexpression,uint32_t subexpressionIndex,bool *foundMatch,SELECTION_UNIVERSE *resultSelectionUniverse,ABORT_TEST_FUNCTION *abortFunction)
// Find searchForText in buffer, starting at the cursor position of selectionUniverse, or the end of
// the current selection.
// If anything is found, return it in resultSelectionUniverse, else, do not modify resultSelectionUniverse.
// foundMatch reflects the status of the search.
// If there is some problem during the search, ErrorSet, return false.
// NOTE: resultSelectionUniverse must be different from selectionUniverse.
// NOTE: if selectionExpr is true, and useSubexpression is true, then
// this will select only the part of the expression that matches the
// subexpression given by subexpressionIndex. If there was no match
// of the subexpression this will put the cursor at the end of
// the match.
{
	TEXT_STREAM
		stream;
	SEARCH_RECORD
		*searchRecord;
	uint64_t
		startPosition,
		endPosition,
		length;
	TEXT_UNIVERSE
		*searchInText;
	uint64_t
		totalBytes;
	bool
		fail;

	fail=*foundMatch=false;
	if((searchRecord=SearchRecordCreate(searchForText,NULL,useUtf8,ignoreCase,selectionExpr,useSubexpression,subexpressionIndex,false,abortFunction)))
	{
		PlatformBusyShow();
		searchInText=buffer->textUniverse;
		totalBytes=TextUniverseTotalBytes(searchInText);
		SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);			// find ends of old selection, or cursor position

		if(!backward)
		{
			if(!limitScope)
			{
				TextStreamPositionSet(&stream,searchInText,endPosition);					// point stream at the start of the search (end of the current selection if any)
				if(SearchForward(&stream,searchRecord,endPosition==0,true,totalBytes-endPosition))
				{
					if((!searchRecord->foundMatch)&&wrapAround)
					{
						TextStreamPositionSet(&stream,searchInText,0);						// point to the start on wrap around
						fail=!SearchForward(&stream,searchRecord,true,startPosition==totalBytes,startPosition);
					}
				}
				else
				{
					fail=true;
				}
			}
			else	// search forward inside the current selection (and ignore wrapAround)
			{
				length=0;
				while(!fail&&!searchRecord->foundMatch&&SelectionAtOrAfterPositionGet(selectionUniverse,startPosition,&startPosition,&length))
				{
					TextStreamPositionSet(&stream,searchInText,startPosition);				// point stream at the start of the search
					fail=!SearchForward(&stream,searchRecord,true,true,length);				// edges of selection are considered edges for the search
					startPosition+=length;
				}
			}
		}
		else
		{
			if(!limitScope)
			{
				TextStreamPositionSet(&stream,searchInText,startPosition);					// point stream to the start of the search (at the top of the current selection)
				if(SearchBackward(&stream,searchRecord,true,startPosition==totalBytes,startPosition))
				{
					if((!searchRecord->foundMatch)&&wrapAround)
					{
						TextStreamPositionSet(&stream,searchInText,totalBytes);				// wrap around to the end
						fail=!SearchBackward(&stream,searchRecord,endPosition==0,true,totalBytes-endPosition);
					}
				}
				else
				{
					fail=true;
				}
			}
			else	// search backward inside the current selection (and ignore wrapAround)
			{
				startPosition=endPosition;	// start at the end
				while(!fail&&!searchRecord->foundMatch&&startPosition&&SelectionAtOrBeforePositionGet(selectionUniverse,startPosition-1,&startPosition,&length))
				{
					TextStreamPositionSet(&stream,searchInText,startPosition+length);		// point stream to the start of the search (at the end of the current selection)
					fail=!SearchBackward(&stream,searchRecord,true,true,length);			// edges of selection are considered edges for the search
				}
			}
		}
		if((!fail)&&searchRecord->foundMatch)
		{
			(*foundMatch)=true;

			SelectionDeleteAll(resultSelectionUniverse);	// remove all old selections

			startPosition=TextStreamPositionGet(&searchRecord->matchUseStream);

			if(SelectionRangeSet(resultSelectionUniverse,startPosition,searchRecord->matchUseLength))
			{
				SelectionCursorPositionSet(resultSelectionUniverse,startPosition);
			}
			else
			{
				fail=true;
			}
		}
		SearchRecordDispose(searchRecord);
		PlatformBusyUnShow();
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static bool FindAllForward(TEXT_UNIVERSE *searchInText,bool leftEdge,bool rightEdge,uint64_t lowPosition,uint64_t highPosition,SEARCH_RECORD *searchRecord,bool useSubexpression,uint32_t subexpressionIndex,SELECTION_UNIVERSE *selectionUniverse,bool *foundMatch,uint64_t *firstMatchPosition)
// Find all occurrences of searchRecord between lowPosition, and highPosition, adding the selections
// to selectionUniverse.
// If there is a problem, ErrorSet and return false.
// If matches were found, return true in foundMatch, and the position of the first match found
// in firstMatchPosition.
// NOTE: even if false is returned, foundMatch, firstMatchPosition, and selectionUniverse are
// valid.
{
	TEXT_STREAM
		stream;
	bool
		fail;

	fail=*foundMatch=false;
	TextStreamPositionSet(&stream,searchInText,lowPosition);				// locate place to start the search
	searchRecord->foundMatch=true;
	while(!fail&&searchRecord->foundMatch)
	{
		if(SearchForward(&stream,searchRecord,leftEdge,rightEdge,highPosition-TextStreamPositionGet(&stream)))
		{
			if(searchRecord->foundMatch)
			{
				if(searchRecord->matchLength)								// finding an empty string is fatal to find all
				{
					leftEdge=false;											// no matter what, we just moved off the left edge
					stream=searchRecord->matchEndStream;					// start searching again at the end of the current match

					if(!(*foundMatch))
					{
						(*firstMatchPosition)=TextStreamPositionGet(&searchRecord->matchUseStream);
						(*foundMatch)=true;
					}
					fail=!SelectionRangeSet(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),searchRecord->matchUseLength);
				}
				else
				{
					ErrorSet(errorDescriptions[MATCHED_NOTHING]);
					fail=true;
				}
			}
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

static bool FindAllBackward(TEXT_UNIVERSE *searchInText,bool leftEdge,bool rightEdge,uint64_t lowPosition,uint64_t highPosition,SEARCH_RECORD *searchRecord,bool useSubexpression,uint32_t subexpressionIndex,SELECTION_UNIVERSE *selectionUniverse,bool *foundMatch,uint64_t *firstMatchPosition)
// Find all occurrences of searchRecord between lowPosition, and highPosition, adding the selections
// to selectionUniverse.
// If there is a problem, ErrorSet and return false.
// If matches were found, return true in foundMatch, and the position of the first match found
// in firstMatchPosition.
// NOTE: even if false is returned, foundMatch, firstMatchPosition, and selectionUniverse are
// valid.
{
	TEXT_STREAM
		stream;
	bool
		fail;

	fail=*foundMatch=false;
	TextStreamPositionSet(&stream,searchInText,highPosition);			// locate place to start the search
	searchRecord->foundMatch=true;
	while(!fail&&searchRecord->foundMatch)
	{
		if(SearchBackward(&stream,searchRecord,leftEdge,rightEdge,TextStreamPositionGet(&stream)-lowPosition))
		{
			if(searchRecord->foundMatch)
			{
				if(searchRecord->matchLength)							// finding an empty string is fatal to find all
				{
					rightEdge=false;									// no matter what, just left the right edge
					stream=searchRecord->matchStartStream;				// start searching again at the beginning of the current match

					if(!(*foundMatch))
					{
						(*firstMatchPosition)=TextStreamPositionGet(&searchRecord->matchUseStream);
						(*foundMatch)=true;
					}
					fail=!SelectionRangeSet(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),searchRecord->matchUseLength);
				}
				else
				{
					ErrorSet(errorDescriptions[MATCHED_NOTHING]);
					fail=true;
				}
			}
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

bool EditorFindAll(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,TEXT_UNIVERSE *searchForText,bool useUtf8,bool backward,bool wrapAround,bool ignoreCase,bool limitScope,bool selectionExpr,bool useSubexpression,uint32_t subexpressionIndex,bool *foundMatch,SELECTION_UNIVERSE *resultSelectionUniverse,ABORT_TEST_FUNCTION *abortFunction)
// Find all occurrences of searchForText using the given rules. If there is an error,
// or the user aborts, ErrorSet, and return false. If nothing is found, set found to false.
// If found returns true, resultSelectionUniverse will be modified to contain the
// list of selections of found text.
// NOTE: resultSelectionUniverse must be different from selectionUniverse.
// NOTE: if selectionExpr is true, and useSubexpression is true, then
// this will select only the part of the expression that matches the
// subexpression given by subexpressionIndex. If there was no match
// of the subexpression this will not select anything for that match.
{
	SEARCH_RECORD
		*searchRecord;
	uint64_t
		startPosition,
		endPosition,
		selectionLength;
	uint64_t
		firstMatchPosition,
		tempFirstMatchPosition;
	TEXT_UNIVERSE
		*searchInText;
	uint64_t
		totalBytes;
	bool
		hadMatch,
		fail;

	fail=*foundMatch=false;
	if((searchRecord=SearchRecordCreate(searchForText,NULL,useUtf8,ignoreCase,selectionExpr,useSubexpression,subexpressionIndex,false,abortFunction)))
	{
		PlatformBusyShow();
		searchInText=buffer->textUniverse;
		totalBytes=TextUniverseTotalBytes(searchInText);

		if(!backward)
		{
			if(!limitScope)
			{
				SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);		// find ends of old selection, or cursor position
				if(FindAllForward(searchInText,endPosition==0,true,endPosition,totalBytes,searchRecord,useSubexpression,subexpressionIndex,resultSelectionUniverse,foundMatch,&firstMatchPosition))
				{
					if(wrapAround)
					{
						fail=!FindAllForward(searchInText,true,startPosition==totalBytes,0,startPosition,searchRecord,useSubexpression,subexpressionIndex,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition);
						if(!(*foundMatch))
						{
							*foundMatch=hadMatch;
							firstMatchPosition=tempFirstMatchPosition;
						}
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				startPosition=0;
				selectionLength=0;
				while(!fail&&SelectionAtOrAfterPositionGet(selectionUniverse,startPosition,&startPosition,&selectionLength))
				{
					if(FindAllForward(searchInText,true,true,startPosition,startPosition+selectionLength,searchRecord,useSubexpression,subexpressionIndex,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition))
					{
						if(!(*foundMatch))
						{
							*foundMatch=hadMatch;
							firstMatchPosition=tempFirstMatchPosition;
						}
					}
					else
					{
						fail=true;
					}
					startPosition+=selectionLength;
				}
			}
		}
		else
		{
			if(!limitScope)
			{
				SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);		// find ends of old selection, or cursor position
				if(FindAllBackward(searchInText,true,startPosition==totalBytes,0,startPosition,searchRecord,useSubexpression,subexpressionIndex,resultSelectionUniverse,foundMatch,&firstMatchPosition))
				{
					if(wrapAround)
					{
						fail=!FindAllBackward(searchInText,endPosition==0,true,endPosition,totalBytes,searchRecord,useSubexpression,subexpressionIndex,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition);
						if(!(*foundMatch))
						{
							*foundMatch=hadMatch;
							firstMatchPosition=tempFirstMatchPosition;
						}
					}
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);
				startPosition=endPosition;	// start at the end

				while(!fail&&startPosition&&SelectionAtOrBeforePositionGet(selectionUniverse,startPosition-1,&startPosition,&selectionLength))
				{
					if(FindAllBackward(searchInText,true,true,startPosition,startPosition+selectionLength,searchRecord,useSubexpression,subexpressionIndex,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition))
					{
						if(!(*foundMatch))
						{
							*foundMatch=hadMatch;
							firstMatchPosition=tempFirstMatchPosition;
						}
					}
					else
					{
						fail=true;
					}
				}
			}
		}
		if(*foundMatch)																				// got something, so change result universe
		{
			SelectionCursorPositionSet(resultSelectionUniverse,firstMatchPosition);
		}
		SearchRecordDispose(searchRecord);
		PlatformBusyUnShow();
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static bool RegisterReferencesReplace(TEXT_UNIVERSE *replaceWithUniverse,SEARCH_RECORD *searchedForText)
// Run through replaceWithUniverse looking for register specifications, if any are seen,
// replace them with what searchForText found.
// If there is a problem, ErrorSet, and return false.
// NOTE: registers are specified in replaceWithUniverse using \0 through \9
// ignore any other \ in replaceWithUniverse.
{
	TEXT_STREAM
		stream,
		startStream;
	uint64_t
		currentPosition,
		matchLength;
	uint8_t
		character;
	bool
		fail;

	fail=false;
	TextStreamPositionSet(&stream,replaceWithUniverse,0);
	while(!TextStreamAtEnd(&stream)&&!fail)
	{
		if(TextStreamChar(&stream)!='\\')
		{
			TextStreamNextChar(&stream);
		}
		else
		{
			currentPosition=TextStreamPositionGet(&stream);
			TextStreamNextChar(&stream);
			if(!TextStreamAtEnd(&stream))		// make sure there is still more data in the stream
			{
				character=TextStreamChar(&stream);
				if(character>='0'&&character<='9')
				{
					character-='0';														// make into an index
					TextUniverseDelete(replaceWithUniverse,currentPosition,2);			// remove the register specification
					if(RegexRegisterMatchGet(searchedForText->expression,character,&startStream,&matchLength))	// see if anything to insert
					{
						fail=!TextUniverseInsertFromUniverse(TextStreamUniverseGet(&startStream),TextStreamPositionGet(&startStream),replaceWithUniverse,currentPosition,matchLength);
						currentPosition+=matchLength;								// move past stuff just inserted
					}
					TextStreamPositionSet(&stream,replaceWithUniverse,currentPosition);	// adjust stream now that the universe has changed
				}
			}
		}
	}
	return(!fail);
}

static bool FoundVariableCreate(EDITOR_BUFFER *replaceInUniverse,uint64_t replaceOffset,uint64_t replaceLength,SEARCH_RECORD *searchRecord)
// Create the "found" variable, and $n variables which are passed to the Tcl procedure as globals.
// NOTE: this routine, and ReplaceSelectedText are the ONLY routines in this file that rely on Tcl...
// If there is a problem, ErrorSet, and return false.
{
	TEXT_STREAM
		startStream;
	uint64_t
		matchLength;
	char
		*locatedText;
	int
		varCount;
	char
		varName[32];
	bool
		fail;

	fail=false;
	if((locatedText=(char *)MNewPtr(replaceLength+1)))	// make pointer to text that was located (this pointer will used to hold all the extracted text)
	{
		TextUniverseTextExtract(replaceInUniverse->textUniverse,replaceOffset,(uint8_t *)locatedText,replaceLength,NULL);
		locatedText[replaceLength]='\0';			// terminate the string
		if(!Tcl_SetVar(tclInterpreter,"found",locatedText,TCL_LEAVE_ERR_MSG))
		{
			ErrorSet("%s",Tcl_GetStringResult(tclInterpreter));
			fail=true;
		}
		MDisposePtr(locatedText);
	}
	else
	{
		fail=true;
	}

	if(!fail&&searchRecord->expression)			// if it is an expression, make a list of 10 more variables (one for each tagged subexpression)
	{
		for(varCount=0;!fail&&varCount<MAX_REGISTERS;varCount++)
		{
			sprintf(varName,"%d",varCount);		// create variable name

			if(RegexRegisterMatchGet(searchRecord->expression,varCount,&startStream,&matchLength))	// see if anything to add
			{
				if((locatedText=(char *)MNewPtr(matchLength+1)))		// make pointer to the text
				{
					TextUniverseTextExtract(replaceInUniverse->textUniverse,TextStreamPositionGet(&startStream),(uint8_t *)locatedText,matchLength,NULL);
					locatedText[matchLength]='\0';					// terminate the string
					if(!Tcl_SetVar(tclInterpreter,varName,locatedText,TCL_LEAVE_ERR_MSG))
					{
						ErrorSet("%s",Tcl_GetStringResult(tclInterpreter));
						fail=true;
					}
					MDisposePtr(locatedText);
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				if(!Tcl_SetVar(tclInterpreter,varName,"",TCL_LEAVE_ERR_MSG))
				{
					ErrorSet("%s",Tcl_GetStringResult(tclInterpreter));
					fail=true;
				}
			}
		}
	}
	return(!fail);
}

static bool SearchedTextReplace(EDITOR_BUFFER *replaceInBuffer,uint64_t replaceOffset,uint64_t replaceLength,SEARCH_RECORD *searchRecord,uint64_t *bytesReplaced)
// Given text that was just searched for by searchForText, replace it
// based on the search type.
// If there is a problem, attempt to leave the text unmodified,
// ErrorSet, and return false.
{
	bool
		fail;
	uint64_t
		length;
	TEXT_UNIVERSE
		*alternateUniverse;
	char
		*stringResult;
	uint64_t
		totalBytes;
	TEXT_STREAM
		stream;

	fail=false;
	if(searchRecord->procedure)								// handle replacement by calling a Tcl procedure
	{
		if(FoundVariableCreate(replaceInBuffer,replaceOffset,replaceLength,searchRecord))
		{
			if(Tcl_Eval(tclInterpreter,searchRecord->procedure)==TCL_OK)
			{
				stringResult=(char *)Tcl_GetStringResult(tclInterpreter);
				length=strlen(stringResult);

				fail=!EditorTextReplace(replaceInBuffer,replaceOffset,replaceLength,(uint8_t *)stringResult,length,false);
				(*bytesReplaced)=length;
			}
			else
			{
				ErrorSet("%s",Tcl_GetStringResult(tclInterpreter));
				fail=true;
			}
			Tcl_ResetResult(tclInterpreter);				// do not leave any Tcl result around (other stuff may append to it)
		}
		else
		{
			fail=true;
		}
	}
	else													// not a Tcl procedure
	{
		if(searchRecord->expression)						// regex search -- if so, deal with tag references in the replacement text
		{
			if((alternateUniverse=TextUniverseOpen()))		// create alternate universe where we make the real text to replace
			{
				if(TextUniverseInsertFromUniverse(searchRecord->replaceWithText,0,alternateUniverse,TextUniverseTotalBytes(alternateUniverse),TextUniverseTotalBytes(searchRecord->replaceWithText)))	// copy replace with text into alternate universe
				{
					if(RegisterReferencesReplace(alternateUniverse,searchRecord))
					{
						totalBytes=TextUniverseTotalBytes(alternateUniverse);
						TextStreamPositionSet(&stream,alternateUniverse,0);
						fail=!EditorTextReplaceWithStream(replaceInBuffer,replaceOffset,replaceLength,&stream,totalBytes,false);
						(*bytesReplaced)=totalBytes;
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
				TextUniverseClose(alternateUniverse);
			}
			else
			{
				fail=true;
			}
		}
		else												// not a regex search
		{
			totalBytes=TextUniverseTotalBytes(searchRecord->replaceWithText);
			TextStreamPositionSet(&stream,searchRecord->replaceWithText,0);
			fail=!EditorTextReplaceWithStream(replaceInBuffer,replaceOffset,replaceLength,&stream,totalBytes,false);
			(*bytesReplaced)=totalBytes;
		}
	}
	return(!fail);
}

bool EditorReplace(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,TEXT_UNIVERSE *searchForText,TEXT_UNIVERSE *replaceWithText,bool useUtf8,bool backward,bool wrapAround,bool ignoreCase,bool limitScope,bool selectionExpr,bool useSubexpression,uint32_t subexpressionIndex,bool replaceProc,bool *foundMatch,SELECTION_UNIVERSE *resultSelectionUniverse,ABORT_TEST_FUNCTION *abortFunction)
// Find searchForText in buffer, starting at the cursor position, or the _start_ of
// the current selection of selectionUniverse.
// If searchForText is not located, return false in foundMatch.
// If it is located, replace what was found with replaceWithText, modify resultSelectionUniverse to
// be the selection of what was replaced, and return true in foundMatch.
// If there is a problem, ErrorSet, and return false.
// NOTE: resultSelectionUniverse must be different from selectionUniverse.
{
	TEXT_STREAM
		stream;
	SEARCH_RECORD
		*searchRecord;
	uint64_t
		startPosition,
		endPosition,
		length,
		bytesReplaced;
	uint64_t
		locatedPosition,
		locatedLength;
	TEXT_UNIVERSE
		*searchInText;
	uint64_t
		totalBytes;
	bool
		fail;

	fail=*foundMatch=false;
	if((searchRecord=SearchRecordCreate(searchForText,replaceWithText,useUtf8,ignoreCase,selectionExpr,useSubexpression,subexpressionIndex,replaceProc,abortFunction)))
	{
		PlatformBusyShow();
		searchInText=buffer->textUniverse;
		totalBytes=TextUniverseTotalBytes(searchInText);
		SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);			// find ends of old selection, or cursor position

		if(!backward)
		{
			if(!limitScope)
			{
				TextStreamPositionSet(&stream,searchInText,startPosition);					// locate place to start the search (at the start of the current selection if any)
				if(SearchForward(&stream,searchRecord,startPosition==0,true,totalBytes-startPosition))
				{
					if(!searchRecord->foundMatch&&wrapAround)
					{
						TextStreamPositionSet(&stream,searchInText,0);						// wrap to the start
						fail=!SearchForward(&stream,searchRecord,true,startPosition==totalBytes,startPosition);
					}
				}
				else
				{
					fail=true;
				}
			}
			else	// replace forward inside the current selection (and ignore wrapAround)
			{
				length=0;
				while(!fail&&!searchRecord->foundMatch&&SelectionAtOrAfterPositionGet(selectionUniverse,startPosition,&startPosition,&length))
				{
					TextStreamPositionSet(&stream,searchInText,startPosition);				// point stream at the start of the search
					fail=!SearchForward(&stream,searchRecord,true,true,length);				// edges of selection are considered edges for the search
					startPosition+=length;
				}
			}
		}
		else
		{
			if(!limitScope)
			{
				TextStreamPositionSet(&stream,searchInText,endPosition);					// locate place to start the search
				if(SearchBackward(&stream,searchRecord,true,endPosition==totalBytes,endPosition))
				{
					if(!searchRecord->foundMatch&&wrapAround)
					{
						TextStreamPositionSet(&stream,searchInText,totalBytes);				// locate place to start the search
						fail=!SearchBackward(&stream,searchRecord,endPosition==0,true,totalBytes-endPosition);
					}
				}
				else
				{
					fail=true;
				}
			}
			else	// search backward inside the current selection (and ignore wrapAround)
			{
				startPosition=endPosition;	// start at the end
				while(!fail&&!searchRecord->foundMatch&&startPosition&&SelectionAtOrBeforePositionGet(selectionUniverse,startPosition-1,&startPosition,&length))
				{
					TextStreamPositionSet(&stream,searchInText,startPosition+length);		// point stream to the start of the search (at the end of the current selection)
					fail=!SearchBackward(&stream,searchRecord,true,true,length);			// edges of selection are considered edges for the search
				}
			}
		}
		if(!fail&&searchRecord->foundMatch)
		{
			(*foundMatch)=true;

			locatedPosition=TextStreamPositionGet(&searchRecord->matchUseStream);
			locatedLength=searchRecord->matchUseLength;

			UndoGroupBegin(buffer);
			EditorReplaceStart(buffer);
			if(!SearchedTextReplace(buffer,locatedPosition,locatedLength,searchRecord,&bytesReplaced))
			{
				bytesReplaced=0;	// if the replace failed for some reason, select nothing
				fail=true;			// report the failure
			}
			EditorReplaceEnd(buffer);
			UndoGroupEndStrict(buffer);

			SelectionDeleteAll(resultSelectionUniverse);	// remove all old selections
			if(SelectionRangeSet(resultSelectionUniverse,locatedPosition,bytesReplaced))
			{
				SelectionCursorPositionSet(resultSelectionUniverse,locatedPosition);
			}
			else
			{
				fail=true;
			}
		}
		SearchRecordDispose(searchRecord);
		PlatformBusyUnShow();
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static bool ReplaceAllForward(EDITOR_BUFFER *buffer,bool leftEdge,bool rightEdge,uint64_t lowPosition,uint64_t *highPosition,SEARCH_RECORD *searchRecord,SELECTION_UNIVERSE *selectionUniverse,bool *foundMatch,uint64_t *firstMatchPosition)
// Replace all occurrences of searchRecord between lowPosition, and highPosition, adding the selections
// to selectionUniverse.
// If there is a problem, ErrorSet and return false.
// If matches were found, return true in foundMatch, and the position of the first match found
// in firstMatchPosition.
// highPosition is modified to reflect the new highPosition after all replacements have been made.
// NOTE: even if false is returned, foundMatch, highPosition, and selectionUniverse are
// valid.
{
	TEXT_STREAM
		stream;
	uint64_t
		nextSearchPosition,
		bytesReplaced;
	uint64_t
		newHighPosition;
	bool
		fail;

	fail=*foundMatch=false;
	nextSearchPosition=lowPosition;
	searchRecord->foundMatch=true;
	newHighPosition=*highPosition;
	while(!fail&&searchRecord->foundMatch)
	{
		TextStreamPositionSet(&stream,buffer->textUniverse,nextSearchPosition);		// locate place to start the search

		if(nextSearchPosition!=lowPosition)
		{
			leftEdge=false;													// no longer at left edge
		}

		if(SearchForward(&stream,searchRecord,leftEdge,rightEdge,newHighPosition-nextSearchPosition))
		{
			if(searchRecord->foundMatch)									// see if a match was found, if not, we are done
			{
				if(searchRecord->matchLength)								// finding an empty string is fatal to replace all
				{
					if(!(*foundMatch))
					{
						(*firstMatchPosition)=TextStreamPositionGet(&searchRecord->matchUseStream);	// remember the first match position so we can home to it when done
						(*foundMatch)=true;
					}
					if(SearchedTextReplace(buffer,TextStreamPositionGet(&searchRecord->matchUseStream),searchRecord->matchUseLength,searchRecord,&bytesReplaced))
					{
						nextSearchPosition=TextStreamPositionGet(&searchRecord->matchStartStream)+searchRecord->matchLength-searchRecord->matchUseLength+bytesReplaced;	// move to the new position after the replacement
						newHighPosition-=searchRecord->matchUseLength;		// subtract off bytes that were replaced
						newHighPosition+=bytesReplaced;						// add back number that replaced them

						SelectionRangeDelete(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),searchRecord->matchUseLength);
						if(SelectionRangeInsert(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),bytesReplaced))
						{
							fail=!SelectionRangeSet(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),bytesReplaced);
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					ErrorSet(errorDescriptions[MATCHED_NOTHING]);
					fail=true;
				}
			}
		}
		else
		{
			fail=true;
		}
	}
	*highPosition=newHighPosition;
	return(!fail);
}

static bool ReplaceAllBackward(EDITOR_BUFFER *buffer,bool leftEdge,bool rightEdge,uint64_t lowPosition,uint64_t *highPosition,SEARCH_RECORD *searchRecord,SELECTION_UNIVERSE *selectionUniverse,bool *foundMatch,uint64_t *firstMatchPosition)
// Replace all occurrences of searchRecord between lowPosition, and highPosition, adding the selections
// to selectionUniverse.
// If there is a problem, ErrorSet and return false.
// If matches were found, return true in foundMatch, and the position of the first match found
// in firstMatchPosition.
// highPosition is modified to reflect the new highPosition after all replacements have been made.
// NOTE: even if false is returned, foundMatch, highPosition, and selectionUniverse are
// valid.
{
	TEXT_STREAM
		stream;
	uint64_t
		nextSearchPosition,
		bytesReplaced;
	uint64_t
		newHighPosition;
	bool
		fail;

	fail=*foundMatch=false;
	nextSearchPosition=*highPosition;
	searchRecord->foundMatch=true;
	newHighPosition=*highPosition;
	while(!fail&&searchRecord->foundMatch)
	{
		TextStreamPositionSet(&stream,buffer->textUniverse,nextSearchPosition);		// locate place to start the search

		if(nextSearchPosition!=newHighPosition)
		{
			rightEdge=false;												// no longer at the right edge
		}

		if(SearchBackward(&stream,searchRecord,leftEdge,rightEdge,nextSearchPosition-lowPosition))
		{
			if(searchRecord->foundMatch)									// see if a match was found, if not, we are done
			{
				if(searchRecord->matchLength)								// finding an empty string is fatal to replace all
				{
					nextSearchPosition=TextStreamPositionGet(&searchRecord->matchStartStream);
					if(!(*foundMatch))										// if no matches found so far ...
					{
						(*firstMatchPosition)=TextStreamPositionGet(&searchRecord->matchUseStream);		// remember end of selections (sort of, will be adjusted at the end)
						(*foundMatch)=true;
					}
					if(SearchedTextReplace(buffer,TextStreamPositionGet(&searchRecord->matchUseStream),searchRecord->matchUseLength,searchRecord,&bytesReplaced))
					{
						newHighPosition-=searchRecord->matchUseLength;		// subtract off bytes that were replaced
						newHighPosition+=bytesReplaced;						// add back number that replaced them

						SelectionRangeDelete(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),searchRecord->matchUseLength);
						if(SelectionRangeInsert(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),bytesReplaced))
						{
							fail=!SelectionRangeSet(selectionUniverse,TextStreamPositionGet(&searchRecord->matchUseStream),bytesReplaced);
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					ErrorSet(errorDescriptions[MATCHED_NOTHING]);
					fail=true;
				}
			}
		}
		else
		{
			fail=true;
		}
	}
	*firstMatchPosition+=newHighPosition-(*highPosition);
	*highPosition=newHighPosition;
	return(!fail);
}

bool EditorReplaceAll(EDITOR_BUFFER *buffer,SELECTION_UNIVERSE *selectionUniverse,TEXT_UNIVERSE *searchForText,TEXT_UNIVERSE *replaceWithText,bool useUtf8,bool backward,bool wrapAround,bool ignoreCase,bool limitScope,bool selectionExpr,bool useSubexpression,uint32_t subexpressionIndex,bool replaceProc,bool *foundMatch,SELECTION_UNIVERSE *resultSelectionUniverse,ABORT_TEST_FUNCTION *abortFunction)
// Replace all occurrences of searchForText with replaceWithText using the given rules, if none are found,
// return false in foundMatch.
// If foundMatch is true, modify resultSelectionUniverse to be the selection of all replaced text.
// If there is a problem, ErrorSet, return false.
// NOTE: resultSelectionUniverse must be different from selectionUniverse.
{
	SEARCH_RECORD
		*searchRecord;
	int64_t
		replaceOffset;
	uint64_t
		startPosition,
		endPosition,
		selectionLength;
	uint64_t
		firstMatchPosition,
		tempFirstMatchPosition,
		highPosition;
	TEXT_UNIVERSE
		*searchInText;
	uint64_t
		totalBytes;
	bool
		hadMatch,
		fail;
	SELECTION_UNIVERSE
		*oldSelectionUniverse;

	fail=*foundMatch=false;
	if((searchRecord=SearchRecordCreate(searchForText,replaceWithText,useUtf8,ignoreCase,selectionExpr,useSubexpression,subexpressionIndex,replaceProc,abortFunction)))
	{
		PlatformBusyShow();
		if((oldSelectionUniverse=SelectionUniverseOpen()))						// copy this, since it must not change during the replacement
		{
			if(SelectionUniverseCopy(selectionUniverse,oldSelectionUniverse))	// copy the old one into it
			{
				searchInText=buffer->textUniverse;
				totalBytes=TextUniverseTotalBytes(searchInText);

				firstMatchPosition=0;
				UndoGroupBegin(buffer);
				EditorReplaceStart(buffer);
				if(!backward)
				{
					if(!limitScope)
					{
						SelectionEndPositionsGet(oldSelectionUniverse,&startPosition,&endPosition);		// find ends of old selection, or cursor position
						highPosition=totalBytes;
						if(ReplaceAllForward(buffer,startPosition==0,true,startPosition,&highPosition,searchRecord,resultSelectionUniverse,foundMatch,&firstMatchPosition))
						{
							if(wrapAround)
							{
								highPosition=startPosition;
								fail=!ReplaceAllForward(buffer,true,highPosition==totalBytes,0,&highPosition,searchRecord,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition);
								if(!(*foundMatch))
								{
									*foundMatch=hadMatch;
									firstMatchPosition=tempFirstMatchPosition;
								}
								else
								{
									firstMatchPosition+=highPosition-startPosition;	// move home position due to text changes above
								}
							}
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						replaceOffset=0;
						startPosition=0;
						selectionLength=0;
						while(!fail&&SelectionAtOrAfterPositionGet(oldSelectionUniverse,startPosition,&startPosition,&selectionLength))
						{
							endPosition=startPosition+replaceOffset+selectionLength;
							if(ReplaceAllForward(buffer,true,true,startPosition+replaceOffset,&endPosition,searchRecord,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition))
							{
								replaceOffset+=endPosition-(startPosition+replaceOffset+selectionLength);
								if(!(*foundMatch))
								{
									*foundMatch=hadMatch;
									firstMatchPosition=tempFirstMatchPosition;
								}
							}
							else
							{
								fail=true;
							}
							startPosition+=selectionLength;
						}
					}
				}
				else
				{
					if(!limitScope)
					{
						SelectionEndPositionsGet(oldSelectionUniverse,&startPosition,&endPosition);		// find ends of old selection, or cursor position
						highPosition=endPosition;
						if(ReplaceAllBackward(buffer,true,highPosition==totalBytes,0,&highPosition,searchRecord,resultSelectionUniverse,foundMatch,&firstMatchPosition))
						{
							if(wrapAround)
							{
								endPosition=highPosition;
								highPosition=totalBytes;
								fail=!ReplaceAllBackward(buffer,endPosition==0,true,endPosition,&highPosition,searchRecord,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition);
								if(!(*foundMatch))
								{
									*foundMatch=hadMatch;
									firstMatchPosition=tempFirstMatchPosition;
								}
							}
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						SelectionEndPositionsGet(selectionUniverse,&startPosition,&endPosition);
						startPosition=endPosition;	// start at the end

						while(!fail&&startPosition&&SelectionAtOrBeforePositionGet(selectionUniverse,startPosition-1,&startPosition,&selectionLength))
						{
							endPosition=startPosition+selectionLength;
							highPosition=endPosition;
							if(ReplaceAllBackward(buffer,true,true,startPosition,&highPosition,searchRecord,resultSelectionUniverse,&hadMatch,&tempFirstMatchPosition))
							{
								if(!(*foundMatch))
								{
									*foundMatch=hadMatch;
									firstMatchPosition=tempFirstMatchPosition;
								}
								else
								{
									firstMatchPosition+=highPosition-endPosition;		// adjust for any replacements just made
								}
							}
							else
							{
								fail=true;
							}
						}
					}
				}
				EditorReplaceEnd(buffer);
				UndoGroupEndStrict(buffer);
				if(*foundMatch)																	// got something, so change selection universe
				{
					SelectionCursorPositionSet(resultSelectionUniverse,firstMatchPosition);		// put the cursor at the start
				}
			}
			else
			{
				fail=true;
			}
			SelectionUniverseClose(oldSelectionUniverse);									// close this universe
		}
		else
		{
			fail=true;
		}
		SearchRecordDispose(searchRecord);
		PlatformBusyUnShow();
	}
	else
	{
		fail=true;
	}
	return(!fail);
}
