// Selection list management functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

typedef struct selectionElement
{
	SPARSE_ARRAY_HEADER
		saHeader;					// only need the header
} SELECTION_ELEMENT;

typedef struct selectionUniverse
{
	EDITOR_MARK
		*mark;						// points to the mark which uses this selection (if any)
	uint64_t
		cursorPosition;				// if there is no selection (selectionArray is empty), this is the cursor's position
	SPARSE_ARRAY
		selectionArray;
} SELECTION_UNIVERSE;

void SelectionDump(SELECTION_UNIVERSE *universe)
// Debug function to dump out all the ranges in a selection.
{
	SELECTION_ELEMENT
		*element;
	uint64_t
		currentPosition;
	uint64_t
		numElements;

	currentPosition=0;
	while(SparseArrayRecordForPositionGet(&universe->selectionArray,currentPosition,&currentPosition,&numElements,(void **)&element))
	{
		if(element)
		{
			fprintf(stderr,"%" PRIu64 ":%" PRIu64 "\n",currentPosition,numElements);
		}
		currentPosition+=numElements;
	}
}

void SelectionDeleteAll(SELECTION_UNIVERSE *universe)
// Remove all selections from universe.
{
	SparseArrayRangeDelete(&universe->selectionArray,0,SparseArrayTotalElementsGet(&universe->selectionArray));
}

void SelectionRangeDelete(SELECTION_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Remove selection array elements from startPosition for numElements.
// NOTE: this does not mind being called for positions outside of the current array.
// This pulls the remaining array elements to the left.
{
	SparseArrayRangeDelete(&universe->selectionArray,startPosition,numElements);
}

bool SelectionRangeInsert(SELECTION_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Insert numElements into the selection array at startPosition.
// The elements inserted will be selected if the insertion was inside a current selection.
// This pushes the existing array elements after startPosition to the right.
{
	return(SparseArrayRangeInsert(&universe->selectionArray,startPosition,numElements));
}

bool SelectionRangeClear(SELECTION_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Clear selection at startPosition for numElements of universe.
// If there is a problem, return false.
{
	return(SparseArrayRangeClear(&universe->selectionArray,startPosition,numElements));
}

bool SelectionRangeSet(SELECTION_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Set selection at startPosition for numElements of universe.
// If there is a problem, return false.
// The selection will overwrite anything currently in the array at
// the given positions (no insertion (right shift) is performed).
{
	SELECTION_ELEMENT
		element;

	return(SparseArrayRangeSet(&universe->selectionArray,startPosition,numElements,&element));
}

bool SelectionAtOrAfterPositionGet(SELECTION_UNIVERSE *universe,uint64_t position,uint64_t *startPosition,uint64_t *numElements)
// Return the selection which occupies position, or the first one after
// the position if the position is not occupied.
// If the passed position is past the end of the array, return false.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other.
{
	SELECTION_ELEMENT
		*element;

	if(SparseArrayRecordAtOrAfterPositionGet(&universe->selectionArray,position,startPosition,numElements,(void **)&element))
	{
		return(true);
	}
	return(false);
}

bool SelectionAtOrBeforePositionGet(SELECTION_UNIVERSE *universe,uint64_t position,uint64_t *startPosition,uint64_t *numElements)
// Return the selection which occupies position, or the first one before
// the position if the position is not occupied.
// If the passed position is before the first element of the array, return false.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other.
{
	SELECTION_ELEMENT
		*element;

	if(SparseArrayRecordAtOrBeforePositionGet(&universe->selectionArray,position,startPosition,numElements,(void **)&element))
	{
		return(true);
	}
	return(false);
}

bool SelectionRangeGet(SELECTION_UNIVERSE *universe,uint64_t position,uint64_t *startPosition,uint64_t *numElements,bool *isActive)
// Return the range of positions in the array which include position
// and are selected (or not selected) the same as position.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other.
// NOTE: if position is off the end of the array, isActive will be returned as false,
// startPosition will contain the position where the array ended and
// this will return false (numElements will not be set).
{
	SELECTION_ELEMENT
		*element;

	*isActive=false;
	if(SparseArrayRecordForPositionGet(&universe->selectionArray,position,startPosition,numElements,(void **)&element))
	{
		if(element)
		{
			*isActive=true;
		}
		return(true);
	}
	return(false);
}

bool SelectionAtPositionGet(SELECTION_UNIVERSE *universe,uint64_t position)
// Call this to return the selection at the given position of
// the given selection universe.
// NOTE: if position is off the end of the array, this will return false.
// NOTE: this uses a cache so it is efficient to call
// repeatedly for positions which are nearby each other.
{
	uint64_t
		startPosition,
		numElements;
	bool
		selection;

	SelectionRangeGet(universe,position,&startPosition,&numElements,&selection);
	return(selection);
}

uint64_t SelectionSegmentsGet(SELECTION_UNIVERSE *universe)
// Return the total number of segments of the passed selection.
{
	return(SparseArraySegmentsGet(&universe->selectionArray));
}

bool SelectionIsEmpty(SELECTION_UNIVERSE *universe)
// Return TRUE if universe contains no selection elements.
{
	return(SparseArrayIsEmpty(&universe->selectionArray));
}

void SelectionEndPositionsGet(SELECTION_UNIVERSE *universe,uint64_t *startPosition,uint64_t *endPosition)
// Return the start and end absolute text positions of the passed selection universe.
// If there are no selections, then the cursor position is returned as both the start and the end.
{
	if(!SparseArrayElementEndsGet(&universe->selectionArray,startPosition,endPosition))
	{
		(*startPosition)=(*endPosition)=universe->cursorPosition;
	}
}

uint64_t SelectionOccupiedElementsGet(SELECTION_UNIVERSE *universe)
// Return the total number of active sparse elements in the sparse array.
{
	return(SparseArrayTotalOccupiedElementsGet(&universe->selectionArray));
}

uint64_t SelectionCursorPositionGet(SELECTION_UNIVERSE *universe)
// Get the cursor position for the passed selection universe.
{
	return(universe->cursorPosition);
}

void SelectionCursorPositionSet(SELECTION_UNIVERSE *universe,uint64_t position)
// Set the cursor position for the passed selection universe.
{
	universe->cursorPosition=position;
}

void SelectionAdjustForChange(SELECTION_UNIVERSE *universe,uint64_t startOffset,uint64_t oldEndOffset,uint64_t newEndOffset)
// Given a change at startOffset, run through the given selection list, and fix it up.
// NOTE: this can cause all the selections to disappear.
// If it does, then the cursor will be left at the end of the change.
{
	uint64_t
		startPosition,
		endPosition;

	if(SparseArrayElementEndsGet(&universe->selectionArray,&startPosition,&endPosition))	// if there is some selection, then move the cursor to the start so if the selection goes away, the cursor is left in the correct spot
	{
		universe->cursorPosition=startPosition;			// push to the start
	}

	SelectionRangeDelete(universe,startOffset,oldEndOffset-startOffset);
	SelectionRangeInsert(universe,startOffset,newEndOffset-startOffset);

	universe->cursorPosition=EditorOffsetFix(universe->cursorPosition,startOffset,oldEndOffset,newEndOffset);
}

bool SelectionUniverseCopy(SELECTION_UNIVERSE *sourceUniverse,SELECTION_UNIVERSE *destUniverse)
// Copy sourceUniverse to destUniverse.
// If there is a problem, ErrorSet, and return false.
{
	bool
		fail;
	uint64_t
		startPosition,
		numElements;

	fail=false;
	SelectionDeleteAll(destUniverse);
	destUniverse->cursorPosition=sourceUniverse->cursorPosition;	// copy the cursor position

	startPosition=0;
	while(!fail&&SelectionAtOrAfterPositionGet(sourceUniverse,startPosition,&startPosition,&numElements))
	{
		fail=!SelectionRangeSet(destUniverse,startPosition,numElements);
		startPosition+=numElements;
	}
	return(!fail);
}

bool SelectionUniverseOr(SELECTION_UNIVERSE *sourceUniverse,SELECTION_UNIVERSE *destUniverse)
// Logical OR sourceUniverse to destUniverse.
// If there is a problem, ErrorSet, and return false.
// NOTE: adjacent selection boundaries that are not shared
// will be coalesced in the result.
{
	bool
		fail;
	uint64_t
		startPosition,
		numElements;
	uint64_t
		destPosition,
		destElements;

	fail=false;

	startPosition=0;
	while(!fail&&SelectionAtOrAfterPositionGet(sourceUniverse,startPosition,&startPosition,&numElements))
	{
		if(SelectionAtOrAfterPositionGet(destUniverse,startPosition,&destPosition,&destElements))
		{
			if(destPosition<startPosition)					// if dest selection has a section that extends back from the start, coalesce it in
			{
				numElements+=startPosition-destPosition;
				startPosition=destPosition;
			}
			if(SelectionAtOrAfterPositionGet(destUniverse,startPosition+numElements,&destPosition,&destElements))
			{
				if(destPosition<startPosition+numElements)	// if dest selection has a section that extends beyond the end, coalesce it in
				{
					numElements=destPosition+destElements-startPosition;
				}
			}
		}
		fail=!SelectionRangeSet(destUniverse,startPosition,numElements);
		startPosition+=numElements;
	}
	return(!fail);
}

bool SelectionUniverseSourceNotAnd(SELECTION_UNIVERSE *sourceUniverse,SELECTION_UNIVERSE *destUniverse)
// Read sourceUniverse, NOT then AND to destUniverse.
// This operation can also be seen as destUniverse - sourceUniverse.
// If there is a problem, ErrorSet, and return false.
{
	bool
		fail;
	uint64_t
		startPosition,
		numElements;

	fail=false;

	startPosition=0;
	while(!fail&&SelectionAtOrAfterPositionGet(sourceUniverse,startPosition,&startPosition,&numElements))
	{
		fail=!SelectionRangeClear(destUniverse,startPosition,numElements);
		startPosition+=numElements;
	}
	return(!fail);
}

bool SelectionUniverseDestNotAnd(SELECTION_UNIVERSE *sourceUniverse,SELECTION_UNIVERSE *destUniverse)
// Read destUniverse, NOT then AND with sourceUniverse, place the result in destUniverse.
// This operation can also be seen as sourceUniverse - destUniverse.
// If there is a problem, ErrorSet, and return false.
{
	bool
		fail;
	uint64_t
		currentPosition,
		startPositionS,
		numElementsS,
		startPositionD,
		numElementsD,
		numToDo;
	uint64_t
		startPosition,
		endPosition;

	fail=false;
	currentPosition=0;
	while(!fail&&SelectionAtOrAfterPositionGet(sourceUniverse,currentPosition,&startPositionS,&numElementsS))
	{
		if(startPositionS<currentPosition)					// move source start up to current position
		{
			numElementsS-=(currentPosition-startPositionS);
			startPositionS=currentPosition;
		}
		if(SelectionAtOrAfterPositionGet(destUniverse,currentPosition,&startPositionD,&numElementsD))
		{
			if(startPositionS<startPositionD)				// copy S to D until we reach D start
			{
				numToDo=startPositionD-startPositionS;
				if(numToDo>numElementsS)
				{
					numToDo=numElementsS;
				}
				fail=!SelectionRangeSet(destUniverse,startPositionS,numToDo);
				currentPosition=startPositionS+numToDo;
			}
			if(!fail&&startPositionS>=startPositionD)		// once S catches up to D, clear D, ignore S
			{
				numToDo=numElementsD;
				if(numToDo<startPositionS-startPositionD)
				{
					numToDo=startPositionS-startPositionD;
				}
				fail=!SelectionRangeClear(destUniverse,startPositionD,numToDo);
				currentPosition=startPositionD+numToDo;		// move to the end of the cleared section
			}
		}
		else	// just copy S to D now
		{
			fail=!SelectionRangeSet(destUniverse,startPositionS,numElementsS);
			currentPosition=startPositionS+numElementsS;
		}
	}

	if(!fail&&SparseArrayElementEndsGet(&destUniverse->selectionArray,&startPosition,&endPosition))	// deal with anything left in the destUniverse
	{
		if(endPosition>currentPosition)
		{
			fail=!SelectionRangeClear(destUniverse,currentPosition,endPosition-currentPosition);
		}
	}
	return(!fail);
}

bool SelectionUniverseAnd(SELECTION_UNIVERSE *sourceUniverse,SELECTION_UNIVERSE *destUniverse)
// Logical AND sourceUniverse to destUniverse.
// If there is a problem, ErrorSet, and return false.
// NOTE: all selections boundaries occurring in either selection will be
// present in the resulting selection.
{
	bool
		fail;
	uint64_t
		clearPosition,
		startPosition,
		endPosition,
		numElements;
	uint64_t
		destPosition,
		destElements;

	fail=false;

	clearPosition=startPosition=0;
	while(!fail&&SelectionAtOrAfterPositionGet(sourceUniverse,startPosition,&startPosition,&numElements))
	{
		fail=!SelectionRangeClear(destUniverse,clearPosition,startPosition-clearPosition);

		// make sure to break dest at end of start if needed (to make sure boundaries in source end up in dest)
		if(SelectionAtOrAfterPositionGet(destUniverse,startPosition+numElements,&destPosition,&destElements))
		{
			if(destPosition<startPosition+numElements)
			{
				fail=!SelectionRangeSet(destUniverse,destPosition,startPosition+numElements-destPosition);
			}
		}
		startPosition+=numElements;
		clearPosition=startPosition;
	}

	if(!fail&&SparseArrayElementEndsGet(&destUniverse->selectionArray,&startPosition,&endPosition))	// deal with anything left in the destUniverse
	{
		if(endPosition>clearPosition)
		{
			fail=!SelectionRangeClear(destUniverse,clearPosition,endPosition-clearPosition);
		}
	}
	return(!fail);
}

bool SelectionUniverseXor(SELECTION_UNIVERSE *sourceUniverse,SELECTION_UNIVERSE *destUniverse)
// Logical XOR sourceUniverse to destUniverse.
// If there is a problem, ErrorSet, and return false.
// NOTE: all selections boundaries occurring in either selection will be
// present in the resulting selection.
{
	bool
		fail;
	uint64_t
		startPositionS,
		numElementsS,
		startPositionD,
		numElementsD,
		numToDo;

	fail=false;

	startPositionS=startPositionD=0;
	while(!fail&&SelectionAtOrAfterPositionGet(sourceUniverse,startPositionS,&startPositionS,&numElementsS))
	{
		while(!fail&&numElementsS)
		{
			if(SelectionAtOrAfterPositionGet(destUniverse,startPositionS,&startPositionD,&numElementsD))
			{
				if(startPositionS<startPositionD)
				{
					numToDo=startPositionD-startPositionS;
					if(numToDo>numElementsS)
					{
						numToDo=numElementsS;
					}
					fail=!SelectionRangeSet(destUniverse,startPositionS,numToDo);
					numElementsS-=numToDo;
					startPositionS+=numToDo;
				}
				if(!fail&&numElementsS)		// still some to do?
				{
					numToDo=startPositionD+numElementsD-startPositionS;
					if(numToDo>numElementsS)
					{
						numToDo=numElementsS;
					}
					fail=!SelectionRangeClear(destUniverse,startPositionS,numToDo);
					numElementsS-=numToDo;
					startPositionS+=numToDo;
				}
			}
			else	// just copy S to D now
			{
				fail=!SelectionRangeSet(destUniverse,startPositionS,numElementsS);
				startPositionS+=numElementsS;
				numElementsS=0;
			}
		}
	}
	return(!fail);
}

EDITOR_MARK *SelectionUniverseMarkGet(SELECTION_UNIVERSE *universe)
// Return the mark associated with this universe (if one exists).
// If none exists, return NULL.
{
	return(universe->mark);
}

void SelectionUniverseClose(SELECTION_UNIVERSE *universe)
// Dispose of a selection universe.
{
	SparseArrayUnInit(&(universe->selectionArray));
	MDisposePtr(universe);
}

SELECTION_UNIVERSE *SelectionUniverseOpen()
// Open a selection universe (with no elements).
// If there is a problem, set the error, and return NULL.
{
	SELECTION_UNIVERSE
		*universe;

	if((universe=(SELECTION_UNIVERSE *)MNewPtr(sizeof(SELECTION_UNIVERSE))))
	{
		universe->mark=NULL;							// no mark associated with this universe currently
		universe->cursorPosition=0;						// no cursor position yet, so just set it to 0
		if(SparseArrayInit(&(universe->selectionArray),sizeof(SELECTION_ELEMENT)))
		{
			return(universe);
		}
		MDisposePtr(universe);
	}
	return(NULL);
}

// ----------------------------------------------------------------------------
// Mark management

static void MarkVisibilityUnlinkFromMark(EDITOR_MARK *mark,MARK_VISIBILITY *visibility)
// Unlink the visibility record from the mark's list.
{
	if(visibility->prevView)
	{
		visibility->prevView->nextView=visibility->nextView;
	}
	else
	{
		mark->headVisibility=visibility->nextView;
	}
	if(visibility->nextView)
	{
		visibility->nextView->prevView=visibility->prevView;
	}
	else
	{
		mark->tailVisibility=visibility->prevView;
	}
}

static void MarkVisibilityLinkToMark(EDITOR_MARK *mark,MARK_VISIBILITY *visibility)
// Link the visibility record to the list of views that this mark is visible in.
{
	visibility->parentMark=mark;

	visibility->nextView=NULL;
	if((visibility->prevView=mark->tailVisibility))
	{
		mark->tailVisibility->nextView=visibility;
	}
	else
	{
		mark->headVisibility=visibility;
	}
	mark->tailVisibility=visibility;
}

static void MarkVisibilityUnlinkFromView(EDITOR_VIEW *view,MARK_VISIBILITY *visibility)
// Unlink the visibility record from the view's list.
{
	if(visibility->prevMark)
	{
		visibility->prevMark->nextMark=visibility->nextMark;
	}
	else
	{
		view->headVisibility=visibility->nextMark;
	}
	if(visibility->nextMark)
	{
		visibility->nextMark->prevMark=visibility->prevMark;
	}
	else
	{
		view->tailVisibility=visibility->prevMark;
	}
}

static void MarkVisibilityLinkToView(EDITOR_VIEW *view,MARK_VISIBILITY *visibility)
// Link the visibility record to the list of marks visible on the
// passed view.
{
	visibility->parentView=view;

	visibility->nextMark=NULL;
	if((visibility->prevMark=view->tailVisibility))
	{
		view->tailVisibility->nextMark=visibility;
	}
	else
	{
		view->headVisibility=visibility;
	}
	view->tailVisibility=visibility;
}

MARK_VISIBILITY *MarkVisibilityLocateInView(EDITOR_VIEW *view,EDITOR_MARK *mark)
// Hunt the passed view to see if mark is present in its visible
// mark list.
{
	MARK_VISIBILITY
		*visibility;

	visibility=view->headVisibility;
	while(visibility)
	{
		if(visibility->parentMark==mark)
		{
			return(visibility);
		}
		visibility=visibility->nextMark;
	}
	return(NULL);
}

static void MarkVisibleDispose(MARK_VISIBILITY *visibility)
// Unlink and remove mark visibility record.
{
	MarkVisibilityUnlinkFromView(visibility->parentView,visibility);
	MarkVisibilityUnlinkFromMark(visibility->parentMark,visibility);
	MDisposePtr(visibility);
}

void MarksVisibleUnlinkFromView(EDITOR_VIEW *view)
// Get rid of all visible marks attached to the passed view (usually done
// when closing the view).
{
	while(view->headVisibility)
	{
		MarkVisibleDispose(view->headVisibility);		// remove all visible marks
	}
}

void EditorMarkVisibleClear(EDITOR_MARK *mark,EDITOR_VIEW *view)
// Remove the mark from the list of those visible on the passed view.
{
	MARK_VISIBILITY
		*visibility;

	if((visibility=MarkVisibilityLocateInView(view,mark)))
	{
		PlatformViewSelectionAreaChangeStart(view,mark->selectionUniverse);
		MarkVisibleDispose(visibility);
		PlatformViewSelectionAreaChangeEnd(view,NULL);
	}
}

bool EditorMarkVisibleSet(EDITOR_MARK *mark,EDITOR_VIEW *view,EDITOR_COLOR backgroundColor)
// Make the passed mark visible in the passed view. If it is already visible, just
// update the colors.
{
	MARK_VISIBILITY
		*visibility;
	bool
		result;

	result=false;

	visibility=MarkVisibilityLocateInView(view,mark);
	PlatformViewSelectionAreaChangeStart(view,visibility?mark->selectionUniverse:NULL);
	if(!visibility)
	{
		if((visibility=(MARK_VISIBILITY *)MNewPtr(sizeof(MARK_VISIBILITY))))	// allocate a record, and link it to the view and mark
		{
			MarkVisibilityLinkToMark(mark,visibility);
			MarkVisibilityLinkToView(view,visibility);
		}
	}
	if(visibility)	// found or allocated record?
	{
		visibility->activeBackgroundColor=backgroundColor;
		visibility->inactiveBackgroundColor=backgroundColor;
		result=true;
	}
	PlatformViewSelectionAreaChangeEnd(view,visibility?mark->selectionUniverse:NULL);
	return(result);
}

EDITOR_MARK *EditorMarkLocate(EDITOR_BUFFER *buffer,const char *markName)
// Attempt to locate a mark with the given name on buffer.
// If one is found, return it, else, return NULL.
{
	bool
		found;
	EDITOR_MARK
		*currentMark;

	currentMark=buffer->marks;
	found=false;
	while(currentMark&&!found)
	{
		if(currentMark->markName&&(strcmp(currentMark->markName,markName)==0))
		{
			found=true;
		}
		else
		{
			currentMark=currentMark->nextMark;		// locate the mark to delete
		}
	}
	return(currentMark);
}

static void EditorMarkUnlinkFromBuffer(EDITOR_BUFFER *buffer,EDITOR_MARK *mark)
// Unlink a mark from the buffer.
{
	EDITOR_MARK
		*previousMark,
		*currentMark;

	previousMark=NULL;
	currentMark=buffer->marks;
	while(currentMark&&(currentMark!=mark))
	{
		previousMark=currentMark;
		currentMark=currentMark->nextMark;			// locate the mark to delete
	}
	if(currentMark)									// make sure we located it
	{
		if(previousMark)							// see if it was at the top
		{
			previousMark->nextMark=currentMark->nextMark;	// link across
		}
		else
		{
			buffer->marks=currentMark->nextMark;	// set new top of list
		}
	}
}

static void EditorMarkLinkToBuffer(EDITOR_BUFFER *buffer,EDITOR_MARK *mark)
// Link a mark to the buffer.
{
	mark->nextMark=buffer->marks;					// link onto start of list
	buffer->marks=mark;
}

static void EditorMarkDispose(EDITOR_BUFFER *buffer,EDITOR_MARK *mark)
// Remove a mark from the buffer.
{
	EDITOR_VIEW
		*view;

	while(mark->headVisibility)
	{
		view=mark->headVisibility->parentView;
		PlatformViewSelectionAreaChangeStart(view,mark->selectionUniverse);	// tell the view this mark is changing
		MarkVisibleDispose(mark->headVisibility);		// remove visibility of this mark
		PlatformViewSelectionAreaChangeEnd(view,NULL);			// tell the view the mark is gone
	}
	EditorMarkUnlinkFromBuffer(buffer,mark);
	MDisposePtr(mark->markName);
	SelectionUniverseClose(mark->selectionUniverse);	// close the selections
	MDisposePtr(mark);									// get rid of the mark
}

static EDITOR_MARK *EditorMarkCreate(EDITOR_BUFFER *buffer,const char *markName)
// Create a mark with the given name in the buffer.
// If there's a problem, ErrorSet, and return NULL.
{
	EDITOR_MARK
		*mark;

	if((mark=(EDITOR_MARK *)MNewPtr(sizeof(EDITOR_MARK))))
	{
		if((mark->selectionUniverse=SelectionUniverseOpen()))
		{
			mark->selectionUniverse->mark=mark;			// point the selection universe back to the mark
			if((mark->markName=(char *)MNewPtr(strlen(markName)+1)))
			{
				strcpy(mark->markName,markName);		// copy the name
				mark->headVisibility=NULL;				// this mark is not yet visible in any view
				mark->tailVisibility=NULL;
				EditorMarkLinkToBuffer(buffer,mark);	// link the mark to its buffer
				return(mark);
			}
			SelectionUniverseClose(mark->selectionUniverse);
		}
		MDisposePtr(mark);
	}
	return(NULL);
}

void EditorMarkClose(EDITOR_BUFFER *buffer,EDITOR_MARK *mark)
// Close a mark opened by EditorMarkOpen.
{
	EditorMarkDispose(buffer,mark);
}

EDITOR_MARK *EditorMarkOpen(EDITOR_BUFFER *buffer,const char *markName)
// Locate, or open a mark with the given name in the buffer.
// If there's a problem, ErrorSet, and return NULL.
{
	EDITOR_MARK
		*mark;

	if((mark=EditorMarkLocate(buffer,markName)))
	{
		return(mark);
	}
	return(EditorMarkCreate(buffer,markName));
}
