// Shell interface for e93 (currently using John Ousterhout's Tcl)
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define MAX_SCROLL_SPEED	2	// maximum number of lines at a time we'll scroll when tracking scroll repeat events

static Tcl_Trace
	tclTrace;					// the TCL trace object we created

static bool Safe_Tcl_Eval(Tcl_Interp *interp, char *cmd, int *tclResult)
// Some Tcl commands can delete the text of the script which
// is running. If this happens, it is possible for things to get confused.
// For example: a menu command could try to delete itself, which would cause
// the script of the command to be deleted mid-execution.
// To solve this problem, the script is copied into a safe temporary place
// then executed, and finally deleted.
// If there is a problem copying the script, this will return false, and not
// execute it. If this returns true, the result from the Tcl_Eval command will
// be placed in tclResult
{
	char
		*scriptCopy;

	if((scriptCopy=(char *)MNewPtr(strlen(cmd)+1)))
	{
		strcpy(scriptCopy,cmd);								// copy the command
		*tclResult=Tcl_Eval(tclInterpreter,scriptCopy);		// execute it
		MDisposePtr(scriptCopy);							// get rid of the copy
		return(true);
	}
	return(false);
}

void KeyBindEventHandle(KEY_BINDING *binding)
// Perform the script given by the bound key.
{
	int
		tclResult;
	const char
		*stringResult;

	AbortClear();
	if(Safe_Tcl_Eval(tclInterpreter,KeyBindTextGet(binding),&tclResult))
	{
		stringResult=Tcl_GetStringResult(tclInterpreter);
		if((tclResult!=TCL_OK)&&stringResult[0])
		{
			PlatformMessageReport("Bound key execution error:\n%.256s\n",stringResult);
		}
	}
	else
	{
		PlatformMessageReport("Bound key execution error:\nOut of memory\n");
	}
}

void MenuEventHandle(EDITOR_MENU *menu)
// When a menu event arrives, this is called (by the menu manager of the GUI) to handle the menu.
{
	int
		tclResult;
	const char
		*stringResult;

	AbortClear();
	if(Safe_Tcl_Eval(tclInterpreter,PlatformMenuDataTextGet(menu),&tclResult))
	{
		stringResult=Tcl_GetStringResult(tclInterpreter);
		if((tclResult!=TCL_OK)&&stringResult[0])
		{
			PlatformMessageReport("Menu execution error:\n%.256s\n",stringResult);
		}
	}
	else
	{
		PlatformMessageReport("Menu execution error:\nOut of memory\n");
	}
}

void NonStandardControlEventHandle(const char *string)
// When the GUI needs to implement some GUI specific control, it can call this to execute shell
// commands for that control.
{
	int
		tclResult;
	const char
		*stringResult;

	AbortClear();
	tclResult=Tcl_Eval(tclInterpreter,string);
	stringResult=Tcl_GetStringResult(tclInterpreter);
	if((tclResult!=TCL_OK)&&stringResult[0])
	{
		PlatformMessageReport("Execution error:\n%.256s\n",stringResult);
	}
}

static void ViewProcessVirtualKey(EDITOR_VIEW *view,uint16_t eventType,uint32_t modifiers,uint32_t vKeyCode)
// Process virtual keys.
{
	static uint32_t
		scrollTimeout;
	static uint32_t
		scrollSpeed;
	uint8_t
		key;
	bool
		didSomething;
	EDITOR_BUFFER
		*clipboard;

	switch(vKeyCode)
	{
		case VVK_SCROLL_UP:
			if(eventType==VET_VKEY_DOWN)
			{
				scrollSpeed=1;
				scrollTimeout=0;
			}
			EditorVerticalScroll(view,-((int32_t)scrollSpeed));
			if(scrollTimeout++>3)
			{
				if(scrollSpeed<MAX_SCROLL_SPEED)
				{
					scrollSpeed++;
				}
				scrollTimeout=0;
			}
			break;
		case VVK_SCROLL_DOWN:
			if(eventType==VET_VKEY_DOWN)
			{
				scrollSpeed=1;
				scrollTimeout=0;
			}
			EditorVerticalScroll(view,scrollSpeed);
			if(scrollTimeout++>3)
			{
				if(scrollSpeed<MAX_SCROLL_SPEED)
				{
					scrollSpeed++;
				}
				scrollTimeout=0;
			}
			break;
		case VVK_DOCUMENT_PAGE_UP:
			EditorVerticalScrollByPages(view,-1);
			break;
		case VVK_DOCUMENT_PAGE_DOWN:
			EditorVerticalScrollByPages(view,1);
			break;
		case VVK_SCROLL_LEFT:
			EditorHorizontalScroll(view,-HORIZONTAL_SCROLL_THRESHOLD);
			break;
		case VVK_SCROLL_RIGHT:
			EditorHorizontalScroll(view,HORIZONTAL_SCROLL_THRESHOLD);
			break;
		case VVK_DOCUMENT_PAGE_LEFT:
			EditorHorizontalScrollByPages(view,-1);
			break;
		case VVK_DOCUMENT_PAGE_RIGHT:
			EditorHorizontalScrollByPages(view,1);
			break;
		case VVK_RETURN:
			EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
			if(EditorAutoIndent(view->parentBuffer))
			{
				EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
			}
			break;
		case VVK_TAB:
			key='\t';				// insert a tab
			EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
			if(EditorInsert(view->parentBuffer,&key,1))
			{
				EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
			}
			break;
		case VVK_LEFT_ARROW:
			if(!(modifiers&EEM_CTRL))
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
						}
					}
				}
			}
			else
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_WORD))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_WORD))
					{
						EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
					}
				}
			}
			break;
		case VVK_RIGHT_ARROW:
			if(!(modifiers&EEM_CTRL))
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
						}
					}
				}
			}
			else
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_WORD))
					{
						EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_WORD))
					{
						EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
					}
				}
			}
			break;
		case VVK_UP_ARROW:
			if(!(modifiers&EEM_CTRL))
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_LINE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_LINE))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_PAGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_PAGE))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
						}
					}
				}
			}
			else
			{
				EditorVerticalScroll(view,-1);
			}
			break;
		case VVK_DOWN_ARROW:
			if(!(modifiers&EEM_CTRL))
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_LINE))
						{
							EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_LINE))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_SEMI_STRICT,HT_SEMI_STRICT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_PAGE))
						{
							EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_PAGE))
						{
							EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
						}
					}
				}
			}
			else
			{
				EditorVerticalScroll(view,1);
			}
			break;
		case VVK_BACK_SPACE:
			if(!(modifiers&EEM_CTRL)&&!(modifiers&EEM_WINDOW))
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
			}
			else
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_WORD))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_WORD))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_DOCEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_DOCEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
			}
			break;
		case VVK_DELETE:
			if(!(modifiers&EEM_CTRL)&&!(modifiers&EEM_WINDOW))
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_CHAR))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_LINEEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
			}
			else
			{
				if(!(modifiers&EEM_ALT))
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_WORD))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_WORD))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
				else
				{
					if(!(modifiers&EEM_SHIFT))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_mBACKWARD|RPM_DOCEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
					else
					{
						EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
						if(EditorDelete(view,RPM_DOCEDGE))
						{
							EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
						}
					}
				}
			}
			break;
		case VVK_HOME:
			if(!(modifiers&EEM_ALT))
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_LINEEDGE))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorNormalSelectionExpand(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_LINEEDGE))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
					}
				}
			}
			else
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_DOCEDGE))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorNormalSelectionExpand(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_DOCEDGE))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
					}
				}
			}
			break;
		case VVK_END:
			if(!(modifiers&EEM_ALT))
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_LINEEDGE))
					{
						EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorNormalSelectionExpand(view,view->parentBuffer->selectionUniverse,RPM_LINEEDGE))
					{
						EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
					}
				}
			}
			else
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_DOCEDGE))
					{
						EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorNormalSelectionExpand(view,view->parentBuffer->selectionUniverse,RPM_DOCEDGE))
					{
						EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
					}
				}
			}
			break;
		case VVK_PAGE_UP:
			if(modifiers&EEM_CTRL)
			{
				EditorVerticalScrollByPages(view,-1);
			}
			else
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_PAGE))
					{
						EditorViewToSelectionEdgeHome(view,false,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_mBACKWARD|RPM_PAGE))
					{
						EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
					}
				}
			}
			break;
		case VVK_PAGE_DOWN:
			if(modifiers&EEM_CTRL)
			{
				EditorVerticalScrollByPages(view,1);
			}
			else
			{
				if(!(modifiers&EEM_SHIFT))
				{
					if(EditorCursorMove(view,view->parentBuffer->selectionUniverse,RPM_PAGE))
					{
						EditorViewToSelectionEdgeHome(view,true,HT_LENIENT,HT_LENIENT);
					}
				}
				else
				{
					if(EditorSelectionMove(view,view->parentBuffer->selectionUniverse,RPM_PAGE))
					{
						EditorViewToSelectionEdgeHome(view,!view->parentBuffer->currentIsStart,HT_LENIENT,HT_LENIENT);
					}
				}
			}
			break;
		case VVK_UNDO:
			if(EditorUndo(view->parentBuffer,&didSomething))
			{
				if(didSomething)
				{
					EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
				}
				else
				{
					PlatformBeep();
				}
			}
			break;
		case VVK_REDO:
			if(EditorRedo(view->parentBuffer,&didSomething))
			{
				if(didSomething)
				{
					EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
				}
				else
				{
					PlatformBeep();
				}
			}
			break;
		case VVK_UNDO_TOGGLE:
			if(EditorUndoToggle(view->parentBuffer,&didSomething))
			{
				if(didSomething)
				{
					EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
				}
				else
				{
					PlatformBeep();
				}
			}
			break;
		case VVK_CUT:
			clipboard=ClipboardCurrentGet();
			if(ClipboardCut(view->parentBuffer,clipboard))
			{
				EditorUndoClear(clipboard);
				EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
			}
			break;
		case VVK_COPY:
			clipboard=ClipboardCurrentGet();
			if(ClipboardCopy(view->parentBuffer,clipboard))
			{
				EditorUndoClear(clipboard);
			}
			break;
		case VVK_PASTE:
			EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
			if(ClipboardColumnarPaste(view->parentBuffer,view,ClipboardCurrentGet()))
			{
				EditorViewToSelectionEdgeHome(view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);
			}
			break;
		case VVK_SELECT_ALL:
			EditorSelectAll(view->parentBuffer,view->parentBuffer->selectionUniverse);
			break;
	}
}

void ViewEventHandle(VIEW_EVENT *event)
// High level view events are handled here.
{
	int64_t
		leftPixel;
	int64_t
		topLine;
	uint16_t
		trackMode;										// tells editor how to place cursor when cursor position keys arrive
	EDITOR_KEY
		*editorKey;
	EDITOR_VKEY
		*editorVKey;
	VIEW_POS_EVENT_DATA
		*posEventData;
	VIEW_CLICK_EVENT_DATA
		*clickEventData;

	editorKey=NULL;										// keep compiler quiet
	switch(event->eventType)
	{
		case VET_KEY_DOWN:								// key arriving
		case VET_KEY_REPEAT:
			editorKey=(EDITOR_KEY *)event->eventData;
			if(editorKey->vKeyCode!=VVK_NOP)			// see if the key mapped to a virtual function
			{
				ViewProcessVirtualKey(event->view,(event->eventType==VET_KEY_DOWN)?VET_VKEY_DOWN:VET_VKEY_REPEAT,editorKey->modifiers,editorKey->vKeyCode);	// process the virtual key
			}
			else if(editorKey->numTextChars)			// see if the key mapped to some text
			{
				EditorViewToSelectionEdgeHome(event->view,false,HT_SEMI_STRICT,HT_SEMI_STRICT);		// if position was off the view, put it back on
				if(EditorInsert(event->view->parentBuffer,&editorKey->text[0],editorKey->numTextChars))		// drop in the text the key mapped to
				{
					EditorViewToSelectionEdgeHome(event->view,false,HT_LENIENT,HT_LENIENT);			// now scroll into position
				}
			}
			break;
		case VET_VKEY_DOWN:								// virtual key arriving
		case VET_VKEY_REPEAT:
			editorVKey=(EDITOR_VKEY *)event->eventData;
			ViewProcessVirtualKey(event->view,event->eventType,editorVKey->modifiers,editorVKey->vKeyCode);
			break;
		case VET_POSITION_VERTICAL:
			posEventData=(VIEW_POS_EVENT_DATA *)event->eventData;
			ViewOriginGet(event->view,&leftPixel,&topLine);
			ViewOriginSet(event->view,leftPixel,posEventData->position);
			break;
		case VET_POSITION_HORIZONTAL:
			posEventData=(VIEW_POS_EVENT_DATA *)event->eventData;
			ViewOriginGet(event->view,&leftPixel,&topLine);
			ViewOriginSet(event->view,posEventData->position,topLine);
			break;
		case VET_CLICK:
		case VET_CLICK_HOLD:
			clickEventData=(VIEW_CLICK_EVENT_DATA *)event->eventData;
			trackMode=0;								// default mode
			if(event->eventType==VET_CLICK_HOLD)		// see if still tracking
			{
				trackMode|=TM_mREPEAT;
			}
			if(clickEventData->modifiers&EEM_CTRL)		// if control is held, it means we're adding or subtracting
			{
				if(clickEventData->modifiers&EEM_SHIFT)	// subtracting?
				{
					trackMode|=TM_SUBTRACT;
				}
				else
				{
					trackMode|=TM_ADD;
				}
			}
			else if(clickEventData->modifiers&EEM_SHIFT)	// continuation of selection?
			{
				trackMode|=TM_mCONTINUE;
			}
			if((clickEventData->modifiers&EEM_ALT)||(clickEventData->button==2))
			{
				trackMode|=TM_COLUMNAR;
			}
			else if((clickEventData->modifiers&EEM_WINDOW)||(clickEventData->button==1))
			{
				trackMode|=TM_HAND;
			}
			switch(clickEventData->numClicks)
			{
				case 0:
					trackMode|=TM_CHAR;
					break;
				case 1:
					trackMode|=TM_WORD;
					break;
				case 2:
					trackMode|=TM_LINE;
					break;
				default:
					trackMode|=TM_ALL;
					break;
			}
			EditorViewPointerTrack(event->view,clickEventData->xClick,clickEventData->yClick,trackMode);
			break;
		case VET_CLICK_RELEASE:
			EditorViewPointerTrackComplete(event->view);
			break;
	}
}

bool ShellCommandHandle(const char *command,int argc,char *argv[])
// Sometimes things in the gui request that the shell should perform
// certain operations. This is the interface to allow that to happen.
// If the operation has some sort of failure (defined by the operation)
// this will report it (if there is a message) and return false.
{
	char
		*list;
	int
		tclResult;
	const char
		*stringResult;

	list=Tcl_Merge(argc,argv);
	AbortClear();
	tclResult=Tcl_VarEval(tclInterpreter,"ShellCommand ",command," ",list,(char *)NULL);
	Tcl_Free(list);
	stringResult=Tcl_GetStringResult(tclInterpreter);
	if(tclResult!=TCL_OK)
	{
		if(stringResult[0])
		{
			PlatformMessageReport("Shell command execution error:\n%.256s\n",stringResult);
		}
		return(false);
	}
	return(true);
}

static bool StartupScriptExecute(Tcl_Interp *interpreter)
// Locate the startup script file, open, and execute it.
// If there is any problem, ErrorSet, and return false.
{
	bool
		fail;
	char
		scriptPath[MAX_PATH_NAME_LENGTH];

	fail=false;
	if(PlatformStartupScriptLocate(scriptPath))
	{
		AbortClear();
		Tcl_SetVar(interpreter,"SCRIPTPATH",scriptPath,TCL_LEAVE_ERR_MSG);
		if(Tcl_EvalFile(interpreter,scriptPath)!=TCL_OK)
		{
			ErrorTraceClear();
			ErrorSet("%s:%d:%s",scriptPath,Tcl_GetErrorLine(interpreter),Tcl_GetStringResult(interpreter));
			fail=true;
		}
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static int TraceCheckAbortProc(ClientData clientData,Tcl_Interp *interpreter,int level,const char *command,Tcl_Command commandToken,int objc,Tcl_Obj *const objv[])
// This is a small trick on Tcl. We tell it we want to trace, but really, we want
// to check to see if the user is trying to abort the execution of a script.
// So, every time we are called, we check to see if the user is aborting.
{
	if(AbortCheck())
	{
		Tcl_AppendResult(interpreter,"Script Aborted",NULL);
		return(TCL_ERROR);
	}
	return(TCL_OK);
}

static void TclAbortHandlingUnSetUp(Tcl_Interp *interpreter)
// Undo what TclAbortHandlingSetUp did.
{
	Tcl_DeleteTrace(interpreter,tclTrace);
}

static bool TclAbortHandlingSetUp(Tcl_Interp *interpreter)
// Set up Tcl so that it manages to call AbortCheck, and if AbortCheck
// returns true, then we force an error in the Tcl control Flow.
{
	if((tclTrace=Tcl_CreateObjTrace(interpreter,0,TCL_ALLOW_INLINE_COMPILATION,TraceCheckAbortProc,(ClientData)NULL,(Tcl_CmdObjTraceDeleteProc *)NULL)))
	{
		return(true);
	}
	return(false);
}

void ShellBackgroundUpdate()
// Handle Tcl's event loop.
{
	while(Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT))
		;							// tell Tcl to check for any events it may care about
}

static void TclUnInit()
// Undo what TclInit did.
{
	TclAbortHandlingUnSetUp(tclInterpreter);
	ChannelsUnInit(tclInterpreter);
	Tcl_DeleteInterp(tclInterpreter);	// have TCL clean up
}

static bool TclInit(char *pathName)
// Initialize Tcl.
{
	if((tclInterpreter=Tcl_CreateInterp()))
	{
		if(ChannelsInit(tclInterpreter))
		{
			if(Tcl_Init(tclInterpreter)==TCL_OK)
			{
			    Tcl_FindExecutable(pathName);

				if(TclAbortHandlingSetUp(tclInterpreter))
				{
					return(true);
				}
			}
			else
			{
				ErrorSet("Failed to Tcl_Init(): %s",Tcl_GetStringResult(tclInterpreter));
			}
			ChannelsUnInit(tclInterpreter);
		}
		else
		{
			ErrorSet("Failed to redirect Tcl's stdout and stderr");
		}
		Tcl_DeleteInterp(tclInterpreter);						// have TCL clean up
	}
	else
	{
		ErrorSet("Failed to create TCL interpreter");
	}
	return(false);
}

void ShellLoop(int argc,char *argv[])
// This is where the editor shell gets and handles events.
// It actually defers to the gui level to call us back with events.
{
	ErrorTraceClear();
	if(TclInit(argv[0]))
	{
		ErrorTraceClear();
		if(EditorShellCommandsCreate(tclInterpreter))
		{
			ErrorTraceClear();
			if(PlatformShellCommandsAdd(tclInterpreter))
			{
				ErrorTraceClear();
				if(StartupScriptExecute(tclInterpreter))			// deal with the start-up script, leave if there is a problem
				{
					PlatformEventLoop(argc,argv);					// pass initial parameters
				}
				else
				{
					fprintf(stderr,"Failed to execute startup script: %s\n",ErrorTraceGet());
				}
			}
			else
			{
				fprintf(stderr,"Failed to add supplemental shell commands: %s\n",ErrorTraceGet());
			}
		}
		else
		{
			fprintf(stderr,"Failed to create editor shell commands: %s\n",ErrorTraceGet());
		}
		TclUnInit();
	}
	else
	{
		fprintf(stderr,"Failed to initialize Tcl: %s\n",ErrorTraceGet());
	}
}
