// Sparse array management functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// Sparse arrays are arrays of arbitrary items of which most
// elements are usually unoccupied.
// These are handy for describing selections, styles, and other
// things within e93.

// To create a sparse array, define a structure of which the first
// element is a SPARSE_ARRAY_HEADER.

#include	"includes.h"

#define	SPARSE_ARRAY_CHUNK_ELEMENTS			256			// number of elements in a sparse array chunk
#define	SPARSE_ARRAY_CACHE_SIZE				2			// number of CARRAY cache entries used for sparse array

#define	CA_SPARSE_ELEMENTS					CA_NUM_INHERENT_PARAMS		// user parameter that tracks the total number sparse elements in the CARRAY (offset+length)
#define	CA_LENGTH_SPARSE_ELEMENTS			(CA_SPARSE_ELEMENTS+1)		// user parameter that tracks the total number of length sparse elements in the CARRAY
#define	CA_NUM_PARAMS						(CA_LENGTH_SPARSE_ELEMENTS+1)

#define	HEADER_OFFSET(sparseArray,chunkVal,offsetVal) ((SPARSE_ARRAY_HEADER *)(&((uint8_t *)chunkVal->data)[(offsetVal)*sparseArray->elementCArray->elementSize]))->offset
#define	HEADER_LENGTH(sparseArray,chunkVal,offsetVal) ((SPARSE_ARRAY_HEADER *)(&((uint8_t *)chunkVal->data)[(offsetVal)*sparseArray->elementCArray->elementSize]))->length

// ----------------------------------------------------------------------------

void SparseArraySanity(const SPARSE_ARRAY *sparseArray)
// Check the sanity of sparseArray.
{
	CARRAY_CHUNK
		*chunk;
	uint32_t
		i;
	uint64_t
		localSparseElements,
		localLengthSparseElements,
		totalSparseElements,
		totalLengthSparseElements;
	uint64_t
		cArrayIndex;
	uint8_t
		*pointer;
	SPARSE_ARRAY_HEADER
		*header;

	CArraySanity(sparseArray->elementCArray);

	if((sparseArray->cacheEntry.cArrayIndex!=0)&&sparseArray->cacheEntry.cArrayIndex>=CArrayTotalElementsGet(sparseArray->elementCArray))
	{
		fprintf(stderr,"SparseArraySanity: cache points off end\n");
		abort();
	}

	cArrayIndex=0;
	totalSparseElements=0;
	totalLengthSparseElements=0;
	chunk=sparseArray->elementCArray->firstChunk;
	while(chunk)
	{
		localSparseElements=0;
		localLengthSparseElements=0;
		for(i=0;i<chunk->params[CA_ELEMENTS];i++)
		{
			if((sparseArray->cacheEntry.cArrayIndex!=0)&&sparseArray->cacheEntry.cArrayIndex==cArrayIndex)	// check that the cache is valid
			{
				if(sparseArray->cacheEntry.sparseIndex!=totalSparseElements)
				{
					fprintf(stderr,"SparseArraySanity: cache is invalid\n");
					abort();
				}
			}
			pointer=&((uint8_t *)chunk->data)[i*sparseArray->elementCArray->elementSize];
			header=(SPARSE_ARRAY_HEADER *)pointer;
			localSparseElements+=header->offset+header->length;
			localLengthSparseElements+=header->length;
			totalSparseElements+=header->offset+header->length;
			totalLengthSparseElements+=header->length;
			cArrayIndex++;
		}
		if(localSparseElements!=chunk->params[CA_SPARSE_ELEMENTS])
		{
			fprintf(stderr,"SparseArraySanity: Mismatch of localSparseElements, %" PRIu64 "!=%" PRIu64 "\n",localSparseElements,chunk->params[CA_SPARSE_ELEMENTS]);
			abort();
		}
		if(localLengthSparseElements!=chunk->params[CA_LENGTH_SPARSE_ELEMENTS])
		{
			fprintf(stderr,"SparseArraySanity: Mismatch of localLengthSparseElements, %" PRIu64 "!=%" PRIu64 "\n",localLengthSparseElements,chunk->params[CA_LENGTH_SPARSE_ELEMENTS]);
			abort();
		}
		chunk=chunk->nextChunk;
	}
	if(totalSparseElements!=sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS])
	{
		fprintf(stderr,"SparseArraySanity: Mismatch of totalSparseElements %" PRIu64 "!=%" PRIu64 "\n",totalSparseElements,sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]);
		abort();
	}
	if(totalLengthSparseElements!=sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS])
	{
		fprintf(stderr,"SparseArraySanity: Mismatch of totalLengthSparseElements %" PRIu64 "!=%" PRIu64 "\n",totalLengthSparseElements,sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]);
		abort();
	}
}

// ----------------------------------------------------------------------------

static void SparseArrayElementsMove(const CARRAY *cArray,const void *source,uint64_t sourceElementIndex,void *dest,uint64_t destElementIndex,uint64_t numElements,uint64_t *params)
// Move function needed by the CARRAY code. This knows how to move
// elements around, and update the needed params.
// In the case of sparse arrays, we manage a single user-parameter
// which tells how many "sparse" elements are used. This is
// the sum of the offset and length fields of the sparse array
// header.
{
	uint64_t
		totalSparseElements,
		totalLengthSparseElements;
	SPARSE_ARRAY_HEADER
		*sourceHeader;
	uint8_t
		*srcPointer,
		*dstPointer;

	if(source)
	{
		if(params)
		{
			params[CA_ELEMENTS]=numElements;
			totalSparseElements=0;
			totalLengthSparseElements=0;
			srcPointer=&((uint8_t *)source)[sourceElementIndex*cArray->elementSize];
			if(dest)							// moving data somewhere?
			{
				dstPointer=&((uint8_t *)dest)[destElementIndex*cArray->elementSize];

				// Check move direction to handle overlaps correctly
				if(dstPointer<srcPointer)		// move forward through the elements
				{
					while(numElements)
					{
						sourceHeader=(SPARSE_ARRAY_HEADER *)srcPointer;
						totalSparseElements+=sourceHeader->offset+sourceHeader->length;
						totalLengthSparseElements+=sourceHeader->length;
						MMoveMem(srcPointer,dstPointer,cArray->elementSize);
						srcPointer+=cArray->elementSize;
						dstPointer+=cArray->elementSize;
						numElements--;
					}
				}
				else							// move backward
				{
					srcPointer+=numElements*cArray->elementSize;
					dstPointer+=numElements*cArray->elementSize;
					while(numElements)
					{
						srcPointer-=cArray->elementSize;
						dstPointer-=cArray->elementSize;
						sourceHeader=(SPARSE_ARRAY_HEADER *)srcPointer;
						totalSparseElements+=sourceHeader->offset+sourceHeader->length;
						totalLengthSparseElements+=sourceHeader->length;
						MMoveMem(srcPointer,dstPointer,cArray->elementSize);
						numElements--;
					}
				}
			}
			else	// just count up elements of the source
			{
				while(numElements)
				{
					sourceHeader=(SPARSE_ARRAY_HEADER *)srcPointer;
					totalSparseElements+=sourceHeader->offset+sourceHeader->length;
					totalLengthSparseElements+=sourceHeader->length;
					srcPointer+=cArray->elementSize;
					numElements--;
				}
			}
			params[CA_SPARSE_ELEMENTS]=totalSparseElements;
			params[CA_LENGTH_SPARSE_ELEMENTS]=totalLengthSparseElements;
		}
		else							// don't have to track the parameters, just do the move
		{
			if(dest)					// only do work if dest is specified
			{
				MMoveMem(&((uint8_t *)source)[sourceElementIndex*cArray->elementSize],&((uint8_t *)dest)[destElementIndex*cArray->elementSize],numElements*cArray->elementSize);
			}
		}
	}
	else
	{
		if(params)
		{
			params[CA_ELEMENTS]=numElements;
			params[CA_SPARSE_ELEMENTS]=0;
			params[CA_LENGTH_SPARSE_ELEMENTS]=0;
		}
		if(dest)					// only do work if dest is specified
		{
			memset(&((uint8_t *)dest)[destElementIndex*cArray->elementSize],0,numElements*cArray->elementSize);
		}
	}
}

// ----------------------------------------------------------------------------

static void SparseArrayRecordLocate(SPARSE_ARRAY *sparseArray,uint64_t position,CARRAY_CHUNK **outChunk,uint64_t *outOffset,uint64_t *elementIndex,uint64_t *startPosition)
// Locate the record in sparseArray which contains the sparse position.
// Return the chunk of the CARRAY which contains the record in question,
// and the offset to the record.
// Return elementIndex as the CARRAY element that chunk/offset point to.
// Also return the actual starting position of the record (the sparse position
// where the record's "offset" begins).
// NOTE: if the passed position is after the last entry in the array, return
// outChunk as NULL, outOffset at 0, elementIndex as the total number of elements
// in the CARRAY, and startPosition as the total number of sparse elements in the array.
{
	uint64_t
		params[sparseArray->elementCArray->numParams];
	uint64_t
		recordLength;

	if(position<CArrayParamTotalGet(sparseArray->elementCArray,CA_SPARSE_ELEMENTS))
	{
		// hunt down the chunk containing the position
		CArrayLocate(sparseArray->elementCArray,CA_SPARSE_ELEMENTS,position,outChunk,params);		// find the chunk that contains the position (should be really quick for repeated calls of nearby entries)
		if((sparseArray->cacheEntry.cArrayIndex)&&(sparseArray->cacheEntry.cArrayIndex>=params[CA_ELEMENTS])&&(sparseArray->cacheEntry.cArrayIndex<params[CA_ELEMENTS]+(*outChunk)->params[CA_ELEMENTS]))	// if the cache points within this chunk, use it
		{
			(*outOffset)=sparseArray->cacheEntry.cArrayIndex-params[CA_ELEMENTS];					// get the offset where we start searching
			(*elementIndex)=sparseArray->cacheEntry.cArrayIndex;									// start at this absolute cArray index
			(*startPosition)=sparseArray->cacheEntry.sparseIndex;									// start at this absolute sparse index
		}
		else	// cache pointed outside the chunk we found, so just start hunting for position from the start
		{
			(*outOffset)=0;																			// start hunting at the start
			(*elementIndex)=params[CA_ELEMENTS];
			(*startPosition)=params[CA_SPARSE_ELEMENTS];
		}

		// now hunt down the actual sparse record within the chunk
		if(*outChunk)													// make sure there was a chunk
		{
			if(position<(*startPosition))								// need to hunt backward?
			{
				while(position<(*startPosition))
				{
					(*outOffset)--;										// step back one record in the chunk
					(*elementIndex)--;									// step back globally
					recordLength=HEADER_OFFSET(sparseArray,(*outChunk),(*outOffset))+HEADER_LENGTH(sparseArray,(*outChunk),(*outOffset));
					(*startPosition)-=recordLength;						// get updated start position
				}
			}
			else
			{
				while(position>=(*startPosition)+(recordLength=HEADER_OFFSET(sparseArray,(*outChunk),(*outOffset))+HEADER_LENGTH(sparseArray,(*outChunk),(*outOffset))))
				{
					(*startPosition)+=recordLength;
					(*elementIndex)++;
					(*outOffset)++;
				}
			}

			// update the cache
			sparseArray->cacheEntry.cArrayIndex=(*elementIndex);
			sparseArray->cacheEntry.sparseIndex=(*startPosition);
		}
	}
	else	// position is past the end
	{
		(*outChunk)=NULL;
		(*outOffset)=0;
		(*elementIndex)=CArrayParamTotalGet(sparseArray->elementCArray,CA_ELEMENTS);
		(*startPosition)=CArrayParamTotalGet(sparseArray->elementCArray,CA_SPARSE_ELEMENTS);
	}
}

static uint64_t RangeIntersect(uint64_t startA,uint64_t numElementsA,uint64_t startB,uint64_t numElementsB)
// Return the count of elements that intersect between the ranges startA/numElementsA and startB/numElementsB.
// This is useful for the delete function below.
{
	uint64_t
		result;

	result=0;			// by default, no intersection
	if(startA>=startB)
	{
		if(startA<startB+numElementsB)
		{
			result=(startB+numElementsB)-startA;
			if(result>numElementsA)
			{
				result=numElementsA;
			}
		}
	}
	else
	{
		if(startB<startA+numElementsA)
		{
			result=(startA+numElementsA)-startB;
			if(result>numElementsB)
			{
				result=numElementsB;
			}
		}
	}
	return(result);
}

void SparseArrayRangeDelete(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements)
// Remove sparse array elements from startPosition for numElements.
// NOTE: this does not mind being called for positions outside of the current array.
// This function was difficult to write.
{
	uint64_t
		totalElements;
	CARRAY_CHUNK
		*startChunk,
		*endChunk;
	uint64_t
		startOffset,
		endOffset;
	uint64_t
		startIndex,
		endIndex,
		deleteIndex;
	uint64_t
		startActualSparse,
		endActualSparse;
	uint64_t
		sparseOffset,
		sparseLength;
	uint64_t
		numToRemoveOffset,
		numToRemoveLength;
	uint64_t
		startOffsetResidue;
	int64_t
		paramChanges[sparseArray->elementCArray->numParams];
	uint32_t
		i;

	totalElements=sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS];
	if(numElements&&(startPosition<totalElements))					// make sure there's work to be done
	{
		if(numElements>(totalElements-startPosition))
		{
			numElements=totalElements-startPosition;				// truncate down the number of elements to the number we have
		}

		for(i=0;i<sparseArray->elementCArray->numParams;i++)		// clear out all param changes
		{
			paramChanges[i]=0;
		}

		SparseArrayRecordLocate(sparseArray,startPosition,&startChunk,&startOffset,&startIndex,&startActualSparse);		// find the record where the delete begins

		sparseOffset=HEADER_OFFSET(sparseArray,startChunk,startOffset);		// get the offset and the length from the record
		sparseLength=HEADER_LENGTH(sparseArray,startChunk,startOffset);

		if(startPosition+numElements<=startActualSparse+sparseOffset+sparseLength)	// see if delete is happening in a single record -- if so, the end record is just the next one over, save some search time and find it directly
		{

			endOffset=startOffset+1;
			endChunk=startChunk;
			if(endOffset>=CArrayChunkElementsGet(endChunk))
			{
				endChunk=endChunk->nextChunk;
				endOffset=0;
			}
			endIndex=startIndex+1;
			endActualSparse=startActualSparse+sparseOffset+sparseLength;
		}
		else
		{
			SparseArrayRecordLocate(sparseArray,startPosition+numElements,&endChunk,&endOffset,&endIndex,&endActualSparse);		// find the record where the delete ends
		}

		// work out how much to remove from the offset of the starting record
		numToRemoveOffset=RangeIntersect(startActualSparse,sparseOffset,startPosition,numElements);
		// work out how much to remove from the length of the starting record
		numToRemoveLength=RangeIntersect(startActualSparse+sparseOffset,sparseLength,startPosition,numElements);

		if(numToRemoveLength<sparseLength)							// some length remaining in first record?
		{
			sparseOffset-=numToRemoveOffset;						// pull from the offset
			sparseLength-=numToRemoveLength;						// pull from the length

			HEADER_OFFSET(sparseArray,startChunk,startOffset)=sparseOffset;	// put back adjusted values
			HEADER_LENGTH(sparseArray,startChunk,startOffset)=sparseLength;
			paramChanges[CA_SPARSE_ELEMENTS]=-(numToRemoveOffset+numToRemoveLength);	// remove this much when updating the cache
			paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=-numToRemoveLength;

			startChunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the number of sparse elements in this chunk
			startChunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
			sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the total number of sparse elements in this array
			sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
			CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);	// adjust the cache for the array
			deleteIndex=startIndex+1;								// begin deleting *after* this record
			startOffsetResidue=0;									// no startOffset residue to carry into the final record
		}
		else
		{
			startOffsetResidue=sparseOffset-numToRemoveOffset;		// this much offset is going to be deleted from the current record, and must be put back into the last record (unless there is no last record)
			deleteIndex=startIndex;									// begin deleting at this record
		}

		if(endChunk)												// make sure the end wasn't off the end of the array
		{
			sparseOffset=HEADER_OFFSET(sparseArray,endChunk,endOffset);	// get the offset and the length from the end record
			sparseLength=HEADER_LENGTH(sparseArray,endChunk,endOffset);

			// work out how much to remove from the offset of the ending record
			numToRemoveOffset=RangeIntersect(endActualSparse,sparseOffset,startPosition,numElements);
			// work out how much to remove from the length of the ending record
			numToRemoveLength=RangeIntersect(endActualSparse+sparseOffset,sparseLength,startPosition,numElements);

			sparseOffset-=numToRemoveOffset;							// pull from the offset
			sparseOffset+=startOffsetResidue;							// add back residue from first record delete if any
			sparseLength-=numToRemoveLength;							// pull from the length
			HEADER_OFFSET(sparseArray,endChunk,endOffset)=sparseOffset;	// put back adjusted values
			HEADER_LENGTH(sparseArray,endChunk,endOffset)=sparseLength;
			paramChanges[CA_SPARSE_ELEMENTS]=-(numToRemoveOffset+numToRemoveLength)+startOffsetResidue;	// adjust this much when updating the cache
			paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=-numToRemoveLength;

			endChunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];		// adjust the number of sparse elements in this chunk
			endChunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
			sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the total number of sparse elements in this array
			sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
			CArrayCacheEntriesUpdate(sparseArray->elementCArray,endIndex,paramChanges);	// adjust the cache for the array
		}

		if(endIndex>deleteIndex)									// elements to delete?
		{
			CArrayElementsDelete(sparseArray->elementCArray,deleteIndex,endIndex-deleteIndex,NULL,NULL);	// get rid of the elements in the middle
		}

		if(startIndex<CArrayTotalElementsGet(sparseArray->elementCArray))	// if there's a record to point to after the delete, then point to it
		{
			sparseArray->cacheEntry.cArrayIndex=startIndex;			// remember the start of the delete as the current cache position (delete below won't mess with it)
			sparseArray->cacheEntry.sparseIndex=startActualSparse;
		}
		else
		{
			sparseArray->cacheEntry.cArrayIndex=0;
		}
	}
//	SparseArraySanity(sparseArray);
}

bool SparseArrayRangeInsert(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements)
// Insert numElements into the array at startPosition.
// This will expand the array as needed, pushing out either the offset or length
// of a record (or doing nothing if the insertion is past the end).
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		offset;
	uint64_t
		startIndex,
		actualStart;
	uint64_t
		numLengthElements;
	int64_t
		paramChanges[sparseArray->elementCArray->numParams];
	uint32_t
		i;
	bool
		fail;

	fail=false;
	if(numElements)			// if no elements to insert, then do nothing
	{
		SparseArrayRecordLocate(sparseArray,startPosition,&chunk,&offset,&startIndex,&actualStart);
		// NOTE: SparseArrayRecordLocate updates local cache to beginning of insert which code
		// below can't change.
		if(chunk)			// did something already exist in the array at this point? If not, then do nothing
		{
			if((startPosition-actualStart)>HEADER_OFFSET(sparseArray,chunk,offset))
			{
				HEADER_LENGTH(sparseArray,chunk,offset)+=numElements;
				numLengthElements=numElements;
			}
			else
			{
				HEADER_OFFSET(sparseArray,chunk,offset)+=numElements;
				numLengthElements=0;
			}

			chunk->params[CA_SPARSE_ELEMENTS]+=numElements;								// bump number of sparse elements in the chunk
			sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=numElements;	// bump number of sparse elements in the universe
			chunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=numLengthElements;				// bump number of length sparse elements in the chunk
			sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=numLengthElements;

			for(i=0;i<sparseArray->elementCArray->numParams;i++)						// clear out all param changes
			{
				paramChanges[i]=0;
			}
			paramChanges[CA_SPARSE_ELEMENTS]=numElements;
			paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=numLengthElements;
			CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);	// adjust the cache for the array
		}
	}
//	SparseArraySanity(sparseArray);
	return(!fail);
}

bool SparseArrayRecordForPositionGet(SPARSE_ARRAY *sparseArray,uint64_t position,uint64_t *startPosition,uint64_t *numElements,void **record)
// Return the pointer to the record of sparseArray which occupies position.
// If the passed position is in an empty part of the array, return record as NULL, with
// startPosition and numElements describing the empty part of the array.
// If the passed position is past the end of the array, return false with
// startPosition as the end of the array, numElements as 0, and *record as NULL.
// NOTE: the pointer is only valid until the next call to any sparse array routine.
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		offset;
	uint64_t
		headerOffset;
	uint64_t
		startIndex;

	SparseArrayRecordLocate(sparseArray,position,&chunk,&offset,&startIndex,startPosition);
	if(chunk)
	{
		headerOffset=HEADER_OFFSET(sparseArray,chunk,offset);
		if(position>=(*startPosition+headerOffset))
		{
			*startPosition+=headerOffset;
			*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
			*record=&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize];
		}
		else
		{
			*numElements=headerOffset;
			*record=NULL;
		}
		return(true);
	}
	else	// off the end
	{
		*numElements=0;
		*record=NULL;
	}
	return(false);
}

bool SparseArrayRecordAtOrAfterPositionGet(SPARSE_ARRAY *sparseArray,uint64_t position,uint64_t *startPosition,uint64_t *numElements,void **record)
// Return the pointer to the record of sparseArray which occupies position,
// or the first one after the position if the position is not occupied.
// If the passed position is past the end of the array, return false.
// NOTE: the pointer is only valid until the next call to any sparse array routine.
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		offset;
	uint64_t
		startIndex;

	SparseArrayRecordLocate(sparseArray,position,&chunk,&offset,&startIndex,startPosition);
	if(chunk)
	{
		*startPosition+=HEADER_OFFSET(sparseArray,chunk,offset);
		*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
		*record=&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize];
		return(true);
	}
	return(false);
}

bool SparseArrayRecordAtOrBeforePositionGet(SPARSE_ARRAY *sparseArray,uint64_t position,uint64_t *startPosition,uint64_t *numElements,void **record)
// Return the pointer to the record of sparseArray which occupies position,
// or the first one before the position if the position is not occupied.
// If the passed position is before the first element of the array, return false.
// NOTE: the pointer is only valid until the next call to any sparse array routine.
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		offset;
	uint64_t
		headerOffset;
	uint64_t
		startIndex;

	SparseArrayRecordLocate(sparseArray,position,&chunk,&offset,&startIndex,startPosition);
	if(chunk)
	{
		headerOffset=HEADER_OFFSET(sparseArray,chunk,offset);
		if(position>=*startPosition+headerOffset)
		{
			*startPosition+=headerOffset;
			*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
			*record=&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize];
			return(true);
		}
		else	// get element that precedes this one
		{
			if(offset)							// step back one
			{
				offset--;
			}
			else
			{
				if((chunk=chunk->previousChunk))
				{
					offset=CArrayChunkElementsGet(chunk)-1;
				}
			}
			if(chunk)
			{
				*startPosition-=HEADER_LENGTH(sparseArray,chunk,offset);
				*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
				*record=&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize];
				return(true);
			}
		}
	}
	else
	{
		if((chunk=sparseArray->elementCArray->lastChunk))
		{
			offset=CArrayChunkElementsGet(sparseArray->elementCArray->lastChunk);
			*startPosition=CArrayParamTotalGet(sparseArray->elementCArray,CA_SPARSE_ELEMENTS)-HEADER_LENGTH(sparseArray,chunk,offset);
			*numElements=HEADER_LENGTH(sparseArray,chunk,offset);
			*record=&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize];
			return(true);
		}
	}
	return(false);
}

bool SparseArrayRangeClear(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements)
// Clear out elements of sparseArray from startPosition for numElements.
// If there is a problem, set the error and return false.
// This will overwrite anything currently in the array at
// the given positions, and will cause items to be split
// if it bisects them.
{
	CARRAY_CHUNK
		*chunk,
		*splitChunk;
	uint64_t
		offset,
		splitOffset;
	uint64_t
		actualStart;
	uint64_t
		startIndex;
	uint64_t
		sparseOffset;
	uint64_t
		newLength;
	uint64_t
		outParams[sparseArray->elementCArray->numParams];
	int64_t
		paramChanges[sparseArray->elementCArray->numParams];
	uint32_t
		i;
	bool
		fail;

	fail=false;
	if(numElements)
	{
		for(i=0;i<sparseArray->elementCArray->numParams;i++)									// clear out all param changes
		{
			paramChanges[i]=0;
		}

		SparseArrayRangeDelete(sparseArray,startPosition,numElements);									// get rid of any elements that were in that spot
		SparseArrayRecordLocate(sparseArray,startPosition,&chunk,&offset,&startIndex,&actualStart);		// find the record containing the start
		// NOTE: SparseArrayRecordLocate updates local cache to beginning of clear which code
		// below can't change.
		if(chunk)																				// if there is a record there, see how we need to clear
		{
			sparseOffset=HEADER_OFFSET(sparseArray,chunk,offset);

			if(startPosition>actualStart+sparseOffset)											// need to split the remaining length?
			{
				if(CArrayEmptyInsert(sparseArray->elementCArray,startIndex,1,&chunk,outParams))	// add 1 element for the split we need (before the record we're going to split)
				{
					offset=startIndex-outParams[CA_ELEMENTS];
					splitOffset=offset;
					splitChunk=chunk;
					splitOffset++;
					if(splitOffset>=CArrayChunkElementsGet(splitChunk))							// move to the next element (which is the original we need to split)
					{
						splitChunk=splitChunk->nextChunk;
						splitOffset=0;
					}

					memcpy(&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize],&((uint8_t *)splitChunk->data)[splitOffset*sparseArray->elementCArray->elementSize],sparseArray->elementCArray->elementSize);	// blast in split value

					// Finally, adjust offsets and lengths

					// Do first chunk
					newLength=startPosition-actualStart-sparseOffset;							// calculate the new length for the first item
					HEADER_LENGTH(sparseArray,chunk,offset)=newLength;							// write into the record (record contains previous offset from the memcpy above)
					paramChanges[CA_SPARSE_ELEMENTS]=sparseOffset+newLength;					// record became this much larger due to memcpy and adjustment we just made
					paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=newLength;
					chunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];							// adjust the sparse elements in the chunk
					chunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the elements in the array
					sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);					// adjust the cache for the array

					// Do second chunk
					HEADER_OFFSET(sparseArray,splitChunk,splitOffset)=numElements;
					HEADER_LENGTH(sparseArray,splitChunk,splitOffset)-=newLength;
					paramChanges[CA_SPARSE_ELEMENTS]=(numElements-sparseOffset)-newLength;		// adjusted this record by this much
					paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=-newLength;
					splitChunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the chunk
					splitChunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the array
					sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex+1,paramChanges);	// adjust the cache for the array
				}
				else
				{
					fail=true;
				}
			}
			else																				// no splitting, just push record over by the amount to clear
			{
				HEADER_OFFSET(sparseArray,chunk,offset)+=numElements;							// bump the offset part of the record
				chunk->params[CA_SPARSE_ELEMENTS]+=numElements;									// adjust the parameters for the chunk
				sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=numElements;		// add back the elements we deleted at the start
				paramChanges[CA_SPARSE_ELEMENTS]=numElements;									// caches need to adjust for this too
				paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=0;
				CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);	// adjust the cache for the array
			}
		}
	}
//	SparseArraySanity(sparseArray);
	return(!fail);
}

bool SparseArrayRangeSet(SPARSE_ARRAY *sparseArray,uint64_t startPosition,uint64_t numElements,void *value)
// Set elements of sparseArray from startPosition for numElements
// to value.
// If there is a problem, set the error and return false.
// This will overwrite anything currently in the array at
// the given positions, and will cause items to be split
// if it bisects them.
// NOTE: if the range goes beyond the current end of the array,
// the array will be extended.
{
	CARRAY_CHUNK
		*chunk,
		*newChunk,
		*splitChunk;
	uint64_t
		offset,
		splitOffset;
	uint64_t
		actualStart;
	uint64_t
		startIndex;
	uint64_t
		sparseOffset;
	uint64_t
		newOffset,
		newLength;
	uint64_t
		outParams[sparseArray->elementCArray->numParams];
	int64_t
		paramChanges[sparseArray->elementCArray->numParams];
	uint32_t
		i;
	bool
		fail;

	fail=false;
	if(numElements)
	{
		for(i=0;i<sparseArray->elementCArray->numParams;i++)									// clear out all param changes
		{
			paramChanges[i]=0;
		}

		SparseArrayRangeDelete(sparseArray,startPosition,numElements);									// get rid of any elements that were in that spot
		SparseArrayRecordLocate(sparseArray,startPosition,&chunk,&offset,&startIndex,&actualStart);		// find the record containing the start
		// NOTE: SparseArrayRecordLocate updates local cache to beginning of clear which code
		// below can't change.
		if(chunk)																				// if there is a record there, see how we need to clear
		{
			sparseOffset=HEADER_OFFSET(sparseArray,chunk,offset);

			if(startPosition>actualStart+sparseOffset)											// need to split? (start is inside the "active" area)
			{
				if(CArrayEmptyInsert(sparseArray->elementCArray,startIndex,2,&chunk,outParams))	// add 2 elements, one for the new stuff, and one for the half of the split we need
				{
					offset=startIndex-outParams[CA_ELEMENTS];
					newOffset=offset;
					newChunk=chunk;
					newOffset++;																// move to the next record
					if(newOffset>=CArrayChunkElementsGet(newChunk))								// next record is in next chunk, so move there
					{
						newChunk=newChunk->nextChunk;
						newOffset=0;
					}

					splitOffset=newOffset;
					splitChunk=newChunk;
					splitOffset++;																// move to the next record (which is the original we need to split)
					if(splitOffset>=CArrayChunkElementsGet(splitChunk))							// next record is in next chunk, so move there
					{
						splitChunk=splitChunk->nextChunk;
						splitOffset=0;
					}

					memcpy(&((uint8_t *)newChunk->data)[newOffset*sparseArray->elementCArray->elementSize],value,sparseArray->elementCArray->elementSize);	// blast in new value
					memcpy(&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize],&((uint8_t *)splitChunk->data)[splitOffset*sparseArray->elementCArray->elementSize],sparseArray->elementCArray->elementSize);	// blast in split value

					// Finally, adjust offsets and lengths

					// Do first chunk
					newLength=startPosition-actualStart-sparseOffset;							// calculate the new length for the first item
					HEADER_LENGTH(sparseArray,chunk,offset)=newLength;							// write into the record (record contains previous offset from the memcpy above)
					paramChanges[CA_SPARSE_ELEMENTS]=sparseOffset+newLength;					// record became this much larger due to memcpy and adjustment we just made
					paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=newLength;
					chunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];							// adjust the sparse elements in the chunk
					chunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the elements in the array
					sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);					// adjust the cache for the array

					// Do second chunk
					HEADER_OFFSET(sparseArray,newChunk,newOffset)=0;
					HEADER_LENGTH(sparseArray,newChunk,newOffset)=numElements;
					paramChanges[CA_SPARSE_ELEMENTS]=numElements;
					paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=numElements;
					newChunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];							// adjust the sparse elements in the chunk
					newChunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the elements in the array
					sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];	// adjust the elements in the array
					CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex+1,paramChanges);					// adjust the cache for the array (+1 because the second chunk is one record over from the first)

					// Do third (original) chunk
					HEADER_OFFSET(sparseArray,splitChunk,splitOffset)=0;
					HEADER_LENGTH(sparseArray,splitChunk,splitOffset)-=newLength;
					paramChanges[CA_SPARSE_ELEMENTS]=(-sparseOffset)-newLength;
					paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=-newLength;
					splitChunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];						// adjust the sparse elements in the chunk
					splitChunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the elements in the array
					sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex+2,paramChanges);					// adjust the cache for the array (+2 because the third chunk is two records over from the first)
				}
				else
				{
					fail=true;
				}
			}
			else	// no splitting, just push other record over, and add this one
			{
				HEADER_OFFSET(sparseArray,chunk,offset)-=startPosition-actualStart;				// reduce the offset of this element
				paramChanges[CA_SPARSE_ELEMENTS]=-(startPosition-actualStart);					// caches need to adjust for this too
				paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=0;
				chunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];							// adjust the sparse elements in the chunk
				sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the elements in the array
				CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);	// adjust the cache for the array

				if(CArrayEmptyInsert(sparseArray->elementCArray,startIndex,1,&chunk,outParams))	// add an element
				{
					offset=startIndex-outParams[CA_ELEMENTS];
					memcpy(&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize],value,sparseArray->elementCArray->elementSize);	// blast in new value
					HEADER_OFFSET(sparseArray,chunk,offset)=startPosition-actualStart;
					HEADER_LENGTH(sparseArray,chunk,offset)=numElements;
					paramChanges[CA_SPARSE_ELEMENTS]=(startPosition-actualStart)+numElements;
					paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=numElements;
					chunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];							// adjust the sparse elements in the chunk
					chunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the elements in the array
					sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
					CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);	// adjust the cache for the array
				}
				else
				{
					fail=true;	// things will not be pretty if this fails, as it leaves the array in an inconsistent state
				}
			}
		}
		else	// no record, so must be at the end, so add a new record there, and set the needed offset at the start
		{
			startIndex=CArrayTotalElementsGet(sparseArray->elementCArray);
			if(CArrayEmptyInsert(sparseArray->elementCArray,startIndex,1,&chunk,outParams))		// add an element at the end
			{
				offset=startIndex-outParams[CA_ELEMENTS];
				memcpy(&((uint8_t *)chunk->data)[offset*sparseArray->elementCArray->elementSize],value,sparseArray->elementCArray->elementSize);	// blast in new value
				newOffset=startPosition-CArrayParamTotalGet(sparseArray->elementCArray,CA_SPARSE_ELEMENTS);
				HEADER_OFFSET(sparseArray,chunk,offset)=newOffset;
				HEADER_LENGTH(sparseArray,chunk,offset)=numElements;
				paramChanges[CA_SPARSE_ELEMENTS]=newOffset+numElements;							// adjusted this record by this much
				paramChanges[CA_LENGTH_SPARSE_ELEMENTS]=numElements;
				chunk->params[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];
				chunk->params[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];
				sparseArray->elementCArray->paramTotals[CA_SPARSE_ELEMENTS]+=paramChanges[CA_SPARSE_ELEMENTS];	// adjust the array
				sparseArray->elementCArray->paramTotals[CA_LENGTH_SPARSE_ELEMENTS]+=paramChanges[CA_LENGTH_SPARSE_ELEMENTS];	// adjust the array
				CArrayCacheEntriesUpdate(sparseArray->elementCArray,startIndex,paramChanges);	// adjust the cache for the array
			}
			else
			{
				fail=true;
			}
		}
	}
//	SparseArraySanity(sparseArray);
	return(!fail);
}

bool SparseArrayElementEndsGet(SPARSE_ARRAY *sparseArray,uint64_t *startElementOffset,uint64_t *endElementOffset)
// If the given sparse array contains elements, fill in startElementOffset as the
// index to the first used element in the array, and endElementOffset as the offset
// of the last used element in the array (+1), then return true.
// Otherwise, return false.
{
	if(CArrayTotalElementsGet(sparseArray->elementCArray))
	{
		(*startElementOffset)=HEADER_OFFSET(sparseArray,sparseArray->elementCArray->firstChunk,0);
		(*endElementOffset)=CArrayParamTotalGet(sparseArray->elementCArray,CA_SPARSE_ELEMENTS);
		return(true);
	}
	return(false);
}

uint64_t SparseArrayTotalOccupiedElementsGet(SPARSE_ARRAY *sparseArray)
// Return the total number of occupied sparse elements in the sparse array.
{
	return(CArrayParamTotalGet(sparseArray->elementCArray,CA_LENGTH_SPARSE_ELEMENTS));
}

uint64_t SparseArrayTotalElementsGet(SPARSE_ARRAY *sparseArray)
// Return the total number of sparse elements in the sparse array.
{
	return(CArrayParamTotalGet(sparseArray->elementCArray,CA_SPARSE_ELEMENTS));
}

uint64_t SparseArraySegmentsGet(SPARSE_ARRAY *sparseArray)
// Return how many segments there are in the passed sparseArray.
{
	return(CArrayTotalElementsGet(sparseArray->elementCArray));
}

bool SparseArrayIsEmpty(SPARSE_ARRAY *sparseArray)
// Return true if the given sparse array contains no sparse elements, false otherwise.
{
	if(CArrayTotalElementsGet(sparseArray->elementCArray))
	{
		return(false);
	}
	return(true);
}

void SparseArrayUnInit(SPARSE_ARRAY *sparseArray)
// Uninitialize a sparse array.
{
	CArrayDispose(sparseArray->elementCArray);
}

bool SparseArrayInit(SPARSE_ARRAY *sparseArray,uint64_t elementSize)
// Initialize a sparse array.
// If there is a problem, set the error, and return false.
{
	if((sparseArray->elementCArray=CArrayCreate(SPARSE_ARRAY_CHUNK_ELEMENTS,elementSize,SparseArrayElementsMove,CA_NUM_PARAMS,SPARSE_ARRAY_CACHE_SIZE)))
	{
		sparseArray->cacheEntry.cArrayIndex=0;				// no cached element yet
		return(true);
	}
	return(false);
}
