// Style list management functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// This code handles keeping track of style in the text
// it uses the sparse-array concept to hold the style information
// NOTE: styles are coalesced in the array to make them work
// in a more intuitive manner

#include	"includes.h"

typedef struct styleElement
{
	SPARSE_ARRAY_HEADER
		saHeader;				// header needed for sparse array code
	uint32_t
		style;					// style index
} STYLE_ELEMENT;

typedef struct styleUniverse
{
	SPARSE_ARRAY
		styleArray;
} STYLE_UNIVERSE;

static void PositionCoalesce(STYLE_UNIVERSE *universe,uint64_t position)
// Look at the style at position, and position-1.
// If they are the same, coalesce the array.
{
	STYLE_ELEMENT
		*leftElement,
		*rightElement;
	STYLE_ELEMENT
		newElement;
	uint64_t
		leftStartPosition,
		leftElements,
		rightStartPosition,
		rightElements;

	if(position>0)
	{
		if(SparseArrayRecordForPositionGet(&universe->styleArray,position-1,&leftStartPosition,&leftElements,(void **)&leftElement))
		{
			if(leftElement&&(leftStartPosition+leftElements==position))	// make sure there's something to the left, and it ends just before position
			{
				if(SparseArrayRecordForPositionGet(&universe->styleArray,position,&rightStartPosition,&rightElements,(void **)&rightElement))
				{
					if(rightElement&&(leftElement->style==rightElement->style))	// make sure there's something to the right, and it matches
					{
						newElement.style=leftElement->style;
						SparseArrayRangeSet(&universe->styleArray,leftStartPosition,leftElements+rightElements,&newElement);
					}
				}
			}
		}
	}
}

void StyleRangeDelete(STYLE_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Remove style array elements from startPosition for numElements.
// NOTE: this does not mind being called for positions outside of the current array.
{
	SparseArrayRangeDelete(&universe->styleArray,startPosition,numElements);
	PositionCoalesce(universe,startPosition);	// if two of the same style ended up next to each other, merge them
}

bool StyleRangeInsert(STYLE_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Insert numElements into the style array at startPosition.
{
	return(SparseArrayRangeInsert(&universe->styleArray,startPosition,numElements));
}

bool StyleRangeGet(STYLE_UNIVERSE *universe,uint64_t position,uint64_t *startPosition,uint64_t *numElements,uint32_t *style)
// Return the range of positions in the array which include position
// and have the same style as position.
// NOTE: this uses a cache so it is efficient to call repeatedly for
// positions which are nearby each other.
// NOTE: if there's no style given for position this will
// return false w/o setting *style. In this case, startPosition and numElements
// will describe the empty area. If the empty area is at the end of the universe,
// startPosition will point to the end, and numElements will be 0.
{
	STYLE_ELEMENT
		*element;

	if(SparseArrayRecordForPositionGet(&universe->styleArray,position,startPosition,numElements,(void **)&element))
	{
		if(element)
		{
			*style=element->style;
			return(true);
		}
	}
	return(false);
}

bool StyleAtPositionGet(STYLE_UNIVERSE *universe,uint64_t position,uint32_t *style)
// Call this to return the style at the given position of
// the given style universe.
// NOTE: this uses a cache so it is efficient to call repeatedly for
// positions which are nearby each other.
// NOTE: if there's no style given for position this will return false
// w/o setting *style.
{
	uint64_t
		startPosition,
		numElements;

	return(StyleRangeGet(universe,position,&startPosition,&numElements,style));
}

bool StyleRangeClear(STYLE_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements)
// Clear style at startPosition for numElements of universe.
// If there is a problem, set the error and return false.
// The style will overwrite anything currently in the array at
// the given positions.
{
	return(SparseArrayRangeClear(&universe->styleArray,startPosition,numElements));
}

bool StyleRangeSet(STYLE_UNIVERSE *universe,uint64_t startPosition,uint64_t numElements,uint32_t style)
// Set style at startPosition for numElements of universe.
// If there is a problem, set the error and return false.
// The style will overwrite anything currently in the array at
// the given positions.
// NOTE: if the range goes beyond the current end of the array,
// the array will be extended.
{
	STYLE_ELEMENT
		element;

	element.style=style;
	if(SparseArrayRangeSet(&universe->styleArray,startPosition,numElements,&element))
	{
		PositionCoalesce(universe,startPosition);
		PositionCoalesce(universe,startPosition+numElements);
		return(true);
	}
	return(false);
}

void StylesAdjustForChange(STYLE_UNIVERSE *universe,uint64_t startOffset,uint64_t oldEndOffset,uint64_t newEndOffset)
// Given a change at startOffset, run through the given style list, and fix it up.
{
	StyleRangeDelete(universe,startOffset,oldEndOffset-startOffset);
	StyleRangeInsert(universe,startOffset,newEndOffset-startOffset);
}

bool StyleCopy(STYLE_UNIVERSE *sourceUniverse,uint64_t sourceOffset,STYLE_UNIVERSE *destUniverse,uint64_t destOffset,uint64_t numElements)
// Copy numElements of style information from sourceUniverse starting at sourceOffset
// to destUniverse at destOffset.
// If there is a problem, set the error, and return false.
// NOTE: if sourceOffset+numElements goes beyond sourceUniverse, that part of destUniverse
// will be cleared.
{
	bool
		fail;
	bool
		hadStyle;
	uint64_t
		startPosition;
	uint64_t
		elementCount;
	uint32_t
		style;
	uint64_t
		elementsToSet;

	fail=false;
	while(!fail&&numElements)
	{
		hadStyle=StyleRangeGet(sourceUniverse,sourceOffset,&startPosition,&elementCount,&style);
		if(elementCount)		// element count will come back as 0 if we fall off the end of sourceUniverse
		{
			elementsToSet=elementCount-(sourceOffset-startPosition);
			if(elementsToSet>numElements)
			{
				elementsToSet=numElements;
			}
			if(hadStyle)
			{
				fail=!StyleRangeSet(destUniverse,destOffset,elementsToSet,style);
			}
			else
			{
				fail=!StyleRangeClear(destUniverse,destOffset,elementsToSet);
			}
			sourceOffset+=elementsToSet;
			destOffset+=elementsToSet;
			numElements-=elementsToSet;
		}
		else		// off the end of the source, just clear the destination
		{
			fail=!StyleRangeClear(destUniverse,destOffset,numElements);
			numElements=0;		// exit the loop now
		}
	}
	return(!fail);
}

void StyleUniverseClose(STYLE_UNIVERSE *universe)
// Dispose of a style universe.
{
	SparseArrayUnInit(&(universe->styleArray));
	MDisposePtr(universe);
}

STYLE_UNIVERSE *StyleUniverseOpen()
// Open a style universe (with no elements).
// If there is a problem, set the error, and return NULL.
{
	STYLE_UNIVERSE
		*universe;

	if((universe=(STYLE_UNIVERSE *)MNewPtr(sizeof(STYLE_UNIVERSE))))
	{
		if(SparseArrayInit(&(universe->styleArray),sizeof(STYLE_ELEMENT)))
		{
			return(universe);
		}
		MDisposePtr(universe);
	}
	return(NULL);
}
