// Lowest level Text handling functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".
//
// NOTE: text is managed using a "all in memory" model.
// So, as far as most functions are concerned, the entire text data
// structure is kept in memory while it is being edited. The OS swapping
// functions are charged with mapping the data in and out if needed.
//
// To manage the text data efficiently, and to keep memory from
// fragmenting, all text is kept in "Chunks" using the carray functions.

#include	"includes.h"

#define	TEXT_CHUNK_SIZE				4096		// size in bytes of the data of text chunks (smaller can be quicker for some operations, but is less memory efficient)
#define	CACHE_ENTRIES				4			// number of entries in the cache

#define	CA_NUM_PARAMS				(CA_NEW_LINES+1)

// ----------------------------------------------------------------------------

static void TextArrayElementsMove(const CARRAY *cArray,const void *source,uint64_t sourceElementIndex,void *dest,uint64_t destElementIndex,uint64_t numElements,uint64_t *params)
// Move function needed by the CARRAY code. This knows how to move
// elements around, and update the needed params.
// In the case of text arrays, we manage a single user-parameter
// which tells how many newlines are seen.
{
	uint64_t
		totalNewLines;
	uint8_t
		*srcPointer,
		*dstPointer;

	if(source)
	{
		if(params)
		{
			params[CA_ELEMENTS]=numElements;
			totalNewLines=0;
			srcPointer=&((uint8_t *)source)[sourceElementIndex];
			if(dest)							// moving data somewhere?
			{
				dstPointer=&((uint8_t *)dest)[destElementIndex];

				// Check move direction to handle overlaps correctly
				if(dstPointer<srcPointer)		// move forward through the elements
				{
					while(numElements)
					{
						if(((*dstPointer)=(*srcPointer))=='\n')
						{
							totalNewLines++;
						}
						srcPointer++;
						dstPointer++;
						numElements--;
					}
				}
				else							// move backward
				{
					srcPointer+=numElements;
					dstPointer+=numElements;
					while(numElements)
					{
						srcPointer--;
						dstPointer--;
						if(((*dstPointer)=(*srcPointer))=='\n')
						{
							totalNewLines++;
						}
						numElements--;
					}
				}
			}
			else	// just count up elements of the source
			{
				while(numElements)
				{
					if((*srcPointer)=='\n')
					{
						totalNewLines++;
					}
					srcPointer++;
					numElements--;
				}
			}
			params[CA_NEW_LINES]=totalNewLines;
		}
		else							// don't have to track the parameters, just do the move
		{
			if(dest)					// only do work if dest is specified
			{
				MMoveMem(&((uint8_t *)source)[sourceElementIndex],&((uint8_t *)dest)[destElementIndex],numElements);
			}
		}
	}
	else
	{
		if(params)
		{
			params[CA_ELEMENTS]=numElements;
			params[CA_NEW_LINES]=0;
		}
		if(dest)					// only do work if dest is specified
		{
			memset(&((uint8_t *)dest)[destElementIndex*cArray->elementSize],0,numElements*cArray->elementSize);
		}
	}
}

// ----------------------------------------------------------------------------

bool TextUniverseSanityCheck(TEXT_UNIVERSE *universe)
// Check the sanity of the text universe, report result.
// If there is something wrong, crash or ErrorSet, and return false.
{
	uint64_t
		totalChunks;
	unsigned int
		efficiency;

	fprintf(stderr,"sanity check\n");
	CArraySanity(universe->textCArray);
	totalChunks=CArrayTotalChunksGet(universe->textCArray);
	fprintf(stderr,"Chunks: %" PRIu64 "\n",totalChunks);
	if(totalChunks)
	{
		efficiency=CArrayTotalElementsGet(universe->textCArray)*100/(totalChunks*TEXT_CHUNK_SIZE);
		fprintf(stderr,"Efficiency: %u%%\n",efficiency);
	}
	return(true);
}

// ----------------------------------------------------------------------------

uint64_t TextUniverseLineToPosition(TEXT_UNIVERSE *universe,uint64_t line,TEXT_STREAM *stream)
// Given a line number, return the position in the text where it starts.
// If the line number is larger than the number of lines in the text, this will
// return the total size of the universe.
// If stream is non-NULL, it will be filled to point to the start
// of the line.
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		lineOffset,				// tells how many lines into the located chunk we need to go
		offset;
	uint64_t
		position;
	uint64_t
		paramTotals[CA_NUM_PARAMS];

	// The locate function will tell us which chunk contains a newline with
	// the given index, and how many newlines there are in the chunk before
	// the one we asked for. For example, if we tried to hunt down
	// index 0 on a universe that contained no newlines, the location function
	// would return NULL/0 (no chunk contains the newline requested).
	// Since this function thinks of line 0 as the start of the universe
	// (no newlines yet), we need to handle that case specially...
	if(line)	// see if we need to just get the start of the universe
	{
		CArrayLocate(universe->textCArray,CA_NEW_LINES,line-1,&chunk,paramTotals);	// find the chunk with the newline at the end of the previous line to the one requested
		offset=0;
		position=paramTotals[CA_ELEMENTS];									// get the start position of the chunk
		if(chunk)															// chunk may be NULL if the requested line was off the end of the universe
		{
			lineOffset=line-paramTotals[CA_NEW_LINES];						// get the number of lines we need to count through
			while(lineOffset)												// move through until enough lines have been counted
			{
				if(((uint8_t *)chunk->data)[offset]=='\n')
				{
					lineOffset--;
				}
				offset++;
			}
			position+=offset;
			if(offset>=chunk->params[CA_ELEMENTS])							// if fell off the end, move to next chunk
			{
				chunk=chunk->nextChunk;
				offset=0;
			}
		}
		if(stream)
		{
			stream->universe=universe;
			stream->chunk=chunk;
			stream->offset=offset;
			stream->position=position;
		}
	}
	else
	{
		position=0;
		if(stream)
		{
			stream->universe=universe;
			stream->chunk=universe->textCArray->firstChunk;
			stream->offset=0;
			stream->position=0;
		}
	}
	return(position);
}

uint64_t TextUniversePositionToLine(TEXT_UNIVERSE *universe,uint64_t position,TEXT_STREAM *stream)
// Given a position in the universe, return the line number in the text
// which contains it.
// If stream is non-NULL, it will be filled to point to the start of the
// line.
{
	CARRAY_CHUNK
		*chunk;
	uint64_t
		i,
		offset;
	uint64_t
		newLinesSoFar;
	bool
		hadLineInChunk;
	uint64_t
		lineOffset;
	uint64_t
		paramTotals[CA_NUM_PARAMS];

	CArrayLocate(universe->textCArray,CA_ELEMENTS,position,&chunk,paramTotals);	// locate the chunk containing the position
	newLinesSoFar=paramTotals[CA_NEW_LINES];						// get number of newlines up to the start of the chunk
	offset=position-paramTotals[CA_ELEMENTS];						// get offset into the chunk

	// move out to the position in the chunk, count up newlines seen
	lineOffset=0;													// keep compiler quiet
	hadLineInChunk=false;
	for(i=0;i<offset;i++)
	{
		if(((uint8_t *)chunk->data)[i]=='\n')
		{
			newLinesSoFar++;
			lineOffset=i+1;											// remember where the last line started in the chunk to avoid extra searching if possible
			hadLineInChunk=true;
		}
	}

	if(stream)
	{
		if(!hadLineInChunk)											// if no newline in this chunk, then the line in question lives in another
		{
			TextUniverseLineToPosition(universe,newLinesSoFar,stream);	// point stream at the start of the line
		}
		else	// line was in the chunk we already found
		{
			stream->universe=universe;
			stream->chunk=chunk;
			stream->offset=lineOffset;
			stream->position=paramTotals[CA_ELEMENTS]+lineOffset;	// stream position points to the start of the line
			if(lineOffset>=stream->chunk->params[CA_ELEMENTS])		// if fell off the end, go find the next chunk
			{
				stream->chunk=stream->chunk->nextChunk;
				stream->offset=0;
			}
		}
	}
	return(newLinesSoFar);
}

void TextStreamToNextLine(TEXT_STREAM *textStream)
// Push textStream forward until it passes the next newline character.
// If the stream runs into the end before hitting the next newline, then
// just return that.
// For speed, this searches for the next line in the current chunk (since it's
// likely to be there). If the end of the chunk is hit, this will perform
// the search with a little more brute force.
{
	uint64_t
		line;
	bool
		found;
	CARRAY_CHUNK
		*chunk;
	uint64_t
		offset;

	found=false;
	if((chunk=textStream->chunk))							// make sure we're not off the end
	{
		offset=textStream->offset;
		while(!found&&(offset<chunk->params[CA_ELEMENTS]))
		{
			if(((uint8_t *)chunk->data)[offset]=='\n')
			{
				found=true;
			}
			offset++;										// move forward one, even if found, since we want to push to the next character
		}
		textStream->position+=offset-textStream->offset;	// push the position forward
		if(offset<chunk->params[CA_ELEMENTS])
		{
			textStream->offset=offset;
		}
		else
		{
			textStream->chunk=chunk->nextChunk;
			textStream->offset=0;
		}
		if(!found)											// if we got to the end w/o finding it, work a bit harder
		{
			line=TextUniversePositionToLine(textStream->universe,textStream->position,NULL);	// find out which line we're currently on
			TextUniverseLineToPosition(textStream->universe,line+1,textStream);					// move to the next one
		}
	}
}

void TextStreamPositionSet(TEXT_STREAM *textStream,TEXT_UNIVERSE *universe,uint64_t position)
// Initialize textStream to point to the given character of the given universe.
// NOTE: if position is past the end of the universe, it'll be pinned to universe->totalBytes.
{
	uint64_t
		totalBytes;
	uint64_t
		paramTotals[CA_NUM_PARAMS];

	textStream->universe=universe;
	totalBytes=TextUniverseTotalBytes(universe);
	if(position<totalBytes)
	{
		CArrayLocate(universe->textCArray,CA_ELEMENTS,position,&textStream->chunk,paramTotals);
		textStream->offset=position-paramTotals[CA_ELEMENTS];
		textStream->position=position;
	}
	else
	{
		textStream->chunk=NULL;	// past end -- no chunk
		textStream->offset=0;
		textStream->position=totalBytes;
	}
}

void TextStreamMoveRelative(TEXT_STREAM *textStream,int64_t amount)
// Move the current position relative to where it is now.
// For single character moves, this is less efficient than
// TextStreamNextChar and TextStreamPreviousChar, so it should only be
// called when the amount to move is arbitrary.
// It follows the same rules as TextStreamNextChar (can go to position==totalBytes),
// and TextStreamPreviousChar (won't back up past the start).
{
	int64_t
		newOffset;
	int64_t
		newPosition;
	uint64_t
		totalBytes;
	uint64_t
		paramTotals[CA_NUM_PARAMS];

	newOffset=(int64_t)textStream->offset+amount;
	if(textStream->chunk&&(newOffset>=0)&&(newOffset<(int64_t)textStream->chunk->params[CA_ELEMENTS]))		// see if we can just move quickly within the chunk
	{
		textStream->offset=newOffset;		// move within the chunk
		textStream->position+=amount;
	}
	else	// need to move to a new chunk, hunt it down
	{
		newPosition=(int64_t)textStream->position+amount;
		if(newPosition<0)
		{
			newPosition=0;
		}
		else
		{
			totalBytes=TextUniverseTotalBytes(textStream->universe);
			if((uint64_t)newPosition>totalBytes)
			{
				newPosition=totalBytes;
			}
		}
		CArrayLocate(textStream->universe->textCArray,CA_ELEMENTS,newPosition,&textStream->chunk,paramTotals);
		textStream->position=newPosition;
		textStream->offset=newPosition-paramTotals[CA_ELEMENTS];
	}
}

// ----------------------------------------------------------------------------

void TextStreamTextExtract(TEXT_STREAM *stream,uint8_t *text,uint64_t numBytes,uint64_t *linesCopied)
// Extract numBytes of text from stream.
// There must be at least numBytes of text in the stream at the given position, and
// text must be large enough to accept them or bad things will happen.
// stream will be updated to point past the extraction.
// If there is a problem, ErrorSet, and return false.
// NOTE: if linesCopied is non-NULL, it'll report how many newlines
// were encountered during the copy operation.
// NOTE: this doesn't handle overlapping source and destination.
{
	uint64_t
		paramTotals[CA_NUM_PARAMS];

	CArrayChunkOffsetExtract(&stream->chunk,&stream->offset,stream->universe->textCArray,text,0,numBytes,linesCopied?paramTotals:NULL);
	if(linesCopied)
	{
		(*linesCopied)=paramTotals[CA_NEW_LINES];
	}
}

void TextUniverseTextExtract(TEXT_UNIVERSE *universe,uint64_t position,uint8_t *text,uint64_t numBytes,uint64_t *linesCopied)
// Extract numBytes of text from universe at position.
// There must be at least numBytes of text in the universe at the given position, and
// text must be large enough to accept them or bad things will happen.
// If there is a problem, ErrorSet, and return false.
{
	uint64_t
		paramTotals[CA_NUM_PARAMS];

	CArrayExtract(universe->textCArray,position,text,0,numBytes,linesCopied?paramTotals:NULL);
	if(linesCopied)
	{
		(*linesCopied)=paramTotals[CA_NEW_LINES];
	}
}

// ----------------------------------------------------------------------------

void TextStreamDelete(TEXT_STREAM *destStream,uint64_t numBytes)
// Delete numBytes of text from destStream at the current position.
// If there is a problem, ErrorSet, and return false.
// This will update destStream to make sure it is still valid after the
// call.
// Otherwise, no saved TEXT_STREAM is valid for the destStream's universe
// after this is called.
{
	CArrayElementsDelete(destStream->universe->textCArray,destStream->position,numBytes,NULL,NULL);
}

void TextUniverseDelete(TEXT_UNIVERSE *universe,uint64_t position,uint64_t numBytes)
// Delete numBytes of text from universe starting at position.
// No saved TEXT_STREAM is valid for universe after this is called.
// If there is a problem, ErrorSet, and return false.
{
	CArrayElementsDelete(universe->textCArray,position,numBytes,NULL,NULL);
}

// ----------------------------------------------------------------------------

bool TextStreamInsertFromStream(TEXT_STREAM *sourceStream,TEXT_STREAM *destStream,uint64_t numBytes)
// Insert numBytes of text from sourceStream into destStream.
// If there is a problem, ErrorSet, and return false.
// NOTE: it is illegal for sourceStream to point to less data than numBytes.
// NOTE ALSO: to accommodate inserting unchunked text, sourceStream is allowed to contain
// chunks larger than TEXT_CHUNK_SIZE with totalLines fields that are inaccurate.
// *sourceStream and *destStream will both be pushed numBytes forward.
{
	return(CArrayInsert(sourceStream->universe->textCArray,sourceStream->position,destStream->universe->textCArray,destStream->position,numBytes,NULL,NULL));
}

bool TextStreamInsertFromText(const uint8_t *text,TEXT_STREAM *destStream,uint64_t numBytes)
// Insert numBytes of text into stream.
// stream will point to the byte immediately following the insertion.
// Otherwise, no saved TEXT_STREAM is valid for the stream's universe
// after this is called.
// If there is a problem, ErrorSet, and return false.
{
	return(CArrayRawInsert(text,destStream->universe->textCArray,destStream->position,numBytes,NULL,NULL));
}

bool TextUniverseInsertFromText(const uint8_t *text,TEXT_UNIVERSE *destUniverse,uint64_t destPosition,uint64_t numBytes)
// Insert numBytes of text into destUniverse at destPosition.
// If there is a problem, ErrorSet, and return false.
{
	return(CArrayRawInsert(text,destUniverse->textCArray,destPosition,numBytes,NULL,NULL));
}

bool TextUniverseInsertFromStream(TEXT_STREAM *sourceStream,TEXT_UNIVERSE *destUniverse,uint64_t destPosition,uint64_t numBytes)
// Insert numBytes of text from sourceStream into destUniverse at destPosition.
// sourceStream must point to a universe other than destUniverse.
// If there is a problem, ErrorSet, and return false.
// No saved TEXT_STREAM is valid for the destUniverse after this is
// called.
// sourceStream will be pushed numBytes forward.
{
	return(CArrayInsert(sourceStream->universe->textCArray,sourceStream->position,destUniverse->textCArray,destPosition,numBytes,NULL,NULL));
}

bool TextUniverseInsertFromUniverse(TEXT_UNIVERSE *sourceUniverse,uint64_t sourcePosition,TEXT_UNIVERSE *destUniverse,uint64_t destPosition,uint64_t numBytes)
// Insert numBytes of text from sourceUniverse, sourcePosition into destUniverse, destPosition.
// sourceUniverse and destUniverse cannot be the same.
// No saved TEXT_STREAM is valid for the destUniverse after this is
// called.
// If there is a problem, ErrorSet, and return false.
{
	return(CArrayInsert(sourceUniverse->textCArray,sourcePosition,destUniverse->textCArray,destPosition,numBytes,NULL,NULL));
}

// ----------------------------------------------------------------------------

void TextUniverseClose(TEXT_UNIVERSE *universe)
// Dispose of a text universe created by TextUniverseOpen.
{
	CArrayDispose(universe->textCArray);
	MDisposePtr(universe);
}

TEXT_UNIVERSE *TextUniverseOpen()
// Create the data structures necessary to manage a universe of text.
// If there is a problem, ErrorSet, and return NULL.
{
	TEXT_UNIVERSE
		*universe;

	if((universe=(TEXT_UNIVERSE *)MNewPtr(sizeof(TEXT_UNIVERSE))))	// create an empty universe
	{
		if((universe->textCArray=CArrayCreate(TEXT_CHUNK_SIZE,1,TextArrayElementsMove,CA_NUM_PARAMS,CACHE_ENTRIES)))
		{
			return(universe);
		}
		MDisposePtr(universe);
	}
	return((TEXT_UNIVERSE *)NULL);
}
