// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#define	CA_NEW_LINES				CA_NUM_INHERENT_PARAMS		// user parameter that keeps track of the number of new lines in a chunk

typedef struct textStream
{
	TEXT_UNIVERSE
		*universe;		// tells which universe this stream points to
	CARRAY_CHUNK
		*chunk;			// the current chunk in the text universe
	uint64_t
		offset;			// the current offset within the chunk
	uint64_t
		position;		// the absolute position within the text universe (not meaningful if chunk==NULL)
} TEXT_STREAM;

typedef struct textUniverse
{
	CARRAY
		*textCArray;							// cArray that holds the text
} TEXT_UNIVERSE;

// ----------------------------------------------------------------------------

bool TextUniverseSanityCheck(TEXT_UNIVERSE *universe);
uint64_t TextUniverseLineToPosition(TEXT_UNIVERSE *universe,uint64_t line,TEXT_STREAM *stream);
uint64_t TextUniversePositionToLine(TEXT_UNIVERSE *universe,uint64_t position,TEXT_STREAM *stream);
void TextStreamToNextLine(TEXT_STREAM *textStream);
void TextStreamPositionSet(TEXT_STREAM *textStream,TEXT_UNIVERSE *universe,uint64_t position);
void TextStreamMoveRelative(TEXT_STREAM *textStream,int64_t amount);
void TextStreamTextExtract(TEXT_STREAM *stream,uint8_t *text,uint64_t numBytes,uint64_t *linesCopied);
void TextUniverseTextExtract(TEXT_UNIVERSE *universe,uint64_t position,uint8_t *text,uint64_t numBytes,uint64_t *linesCopied);
void TextStreamDelete(TEXT_STREAM *destStream,uint64_t numBytes);
void TextUniverseDelete(TEXT_UNIVERSE *universe,uint64_t position,uint64_t numBytes);
bool TextStreamInsertFromStream(TEXT_STREAM *sourceStream,TEXT_STREAM *destStream,uint64_t numBytes);
bool TextStreamInsertFromText(const uint8_t *text,TEXT_STREAM *destStream,uint64_t numBytes);
bool TextUniverseInsertFromText(const uint8_t *text,TEXT_UNIVERSE *destUniverse,uint64_t destPosition,uint64_t numBytes);
bool TextUniverseInsertFromStream(TEXT_STREAM *sourceStream,TEXT_UNIVERSE *destUniverse,uint64_t destPosition,uint64_t numBytes);
bool TextUniverseInsertFromUniverse(TEXT_UNIVERSE *sourceUniverse,uint64_t sourcePosition,TEXT_UNIVERSE *destUniverse,uint64_t destPosition,uint64_t numBytes);
void TextUniverseClose(TEXT_UNIVERSE *universe);
TEXT_UNIVERSE *TextUniverseOpen();

// ----------------------------------------------------------------------------

static inline uint64_t TextUniverseTotalBytes(TEXT_UNIVERSE *universe)
// Returns the total number of bytes in a text universe.
{
	return(CArrayParamTotalGet(universe->textCArray,CA_ELEMENTS));
}

static inline uint64_t TextUniverseTotalLines(TEXT_UNIVERSE *universe)
// Returns the total number of lines in a text universe.
{
	return(CArrayParamTotalGet(universe->textCArray,CA_NEW_LINES));
}

// Text stream functions which allow access to text universes as streams of
// bytes -- useful for searching, etc...
// These are inlined for speed.

static inline uint8_t TextStreamChar(const TEXT_STREAM *textStream)
// Return the character at the current position in the text stream.
// This should only be called when the position is known to be valid.
{
	return(((uint8_t *)textStream->chunk->data)[textStream->offset]);
}

static inline void TextStreamNextChar(TEXT_STREAM *textStream)
// Move to the next character in the stream. This can run
// off the end of the stream (position==totalBytes). Once
// it does (and it is the caller's responsibility to know), calling
// TextStreamChar is not allowed.
{
	if(textStream->chunk)
	{
		if((++textStream->offset)>=textStream->chunk->params[CA_ELEMENTS])
		{
			textStream->chunk=textStream->chunk->nextChunk;
			textStream->offset=0;
		}
		textStream->position++;
	}
}

static inline bool TextStreamPreviousChar(TEXT_STREAM *textStream)
// Move to the previous character in the stream. This will not back
// up past the start of the stream. If the stream is off the end
// when it is called, it will back up to the last character (if possible).
// This will return false if it is impossible to back up from the
// current position.
{
	bool
		result;
	uint64_t
		totalBytes;

	result=true;
	if(textStream->chunk)					// in the stream?
	{
		if(textStream->offset)				// room to back up in the chunk?
		{
			textStream->offset--;
			textStream->position--;
		}
		else
		{
			if(textStream->chunk->previousChunk)	// current chunk not at the start?
			{
				textStream->chunk=textStream->chunk->previousChunk;
				textStream->offset=textStream->chunk->params[CA_ELEMENTS]-1;
				textStream->position--;
			}
			else
			{
				result=false;
			}
		}
	}
	else if(textStream->universe&&(totalBytes=TextUniverseTotalBytes(textStream->universe)))	// can we back up into the last chunk?
	{
		TextStreamPositionSet(textStream,textStream->universe,totalBytes-1);
	}
	else
	{
		result=false;
	}
	return(result);
}

static inline bool TextStreamAtEnd(const TEXT_STREAM *textStream)
// Return true of the current stream is positioned at the end.
{
	return(textStream->chunk==NULL);
}

static inline uint64_t TextStreamPositionGet(const TEXT_STREAM *textStream)
// Return the current universe position that the text stream points to.
{
	return(textStream->position);
}

static inline TEXT_UNIVERSE *TextStreamUniverseGet(const TEXT_STREAM *textStream)
// Return the current universe associated with the textStream.
{
	return(textStream->universe);
}

static inline void TextStreamPositionClear(TEXT_STREAM *textStream)
// Initialize textStream to point to nothing.
{
	textStream->universe=NULL;
	textStream->chunk=NULL;
	textStream->offset=0;
	textStream->position=0;
}
