// Undo/redo functions
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

#define	UNDO_CHUNK_ELEMENTS			256			// number of undo elements in an undo chunk

typedef struct undoElement
{
	uint64_t
		frameNumber;				// tells which frame of undo this belongs to
	uint64_t
		textStartPosition,			// position in text where replacement begins
		textLength,					// number of bytes to replace
		undoStartPosition,			// position in undo buffer of replacement text/style
		undoLength;					// length of replacement text
} UNDO_ELEMENT;

typedef struct undoGroup
{
	CARRAY
		*undoCArray;				// array which is used to keep track of the individual undos
	TEXT_UNIVERSE
		*undoText;					// text used while undoing
	STYLE_UNIVERSE
		*undoStyles;				// styles used while undoing
} UNDO_GROUP;

typedef struct undoUniverse
{
	uint64_t
		currentFrame;				// keeps track of the current undo frame index
	uint64_t
		groupingDepth;				// knows the depth of begin/end undo group calls
	uint64_t
		cleanElements;				// tells the number of elements in the undo-array when the buffer is considered "clean"
	bool
		cleanKnown;					// if false, cleanElements is meaningless and buffer is considered dirty
	bool
		undoActive;					// set true while undoing/redoing (tells not to delete redo information)
	bool
		undoing;					// set true while undoing (tells to place undo info into redo buffer)
	UNDO_GROUP
		undoGroup;					// used to keep track of undos
	UNDO_GROUP
		redoGroup;					// used to keep track of redos
} UNDO_UNIVERSE;

// ----------------------------------------------------------------------------

static void UndoElementsMove(const CARRAY *cArray,const void *source,uint64_t sourceElementIndex,void *dest,uint64_t destElementIndex,uint64_t numElements,uint64_t *params)
// Move function needed by the CARRAY code. This knows how to move
// elements around, and update the needed params.
// NOTE: the undo code doesn't maintain any params of its own, so
// this is quite simple.
{
	const UNDO_ELEMENT
		*sourceElement;
	UNDO_ELEMENT
		*destElement;

	if(params)
	{
		params[CA_ELEMENTS]=numElements;
	}

	if(dest)						// only do work if dest is specified
	{
		destElement=(UNDO_ELEMENT *)dest;
		if((sourceElement=(UNDO_ELEMENT *)source))
		{
			MMoveMem(&sourceElement[sourceElementIndex],&destElement[destElementIndex],numElements*sizeof(UNDO_ELEMENT));
		}
		else	// no source specified, fill with zeroes (doesn't ever happen, since code here never calls CArrayEmptyInsert)
		{
			memset(&destElement[destElementIndex],0,numElements*sizeof(UNDO_ELEMENT));
		}
	}
}

// ----------------------------------------------------------------------------

static void UndoClear(UNDO_GROUP *group)
// Clear the text, style, and undo arrays from group.
{
	CArrayElementsDelete(group->undoCArray,0,CArrayTotalElementsGet(group->undoCArray),NULL,NULL);	// delete all elements in the array
	StyleRangeDelete(group->undoStyles,0,TextUniverseTotalBytes(group->undoText));
	TextUniverseDelete(group->undoText,0,TextUniverseTotalBytes(group->undoText));
}

void UndoGroupBegin(EDITOR_BUFFER *buffer)
// All of the things that are done between now, and the matching UndoGroupEnd
// will be remembered as a group in the undo buffer.
{
	if(!buffer->undoUniverse->groupingDepth++)			// increment the depth, if it was 0, then start a new group
	{
		buffer->undoUniverse->currentFrame++;			// bump the frame number
	}
}

void UndoGroupEnd(EDITOR_BUFFER *buffer)
// Call this to balance a call to UndoGroupBegin.
// It ends the undo group, but will allow subsequent undos to be joined.
{
	if(buffer->undoUniverse->groupingDepth)
	{
		--buffer->undoUniverse->groupingDepth;			// decrement the group nesting count
	}
}

void UndoGroupEndStrict(EDITOR_BUFFER *buffer)
// Call this to balance a call to UndoGroupBegin.
// It ends the undo group, and will NOT allow subsequent undos to be joined.
{
	if(buffer->undoUniverse->groupingDepth)
	{
		if(!(--buffer->undoUniverse->groupingDepth))	// decrement the group nesting count, if it went to 0, force new group
		{
			buffer->undoUniverse->currentFrame++;		// bump the frame number
		}
	}
}

static bool UndoDeleteElementCreate(EDITOR_BUFFER *buffer,uint64_t startPosition,uint64_t numBytes,UNDO_GROUP *group)
// Create a new element at the end of the undo array.
// If there is a problem, set the error, and return false.
{
	TEXT_STREAM
		stream;
	UNDO_ELEMENT
		undoElement;

	if(!buffer->undoUniverse->groupingDepth)							// if not currently in a group, then make this a new frame
	{
		buffer->undoUniverse->currentFrame++;							// bump the frame number
	}

	undoElement.frameNumber=buffer->undoUniverse->currentFrame;			// set the frame number to the current undo frame
	undoElement.textStartPosition=startPosition;						// when the text is blasted back into place, this is where it goes
	undoElement.textLength=0;											// number of bytes to overwrite (none, we are deleting)
	undoElement.undoStartPosition=TextUniverseTotalBytes(group->undoText);	// start at what is now the current end of the undo text
	undoElement.undoLength=numBytes;									// number of bytes to put back
	if(CArrayRawInsert(&undoElement,group->undoCArray,CArrayTotalElementsGet(group->undoCArray),1,NULL,NULL))	// insert the element at the end
	{
		TextStreamPositionSet(&stream,buffer->textUniverse,startPosition);	// point to bytes about to be deleted
		if(StyleCopy(buffer->styleUniverse,startPosition,group->undoStyles,TextUniverseTotalBytes(group->undoText),numBytes))
		{
			if(TextUniverseInsertFromStream(&stream,group->undoText,TextUniverseTotalBytes(group->undoText),numBytes))		// copy into undo holding area
			{
				return(true);											// if we got them, then we are done
			}
			StyleRangeDelete(group->undoStyles,TextUniverseTotalBytes(group->undoText),numBytes);	// kill the styles we copied
		}
		CArrayElementsDelete(group->undoCArray,CArrayTotalElementsGet(group->undoCArray)-1,1,NULL,NULL);	// we failed, so get rid of the last element
	}
	return(false);
}

static bool UndoDeleteElementMerge(EDITOR_BUFFER *buffer,UNDO_ELEMENT *lastElement,uint64_t startPosition,uint64_t numBytes,UNDO_GROUP *group)
// A delete is happening, and the last element of the undo list is in the current
// frame. See if the delete can be combined with the last element of the undo list,
// or if another entirely new element needs to be created.
// If there is a problem, set the error, and return false.
{
	TEXT_STREAM
		stream;
	uint64_t
		bytesToEdge;

	if((lastElement->textStartPosition==startPosition)||(lastElement->textStartPosition+lastElement->textLength==startPosition+numBytes))	// for a delete to be grouped, it must start exactly in line with last element, or end exactly in line with it
	{
		if(startPosition>=lastElement->textStartPosition)
		{
			bytesToEdge=lastElement->textStartPosition+lastElement->textLength-startPosition;
			if(bytesToEdge>=numBytes)
			{
				lastElement->textLength-=numBytes;
				if(!lastElement->textLength&&!lastElement->undoLength)	// if this element has become empty, then get rid of it
				{
					CArrayElementsDelete(group->undoCArray,CArrayTotalElementsGet(group->undoCArray)-1,1,NULL,NULL);
				}
				return(true);											// all that was needed was to remove bytes
			}
			else
			{
				startPosition+=bytesToEdge;
				numBytes-=bytesToEdge;
				TextStreamPositionSet(&stream,buffer->textUniverse,startPosition);	// point to bytes about to be deleted

				if(StyleCopy(buffer->styleUniverse,startPosition,group->undoStyles,TextUniverseTotalBytes(group->undoText),numBytes))
				{
					if(TextUniverseInsertFromStream(&stream,group->undoText,TextUniverseTotalBytes(group->undoText),numBytes))	// copy into undo holding area
					{
						lastElement->textLength-=bytesToEdge;
						lastElement->undoLength+=numBytes;
						return(true);																// if we got them, then we are done
					}
					StyleRangeDelete(group->undoStyles,TextUniverseTotalBytes(group->undoText),numBytes);		// kill the styles we copied
				}
				return(false);
			}
		}
		else
		{
			if((startPosition+numBytes)<=(lastElement->textStartPosition+lastElement->textLength))	// make sure it does not span entire element
			{
				bytesToEdge=startPosition+numBytes-lastElement->textStartPosition;
				numBytes-=bytesToEdge;
				TextStreamPositionSet(&stream,buffer->textUniverse,startPosition);					// point to bytes about to be deleted

				if(StyleRangeInsert(group->undoStyles,lastElement->undoStartPosition,numBytes))		// move old styles over to account for new ones
				{
					if(StyleCopy(buffer->styleUniverse,startPosition,group->undoStyles,lastElement->undoStartPosition,numBytes))
					{
						if(TextUniverseInsertFromStream(&stream,group->undoText,lastElement->undoStartPosition,numBytes))	// copy into undo holding area
						{
							lastElement->textStartPosition=startPosition;
							lastElement->textLength-=bytesToEdge;
							lastElement->undoLength+=numBytes;
							return(true);															// if we got them, then we are done
						}
					}
					StyleRangeDelete(group->undoStyles,lastElement->undoStartPosition,numBytes);	// kill the styles we inserted
				}
				return(false);
			}
			else
			{
				return(UndoDeleteElementCreate(buffer,startPosition,numBytes,group));				// this delete spans entire previous element, so make new one
			}
		}
	}
	// if we cannot merge, then create a new element
	return(UndoDeleteElementCreate(buffer,startPosition,numBytes,group));	// delete in a different place
}

static bool DeleteRegister(EDITOR_BUFFER *buffer,uint64_t startPosition,uint64_t numBytes,UNDO_GROUP *group,bool allowMerge)
// Register the delete into the given text buffer, and chunk array.
{
	CARRAY_CHUNK
		*lastUndoChunk;
	UNDO_ELEMENT
		*lastElement;

	if((lastUndoChunk=group->undoCArray->lastChunk))										// is there a current item in the array?
	{
		lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[CArrayChunkElementsGet(lastUndoChunk)-1]);	// point to last element of undo array
		if(allowMerge&&(buffer->undoUniverse->currentFrame==lastElement->frameNumber))		// is it in the current frame?
		{
			return(UndoDeleteElementMerge(buffer,lastElement,startPosition,numBytes,group));
		}
		else
		{
			return(UndoDeleteElementCreate(buffer,startPosition,numBytes,group));
		}
	}
	else
	{
		return(UndoDeleteElementCreate(buffer,startPosition,numBytes,group));
	}
}

bool UndoDeleteRegister(EDITOR_BUFFER *buffer,uint64_t startPosition,uint64_t numBytes,bool allowMerge)
// Call this BEFORE deleting numBytes of text from startPosition of buffer.
// If there is a problem, return false.
// NOTE: for speed, it is the caller's responsibility to determine chunk and offset
// which point to startPosition of buffer.
{
	UNDO_UNIVERSE
		*undoUniverse;

	if(numBytes)
	{
		undoUniverse=buffer->undoUniverse;

		if(!undoUniverse->undoActive)
		{
			if(undoUniverse->cleanElements>CArrayTotalElementsGet(undoUniverse->undoGroup.undoCArray))	// if position which is considered "clean" is beyond where we are now, it becomes meaningless
			{
				undoUniverse->cleanKnown=false;
			}
			UndoClear(&(undoUniverse->redoGroup));	// get rid of redo information
		}
		if(undoUniverse->undoing)						// if undoing, then place undo info into redo buffers
		{
			return(DeleteRegister(buffer,startPosition,numBytes,&(undoUniverse->redoGroup),allowMerge));
		}
		else
		{
			return(DeleteRegister(buffer,startPosition,numBytes,&(undoUniverse->undoGroup),allowMerge));
		}
	}
	else
	{
		return(true);
	}
}

static bool UndoInsertElementCreate(EDITOR_BUFFER *buffer,uint64_t startPosition,uint64_t numBytes,UNDO_GROUP *group)
// Create a new element at the end of the undo array.
// If there is a problem, set the error, and return false.
{
	UNDO_ELEMENT
		undoElement;

	if(!buffer->undoUniverse->groupingDepth)							// if not currently in a group, then make this a new frame
	{
		buffer->undoUniverse->currentFrame++;							// bump the frame number
	}

	undoElement.frameNumber=buffer->undoUniverse->currentFrame;			// set the frame number to the current undo frame
	undoElement.textStartPosition=startPosition;						// when the text is blasted back into place, this is where it goes
	undoElement.textLength=numBytes;									// number of bytes to overwrite (all that is being inserted)
	undoElement.undoStartPosition=TextUniverseTotalBytes(group->undoText);	// start at what is now the current end of the undo text
	undoElement.undoLength=0;											// number of bytes to put back (none, we are inserting)
	if(CArrayRawInsert(&undoElement,group->undoCArray,CArrayTotalElementsGet(group->undoCArray),1,NULL,NULL))	// insert the element at the end
	{
		return(true);
	}
	return(false);
}

static bool UndoInsertElementMerge(EDITOR_BUFFER *buffer,UNDO_ELEMENT *lastElement,uint64_t startPosition,uint64_t numBytes,UNDO_GROUP *group)
// An insert is happening, and the last element of the undo list is in the current
// frame. See if the insert can be combined with the last element of the undo list,
// or if another entirely new element needs to be created.
// If there is a problem, set the error, and return false.
{
	if(startPosition==lastElement->textStartPosition+lastElement->textLength)	// see if insert just after end of current start
	{
		lastElement->textLength+=numBytes;									// add these bytes to what should be deleted during undo
		return(true);
	}
	// if we cannot merge, then create a new element
	return(UndoInsertElementCreate(buffer,startPosition,numBytes,group));	// insert in a different place
}

static bool InsertRegister(EDITOR_BUFFER *buffer,uint64_t startPosition,uint64_t numBytes,UNDO_GROUP *group,bool allowMerge)
// Register the insert into the given text buffer, and chunk array.
// If there is a problem, set the error, and return false.
{
	CARRAY_CHUNK
		*lastUndoChunk;
	UNDO_ELEMENT
		*lastElement;

	if((lastUndoChunk=group->undoCArray->lastChunk))										// is there a current item in the array?
	{
		lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[CArrayChunkElementsGet(lastUndoChunk)-1]);	// point to last element of undo array
		if(allowMerge&&(buffer->undoUniverse->currentFrame==lastElement->frameNumber))		// is it in the current frame?
		{
			return(UndoInsertElementMerge(buffer,lastElement,startPosition,numBytes,group));
		}
		else
		{
			return(UndoInsertElementCreate(buffer,startPosition,numBytes,group));
		}
	}
	else
	{
		return(UndoInsertElementCreate(buffer,startPosition,numBytes,group));
	}
}

bool UndoInsertRegister(EDITOR_BUFFER *buffer,uint64_t startPosition,uint64_t numBytes,bool allowMerge)
// Call this AFTER inserting numBytes of text into buffer at startPosition.
// If there is a problem, set the error, return false.
{
	UNDO_UNIVERSE
		*undoUniverse;

	if(numBytes)
	{
		undoUniverse=buffer->undoUniverse;

		if(!undoUniverse->undoActive)
		{
			if(undoUniverse->cleanElements>CArrayTotalElementsGet(undoUniverse->undoGroup.undoCArray))	// if position which is considered "clean" is beyond where we are now, it becomes meaningless
			{
				undoUniverse->cleanKnown=false;
			}
			UndoClear(&(undoUniverse->redoGroup));		// get rid of redo information
		}
		if(undoUniverse->undoing)						// if undoing, then place undo info into redo buffers
		{
			return(InsertRegister(buffer,startPosition,numBytes,&(undoUniverse->redoGroup),allowMerge));
		}
		else
		{
			return(InsertRegister(buffer,startPosition,numBytes,&(undoUniverse->undoGroup),allowMerge));
		}
	}
	else
	{
		return(true);
	}
}

static bool UndoSelectionChange(SELECTION_UNIVERSE *selectionUniverse,uint64_t position,uint64_t numBytesOut,uint64_t numBytesIn)
// Handle updating the selection, an creating new selections during undo.
// NOTE: if something goes wrong, attempt to leave the selection universe in a sane state.
{
	SelectionAdjustForChange(selectionUniverse,position,position+numBytesOut,position+numBytesIn);
	if(SelectionRangeSet(selectionUniverse,position,numBytesIn))
	{
		return(true);
	}
	return(false);
}

static bool Undo(EDITOR_BUFFER *buffer,UNDO_GROUP *group,bool *didSomething)
// Undo from the given text and array into buffer.
// NOTE: if there is nothing to undo, return false in *didSomething
// else return true.
// If something bad happens while trying to undo, set the error, and return false.
{
	SELECTION_UNIVERSE
		*selectionUniverse,
		*newSelectionUniverse;
	CARRAY_CHUNK
		*lastUndoChunk;
	UNDO_ELEMENT
		*lastElement;
	uint64_t
		frameToUse;
	bool
		haveMinimumPosition;
	uint64_t
		minimumPosition;													// used to set the cursor position after the change
	TEXT_STREAM
		stream;
	bool
		fail;

	fail=*didSomething=false;
	haveMinimumPosition=false;
	minimumPosition=0;
	if((lastUndoChunk=group->undoCArray->lastChunk))						// is there a current item in the array?
	{
		lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[CArrayChunkElementsGet(lastUndoChunk)-1]);	// point to last element of undo array
		frameToUse=lastElement->frameNumber;								// undo all elements of this frame

		if((newSelectionUniverse=SelectionUniverseOpen()))					// create a new selection universe where the output selection will be created
		{
			selectionUniverse=buffer->selectionUniverse;					// point to current selection universe
			EditorSelectionChangeStart(buffer,selectionUniverse);
			SelectionDeleteAll(selectionUniverse);							// remove all old selections
			EditorSelectionChangeEnd(buffer,selectionUniverse);

			EditorReplaceStart(buffer);
			UndoGroupBegin(buffer);											// group the redo of this undo
			while(!fail&&lastElement&&lastElement->frameNumber==frameToUse)
			{
				TextStreamPositionSet(&stream,group->undoText,lastElement->undoStartPosition);	// point to bytes about to be placed into the text

				if(EditorTextReplaceWithStream(buffer,lastElement->textStartPosition,lastElement->textLength,&stream,lastElement->undoLength,false))
				{
					if(StyleCopy(group->undoStyles,lastElement->undoStartPosition,buffer->styleUniverse,lastElement->textStartPosition,lastElement->undoLength))
					{
						if(UndoSelectionChange(newSelectionUniverse,lastElement->textStartPosition,lastElement->textLength,lastElement->undoLength))
						{
							if(haveMinimumPosition)
							{
								if(lastElement->textStartPosition<minimumPosition)
								{
									minimumPosition=lastElement->textStartPosition;
								}
							}
							else
							{
								minimumPosition=lastElement->textStartPosition;
								haveMinimumPosition=true;
							}

							StyleRangeDelete(group->undoStyles,lastElement->undoStartPosition,lastElement->undoLength);
							TextUniverseDelete(group->undoText,lastElement->undoStartPosition,lastElement->undoLength);
							CArrayElementsDelete(group->undoCArray,CArrayTotalElementsGet(group->undoCArray)-1,1,NULL,NULL);
							if((lastUndoChunk=group->undoCArray->lastChunk))	// fetch the last element again
							{
								lastElement=&(((UNDO_ELEMENT *)lastUndoChunk->data)[CArrayChunkElementsGet(lastUndoChunk)-1]);	// point to last element of undo array
							}
							else
							{
								lastElement=NULL;
							}
						}
						else
						{
							fail=true;
						}
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
			}

			EditorReplaceEnd(buffer);
			UndoGroupEndStrict(buffer);										// end this group, do not let anything combine with it

			// place new selection into the universe

			EditorSelectionChangeStart(buffer,buffer->selectionUniverse);
			SelectionCursorPositionSet(newSelectionUniverse,minimumPosition);	// place cursor at the start
			buffer->selectionUniverse=newSelectionUniverse;
			SelectionUniverseClose(selectionUniverse);						// close old universe
			EditorSelectionChangeEnd(buffer,buffer->selectionUniverse);

			*didSomething=true;
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

bool EditorUndo(EDITOR_BUFFER *buffer,bool *didSomething)
// Undo whatever was done last (if there is nothing in the undo buffer
// then do not do anything).
// If there is no undo information, return false.
// otherwise return true.
// If there is a problem, set the error, return false.
{
	bool
		result;

	*didSomething=false;
	PlatformBusyShow();
	buffer->undoUniverse->undoActive=true;							// changes are coming from undo/redo, not user
	buffer->undoUniverse->undoing=true;								// remember that we are undoing so if register calls come in, we can place them in redo buffer
	result=Undo(buffer,&(buffer->undoUniverse->undoGroup),didSomething);
	buffer->undoUniverse->undoing=false;
	buffer->undoUniverse->undoActive=false;
	PlatformBusyUnShow();
	return(result);
}

bool EditorRedo(EDITOR_BUFFER *buffer,bool *didSomething)
// Redo whatever was undone last.
// If there is no undo information, return false in *didSomething
// otherwise return true.
// If there is a problem, set the error, return false.
{
	bool
		result;

	*didSomething=false;
	PlatformBusyShow();
	buffer->undoUniverse->undoActive=true;							// changes are coming from undo/redo, not user
	result=Undo(buffer,&(buffer->undoUniverse->redoGroup),didSomething);
	buffer->undoUniverse->undoActive=false;
	PlatformBusyUnShow();
	return(result);
}

bool EditorUndoToggle(EDITOR_BUFFER *buffer,bool *didSomething)
// If there is a redo that can be done, then do it, otherwise, do an undo.
// If neither can be done, return didSomething=false.
// If there is a problem, set the error, return false.
{
	bool
		result;

	if((result=EditorRedo(buffer,didSomething)))	// try to redo
	{
		if(!(*didSomething))						// if nothing to redo, then try to undo
		{
			return(EditorUndo(buffer,didSomething));
		}
	}
	return(result);
}

void EditorUndoClear(EDITOR_BUFFER *buffer)
// When, undo/redo must be cleared, call this to do it.
{
	if(buffer->undoUniverse->cleanKnown)				// if clean position is known, then make sure we are there, otherwise we don't know it anymore
	{
		if(buffer->undoUniverse->cleanElements!=CArrayTotalElementsGet(buffer->undoUniverse->undoGroup.undoCArray))
		{
			buffer->undoUniverse->cleanKnown=false;
		}
	}
	buffer->undoUniverse->cleanElements=0;				// clean position moves to 0

	UndoClear(&(buffer->undoUniverse->undoGroup));		// get rid of undo information
	UndoClear(&(buffer->undoUniverse->redoGroup));		// get rid of redo information
	buffer->undoUniverse->currentFrame=0;				// reset the frame number
}

void UndoCleanPointSet(EDITOR_BUFFER *buffer)
// Mark the current position in the undo data as the buffer "clean" position.
// NOTE: this will force a break in the undo flow if needed.
{
	buffer->undoUniverse->currentFrame++;				// bump the frame number to force a break in the undo flow
	buffer->undoUniverse->cleanKnown=true;
	buffer->undoUniverse->cleanElements=CArrayTotalElementsGet(buffer->undoUniverse->undoGroup.undoCArray);
	ViewsMetaChangeCallbacks(buffer);					// inform views that meta information about the buffer has changed (dirty flag cleared)
}

bool UndoCleanPointAt(EDITOR_BUFFER *buffer)
// Tells if the buffer is at the "clean" point in the undo flow.
{
	if(buffer->undoUniverse->cleanKnown&&buffer->undoUniverse->cleanElements==CArrayTotalElementsGet(buffer->undoUniverse->undoGroup.undoCArray))
	{
		return(true);
	}
	return(false);
}

static void UndoGroupUnInit(UNDO_GROUP *group)
// Undo what UndoGroupInit did.
{
	StyleUniverseClose(group->undoStyles);
	TextUniverseClose(group->undoText);
	CArrayDispose(group->undoCArray);
}

static bool UndoGroupInit(UNDO_GROUP *group)
// Initialize an undo group.
{
	if((group->undoCArray=CArrayCreate(UNDO_CHUNK_ELEMENTS,sizeof(UNDO_ELEMENT),UndoElementsMove,CA_NUM_INHERENT_PARAMS,0)))
	{
		if((group->undoText=TextUniverseOpen()))
		{
			if((group->undoStyles=StyleUniverseOpen()))
			{
				return(true);
			}
			TextUniverseClose(group->undoText);
		}
		CArrayDispose(group->undoCArray);
	}
	return(false);
}

void UndoUniverseClose(UNDO_UNIVERSE *universe)
// Dispose of an undo universe.
{
	UndoGroupUnInit(&(universe->redoGroup));
	UndoGroupUnInit(&(universe->undoGroup));
	MDisposePtr(universe);
}

UNDO_UNIVERSE *UndoUniverseOpen()
// Open an undo universe.
// If there is a problem, set the error, and return NULL.
{
	UNDO_UNIVERSE
		*universe;

	if((universe=(UNDO_UNIVERSE *)MNewPtr(sizeof(UNDO_UNIVERSE))))
	{
		universe->currentFrame=0;
		universe->groupingDepth=0;
		universe->cleanElements=0;
		universe->cleanKnown=true;
		universe->undoActive=false;
		universe->undoing=false;
		if(UndoGroupInit(&(universe->undoGroup)))
		{
			if(UndoGroupInit(&(universe->redoGroup)))
			{
				return(universe);
			}
			UndoGroupUnInit(&(universe->undoGroup));
		}
		MDisposePtr(universe);
	}
	return(NULL);
}
