// UTF-8 code point decoder
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static bool Utf8CodePointExtract(TEXT_STREAM *stream,uint8_t startCode,uint32_t *codePoint)
// Passed a UTF-8 start code, and a stream which points to
// the potential data bytes for the start code, attempt to extract
// the data bytes and assemble codePoint. If successful, return true with
// stream updated to point past the last data byte. If unsuccessful,
// return false with stream left pointing past the last byte read,
// and codePoint unmodified.
// NOTE: this code explicitly disallows over-long sequences (sequences
// that generate code points which could have been generated with a
// smaller sequence). This is done so that other code which relies on
// spotting control characters directly (like new lines) won't be in
// conflict with high level code which is interpreting just the UTF-8
// output.
{
	uint32_t
		codePointAccumulator,
		minCodePoint;
	uint8_t
		value;
	uint8_t
		bytesRemaining;

	if(!(startCode&0b00100000))
	{
		codePointAccumulator=startCode&0b00011111;
		minCodePoint=0x80;
		bytesRemaining=1;
	}
	else if(!(startCode&0b00010000))
	{
		codePointAccumulator=startCode&0b00001111;
		minCodePoint=0x800;
		bytesRemaining=2;
	}
	else if(!(startCode&0b00001000))
	{
		codePointAccumulator=startCode&0b00000111;
		minCodePoint=0x10000;
		bytesRemaining=3;
	}
	else
	{
		return(false);							// not a valid start code (UTF-8 only codes for 21 bits)
	}
	while(bytesRemaining&&!TextStreamAtEnd(stream))
	{
		value=TextStreamChar(stream);			// fetch the next piece
		TextStreamNextChar(stream);
		if((value&0b11000000)==0b10000000)		// valid continuation?
		{
			codePointAccumulator=(codePointAccumulator<<6)|(value&0b00111111);	// accumulate it
			bytesRemaining--;
		}
		else
		{
			break;								// got bad value, exit the loop
		}
	}
	if(bytesRemaining||(codePointAccumulator<minCodePoint))		// if bytes remaining, or over-long sequence, reject
	{
		return(false);							// report failure to extract
	}
	*codePoint=codePointAccumulator;			// set the return result
	return(true);								// report success
}

uint32_t Utf8NextCodePoint(TEXT_STREAM *stream)
// Read the next UTF-8 code point from the passed (synchronized) stream (which must not be at the end).
// If this encounters something which doesn't look like UTF-8, it'll return the
// first 8-bit value it runs into.
// stream will be left pointing to the byte after the code point read.
{
	uint32_t
		codePoint;
	TEXT_STREAM
		savedStream;

	codePoint=TextStreamChar(stream);
	TextStreamNextChar(stream);				
	if((codePoint&0b11000000)==0b11000000)		// did we get a UTF-8 start code?
	{
		savedStream=*stream;						// make a copy of the stream in case this goes wrong
		if(!Utf8CodePointExtract(stream,codePoint,&codePoint))	// try to extract a codePoint from here
		{
			*stream=savedStream;					// put the stream back where it was at the start (codePoint is still the start code)
		}
	}
	return(codePoint);
}

bool Utf8LimitedNextCodePoint(TEXT_STREAM *stream,uint64_t maxPosition,uint32_t *codePoint)
// Read the next UTF-8 code point from the passed (synchronized) stream (which must not be at the end).
// If this encounters something which doesn't look like UTF-8, it'll return the
// first 8-bit value it runs into.
// stream will be left pointing to the byte after the code point read.
// NOTE: this will fail if the stream moves beyond maxPosition.
{
	uint32_t
		tempCodePoint;
	TEXT_STREAM
		savedStream;

	if(TextStreamPositionGet(stream)<maxPosition)
	{
		tempCodePoint=TextStreamChar(stream);
		if((tempCodePoint&0b11000000)==0b11000000)		// did we get a UTF-8 start code?
		{
			savedStream=*stream;						// make a copy of the stream in case this goes wrong
			TextStreamNextChar(stream);					// step over the start code
			if(Utf8CodePointExtract(stream,tempCodePoint,&tempCodePoint))	// try to extract a codePoint from here
			{
				// got a valid UTF-8 code point, make sure it didn't cause us to move beyond maxPosition
				if(TextStreamPositionGet(stream)>maxPosition)	// code went past the maxPosition, so don't read anything
				{
					*stream=savedStream;				// put the stream back where it was at the start
					return(false);
				}
			}
			else	// could not extact a valid code point, so just return the start code as the code point
			{
				*stream=savedStream;					// put the stream back where it was at the start (tempCodePoint is still the start code)
				TextStreamNextChar(stream);				// step over the start code
			}
		}
		else
		{
			TextStreamNextChar(stream);				
		}
		*codePoint=tempCodePoint;
		return(true);
	}
	return(false);
}

void Utf8Synchronize(TEXT_STREAM *stream)
// Move stream back to the start of the code point which overlaps the current position.
// This only does any real work if the character at the current position is a continuation
// character.
{
	uint8_t
		byte;
	uint32_t
		codePoint;
	TEXT_STREAM
		savedStream,
		testStream;
	uint32_t
		distance;

	byte=TextStreamChar(stream);				// read the byte at the current position
	if((byte&0b11000000)==0b10000000)			// only ever do any work if we start in a continuation character
	{
		savedStream=*stream;					// copy the stream in case we discover the sequence is invalid
		distance=0;								// keeps track of how far we've moved back
		while((distance<3)&&TextStreamPreviousChar(stream)&&(((byte=TextStreamChar(stream))&0b11000000)==0b10000000))
		{
			distance++;							// keep moving back
		}
		if((byte&0b11000000)==0b11000000)		// successfully found a start code?
		{
			testStream=*stream;					// read forward from the test stream
			TextStreamNextChar(&testStream);	// skip over the start code
			if(Utf8CodePointExtract(&testStream,byte,&codePoint))
			{
				if(TextStreamPositionGet(&testStream)>TextStreamPositionGet(&savedStream))
				{
					// success!
				}
				else
				{
					*stream=savedStream;		// start code didn't use the continuation code we found
				}
			}
			else
			{
				*stream=savedStream;			// failed to extract a sequence, point to original character
			}
		}
		else
		{
			*stream=savedStream;				// failed to locate start, point to original character
		}
	}
}
