// Buffer variable binding interface
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// This file handles setting variables onto buffers.

#include	"includes.h"

typedef struct variableBinding
{
	struct variableBinding
		*previousBinding,					// pointer to previous binding in the global binding list
		*nextBinding,						// pointer to next binding in the global binding list (in user defined order)
		*nextHashBinding;					// pointer to next binding in local hash list
	char
		*variableName,						// name of this variable
		*variableText;						// data contained in this variable
} VARIABLE_BINDING;

static VARIABLE_BINDING **BindingHashListHeadGet(VARIABLE_TABLE *variableTable,char *variableName)
// Perform hashing function on variableName, return pointer to list head
// for bindings.
{
	uint32_t
		count,
		length,
		sum;

	length=strlen(variableName);
	sum=0;
	for(count=0;count<length;count++)			// very crude hash function
	{
		sum+=variableName[count];
	}
	return(&(variableTable->variableBindingTable[sum&0xFF]));
}

char *VariableBindingNameGet(VARIABLE_BINDING *binding)
// Return a pointer to the variable binding name for binding.
{
	return(binding->variableName);
}

char *VariableBindingTextGet(VARIABLE_BINDING *binding)
// Return a pointer to the variable binding text for binding.
{
	return(binding->variableText);
}

static bool VariableBindingLocationLocate(VARIABLE_TABLE *variableTable,char *variableName,VARIABLE_BINDING ***hashListHead,VARIABLE_BINDING **previousHashEntry,VARIABLE_BINDING **entry)
// Locate a variable binding in the binding table, return pointers to important entries.
{
	bool
		found;
	VARIABLE_BINDING
		*currentEntry;

	currentEntry=*(*hashListHead=BindingHashListHeadGet(variableTable,variableName));	// point to the head of the linked list at this hash entry
	*previousHashEntry=NULL;
	found=false;
	while(currentEntry&&!found)
	{
		if(strcmp(variableName,currentEntry->variableName)==0)
		{
			found=true;
			*entry=currentEntry;
		}
		else
		{
			*previousHashEntry=currentEntry;
			currentEntry=currentEntry->nextHashBinding;
		}
	}
	return(found);
}

VARIABLE_BINDING *VariableBindingNextLocate(VARIABLE_BINDING *currentBinding)
// Return the next variable binding from the current one.
// If there is none, return NULL.
{
	return(currentBinding->nextBinding);
}

VARIABLE_BINDING *VariableBindingLocate(VARIABLE_TABLE *variableTable,char *variableName)
// Find a key binding that matches the variableName, and return it.
// If there is none, return NULL.
{
	bool
		found;
	VARIABLE_BINDING
		*currentEntry;

	currentEntry=*BindingHashListHeadGet(variableTable,variableName);	// point to the head of the linked list at this hash entry
	found=false;
	while(currentEntry&&!found)
	{
		if(strcmp(variableName,currentEntry->variableName)==0)
		{
			found=true;
		}
		else
		{
			currentEntry=currentEntry->nextHashBinding;
		}
	}
	return(currentEntry);
}

bool VariableBindingDelete(VARIABLE_TABLE *variableTable,char *variableName)
// Attempt to locate the given binding in variableTable, and
// remove it if it is found.
// If it is not located, return false.
{
	VARIABLE_BINDING
		**hashListHead,
		*previousHashEntry,
		*entry;

	if(VariableBindingLocationLocate(variableTable,variableName,&hashListHead,&previousHashEntry,&entry))
	{
		if(previousHashEntry)
		{
			previousHashEntry->nextHashBinding=entry->nextHashBinding;		// unlink the located entry
		}
		else
		{
			*hashListHead=entry->nextHashBinding;
		}
		if(entry->previousBinding)
		{
			if((entry->previousBinding->nextBinding=entry->nextBinding))
			{
				entry->nextBinding->previousBinding=entry->previousBinding;
			}
		}
		else
		{
			if((variableTable->variableBindingListHead=entry->nextBinding))
			{
				variableTable->variableBindingListHead->previousBinding=NULL;
			}
		}
		MDisposePtr(entry->variableName);
		MDisposePtr(entry->variableText);
		MDisposePtr(entry);
		return(true);
	}
	return(false);
}

bool VariableBindingCreate(VARIABLE_TABLE *variableTable,char *variableName,char *variableText)
// Search the variable binding table for an entry that matches the one given, and delete it if
// it exists. Then add this entry to the table.
// If there is a problem, ErrorSet, and return false.
{
	VARIABLE_BINDING
		**hashListHead,
		*newEntry;

	VariableBindingDelete(variableTable,variableName);					// try to delete one if it exists
	hashListHead=BindingHashListHeadGet(variableTable,variableName);	// point to head of list in hash table where this binding will go
	if((newEntry=(VARIABLE_BINDING *)MNewPtr(sizeof(VARIABLE_BINDING))))
	{
		if((newEntry->variableName=(char *)MNewPtr(strlen(variableName)+1)))
		{
			if((newEntry->variableText=(char *)MNewPtr(strlen(variableText)+1)))
			{
				strcpy(newEntry->variableName,variableName);			// copy in name
				strcpy(newEntry->variableText,variableText);			// copy in text
				newEntry->nextHashBinding=*hashListHead;
				*hashListHead=newEntry;									// link to head of hash list
				newEntry->previousBinding=NULL;
				if((newEntry->nextBinding=variableTable->variableBindingListHead))	// point to the next entry
				{
					variableTable->variableBindingListHead->previousBinding=newEntry;
				}
				variableTable->variableBindingListHead=newEntry;
				return(true);
			}
			MDisposePtr(newEntry->variableName);
		}
		MDisposePtr(newEntry);
	}
	return(false);
}

bool VariableBindingTableInit(VARIABLE_TABLE *variableTable)
// Clear the variable binding table so that there are no bindings present.
// If there is a problem, ErrorSet, and return false.
{
	int
		i;

	for(i=0;i<256;i++)
	{
		variableTable->variableBindingTable[i]=NULL;
	}
	variableTable->variableBindingListHead=NULL;
	return(true);
}

void VariableBindingTableUnInit(VARIABLE_TABLE *variableTable)
// Remove all variable bindings from the bindings table.
{
	int
		i;
	VARIABLE_BINDING
		*currentEntry,
		*nextEntry;

	for(i=0;i<256;i++)
	{
		currentEntry=variableTable->variableBindingTable[i];
		while(currentEntry)
		{
			nextEntry=currentEntry->nextHashBinding;
			MDisposePtr(currentEntry->variableName);
			MDisposePtr(currentEntry->variableText);
			MDisposePtr(currentEntry);
			currentEntry=nextEntry;
		}
		variableTable->variableBindingTable[i]=NULL;
	}
	variableTable->variableBindingListHead=NULL;
}
