// View color scheme functions
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// All views in e93 are drawn with a "color scheme".
// A color scheme tells which fonts and colors the view should be
// drawn in, including things like colors for all the styles, and
// the selection colors.
// Each view has its own (which can be modified independently from
// all others). Further, lists of named color schemes exist which
// can be used as templates for new views when they are created.
// These lists can be created from Tcl.

#include	"includes.h"

typedef struct
{
	bool
		fontBold,								// tells if font should be bold
		fontOblique;							// tells if font should be oblique
	EDITOR_COLOR
		activeBackgroundColor,					// color to use when drawing background of active view (if 0x00000000, color scheme's background will be used)
		activeForegroundColor,					// color to use when drawing foreground of active view (if 0x00000000, color scheme's foreground will be used)
		inactiveBackgroundColor,				// color to use when drawing background of inactive view (if 0x00000000, active color will be used)
		inactiveForegroundColor;				// color to use when drawing foreground of inactive view (if 0x00000000, active color will be used)
} STYLE_COLOR_SCHEME;

typedef struct viewColorScheme
{
	NAMED_LIST_ITEM
		*listItem;								// points back to the list item for the named list this scheme is on (NULL if it is on no list)
	EDITOR_COLOR
		activeBackgroundColor,					// if nothing else overrides it, the background color for the view when it is active
		activeForegroundColor,					// if nothing else overrides it, the foreground color of the text in the view when the view is active
		inactiveBackgroundColor,				// if nothing else overrides it, the background color for the view when it is inactive
		inactiveForegroundColor;				// if nothing else overrides it, the foreground color of the text in the view when the view is inactive
	EDITOR_COLOR
		activeSelectionBackgroundColor,			// color to use when drawing active selection background (if 0x00000000, the view's foreground will be used)
		activeSelectionForegroundColor,			// color to use when drawing active selection foreground (if 0x00000000, leave selected foreground unmodified)
		inactiveSelectionBackgroundColor,		// color to use when drawing inactive selection background (if 0x00000000, active color will be used)
		inactiveSelectionForegroundColor;		// color to use when drawing inactive selection foreground (if 0x00000000, active color will be used)
	STYLE_COLOR_SCHEME
		styleColorScheme[VIEW_MAX_STYLES];		// array of style color schemes
} VIEW_COLOR_SCHEME;

static NAMED_LIST
	*viewColorSchemeList;						// the list of color schemes

// ----------------------------------------------------------------------------

void ViewColorSchemeStyleDrawColorsGet(const VIEW_COLOR_SCHEME *colorScheme,uint32_t style,bool active,EDITOR_COLOR *backgroundColor,EDITOR_COLOR *foregroundColor)
// Get the active, or inactive background and foreground colors for the given style of the given colorScheme.
{
	if(active)
	{
		*backgroundColor=colorScheme->styleColorScheme[style].activeBackgroundColor;
		*foregroundColor=colorScheme->styleColorScheme[style].activeForegroundColor;
	}
	else
	{
		*backgroundColor=colorScheme->styleColorScheme[style].inactiveBackgroundColor;
		*foregroundColor=colorScheme->styleColorScheme[style].inactiveForegroundColor;
	}
}

void ViewColorSchemeStyleDrawColorsSet(VIEW_COLOR_SCHEME *colorScheme,uint32_t style,bool active,EDITOR_COLOR backgroundColor,EDITOR_COLOR foregroundColor)
// Set the active, or inactive background and foreground colors for the given style of the given colorScheme.
{
	if(active)
	{
		colorScheme->styleColorScheme[style].activeBackgroundColor=backgroundColor;
		colorScheme->styleColorScheme[style].activeForegroundColor=foregroundColor;
	}
	else
	{
		colorScheme->styleColorScheme[style].inactiveBackgroundColor=backgroundColor;
		colorScheme->styleColorScheme[style].inactiveForegroundColor=foregroundColor;
	}
}

void ViewColorSchemeStyleFontAttributesGet(const VIEW_COLOR_SCHEME *colorScheme,uint32_t style,bool *bold,bool *oblique)
// Return the font attributes associated with the given style.
{
	(*bold)=colorScheme->styleColorScheme[style].fontBold;
	(*oblique)=colorScheme->styleColorScheme[style].fontOblique;
}

void ViewColorSchemeStyleFontAttributesSet(VIEW_COLOR_SCHEME *colorScheme,uint32_t style,bool bold,bool oblique)
// Set the font attributes for the given style of the given colorScheme.
{
	colorScheme->styleColorScheme[style].fontBold=bold;
	colorScheme->styleColorScheme[style].fontOblique=oblique;
}

void ViewColorSchemeSelectionColorsGet(const VIEW_COLOR_SCHEME *colorScheme,bool active,EDITOR_COLOR *backgroundColor,EDITOR_COLOR *foregroundColor)
// Get the active, or inactive selection background and foreground colors for the given colorScheme.
{
	if(active)
	{
		*backgroundColor=colorScheme->activeSelectionBackgroundColor;
		*foregroundColor=colorScheme->activeSelectionForegroundColor;
	}
	else
	{
		*backgroundColor=colorScheme->inactiveSelectionBackgroundColor;
		*foregroundColor=colorScheme->inactiveSelectionForegroundColor;
	}
}

void ViewColorSchemeSelectionColorsSet(VIEW_COLOR_SCHEME *colorScheme,bool active,EDITOR_COLOR backgroundColor,EDITOR_COLOR foregroundColor)
// Set the active, or inactive selection background and foreground colors for the given colorScheme.
{
	if(active)
	{
		colorScheme->activeSelectionBackgroundColor=backgroundColor;
		colorScheme->activeSelectionForegroundColor=foregroundColor;
	}
	else
	{
		colorScheme->inactiveSelectionBackgroundColor=backgroundColor;
		colorScheme->inactiveSelectionForegroundColor=foregroundColor;
	}
}

void ViewColorSchemeDrawColorsGet(const VIEW_COLOR_SCHEME *colorScheme,bool active,EDITOR_COLOR *backgroundColor,EDITOR_COLOR *foregroundColor)
// Get the active, or inactive background and foreground colors for the given colorScheme.
{
	if(active)
	{
		*backgroundColor=colorScheme->activeBackgroundColor;
		*foregroundColor=colorScheme->activeForegroundColor;
	}
	else
	{
		*backgroundColor=colorScheme->inactiveBackgroundColor;
		*foregroundColor=colorScheme->inactiveForegroundColor;
	}
}

void ViewColorSchemeDrawColorsSet(VIEW_COLOR_SCHEME *colorScheme,bool active,EDITOR_COLOR backgroundColor,EDITOR_COLOR foregroundColor)
// Set the active, or inactive background and foreground colors for the given colorScheme.
{
	if(active)
	{
		colorScheme->activeBackgroundColor=backgroundColor;
		colorScheme->activeForegroundColor=foregroundColor;
	}
	else
	{
		colorScheme->inactiveBackgroundColor=backgroundColor;
		colorScheme->inactiveForegroundColor=foregroundColor;
	}
}

void ViewColorSchemeDestroy(VIEW_COLOR_SCHEME *colorScheme)
// Destroy a color scheme created by ViewColorSchemeCreate.
{
	MDisposePtr(colorScheme);
}

VIEW_COLOR_SCHEME *ViewColorSchemeCreate()
// Create a color scheme with no fonts selected, and all colors reset to 0x00000000.
// If there's a problem, ErrorSet and return NULL.
{
	VIEW_COLOR_SCHEME
		*colorScheme;
	uint32_t
		i;

	if((colorScheme=(VIEW_COLOR_SCHEME *)MNewPtr(sizeof(VIEW_COLOR_SCHEME))))
	{
		colorScheme->listItem=NULL;

		colorScheme->activeBackgroundColor=0x00000000;
		colorScheme->activeForegroundColor=0x00000000;
		colorScheme->inactiveBackgroundColor=0x00000000;
		colorScheme->inactiveForegroundColor=0x00000000;

		colorScheme->activeSelectionBackgroundColor=0x00000000;
		colorScheme->activeSelectionForegroundColor=0x00000000;
		colorScheme->inactiveSelectionBackgroundColor=0x00000000;
		colorScheme->inactiveSelectionForegroundColor=0x00000000;

		for(i=0;i<VIEW_MAX_STYLES;i++)
		{
			colorScheme->styleColorScheme[i].fontBold=false;
			colorScheme->styleColorScheme[i].fontOblique=false;

			colorScheme->styleColorScheme[i].activeBackgroundColor=0x00000000;
			colorScheme->styleColorScheme[i].activeForegroundColor=0x00000000;
			colorScheme->styleColorScheme[i].inactiveBackgroundColor=0x00000000;
			colorScheme->styleColorScheme[i].inactiveForegroundColor=0x00000000;
		}
	}
	return(colorScheme);
}

VIEW_COLOR_SCHEME *ViewColorSchemeDuplicate(const VIEW_COLOR_SCHEME *original)
// Create a color scheme which is a copy of original.
// If there's a problem, ErrorSet and return NULL.
// The returned color scheme must eventually be destroyed by calling
// ViewColorSchemeDestroy.
{
	VIEW_COLOR_SCHEME
		*duplicate;
	uint32_t
		i;

	if((duplicate=ViewColorSchemeCreate()))
	{
		duplicate->activeBackgroundColor=original->activeBackgroundColor;
		duplicate->activeForegroundColor=original->activeForegroundColor;
		duplicate->inactiveBackgroundColor=original->inactiveBackgroundColor;
		duplicate->inactiveForegroundColor=original->inactiveForegroundColor;

		duplicate->activeSelectionBackgroundColor=original->activeSelectionBackgroundColor;
		duplicate->activeSelectionForegroundColor=original->activeSelectionForegroundColor;
		duplicate->inactiveSelectionBackgroundColor=original->inactiveSelectionBackgroundColor;
		duplicate->inactiveSelectionForegroundColor=original->inactiveSelectionForegroundColor;

		for(i=0;i<VIEW_MAX_STYLES;i++)
		{
			duplicate->styleColorScheme[i].fontBold=original->styleColorScheme[i].fontBold;
			duplicate->styleColorScheme[i].fontOblique=original->styleColorScheme[i].fontOblique;

			duplicate->styleColorScheme[i].activeBackgroundColor=original->styleColorScheme[i].activeBackgroundColor;
			duplicate->styleColorScheme[i].activeForegroundColor=original->styleColorScheme[i].activeForegroundColor;
			duplicate->styleColorScheme[i].inactiveBackgroundColor=original->styleColorScheme[i].inactiveBackgroundColor;
			duplicate->styleColorScheme[i].inactiveForegroundColor=original->styleColorScheme[i].inactiveForegroundColor;
		}
		return(duplicate);
	}
	return(NULL);
}

const char *ViewColorSchemeListNameGet(VIEW_COLOR_SCHEME *colorScheme)
// Return the name of the color scheme. If the scheme is on no list,
// return NULL.
{
	if(colorScheme->listItem)
	{
		return(NamedListItemNameGet(colorScheme->listItem));
	}
	return(NULL);
}

VIEW_COLOR_SCHEME *ViewColorSchemeListLocate(const char *name)
// Locate the color scheme with the given name and return it.
// If none can be located, return NULL.
{
	NAMED_LIST_ITEM
		*listItem;

	if((listItem=NamedListItemLocateName(viewColorSchemeList,name)))
	{
		return((VIEW_COLOR_SCHEME *)NamedListItemDataGet(listItem));
	}
	return(NULL);
}

VIEW_COLOR_SCHEME *ViewColorSchemeListLocateNext(VIEW_COLOR_SCHEME *colorScheme)
// Return the next color scheme on the global list.
// If there is no next item, return NULL.
{
	NAMED_LIST_ITEM
		*listItem;

	if((listItem=NamedListItemLocateNext(colorScheme->listItem)))
	{
		return((VIEW_COLOR_SCHEME *)NamedListItemDataGet(listItem));
	}
	return(NULL);
}

VIEW_COLOR_SCHEME *ViewColorSchemeListLocateFirst()
// Return the first color scheme on the global list.
// If there are no color schemes on the list, NULL will be returned.
{
	NAMED_LIST_ITEM
		*listItem;

	if((listItem=NamedListItemLocateFirst(viewColorSchemeList)))
	{
		return((VIEW_COLOR_SCHEME *)NamedListItemDataGet(listItem));
	}
	return(NULL);
}

void ViewColorSchemeListRemove(VIEW_COLOR_SCHEME *colorScheme)
// Remove colorScheme from the named list it is on (assuming it is on one).
// NOTE: this does not delete the passed colorScheme -- it just takes it off
// the list.
{
	if(colorScheme->listItem)
	{
		NamedListItemDestroy(viewColorSchemeList,colorScheme->listItem);
		colorScheme->listItem=NULL;
	}
}

bool ViewColorSchemeListAdd(VIEW_COLOR_SCHEME *colorScheme,const char *listName)
// Add the given colorScheme to the global list with the given name.
// If there's a problem, ErrorSet and return false.
{
	NAMED_LIST_ITEM
		*listItem;

	if((listItem=NamedListItemCreate(viewColorSchemeList,listName)))
	{
		NamedListItemDataSet(listItem,colorScheme);
		colorScheme->listItem=listItem;
		return(true);
	}
	return(false);
}

void ViewColorSchemeUnInit()
// Undo what ViewColorSchemeInit did.
{
	NAMED_LIST_ITEM
		*listItem;

	while((listItem=NamedListItemLocateFirst(viewColorSchemeList)))			// remove all items from the list
	{
		ViewColorSchemeDestroy((VIEW_COLOR_SCHEME *)NamedListItemDataGet(listItem));
		NamedListItemDestroy(viewColorSchemeList,listItem);
	}
	NamedListDestroy(viewColorSchemeList);									// kill the list itself
}

bool ViewColorSchemeInit()
// Initialize the color schemes.
// If there's a problem, ErrorSet and return false.
{
	if((viewColorSchemeList=NamedListCreate()))
	{
		return(true);
	}
	return(false);
}
