// View color scheme font loading functions
// Copyright © 2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// The code here manages loading and unloading fonts associated
// with view color schemes.

#include	"includes.h"

typedef struct viewColorSchemeFonts
{
	EDITOR_FONT
		*font,									// font for the view when no style is chosen
		*styleFonts[VIEW_MAX_STYLES];			// font for each style (NULL if default view font should be used)
} VIEW_COLOR_SCHEME_FONTS;

EDITOR_FONT *ViewColorSchemeFontsStyleFontGet(const VIEW_COLOR_SCHEME_FONTS *colorSchemeFonts,uint32_t style)
// Return a pointer to the font for the given style in the given loaded fonts, or NULL
// if none has been set.
{
	return(colorSchemeFonts->styleFonts[style]);
}

EDITOR_FONT *ViewColorSchemeFontsViewFontGet(const VIEW_COLOR_SCHEME_FONTS *colorSchemeFonts)
// Return a pointer to the font for the view.
{
	return(colorSchemeFonts->font);
}

void ViewColorSchemeFontsUnload(VIEW_COLOR_SCHEME_FONTS *colorSchemeFonts)
// Free up any fonts loaded by ViewColorSchemeFontLoad.
{
	uint32_t
		i;

	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		if(colorSchemeFonts->styleFonts[i])
		{
			PlatformFontFree(colorSchemeFonts->styleFonts[i]);
			colorSchemeFonts->styleFonts[i]=NULL;
		}
	}
	if(colorSchemeFonts->font)
	{
		PlatformFontFree(colorSchemeFonts->font);
		colorSchemeFonts->font=NULL;
	}
}

void ViewColorSchemeFontsLoad(VIEW_COLOR_SCHEME_FONTS *colorSchemeFonts,VIEW_COLOR_SCHEME *colorScheme,const char *fontName)
// Run through all the fonts in VIEW_COLOR_SCHEME, and load up colorSchemeFonts.
// NOTE: colorSchemeFonts must be empty at the time this is called.
// NOTE: this isn't allowed to fail -- if fonts fail to load, they'll just end up as
// NULL in colorSchemeFonts.
{
	uint32_t
		i;
	char
		familyName[MAX_FONT_NAME_LENGTH],
		compositeName[MAX_FONT_NAME_LENGTH];
	uint32_t
		size;
	bool
		bold,
		oblique;
	bool
		styleBold,
		styleOblique;

	colorSchemeFonts->font=PlatformFontLoad(fontName);			// load up the main font
	PlatformFontCompositeNameSplit(fontName,familyName,&size,&bold,&oblique);	// split into components to merge with style attributes
	for(i=0;i<VIEW_MAX_STYLES;i++)
	{
		ViewColorSchemeStyleFontAttributesGet(colorScheme,i,&styleBold,&styleOblique);
		PlatformFontComponentNameJoin(familyName,size,styleBold,styleOblique,compositeName);
		colorSchemeFonts->styleFonts[i]=PlatformFontLoad(compositeName);
	}
}

void ViewColorSchemeFontsReload(VIEW_COLOR_SCHEME_FONTS *colorSchemeFonts,VIEW_COLOR_SCHEME *colorScheme,const char *fontName)
// When the fonts or font attributes of colorScheme change, call this to
// reload colorSchemeFonts with updated font information.
// NOTE: this isn't allowed to fail -- if fonts fail to load, they'll just end up as
// NULL in colorSchemeFonts.
{
	ViewColorSchemeFontsUnload(colorSchemeFonts);
	ViewColorSchemeFontsLoad(colorSchemeFonts,colorScheme,fontName);
}

void ViewColorSchemeFontsDestroy(VIEW_COLOR_SCHEME_FONTS *colorSchemeFonts)
// Destroy a set of color scheme fonts created by ViewColorSchemeFontsCreate.
{
	ViewColorSchemeFontsUnload(colorSchemeFonts);
	MDisposePtr(colorSchemeFonts);
}

VIEW_COLOR_SCHEME_FONTS *ViewColorSchemeFontsCreate()
// Create an empty set of color scheme fonts.
// If there's a problem, ErrorSet and return NULL.
{
	VIEW_COLOR_SCHEME_FONTS
		*colorSchemeFonts;
	uint32_t
		i;

	if((colorSchemeFonts=(VIEW_COLOR_SCHEME_FONTS *)MNewPtr(sizeof(VIEW_COLOR_SCHEME_FONTS))))
	{
		colorSchemeFonts->font=NULL;
		for(i=0;i<VIEW_MAX_STYLES;i++)
		{
			colorSchemeFonts->styleFonts[i]=NULL;
		}
	}
	return(colorSchemeFonts);
}
