// High level management of views.
// Copyright © 2004-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

// The code in this file is the only platform independent code
// which should be calling the platform view code.

#include	"includes.h"

void ViewMetaChangeCallback(EDITOR_VIEW *view)
// Tell the given view that meta information has changed.
{
	if(view->viewMetaChangedVector)
	{
		view->viewMetaChangedVector(view);
	}
}

void ViewsMetaChangeCallbacks(EDITOR_BUFFER *buffer)
// Tell all the views that meta information about the buffer
// has changed.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		ViewMetaChangeCallback(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

bool ViewUtf8Get(EDITOR_VIEW *view)
// Get the flag associated with the view of a window that tells if the
// view is interpreting text as UTF8, or just ASCII.
{
	return(view->useUtf8);
}

bool ViewUtf8Set(EDITOR_VIEW *view,bool useUtf8)
// Set the flag associated with the view of a window that tells if the
// view is interpreting text as UTF8, or just ASCII.
// If there is a problem, set the error, and return false.
{
	PlatformViewMetaChangeStart(view);
	view->useUtf8=useUtf8;
	PlatformViewMetaChangeEnd(view);
	ViewMetaChangeCallback(view);
	return(true);
}

uint32_t ViewTabSizeGet(EDITOR_VIEW *view)
// Get the current tab size for the given view.
{
	return(view->tabSize);
}

bool ViewTabSizeSet(EDITOR_VIEW *view,uint32_t size)
// Set the tab size to be used in view.
// If there is a problem, leave the tab size unchanged, set the error
// and return false.
{
	PlatformViewMetaChangeStart(view);
	view->tabSize=size;
	PlatformViewMetaChangeEnd(view);
	ViewMetaChangeCallback(view);
	return(true);
}

void ViewFontNameGet(EDITOR_VIEW *view,char *fontName)
// Get the font name for the given view.
// NOTE: this collects the name from the font itself --
// that way it'll get the canonicalized version.
{
	EDITOR_FONT
		*font;

	if((font=ViewColorSchemeFontsViewFontGet(view->colorSchemeFonts)))
	{
		snprintf(fontName,MAX_FONT_NAME_LENGTH,"%s",PlatformFontCompositeNameGet(font));
	}
	else
	{
		fontName[0]='\0';		// shouldn't really happen unless font load fails for some reason
	}
}

void ViewFontNameSet(EDITOR_VIEW *view,const char *fontName)
// Set the font name to be used in view.
// This will update all the loaded fonts for the view.
{
	PlatformViewColorSchemeChangeStart(view);
	ViewColorSchemeFontsReload(view->colorSchemeFonts,view->colorScheme,fontName);
	PlatformViewColorSchemeChangeEnd(view);
}

void ViewColorsGet(EDITOR_VIEW *view,EDITOR_COLOR *backgroundColor,EDITOR_COLOR *foregroundColor)
// Get the foreground and background colors for the given view.
{
	ViewColorSchemeDrawColorsGet(view->colorScheme,true,backgroundColor,foregroundColor);
}

void ViewColorsSet(EDITOR_VIEW *view,EDITOR_COLOR backgroundColor,EDITOR_COLOR foregroundColor)
// Set the foreground and background colors for the given view.
// NOTE: currently this sets the both the active and inactive colors.
{
	PlatformViewColorSchemeChangeStart(view);
	ViewColorSchemeDrawColorsSet(view->colorScheme,false,backgroundColor,foregroundColor);
	ViewColorSchemeDrawColorsSet(view->colorScheme,true,backgroundColor,foregroundColor);
	PlatformViewColorSchemeChangeEnd(view);
}

bool ViewColorSchemeSet(EDITOR_VIEW *view,const char *colorSchemeName)
// Set the color scheme in use by the view to the new one passed in.
// This will replace the current color scheme.
// If there's a problem, leave the current color scheme unchanged, ErrorSet,
// and return false.
{
	char
		fontName[MAX_FONT_NAME_LENGTH];
	VIEW_COLOR_SCHEME
		*colorScheme,
		*newColorScheme;

	ViewFontNameGet(view,fontName);									// get current font name
	if((colorScheme=ViewColorSchemeListLocate(colorSchemeName)))
	{
		if((newColorScheme=ViewColorSchemeDuplicate(colorScheme)))	// make a duplicate of the passed scheme
		{
			PlatformViewColorSchemeChangeStart(view);
			ViewColorSchemeDestroy(view->colorScheme);				// get rid of the old one
			view->colorScheme=newColorScheme;						// save the new one
			ViewColorSchemeFontsReload(view->colorSchemeFonts,view->colorScheme,fontName);	// update the fonts
			PlatformViewColorSchemeChangeEnd(view);
			return(true);
		}
	}
	else
	{
		ErrorSet("Could not locate color scheme '%s'",colorSchemeName);
	}
	return(false);
}

void ViewsStyleAreaChangeEnd(EDITOR_BUFFER *buffer)
// The styles in buffer have finished being changed.
// Do whatever needs to be done to get it up to date.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewStyleAreaChangeEnd(currentView,currentView->parentBuffer->styleUniverse);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsStyleAreaChangeStart(EDITOR_BUFFER *buffer)
// The area covered by different styles in buffer are about to be changed in some way.
// Do whatever is needed.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewStyleAreaChangeStart(currentView,currentView->parentBuffer->styleUniverse);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsMarkChangeEnd(EDITOR_MARK *mark)
// The passed mark has finished being changed, update all affected views.
{
	MARK_VISIBILITY
		*currentVisibility;

	currentVisibility=mark->headVisibility;
	while(currentVisibility)							// walk through all views, update each one as needed
	{
		PlatformViewSelectionAreaChangeEnd(currentVisibility->parentView,currentVisibility->parentMark->selectionUniverse);
		currentVisibility=currentVisibility->nextView;	// locate next view this mark is visible in
	}
}

void ViewsMarkChangeStart(EDITOR_MARK *mark)
// A mark (which may be visible) in buffer is about to be changed in some way.
// Do whatever is needed.
// NOTE: mark changes cannot be nested, and nothing except a mark is allowed
// to be altered during a mark change.
{
	MARK_VISIBILITY
		*currentVisibility;

	currentVisibility=mark->headVisibility;
	while(currentVisibility)							// walk through all views, update each one as needed
	{
		PlatformViewSelectionAreaChangeStart(currentVisibility->parentView,currentVisibility->parentMark->selectionUniverse);
		currentVisibility=currentVisibility->nextView;	// locate next view this mark is visible in
	}
}

void ViewsSelectionAreaChangeEnd(EDITOR_BUFFER *buffer)
// The selection in buffer has finished being changed.
// Do whatever needs to be done to get it up to date.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewSelectionAreaChangeEnd(currentView,currentView->parentBuffer->selectionUniverse);
		if(currentView->viewSelectionChangedVector)		// see if someone needs to be called because of the change
		{
			currentView->viewSelectionChangedVector(currentView);		// call him
		}
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsSelectionAreaChangeStart(EDITOR_BUFFER *buffer)
// The selection in buffer is about to be changed in some way.
// Do whatever is needed.
// NOTE: a change in the selection could mean something as simple as a cursor position change.
// NOTE ALSO: selection changes cannot be nested, and nothing except the selection is allowed
// to be altered during a selection change.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewSelectionAreaChangeStart(currentView,currentView->parentBuffer->selectionUniverse);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsTextPieceReplaceEnd(EDITOR_BUFFER *buffer,uint64_t position,uint64_t startLength,uint64_t endLength)
// The piece of text in buffer has finished being changed, and the views have been
// updated. Do whatever needs to be done.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewTextPieceReplaceEnd(currentView,position,startLength,endLength);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsTextPieceReplaceStart(EDITOR_BUFFER *buffer,uint64_t position,uint64_t startLength,uint64_t endLength)
// A piece of text in buffer is about to be changed in some way.
// Do whatever is needed.
// NOTE: when the text changes, it is possible that the selection information
// will also change, so that updates will be handled correctly.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewTextPieceReplaceStart(currentView,position,startLength,endLength);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsTextReplaceEnd(EDITOR_BUFFER *buffer)
// The changes to the text in buffer are complete.
// Call the views to let them know.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewTextReplaceEnd(currentView);
		if(currentView->viewTextChangedVector)			// see if a function needs to be called because of the change
		{
			currentView->viewTextChangedVector(currentView);	// call it
		}
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewsTextReplaceStart(EDITOR_BUFFER *buffer)
// The text for buffer is about to be changed (possibly in multiple pieces).
// Call the views to let them prepare.
{
	EDITOR_VIEW
		*currentView;

	currentView=buffer->firstView;
	while(currentView)									// walk through all views, update each one as needed
	{
		PlatformViewTextReplaceStart(currentView);
		currentView=currentView->nextBufferView;		// locate next view of this buffer
	}
}

void ViewActiveSet(EDITOR_VIEW *view,bool active)
// Activate or deactivate the given view (make it draw with focus or not).
{
	view->active=active;
	PlatformViewFocusChange(view);
}

void ViewOriginGet(EDITOR_VIEW *view,int64_t *leftPixel,int64_t *topLine)
// Get the current origin of the given view.
{
	PlatformViewOriginGet(view,leftPixel,topLine);
}

void ViewOriginSet(EDITOR_VIEW *view,int64_t newLeftPixel,int64_t newTopLine)
// Set the given view's origin. Make callback as needed.
{
	PlatformViewOriginSet(view,newLeftPixel,newTopLine);
	if(view->viewPositionChangedVector)				// see if some function needs to be called because of the change
	{
		view->viewPositionChangedVector(view);		// call the function
	}
}

void ViewDispose(EDITOR_VIEW *view)
// Unlink view from everywhere, and dispose of it.
{
	PlatformViewInfoDispose(view);					// tell GUI to tear it down
	MarksVisibleUnlinkFromView(view);				// get rid of any visible marks hanging off the view
	ViewColorSchemeFontsUnload(view->colorSchemeFonts);		// unload the fonts
	ViewColorSchemeFontsDestroy(view->colorSchemeFonts);	// get rid of loaded fonts data structure
	ViewColorSchemeDestroy(view->colorScheme);		// get rid of color scheme for view
	EditorBufferViewUnlink(view);					// remove the view from its buffer
	MDisposePtr(view);
}

EDITOR_VIEW *ViewCreate(EDITOR_WINDOW *window,EDITOR_VIEW_DESCRIPTOR *descriptor)
// Create a view in window using descriptor.
// If there is a problem, ErrorSet, and return NULL.
{
	EDITOR_VIEW
		*view;
	VIEW_COLOR_SCHEME
		*colorScheme;

	if((view=(EDITOR_VIEW *)MNewPtr(sizeof(EDITOR_VIEW))))
	{
		EditorBufferViewLink(descriptor->buffer,view);	// link the view onto the buffer
		view->parentWindow=window;						// remember the window this view is in
		view->wantHome=false;

		view->active=descriptor->active;
		view->useUtf8=descriptor->useUtf8;
		view->tabSize=descriptor->tabSize;				// set the tab size

		if((colorScheme=ViewColorSchemeListLocate(descriptor->colorSchemeName)))	// locate the given color scheme
		{
			if((view->colorScheme=ViewColorSchemeDuplicate(colorScheme)))			// make a copy into the view
			{
				if((view->colorSchemeFonts=ViewColorSchemeFontsCreate()))
				{
					ViewColorSchemeFontsLoad(view->colorSchemeFonts,view->colorScheme,descriptor->fontName);
					view->viewTextChangedVector=descriptor->viewTextChangedVector;
					view->viewSelectionChangedVector=descriptor->viewSelectionChangedVector;
					view->viewPositionChangedVector=descriptor->viewPositionChangedVector;
					view->viewMetaChangedVector=descriptor->viewMetaChangedVector;

					view->headVisibility=NULL;					// no visible marks hanging off of this view yet
					view->tailVisibility=NULL;
					if(PlatformViewInfoCreate(view,&descriptor->rect))		// allow platform code to add whatever it likes
					{
						return(view);
					}
					ViewColorSchemeFontsUnload(view->colorSchemeFonts);
					ViewColorSchemeFontsDestroy(view->colorSchemeFonts);	// get rid of loaded fonts
				}
				ViewColorSchemeDestroy(view->colorScheme);
			}
		}
		else
		{
			ErrorSet("Could not locate color scheme '%s'",descriptor->colorSchemeName);
		}
		EditorBufferViewUnlink(view);
		MDisposePtr(view);
	}
	return(NULL);
}
