// TCL commands that e93 supports
// Copyright © 2000-2020 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

// NOTES about TCL, and its subtle interactions with the editor
//
// First, Tcl should never be allowed to evaluate script while an editor
// window, view, buffer, selectionlist, etc... is in a state of flux that
// would cause it to draw improperly if the script caused the editor to
// update its windows. This is because certain editor operations that
// scripts can call, can force updates of all editor windows.
//
// Second, if a script is running, and decides to operate on a window that
// is in some state of flux, (say in the middle of a replace-all), it must
// be prohibited from doing so, and the operation must fail. NOTE, that by
// rule one, if you are executing script in the middle of replace-all, that
// all buffers must be in a state where they can be drawn.
//
// Finally, if a script is run while some code of the editor is relying on
// a given window, or buffer to be around, the code MUST set the
// buffer Busy, or check that the buffer still exists after
// the script is run.
// If a buffer is set Busy, no script command, or side-effect can allow the
// buffer to be modified in ANY way, including text/selection changes, undo
// changes, new views, deleting views, etc...

#define	TEMP_STRING_BUFFER_LENGTH	1024

typedef bool OptionParseFunction(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex);

typedef struct
{
	const char
		*optionString;
	int
		optionValue;
	OptionParseFunction
		*function;
} OPTION_DESCRIPTOR_RECORD;

typedef struct
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;
} PARSE_MARK;

typedef struct
{
	bool
		parsed;
	int64_t
		value;
} PARSE_INT64;

static TOKEN_LIST
	menuRelationshipTokens[]=
	{
		{"nextsibling",MR_NEXT_SIBLING},
		{"previoussibling",MR_PREVIOUS_SIBLING},
		{"firstchild",MR_FIRST_CHILD},
		{"lastchild",MR_LAST_CHILD},
		{"",0}
	};

static TOKEN_LIST
	cursorMovementTokens[]=
	{
		{"leftchar",RPM_mBACKWARD|RPM_CHAR},
		{"rightchar",RPM_CHAR},
		{"leftword",RPM_mBACKWARD|RPM_WORD},
		{"rightword",RPM_WORD},
		{"upline",RPM_mBACKWARD|RPM_LINE},
		{"downline",RPM_LINE},
		{"startline",RPM_mBACKWARD|RPM_LINEEDGE},
		{"endline",RPM_LINEEDGE},
		{"startparagraph",RPM_mBACKWARD|RPM_PARAGRAPHEDGE},
		{"endparagraph",RPM_PARAGRAPHEDGE},
		{"uppage",RPM_mBACKWARD|RPM_PAGE},
		{"downpage",RPM_PAGE},
		{"startdoc",RPM_mBACKWARD|RPM_DOCEDGE},
		{"enddoc",RPM_DOCEDGE},
		{"",0}
	};

typedef enum
{
	SM_EXPRESSION=0,
	SM_MAPPING,
	SM_AT,
	SM_ROOT
} SYNTAX_MAP_COMMANDS;

static TOKEN_LIST
	syntaxMapCommands[]=
	{
		{"exp",SM_EXPRESSION},
		{"map",SM_MAPPING},
		{"at",SM_AT},
		{"root",SM_ROOT},
		{"",0}
	};

static uint32_t ViewStyleForceIntoRange(uint32_t style)
// Make sure the style is in range (for externally called routines).
{
	if(style>=VIEW_MAX_STYLES)
	{
		style=VIEW_MAX_STYLES-1;
	}
	return(style);
}

static uint64_t PositionForceIntoRange(EDITOR_BUFFER *buffer,uint64_t position)
// Take position, and make sure it is in range for buffer.
{
	uint64_t
		totalBytes;

	totalBytes=TextUniverseTotalBytes(buffer->textUniverse);
	if(position>totalBytes)
	{
		position=totalBytes;
	}
	return(position);
}

static void PositionsArrange(uint64_t *startPosition,uint64_t *endPosition)
// Make sure startPosition is <= endPosition.
// If not, flip them around.
{
	uint64_t
		temp;

	if(*startPosition>*endPosition)
	{
		temp=*startPosition;
		*startPosition=*endPosition;
		*endPosition=temp;
	}
}

static EDITOR_BUFFER *TclEditorBufferGet(Tcl_Interp *localInterpreter,Tcl_Obj *const name)
// See if string is a buffer, if so, return it.
// If not, return NULL, and set the Tcl result.
{
	EDITOR_BUFFER
		*buffer;
	char
		*string;

	string=Tcl_GetString(name);
	if((buffer=EditorBufferLocate(string)))
	{
		return(buffer);
	}
	Tcl_AppendResult(localInterpreter,"Failed to locate buffer '",string,"'",NULL);
	return(NULL);
}

static EDITOR_WINDOW *TclEditorWindowGet(Tcl_Interp *localInterpreter,Tcl_Obj *const name)
// See if string is a document window, if so, return it.
// If not, return NULL, and set the Tcl result.
{
	EDITOR_WINDOW
		*window;
	char
		*string;

	string=Tcl_GetString(name);
	if((window=EditorDocumentWindowLocate(string)))
	{
		return(window);
	}
	Tcl_AppendResult(localInterpreter,"Failed to locate window '",string,"'",NULL);
	return(NULL);
}

static bool TclInt32Get(Tcl_Interp *localInterpreter,Tcl_Obj *const object,int32_t *number)
// convert object to a int32_t.
// Return false if the conversion failed, and fill the Tcl result.
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(int32_t)longValue;
		return(true);
	}
	return(false);
}

static bool TclUInt32Get(Tcl_Interp *localInterpreter,Tcl_Obj *const object,uint32_t *number)
// Convert object to a uint32_t.
// Return false if the conversion failed, and fill the Tcl result.
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(uint32_t)longValue;					// ### this should really be fixed to be a true unsigned
		return(true);
	}
	return(false);
}

static bool TclInt64Get(Tcl_Interp *localInterpreter,Tcl_Obj *const object,int64_t *number)
// convert object to a int64_t.
// Return false if the conversion failed, and fill the Tcl result.
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(uint64_t)longValue;
		return(true);
	}
	return(false);
}

static bool TclUInt64Get(Tcl_Interp *localInterpreter,Tcl_Obj *const object,uint64_t *number)
// Convert object to a uint64_t.
// Return false if the conversion failed, and fill the Tcl result.
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		*number=(uint64_t)longValue;					// ### this should really be fixed to be a true unsigned
		return(true);
	}
	return(false);
}

static bool TclTabSizeGet(Tcl_Interp *localInterpreter,Tcl_Obj *const object,uint32_t *tabSize)
// Convert object to a tabSize.
// This forces the incoming tab size into range.
// Return false if the conversion failed, and fill the Tcl result.
{
	long
		longValue;

	if(Tcl_ExprLongObj(localInterpreter,object,&longValue)==TCL_OK)
	{
		if(longValue<0)
		{
			longValue=0;
		}
		else if(longValue>VIEW_MAX_TAB_SIZE)
		{
			longValue=VIEW_MAX_TAB_SIZE;
		}
		*tabSize=(uint32_t)longValue;
		return(true);
	}
	return(false);
}

static bool TclRectangleGet(Tcl_Interp *localInterpreter,Tcl_Obj *const *objs,EDITOR_RECT *rect)
// Convert objs (of which there must be at least 4) to an EDITOR_RECT.
// Return false if the conversion failed, and fill the Tcl result.
{
	if(TclInt32Get(localInterpreter,objs[0],&rect->x))
	{
		if(TclInt32Get(localInterpreter,objs[1],&rect->y))
		{
			if(TclUInt32Get(localInterpreter,objs[2],&rect->w))
			{
				if(TclUInt32Get(localInterpreter,objs[3],&rect->h))
				{
					return(true);
				}
			}
		}
	}
	return(false);
}

static bool NibbleGet(char character,uint32_t *nibble)
// Interpret character as a hex nibble, and return it.
// If there is a problem return false.
{
	if(character>='0'&&character<='9')
	{
		*nibble=character-'0';
		return(true);
	}
	else
	{
		if(character>='A'&&character<='F')
		{
			*nibble=10+character-'A';
			return(true);
		}
		else
		{
			if(character>='a'&&character<='f')
			{
				*nibble=10+character-'a';
				return(true);
			}
		}
	}
	return(false);
}

static bool TclColorGet(Tcl_Interp *localInterpreter,Tcl_Obj *const object,EDITOR_COLOR *color)
// Convert object to a color value.
// Return false if the conversion failed, and fill the Tcl result.
{
	uint32_t
		index,
		nibble;
	char
		character;
	bool
		fail;
	char
		*string;

	fail=false;
	string=Tcl_GetString(object);
	if(!PlatformColorNameToColor(string,color))
	{
		*color=0;
		index=0;
		while(!fail&&(character=string[index]))
		{
			if(NibbleGet(character,&nibble))
			{
				(*color)<<=4;
				(*color)|=nibble;
				index++;
			}
			else
			{
				fail=true;
			}
		}
		if(!index)					// no characters in the input string
		{
			fail=true;
		}
	}
	if(fail)
	{
		Tcl_AppendResult(localInterpreter,"Invalid color: '",string,"'",NULL);
	}
	return(!fail);
}

static bool TclColorOrEmptyGet(Tcl_Interp *localInterpreter,Tcl_Obj *const object,EDITOR_COLOR *color)
// Convert object to a color value.
// If the object is the empty string, return it as color 0x00000000.
// This is useful for styles, when the color specified should not override the default.
// Return false if the conversion failed, and fill the Tcl result.
{
	char
		*string;

	string=Tcl_GetString(object);
	if(string[0]!='\0')			// something specified?
	{
		return(TclColorGet(localInterpreter,object,color));
	}
	else
	{
		(*color)=0x00000000;
	}
	return(true);
}

static bool TclBooleanGet(Tcl_Interp *localInterpreter,Tcl_Obj *const object,bool *value)
// Convert object to a boolean value.
// Return false if the conversion failed, and fill the Tcl result.
{
	int
		intValue;

	if(Tcl_GetBoolean(localInterpreter,Tcl_GetString(object),&intValue)==TCL_OK)
	{
		*value=intValue?true:false;
		return(true);
	}
	return(false);
}

static bool TclOptionBoolean(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Handle parsing of a boolean option.
{
	*((bool *)variable)=(optionValue!=0);
	return(true);
}

static bool TclOptionEnum(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Handle parsing of a enumeration option.
{
	*((int *)variable)=optionValue;
	return(true);
}

static bool TclOptionInt64(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Return a signed 64-bit integer.
{
	PARSE_INT64
		*parseInt64;

	if((*optionIndex)<objc)
	{
		parseInt64=(PARSE_INT64 *)variable;
		if(TclInt64Get(localInterpreter,objv[*optionIndex],&parseInt64->value))
		{
			parseInt64->parsed=true;
			(*optionIndex)++;
			return(true);
		}
	}
	else
	{
		Tcl_AppendResult(localInterpreter,optionName,": Missing arguments",NULL);
	}
	return(false);
}

static bool TclOptionMark(Tcl_Interp *localInterpreter,char *optionName,int optionValue,void *variable,int objc,Tcl_Obj *const objv[],int *optionIndex)
// Attempt to locate a currently existing mark within an editor buffer.
{
	char
		*markName;
	PARSE_MARK
		*parseMark;

	if((*optionIndex)<objc)
	{
		parseMark=(PARSE_MARK *)variable;
		markName=Tcl_GetString(objv[*optionIndex]);
		(*optionIndex)++;
		if((parseMark->mark=EditorMarkLocate(parseMark->buffer,markName)))
		{
			return(true);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Failed to locate mark '",markName,"'",NULL);
		}
	}
	else
	{
		Tcl_AppendResult(localInterpreter,optionName,": Missing arguments",NULL);
	}
	return(false);
}

static bool TclOptionMatchLocate(const OPTION_DESCRIPTOR_RECORD *options,char *option,int *matchIndex)
// Attempt to locate an option that matches option.
// If one was located, return true with the index in matchIndex.
// Else, return false.
{
	int
		i;
	bool
		found,
		reachedEnd;

	i=0;
	found=reachedEnd=false;
	while(!found&&!reachedEnd)
	{
		if(options[i].optionString)
		{
			if(strcmp(option,options[i].optionString)==0)
			{
				*matchIndex=i;
				found=true;
			}
			i++;
		}
		else
		{
			reachedEnd=true;
		}
	}
	return(found);
}

static void TclOptionsAddToResult(Tcl_Interp *localInterpreter,const OPTION_DESCRIPTOR_RECORD *options)
// Used when reporting bad options.
{
	int
		i;

	Tcl_AppendResult(localInterpreter,options[0].optionString,NULL);
	i=1;
	while(options[i].optionString)
	{
		if(options[i+1].optionString)
		{
			Tcl_AppendResult(localInterpreter,", ",NULL);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,", or ",NULL);
		}
		Tcl_AppendResult(localInterpreter,options[i].optionString,NULL);
		i++;
	}
}

static bool TclOptionsParse(Tcl_Interp *localInterpreter,const OPTION_DESCRIPTOR_RECORD *options,void **variables,int objc,Tcl_Obj *const objv[],int *optionIndex,bool unrecognizedIsError)
// Parse command options given a description of the options, and a list of parameters for them
// to modify.
// If there is a problem, set the error string in localInterpreter, and return false.
// NOTE: if unrecognizedIsError is false, this will stop parsing at the first
// unrecognized token, and return true.
// If unrecognizedIsError is true, this will only return true if the list of
// options can be parsed correctly until the end.
{
	bool
		fail,
		mismatch;
	char
		*optionName;
	int
		matchIndex;

	fail=mismatch=false;
	while(!fail&&!mismatch&&((*optionIndex)<objc))
	{
		optionName=Tcl_GetString(objv[*optionIndex]);
		(*optionIndex)++;
		matchIndex=0;			// keep GCC quiet
		if(TclOptionMatchLocate(options,optionName,&matchIndex))
		{
			fail=!options[matchIndex].function(localInterpreter,optionName,options[matchIndex].optionValue,variables[matchIndex],objc,objv,optionIndex);
		}
		else
		{
			mismatch=true;
			if(unrecognizedIsError)
			{
				Tcl_AppendResult(localInterpreter,"bad option \"",optionName,"\": must be ",NULL);
				TclOptionsAddToResult(localInterpreter,options);
				fail=true;
			}
		}
	}
	return(!fail);
}

static bool TclVarListGet(Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[],int64_t *varList)
// Write values from the given variable array to the given array of integers.
// If there is a problem, set the TCL error, and return false.
{
	Tcl_Obj
		*varValue;
	int
		currentCount;
	bool
		fail;

	fail=false;
	for(currentCount=0;currentCount<objc&&!fail;currentCount++)
	{
		if((varValue=Tcl_ObjGetVar2(localInterpreter,objv[currentCount],NULL,TCL_LEAVE_ERR_MSG)))
		{
			fail=!TclInt64Get(localInterpreter,varValue,&(varList[currentCount]));
		}
		else
		{
			fail=true;
		}
	}
	return(!fail);
}

static bool TclVarListPut(Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[],int64_t *varList)
// Write the variables back into Tcl variables.
// If there is a problem, set the TCL error, and return false.
{
	int
		currentCount;
	bool
		fail;

	fail=false;
	for(currentCount=0;currentCount<objc&&!fail;currentCount++)
	{
		fail=!Tcl_ObjSetVar2(localInterpreter,objv[currentCount],NULL,Tcl_NewLongObj(varList[currentCount]),TCL_LEAVE_ERR_MSG);
	}
	return(!fail);
}

static bool ShellBufferNotBusy(Tcl_Interp *localInterpreter,EDITOR_BUFFER *buffer)
// Check the given buffer's busy state.
// If it is true, set the Tcl result, and return false.
{
	if(!EditorBufferBusy(buffer))
	{
		return(true);
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Buffer '",buffer->contentName,"' is busy",NULL);
	}
	return(false);
}

static bool CurrentClipboardNotBusy(Tcl_Interp *localInterpreter)
// Make sure the current clipboard is not busy. If it is, set the Tcl result
// and return false.
{
	EDITOR_BUFFER
		*clipboardBuffer;

	if((clipboardBuffer=ClipboardCurrentGet()))
	{
		return(ShellBufferNotBusy(localInterpreter,clipboardBuffer));
	}
	return(true);
}

// ACTUAL TCL COMMAND IMPLEMENTATIONS

static int Cmd_DeleteMenu(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Delete a menu from the menu list.
// NOTE: if this deletes an element, it will also delete the element's children.
{
	EDITOR_MENU
		*menu;
	uint32_t
		index;
	int
		pathArgc;
	const char
		**pathArgv;
	bool
		fail;

	fail=false;
	if(objc>1)
	{
		for(index=1;!fail&&(int)index<objc;index++)
		{
			if(Tcl_SplitList(localInterpreter,Tcl_GetString(objv[index]),&pathArgc,&pathArgv)==TCL_OK)
			{
				menu=NULL;										// start at the root
				if(PlatformMenuGet(pathArgc,pathArgv,&menu))
				{
					PlatformMenuDispose(menu);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid menu specification",NULL);
					fail=true;
				}
				Tcl_Free((char *)pathArgv);
			}
			else
			{
				fail=true;
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"menuPathList ?menuPathList ...?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_AddMenu(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Add a menu to the menu list.
// NOTE: if a menu already exists, this will delete it before attempting
// a creation,
// NOTE ALSO: if this deletes an element, it will also delete the element's children.
{
	int
		relationship;
	bool
		active;
	int
		pathArgc;
	const char
		**pathArgv;
	bool
		succeed;

	succeed=false;
	if(objc==7)
	{
		if(TokenMatch(Tcl_GetString(objv[2]),menuRelationshipTokens,&relationship))
		{
			if(TclBooleanGet(localInterpreter,objv[3],&active))
			{
				if(Tcl_SplitList(localInterpreter,Tcl_GetString(objv[1]),&pathArgc,&pathArgv)==TCL_OK)
				{
					ErrorTraceClear();
					if(PlatformMenuCreate(pathArgc,pathArgv,relationship,Tcl_GetString(objv[4]),Tcl_GetString(objv[5]),Tcl_GetString(objv[6]),active))
					{
						succeed=true;
					}
					else
					{
						Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
					}
					Tcl_Free((char *)pathArgv);
				}
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Syntax error in menu relationship field",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"menuPathList relationship active menuName menuModifiers menuFunction");
	}
	return(succeed?TCL_OK:TCL_ERROR);
}

static int Cmd_BindKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Add a key binding to the bindings list.
// NOTE: if a binding already exists, this will delete it before attempting
// a creation.
{
	uint32_t
		modifiers,
		keyCode;

	if(objc==3)
	{
		if(KeyBindNameToKey(Tcl_GetString(objv[1]),&modifiers,&keyCode))		// generate values from the name
		{
			ErrorTraceClear();
			if(KeyBindCreate(modifiers,keyCode,Tcl_GetString(objv[2])))
			{
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"keyDesignation keyFunction");
	}
	return(TCL_ERROR);
}

static int Cmd_UnBindKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Delete a key binding from the bindings list.
{
	uint32_t
		modifiers,
		keyCode;

	if(objc==2)
	{
		if(KeyBindNameToKey(Tcl_GetString(objv[1]),&modifiers,&keyCode))		// generate modifiers and keyCode from the string
		{
			if(KeyBindDelete(modifiers,keyCode))
			{
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Key binding does not exist",NULL);
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"keyDesignation");
	}
	return(TCL_ERROR);
}

static int Cmd_KeyBindings(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Dump all key bindings.
{
	KEY_BINDING
		*currentBinding;
	uint32_t
		huntModifiers,
		huntKeyCode,
		modifiers,
		keyCode;
	char
		nameString[TEMP_STRING_BUFFER_LENGTH];
	bool
		haveName;
	bool
		fail;
	bool
		didOne;

	if(objc>=1&&objc<=2)
	{
		fail=false;
		haveName=false;
		if(objc>=2)
		{
			haveName=true;
			if(!KeyBindNameToKey(Tcl_GetString(objv[1]),&huntModifiers,&huntKeyCode))			// find the key to match
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				fail=true;
			}
		}
		if(!fail)
		{
			currentBinding=KeyBindLocateFirst();
			didOne=false;
			while(currentBinding)
			{
				KeyBindEditorKeyGet(currentBinding,&modifiers,&keyCode);

				if(!haveName||((modifiers==huntModifiers)&&(keyCode==huntKeyCode)))
				{
					KeyBindKeyToName(modifiers,keyCode,nameString,sizeof(nameString));
					if(didOne)				// do not put return on last line
					{
						Tcl_AppendResult(localInterpreter,"\n",NULL);
					}
					Tcl_AppendResult(localInterpreter,"bindkey",NULL);
					Tcl_AppendElement(localInterpreter,nameString);
					Tcl_AppendElement(localInterpreter,KeyBindTextGet(currentBinding));
					didOne=true;
				}
				currentBinding=KeyBindLocateNext(currentBinding);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"?keyDesignation?");
	}
	return(TCL_ERROR);
}

static int Cmd_WaitKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Wait for a keyboard press, return the string for the code, and modifiers.
{
	EDITOR_KEY
		editorKey;
	char
		nameString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		PlatformKeyPressGet(&editorKey,true,false);
		KeyBindKeyToName(editorKey.modifiers,editorKey.keyCode,nameString,sizeof(nameString));
		Tcl_AppendElement(localInterpreter,nameString);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_GetKey(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Check to see if there is a keyboard event in the queue, if so, get it, if not, return an error.
{
	EDITOR_KEY
		editorKey;
	char
		nameString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		if(PlatformKeyPressGet(&editorKey,false,false))
		{
			KeyBindKeyToName(editorKey.modifiers,editorKey.keyCode,nameString,sizeof(nameString));
			Tcl_AppendElement(localInterpreter,nameString);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_DialogCenter(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the center point for all dialog boxes.
{
	int32_t
		xPos,
		yPos;

	if(objc==3)
	{
		if(TclInt32Get(localInterpreter,objv[1],&xPos))
		{
			if(TclInt32Get(localInterpreter,objv[2],&yPos))
			{
				PlatformDialogCenterPointSet(xPos,yPos);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"centerX centerY");
	}
	return(TCL_ERROR);
}

static int Cmd_OkDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a dialog with some text, and an OK button.
{
	bool
		fail;

	fail=false;
	if(objc==2)
	{
		ErrorTraceClear();
		if(PlatformDialogOk(Tcl_GetString(objv[1])))
		{
			AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogText");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_OkCancelDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a dialog with some text, and an OK button, and a cancel button.
{
	bool
		cancel,
		fail;

	fail=false;
	if(objc==2)
	{
		ErrorTraceClear();
		if(PlatformDialogOkCancel(Tcl_GetString(objv[1]),&cancel))
		{
			if(cancel)
			{
				fail=true;
			}
			AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogText");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_YesNoDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a dialog with some text, and yes, no, and cancel buttons.
{
	bool
		yes,
		cancel,
		fail;

	fail=false;
	if(objc==2)
	{
		ErrorTraceClear();
		if(PlatformDialogYesNoCancel(Tcl_GetString(objv[1]),&yes,&cancel))
		{
			if(cancel)
			{
				fail=true;
			}
			else
			{
				Tcl_AppendResult(localInterpreter,yes?"1":"0",NULL);
			}
			AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogText");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_TextDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a get text dialog.
// If InitialText is not specified, then no initial text will be present in the dialog.
// NOTE: since this might indirectly modify the current clipboard (cut/copy might be allowed
// in the dialog), the clipboard must not be busy.
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetString(objv[2]),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ErrorTraceClear();
			if(PlatformDialogSimpleText(Tcl_GetString(objv[1]),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(cancel)
				{
					fail=true;
				}
				else
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SearchDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a search dialog.
// findBuffer, and replaceBuffer must exist, or an error will be returned.
// NOTE: currently DialogTitle is ignored, later it should be implemented.
// NOTE: since this might indirectly modify the current clipboard (cut/copy might be allowed
// in the dialog), the clipboard must not be busy.
{
	int64_t
		varList[8];
	bool
		backwards,
		wrapAround,
		ignoreCase,
		limitScope,
		selectionExpr,
		useSubexpression,
		replaceProc;
	uint32_t
		subexpressionIndex;
	uint16_t
		searchMode;
	EDITOR_BUFFER
		*findBuffer,
		*replaceBuffer;
	bool
		cancel,
		result;

	result=false;
	if(objc==12)
	{
		if((findBuffer=TclEditorBufferGet(localInterpreter,objv[2])))
		{
			if(ShellBufferNotBusy(localInterpreter,findBuffer))
			{
				if((replaceBuffer=TclEditorBufferGet(localInterpreter,objv[3])))
				{
					if(ShellBufferNotBusy(localInterpreter,replaceBuffer))
					{
						if(TclVarListGet(localInterpreter,8,&(objv[4]),&(varList[0])))
						{
							if(CurrentClipboardNotBusy(localInterpreter))
							{
								ErrorTraceClear();

								backwards=varList[0]?true:false;
								wrapAround=varList[1]?true:false;
								ignoreCase=varList[2]?true:false;
								limitScope=varList[3]?true:false;
								selectionExpr=varList[4]?true:false;
								useSubexpression=varList[5]?true:false;
								subexpressionIndex=(uint32_t)varList[6];
								replaceProc=varList[7]?true:false;

								if(PlatformDialogSearchReplace(findBuffer,replaceBuffer,&backwards,&wrapAround,&ignoreCase,&limitScope,&selectionExpr,&useSubexpression,&subexpressionIndex,&replaceProc,&searchMode,&cancel))
								{
									// Make an undo boundary for the buffers used in the dialog
									UndoGroupBegin(findBuffer);
									UndoGroupEndStrict(findBuffer);
									UndoGroupBegin(replaceBuffer);
									UndoGroupEndStrict(replaceBuffer);

									if(!cancel)
									{
										result=true;
									}
									AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!

									varList[0]=backwards;
									varList[1]=wrapAround;
									varList[2]=ignoreCase;
									varList[3]=limitScope;
									varList[4]=selectionExpr;
									varList[5]=useSubexpression;
									varList[6]=subexpressionIndex;
									varList[7]=replaceProc;

									TclVarListPut(localInterpreter,8,&(objv[4]),&(varList[0]));	// put these back even if user canceled
								}
								else
								{
									Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
								}
							}
							if(result)
							{
								switch(searchMode)
								{
									case ST_FIND:
										Tcl_AppendResult(localInterpreter,"find",NULL);
										break;
									case ST_FIND_ALL:
										Tcl_AppendResult(localInterpreter,"findall",NULL);
										break;
									case ST_REPLACE:
										Tcl_AppendResult(localInterpreter,"replace",NULL);
										break;
									case ST_REPLACE_ALL:
										Tcl_AppendResult(localInterpreter,"replaceall",NULL);
										break;
								}
								return(TCL_OK);
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle findBuffer replaceBuffer backwardsVar wrapAroundVar ignoreCaseVar limitScopeVar selectionExprVar useSubexpressionVar subexpressionIndexVar replaceProcVar");
	}
	return(TCL_ERROR);
}

static int Cmd_ListDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a list dialog, handle it, and if all is well, return a new list
// that contains the chosen elements of the old list.
{
	uint32_t
		index;
	int
		listArgc;
	const char
		**listArgv;
	bool
		*selectedElements;
	bool
		cancel,
		fail;

	fail=false;
	if(objc==3)
	{
		if(Tcl_SplitList(localInterpreter,Tcl_GetString(objv[2]),&listArgc,&listArgv)==TCL_OK)
		{
			if((selectedElements=(bool *)MNewPtrClr(sizeof(bool)*listArgc)))	// get array of "falses"
			{
				if(listArgc)
				{
					selectedElements[0]=true;	// select the first element if there is one
				}
				ErrorTraceClear();
				if(PlatformDialogSimpleList(Tcl_GetString(objv[1]),listArgc,listArgv,selectedElements,&cancel))	// let user choose elements
				{
					if(!cancel)
					{
						for(index=0;(int)index<listArgc;index++)
						{
							if(selectedElements[index])
							{
								Tcl_AppendElement(localInterpreter,listArgv[index]);	// make a result that contains the selected elements
							}
						}
					}
					else
					{
						fail=true;
					}
					AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
					fail=true;
				}
				MDisposePtr(selectedElements);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Failed to allocate memory for list result",NULL);
				fail=true;
			}
			Tcl_Free((char *)listArgv);
		}
		else
		{
			// SplitList will leave a result if it fails
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle inputList");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_OpenDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a get file(s) dialog.
// If InitialText is not specified, then no initial text will be present in the dialog.
// NOTE: since this might indirectly modify the current clipboard (cut/copy might be allowed
// in the dialog), the clipboard must not be busy.
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	uint32_t
		index;
	char
		**paths;

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetString(objv[2]),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ErrorTraceClear();
			if(PlatformDialogOpenFile(Tcl_GetString(objv[1]),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&paths,&cancel))
			{
				if(!cancel)
				{
					index=0;
					while(paths[index])
					{
						Tcl_AppendElement(localInterpreter,paths[index++]);
					}
					PlatformDialogFreeOpenFilePaths(paths);
				}
				else
				{
					fail=true;
				}
				AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SaveDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a save file dialog.
// If InitialText is not specified, then no initial text will be present in the dialog.
// NOTE: since this might indirectly modify the current clipboard (cut/copy might be allowed
// in the dialog), the clipboard must not be busy.
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetString(objv[2]),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ErrorTraceClear();
			if(PlatformDialogSaveFile(Tcl_GetString(objv[1]),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(!cancel)
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				else
				{
					fail=true;
				}
				AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_PathDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a path choose dialog.
// If InitialText is not specified, then no initial text will be present in the dialog.
// NOTE: since this might indirectly modify the current clipboard (cut/copy might be allowed
// in the dialog), the clipboard must not be busy.
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetString(objv[2]),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ErrorTraceClear();
			if(PlatformDialogChoosePath(Tcl_GetString(objv[1]),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(!cancel)
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				else
				{
					fail=true;
				}
				AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialText?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_FontDialog(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a font family selection dialog.
// If InitialFont is not specified, then no initial family will be selected in the dialog.
// NOTE: since this might indirectly modify the current clipboard (cut/copy might be allowed
// in the dialog), the clipboard must not be busy.
{
	bool
		cancel,
		fail;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	fail=false;
	if(objc==2||objc==3)
	{
		if(objc==3)
		{
			strncpy(tempString,Tcl_GetString(objv[2]),TEMP_STRING_BUFFER_LENGTH-1);
			tempString[TEMP_STRING_BUFFER_LENGTH-1]='\0';
		}
		else
		{
			tempString[0]='\0';
		}
		if(CurrentClipboardNotBusy(localInterpreter))
		{
			ErrorTraceClear();
			if(PlatformDialogChooseFont(Tcl_GetString(objv[1]),&(tempString[0]),TEMP_STRING_BUFFER_LENGTH,&cancel))
			{
				if(!cancel)
				{
					Tcl_AppendResult(localInterpreter,&(tempString[0]),NULL);
				}
				else
				{
					fail=true;
				}
				AbortClear();				// if user managed to send break sequence during the dialog, we would rather not abort now!
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"dialogTitle ?initialFont?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_CloseBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Close the passed buffer as long as it is not in use, or being modified.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorBufferClose(buffer);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_OpenBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Open the passed file into a buffer, return the buffer name.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		ErrorTraceClear();
		if((buffer=EditorBufferOpen(Tcl_GetString(objv[1]))))
		{
			Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"pathName");
	}
	return(TCL_ERROR);
}

static int Cmd_NewBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a new buffer with the passed name.
// If a buffer already exists with the passed name fail.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		ErrorTraceClear();
		if((buffer=EditorBufferNew(Tcl_GetString(objv[1]))))				// create a new buffer
		{
			Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the current list of document buffers.
{
	bool
		fail;
	EDITOR_BUFFER
		*testBuffer;

	fail=false;
	if(objc==1)
	{
		testBuffer=EditorBufferGetFirst();
		while(testBuffer)
		{
			Tcl_AppendElement(localInterpreter,testBuffer->contentName);
			testBuffer=EditorBufferGetNext(testBuffer);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SaveBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Save the passed buffer.
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				result=EditorBufferSave(buffer);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SaveBufferAs(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Save the passed buffer to newPath.
// Rename the buffer to the new name, set clean.
// NOTE: this returns the new name of buffer after it is saved.
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				result=EditorBufferSaveAs(buffer,Tcl_GetString(objv[2]));
				if(result)
				{
					Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName newPath");
	}
	return(TCL_ERROR);
}

static int Cmd_SaveBufferTo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Save the passed buffer to a file, do not modify the buffer's "clean" status.
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				result=EditorBufferSaveTo(buffer,Tcl_GetString(objv[2]));
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName pathName");
	}
	return(TCL_ERROR);
}

static int Cmd_RevertBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Revert the passed buffer.
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				result=EditorBufferRevert(buffer);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferVariables(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Show all the variables linked to the passed buffer.
{
	EDITOR_BUFFER
		*buffer;
	VARIABLE_BINDING
		*currentBinding;
	bool
		didOne;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			currentBinding=buffer->variableTable.variableBindingListHead;
			didOne=false;
			while(currentBinding)
			{
				if(didOne)				// do not put return on last line
				{
					Tcl_AppendResult(localInterpreter,"\n",NULL);
				}
				Tcl_AppendElement(localInterpreter,VariableBindingNameGet(currentBinding));
				Tcl_AppendElement(localInterpreter,VariableBindingTextGet(currentBinding));
				didOne=true;
				currentBinding=VariableBindingNextLocate(currentBinding);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetBufferVariable(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set/remove a buffer relative variable.
{
	EDITOR_BUFFER
		*buffer;
	char
		*variableName;

	if(objc==3||objc==4)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			variableName=Tcl_GetString(objv[2]);
			if(objc==3)
			{
				if(VariableBindingDelete(&(buffer->variableTable),Tcl_GetString(objv[2])))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to locate variable '",Tcl_GetString(objv[2]),"'",NULL);
				}
			}
			else
			{
				ErrorTraceClear();
				if(VariableBindingCreate(&(buffer->variableTable),variableName,Tcl_GetString(objv[3])))
				{
					Tcl_AppendResult(localInterpreter,Tcl_GetString(objv[3]),NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName variableName ?variableText?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetBufferVariable(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get a variable hung off of a buffer.
// These variables can be used by Tcl scripts to mark buffers, and associate
// arbitrary data with them.
{
	EDITOR_BUFFER
		*buffer;
	VARIABLE_BINDING
		*bufferVariable;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if((bufferVariable=VariableBindingLocate(&(buffer->variableTable),Tcl_GetString(objv[2]))))
			{
				Tcl_AppendResult(localInterpreter,VariableBindingTextGet(bufferVariable),NULL);
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Failed to locate variable '",Tcl_GetString(objv[2]),"'",NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName variableName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferDirty(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return 1 if the passed buffer is dirty, 0 if not.
// If the buffer cannot be located, fail.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(UndoCleanPointAt(buffer))
			{
				Tcl_AppendResult(localInterpreter,"0",NULL);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"1",NULL);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_FromFile(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return 1 if the passed buffer is linked to a file, 0 if not.
// If the buffer cannot be located, fail.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(buffer->fromFile)
			{
				Tcl_AppendResult(localInterpreter,"1",NULL);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"0",NULL);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BufferExists(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return 1 if the passed buffer exists, 0 if not.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=EditorBufferLocate(Tcl_GetString(objv[1]))))
		{
			Tcl_AppendResult(localInterpreter,"1",NULL);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"0",NULL);
		}
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_HasWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return 1 if the passed buffer has a document window associated with it, 0 if not.
// If the buffer cannot be located, fail.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(PlatformDocumentWindowLocate(buffer))
			{
				Tcl_AppendResult(localInterpreter,"1",NULL);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"0",NULL);
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_ClearDirty(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Clear the dirty flag for the passed buffer.
// If the buffer cannot be located, fail.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				UndoCleanPointSet(buffer);				// make it clean again
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_TextInfo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the number of lines, and number of characters of text in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				snprintf(tempString,sizeof(tempString),"%" PRIu64 " %" PRIu64 "",TextUniverseTotalLines(buffer->textUniverse),TextUniverseTotalBytes(buffer->textUniverse));
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_PositionToLineOffset(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Given a position in a buffer, return a line number, and offset in characters into the line
// that matches the position.
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here.
{
	TEXT_STREAM
		stream;
	EDITOR_BUFFER
		*buffer;
	uint64_t
		position,
		line;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&position))
			{
				if(ShellBufferNotBusy(localInterpreter,buffer))
				{
					line=TextUniversePositionToLine(buffer->textUniverse,position,&stream);
					snprintf(tempString,sizeof(tempString),"%" PRIu64 " %" PRIu64 "",line+1,position-TextStreamPositionGet(&stream));
					Tcl_AppendResult(localInterpreter,tempString,NULL);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName position");
	}
	return(TCL_ERROR);
}

static int Cmd_LineOffsetToPosition(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Given a line, and offset in a buffer, return a position in the buffer that matches.
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here.
{
	TEXT_STREAM
		stream;
	EDITOR_BUFFER
		*buffer;
	uint64_t
		position,
		line,
		lineOffset,
		distanceMoved;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==4)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&line))
			{
				if(TclUInt64Get(localInterpreter,objv[3],&lineOffset))
				{
					if(ShellBufferNotBusy(localInterpreter,buffer))
					{
						if(line==0)
						{
							line=1;
						}
						position=TextUniverseLineToPosition(buffer->textUniverse,line-1,&stream);
						TextStreamToNextLine(&stream);
						distanceMoved=TextStreamPositionGet(&stream)-position;
						if(lineOffset>distanceMoved)		// see if offset is past end of line, if so, pin it
						{
							lineOffset=distanceMoved;
						}
						snprintf(tempString,sizeof(tempString),"%" PRIu64 "",position+lineOffset);
						Tcl_AppendResult(localInterpreter,tempString,NULL);
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName lineNumber offset");
	}
	return(TCL_ERROR);
}

static int Cmd_CheckBuffer(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Check the text universe on the given buffer, report the results.
// NOTE: this is mainly for debugging.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(TextUniverseSanityCheck(buffer->textUniverse))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_CloseWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Close a given window.
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			EditorDocumentWindowClose(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_OpenWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Open a window onto the given buffer.
// If there is already a window open on the buffer, complain.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	bool
		useUtf8;
	uint32_t
		tabSize;

	if(objc==10)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(TclRectangleGet(localInterpreter,&(objv[2]),&rect))
			{
				if(TclBooleanGet(localInterpreter,objv[6],&useUtf8))
				{
					if(TclTabSizeGet(localInterpreter,objv[7],&tabSize))
					{
						if(ShellBufferNotBusy(localInterpreter,buffer))
						{
							ErrorTraceClear();
							window=EditorDocumentWindowOpen(buffer,&rect,useUtf8,tabSize,Tcl_GetString(objv[8]),Tcl_GetString(objv[9]));
							if(window)
							{
								return(TCL_OK);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName x y w h useUtf8 tabSize fontName colorScheme");
	}
	return(TCL_ERROR);
}

static int Cmd_WindowList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the current list of document windows.
{
	bool
		fail;
	uint32_t
		windowIndex,
		numWindows;
	EDITOR_WINDOW
		**windowList;
	EDITOR_BUFFER
		*buffer;

	fail=false;
	if(objc==1)
	{
		if(PlatformDocumentWindowSortedListGet(&numWindows,&windowList))
		{
			for(windowIndex=0;windowIndex<numWindows;windowIndex++)
			{
				buffer=PlatformDocumentWindowBufferGet(windowList[windowIndex]);
				Tcl_AppendElement(localInterpreter,buffer->contentName);
			}
			MDisposePtr(windowList);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_GetTopLeft(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the top line, and lefthand coordinate of the active view in the given window.
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	int64_t
		leftPixel;
	int64_t
		topLine;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			ViewOriginGet(view,&leftPixel,&topLine);
			snprintf(tempString,sizeof(tempString),"%" PRIu64 " %" PRIi64 "",topLine+1,leftPixel);
			Tcl_AppendResult(localInterpreter,tempString,NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetTopLeft(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the top line, and lefthand coordinate of the active view in the given window.
// NOTE: to be consistent with line numbering starting at 1, we fudge the number in here.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	uint64_t
		topLine;
	int32_t
		leftPixel;

	if(objc==4)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&topLine))
			{
				if(TclInt32Get(localInterpreter,objv[3],&leftPixel))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);
					if(topLine==0)				// if line 0 was given, just make it line 1
					{
						topLine=1;
					}
					ViewOriginSet(view,leftPixel,topLine-1);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName topLine leftPixel");
	}
	return(TCL_ERROR);
}

static const OPTION_DESCRIPTOR_RECORD
	homeOptions[]=
	{
		{"-lenient",		HT_LENIENT,		TclOptionEnum},
		{"-semistrict",		HT_SEMI_STRICT,	TclOptionEnum},
		{"-strict",			HT_STRICT,		TclOptionEnum},
		{"-start",			false,			TclOptionBoolean},
		{"-end",			true,			TclOptionBoolean},
		{NULL,0,NULL}
	};

static int Cmd_HomeWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Home a document view to a given position, in the given way.
// By default, the mode is HT_SEMI_STRICT, and useEnd is false.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_VIEW
		*view;
	uint64_t
		startPosition,
		endPosition;
	bool
		fail;
	int
		mode;
	bool
		useEnd;
	void
		*parsedVars[]=
		{
			(void *)&mode,
			(void *)&mode,
			(void *)&mode,
			(void *)&useEnd,
			(void *)&useEnd,
		};
	int
		optionIndex;

	fail=false;
	if(objc>=4)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&startPosition))
			{
				if(TclUInt64Get(localInterpreter,objv[3],&endPosition))
				{
					buffer=PlatformDocumentWindowBufferGet(window);
					startPosition=PositionForceIntoRange(buffer,startPosition);
					endPosition=PositionForceIntoRange(buffer,endPosition);
					PositionsArrange(&startPosition,&endPosition);
					optionIndex=4;

					mode=HT_SEMI_STRICT;
					useEnd=false;

					if(TclOptionsParse(localInterpreter,homeOptions,parsedVars,objc,objv,&optionIndex,true))
					{
						view=PlatformDocumentWindowCurrentViewGet(window);
						EditorViewToRangeHome(view,startPosition,endPosition,useEnd,mode,mode);
					}
					else
					{
						fail=true;
					}

				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName startPosition endPosition ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_GetRect(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the rectangle of a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	bool
		result;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			ErrorTraceClear();
			result=PlatformDocumentWindowRectGet(window,&rect);
			if(result)
			{
				snprintf(tempString,sizeof(tempString),"%d %d %d %d",(int)rect.x,(int)rect.y,(int)rect.w,(int)rect.h);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetRect(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the rectangle of a window (relative to the primary screen).
{
	EDITOR_WINDOW
		*window;
	EDITOR_RECT
		rect;
	bool
		result;

	if(objc==6)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			if(TclRectangleGet(localInterpreter,&(objv[2]),&rect))
			{
				ErrorTraceClear();
				result=PlatformDocumentWindowRectSet(window,&rect);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName x y w h");
	}
	return(TCL_ERROR);
}

static int Cmd_GetUtf8(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the flag associated with the view of a window that tells if the
// view is interpreting text as UTF8, or just ASCII.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	bool
		useUtf8;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			useUtf8=ViewUtf8Get(view);
			snprintf(tempString,sizeof(tempString),"%d",(int)useUtf8);
			Tcl_AppendResult(localInterpreter,tempString,NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetUtf8(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the flag associated with the view of a window that tells if the
// view is interpreting text as UTF8, or just ASCII.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	bool
		useUtf8;
	bool
		result;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			if(TclBooleanGet(localInterpreter,objv[2],&useUtf8))
			{
				view=PlatformDocumentWindowCurrentViewGet(window);
				ErrorTraceClear();
				result=ViewUtf8Set(view,useUtf8);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName useUtf8");
	}
	return(TCL_ERROR);
}

static int Cmd_GetTabSize(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the tab size of a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	uint32_t
		tabSize;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			tabSize=ViewTabSizeGet(view);
			snprintf(tempString,sizeof(tempString),"%d",(int)tabSize);
			Tcl_AppendResult(localInterpreter,tempString,NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetTabSize(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the tab size of a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	uint32_t
		tabSize;
	bool
		result;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			if(TclTabSizeGet(localInterpreter,objv[2],&tabSize))
			{
				view=PlatformDocumentWindowCurrentViewGet(window);
				ErrorTraceClear();
				result=ViewTabSizeSet(view,tabSize);
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName tabSize");
	}
	return(TCL_ERROR);
}

static int Cmd_GetFont(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the font of a document window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	char
		fontName[MAX_FONT_NAME_LENGTH];

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			ViewFontNameGet(view,fontName);
			Tcl_AppendResult(localInterpreter,fontName,NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetFont(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the font of a document window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	char
		*fontName;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			fontName=Tcl_GetString(objv[2]);
			view=PlatformDocumentWindowCurrentViewGet(window);
			ViewFontNameSet(view,fontName);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName fontName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetColors(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get the colors of a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			ViewColorsGet(view,&backgroundColor,&foregroundColor);

			snprintf(tempString,sizeof(tempString),"%08X ",(int)backgroundColor);
			Tcl_AppendResult(localInterpreter,tempString,NULL);

			snprintf(tempString,sizeof(tempString),"%08X",(int)foregroundColor);
			Tcl_AppendResult(localInterpreter,tempString,NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_SetColors(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the foreground and background colors of a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor;

	if(objc==4)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			if(TclColorGet(localInterpreter,objv[2],&backgroundColor))
			{
				if(TclColorGet(localInterpreter,objv[3],&foregroundColor))
				{
					ViewColorsSet(view,backgroundColor,foregroundColor);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName backgroundColor foregroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_SetColorScheme(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the colors scheme assigned to the view of a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			view=PlatformDocumentWindowCurrentViewGet(window);
			ErrorTraceClear();
			if(ViewColorSchemeSet(view,Tcl_GetString(objv[2])))
			{
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName colorScheme");
	}
	return(TCL_ERROR);
}

static int Cmd_SetTopWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Make the passed window the top/active window.
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			PlatformDocumentWindowTopSet(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_MinimizeWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Attempt to minimize (in whatever way the current OS supports) the given document window.
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			PlatformDocumentWindowMinimize(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_UnminimizeWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Attempt to un-minimize (in whatever way the current OS supports) the given document window.
{
	EDITOR_WINDOW
		*window;

	if(objc==2)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			PlatformDocumentWindowUnminimize(window);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetActiveWindow(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the active window name, or error if none.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;

	if(objc==1)
	{
		if((window=PlatformDocumentWindowActiveGet()))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			Tcl_AppendResult(localInterpreter,buffer->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"No active window",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_FontNameSplit(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Split a composite font name into its component parts.
{
	char
		familyName[MAX_FONT_NAME_LENGTH];
	uint32_t
		size;
	bool
		bold,
		oblique;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==2)
	{
		PlatformFontCompositeNameSplit(Tcl_GetString(objv[1]),familyName,&size,&bold,&oblique);
		Tcl_AppendElement(localInterpreter,familyName);

		snprintf(tempString,sizeof(tempString),"%d",size);
		Tcl_AppendElement(localInterpreter,tempString);

		snprintf(tempString,sizeof(tempString),"%d",bold);
		Tcl_AppendElement(localInterpreter,tempString);

		snprintf(tempString,sizeof(tempString),"%d",oblique);
		Tcl_AppendElement(localInterpreter,tempString);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"fontName");
	}
	return(TCL_ERROR);
}

static int Cmd_FontNameJoin(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Join the components of a font specification into a composite name.
{
	uint32_t
		size;
	bool
		bold,
		oblique;
	char
		compositeName[MAX_FONT_NAME_LENGTH];

	if(objc==5)
	{
		if(TclUInt32Get(localInterpreter,objv[2],&size))
		{
			if(TclBooleanGet(localInterpreter,objv[3],&bold))
			{
				if(TclBooleanGet(localInterpreter,objv[4],&oblique))
				{
					PlatformFontComponentNameJoin(Tcl_GetString(objv[1]),size,bold,oblique,compositeName);
					Tcl_AppendResult(localInterpreter,compositeName,NULL);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"familyName size bold oblique");
	}
	return(TCL_ERROR);
}

static int Cmd_Extract(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Returns the bytes given in a range of positions for a buffer.
{
	TEXT_STREAM
		stream;
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	uint8_t
		*tempBuffer;

	if(objc==4)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&startPosition))
			{
				if(TclUInt64Get(localInterpreter,objv[3],&endPosition))
				{
					startPosition=PositionForceIntoRange(buffer,startPosition);
					endPosition=PositionForceIntoRange(buffer,endPosition);
					PositionsArrange(&startPosition,&endPosition);

					if((tempBuffer=(uint8_t *)MNewPtr(endPosition-startPosition+1)))
					{
						TextStreamPositionSet(&stream,buffer->textUniverse,startPosition);
						TextStreamTextExtract(&stream,tempBuffer,endPosition-startPosition,NULL);
		 				tempBuffer[endPosition-startPosition]='\0';
						Tcl_AppendElement(localInterpreter,(char *)tempBuffer);
						MDisposePtr(tempBuffer);
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Failed to allocate memory for result",NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName startPosition endPosition");
	}
	return(TCL_ERROR);
}

static int Cmd_Insert(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Insert the given text into the current selection of the passed buffer.
// NOTE: Tcl will convert binary values in the text to UTF-8 representation
// for the insertion. So, if the passed text contains codepoints above 127,
// they'll be inserted as UTF-8.
{
	EDITOR_BUFFER
		*buffer;
	char
		*dataToAdd;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				dataToAdd=Tcl_GetString(objv[2]);
				ErrorTraceClear();
				if(EditorInsert(buffer,(uint8_t *)dataToAdd,strlen(dataToAdd)))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName textToInsert");
	}
	return(TCL_ERROR);
}

static int Cmd_InsertBinary(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Insert the given binary data into the current selection of the passed buffer.
// This treats the passed Tcl data as a chunk of binary bytes, and dumps them
// directly into the buffer.
{
	EDITOR_BUFFER
		*buffer;
	uint8_t
		*dataToAdd;
	int
		length;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				dataToAdd=Tcl_GetByteArrayFromObj(objv[2],&length);
				ErrorTraceClear();
				if(EditorInsert(buffer,dataToAdd,length))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName binaryToInsert");
	}
	return(TCL_ERROR);
}

static int Cmd_InsertFile(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Insert the given file into the current selection
// of the passed buffer.
{
	EDITOR_BUFFER
		*buffer;
	bool
		result;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				result=EditorFileInsert(buffer,Tcl_GetString(objv[2]));
				if(result)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName pathName");
	}
	return(TCL_ERROR);
}

static int Cmd_AutoIndent(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform an autoindent on the passed buffer.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorAutoIndent(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_MoveCursor(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Move the cursor in the given window by the relative mode.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_VIEW
		*view;
	int
		relativeMode;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TokenMatch(Tcl_GetString(objv[2]),cursorMovementTokens,&relativeMode))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);

					parseMark.buffer=buffer;
					parseMark.mark=NULL;

					optionIndex=3;
					if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
					{
						selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

						ErrorTraceClear();
						if(EditorCursorMove(view,selection,(uint16_t)relativeMode))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_Clear(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a clear in the given buffer. This deletes any characters
// which are currently selected. If none are selected, nothing happens.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorSelectionDelete(buffer,buffer->selectionUniverse))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Delete(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Delete the characters in the window between the current cursor position, and
// the given relative mode away, if there is a selection, it will be deleted instead.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TokenMatch(Tcl_GetString(objv[2]),cursorMovementTokens,&relativeMode))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);
					ErrorTraceClear();
					if(EditorDelete(view,(uint16_t)relativeMode))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_ExpandSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Expand the selection in the given window in the given relative direction.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TokenMatch(Tcl_GetString(objv[2]),cursorMovementTokens,&relativeMode))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);
					ErrorTraceClear();
					if(EditorNormalSelectionExpand(view,view->parentBuffer->selectionUniverse,(uint16_t)relativeMode))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_ReduceSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Reduce the selection in the given window in the given relative direction.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_VIEW
		*view;
	int
		relativeMode;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TokenMatch(Tcl_GetString(objv[2]),cursorMovementTokens,&relativeMode))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);
					ErrorTraceClear();
					if(EditorNormalSelectionReduce(view,view->parentBuffer->selectionUniverse,(uint16_t)relativeMode))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Invalid relative mode",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName relativeMode");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectionInfo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return information about the selection:
// startPosition,endPosition,startLine,endLine,startLinePosition,endLinePosition,totalSegments,totalSpan
// NOTE: to be consistent with line numbering starting at 1, we fudge the numbers in here
// ### obsolete -- build out of Tcl
{
	uint64_t
		startPosition,						// selection information
		endPosition,
		startLine,
		endLine,
		startLinePosition,
		endLinePosition,
		totalSegments,
		totalSpan;
	EDITOR_BUFFER
		*buffer;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			parseMark.buffer=buffer;
			parseMark.mark=NULL;

			optionIndex=2;
			if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
			{
				selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

				EditorSelectionInfoGet(buffer,selection,&startPosition,&endPosition,&startLine,&endLine,&startLinePosition,&endLinePosition,&totalSegments,&totalSpan);
				snprintf(tempString,sizeof(tempString),"%" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 "",startPosition,endPosition,startLine+1,endLine+1,startLinePosition,endLinePosition,totalSegments,totalSpan);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectAll(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a select all in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				parseMark.buffer=buffer;
				parseMark.mark=NULL;

				optionIndex=2;
				if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
				{
					selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

					ErrorTraceClear();
					if(EditorSelectAll(buffer,selection))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectedTextList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the currently selected text as a list -- one element per selection.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		currentPosition;
	uint64_t
		actualLength;
	uint8_t
		*selectionBuffer;
	bool
		done;
	bool
		fail;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	fail=false;
	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			parseMark.buffer=buffer;
			parseMark.mark=NULL;

			optionIndex=2;
			if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
			{
				selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

				currentPosition=0;
				done=false;
				while(!done&&!fail)
				{
					ErrorTraceClear();
					if((selectionBuffer=EditorNextSelectionToBuffer(buffer,selection,&currentPosition,1,&actualLength,&done)))
					{
						selectionBuffer[actualLength-1]='\0';					// terminate the string
						Tcl_AppendElement(localInterpreter,(char *)selectionBuffer);
						MDisposePtr(selectionBuffer);
					}
					else
					{
						if(!done)
						{
							Tcl_ResetResult(localInterpreter);
							Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
							fail=true;
						}
					}
				}
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SelectLine(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Locate the given line, and select it.
// If lineNumber is out of range, go to the nearest one.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		line;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TclUInt64Get(localInterpreter,objv[2],&line))
				{
					if(line==0)		// if user enters 0, just take it as 1
					{
						line=1;
					}
					parseMark.buffer=buffer;
					parseMark.mark=NULL;

					optionIndex=3;
					if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
					{
						selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

						ErrorTraceClear();
						if(EditorLineLocate(buffer,selection,line-1))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName lineNumber ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionEnds(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the start and end of the current selection
// of the passed buffer.
// If there is no selection, the cursor position is returned in both values.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			parseMark.buffer=buffer;
			parseMark.mark=NULL;

			optionIndex=2;
			if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
			{
				selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

				SelectionEndPositionsGet(selection,&startPosition,&endPosition);
				snprintf(tempString,sizeof(tempString),"%" PRIu64  " %" PRIu64 "",startPosition,endPosition);
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_SetSelectionEnds(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the selection of the passed buffer.
// If the given end position is < the start position,
// the positions will be reversed.
// If any position is out of range, it will be forced in range.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=4)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TclUInt64Get(localInterpreter,objv[2],&startPosition))
				{
					if(TclUInt64Get(localInterpreter,objv[3],&endPosition))
					{
						startPosition=PositionForceIntoRange(buffer,startPosition);
						endPosition=PositionForceIntoRange(buffer,endPosition);
						PositionsArrange(&startPosition,&endPosition);

						parseMark.buffer=buffer;
						parseMark.mark=NULL;

						optionIndex=4;
						if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
						{
							selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

							ErrorTraceClear();
							if(EditorSelectionSet(buffer,selection,startPosition,endPosition))
							{
								return(TCL_OK);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
							}
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName startPosition endPosition ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionEndList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return offsets of the ends of each selection segment.
// If there is no selection, an empty list is returned.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		currentPosition,
		currentLength;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			parseMark.buffer=buffer;
			parseMark.mark=NULL;

			optionIndex=2;
			if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
			{
				selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

				currentPosition=0;
				while(SelectionAtOrAfterPositionGet(selection,currentPosition,&currentPosition,&currentLength))
				{
					snprintf(tempString,sizeof(tempString),"%" PRIu64 " %" PRIu64 "",currentPosition,currentPosition+currentLength);
					Tcl_AppendElement(localInterpreter,tempString);
					currentPosition+=currentLength;
				}
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_AddSelectionEndList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Add list of ends given to the selection.
// If any position is out of range, it will be forced in range.
// If there is any overlap, the new segment boundaries will remain.
{
	EDITOR_BUFFER
		*buffer;
	int
		listArgc;
	Tcl_Obj
		**listArgv;
	int
		elementArgc;
	Tcl_Obj
		**elementArgv;
	uint64_t
		startPosition,
		endPosition;
	int
		i;
	bool
		fail;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(Tcl_ListObjGetElements(localInterpreter,objv[2],&listArgc,&listArgv)==TCL_OK)
				{
					parseMark.buffer=buffer;
					parseMark.mark=NULL;

					optionIndex=3;
					if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
					{
						selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

						EditorSelectionChangeStart(buffer,selection);
						fail=false;
						for(i=0;!fail&&i<listArgc;i++)
						{
							if(Tcl_ListObjGetElements(localInterpreter,listArgv[i],&elementArgc,&elementArgv)==TCL_OK)
							{
								if(elementArgc==2)
								{
									if(TclUInt64Get(localInterpreter,elementArgv[0],&startPosition))
									{
										if(TclUInt64Get(localInterpreter,elementArgv[1],&endPosition))
										{
											startPosition=PositionForceIntoRange(buffer,startPosition);
											endPosition=PositionForceIntoRange(buffer,endPosition);
											PositionsArrange(&startPosition,&endPosition);
											if(endPosition>startPosition)	// make sure there is at least something to select
											{
												if(!SelectionRangeSet(selection,startPosition,endPosition-startPosition))
												{
													Tcl_AppendResult(localInterpreter,"Failed to set selection",NULL);
													fail=true;
												}
											}
										}
										else
										{
											fail=true;
										}
									}
									else
									{
										fail=true;
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Exactly two elements are allowed per list entry",NULL);
									fail=true;
								}
							}
							else
							{
								fail=true;
							}
						}
						EditorSelectionChangeEnd(buffer,selection);
						if(!fail)
						{
							return(TCL_OK);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName selectionList ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSelectionAtPosition(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return information about the selection at the given character position in the text.
// The result is a value (0 if no selection, 1 if selection) and the position <= the given
// one where the current selection begins, and the position just after the end of the
// selection.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		position;
	uint64_t
		startPosition,
		numElements;
	bool
		isSelected;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;


	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&position))
			{
				parseMark.buffer=buffer;
				parseMark.mark=NULL;

				optionIndex=3;
				if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
				{
					selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

					position=PositionForceIntoRange(buffer,position);
					if(!SelectionRangeGet(selection,position,&startPosition,&numElements,&isSelected))
					{
						numElements=TextUniverseTotalBytes(buffer->textUniverse)-startPosition;
					}
					snprintf(tempString,sizeof(tempString),"%u %" PRIu64 " %" PRIu64 "",isSelected?1:0,startPosition,startPosition+numElements);
					Tcl_AppendResult(localInterpreter,tempString,NULL);
					return(TCL_OK);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName position ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_CopySelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Copy selection data from one selection to another.
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-sourcemark",		0,		TclOptionMark},
			{"-destmark",		0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					EditorSelectionChangeStart(buffer,destSelection);
					SelectionUniverseCopy(sourceSelection,destSelection);	// perform the operation
					EditorSelectionChangeEnd(buffer,destSelection);
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_AndSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Logical AND selection data from one selection to another.
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-sourcemark",		0,		TclOptionMark},
			{"-destmark",		0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					EditorSelectionChangeStart(buffer,destSelection);
					SelectionUniverseAnd(sourceSelection,destSelection);	// perform the operation
					EditorSelectionChangeEnd(buffer,destSelection);
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_OrSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Logical OR selection data from one selection to another.
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-sourcemark",		0,		TclOptionMark},
			{"-destmark",		0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					EditorSelectionChangeStart(buffer,destSelection);
					SelectionUniverseOr(sourceSelection,destSelection);	// perform the operation
					EditorSelectionChangeEnd(buffer,destSelection);
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_XorSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Logical OR selection data from one selection to another.
{
	EDITOR_BUFFER
		*buffer;
	PARSE_MARK
		sourceMark,
		destMark;
	SELECTION_UNIVERSE
		*sourceSelection,
		*destSelection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-sourcemark",		0,		TclOptionMark},
			{"-destmark",		0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&sourceMark,
			(void *)&destMark,
		};
	int
		optionIndex;
	bool
		fail;

	fail=false;
	if(objc>=2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				optionIndex=2;

				sourceMark.buffer=buffer;
				sourceMark.mark=NULL;

				destMark.buffer=buffer;
				destMark.mark=NULL;

				if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
				{
					sourceSelection=sourceMark.mark?sourceMark.mark->selectionUniverse:buffer->selectionUniverse;
					destSelection=destMark.mark?destMark.mark->selectionUniverse:buffer->selectionUniverse;

					EditorSelectionChangeStart(buffer,destSelection);
					SelectionUniverseXor(sourceSelection,destSelection);	// perform the operation
					EditorSelectionChangeEnd(buffer,destSelection);
				}
				else
				{
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_MarkInvisible(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Make a mark invisible within a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;
	EDITOR_VIEW
		*view;

	if(objc==3)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			if((mark=EditorMarkLocate(buffer,Tcl_GetString(objv[2]))))
			{
				view=PlatformDocumentWindowCurrentViewGet(window);
				EditorMarkVisibleClear(mark,view);
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Failed to locate mark '",Tcl_GetString(objv[2]),"'",NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName markName");
	}
	return(TCL_ERROR);
}

static int Cmd_MarkVisible(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Make a mark visible in a window.
{
	EDITOR_WINDOW
		*window;
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;
	EDITOR_VIEW
		*view;
	EDITOR_COLOR
		background;

	if(objc==4)
	{
		if((window=TclEditorWindowGet(localInterpreter,objv[1])))
		{
			buffer=PlatformDocumentWindowBufferGet(window);
			if((mark=EditorMarkLocate(buffer,Tcl_GetString(objv[2]))))
			{
				if(TclColorGet(localInterpreter,objv[3],&background))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);
					EditorMarkVisibleSet(mark,view,background);
					return(TCL_OK);
				}
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Failed to locate mark '",Tcl_GetString(objv[2]),"'",NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"windowName markName backgroundColor");
	}
	return(TCL_ERROR);
}

static int Cmd_MarkList(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return a list of marks for the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			mark=buffer->marks;
			while(mark)
			{
				if(mark->markName)
				{
					Tcl_AppendElement(localInterpreter,mark->markName);
				}
				mark=mark->nextMark;
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_CloseMark(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Remove a mark from a buffer.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if((mark=EditorMarkLocate(buffer,Tcl_GetString(objv[2]))))
				{
					EditorMarkClose(buffer,mark);
				}
				if(mark)
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to locate mark '",Tcl_GetString(objv[2]),"'",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName markName");
	}
	return(TCL_ERROR);
}

static int Cmd_OpenMark(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Open a mark in the given buffer. If there's already a mark with the
// given name, do nothing.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_MARK
		*mark;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			ErrorTraceClear();
			mark=EditorMarkOpen(buffer,Tcl_GetString(objv[2]));
			if(mark)
			{
				return(TCL_OK);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName markName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetStyleAtPosition(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return information about the style at the given character position in the text.
// The result is a value (style index), the position <= the given
// one where the style begins, and the position just after the end of the
// style.
// If there is no style at the position, return an empty string for
// the style field.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		position;
	uint64_t
		startPosition,
		numElements;
	uint32_t
		style;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(TclUInt64Get(localInterpreter,objv[2],&position))
			{
				position=PositionForceIntoRange(buffer,position);
				if(!StyleRangeGet(buffer->styleUniverse,position,&startPosition,&numElements,&style))
				{
					snprintf(tempString,sizeof(tempString),"\"\" %" PRIu64 " %" PRIu64 "",startPosition,startPosition+numElements);
				}
				else
				{
					snprintf(tempString,sizeof(tempString),"%u %" PRIu64 " %" PRIu64 "",style,startPosition,startPosition+numElements);
				}
				Tcl_AppendResult(localInterpreter,tempString,NULL);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName position");
	}
	return(TCL_ERROR);
}

static int Cmd_SetStyle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the style of text of the passed buffer.
// If the given end position is < the start position,
// the positions will be reversed.
// If any position is out of range, it will be forced in range.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	uint32_t
		style;

	if(objc==5)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TclUInt64Get(localInterpreter,objv[2],&startPosition))
				{
					if(TclUInt64Get(localInterpreter,objv[3],&endPosition))
					{
						if(TclUInt32Get(localInterpreter,objv[4],&style))
						{
							style=ViewStyleForceIntoRange(style);
							startPosition=PositionForceIntoRange(buffer,startPosition);
							endPosition=PositionForceIntoRange(buffer,endPosition);
							PositionsArrange(&startPosition,&endPosition);
							StyleRangeSet(buffer->styleUniverse,startPosition,endPosition-startPosition,style);
							return(TCL_OK);
						}
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName startPosition endPosition styleNum");
	}
	return(TCL_ERROR);
}

static int Cmd_SelectionToStyle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the style of a buffer to a given value anywhere text is selected.
{
	EDITOR_BUFFER
		*buffer;
	uint32_t
		style;
	bool
		fail;
	uint64_t
		startPosition,
		numElements;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TclUInt32Get(localInterpreter,objv[2],&style))
				{
					style=ViewStyleForceIntoRange(style);
					fail=false;

					parseMark.buffer=buffer;
					parseMark.mark=NULL;

					optionIndex=3;
					if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
					{
						selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

						startPosition=0;
						while(!fail&&SelectionAtOrAfterPositionGet(selection,startPosition,&startPosition,&numElements))
						{
							fail=!StyleRangeSet(buffer->styleUniverse,startPosition,numElements,style);
							startPosition+=numElements;
						}
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName styleNum ?options?");
	}
	return(TCL_ERROR);
}

static int Cmd_StyleToSelection(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the selection in a buffer everywhere a given style is in the buffer.
// NOTE: this does not clear the old selection, it will just add to it.
{
	EDITOR_BUFFER
		*buffer;
	uint32_t
		style;
	bool
		fail;
	uint64_t
		startPosition,
		numElements;
	uint32_t
		sawStyle;
	PARSE_MARK
		parseMark;
	SELECTION_UNIVERSE
		*selection;
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-mark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};
	void
		*parsedVars[]=
		{
			(void *)&parseMark,
		};
	int
		optionIndex;

	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(TclUInt32Get(localInterpreter,objv[2],&style))
				{
					style=ViewStyleForceIntoRange(style);

					parseMark.buffer=buffer;
					parseMark.mark=NULL;

					optionIndex=3;
					if(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,&optionIndex,true))
					{
						selection=parseMark.mark?parseMark.mark->selectionUniverse:buffer->selectionUniverse;

						EditorSelectionChangeStart(buffer,selection);

						fail=false;
						startPosition=0;
						do
						{
							if(StyleRangeGet(buffer->styleUniverse,startPosition,&startPosition,&numElements,&sawStyle))
							{
								if(sawStyle==style)
								{
									fail=!SelectionRangeSet(selection,startPosition,numElements);
								}
							}
							startPosition+=numElements;
						}
						while(!fail&&numElements);

						EditorSelectionChangeEnd(buffer,selection);
						return(TCL_OK);
					}
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName styleNum ?options?");
	}
	return(TCL_ERROR);
}

static void SplitExpressionNameDispose(char *name)
// Get rid of a split expression name created by ExpressionNameSplit.
{
	MDisposePtr(name);
}

static char *ExpressionNameSplit(const char *name,bool *useRegisterMatch,uint32_t *registerIndex)
// Split name into a name, and information about any register match.
// Expression names may be specified as just a name, or a name followed by a ':'
// and then a register number.
// NOTE: if this fails, it will ErrorSet, and return NULL.
// NOTE: if this succeeds, the returned name should be disposed of by calling
// SplitExpressionNameDispose.
{
	int
		length;
	char
		*newName;

	*useRegisterMatch=false;

	length=strlen(name);
	if((newName=(char *)MNewPtr(length+1)))		// +1 to get space for termination
	{
		strcpy(newName,name);					// copy it over
		if(length>=2)
		{
			if(newName[length-2]==':'&&newName[length-1]>='0'&&newName[length-1]<='9')	// see if indication of register specification
			{
				*useRegisterMatch=true;
				*registerIndex=newName[length-1]-'0';
				newName[length-2]='\0';	// terminate
			}
		}
	}
	return(newName);
}

static EXPRESSION_PIECE *ExpressionPieceGet(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,Tcl_Obj *const name)
// See if an expression piece of the given name can be added to the list.
// If there is a problem, add it to the Tcl result, and return NULL.
{
	char
		*newName;
	bool
		useRegisterMatch;
	uint32_t
		registerIndex;
	SYNTAX_EXPRESSION
		*expression;
	EXPRESSION_PIECE
		*piece;

	piece=NULL;
	registerIndex=0;		// keep GCC quiet
	if((newName=ExpressionNameSplit(Tcl_GetString(name),&useRegisterMatch,&registerIndex)))
	{
		if((expression=SyntaxMapExpressionLocate(map,newName)))
		{
			ErrorTraceClear();
			if(!(piece=SyntaxMapExpressionAddPiece(expression,useRegisterMatch,registerIndex)))
			{
				Tcl_AppendResult(localInterpreter,"Failed to add piece to expression '",newName,"': ",ErrorTraceGet(),NULL);
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Failed to locate expression '",newName,"'",NULL);
		}
		SplitExpressionNameDispose(newName);
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Failed to split name '",Tcl_GetString(name),"'",NULL);
	}
	return(piece);
}

static bool EndExpressionsParse(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,SYNTAX_STYLE_MAPPING *mapping,Tcl_Obj *const endExpressionList)
// Parse up the ending expressions, and add them to mapping.
// If there is a problem, add it to the Tcl result, and return false.
{
	EXPRESSION_PIECE
		*expressionPiece;
	int
		argc;
	Tcl_Obj
		**argv;
	int
		i;
	bool
		fail;

	fail=false;
	if(Tcl_ListObjGetElements(localInterpreter,endExpressionList,&argc,&argv)==TCL_OK)	// get list of end expressions
	{
		for(i=0;!fail&&i<argc;i++)
		{
			if((expressionPiece=ExpressionPieceGet(localInterpreter,map,argv[i])))
			{
				ErrorTraceClear();
				if(SyntaxMapMappingEndExpressionPieceAdd(mapping,expressionPiece))
				{
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Failed to end expression '",Tcl_GetString(argv[i]),"' to mapping: ",ErrorTraceGet(),NULL);
					fail=true;
				}
			}
			else
			{
				fail=true;
			}
		}
	}
	else
	{
		fail=true;				// SplitList will leave a result if it fails
	}
	return(!fail);
}

static bool MappingParse(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,int argc,Tcl_Obj *const argv[])
// Use the arguments to build a mapping onto map.
// If there is a problem, add it to the Tcl result, and return false.
{
	SYNTAX_STYLE_MAPPING
		*mapping;
	EXPRESSION_PIECE
		*expressionPiece;
	uint32_t
		startStyle,
		betweenStyle,
		endStyle;

	ErrorTraceClear();
	if((mapping=SyntaxMapMappingAdd(map,Tcl_GetString(argv[0]))))
	{
		if((expressionPiece=ExpressionPieceGet(localInterpreter,map,argv[1])))
		{
			SyntaxMapMappingStartExpressionPieceSet(mapping,expressionPiece);
			if(EndExpressionsParse(localInterpreter,map,mapping,argv[2]))
			{
				if(TclUInt32Get(localInterpreter,argv[3],&startStyle))
				{
					if(TclUInt32Get(localInterpreter,argv[4],&betweenStyle))
					{
						if(TclUInt32Get(localInterpreter,argv[5],&endStyle))
						{
							SyntaxMapMappingStylesSet(mapping,startStyle,betweenStyle,endStyle);
							return(true);
						}
					}
				}
			}
		}
		SyntaxMapMappingRemove(map,mapping);
	}
	else
	{
		Tcl_AppendResult(localInterpreter,"Failed to add mapping '",argv[0],"' to map: ",ErrorTraceGet(),NULL);
	}
	return(false);
}

static bool SyntaxMapParse(Tcl_Interp *localInterpreter,SYNTAX_MAP *map,Tcl_Obj *const mapContents)
// Parse the mapContents, and use it to build map.
// If there is a problem, add it to the Tcl result, and return false.
{
	bool
		fail;
	int
		listArgc;
	Tcl_Obj
		**listArgv;
	int
		argc;
	Tcl_Obj
		**argv;
	int
		mappinglistArgc;
	Tcl_Obj
		**mappinglistArgv;
	int
		syntaxMapToken;
	SYNTAX_STYLE_MAPPING
		*mapping,
		*otherMapping;
	int
		i,j;

	fail=false;
	if(Tcl_ListObjGetElements(localInterpreter,mapContents,&listArgc,&listArgv)==TCL_OK)	// make list of command/parameter
	{
		for(i=0;(i<listArgc)&&!fail;i+=2)					// step over each command, parameter list pair
		{
			if(TokenMatch(Tcl_GetString(listArgv[i]),syntaxMapCommands,&syntaxMapToken))
			{
				if(Tcl_ListObjGetElements(localInterpreter,listArgv[i+1],&argc,&argv)==TCL_OK)	// make parameters into arguments
				{
					switch(syntaxMapToken)
					{
						case SM_EXPRESSION:
							if(argc==2)
							{
								ErrorTraceClear();
								if(SyntaxMapExpressionAdd(map,Tcl_GetString(argv[0]),Tcl_GetString(argv[1])))
								{
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Failed to add expression '",Tcl_GetString(argv[0]),"' to map: ",ErrorTraceGet(),NULL);
									fail=true;
								}
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"wrong # args: should be \"",Tcl_GetString(listArgv[i])," expressionName expression\"",NULL);
								fail=true;
							}
							break;
						case SM_MAPPING:
							if(argc==6)
							{
								fail=!MappingParse(localInterpreter,map,argc,argv);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"wrong # args: should be \"",Tcl_GetString(listArgv[i])," mapName startExpressionName endExpressionList startStyle betweenStyle endStyle\"",NULL);
								fail=true;
							}
							break;
						case SM_AT:
							if(argc==2)
							{
								if((mapping=SyntaxMapStyleMappingLocate(map,Tcl_GetString(argv[0]))))
								{
									if(Tcl_ListObjGetElements(localInterpreter,argv[1],&mappinglistArgc,&mappinglistArgv)==TCL_OK)	// make parameters into arguments
									{
										for(j=0;j<mappinglistArgc&&!fail;j++)
										{
											if((otherMapping=SyntaxMapStyleMappingLocate(map,Tcl_GetString(mappinglistArgv[j]))))
											{
												ErrorTraceClear();
												if(SyntaxMapBetweenListElementAdd(mapping,otherMapping))
												{
												}
												else
												{
													Tcl_AppendResult(localInterpreter,"Failed to add mapping '",Tcl_GetString(mappinglistArgv[j]),"': ",ErrorTraceGet(),NULL);
													fail=true;
												}
											}
											else
											{
												Tcl_AppendResult(localInterpreter,"Failed to locate mapping '",Tcl_GetString(mappinglistArgv[j]),"'",NULL);
												fail=true;
											}
										}
									}
									else
									{
										fail=true;
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Failed to locate mapping '",Tcl_GetString(argv[0]),"'",NULL);
									fail=true;
								}
							}
							else
							{
								Tcl_AppendResult(localInterpreter,"wrong # args: should be \"",Tcl_GetString(listArgv[i])," mapName betweenList\"",NULL);
								fail=true;
							}
							break;
						case SM_ROOT:
							for(j=0;j<argc&&!fail;j++)
							{
								if((otherMapping=SyntaxMapStyleMappingLocate(map,Tcl_GetString(argv[j]))))
								{
									ErrorTraceClear();
									if(SyntaxMapRootListElementAdd(map,otherMapping))
									{
									}
									else
									{
										Tcl_AppendResult(localInterpreter,"Failed to add root mapping '",Tcl_GetString(argv[j]),"': ",ErrorTraceGet(),NULL);
										fail=true;
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,"Failed to locate mapping '",Tcl_GetString(argv[j]),"'",NULL);
									fail=true;
								}
							}
							break;
					}
				}
				else
				{
					fail=true;		// Tcl_ListObjGetElements will leave a result if it fails
				}
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Invalid syntax map command '",Tcl_GetString(listArgv[i]),"'",NULL);
				fail=true;
			}
		}
	}
	else
	{
		fail=true;		// Tcl_ListObjGetElements will leave a result if it fails
	}
	return(!fail);
}

static int Cmd_AddColorScheme(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a color scheme from a base mapping, and a list of styles.
// NOTE: this can be used to replace an existing color scheme.
{
	VIEW_COLOR_SCHEME
		*oldColorScheme,
		*colorScheme;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor;

	if(objc==10)
	{
		ErrorTraceClear();
		if((colorScheme=ViewColorSchemeCreate()))
		{
			if((oldColorScheme=ViewColorSchemeListLocate(Tcl_GetString(objv[1]))))	// see if there is an old scheme with this name that is about to be replaced
			{
				ViewColorSchemeListRemove(oldColorScheme);
				ViewColorSchemeDestroy(oldColorScheme);
			}
			if(ViewColorSchemeListAdd(colorScheme,Tcl_GetString(objv[1])))
			{
				if(TclColorGet(localInterpreter,objv[2],&backgroundColor))
				{
					if(TclColorGet(localInterpreter,objv[3],&foregroundColor))
					{
						ViewColorSchemeDrawColorsSet(colorScheme,true,backgroundColor,foregroundColor);
						if(TclColorOrEmptyGet(localInterpreter,objv[4],&backgroundColor))
						{
							if(TclColorOrEmptyGet(localInterpreter,objv[5],&foregroundColor))
							{
								ViewColorSchemeDrawColorsSet(colorScheme,false,backgroundColor,foregroundColor);
								if(TclColorOrEmptyGet(localInterpreter,objv[6],&backgroundColor))
								{
									if(TclColorOrEmptyGet(localInterpreter,objv[7],&foregroundColor))
									{
										ViewColorSchemeSelectionColorsSet(colorScheme,true,backgroundColor,foregroundColor);
										if(TclColorOrEmptyGet(localInterpreter,objv[8],&backgroundColor))
										{
											if(TclColorOrEmptyGet(localInterpreter,objv[9],&foregroundColor))
											{
												ViewColorSchemeSelectionColorsSet(colorScheme,false,backgroundColor,foregroundColor);
												return(TCL_OK);
											}
										}
									}
								}
							}
						}
					}
				}
				ViewColorSchemeListRemove(colorScheme);
			}
			else
			{
				Tcl_AppendResult(localInterpreter,"Error adding color scheme '",Tcl_GetString(objv[1]),"': ",ErrorTraceGet(),NULL);
			}
			ViewColorSchemeDestroy(colorScheme);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Error creating color scheme '",Tcl_GetString(objv[1]),"': ",ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"colorSchemeName activeTextBackground activeTextForeground inactiveTextBackground inactiveTextForeground activeSelectionBackground activeSelectionForeground inactiveSelectionBackground inactiveSelectionForeground");
	}
	return(TCL_ERROR);
}

static int Cmd_ColorSchemeStyle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Adjust the colors used by a style of the given color scheme.
{
	VIEW_COLOR_SCHEME
		*colorScheme;
	uint32_t
		style;
	EDITOR_COLOR
		backgroundColor,
		foregroundColor;
	bool
		bold,
		oblique;

	if(objc==9)
	{
		if((colorScheme=ViewColorSchemeListLocate(Tcl_GetString(objv[1]))))
		{
			if(TclUInt32Get(localInterpreter,objv[2],&style))
			{
				style=ViewStyleForceIntoRange(style);
				if(TclColorOrEmptyGet(localInterpreter,objv[3],&backgroundColor))
				{
					if(TclColorOrEmptyGet(localInterpreter,objv[4],&foregroundColor))
					{
						ViewColorSchemeStyleDrawColorsSet(colorScheme,style,true,backgroundColor,foregroundColor);
						if(TclColorOrEmptyGet(localInterpreter,objv[5],&backgroundColor))
						{
							if(TclColorOrEmptyGet(localInterpreter,objv[6],&foregroundColor))
							{
								ViewColorSchemeStyleDrawColorsSet(colorScheme,style,false,backgroundColor,foregroundColor);
								if(TclBooleanGet(localInterpreter,objv[7],&bold))
								{
									if(TclBooleanGet(localInterpreter,objv[8],&oblique))
									{
										ViewColorSchemeStyleFontAttributesSet(colorScheme,style,bold,oblique);
										return(TCL_OK);
									}
								}
							}
						}
					}
				}
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Color scheme not found: '",Tcl_GetString(objv[1]),"': ",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"colorSchemeName styleNumber activeTextBackground activeTextForeground inactiveTextBackground inactiveTextForeground fontBold fontOblique");
	}
	return(TCL_ERROR);
}

static int Cmd_DeleteColorScheme(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Delete a color scheme that was created by "addcolorscheme".
{
	VIEW_COLOR_SCHEME
		*colorScheme;

	if(objc==2)
	{
		if((colorScheme=ViewColorSchemeListLocate(Tcl_GetString(objv[1]))))	// find scheme to delete
		{
			ViewColorSchemeListRemove(colorScheme);
			ViewColorSchemeDestroy(colorScheme);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Could not locate color scheme '",Tcl_GetString(objv[1]),"'",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"colorSchemeName");
	}
	return(TCL_ERROR);
}

static int Cmd_ColorSchemes(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the list of currently defined color schemes.
{
	VIEW_COLOR_SCHEME
		*colorScheme;
	bool
		fail;

	fail=false;
	if(objc==1)
	{
		colorScheme=ViewColorSchemeListLocateFirst();
		while(colorScheme)
		{
			Tcl_AppendElement(localInterpreter,ViewColorSchemeListNameGet(colorScheme));
			colorScheme=ViewColorSchemeListLocateNext(colorScheme);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_AddSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Create a syntax highlighting map from a list of expressions and a list of mappings from expressions to styles.
// NOTE: this can be used to replace an existing syntax map, but
// if the existing map is in use by any buffer which is busy, this will fail
// before it does anything.
{
	EDITOR_BUFFER
		*buffer,
		*nextBuffer;
	SYNTAX_MAP
		*oldMap,
		*map;
	bool
		fail;

	if(objc==3)
	{
		ErrorTraceClear();
		if((map=SyntaxMapOpen(Tcl_GetString(objv[1]))))
		{
			if(SyntaxMapParse(localInterpreter,map,objv[2]))
			{
				fail=false;
				if((oldMap=SyntaxMapLocate(Tcl_GetString(objv[1]))))	// see if there is an old map with this name that is about to be replaced
				{
					buffer=NULL;
					while(!fail&&(buffer=SyntaxMapLocateNextEditorBuffer(oldMap,buffer)))	// verify all buffers are not busy
					{
						fail=!ShellBufferNotBusy(localInterpreter,buffer);
					}
					if(!fail)							// move all buffers from old map to new map
					{
						nextBuffer=SyntaxMapLocateNextEditorBuffer(oldMap,NULL);
						while(nextBuffer)				// re-assign the maps
						{
							buffer=nextBuffer;			// point at the current buffer
							nextBuffer=SyntaxMapLocateNextEditorBuffer(oldMap,nextBuffer);	// move to the next before we do any modification to the current
							SyntaxMapAssign(buffer,map);	// if this fails, there's not much we can do anyway, so ignore it
						}
						SyntaxMapUnlink(oldMap);
						SyntaxMapClose(oldMap);
					}
				}
				if(!fail)
				{
					ErrorTraceClear();
					if(SyntaxMapLink(map))				// link this map onto the global list
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Error linking syntax map '",Tcl_GetString(objv[1]),"': ",ErrorTraceGet(),NULL);

						// failed to link new map to list, so untie any buffers that got linked to it
						buffer=NULL;
						while((buffer=SyntaxMapLocateNextEditorBuffer(map,buffer)))
						{
							SyntaxMapAssign(buffer,NULL);
						}
					}
				}
			}
			SyntaxMapClose(map);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Error creating syntax map '",Tcl_GetString(objv[1]),"': ",ErrorTraceGet(),NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"syntaxMapName syntaxMap");
	}
	return(TCL_ERROR);
}

static int Cmd_DeleteSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Delete a syntax map that was created by "addsyntaxmap".
// Remove the map from all buffers which are using it.
// NOTE: if any buffer which is using the map is currently busy, this will
// fail before it does anything.
{
	EDITOR_BUFFER
		*buffer,
		*nextBuffer;
	SYNTAX_MAP
		*map;
	bool
		fail;

	if(objc==2)
	{
		if((map=SyntaxMapLocate(Tcl_GetString(objv[1]))))	// find map to delete
		{
			fail=false;
			buffer=NULL;
			while(!fail&&(buffer=SyntaxMapLocateNextEditorBuffer(map,buffer)))	// verify all buffers are not busy
			{
				fail=!ShellBufferNotBusy(localInterpreter,buffer);
			}
			if(!fail)							// remove all buffers from the map, and kill the map
			{
				nextBuffer=SyntaxMapLocateNextEditorBuffer(map,NULL);
				while(nextBuffer)				// re-assign the maps
				{
					buffer=nextBuffer;			// point at the current buffer
					nextBuffer=SyntaxMapLocateNextEditorBuffer(map,nextBuffer);	// move to the next before we do any modification to the current
					SyntaxMapAssign(buffer,NULL);	// if this fails, there's not much we can do anyway, so ignore it
				}
				SyntaxMapUnlink(map);
				SyntaxMapClose(map);
				return(TCL_OK);
			}
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Could not locate syntax map '",Tcl_GetString(objv[1]),"'",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"syntaxMap");
	}
	return(TCL_ERROR);
}

static int Cmd_SyntaxMaps(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the list of currently defined syntax maps.
{
	bool
		fail;
	SYNTAX_MAP
		*currentMap;

	fail=false;
	if(objc==1)
	{
		currentMap=NULL;
		while((currentMap=SyntaxMapLocateNext(currentMap)))
		{
			Tcl_AppendElement(localInterpreter,SyntaxMapNameGet(currentMap));
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_SetSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the syntax map in use for the given buffer.
// If an empty map string is specified, then remove any map from the current buffer.
{
	EDITOR_BUFFER
		*buffer;
	SYNTAX_MAP
		*map;
	char
		*mapName;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				map=NULL;
				mapName=Tcl_GetString(objv[2]);
				if(mapName[0]=='\0'||(map=SyntaxMapLocate(mapName)))
				{
					ErrorTraceClear();
					if(SyntaxMapAssign(buffer,map))
					{
						return(TCL_OK);
					}
					else
					{
						Tcl_AppendResult(localInterpreter,"Failed to assign syntax map: ",ErrorTraceGet(),NULL);
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"Could not locate syntax map '",mapName,"'",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName syntaxMapName");
	}
	return(TCL_ERROR);
}

static int Cmd_GetSyntaxMap(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Read back the syntax map in use for the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	SYNTAX_MAP
		*map;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if((map=SyntaxMapGetAssigned(buffer)))
			{
				Tcl_AppendElement(localInterpreter,SyntaxMapNameGet(map));
			}
			else
			{
				Tcl_AppendElement(localInterpreter,"");
			}
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_SyntaxMapBuffers(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the list of buffer names which are currently using the given syntax map.
{
	EDITOR_BUFFER
		*buffer;
	SYNTAX_MAP
		*map;

	if(objc==2)
	{
		if((map=SyntaxMapLocate(Tcl_GetString(objv[1]))))
		{
			buffer=NULL;
			while((buffer=SyntaxMapLocateNextEditorBuffer(map,buffer)))
			{
				Tcl_AppendElement(localInterpreter,buffer->contentName);
			}
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"Could not locate syntax map '",Tcl_GetString(objv[1]),"'",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"syntaxMapName");
	}
	return(TCL_ERROR);
}

static void CreateSearchResult(char *string,int stringSize,bool foundMatch,SELECTION_UNIVERSE *selection)
// When a find, findall, replace, or replaceall has finished executing, this
// is called to create the Tcl result for the search.
// If the search was successful, this will return the range of the first
// match/replacement.
// If the search was not successful, this will return an empty string.
{
	uint64_t
		matchPosition;
	uint64_t
		startPosition,
		numElements;
	bool
		isActive;

	if(foundMatch)
	{
		matchPosition=SelectionCursorPositionGet(selection);	// get the position of the match/first match
		SelectionRangeGet(selection,matchPosition,&startPosition,&numElements,&isActive);
		if(!isActive)
		{
			startPosition=matchPosition;
			numElements=0;
		}
		snprintf(string,stringSize,"%" PRIu64 " %" PRIu64 "",startPosition,startPosition+numElements);
	}
	else
	{
		string[0]='\0';					// just return an empty string
	}
}

typedef struct
{
	bool
		useUtf8,
		backwards,
		wrap,
		ignoreCase,
		limitScope,
		regex;
	PARSE_INT64
		useSubExpression;
	bool
		replaceScript;
	PARSE_MARK
		sourceMark,
		destMark;
} SEARCH_PARAMETERS;

static bool SearchParametersProcess(Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[],int *optionIndex,EDITOR_BUFFER *buffer,SEARCH_PARAMETERS *searchParameters)
// Process command line options to set up searchParameters.
// If there is a problem, set the error string in localInterpreter, and return false.
{
	const OPTION_DESCRIPTOR_RECORD
		options[]=
		{
			{"-ascii",				false,	TclOptionBoolean},
			{"-utf8",				true,	TclOptionBoolean},
			{"-forward",			false,	TclOptionBoolean},
			{"-backward",			true,	TclOptionBoolean},
			{"-nowrap",				false,	TclOptionBoolean},
			{"-wrap",				true,	TclOptionBoolean},
			{"-case",				false,	TclOptionBoolean},
			{"-ignorecase",			true,	TclOptionBoolean},
			{"-globalscope",		false,	TclOptionBoolean},
			{"-limitscope",			true,	TclOptionBoolean},
			{"-literal",			false,	TclOptionBoolean},
			{"-regex",				true,	TclOptionBoolean},
			{"-usesubexpression",	0,		TclOptionInt64},
			{"-replaceliteral",		false,	TclOptionBoolean},
			{"-replacescript",		true,	TclOptionBoolean},
			{"-sourcemark",			0,		TclOptionMark},
			{"-destmark",			0,		TclOptionMark},
			{NULL,0,NULL}
		};

	void
		*parsedVars[]=
		{
			(void *)&searchParameters->useUtf8,
			(void *)&searchParameters->useUtf8,
			(void *)&searchParameters->backwards,
			(void *)&searchParameters->backwards,
			(void *)&searchParameters->wrap,
			(void *)&searchParameters->wrap,
			(void *)&searchParameters->ignoreCase,
			(void *)&searchParameters->ignoreCase,
			(void *)&searchParameters->limitScope,
			(void *)&searchParameters->limitScope,
			(void *)&searchParameters->regex,
			(void *)&searchParameters->regex,
			(void *)&searchParameters->useSubExpression,
			(void *)&searchParameters->replaceScript,
			(void *)&searchParameters->replaceScript,
			(void *)&searchParameters->sourceMark,
			(void *)&searchParameters->destMark,
		};

	searchParameters->useUtf8=false;
	searchParameters->backwards=false;
	searchParameters->wrap=false;
	searchParameters->ignoreCase=false;
	searchParameters->limitScope=false;
	searchParameters->regex=false;
	searchParameters->useSubExpression.parsed=false;
	searchParameters->useSubExpression.value=0;
	searchParameters->replaceScript=false;

	searchParameters->sourceMark.buffer=buffer;
	searchParameters->sourceMark.mark=NULL;

	searchParameters->destMark.buffer=buffer;
	searchParameters->destMark.mark=NULL;

	return(TclOptionsParse(localInterpreter,options,parsedVars,objc,objv,optionIndex,true));
}

static int Cmd_Find(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Find within the given buffer.
// findBuffer must exist, or an error will be returned.
// If there is a match, the current selection (or destMask if specified)
// will be modified to hold the match. Also, the position of the first
// located match will be returned to Tcl (for homing purposes).
// If there is no match, the dest selection will not be modified, and
// an empty string will be returned to Tcl.
// If there is an error, an error message will be returned to Tcl.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer;
	SEARCH_PARAMETERS
		searchParameters;
	int
		optionIndex;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection,
		*destSelection;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorBufferBusySet(buffer);
				if((findBuffer=TclEditorBufferGet(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						EditorBufferBusySet(findBuffer);
						optionIndex=3;

						if(SearchParametersProcess(localInterpreter,objc,objv,&optionIndex,buffer,&searchParameters))
						{
							sourceSelection=searchParameters.sourceMark.mark?searchParameters.sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

							ErrorTraceClear();
							if((resultSelection=SelectionUniverseOpen()))	// get a selection for the result
							{
								ErrorTraceClear();
								if(EditorFind(buffer,sourceSelection,findBuffer->textUniverse,searchParameters.useUtf8,searchParameters.backwards,searchParameters.wrap,searchParameters.ignoreCase,searchParameters.limitScope,searchParameters.regex,searchParameters.useSubExpression.parsed,searchParameters.useSubExpression.value,&foundMatch,resultSelection,AbortCheck))
								{
									CreateSearchResult(tempString,sizeof(tempString),foundMatch,resultSelection);
									Tcl_AppendResult(localInterpreter,tempString,NULL);
									if(foundMatch)
									{
										destSelection=searchParameters.destMark.mark?searchParameters.destMark.mark->selectionUniverse:buffer->selectionUniverse;
										EditorSelectionChangeStart(buffer,destSelection);
										SelectionUniverseCopy(resultSelection,destSelection);
										EditorSelectionChangeEnd(buffer,destSelection);
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
									fail=true;
								}
								SelectionUniverseClose(resultSelection);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						EditorBufferBusyClear(findBuffer);
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
				EditorBufferBusyClear(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer findBuffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_FindAll(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Find all within the given buffer.
// findBuffer must exist, or an error will be returned.
// If there is a match, the current selection (or destMask if specified)
// will be modified to hold the match. Also, the position of the first
// located match will be returned to Tcl (for homing purposes).
// If there is no match, the dest selection will not be modified, and
// an empty string will be returned to Tcl.
// If there is an error, an error message will be returned to Tcl.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer;
	SEARCH_PARAMETERS
		searchParameters;
	int
		optionIndex;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection,
		*destSelection;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorBufferBusySet(buffer);
				if((findBuffer=TclEditorBufferGet(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						EditorBufferBusySet(findBuffer);
						optionIndex=3;

						if(SearchParametersProcess(localInterpreter,objc,objv,&optionIndex,buffer,&searchParameters))
						{
							sourceSelection=searchParameters.sourceMark.mark?searchParameters.sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

							ErrorTraceClear();
							if((resultSelection=SelectionUniverseOpen()))	// get a selection for the result
							{
								ErrorTraceClear();
								if(EditorFindAll(buffer,sourceSelection,findBuffer->textUniverse,searchParameters.useUtf8,searchParameters.backwards,searchParameters.wrap,searchParameters.ignoreCase,searchParameters.limitScope,searchParameters.regex,searchParameters.useSubExpression.parsed,searchParameters.useSubExpression.value,&foundMatch,resultSelection,AbortCheck))
								{
									CreateSearchResult(tempString,sizeof(tempString),foundMatch,resultSelection);
									Tcl_AppendResult(localInterpreter,tempString,NULL);
									if(foundMatch)
									{
										destSelection=searchParameters.destMark.mark?searchParameters.destMark.mark->selectionUniverse:buffer->selectionUniverse;
										EditorSelectionChangeStart(buffer,destSelection);
										SelectionUniverseCopy(resultSelection,destSelection);
										EditorSelectionChangeEnd(buffer,destSelection);
									}
								}
								else
								{
									Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
									fail=true;
								}
								SelectionUniverseClose(resultSelection);
							}
							else
							{
								Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						EditorBufferBusyClear(findBuffer);
					}
					else
					{
						fail=true;
					}
				}
				else
				{
					fail=true;
				}
				EditorBufferBusyClear(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer findBuffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_Replace(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Replace within the given buffer.
// findBuffer and replaceBuffer must exist, or an error will be returned.
// If there is a match, the current selection (or destMask if specified)
// will be modified to indicate where replacements were made. Also, the
// position of the first replacement will be returned to Tcl (for homing purposes).
// If there is no match, the dest selection will not be modified, and
// an empty string will be returned to Tcl.
// If there is an error, an error message will be returned to Tcl.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer,
		*replaceBuffer;
	SEARCH_PARAMETERS
		searchParameters;
	int
		optionIndex;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection,
		*destSelection;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorBufferBusySet(buffer);
				if((findBuffer=TclEditorBufferGet(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						EditorBufferBusySet(findBuffer);
						if((replaceBuffer=TclEditorBufferGet(localInterpreter,objv[3])))
						{
							if(ShellBufferNotBusy(localInterpreter,replaceBuffer))
							{
								EditorBufferBusySet(replaceBuffer);
								optionIndex=4;

								if(SearchParametersProcess(localInterpreter,objc,objv,&optionIndex,buffer,&searchParameters))
								{
									sourceSelection=searchParameters.sourceMark.mark?searchParameters.sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

									ErrorTraceClear();
									if((resultSelection=SelectionUniverseOpen()))	// get a selection for the result
									{
										ErrorTraceClear();
										if(EditorReplace(buffer,sourceSelection,findBuffer->textUniverse,replaceBuffer->textUniverse,searchParameters.useUtf8,searchParameters.backwards,searchParameters.wrap,searchParameters.ignoreCase,searchParameters.limitScope,searchParameters.regex,searchParameters.useSubExpression.parsed,searchParameters.useSubExpression.value,searchParameters.replaceScript,&foundMatch,resultSelection,AbortCheck))
										{
											CreateSearchResult(tempString,sizeof(tempString),foundMatch,resultSelection);
											Tcl_AppendResult(localInterpreter,tempString,NULL);
											if(foundMatch)
											{
												destSelection=searchParameters.destMark.mark?searchParameters.destMark.mark->selectionUniverse:buffer->selectionUniverse;
												EditorSelectionChangeStart(buffer,destSelection);
												SelectionUniverseCopy(resultSelection,destSelection);
												EditorSelectionChangeEnd(buffer,destSelection);
											}
										}
										else
										{
											Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
											fail=true;
										}
										SelectionUniverseClose(resultSelection);
									}
									else
									{
										Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
										fail=true;
									}
								}
								else
								{
									fail=true;
								}
								EditorBufferBusyClear(replaceBuffer);
							}
							else
							{
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						EditorBufferBusyClear(findBuffer);
					}
				}
				else
				{
					fail=true;
				}
				EditorBufferBusyClear(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer findBuffer replaceBuffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_ReplaceAll(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Replace all within the given buffer.
// findBuffer and replaceBuffer must exist, or an error will be returned.
// If there is a match, the current selection (or destMask if specified)
// will be modified to indicate where replacements were made. Also, the
// position of the first replacement will be returned to Tcl (for homing purposes).
// If there is no match, the dest selection will not be modified, and
// an empty string will be returned to Tcl.
// If there is an error, an error message will be returned to Tcl.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*findBuffer,
		*replaceBuffer;
	SEARCH_PARAMETERS
		searchParameters;
	int
		optionIndex;
	SELECTION_UNIVERSE
		*sourceSelection,
		*resultSelection,
		*destSelection;
	bool
		foundMatch;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	bool
		fail;

	fail=false;
	if(objc>=3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorBufferBusySet(buffer);
				if((findBuffer=TclEditorBufferGet(localInterpreter,objv[2])))
				{
					if(ShellBufferNotBusy(localInterpreter,findBuffer))
					{
						EditorBufferBusySet(findBuffer);
						if((replaceBuffer=TclEditorBufferGet(localInterpreter,objv[3])))
						{
							if(ShellBufferNotBusy(localInterpreter,replaceBuffer))
							{
								EditorBufferBusySet(replaceBuffer);
								optionIndex=4;

								if(SearchParametersProcess(localInterpreter,objc,objv,&optionIndex,buffer,&searchParameters))
								{
									sourceSelection=searchParameters.sourceMark.mark?searchParameters.sourceMark.mark->selectionUniverse:buffer->selectionUniverse;

									ErrorTraceClear();
									if((resultSelection=SelectionUniverseOpen()))	// get a selection for the result
									{
										ErrorTraceClear();
										if(EditorReplaceAll(buffer,sourceSelection,findBuffer->textUniverse,replaceBuffer->textUniverse,searchParameters.useUtf8,searchParameters.backwards,searchParameters.wrap,searchParameters.ignoreCase,searchParameters.limitScope,searchParameters.regex,searchParameters.useSubExpression.parsed,searchParameters.useSubExpression.value,searchParameters.replaceScript,&foundMatch,resultSelection,AbortCheck))
										{
											CreateSearchResult(tempString,sizeof(tempString),foundMatch,resultSelection);
											Tcl_AppendResult(localInterpreter,tempString,NULL);
											if(foundMatch)
											{
												destSelection=searchParameters.destMark.mark?searchParameters.destMark.mark->selectionUniverse:buffer->selectionUniverse;
												EditorSelectionChangeStart(buffer,destSelection);
												SelectionUniverseCopy(resultSelection,destSelection);
												EditorSelectionChangeEnd(buffer,destSelection);
											}
										}
										else
										{
											Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
											fail=true;
										}
										SelectionUniverseClose(resultSelection);
									}
									else
									{
										Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
										fail=true;
									}
								}
								else
								{
									fail=true;
								}
								EditorBufferBusyClear(replaceBuffer);
							}
							else
							{
								fail=true;
							}
						}
						else
						{
							fail=true;
						}
						EditorBufferBusyClear(findBuffer);
					}
				}
				else
				{
					fail=true;
				}
				EditorBufferBusyClear(buffer);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"buffer findBuffer replaceBuffer ?options?");
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_UndoToggle(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform an undo toggle in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	bool
		didSomething;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorUndoToggle(buffer,&didSomething))
				{
					Tcl_AppendResult(localInterpreter,didSomething?"1":"0",NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Undo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform an undo in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	bool
		didSomething;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorUndo(buffer,&didSomething))
				{
					Tcl_AppendResult(localInterpreter,didSomething?"1":"0",NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Redo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a redo in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	bool
		didSomething;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorRedo(buffer,&didSomething))
				{
					Tcl_AppendResult(localInterpreter,didSomething?"1":"0",NULL);
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_BreakUndo(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Cause a break in the undo flow for the passed buffer.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				UndoGroupBegin(buffer);
				UndoGroupEndStrict(buffer);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_FlushUndos(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a clear of the undo/redo buffers in the given buffer.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				EditorUndoClear(buffer);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Cut(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a cut in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*clipboard;

	if(objc==2||objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(objc==3)
				{
					clipboard=EditorBufferLocate(Tcl_GetString(objv[2]));
				}
				else
				{
					clipboard=ClipboardCurrentGet();
				}
				if(clipboard)
				{
					if(ShellBufferNotBusy(localInterpreter,clipboard))
					{
						ErrorTraceClear();
						if(ClipboardCut(buffer,clipboard))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_Copy(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a copy in the given buffer.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*clipboard;

	if(objc==2||objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if(objc==3)
				{
					clipboard=EditorBufferLocate(Tcl_GetString(objv[2]));
				}
				else
				{
					clipboard=ClipboardCurrentGet();
				}
				if(clipboard)
				{
					if(ShellBufferNotBusy(localInterpreter,clipboard))
					{
						ErrorTraceClear();
						if(ClipboardCopy(buffer,clipboard))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_Paste(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Perform a columnar paste in the given window/buffer.
{
	EDITOR_BUFFER
		*buffer;
	EDITOR_BUFFER
		*clipboard;
	EDITOR_WINDOW
		*window;
	EDITOR_VIEW
		*view;

	if(objc==2||objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				if((window=PlatformDocumentWindowLocate(buffer)))
				{
					view=PlatformDocumentWindowCurrentViewGet(window);
				}
				else
				{
					view=NULL;
				}
				if(objc==3)
				{
					clipboard=EditorBufferLocate(Tcl_GetString(objv[2]));
				}
				else
				{
					clipboard=ClipboardCurrentGet();
				}
				if(clipboard)
				{
					if(ShellBufferNotBusy(localInterpreter,clipboard))
					{
						ErrorTraceClear();
						if(ClipboardColumnarPaste(buffer,view,clipboard))
						{
							return(TCL_OK);
						}
						else
						{
							Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
						}
					}
				}
				else
				{
					Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName ?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_SetClipboard(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the current clipboard to the given buffer (or NULL).
{
	EDITOR_BUFFER
		*clipboard;

	if(objc==1)
	{
		ClipboardCurrentSet(NULL);
		return(TCL_OK);
	}
	else if(objc==2)
	{
		if((clipboard=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,clipboard))
			{
				ClipboardCurrentSet(clipboard);
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"?clipboardBuffer?");
	}
	return(TCL_ERROR);
}

static int Cmd_GetClipboard(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the current clipboard, fail if there is none.
{
	EDITOR_BUFFER
		*clipboard;

	if(objc==1)
	{
		if((clipboard=ClipboardCurrentGet()))
		{
			Tcl_AppendResult(localInterpreter,clipboard->contentName,NULL);
			return(TCL_OK);
		}
		else
		{
			Tcl_AppendResult(localInterpreter,"No clipboard",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_ScreenSize(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the dimensions of the screen that the editor is running on.
{
	int32_t
		xOffset,
		yOffset;
	uint32_t
		width,
		height;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		PlatformScreenOffsetAndDimensionsGet(&xOffset,&yOffset,&width,&height);
		snprintf(tempString,sizeof(tempString),"%" PRIu32 " %" PRIu32 "",width,height);
		Tcl_AppendResult(localInterpreter,tempString,NULL);
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_SetWordSpaceChars(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Set the wordSpaceList so that all of the characters in the passed argument
// are considered space.
{
	int
		result;
	int
		i;
	int
		listArgc;
	Tcl_Obj
		**listArgv;
	uint32_t
		startPoint,
		endPoint;

	result=TCL_OK;
	CodePointListRangeClear(wordSpaceList,0,0x100000000);	// clear the table
	for(i=1;(result==TCL_OK)&&(i<objc);i++)
	{
		if((result=Tcl_ListObjGetElements(localInterpreter,objv[i],&listArgc,&listArgv))==TCL_OK)
		{
			if(listArgc)	// any elements in the list?
			{
				if(TclUInt32Get(localInterpreter,listArgv[0],&startPoint))
				{
					endPoint=startPoint;
					if(listArgc>1)
					{
						if(!TclUInt32Get(localInterpreter,listArgv[1],&endPoint))
						{
							result=TCL_ERROR;
						}
					}
				}
				else
				{
					result=TCL_ERROR;
				}
				if(result==TCL_OK)
				{
					if(endPoint>=startPoint)
					{
						if(!CodePointListRangeSet(wordSpaceList,startPoint,(endPoint-startPoint)+1))
						{
							result=TCL_ERROR;
						}
					}
				}
			}
		}
	}
	return(result);
}

static int Cmd_GetWordSpaceChars(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Return the list of ranges of characters that are considered word separators.
{
	uint32_t
		startPosition;
	uint64_t
		testPosition,
		numElements;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];

	if(objc==1)
	{
		testPosition=0;
		while((testPosition<0x100000000)&&CodePointListTestAtOrAfterPositionGet(wordSpaceList,testPosition,&startPosition,&numElements))
		{
			snprintf(tempString,sizeof(tempString),"%" PRIu32 " %" PRIu32 "",(uint32_t)startPosition,(uint32_t)(startPosition+numElements-1));
			Tcl_AppendElement(localInterpreter,&(tempString[0]));
			testPosition=startPosition+numElements;
		}
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_Execute(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Execute the given tcl script, place the results in the given buffer.
// If it still exists after the script is finished, otherwise throw them away.
{
	EDITOR_BUFFER_HANDLE
		handle;
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	char
		tempString[TEMP_STRING_BUFFER_LENGTH];
	int
		tclResult;
	const char
		*stringResult;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				// NOTE: do not set busy here, we will allow executed script to mess with the window, but we must be careful:
				// If the buffer goes away as the result of executing the script, we have to make sure it is not referenced again

				EditorBufferHandleInit(&handle);
				EditorBufferHandleGrab(&handle,buffer);					// hold on to this buffer

				SelectionEndPositionsGet(buffer->selectionUniverse,&startPosition,&endPosition);
				SelectionCursorPositionSet(buffer->auxSelectionUniverse,endPosition);	// move the aux cursor to the end of any selection/cursor, and start output there

				// do not call AbortClear here, it will have already been called when the script that got us here was started
				tclResult=Tcl_EvalObj(localInterpreter,objv[2]);
				stringResult=Tcl_GetStringResult(localInterpreter);
				if(*stringResult)										// if no result, do not try to insert into buffer
				{
					if((buffer=EditorBufferGetFromHandle(&handle)))		// see if this buffer is still alive
					{
						if(tclResult!=TCL_OK)
						{
							snprintf(tempString,sizeof(tempString),"Error in line %d: ",(int)Tcl_GetErrorLine(localInterpreter));
							EditorAuxInsert(buffer,(uint8_t *)tempString,(uint32_t)strlen(tempString));
						}
						EditorAuxInsert(buffer,(uint8_t *)stringResult,(uint32_t)strlen(stringResult));
						EditorAuxInsert(buffer,(uint8_t *)"\n",1);
					}
				}
				EditorBufferHandleRelease(&handle);						// do not need this any more
				return(TCL_OK);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName script");
	}
	return(TCL_ERROR);
}

static int Cmd_Task(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Start a task running in the given buffer, or send data to a task in a buffer.
{
	EDITOR_BUFFER
		*buffer;
	uint64_t
		startPosition,
		endPosition;
	char
		*taskData;
	int
		taskDataLength;

	if(objc==3)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				SelectionEndPositionsGet(buffer->selectionUniverse,&startPosition,&endPosition);
				SelectionCursorPositionSet(buffer->auxSelectionUniverse,endPosition);	// move the aux cursor to the end of any selection/cursor, and start output there
				taskData=Tcl_GetStringFromObj(objv[2],&taskDataLength);
				ErrorTraceClear();
				if(!EditorBufferHasTask(buffer))
				{
					if(EditorBufferTaskSpawn(buffer,taskData))
					{
						return(TCL_OK);
					}
				}
				else
				{
					if(EditorBufferTaskDataWrite(buffer,(uint8_t *)taskData,taskDataLength))
					{
						return(TCL_OK);
					}
				}
				Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName taskData");
	}
	return(TCL_ERROR);
}

static int Cmd_EOFTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Send an EOF to the task on the passed buffer, if there is a problem, complain.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorBufferTaskEofWrite(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_KillTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Send a signal to the task on the passed buffer, if there is a problem, complain.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			if(ShellBufferNotBusy(localInterpreter,buffer))
			{
				ErrorTraceClear();
				if(EditorBufferTaskKill(buffer))
				{
					return(TCL_OK);
				}
				else
				{
					Tcl_AppendResult(localInterpreter,ErrorTraceGet(),NULL);
				}
			}
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_HasTask(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// See if a task is running in the given buffer, if so, return 1, else 0.
{
	EDITOR_BUFFER
		*buffer;

	if(objc==2)
	{
		if((buffer=TclEditorBufferGet(localInterpreter,objv[1])))
		{
			Tcl_AppendResult(localInterpreter,EditorBufferHasTask(buffer)?"1":"0",NULL);
			return(TCL_OK);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,"bufferName");
	}
	return(TCL_ERROR);
}

static int Cmd_Beep(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Make a noise.
{
	if(objc==1)
	{
		PlatformBeep();
		return(TCL_OK);
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
	}
	return(TCL_ERROR);
}

static int Cmd_Version(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Get version numbers.
{
	const char
		*localVersion;
	bool
		fail;

	fail=false;
	if(objc==1)
	{
		Tcl_AppendResult(localInterpreter,programName," ",VERSION,EDITION,NULL);
		if((localVersion=PlatformVersionGet()))
		{
			Tcl_AppendResult(localInterpreter," (",localVersion,")",NULL);
		}
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

static int Cmd_ShellCommand(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// This exists to provide a stub for the shell commands which e93 will call internally
// for various functions. Normally, when the initial script is executed, it will define
// this function, but if the script does not define it, or fails before defining it,
// this acts as a safety net.
{
	return(TCL_OK);
}

static int Cmd_ForceQuit(ClientData clientData,Tcl_Interp *localInterpreter,int objc,Tcl_Obj *const objv[])
// Execute editor quit function (the editor WILL quit, no questions asked if you call this).
{
	bool
		fail;

	fail=false;
	if(objc==1)
	{
		PlatformQuit();
	}
	else
	{
		Tcl_WrongNumArgs(localInterpreter,1,objv,NULL);
		fail=true;
	}
	return(fail?TCL_ERROR:TCL_OK);
}

// Create commands so TCL can find them

bool EditorShellCommandsCreate(Tcl_Interp *interpreter)
// Create the shell commands that are implemented by the editor.
// If there is a problem, ErrorSet, return false.
{
// menus and key bindings
	Tcl_CreateObjCommand(interpreter,"deletemenu",Cmd_DeleteMenu,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"addmenu",Cmd_AddMenu,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"bindkey",Cmd_BindKey,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"unbindkey",Cmd_UnBindKey,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"keybindings",Cmd_KeyBindings,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"waitkey",Cmd_WaitKey,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getkey",Cmd_GetKey,NULL,NULL);

// dialogs
	Tcl_CreateObjCommand(interpreter,"dialogcenter",Cmd_DialogCenter,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"okdialog",Cmd_OkDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"okcanceldialog",Cmd_OkCancelDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"yesnodialog",Cmd_YesNoDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"textdialog",Cmd_TextDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"searchdialog",Cmd_SearchDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"listdialog",Cmd_ListDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"opendialog",Cmd_OpenDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"savedialog",Cmd_SaveDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"pathdialog",Cmd_PathDialog,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"fontdialog",Cmd_FontDialog,NULL,NULL);

// buffers, windows, and views
	Tcl_CreateObjCommand(interpreter,"closebuffer",Cmd_CloseBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"openbuffer",Cmd_OpenBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"newbuffer",Cmd_NewBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"bufferlist",Cmd_BufferList,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"savebuffer",Cmd_SaveBuffer,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"savebufferas",Cmd_SaveBufferAs,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"savebufferto",Cmd_SaveBufferTo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"revertbuffer",Cmd_RevertBuffer,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"buffervariables",Cmd_BufferVariables,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setbuffervariable",Cmd_SetBufferVariable,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getbuffervariable",Cmd_GetBufferVariable,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"isdirty",Cmd_BufferDirty,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"fromfile",Cmd_FromFile,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"bufferexists",Cmd_BufferExists,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"haswindow",Cmd_HasWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"cleardirty",Cmd_ClearDirty,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"textinfo",Cmd_TextInfo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"positiontolineoffset",Cmd_PositionToLineOffset,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"lineoffsettoposition",Cmd_LineOffsetToPosition,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"checkbuffer",Cmd_CheckBuffer,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"closewindow",Cmd_CloseWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"openwindow",Cmd_OpenWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"windowlist",Cmd_WindowList,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"gettopleft",Cmd_GetTopLeft,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"settopleft",Cmd_SetTopLeft,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"homewindow",Cmd_HomeWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getrect",Cmd_GetRect,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setrect",Cmd_SetRect,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getutf8",Cmd_GetUtf8,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setutf8",Cmd_SetUtf8,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"gettabsize",Cmd_GetTabSize,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"settabsize",Cmd_SetTabSize,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"getfont",Cmd_GetFont,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setfont",Cmd_SetFont,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getcolors",Cmd_GetColors,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setcolors",Cmd_SetColors,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setcolorscheme",Cmd_SetColorScheme,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"settopwindow",Cmd_SetTopWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"minimizewindow",Cmd_MinimizeWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"unminimizewindow",Cmd_UnminimizeWindow,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"activewindow",Cmd_GetActiveWindow,NULL,NULL);

// fonts
	Tcl_CreateObjCommand(interpreter,"fontnamesplit",Cmd_FontNameSplit,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"fontnamejoin",Cmd_FontNameJoin,NULL,NULL);

// editing, cursor, and selection functions
	Tcl_CreateObjCommand(interpreter,"extract",Cmd_Extract,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"insert",Cmd_Insert,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"insertbinary",Cmd_InsertBinary,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"insertfile",Cmd_InsertFile,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"autoindent",Cmd_AutoIndent,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"movecursor",Cmd_MoveCursor,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"clear",Cmd_Clear,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"delete",Cmd_Delete,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"expandselection",Cmd_ExpandSelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"reduceselection",Cmd_ReduceSelection,NULL,NULL);

// selection/marks
	Tcl_CreateObjCommand(interpreter,"selectioninfo",Cmd_SelectionInfo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"selectall",Cmd_SelectAll,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"selectedtextlist",Cmd_SelectedTextList,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"selectline",Cmd_SelectLine,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getselectionends",Cmd_GetSelectionEnds,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setselectionends",Cmd_SetSelectionEnds,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getselectionendlist",Cmd_GetSelectionEndList,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"addselectionendlist",Cmd_AddSelectionEndList,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getselectionatposition",Cmd_GetSelectionAtPosition,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"copyselection",Cmd_CopySelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"andselection",Cmd_AndSelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"orselection",Cmd_OrSelection,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"xorselection",Cmd_XorSelection,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"markinvisible",Cmd_MarkInvisible,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"markvisible",Cmd_MarkVisible,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"marklist",Cmd_MarkList,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"closemark",Cmd_CloseMark,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"openmark",Cmd_OpenMark,NULL,NULL);

// styles
	Tcl_CreateObjCommand(interpreter,"getstyleatposition",Cmd_GetStyleAtPosition,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setstyle",Cmd_SetStyle,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"selectiontostyle",Cmd_SelectionToStyle,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"styletoselection",Cmd_StyleToSelection,NULL,NULL);

// color schemes
	Tcl_CreateObjCommand(interpreter,"addcolorscheme",Cmd_AddColorScheme,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"colorschemestyle",Cmd_ColorSchemeStyle,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"deletecolorscheme",Cmd_DeleteColorScheme,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"colorschemes",Cmd_ColorSchemes,NULL,NULL);

// syntax maps
	Tcl_CreateObjCommand(interpreter,"addsyntaxmap",Cmd_AddSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"deletesyntaxmap",Cmd_DeleteSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"syntaxmaps",Cmd_SyntaxMaps,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setsyntaxmap",Cmd_SetSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getsyntaxmap",Cmd_GetSyntaxMap,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"syntaxmapbuffers",Cmd_SyntaxMapBuffers,NULL,NULL);

// search/replace
	Tcl_CreateObjCommand(interpreter,"find",Cmd_Find,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"findall",Cmd_FindAll,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"replace",Cmd_Replace,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"replaceall",Cmd_ReplaceAll,NULL,NULL);

// undo
	Tcl_CreateObjCommand(interpreter,"undotoggle",Cmd_UndoToggle,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"undo",Cmd_Undo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"redo",Cmd_Redo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"breakundo",Cmd_BreakUndo,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"flushundos",Cmd_FlushUndos,NULL,NULL);

// clipboard
	Tcl_CreateObjCommand(interpreter,"cut",Cmd_Cut,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"copy",Cmd_Copy,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"paste",Cmd_Paste,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setclipboard",Cmd_SetClipboard,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getclipboard",Cmd_GetClipboard,NULL,NULL);

// misc functions
	Tcl_CreateObjCommand(interpreter,"screensize",Cmd_ScreenSize,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"setwordspacechars",Cmd_SetWordSpaceChars,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"getwordspacechars",Cmd_GetWordSpaceChars,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"execute",Cmd_Execute,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"task",Cmd_Task,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"eoftask",Cmd_EOFTask,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"killtask",Cmd_KillTask,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"hastask",Cmd_HasTask,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"beep",Cmd_Beep,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"version",Cmd_Version,NULL,NULL);

	Tcl_CreateObjCommand(interpreter,"ShellCommand",Cmd_ShellCommand,NULL,NULL);
	Tcl_CreateObjCommand(interpreter,"forceQUIT",Cmd_ForceQuit,NULL,NULL);

// Built-in Tcl commands we DONT want
	Tcl_DeleteCommand(interpreter,"exit");

	return(true);
}
